Package: qol
Title: Powerful 'SAS' Inspired Concepts for more Efficient Bigger
        Outputs
Version: 1.3.4
Authors@R: 
    person("Tim", "Siebenmorgen", email = "qol_package@proton.me", role = c("aut", "cre", "cph"))
Description: The main goal is to make descriptive evaluations easier to create bigger and more complex outputs in less time with less code. Introducing format containers with multilabels <https://documentation.sas.com/doc/en/pgmsascdc/v_067/proc/p06ciqes4eaqo6n0zyqtz9p21nfb.htm>, a more powerful summarise which is capable to output every possible combination of the provided grouping variables in one go <https://documentation.sas.com/doc/en/pgmsascdc/v_067/proc/p0jvbbqkt0gs2cn1lo4zndbqs1pe.htm>, tabulation functions which can create any table in different styles <https://documentation.sas.com/doc/en/pgmsascdc/v_067/proc/n1ql5xnu0k3kdtn11gwa5hc7u435.htm> and other more readable functions. The code is optimized to work fast even with datasets of over a million observations.
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
URL: https://github.com/s3rdia/qol, https://s3rdia.github.io/qol/,
        https://deepwiki.com/s3rdia/qol,
        https://s3rdia.github.io/qol_blog/
BugReports: https://github.com/s3rdia/qol/issues
Imports: data.table (>= 1.17.8), collapse (>= 2.1.2), openxlsx2 (>=
        1.29), fst (>= 0.9.8)
Depends: R (>= 4.1.0)
RoxygenNote: 7.3.3
Suggests: tinytest (>= 1.4.3), yyjsonr (>= 0.1.22), rstudioapi (>=
        0.19.0)
Config/tinytest/edition: 1
Config/Needs/website: rmarkdown
NeedsCompilation: no
Packaged: 2026-08-20 18:32:27 UTC; User
Author: Tim Siebenmorgen [aut, cre, cph]
Maintainer: Tim Siebenmorgen <qol_package@proton.me>
Repository: CRAN
Date/Publication: 2026-08-21 05:41:02 UTC
Built: R 4.6.0; ; 2026-08-21 07:31:18 UTC; unix

# Generated by roxygen2: do not edit by hand

export(add_extension)
export(add_variable_range)
export(any_table)
export(apply_macro)
export(args_to_char)
export(build_master)
export(build_rstheme)
export(check_required_package)
export(check_weight)
export(close_file)
export(code_statistics)
export(combine_into_workbook)
export(compute.)
export(concat)
export(content_report)
export(convert_factor)
export(convert_numeric)
export(convert_square_brackets)
export(create_table_of_contents)
export(crosstabs)
export(discrete_format)
export(do_if)
export(dots_to_char)
export(drop_type_vars)
export(dropp)
export(dummy_data)
export(else.)
export(else_do)
export(else_if.)
export(end_all_do)
export(end_do)
export(excel_output_style)
export(expand_formats)
export(export_data)
export(export_multi)
export(export_with_style)
export(first_row_as_names)
export(free_memory)
export(frequencies)
export(fuse_variables)
export(get_duplicate_var_count)
export(get_duplicate_var_names)
export(get_excel_range)
export(get_footnotes)
export(get_labels)
export(get_message_stack)
export(get_monitor)
export(get_na.rm)
export(get_origin_as_char)
export(get_output)
export(get_print)
export(get_print_miss)
export(get_style_options)
export(get_threads)
export(get_titles)
export(hex_to_256)
export(hex_to_ansi)
export(if.)
export(ifelse_multi)
export(import_data)
export(import_multi)
export(interval_format)
export(inverse)
export(keep)
export(libname)
export(load_file)
export(load_file_multi)
export(macro)
export(mark_case)
export(modify_number_formats)
export(modify_output_style)
export(monitor_end)
export(monitor_next)
export(monitor_plot)
export(monitor_start)
export(multi_join)
export(number_format_style)
export(part_of_df)
export(print_closing)
export(print_headline)
export(print_message)
export(print_stack_as_messages)
export(print_start_message)
export(print_step)
export(qol_chat)
export(qol_news)
export(recode.)
export(recode_multi)
export(remove_blanks)
export(remove_doubled_values)
export(remove_stat_extension)
export(rename_multi)
export(rename_pattern)
export(replace_except)
export(report_test_results)
export(reset_qol_options)
export(reset_style_options)
export(resolve_intersection)
export(retain_stat)
export(retain_value)
export(retain_variables)
export(round_multi)
export(round_values)
export(row_calculation)
export(run_folder)
export(run_project)
export(run_project_parallel)
export(run_script)
export(running_number)
export(save_file)
export(save_file_multi)
export(set_footnotes)
export(set_labels)
export(set_monitor)
export(set_na.rm)
export(set_no_color)
export(set_no_print)
export(set_output)
export(set_print)
export(set_print_miss)
export(set_style_options)
export(set_threads)
export(set_titles)
export(set_up_custom_message)
export(setcolorder_by_pattern)
export(sort_plus)
export(split_by)
export(stack_data)
export(sub_string)
export(summarise_plus)
export(test_package)
export(test_single_file)
export(transpose_plus)
export(vars_between)
export(where.)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# qol - Quality of Life <img src='man/figures/logo.png' width="150px" align="right" alt='Logo'/>

<!-- badges: start -->

[![SAS](https://img.shields.io/badge/SAS-R-orange.svg)](https://github.com/s3rdia/qol)
[![CRAN
Version](https://www.r-pkg.org/badges/version/qol?color=green)](https://cran.r-project.org/package=qol)
[![r-universe](https://s3rdia.r-universe.dev/badges/qol)](https://s3rdia.r-universe.dev/qol)
[![DEVELOPMENT
Version](https://img.shields.io/badge/GitHub-1.3.4-blue.svg)](https://github.com/s3rdia/qol)
[![Ask
DeepWiki](https://deepwiki.com/badge.svg)](https://deepwiki.com/s3rdia/qol)
[![qol_blog](https://img.shields.io/badge/qol-Blog-036f30)](https://s3rdia.github.io/qol_blog/)
[![status](https://tinyverse.netlify.app/badge/qol)](https://CRAN.R-project.org/package=qol)
<!-- badges: end -->

Bringing Powerful ‘SAS’ Inspired Concepts for more Efficient Bigger
Outputs to ‘R’.

The main goal is to make descriptive evaluations easier to create bigger
and more complex outputs in less time with less code. Introducing format
containers with multilabels, a more powerful summarise which is capable
to output every possible combination of the provided grouping variables
in one go, tabulation functions which can create any table in different
styles and other more readable functions. The code is optimized to work
fast even with datasets of over a million observations.

## Installation

``` r
# Official release
install.packages("qol")

# Development version
devtools::install_github("s3rdia/qol")

install.packages('qol', repos = c('https://s3rdia.r-universe.dev', 'https://cloud.r-project.org'))

pak::pak("s3rdia/qol")
```

## Format Containers

Create a format container independent from any data frame. Define which
values should be recoded into which new categories, if the format is
applied to a variable in a data frame. It is possible to assign a single
value to multiple new categories to create a multilabel. With these
format containers, you just keep a small reference of original values
and result categories. Formats and data find their way together only
just before computing the results. This method is very memory efficient,
readable and user friendly for creating larger and more complex outputs
at the same time.

``` r
# Creating format containers
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)
```

## Massive Outputs: Simple and fast

The package builds on the incredibly fast
[collapse](https://github.com/fastverse/collapse) and
[data.table](https://github.com/Rdatatable/data.table) packages. In
addition the code is optimized to handle big datasets efficiently with
the format concept.

``` r
library(qol)

# If you want to test the raw speed in combination with creating big outputs try this:
# Lets crank up the observations to 10 Millions
my_data <- dummy_data(10000000)

# Create format containers
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

education. <- discrete_format(
    "Total"            = c("low", "middle", "high"),
    "low education"    = "low",
    "middle education" = "middle",
    "high education"   = "high")
    
# And now let's take a second and see what massive outputs we can get in no time
summary_df <- my_data |>
    summarise_plus(class      = c(year, sex, age, education),
                   values     = c(income, probability, weight),
                   statistics = c("freq", "sum", "sum_wgt", "pct_group", "pct_total", "missing"),
                   formats    = list(age = age.,
                                     sex = sex.,
                                     education = education.),
                   weight     = "weight",
                   nesting    = "all",
                   notes      = FALSE)
```

<img src='man/figures/output.png' alt="Example output data frame"/>

The operations based on summarisation are the fastest. Other operations
take a bit longer but still work fast with big datasets.

## Powerful tabulation

Using the wonderful [openxlsx2](https://github.com/JanMarvin/openxlsx2)
package for maximum style, you can basically output any table fully
styled with little effort. Combine any number of variables in any
possible way, all at once. Setting up a custom, reusable style is as
easy as setting up options like: provide a color for the table header,
set the font size for the row header, should borders be drawn for the
table cells yes/no, and so on. You can fully concentrate on designing a
table, instead of thinking hard about how to calculate where to put a
border or to even manually prepare a designed workbook.

``` r
my_data <- dummy_data(100000)

# Create format containers
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

education. <- discrete_format(
    "Total"            = c("low", "middle", "high"),
    "low education"    = "low",
    "middle education" = "middle",
    "high education"   = "high")
    
# Define style
set_style_options(column_widths = c(2, 15, 15, 15, 9))

# Define titles and footnotes. If you want to add hyperlinks you can do so by
# adding "link:" followed by the hyperlink to the main text.
set_titles("This is title number 1 link: https://cran.r-project.org/",
           "This is title number 2",
           "This is title number 3")

set_footnotes("This is footnote number 1",
              "This is footnote number 2",
              "This is footnote number 3 link: https://cran.r-project.org/")

# Output complex tables with different percentages
my_data |> any_table(rows       = c("sex + age", "sex", "age"),
                     columns    = c("year", "education + year"),
                     values     = weight,
                     statistics = c("sum", "pct_group"),
                     pct_group  = c("sex", "age"),
                     formats    = list(sex = sex., age = age.,
                                       education = education.),
                     na.rm      = TRUE)

reset_style_options()
reset_qol_options()
```

<img src='man/figures/tabulation.png' alt="Example output table"/>

You can also combine multiple tables in one workbook, each on a
different sheet, and save the file, instead of just viewing it.

``` r
# Set output and style options globally
set_style_options(save_path  = "C:/My_folder/",
                  file       = "My_workbook.xlsx",
                  sheet_name = "age_sex")

# Capture the output of any_table() to get a list which contains the data frame
# as well as the formatted workbook.
# Note: Set print to FALSE to prevent the workbook from being saved to early.
result_list <- my_data |>
           any_table(rows       = c("age"),
                     columns    = c("sex"),
                     values     = weight,
                     statistics = c("sum"),
                     formats    = list(sex = sex., age = age.),
                     na.rm      = TRUE,
                     print      = FALSE)

# Set a new sheet name for the next table
set_style_options(sheet_name = "edu_year")

# Pass on the workbook from before to the next any_table()
# print is now TRUE so the workbook will be saved.
my_data |> any_table(workbook   = result_list[["workbook"]],
                     rows       = c("education"),
                     columns    = c("year"),
                     values     = weight,
                     statistics = c("pct_group"),
                     formats    = list(education = education.),
                     na.rm      = TRUE)
                     
# For safety set file back to NULL at the end. Otherwise the next any_table() will overwrite
# the existing file.
close_file()
```

In case you have a good amount of tables, you want to combine in a
single workbook, you can also catch the outputs and combine them
afterwards in one go:

``` r
# Set output and style options globally
set_style_options(sheet_name = "age_sex")

# Catch the output as shown before, but additionally use the option -> output = "excel_nostyle".
# This skips the styling part, so that the function runs faster. The styling is done later on.
tab1 <- my_data |>
       any_table(rows       = c("age"),
                 columns    = c("sex"),
                 values     = weight,
                 statistics = c("sum"),
                 formats    = list(sex = sex., age = age.),
                 na.rm      = TRUE,
                 print      = FALSE,
                 output     = "excel_nostyle")

# Now let's asume you create a bunch of different tables
set_style_options(sheet_name = "sheet2")
tab2 <- my_data  |> any_table(..., print = FALSE, output = "excel_nostyle")

set_style_options(sheet_name = "sheet3")
tab3 <- my_data  |> any_table(..., print = FALSE, output = "excel_nostyle")

set_style_options(sheet_name = "sheet4")
tab4 <- my_data  |> any_table(..., print = FALSE, output = "excel_nostyle")

...

# Every of the above tabs is a list, which contains the data table, an unstyled workbook and the meta
# information needed for the individual styling. These tabs can be input into the following function,
# which reads the meta information, styles each table individually and combines them as separate sheets
# into a single workbook.
combine_into_workbook(tab1, tab2, tab3, tab4, tab5, tab6, tab7, tab8,
                      file = "C:/My_folder/My_workbook.xlsx")
```

## Save with protection and load as you like

The package offers file saving in the standard rds format or the
incredibly fast [fst](https://github.com/fstpackage/fst) format. By
default the saving function write protects any existing files. So
nothing gets overwritten unless you say so.  
The loading functions are able to keep only the desired variables, but
with a twist: You can write the variable names as you like. They are
always found, whether you use small letters or big ones or something in
between. And they are retrieved in provided order.

``` r
# Save file
my_data |> save_file(path    = tempdir(),
                     file    = "testfile.fst",
                     keep    = c(income_class, sex, age, state, weight, NUTS3),
                     where   = sex == 1)
                     
# Load file
my_fst_where <- load_file(path  = tempdir(),
                          file  = "testfile.fst",
                          keep  = c(AGE, INCOME_class, State, weight),
                          where = age > 65)
```

## One function to join them all

Join two or more data frames together in one operation with multiple
different join methods and join on differently named variables.
Available methods are: left, right, inner, full, outer, left_inner,
right_inner.

``` r
# Simple join
df1 <- data.frame(key = c(1, 1, 1, 2, 2, 2),
                  a   = c("a", "a", "a", "a", "a", "a"))

df2 <- data.frame(key = c(2, 3),
                  b   = c("b", "b"))

left_joined <- multi_join(list(df1, df2), on = "key")

# Joining more than two data frames on different variable names with different methods
df1b <- data.frame(key1 = c(1, 1, 1, 2, 2, 2),
                   key2 = c("a", "a", "a", "a", "a", "a"),
                   a    = c("a", "a", "a", "a", "a", "a"))

df2b <- data.frame(var1 = c(1, 2),
                   var2 = c("a", "a"),
                   c    = c("c", "c"))

df3b <- data.frame(any  = c(2, 3),
                   name = c("a", "a"),
                   b    = c("b", "b"))

multiple_joined3 <- multi_join(list(df1b, df2b, df3b),
                               on = list(df1b = c("key1", "key2"),
                                         df2b = c("var1", "var2"),
                                         df3b = c("any", "name")),
                               how = c("left", "right"))
```

## Transpose like never before

Transpose, weight results and generate additional categories all in one
operation. Put variables side by side or nest them or both, just as you
need it. And of course you can transpose multiple variables from wide
into long format in one go.

``` r
# Transpose from long to wide and use a multilabel to generate additional categories
long_to_wide <- my_data |>
    transpose_plus(preserve = c(year, age),
                   pivot    = c("sex + education", "sex", "education"),
                   values   = income,
                   formats  = list(sex = sex., age = age.),
                   weight   = weight,
                   na.rm    = TRUE)

# Transpose back from wide to long
wide_to_long <- long_to_wide |>
    transpose_plus(preserve = c(year, age),
                   pivot    = list(sex       = c("Total", "Male", "Female"),
                                   education = c("low", "middle", "high")))
```

## Sorting with additions

Sort cases while preserving the order of certain variables or make use
of formats to sort in format order, which can be used to e.g. sort a
character variable in another than alphabetical order without creating a
temporary variable just for sorting.

``` r
education. <- discrete_format(
    "1" = "low",
    "2" = "middle",
    "3" = "high")

# Simple sorting
sort_df1 <- my_data |> sort_plus(by = c(state, sex, age))
sort_df2 <- my_data |> sort_plus(by    = c(state, sex, age),
                                 order = c("ascending", "descending"))

# Character variables will normally be sorted alphabetically. With the help
# of a format this variable can be sorted in a completely different way.
sort_df3 <- my_data |> sort_plus(by      = education,
                                 formats = list(education = education.))

# Preserve the order of the character variable, otherwise it couldn't stay in
# it's current order.
sort_df4 <- sort_df3 |> sort_plus(by       = age,
                                  preserve = education)
```

## Readability

There are also some functions which enhance the readability of the code.
For example if - else if - else statements like in other languages:

``` r
# Example data frame
my_data <- dummy_data(1000)

# Call function
new_df <- my_data |>
         if.(age < 18,             age_group = "under 18") |>
    else_if.(age >= 18 & age < 65, age_group = "18 to under 65") |>
    else.   (                      age_group = "65 and older")

# Or with multiple variables
new_df <- my_data |>
         if.(age < 18,             age_group = "under 18"      , age_num = 1L) |>
    else_if.(age >= 18 & age < 65, age_group = "18 to under 65", age_num = 2L) |>
    else.   (                      age_group = "65 and older",   age_num = 3L)

# NOTE: As in other languages the following if blocks won't produce the same result.
#       if.() will overwrite existing values while else_if.() will not.
state_df <- my_data |>
         if.(state == 1, state_a = "State 1") |>
    else_if.(state < 11, state_a = "West") |>
    else.   (            state_a = "East")

state_df <- state_df |>
      if.(state == 1, state_b = "State 1") |>
      if.(state < 11, state_b = "West") |>
    else.(            state_b = "East")
    
# Use if.() as a do-over-loop. In this kind of loop all vectors will be
# advanced one iteration at a time in parallel.
money    <- c("income", "expenses", "balance", "probability")
new_vars <- c("var1", "var2", "var3", "var4")
result   <- c(1, 2, 3, 4)

do_over_df <- my_data |>
      if.(money > 0, new_vars = result) |>
    else.(           new_vars = 0)
    
# All these functions can be used in a do_if() situation and are aware of
# overarching conditions.
do_if_df <- my_data |>
    do_if(state < 11) |>
          if.(age < 18, age_cat = 1) |>
        else.(          age_cat = 2) |>
    else_do() |>
          if.(age < 18, age_cat = 3) |>
        else.(          age_cat = 4) |>
    end_do()
```

The if condition functions also enable a SAS like writing style:

``` r
# Example data frame
my_data <- dummy_data(100)

# Write in intervals
my_data[["age_gr"]] <- my_data |> ifelse_multi(" age < 18 "       = "under 18",
                                               " 18 <= age < 25 " = "18 to under 25",
                                               " 25 <= age < 50 " = "25 to under 50",
                                               " 50 <= age < 65 " = "50 to under 65",
                                               " 65 <= age      " = "65 and more")

# With overarching do_if condition
my_data[["age_gr_edu"]] <- my_data |> ifelse_multi(do_if = " education in ('middle' 'high') ",
                                                       " age < 18 "       = "under 18",
                                                       " 18 <= age < 25 " = "18 to under 25",
                                                       " 25 <= age < 50 " = "25 to under 50",
                                                       " 50 <= age < 65 " = "50 to under 65",
                                                       " 65 <= age      " = "65 and more")

# And/or translation
my_data[["and"]] <- my_data |> ifelse_multi(" age > 65 and sex = 1 " = 1,
                                            " age > 65 and sex = 2 " = 2,
                                            else. = 0)

my_data[["or"]] <- my_data |> ifelse_multi(" age > 65 or sex = 1 " = 1,
                                           " age > 65 or sex = 2 " = 2,
                                           else. = 0)

# "in" translation
my_data[["in"]] <- my_data |> ifelse_multi(" age in (1 10 25 65 90) " = 1, else. = 0)

# Colon translation: start/ends with and contains
my_data[["start"]]    <- my_data |> ifelse_multi(" education == 'lo:' "  = 1, else. = 0)
my_data[["end"]]      <- my_data |> ifelse_multi(" education == ':le' "  = 1, else. = 0)
my_data[["contains"]] <- my_data |> ifelse_multi(" education == ':ig:' " = 1, else. = 0)

# Macro variables can be integrated in any place
variable     <- "age"
age_to_check <- 18
value_to_set <- "under 18"

my_data[["macro"]] <- my_data |>
ifelse_multi(" &variable < &age_to_check " = "&value_to_set", else. = "other")

# NA translation
my_data[["NA"]]    <- my_data |> ifelse_multi(" age == .       " = 1, else. = 0)
my_data[["notNA"]] <- my_data |> ifelse_multi(" education != . " = 1, else. = 0)

# Pass in existing variable values
my_data[["income_mix"]] <- my_data |> ifelse_multi(" age < 50 " = income, else. = expenses)
```

## Styled messages

The functions in this package actually talk to the user and show what
they are doing during runtime. The message system relies on pure base R
and does not only provide built in message types, but also custom ones.

The standalone, lightweight and zero dependency version of this message
system is also available with the package
[printify](https://github.com/s3rdia/printify) and can be downloaded via
[CRAN](https://CRAN.R-project.org/package=printify).

<img src='man/figures/messages.png' alt='Messages'/>

## Monitoring

This package also includes some basic yet very effective performance
monitoring functions. The heavier functions in this package already make
use of them and can show how they work internally like this:

<img src='man/figures/monitor.png' alt='Monitoring example'/>

## Customizing visual appearance

Don’t like the colors of the built-in in RStudio themes? Can’t find a
theme online that fits your liking? No Problem. With a simple function
you can create a full theme file and decide which parts of the editor
receive which colors. The themes shown below and the corresponding code
can be found on GitHub.

<img src='man/figures/theme.png' alt='Theme examples'/>

#' Create Format Container
#'
#' @name formats
#'
#' @description
#' Create a format container which stores discrete or interval values
#' with corresponding labels that can be applied by using functions capable of
#' using formats.
#'
#' Create a format container independent from any data frame. Define which values
#' should be recoded into which new categories, if the format is applied to
#' a variable in a data frame.
#' It is possible to assign a single value to multiple new categories to create
#' a multilabel.
#' It is recommended to let format names end with a dot to make them stand out.
#'
#' @param ... List all the desired recodings/recoding ranges. Every element containsa text for
#' the new category name and the values/value ranges which should be recoded into this new category.
#' @param include_lower TRUE by default. Whether to include lower bound value in
#' interval formats.
#' @param include_upper FALSE by default. Whether to include upper bound value in
#' interval formats.
#'
#' @details
#' The concept of having formats as molds or stencils to put the data through, is inspired by
#' 'SAS' formats. In 'SAS' formats are defined with the procedure Proc Formats, which is adapted
#' with [discrete_format()] and [interval_format()]. Here you can define, which values
#' should be transferred into which result categories. This is completely detached from the data
#' your working with.
#'
#' The great thing about this is, that one can not only label and recode values, but one can also
#' define so called multilabels. Meaning, one original value can be transferred into multiple
#' result categories.
#'
#' A cell in a data frame can only hold one distinct value, which is normally a good thing.
#' But let's say you want to convert single ages into age categories. The age "3" for example
#' could go into the category "under 6", but also in "under 12", "under 18" and "total". Normally
#' you would compute additional variables, which hold the different categorizations, or you could also
#' double up the observations for each category. Both ways would just bloat up the data frame and
#' cost additional memory, particularly if you work with big data sets.
#'
#' With these format containers, you just keep a small reference of original values and result
#' categories. Formats and data find their way together only just before computing the results,
#' meaning the original data frame can be passed into a function capable of handling formats (see below),
#' without any data transformation beforehand. You just tell the function which format should
#' be applied to which variable. That's it. The function handles the rest and outputs all the
#' desired categories.
#'
#' This method is very memory efficient, readable and user friendly for creating larger and more
#' complex outputs at the same time.
#'
#' @return
#' Returns a data table which contains the values/value ranges with the corresponding labels
#'
#' @seealso
#' Functions that can handle formats: [summarise_plus()], [frequencies()], [crosstabs()],
#' [any_table()], [recode.()], [recode_multi()], [transpose_plus()], [sort_plus()].
#'
#' @examples
#' age. <- discrete_format(
#'     "Total"          = 0:100,
#'     "under 18"       = 0:17,
#'     "18 to under 25" = 18:24,
#'     "25 to under 55" = 25:54,
#'     "55 to under 65" = 55:64,
#'     "65 and older"   = 65:100)
#'
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' education. <- discrete_format(
#'     "Total"            = c("low", "middle", "high"),
#'     "low education"    = "low",
#'     "middle education" = "middle",
#'     "high education"   = "high")
#'
#' income. <- interval_format(
#'     "Total"              =    0:100000,
#'     "below 500"          =    0:500,
#'     "500 to under 1000"  =  500:1000,
#'     "1000 to under 2000" = 1000:2000,
#'     "2000 and more"      = 2000:100000)
#'
#' state. <- discrete_format(
#'     "Germany"                       = 1:16,
#'     "Schleswig-Holstein"            = 1,
#'     "Hamburg"                       = 2,
#'     "Lower Saxony"                  = 3,
#'     "Bremen"                        = 4,
#'     "North Rhine-Westphalia"        = 5,
#'     "Hesse"                         = 6,
#'     "Rhineland-Palatinate"          = 7,
#'     "Baden-Württemberg"             = 8,
#'     "Bavaria"                       = 9,
#'     "Saarland"                      = 10,
#'     "West"                          = 1:10,
#'     "Berlin"                        = 11,
#'     "Brandenburg"                   = 12,
#'     "Mecklenburg-Western Pomerania" = 13,
#'     "Saxony"                        = 14,
#'     "Saxony-Anhalt"                 = 15,
#'     "Thuringia"                     = 16,
#'     "East"                          = 11:16)
#'
#' # With discrete formats you can specify the keyword "other" to
#' # catch any other value not covered by the explicitly specified values.
#' age. <- discrete_format(
#'     "under 18"       = 0:17,
#'     "18 to under 25" = 18:24,
#'     "25 to under 55" = "other")
#'
#' # With interval formats you can also use the keywords "low" and "high" to
#' # catch everything from the lowest to the highest values, in case one doesn't
#' # know exactly what the lowest and highest values are.
#' income. <- interval_format(
#'     "Total"              = c("low", "high"),
#'     "below 500"          = c("low", 500),
#'     "500 to under 1000"  = 500:1000,
#'     "1000 to under 2000" = 1000:2000,
#'     "2000 and more"      = c(2000, "high"))
#'
#' @rdname formats
#'
#' @keywords internal
NULL


#' @rdname formats
#'
#' @export
discrete_format <- function(...){
    # Measure the time
    print_start_message(suppress = TRUE)

    # Translate ... into separately controllable arguments
    list_of_groupings <- tryCatch({
        # Force evaluation to see if it exists
        list(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(list_of_groupings)){
        print_message("ERROR", c("Formats must be provided in the form [target category] = [original expressions].",
								 "Creating format will be aborted."))
        return(invisible(NULL))
    }

    # Check if any element has a missing name. If that is the case the list wasn't provided
    # in the correct way.
    labels <- names(list_of_groupings)

    if (is.null(labels) || any(is.na(labels) | labels == "")){
        print_message("ERROR", c("Formats must be provided in the form [target category] = [original expressions].",
								 "Creating format will be aborted."))
        return(invisible(NULL))
    }

    unwrap_grouping <- function(single_label){
        data.table::data.table(value = list_of_groupings[[single_label]],
                   label = single_label,
                   stringsAsFactors = FALSE)
    }

    unwrap_all_groupings <- lapply(names(list_of_groupings), unwrap_grouping)

    # Flatten to long format
    unwrapped_format <- data.table::rbindlist(unwrap_all_groupings)

    # If label column is all numeric then convert it to numeric
    unwrapped_format <- unwrapped_format |> convert_numeric("label")

    # Convert "other" keyword to integer max. This should be a value no one would pick normally.
    if ("other" %in% tolower(unwrapped_format[["value"]])){
        unwrapped_format[["value"]] <- sub("other", .Machine[["integer.max"]], tolower(unwrapped_format[["value"]]))

        # If value column is all numeric then convert it to numeric
        unwrapped_format <- unwrapped_format |> convert_numeric("value")
    }

    print_closing()

    unwrapped_format
}


#' @rdname formats
#'
#' @export
interval_format <- function(...,
                            include_lower = TRUE,
                            include_upper = FALSE){
    # Measure the time
    print_start_message(suppress = TRUE)

    # Translate ... into separately controllable arguments
    ranges <- tryCatch({
        # Force evaluation to see if it exists
        list(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(ranges)){
        print_message("ERROR", c("Formats must be provided in the form [target category] = [original expressions].",
								 "Creating format will be aborted."))
        return(invisible(NULL))
    }

    # Check if any element has a missing name. If that is the case the list wasn't provided
    # in the correct way.
    labels <- names(ranges)

    if (is.null(labels) || any(is.na(labels) | labels == "")){
        print_message("ERROR", c("Formats must be provided in the form [target category] = [original expressions].",
								 "Creating format will be aborted."))
        return(invisible(NULL))
    }

    # Get from - to value as vectors
    from <- sapply(ranges, function(x) min(x, na.rm = TRUE))
    to   <- sapply(ranges, function(x) max(x, na.rm = TRUE))

    # Insert pseudo low and high numbers for keywords
    if (is.character(to)){
        # First check if there are any other words than low and high in the format. If yes, abort.
        if (!any(c("low", "high") %in% tolower(to))){
            print_message("ERROR", "Unknown keyword found. Creating interval format will be aborted.")
            return(invisible(NULL))
        }

        # Low always ends up in "to", because it is a character value and comes alphabetically after high
        if ("low" %in% tolower(to)){
            # Swap the real not "low" value to "to" and insert a pseudo low number in "from".
            # Capture the original "low" positions first, to just replace these and
            # preserve single value entries.
            low_entries              <- to == "low"
            to[tolower(to) == "low"] <- from[tolower(to) == "low"]
            from[low_entries & tolower(from) == tolower(to)] <- -.Machine[["double.xmax"]]

            from <- as.numeric(from)
        }

        # If there is the "high" keyword, at this point it will always be in "to" either the same way as low,
        # just because it is character or because it is swapped here by the low code above.
        if ("high" %in% tolower(to)){
            # Enter a pseudo high number for "to"
            to[tolower(to) == "high"] <- .Machine[["double.xmax"]]

            from <- as.numeric(from)
        }

        to <- as.numeric(to)
    }

    # Get positions where "from" and "to" are equal. This is necessary to capture
    # occasions were a single value and not a range should be included in the format.
    single_value <- from == to

    # Add the smallest piece to "from" in case lower bound should not be included
    if (!include_lower){
        from[!single_value] <- from[!single_value] + .Machine[["double.eps"]] * 1000
    }

    # Subtract the smallest piece from "to" in case upper bound should not be included
    if (!include_upper){
        to[!single_value] <- to[!single_value] - .Machine[["double.eps"]] * 1000
    }

    print_closing()

    # Put everything together in a data frame
    data.table::data.table(from  = from,
                           to    = to,
                           label = labels) |> convert_numeric("label")
}


#' Evaluate Formats
#'
#' @description
#' Get the list of formats with their corresponding data frames.
#'
#' @param formats_list A list containing format names.
#'
#' @return
#' Returns a list of format data frames.
#'
#' @noRd
evaluate_formats <- function(formats_list){
    if (length(formats_list) == 0){
        return(c())
    }

    # Fetch all the provided formats and get the original data frames and store them into a list
    formats <- lapply(formats_list, function(format){
        if (length(format) > 1){
            print_message("WARNING", c("Formats not passed correctly. Create format object first and then",
									   "pass it to the function parameter like: list(my_variable = my_format)"),
									   always_print = TRUE)
            return(c())
        }
        dynGet(as.character(format), ifnotfound = NULL, inherits = TRUE)
    })

    names(formats) <- names(formats_list)

    formats
}


#' Check If Format List Is Provided
#'
#' @description
#' Check if the provided list is a list of data frames.
#'
#' @param formats_list A list containing format data frames.
#'
#' @return
#' Returns TRUE or FALSE.
#'
#' @noRd
is_list_of_dfs <- function(formats_list){
    # This errors if formats_list can't be evaluated because of a missing object.
    # In this case return FALSE so that the function using this check can evaluate to NULL.
    tryCatch({
        (is.list(formats_list)
         && length(formats_list) > 0
         && all(vapply(formats_list, data.table::is.data.table, logical(1))))
    }, error = function(e){
        FALSE
    })

}


#' Expand A List Of Formats To Generate All Possible Combinations
#'
#' @description
#' Generates a data frame which contains all nested combinations of the provided
#' format labels.
#'
#' @param ... A list containing format data frames.
#' @param names Custom variable names for the newly generated data frame.
#'
#' @return
#' Returns a data frames with all format label combinations.
#'
#' @examples
#' # Example formats
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' age. <- discrete_format(
#'     "Total"          = 0:100,
#'     "under 18"       = 0:17,
#'     "18 to under 25" = 18:24,
#'     "25 to under 55" = 25:54,
#'     "55 to under 65" = 55:64,
#'     "65 and older"   = 65:100)
#'
#' # Expand formats
#' expand_df <- expand_formats(sex., age.,
#'                             names = c("sex", "age"))
#'
#' @export
expand_formats <- function(..., names = NULL){
    # Measure the time
    print_start_message()

    # Translate ... into a list if possible
    format_list <- tryCatch({
        # Force evaluation to see if it exists
        formats      <- list(...)
        format_names <- names(formats)

        # If ... is passed as a list instead of single format data frames, it has
        # to be unwrapped first. Otherwise there would be a list inside a list.
        if (length(formats) == 1 && is.list(formats[[1]])){
            if (is.null(format_names) || identical(format_names, "")){
                formats <- formats[[1]]
            }

            # If there is only one format data frame provided there is nothing to expand,
            # so the function just returns the unique values of the label column.
            if (data.table::is.data.table(formats)){
                print_closing()
                return(formats |>
                           keep("label") |>
                           collapse::funique())
            }
        }

        formats
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(format_list)){
        print_message("ERROR", c("Formats must be provided as data frames or as a list of data frames.",
								 "Format expansion will be aborted."))
        return(invisible(NULL))
    }

    # Create a list of format data frames which only contain the "label" column
    format_list <- lapply(format_list, function(format){
            if ("label" %in% names(format)){
                format[["label"]]
            }
            else{
                NULL
            }
        })

    # Abort if there is a missing "label" column in one of the data frames. Normally I could drop this
    # and go on, but if used in other functions this could lead to follow up errors. So abort to be safe.
    if (length(format_list[sapply(format_list, is.null)]) > 0){
        print_message("ERROR", c("A data frame is missing the 'label' column. This function is especially for expanding formats created",
								 "with 'discrete_format' and 'interval_format'. If you want to include another non format data frame, you have",
								 "to name the desired expansion column 'label'. Format expansion will be aborted."))

        return(invisible(NULL))
    }

    # Calculate the final number of elements beforehand, to give a hint whether
    # the output becomes very large.
    final_number_of_rows <- prod(lengths(format_list))

    if (final_number_of_rows > 1000000){
        final_number_of_rows <- format(final_number_of_rows, big.mark = ".", decimal.mark = ",")

        print_message("NOTE", c("expand_formats() is about to calculate a cartesian product which produces [number] rows.",
                                "If you didn't call the function directly you should consider setting the print_miss parameter to FALSE.",
                                "Otherwise this process might take a long time to finish and likely produce lots of empty cells."),
                      number = final_number_of_rows, always_print = TRUE)
    }

    # Generate the cartesian product of all given labels to get all possible
    # combinations of categories.
    expand_df <- do.call(data.table::CJ, c(format_list,
                                           list(unique = TRUE, sorted = FALSE)))

    # Name the columns after the given names
    if (!is.null(names)){
        names(expand_df) <- names
    }

    # Convert to factor so that they are sorted in format order
    var_names <- names(expand_df)
    expand_df <- expand_df |> convert_factor(var_names)

    print_closing()

    expand_df
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

# Create single discrete label
sex. <- discrete_format(
        "Male"   = 1,
        "Female" = 2)

expect_true(all(c("value", "label") %in% names(sex.)), info = "Create single discrete label")
expect_equal(nrow(sex.), 2, info = "Create single discrete label")
expect_equal(ncol(sex.), 2, info = "Create single discrete label")


# Create discrete multilabel
sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

expect_true(all(c("value", "label") %in% names(sex.)), info = "Create discrete multilabel")
expect_equal(nrow(sex.), 4, info = "Create discrete multilabel")
expect_equal(ncol(sex.), 2, info = "Create discrete multilabel")


# Create single interval label
income. <- interval_format(
    "below 500"          =    0:500,
    "500 to under 1000"  =  500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

expect_true(all(c("from", "to", "label") %in% names(income.)), info = "Create single interval label")
expect_equal(nrow(income.), 4, info = "Create single interval label")
expect_equal(ncol(income.), 3, info = "Create single interval label")


# Create interval multilabel
income. <- interval_format(
    "Total"              =    0:100000,
    "below 500"          =    0:500,
    "500 to under 1000"  =  500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

expect_true(all(c("from", "to", "label") %in% names(income.)), info = "Create interval multilabel")
expect_equal(nrow(income.), 5, info = "Create interval multilabel")
expect_equal(ncol(income.), 3, info = "Create interval multilabel")


# Create interval format with low and high keywords
income. <- interval_format(
    "Total"              = c("low", "high"),
    "below 500"          = c("low", 500),
    "500 to under 1000"  = 500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = c(2000, "high"))

expect_true(all(c("from", "to", "label") %in% names(income.)), info = "Create interval format with low and high keywords")
expect_equal(income.[["from"]][1], -Inf, info = "Create interval format with low and high keywords")
expect_equal(income.[["to"]][1],    Inf, info = "Create interval format with low and high keywords")


# Expand formats
sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

income. <- interval_format(
    "below 500"          =    0:500,
    "500 to under 1000"  =  500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

expand_df <- expand_formats(sex., income.,
                            names = c("sex", "income"))

expect_true(all(c("sex", "income") %in% names(expand_df)), info = "Expand formats")
expect_equal(collapse::fncol(expand_df), 2, info = "Expand formats")
expect_equal(collapse::fnrow(expand_df), 12, info = "Expand formats")


# Expand formats provided as list
sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

income. <- interval_format(
    "below 500"          =    0:500,
    "500 to under 1000"  =  500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

expand_df <- expand_formats(list(sex., income.), names = c("sex", "income"))

expect_true(all(c("sex", "income") %in% names(expand_df)), info = "Expand formats provided as list")
expect_equal(collapse::fncol(expand_df), 2, info = "Expand formats provided as list")
expect_equal(collapse::fnrow(expand_df), 12, info = "Expand formats provided as list")


# Expand formats returns unique label values if only one format is provided
sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

expand_df <- expand_formats(sex.)

expect_equal(expand_df[[1]], c("Total", "Male", "Female"), info = "Expand formats returns unique label values if only one format is provided")


# Interval formats including lower bounds and excluding upper bounds
age. <- interval_format(
    "under 65"    = c(0,  65),
    "65 und mehr" = c(65, 100))

expect_true(age.[1, "from"] ==  0 && age.[1, "to"] <  65, info = "Interval formats including lower bounds and excluding upper bounds")
expect_true(age.[2, "from"] == 65 && age.[2, "to"] < 100, info = "Interval formats including lower bounds and excluding upper bounds")


# Interval formats including upper bounds and excluding lower bounds
age. <- interval_format(
    "under 65"    = c(0,  65),
    "65 und mehr" = c(65, 100),
    include_lower = FALSE,
    include_upper = TRUE)

expect_true(age.[1, "from"] >  0 && age.[1, "to"] ==  65, info = "Interval formats including upper bounds and excluding lower bounds")
expect_true(age.[2, "from"] > 65 && age.[2, "to"] == 100, info = "Interval formats including upper bounds and excluding lower bounds")

###############################################################################
# Abort checks
###############################################################################

# Abort format expansion, if a data frame has no label column
sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

sex. <- sex. |> dropp("label")

income. <- interval_format(
    "below 500"          =    0:500,
    "500 to under 1000"  =  500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

expand_df <- expand_formats(sex., income., names = c("sex", "income"))

expect_error(print_stack_as_messages("ERROR"), "A data frame is missing the 'label' column. This function is especially for expanding formats created", info = "Abort format expansion, if a data frame has no label column")


# Create single discrete label aborts, if elements not provided in the correct way
discrete_format(test == 1)

expect_error(print_stack_as_messages("ERROR"), "Formats must be provided in the form.", info = "Create single discrete label aborts, if elements not provided in the correct way")


# Create single discrete label aborts, if list element is missing a name
discrete_format(1)

expect_error(print_stack_as_messages("ERROR"), "Formats must be provided in the form.", info = "Create single discrete label aborts, if list element is missing a name")


# Create single interval label aborts, if elements not provided in the correct way
interval_format(test == 1)

expect_error(print_stack_as_messages("ERROR"), "Formats must be provided in the form.", info = "Create single interval label aborts, if elements not provided in the correct way")


# Create single interval label aborts, if list element is missing a name
discrete_format(1)

expect_error(print_stack_as_messages("ERROR"), "Formats must be provided in the form.", info = "Create single interval label aborts, if list element is missing a name")


set_no_print()

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
# Printing the output is always set to FALSE, so the overall code can be tested
# but without drawing the whole outputs on screen.
###############################################################################

set_style_options(as_heatmap = TRUE)

dummy_df <- dummy_data(1000)


# Different way of passing variables in
result_list1 <- dummy_df |>
    frequencies(variables = age,
                by        = sex,
                weight    = weight,
                print     = FALSE)

result_list2 <- dummy_df |>
    frequencies(variables = "age",
                by        = "sex",
                weight    = "weight",
                print     = FALSE)

expect_identical(result_list1, result_list2, info = "Different way of passing variables in")


# Simplest form of frequencies
result_list <- dummy_df |>
      frequencies(variables = sex,
                  print     = FALSE)

expect_inherits(result_list, "qol_freq", info = "Simplest form of frequencies")
expect_equal(length(result_list), 3, info = "Simplest form of frequencies")

expect_true(is.null(result_list[["mean"]]), info = "Simplest form of frequencies")
expect_true(all(c("fused_vars", "TYPE", "TYPE_NR", "DEPTH",
                  "var_sum", "var_pct_group", "var_freq")
                %in% names(result_list[["freq"]])), info = "Simplest form of frequencies")


# Simplest form of frequencies with means
result_list <- dummy_df |>
        frequencies(variables = sex,
                    means     = TRUE,
                    full_precision = TRUE,
                    print     = FALSE)

expect_true(all(c("variable", "mean", "sd", "min", "max", "freq", "miss")
                %in% names(result_list[["mean"]])), info = "Simplest form of frequencies with means")


# frequencies with titles and footnotes
result_list <- dummy_df |>
        frequencies(variables = sex,
                    titles    = "Hello world",
                    footnotes = "This is a footnote",
                    print     = FALSE)

expect_inherits(result_list, "qol_freq", info = "frequencies with titles and footnotes")
expect_equal(length(result_list), 3, info = "frequencies with titles and footnotes")


# frequencies with multiple variables
result_list <- dummy_df |>
        frequencies(variables = c(sex, age),
                    means     = TRUE,
                    print     = FALSE)

expect_true(all(c("sex", "age") %in% result_list[["freq"]][["TYPE"]]), info = "frequencies with multiple variables")


# frequencies with multiple variables and means
result_list <- dummy_df |>
        frequencies(variables = c(sex, age),
                    means     = TRUE,
                    print     = FALSE)

expect_equal(collapse::fnrow(result_list[["mean"]]), 2, info = "frequencies with multiple variables and means")
expect_true(all(c("sex", "age") %in% result_list[["mean"]][["variable"]]), info = "frequencies with multiple variables and means")


# Character variables won't be evaluated in mean tab
result_list <- dummy_df |>
        frequencies(variables = c(sex, education),
                    means     = TRUE,
                    print     = FALSE)

expect_equal(collapse::fnrow(result_list[["mean"]]), 1, info = "Character variables won't be evaluated in mean tab")
expect_true(!"education" %in% result_list[["mean"]][["variable"]], info = "Character variables won't be evaluated in mean tab")


# frequencies with by variables
result_list <- dummy_df |>
        frequencies(variables = age,
                    by        = sex,
                    print     = FALSE)

expect_true(all(c("by_vars", "BY") %in% names(result_list[["freq"]])), info = "frequencies with by variables")


# frequencies with multiple by variables
result_list <- dummy_df |>
        frequencies(variables = c(age, education),
                    by        = c(sex, year),
                    means     = TRUE,
                    print     = FALSE)

expect_true(all(c("sex", "year") %in% result_list[["freq"]][["BY"]]), info = "frequencies with multiple by variables")


# frequencies where by is also part of freq variables is aborted
result_list <- dummy_df |>
    frequencies(variables = c(age, sex),
                by        = c(sex, year),
                print     = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "The provided <by> variable '", info = "frequencies where by is also part of freq variables is aborted")


# frequencies throws a NOTE, if print_miss and means option are both TRUE
result_list <- dummy_df |>
        frequencies(variables  = age,
                    by         = sex,
                    print_miss = TRUE,
                    means      = TRUE,
                    print      = FALSE)

expect_message(print_stack_as_messages("NOTE"), "Wenn <print_miss> is TRUE, there will be no mean tables. <Means> is set to FALSE.",
               info = "frequencies throws a NOTE, if print_miss and means option are both TRUE")


# frequencies with weighted results
result_list1 <- dummy_df |>
        frequencies(variables = age,
                    weight    = weight,
                    print     = FALSE)

result_list2 <- dummy_df |>
         frequencies(variables = age,
                     print     = FALSE)

expect_false(identical(result_list1[["freq"]][["var_sum"]],
                       result_list2[["freq"]][["var_sum"]]), info = "frequencies with weighted results")


# frequencies with NAs removed
result_list <- dummy_df |>
        frequencies(variables = sex,
                    na.rm     = TRUE,
                    print     = FALSE)

expect_true(sum(is.na(result_list[["freq"]][["fused_vars"]])) <= 1, info = "frequencies with NAs removed")


# frequencies with character variable omits mean table
result_list <- dummy_df |>
        frequencies(variables = education,
                    means     = TRUE,
                    print     = FALSE)

expect_true(is.null(collapse::fnrow(result_list[["means"]])), info = "frequencies with character variable omits mean table")


# Apply single discrete labels
format_list <- dummy_df |>
         frequencies(variables = sex,
                     formats   = list(sex = discrete_format(
                        "Male"   = 1,
                        "Female" = 2)),
                     print     = FALSE)

no_format_list <- dummy_df |>
          frequencies(variables = age,
                      print     = FALSE)

expect_equal(collapse::fsum(format_list[["freq"]][["var_sum"]]),
             collapse::fsum(no_format_list[["freq"]][["var_sum"]]), info = "Apply single discrete labels")
expect_true(all(c("Male", "Female")
                %in% format_list[["freq"]][["fused_vars"]]), info = "Apply single discrete labels")


# Apply single interval labels
format_list <- dummy_df |>
        frequencies(variables = income,
                    formats   = list(income = interval_format(
                        "below 500"          = 0:500,
                        "500 to under 1000"  = 500:1000,
                        "1000 to under 2000" = 1000:2000,
                        "2000 and more"      = 2000:100000)),
                    print     = FALSE)

no_format_list <- dummy_df |>
           frequencies(variables = age,
                       print     = FALSE)

expect_equal(collapse::fsum(format_list[["freq"]][["var_sum"]]),
             collapse::fsum(no_format_list[["freq"]][["var_sum"]]), info = "Apply single interval labels")
expect_true(all(c("below 500", "2000 and more")
                %in% format_list[["freq"]][["fused_vars"]]), info = "Apply single interval labels")


# Apply discrete multilabel
format_list <- dummy_df |>
        frequencies(variables = sex,
                    formats   = list(sex = discrete_format(
                        "Total"  = 1:2,
                        "Male"   = 1,
                        "Female" = 2)),
                    print     = FALSE)

expect_true(all(c("Total", "Male", "Female")
                %in% format_list[["freq"]][["fused_vars"]]), info = "Apply discrete multilabel")


# Apply interval multilabel
format_list <- dummy_df |>
        frequencies(variables = income,
                    formats   = list(income = interval_format(
                        "Total"              = 0:100000,
                        "below 500"          = 0:500,
                        "500 to under 1000"  = 500:1000,
                        "1000 to under 2000" = 1000:2000,
                        "2000 and more"      = 2000:100000)),
                    print     = FALSE)

no_format_list <- dummy_df |>
           frequencies(variables = income,
                       print     = FALSE)

expect_true(all(c("Total", "below 500", "2000 and more")
                %in% format_list[["freq"]][["fused_vars"]]), info = "Apply interval multilabel")


# frequencies throws a warning, if invalid format is used
result_list <- dummy_df |>
    frequencies(variables = sex,
                formats   = list(sex = test),
                print     = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Format for variable 'sex' does not exist and can't be applied.",
               info = "frequencies throws a warning, if invalid format is used")

###############################################################################
# The Excel tests are kept simpler because the whole stat evaluation is the
# same as before. It serves mainly to let the code run through all the formatting
# lines.
###############################################################################

# frequencies with excel output
result_list <- dummy_df |>
    frequencies(variables = age,
                means     = TRUE,
                output    = "excel",
                print     = FALSE)

expect_inherits(result_list, "qol_freq", info = "frequencies with excel output")
expect_equal(length(result_list), 3, info = "frequencies with excel output")


# frequencies with titles and footnotes and weight
result_list <- dummy_df |>
    frequencies(variables = sex,
                output    = "excel_nostyle",
                titles    = "Hello world",
                footnotes = "This is a footnote",
                weight    = weight,
                print     = FALSE)

expect_inherits(result_list, "qol_freq", info = "frequencies with titles and footnotes and weight")
expect_equal(length(result_list), 3, info = "frequencies with titles and footnotes and weight")


# frequencies with excel output and by variables
result_list <- dummy_df |>
    frequencies(variables  = age,
                by         = sex,
                output     = "excel",
                print      = FALSE)

expect_true(all(c("by_vars", "BY") %in% names(result_list[["freq"]])), info = "frequencies with excel output and by variables")


# frequencies with excel output and by variables and print_miss
result_list <- dummy_df |>
    frequencies(variables  = age,
                by         = sex,
                print_miss = TRUE,
                means      = TRUE,
                output     = "excel_nostyle",
                print      = FALSE)

expect_true(all(c("by_vars", "BY") %in% names(result_list[["freq"]])), info = "frequencies with excel output and by variables and print_miss")


# frequencies with borders in excel output
result_list <- dummy_df |>
    frequencies(variables = age,
                output    = "excel",
                print     = FALSE,
                style = excel_output_style(header_borders   = TRUE,
                                           box_borders      = TRUE,
                                           cat_col_borders  = TRUE,
                                           table_borders    = TRUE,
                                           column_widths    = c(2, 3),
                                           row_heights      = c(2, 3)))

expect_inherits(result_list, "qol_freq", info = "frequencies with borders in excel output")
expect_equal(length(result_list), 3, info = "frequencies with borders in excel output")


# frequencies with set column width/row heights
result_list <- dummy_df |>
    frequencies(variables = age,
                output    = "excel",
                print     = FALSE,
                style = excel_output_style(column_widths = 2,
                                           row_heights   = 3))

expect_inherits(result_list, "qol_freq", info = "frequencies with set column width/row heights")
expect_equal(length(result_list), 3, info = "frequencies with set column width/row heights")


# frequencies with set partial row heights
result_list <- dummy_df |>
    frequencies(variables = age,
                output    = "excel",
                titles    = "Hello world",
                footnotes = "This is a footnote",
                print     = FALSE,
                style = excel_output_style(title_heights    = c(10, 10),
                                           header_heights   = c(10, 10),
                                           table_heights    = c(10),
                                           footnote_heights = 10))

expect_inherits(result_list, "qol_freq", info = "frequencies with set partial row heights")
expect_equal(length(result_list), 3, info = "frequencies with set partial row heights")


# frequencies with fast none styled excel output
result_list <- dummy_df |>
    frequencies(variables = age,
                output    = "excel_nostyle",
                print     = FALSE)

expect_inherits(result_list, "qol_freq", info = "frequencies with fast none styled excel output")
expect_equal(length(result_list), 3, info = "frequencies with fast none styled excel output")


# Invalid output format leads to console output
result_list <- dummy_df |>
        frequencies(variables = age,
                    by        = sex,
                    output    = "test",
                    print     = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "<Output> format 'test' not available. Using 'console' instead.", info = "Invalid output format leads to console output")


# Save frequencies as Excel file
temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

dummy_df |>
     frequencies(variables = age,
                 output    = "excel",
                 style     = excel_output_style(save_path = dirname(temp_file),
                                                file      = basename(temp_file)))

expect_true(file.exists(temp_file), info = "Save frequencies as Excel file")


set_style_options(as_heatmap = FALSE)


set_no_print()

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

test_df <- dummy_data(1000)


# Recode discrete values into groups with all values covered
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

test_df2 <- test_df
test_df2[["age_group"]] <- test_df2 |> recode.(age = age.)

expect_true("age_group" %in% names(test_df2), info = "Recode discrete values into groups with all values covered")

expect_true(length(unique(test_df2[["age_group"]])) <= 6, info = "Recode discrete values into groups with all values covered")


# Recode interval values into groups with all values covered
income. <- interval_format(
    "below 500"          = 0:500,
    "500 to under 1000"  = 500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

test_df2 <- test_df
test_df2 <- test_df2 |> collapse::fsubset(!is.na(income))
test_df2[["income_group"]] <- test_df2 |> recode.(income = income.)

expect_true("income_group" %in% names(test_df2), info = "Recode interval values into groups with all values covered")

expect_true(length(unique(test_df2[["income_group"]])) <= 5, info = "Recode interval values into groups with all values covered")


# Recode interval values with NA values
income. <- interval_format(
    "00. no income"          = 0,
    "01.         under  500" = c(0.000001, 500),
    "02.  500 to under 1000" =  500:1000,
    "03. 1000 to under 1500" = 1000:1500,
    "04. 1500 to under 2000" = 1500:2000,
    "05. 2000 to under 2500" = 2000:2500,
    "06. 2500 to under 3000" = 2500:3000,
    "07. 3000 to under 3500" = 3000:3500,
    "08. 3500 to under 4000" = 3500:4000,
    "09. 4000 to under 4500" = 4000:4500,
    "10. 4500 to under 5000" = 4500:5000,
    "11. 5000 and more"      = c(5000, "high"),
    "99. Total"              = c("low", "high"))

test_df2 <- test_df
test_df2[["income_group"]] <- test_df2 |> recode.(income = income.)

expect_equal(test_df2[["income_class"]], test_df2[["income_group"]],
             info = "Recode interval values into groups with all values covered")
expect_warning(print_stack_as_messages("WARNING"), "The format for 'income' is a multilabel. A multilabel can't be fully applied in recode.",
               info = "Recode interval values with NA values")


# Recode discrete values into groups with not all values covered
age. <- discrete_format(
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

test_df2 <- test_df
test_df2 <- test_df2 |> collapse::fsubset(!is.na(income))
test_df2[["age_group"]] <- test_df2 |> recode.(age = age.)

expect_true("age_group" %in% names(test_df2), info = "Recode discrete values into groups with not all values covered")

unique_age       <- length(unique(test_df2[["age"]]))
unique_age_group <- length(unique(test_df2[["age_group"]]))
expect_true(unique_age_group >= 4 & unique_age_group < unique_age, info = "Recode discrete values into groups with not all values covered")


# Recode interval values into groups with not all values covered
income. <- interval_format(
    "below 500"          = 0:500,
    "500 to under 1000"  = 500:1000,
    "1000 to under 2000" = 1000:2000)

test_df2 <- test_df
test_df2 <- test_df2 |> collapse::fsubset(!is.na(income))
test_df2[["income_group"]] <- test_df2 |> recode.(income = income.)

expect_true("income_group" %in% names(test_df2), info = "Recode interval values into groups with not all values covered")

unique_income       <- length(unique(test_df2[["income"]]))
unique_income_group <- length(unique(test_df2[["income_group"]]))
expect_true(unique_income_group >= 4 & unique_income_group < unique_income, info = "Recode interval values into groups with not all values covered")


# Providing none format data frames
age1 <- c(1, 2, 3, 4)
age2 <- c("a", "b", "c", "d")

test_df2 <- test_df
test_df2[["age_group"]] <- test_df2 |> recode.(age = age1)

expect_error(print_stack_as_messages("ERROR"), "The format for 'age' must be a data table", info = "Providing none format data frames")
test_df2[["age_group"]] <- test_df2 |> recode.(age = age2)

expect_error(print_stack_as_messages("ERROR"), "The format for 'age' must be a data table", info = "Providing none format data frames")

expect_true(!"age_group" %in% names(test_df2), info = "Providing none format data frames")
expect_true(!"age_group" %in% names(test_df2), info = "Providing none format data frames")


# Given variable is not in data frame
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

test_df2 <- test_df
test_df2[["age_group"]] <- test_df2 |> recode.(dog = age.)

expect_error(print_stack_as_messages("ERROR"), "Variable 'dog' not found in the input data frame", info = "Given variable is not in data frame")

expect_true(!"age_group" %in% names(test_df2), info = "Given variable is not in data frame")


# Recoding with multilabel gives a warning
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

test_df2 <- test_df
test_df2[["age_group"]] <- test_df2 |> recode.(age = age.)

expect_warning(print_stack_as_messages("WARNING"), "The format for 'age' is a multilabel", info = "Recoding with multilabel gives a warning")


# Recode will overwrite existing variable
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

test_df2 <- test_df
test_df2[["age"]] <- test_df2 |> recode.(age = age.)

expect_true(all(c("under 18",
                  "18 to under 25",
                  "25 to under 55",
                  "55 to under 65",
                  "65 and older") %in% test_df2[["age"]]), info = "Recode will overwrite existing variable")


# Recode multiple variables at once into multiple new variables
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

income. <- interval_format(
    "below 500"          = 0:500,
    "500 to under 1000"  = 500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

test_df2 <- test_df
test_df2[, c("age_group", "income_group")] <- test_df2 |>
    recode.(age = age., income = income.)

expect_true("age_group"    %in% names(test_df2), info = "Recode multiple variables into multiple new variables (discrete)")
expect_true("income_group" %in% names(test_df2), info = "Recode multiple variables into multiple new variables (interval)")
expect_true(length(unique(test_df2[["age_group"]]))    <= 6, info = "Recode multiple variables into multiple new variables (discrete values)")
expect_true(length(unique(test_df2[["income_group"]])) <= 5, info = "Recode multiple variables into multiple new variables (interval values)")


# recode. keeps original values which are not covered by a discrete format
sex. <- discrete_format(
    "Male"   = 1,
    "Female" = 2)

test_df2 <- test_df
test_df2[["persons"]] <- test_df |> recode.(number_of_persons = sex.)

expect_true(all(c("Male", "Female", as.character(3:5)) %in% test_df2[["persons"]]),
            info = "recode. keeps original values which are not covered by a format")


###############################################################################
# Multi recode
###############################################################################

# Recode a variable with a multilabel format (discrete)
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

recode_df <- test_df |> recode_multi(age = age.)

expect_true(collapse::fncol(recode_df) == collapse::fncol(test_df), info = "Recode a variable with a multilabel format (discrete)")
expect_true(collapse::fnrow(recode_df) > collapse::fnrow(test_df),  info = "Recode a variable with a multilabel format (discrete)")


# Recode a variable with a multilabel format (interval)
income. <- interval_format(
    "Total"              = 0:100000,
    "below 500"          = 0:500,
    "500 to under 1000"  = 500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

recode_df <- test_df |> recode_multi(income = income.)

expect_true(collapse::fncol(recode_df) == collapse::fncol(test_df), info = "Recode a variable with a multilabel format (interval)")
expect_true(collapse::fnrow(recode_df) > collapse::fnrow(test_df),  info = "Recode a variable with a multilabel format (interval)")


# Recode multiple variables at once
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

income. <- interval_format(
    "Total"              = 0:100000,
    "below 500"          = 0:500,
    "500 to under 1000"  = 500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

recode_df <- test_df |> recode_multi(age = age., income = income.)

expect_true(collapse::fncol(recode_df) == collapse::fncol(test_df), info = "Recode multiple variables at once")
expect_true(collapse::fnrow(recode_df) > collapse::fnrow(test_df),  info = "Recode multiple variables at once")


# recode_multi keeps original values which are not covered by a discrete format
sex. <- discrete_format(
    "Male"   = 1,
    "Female" = 2)

recode_df <- test_df |> recode_multi(number_of_persons = sex.)

expect_true(all(c("Male", "Female", as.character(3:5)) %in% recode_df[["number_of_persons"]]),
            info = "recode_multi keeps original values which are not covered by a format")


set_no_print()

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
# Printing the output is always set to FALSE, so the overall code can be tested
# but without drawing the whole outputs on screen.
###############################################################################

set_style_options(as_heatmap = TRUE)

dummy_df <- dummy_data(1000)


# Different way of passing variables in
result_list1 <- dummy_df |>
     crosstabs(rows    = age,
               columns = sex,
               by      = education,
               weight  = weight,
               print   = FALSE)

result_list2 <- dummy_df |>
     crosstabs(rows    = "age",
               columns = "sex",
               by      = "education",
               weight  = "weight",
               print   = FALSE)

expect_identical(result_list1, result_list2, info = "Different way of passing variables in")


# Simplest form of crosstabs
result_list <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              print   = FALSE)

expect_inherits(result_list, "qol_cross", info = "Simplest form of crosstabs")
expect_equal(length(result_list), 2, info = "Simplest form of crosstabs")

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_list[[1]]), (values * 4) + 1, info = "Simplest form of crosstabs")


# crosstabs with titles and footnotes
result_list <- dummy_df |>
    crosstabs(rows      = age,
              columns   = sex,
              titles    = "Hello world",
              footnotes = "This is a footnote",
              full_precision = TRUE,
              print     = FALSE)

expect_inherits(result_list, "qol_cross", info = "crosstabs with titles and footnotes")

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_list[[1]]), (values * 4) + 1, info = "crosstabs with titles and footnotes")


# crosstabs not allowed with multiple row variables
result_list <- dummy_df |>
    crosstabs(rows    = c(age, state),
              columns = sex,
              print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Only one variable for <rows> allowed. First variable will be used.", info = "crosstabs not allowed with multiple row variables")


# crosstabs not allowed with multiple column variables
result_list <- dummy_df |>
     crosstabs(rows    = sex,
               columns = c(age, state),
               print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Only one variable for <columns> allowed. First variable will be used.", info = "crosstabs not allowed with multiple column variables")


# crosstabs with by variables
result_list <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              by      = education,
              print   = FALSE)

expect_true("BY" %in% names(result_list[[1]]), info = "crosstabs with by variables")
expect_equal(length(unique(result_list[[1]][["BY"]])), 1, info = "crosstabs with by variables")


# crosstabs with multiple by variables
result_list <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              by      = c(education, year),
              print   = FALSE)

expect_true("BY" %in% names(result_list[[1]]), info = "crosstabs with multiple by variables")
expect_equal(length(unique(result_list[[1]][["BY"]])), 2, info = "crosstabs with multiple by variables")


# crosstabs where by is also part of rows or columns is aborted
result_list <- dummy_df |>
       crosstabs(rows    = age,
                 columns = sex,
                 by      = c(sex, year),
                 print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "The provided <by> variable '", info = "crosstabs where by is also part of rows or columns is aborted")


# crosstabs with weighted results
result_list1 <- dummy_df |>
     crosstabs(rows    = age,
               columns = sex,
               weight  = weight,
               print   = FALSE)

result_list2 <- dummy_df |>
     crosstabs(rows    = age,
               columns = sex,
               print   = FALSE)

expect_false(identical(as.numeric(result_list1[[1]][["var_sum_1"]]),
                       as.numeric(result_list2[[1]][["var_sum_1"]])), info = "crosstabs with weighted results")


# crosstabs with NAs removed
result_list <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              na.rm   = TRUE,
              print   = FALSE)

expect_true(sum(is.na(result_list[[1]][["fused_vars"]])) == 0, info = "crosstabs with NAs removed")


# crosstabs with multiple statistics
result_list <- dummy_df |>
      crosstabs(rows       = age,
                columns    = sex,
                statistics = c("sum", "freq", "pct_row", "pct_column", "pct_total"),
                print      = FALSE)

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_list[[1]]), (values * 4) + 1, info = "crosstabs with multiple statistics")


# Apply single discrete labels
format_df <- dummy_df |>
    crosstabs(rows    = sex,
              columns = age,
              formats = list(sex = discrete_format(
                  "Male"   = 1,
                  "Female" = 2)),
              print   = FALSE)

no_format_df <- dummy_df |>
   crosstabs(rows    = sex,
             columns = age,
             print   = FALSE)

expect_equal(collapse::fsum(format_df[["var_sum_1"]]),
             collapse::fsum(no_format_df[["var_sum_1"]]), info = "Apply single discrete labels")
expect_true(all(c("Male", "Female")
                %in% format_df[[1]][["sex"]]), info = "Apply single discrete labels")


# Apply discrete multilabel
format_df <- dummy_df |>
    crosstabs(rows    = sex,
              columns = age,
              formats = list(sex = discrete_format(
                                  "Total"  = 1:2,
                                  "Male"   = 1,
                                  "Female" = 2),
                              age = discrete_format(
                                  "Total"          = 0:100,
                                  "under 18"       = 0:17,
                                  "18 to under 25" = 18:24,
                                  "25 to under 55" = 25:54,
                                  "55 to under 65" = 55:64,
                                  "65 and older"   = 65:100)),
              print   = FALSE)

expect_true(all(c("Total", "Male", "Female")
                %in% format_df[[1]][["sex"]]), info = "Apply discrete multilabel")


# Apply interval multilabel
format_df <- dummy_df |>
    crosstabs(rows    = income,
              columns = sex,
              formats = list(income = interval_format(
                    "Total"              = 0:100000,
                    "below 500"          = 0:500,
                    "500 to under 1000"  = 500:1000,
                    "1000 to under 2000" = 1000:2000,
                    "2000 and more"      = 2000:100000)),
              print   = FALSE)

expect_true(all(c("Total", "below 500", "2000 and more")
                %in% format_df[[1]][["income"]]), info = "Apply interval multilabel")


# crosstabs row multilabel leads to notification
format_df <- dummy_df |>
      crosstabs(rows    = sex,
                columns = age,
                formats = list(sex = discrete_format(
                    "Total"  = 1:2,
                    "Male"   = 1,
                    "Female" = 2)),
                statistics = c("pct_row"),
                print   = FALSE)

expect_message(print_stack_as_messages("NOTE"), "The format for variable 'sex' is a multilabel", info = "crosstabs row multilabel leads to notification")


# crosstabs row multilabel leads to notification with by variables
format_df <- dummy_df |>
       crosstabs(rows    = sex,
                 columns = age,
                 by      = year,
                 formats = list(sex = discrete_format(
                     "Total"  = 1:2,
                     "Male"   = 1,
                     "Female" = 2)),
                 statistics = c("pct_row"),
                 print   = FALSE)

expect_message(print_stack_as_messages("NOTE"), "The format for variable 'sex' is a multilabel", info = "crosstabs row multilabel leads to notification with by variables")


# crosstabs throws a warning, if invalid format is used
result_list <- dummy_df |>
    crosstabs(rows    = sex,
              columns = age,
              formats = list(age = age., sex = test),
              print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Format for variable 'sex' does not exist and can't be applied.",
               info = "crosstabs throws a warning, if invalid format is used")

###############################################################################
# The Excel tests are kept simpler because the whole stat evaluation is the
# same as before. It serves mainly to let the code run through all the formatting
# lines.
###############################################################################

# crosstabs with excel output
result_list <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              output  = "excel_nostyle",
              print   = FALSE)

expect_inherits(result_list, "qol_cross", info = "crosstabs with excel output")

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_list[[1]]), (values * 4) + 1, info = "crosstabs with excel output")


# crosstabs with titles and footnotes and weight (excel)
result_list <- dummy_df |>
    crosstabs(rows      = age,
              columns   = sex,
              output    = "excel",
              titles    = "Hello world",
              footnotes = "This is a footnote",
              weight    = weight,
              print     = FALSE)

expect_inherits(result_list, "qol_cross", info = "crosstabs with titles and footnotes and weight (excel)")

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_list[[1]]), (values * 4) + 1, info = "crosstabs with titles and footnotes and weight (excel)")


# crosstabs with excel output and by variables (excel)
result_list <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              by      = education,
              output  = "excel_nostyle",
              print_miss = TRUE, # Has no value here but the test runs faster
              print   = FALSE)

expect_true("BY" %in% names(result_list[[1]]), info = "crosstabs with excel output and by variables (excel)")


# crosstabs with fast none styled excel output (excel)
result_list <- dummy_df |>
      crosstabs(rows    = age,
                columns = sex,
                output  = "excel_nostyle",
                print   = FALSE)

expect_inherits(result_list, "qol_cross", info = "crosstabs with fast none styled excel output (excel)")

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_list[[1]]), (values * 4) + 1, info = "crosstabs with fast none styled excel output (excel)")


# crosstabs row multilabel leads to notification (excel)
format_df <- dummy_df |>
       crosstabs(rows    = sex,
                 columns = age,
                 formats = list(sex = discrete_format(
                                     "Total"  = 1:2,
                                     "Male"   = 1,
                                     "Female" = 2),
                                 age = discrete_format(
                                     "Total"          = 0:100,
                                     "under 18"       = 0:17,
                                     "18 to under 25" = 18:24,
                                     "25 to under 55" = 25:54,
                                     "55 to under 65" = 55:64,
                                     "65 and older"   = 65:100)),
                 statistics = c("sum", "freq", "pct_row", "pct_column", "pct_total"),
                 output  = "excel_nostyle",
                 print   = FALSE)

expect_message(print_stack_as_messages("NOTE"), "The format for variable 'sex' is a multilabel", info = "crosstabs row multilabel leads to notification (excel)")


# crosstabs row multilabel leads to notification (excel)
format_df <- dummy_df |>
       crosstabs(rows    = sex,
                 columns = age,
                 formats = list(sex = discrete_format(
                     "Total"  = 1:2,
                     "Male"   = 1,
                     "Female" = 2)),
                 statistics = c("pct_row"),
                 output  = "excel",
                 print   = FALSE)

expect_message(print_stack_as_messages("NOTE"), "The format for variable 'sex' is a multilabel", info = "crosstabs row multilabel leads to notification (excel)")


# crosstabs row multilabel leads to notification with by variables (excel)
format_df <- dummy_df |>
       crosstabs(rows    = sex,
                 columns = age,
                 by      = year,
                 formats = list(sex = discrete_format(
                     "Total"  = 1:2,
                     "Male"   = 1,
                     "Female" = 2)),
                 statistics = c("pct_row"),
                 output     = "excel_nostyle",
                 print_miss = TRUE,
                 print      = FALSE)

expect_message(print_stack_as_messages("NOTE"), "The format for variable 'sex' is a multilabel", info = "crosstabs row multilabel leads to notification with by variables (excel)")


# Invalid output format leads to console output
result_list <- dummy_df |>
       crosstabs(rows    = age,
                 columns = sex,
                 by      = education,
                 output  = "test",
                 print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "<Output> format 'test' not available. Using 'console' instead.", info = "Invalid output format leads to console output")


# Save crosstabs as Excel file
temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

dummy_df |>
     crosstabs(rows    = age,
               columns = sex,
               output  = "excel",
               style   = excel_output_style(save_path = dirname(temp_file),
                                            file      = basename(temp_file)))

expect_true(file.exists(temp_file), info = "Save crosstabs as Excel file")


set_style_options(as_heatmap = FALSE)
set_no_print()

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
# Printing the output is always set to FALSE, so the overall code can be tested
# but without drawing the whole outputs on screen.
###############################################################################

set_style_options(as_heatmap = TRUE)

dummy_df <- dummy_data(1000)


# Different way of passing variables in
result_df1 <- dummy_df |>
     crosstabs(rows    = age,
               columns = sex,
               by      = education,
               weight  = weight,
               print   = FALSE)

result_df2 <- dummy_df |>
     crosstabs(rows    = "age",
               columns = "sex",
               by      = "education",
               weight  = "weight",
               print   = FALSE)

expect_identical(result_df1, result_df2, info = "Different way of passing variables in")


# Simplest form of crosstabs
result_df <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              print   = FALSE)

expect_inherits(result_df, "data.table", info = "Simplest form of crosstabs")

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_df), (values * 4) + 1, info = "Simplest form of crosstabs")


# crosstabs with titles and footnotes
result_df <- dummy_df |>
    crosstabs(rows      = age,
              columns   = sex,
              titles    = "Hello world",
              footnotes = "This is a footnote",
              print     = FALSE)

expect_inherits(result_df, "data.table", info = "crosstabs with titles and footnotes")

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_df), (values * 4) + 1, info = "crosstabs with titles and footnotes")


# crosstabs not allowed with multiple row variables
result_df <- dummy_df |>
    crosstabs(rows    = c(age, state),
              columns = sex,
              print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Only one variable for <rows> allowed. First variable will be used.", info = "crosstabs not allowed with multiple row variables")


# crosstabs not allowed with multiple column variables
result_df <- dummy_df |>
     crosstabs(rows    = sex,
               columns = c(age, state),
               print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Only one variable for <columns> allowed. First variable will be used.", info = "crosstabs not allowed with multiple column variables")


# crosstabs with by variables
result_df <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              by      = education,
              print   = FALSE)

expect_true("BY" %in% names(result_df), info = "crosstabs with by variables")
expect_equal(length(unique(result_df[["BY"]])), 1, info = "crosstabs with by variables")


# crosstabs with multiple by variables
result_df <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              by      = c(education, year),
              print   = FALSE)

expect_true("BY" %in% names(result_df), info = "crosstabs with multiple by variables")
expect_equal(length(unique(result_df[["BY"]])), 2, info = "crosstabs with multiple by variables")


# crosstabs where by is also part of rows or columns is aborted
result_list <- dummy_df |>
       crosstabs(rows    = age,
                 columns = sex,
                 by      = c(sex, year),
                 print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "The provided <by> variable '", info = "crosstabs where by is also part of rows or columns is aborted")


# crosstabs with weighted results
result_df1 <- dummy_df |>
     crosstabs(rows    = age,
               columns = sex,
               weight  = weight,
               print   = FALSE)

result_df2 <- dummy_df |>
     crosstabs(rows    = age,
               columns = sex,
               print   = FALSE)

expect_false(identical(as.numeric(result_df1[["var_sum_1"]]),
                       as.numeric(result_df2[["var_sum_1"]])), info = "crosstabs with weighted results")


# crosstabs with NAs removed
result_df <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              na.rm   = TRUE,
              print   = FALSE)

expect_true(sum(is.na(result_df[["fused_vars"]])) == 0, info = "crosstabs with NAs removed")


# crosstabs with multiple statistics
result_df <- dummy_df |>
      crosstabs(rows       = age,
                columns    = sex,
                statistics = c("sum", "freq", "pct_row", "pct_column", "pct_total"),
                print      = FALSE)

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_df), (values * 4) + 1, info = "crosstabs with multiple statistics")


# Apply single discrete labels
format_df <- dummy_df |>
    crosstabs(rows    = sex,
              columns = age,
              formats = list(sex = discrete_format(
                  "Male"   = 1,
                  "Female" = 2)),
              print   = FALSE)

no_format_df <- dummy_df |>
   crosstabs(rows    = sex,
             columns = age,
             print   = FALSE)

expect_equal(collapse::fsum(format_df[["var_sum_1"]]),
             collapse::fsum(no_format_df[["var_sum_1"]]), info = "Apply single discrete labels")
expect_true(all(c("Male", "Female")
                %in% format_df[["sex"]]), info = "Apply single discrete labels")


# Apply discrete multilabel
format_df <- dummy_df |>
    crosstabs(rows    = sex,
              columns = age,
              formats = list(sex = discrete_format(
                                  "Total"  = 1:2,
                                  "Male"   = 1,
                                  "Female" = 2),
                              age = discrete_format(
                                  "Total"          = 0:100,
                                  "under 18"       = 0:17,
                                  "18 to under 25" = 18:24,
                                  "25 to under 55" = 25:54,
                                  "55 to under 65" = 55:64,
                                  "65 and older"   = 65:100)),
              print   = FALSE)

expect_true(all(c("Total", "Male", "Female")
                %in% format_df[["sex"]]), info = "Apply discrete multilabel")


# Apply interval multilabel
format_df <- dummy_df |>
    crosstabs(rows    = income,
              columns = sex,
              formats = list(income = interval_format(
                    "Total"              = 0:100000,
                    "below 500"          = 0:500,
                    "500 to under 1000"  = 500:1000,
                    "1000 to under 2000" = 1000:2000,
                    "2000 and more"      = 2000:100000)),
              print   = FALSE)

expect_true(all(c("Total", "below 500", "2000 and more")
                %in% format_df[["income"]]), info = "Apply interval multilabel")


# crosstabs row multilabel leads to notification
format_df <- dummy_df |>
      crosstabs(rows    = sex,
                columns = age,
                formats = list(sex = discrete_format(
                    "Total"  = 1:2,
                    "Male"   = 1,
                    "Female" = 2)),
                statistics = c("pct_row"),
                print   = FALSE)

expect_message(print_stack_as_messages("NOTE"), "The format for variable 'sex' is a multilabel", info = "crosstabs row multilabel leads to notification")


# crosstabs row multilabel leads to notification with by variables
format_df <- dummy_df |>
       crosstabs(rows    = sex,
                 columns = age,
                 by      = year,
                 formats = list(sex = discrete_format(
                     "Total"  = 1:2,
                     "Male"   = 1,
                     "Female" = 2)),
                 statistics = c("pct_row"),
                 print   = FALSE)

expect_message(print_stack_as_messages("NOTE"), "The format for variable 'sex' is a multilabel", info = "crosstabs row multilabel leads to notification with by variables")


# crosstabs throws a warning, if invalid format is used
result_list <- dummy_df |>
    crosstabs(rows    = sex,
              columns = age,
              formats = list(age = age., sex = test),
              print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Format for variable 'sex' does not exist and can't be applied.",
               info = "crosstabs throws a warning, if invalid format is used")

###############################################################################
# The Excel tests are kept simpler because the whole stat evaluation is the
# same as before. It serves mainly to let the code run through all the formatting
# lines.
###############################################################################

# crosstabs with excel output
result_df <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              output  = "excel_nostyle",
              print   = FALSE)

expect_inherits(result_df, "data.table", info = "crosstabs with excel output")

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_df), (values * 4) + 1, info = "crosstabs with excel output")


# crosstabs with titles and footnotes and weight (excel)
result_df <- dummy_df |>
    crosstabs(rows      = age,
              columns   = sex,
              output    = "excel",
              titles    = "Hello world",
              footnotes = "This is a footnote",
              weight    = weight,
              print     = FALSE)

expect_inherits(result_df, "data.table", info = "crosstabs with titles and footnotes and weight (excel)")

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_df), (values * 4) + 1, info = "crosstabs with titles and footnotes and weight (excel)")


# crosstabs with excel output and by variables (excel)
result_df <- dummy_df |>
    crosstabs(rows    = age,
              columns = sex,
              by      = education,
              output  = "excel_nostyle",
              print_miss = TRUE, # Has no value here but the test runs faster
              print   = FALSE)

expect_true("BY" %in% names(result_df), info = "crosstabs with excel output and by variables (excel)")


# crosstabs with fast none styled excel output (excel)
result_df <- dummy_df |>
      crosstabs(rows    = age,
                columns = sex,
                output  = "excel_nostyle",
                print   = FALSE)

expect_inherits(result_df, "data.table", info = "crosstabs with fast none styled excel output (excel)")

values <- length(unique(dummy_df[["sex"]]))
expect_equal(collapse::fncol(result_df), (values * 4) + 1, info = "crosstabs with fast none styled excel output (excel)")


# crosstabs row multilabel leads to notification (excel)
format_df <- dummy_df |>
       crosstabs(rows    = sex,
                 columns = age,
                 formats = list(sex = discrete_format(
                                     "Total"  = 1:2,
                                     "Male"   = 1,
                                     "Female" = 2),
                                 age = discrete_format(
                                     "Total"          = 0:100,
                                     "under 18"       = 0:17,
                                     "18 to under 25" = 18:24,
                                     "25 to under 55" = 25:54,
                                     "55 to under 65" = 55:64,
                                     "65 and older"   = 65:100)),
                 statistics = c("sum", "freq", "pct_row", "pct_column", "pct_total"),
                 output  = "excel_nostyle",
                 print   = FALSE)

expect_message(print_stack_as_messages("NOTE"), "The format for variable 'sex' is a multilabel", info = "crosstabs row multilabel leads to notification (excel)")


# crosstabs row multilabel leads to notification (excel)
format_df <- dummy_df |>
       crosstabs(rows    = sex,
                 columns = age,
                 formats = list(sex = discrete_format(
                     "Total"  = 1:2,
                     "Male"   = 1,
                     "Female" = 2)),
                 statistics = c("pct_row"),
                 output  = "excel",
                 print   = FALSE)

expect_message(print_stack_as_messages("NOTE"), "The format for variable 'sex' is a multilabel", info = "crosstabs row multilabel leads to notification (excel)")


# crosstabs row multilabel leads to notification with by variables (excel)
format_df <- dummy_df |>
       crosstabs(rows    = sex,
                 columns = age,
                 by      = year,
                 formats = list(sex = discrete_format(
                     "Total"  = 1:2,
                     "Male"   = 1,
                     "Female" = 2)),
                 statistics = c("pct_row"),
                 output     = "excel_nostyle",
                 print_miss = TRUE,
                 print      = FALSE)

expect_message(print_stack_as_messages("NOTE"), "The format for variable 'sex' is a multilabel", info = "crosstabs row multilabel leads to notification with by variables (excel)")


# Invalid output format leads to console output
result_df <- dummy_df |>
       crosstabs(rows    = age,
                 columns = sex,
                 by      = education,
                 output  = "test",
                 print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "<Output> format 'test' not available. Using 'console' instead.", info = "Invalid output format leads to console output")


# Save crosstabs as Excel file
temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

dummy_df |>
     crosstabs(rows    = age,
               columns = sex,
               output  = "excel",
               style   = excel_output_style(save_path = dirname(temp_file),
                                            file      = basename(temp_file)))

expect_true(file.exists(temp_file), info = "Save crosstabs as Excel file")


set_style_options(as_heatmap = FALSE)


set_no_print()

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

test_df <- dummy_data(1000)


# Recode discrete values into groups with all values covered
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

test_df2 <- test_df
test_df2[["age_group"]] <- test_df2 |> recode.(age = age.)

expect_true("age_group" %in% names(test_df2), info = "Recode discrete values into groups with all values covered")

expect_true(length(unique(test_df2[["age_group"]])) <= 6, info = "Recode discrete values into groups with all values covered")


# Recode interval values into groups with all values covered
income. <- interval_format(
    "below 500"          = 0:500,
    "500 to under 1000"  = 500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

test_df2 <- test_df
test_df2 <- test_df2 |> collapse::fsubset(!is.na(income))
test_df2[["income_group"]] <- test_df2 |> recode.(income = income.)

expect_true("income_group" %in% names(test_df2), info = "Recode interval values into groups with all values covered")

expect_true(length(unique(test_df2[["income_group"]])) <= 5, info = "Recode interval values into groups with all values covered")


# Recode interval values with NA values
income. <- interval_format(
    "00. no income"          = 0,
    "01.         under  500" = c(0.000001, 500),
    "02.  500 to under 1000" =  500:1000,
    "03. 1000 to under 1500" = 1000:1500,
    "04. 1500 to under 2000" = 1500:2000,
    "05. 2000 to under 2500" = 2000:2500,
    "06. 2500 to under 3000" = 2500:3000,
    "07. 3000 to under 3500" = 3000:3500,
    "08. 3500 to under 4000" = 3500:4000,
    "09. 4000 to under 4500" = 4000:4500,
    "10. 4500 to under 5000" = 4500:5000,
    "11. 5000 and more"      = c(5000, "high"),
    "99. Total"              = c("low", "high"))

test_df2 <- test_df
test_df2[["income_group"]] <- test_df2 |> recode.(income = income.)

expect_equal(test_df2[["income_class"]], test_df2[["income_group"]],
             info = "Recode interval values into groups with all values covered")
expect_warning(print_stack_as_messages("WARNING"), "The format for 'income' is a multilabel. A multilabel can't be fully applied in recode.",
               info = "Recode interval values with NA values")


# Recode discrete values into groups with not all values covered
age. <- discrete_format(
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

test_df2 <- test_df
test_df2 <- test_df2 |> collapse::fsubset(!is.na(income))
test_df2[["age_group"]] <- test_df2 |> recode.(age = age.)

expect_true("age_group" %in% names(test_df2), info = "Recode discrete values into groups with not all values covered")

unique_age       <- length(unique(test_df2[["age"]]))
unique_age_group <- length(unique(test_df2[["age_group"]]))
expect_true(unique_age_group >= 4 & unique_age_group < unique_age, info = "Recode discrete values into groups with not all values covered")


# Recode interval values into groups with not all values covered
income. <- interval_format(
    "below 500"          = 0:500,
    "500 to under 1000"  = 500:1000,
    "1000 to under 2000" = 1000:2000)

test_df2 <- test_df
test_df2 <- test_df2 |> collapse::fsubset(!is.na(income))
test_df2[["income_group"]] <- test_df2 |> recode.(income = income.)

expect_true("income_group" %in% names(test_df2), info = "Recode interval values into groups with not all values covered")

unique_income       <- length(unique(test_df2[["income"]]))
unique_income_group <- length(unique(test_df2[["income_group"]]))
expect_true(unique_income_group >= 4 & unique_income_group < unique_income, info = "Recode interval values into groups with not all values covered")


# Providing none format data frames
age1 <- c(1, 2, 3, 4)
age2 <- c("a", "b", "c", "d")

test_df2 <- test_df
test_df2[["age_group"]] <- test_df2 |> recode.(age = age1)

expect_error(print_stack_as_messages("ERROR"), "The format for 'age' must be a data table", info = "Providing none format data frames")
test_df2[["age_group"]] <- test_df2 |> recode.(age = age2)

expect_error(print_stack_as_messages("ERROR"), "The format for 'age' must be a data table", info = "Providing none format data frames")

expect_true(!"age_group" %in% names(test_df2), info = "Providing none format data frames")
expect_true(!"age_group" %in% names(test_df2), info = "Providing none format data frames")


# Given variable is not in data frame
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

test_df2 <- test_df
test_df2[["age_group"]] <- test_df2 |> recode.(dog = age.)

expect_error(print_stack_as_messages("ERROR"), "Variable 'dog' not found in the input data frame", info = "Given variable is not in data frame")

expect_true(!"age_group" %in% names(test_df2), info = "Given variable is not in data frame")


# Recoding with multilabel gives a warning
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

test_df2 <- test_df
test_df2[["age_group"]] <- test_df2 |> recode.(age = age.)

expect_warning(print_stack_as_messages("WARNING"), "The format for 'age' is a multilabel", info = "Recoding with multilabel gives a warning")


# Recode will overwrite existing variable
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

test_df2 <- test_df
test_df2[["age"]] <- test_df2 |> recode.(age = age.)

expect_true(all(c("under 18",
                  "18 to under 25",
                  "25 to under 55",
                  "55 to under 65",
                  "65 and older") %in% test_df2[["age"]]), info = "Recode will overwrite existing variable")

###############################################################################
# Multi recode
###############################################################################

# Recode a variable with a multilabel format (discrete)
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

recode_df <- test_df |> recode_multi(age = age.)

expect_true(collapse::fncol(recode_df) == collapse::fncol(test_df), info = "Recode a variable with a multilabel format (discrete)")
expect_true(collapse::fnrow(recode_df) > collapse::fnrow(test_df),  info = "Recode a variable with a multilabel format (discrete)")


# Recode a variable with a multilabel format (interval)
income. <- interval_format(
    "Total"              = 0:100000,
    "below 500"          = 0:500,
    "500 to under 1000"  = 500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

recode_df <- test_df |> recode_multi(income = income.)

expect_true(collapse::fncol(recode_df) == collapse::fncol(test_df), info = "Recode a variable with a multilabel format (interval)")
expect_true(collapse::fnrow(recode_df) > collapse::fnrow(test_df),  info = "Recode a variable with a multilabel format (interval)")


# Recode multiple variables at once
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

income. <- interval_format(
    "Total"              = 0:100000,
    "below 500"          = 0:500,
    "500 to under 1000"  = 500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

recode_df <- test_df |> recode_multi(age = age., income = income.)

expect_true(collapse::fncol(recode_df) == collapse::fncol(test_df), info = "Recode multiple variables at once")
expect_true(collapse::fnrow(recode_df) > collapse::fnrow(test_df),  info = "Recode multiple variables at once")


set_no_print()

#' Join Multiple Data Frames In One Go
#'
#' @description
#' Join two or more data frames together in one operation. [multi_join()] can handle
#' multiple different join methods and can join on differently named variables.
#'
#' @param data_frames A list of data frames to join together. The second and all
#' following data frames will be joined on the first one.
#' @param on The key variables on which the data frames should be joined. If a
#' character vector is provided, the function assumes all the variables are in every
#' data frame. To join on different  variable names a list of character vectors has
#' to be provided.
#' @param how A character vector containing the join method names. Available methods are:
#' left, right, inner, full, outer, left_inner and right_inner.
#' @param keep_indicators FALSE by default. If TRUE, a variable for each data frame
#' is created, which indicates whether a data frame provides values.
#' @param monitor FALSE by default. If TRUE, outputs two charts to visualize the
#' functions time consumption.
#'
#' @details
#' [multi_join()] is based on the 'SAS' Data-Step function Merge. Merge is capable of
#' joining multiple data sets together at once, with a very basic syntax.
#'
#' Provide the dataset names, the variables, on which they should be joined and after
#' a full join is complete, the user can decide which parts of the joins should remain
#' in the final dataset.
#'
#' [multi_join()] tries to keep the simplicity, while giving the user the power, to
#' do more joins at the same time. Additionally to what Merge can do, this function
#' also makes use of the Proc SQL possibility to join datasets on different variable
#' names.
#'
#' @return
#' Returns a single data frame with joined variables from all given data frames.
#'
#' @examples
#' # Example data frames
#' df1 <- data.frame(key = c(1, 1, 1, 2, 2, 2),
#'                   a   = c("a", "a", "a", "a", "a", "a"))
#'
#' df2 <- data.frame(key = c(2, 3),
#'                   b   = c("b", "b"))
#'
#' # See all different joins in action
#' join_methods <- c("left", "right", "inner", "full", "outer", "left_inner", "right_inner")
#' joined_data  <- list()
#'
#' for (method in seq_along(join_methods)){
#'     joined_data[[method]] <- multi_join(list(df1, df2),
#'                                         on  = "key",
#'                                         how = join_methods[[method]])
#' }
#'
#' # Left join on more than one key
#' df1b <- data.frame(key1 = c(1, 1, 1, 2, 2, 2),
#'                    key2 = c("a", "a", "a", "a", "a", "a"),
#'                    a    = c("a", "a", "a", "a", "a", "a"))
#'
#' df2b <- data.frame(key1 = c(2, 3),
#'                    key2 = c("a", "a"),
#'                    b    = c("b", "b"))
#'
#' left_joined <- multi_join(list(df1b, df2b), on = c("key1", "key2"))
#'
#' # Join more than two data frames
#' df3 <- data.frame(key = c(1, 2),
#'                   c   = c("c", "c"))
#'
#' multiple_joined <- multi_join(list(df1, df2, df3), on = "key")
#'
#' # You can also use different methods for each join
#' multiple_joined2 <- multi_join(list(df1, df3, df2),
#'                                on  = "key",
#'                                how = c("left", "right"))
#'
#' # Joining on different variable names
#' df1c <- data.frame(key1 = c(1, 1, 1, 2, 2, 2),
#'                    key2 = c("a", "a", "a", "a", "a", "a"),
#'                    a    = c("a", "a", "a", "a", "a", "a"))
#'
#' df2c <- data.frame(var1 = c(2, 3),
#'                    var2 = c("a", "a"),
#'                    b    = c("b", "b"))
#'
#' df3c <- data.frame(any  = c(1, 2),
#'                    name = c("a", "a"),
#'                    c    = c("c", "c"))
#'
#' multiple_joined3 <- multi_join(list(df1c, df2c, df3c),
#'                                on = list(df1c = c("key1", "key2"),
#'                                          df2c = c("var1", "var2"),
#'                                          df3c = c("any", "name")))
#'
#' @export
multi_join <- function(data_frames,
                       on,
                       how             = "left",
                       keep_indicators = FALSE,
                       monitor         = .qol_options[["monitor"]]){
    # Measure the time
    print_start_message()
    print_step("GREY", "Error handling")

    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Error handling", "Preparation")
    #-------------------------------------------------------------------------#

    # Check if a valid list of data frames is given
    if (!all(vapply(data_frames, is.data.frame, logical(1)))){
        print_message("ERROR", "Data frames must be provided as a list. Join will be aborted.")
        return(invisible(NULL))
    }

    if (length(data_frames) < 2L){
        print_message("ERROR", "At least two data frames are required. Join will be aborted.")
        return(invisible(NULL))
    }

    # First convert data frame to data table
    for (i in seq_along(data_frames)){
        if (!data.table::is.data.table(data_frames[[i]])){
            data_frames[[i]] <- data.table::as.data.table(data_frames[[i]])
        }
    }

    # If the user passed an unnamed list, get the data frame names back and set
    # them as names.
    if (!is_named_list(data_frames)){
        # Grab the raw call made to the function
        complete_call <- match.call()

        # Extract the passed arguments of the data_frames parameter and set as names
        names(data_frames) <- sapply(as.list(complete_call[["data_frames"]][-1]), deparse)
    }

    ###########################################################################
    # Error handling
    ###########################################################################

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # If a named list is given, a join with unequal variable names will be performed
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    unequal_names <- FALSE

    if (is.list(on)){
        base_on <- on[[1]]

        # If all list entries have a name
        if (!is.null(names(on)) && all(nzchar(names(on)))){
            # Check if number of list entries matches number of data frames
            if (length(data_frames) != length(on)){
                print_message("ERROR", "Length of <on> doesn't match the number of provided data frames. Join will be aborted.")
                return(invisible(NULL))
            }

            unequal_names <- TRUE
        }
        # If a list entry is missing a name abort
        else{
            print_message("ERROR", c("If all data frames have the same variable names for the <on> variables,",
								     "provide them as a vector instead of a list. For unequal names provide a",
								     "named list. Join will be aborted."))
            return(invisible(NULL))
        }
    }
    else{
        base_on <- on
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Check if all data frames (except the first) one only have unique value
    # combinations for the 'on' variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    for (i in seq_along(data_frames)){
        # Check if provided variable names are in the base data frame
        if (!unequal_names){
            if (!all(on %in% names(data_frames[[i]]))){
                print_message("ERROR", c("Not all <on> variables ([on]) appear in data frame [name].",
										 "Join will be aborted."), on = on, name = i)
                return(invisible(NULL))
            }
        }
        else{
            if (!all(on[[i]] %in% names(data_frames[[i]]))){
                print_message("ERROR", c("Not all <on> variables ([on]) appear in data frame [name].",
										 "Join will be aborted."), on = on, name = i)
                return(invisible(NULL))
            }
        }

        # Skip first data frame. This is the only one that is allowed to have duplicate combinations.
        if (i == 1){
            next
        }

        # On equal names there is just one combination which needs to be checked
        if (!unequal_names){
            # Check for duplicate combinations
            if (collapse::any_duplicated(data_frames[[i]][on])){
                print_message("ERROR", c("The second and all following data frames need to have unique combinations",
										 "in the provided <on> variables. Join will be aborted."))
                return(invisible(NULL))
            }
        }
        # On unequal names each individual variable combination has to be checked on the corresponding data frame
        else{
            if (collapse::any_duplicated(data_frames[[i]][on[[i]]])){
                print_message("ERROR", c("The second and all following data frames need to have unique combinations",
										 "in the provided <on> variables. Join will be aborted."))
                return(invisible(NULL))
            }
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Join methods
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Only keep valid join methods
    valid_join_methods <- c("left", "right", "inner", "full", "outer", "left_inner", "right_inner")
    valid_methods <- tolower(how) %in% valid_join_methods
    invalid_how   <- how[!valid_methods]
    how           <- how[valid_methods]

    if (length(invalid_how) > 0){
        print_message("WARNING", "The provided join method '[invalid]' is not valid.", invalid = invalid_how)
    }

    # If length of provided joins is lesser than number of data frames minus one
    if (length(how) == 0){
        print_message("WARNING", "No valid join method provided, 'left' will be used.")

        how <- "left"
    }
    # If length of provided joins is lesser than number of data frames minus one
    else if (length(how) < length(data_frames) - 1){
        # Repeat the last vector element until number of data frames minus one is reached
        how <- c(how, rep(utils::tail(how, 1), length(data_frames) - 1))

        # I am not printing the warning here, so one can be lazy on the function call. Meaning one can just
        # input one join method, if it is the same for all joins.
        # print_message("WARNING", "Not enough join methods in 'how' given. Last join method will be repeated.")
    }
    # If length of provided joins is greater number of data frames minus one
    else if (length(how) > length(data_frames) - 1){
        # Cut elements down to number of data frames minus one
        how <- utils::head(how, length(data_frames) - 1)

        print_message("NOTE", "Too many join methods given in <how>. Excess methods will remain unused.")
    }

    ###########################################################################
    # Join starts
    ###########################################################################

    join_keys <- paste0(".", letters)

    print_step("MAJOR", "Begin joining.")

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # First perform joins as a full joins. This way all combinations are in the
    # final data frame. Afterwards the data frame is conditionally filtered.
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    base_df_name <- names(data_frames)[1]

    # Start the actual joining
    for (i in seq_along(data_frames)){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next(paste0("Join ", i - 1), "Join")
        #---------------------------------------------------------------------#

        # Skip first data frame, because it is not joined on itself
        if (i == 1){
            joined_df <- data_frames[[i]]
            joined_df[[join_keys[[i]]]] <- 1

            next
        }

        to_join_df  <- data_frames[[i]]
        join_method <- how[[i - 1]]

        # Depending on the join the printed data frame names have to be swapped
        if (!join_method %in% c("right", "inner_right")){
            print_step("MINOR", "{how} joining [data_frame1] to [data_frame2].",
                       how = join_method, data_frame1 = names(data_frames)[i], data_frame2 = base_df_name)
        }
        else{
            print_step("MINOR", "{how} joining [data_frame1] to [data_frame2].",
                       how = join_method, data_frame1 = base_df_name, data_frame2 = names(data_frames)[i])
        }

        # Create filter variable which indicates, which data frame provides observations
        to_join_df[[join_keys[[i]]]] <- 1

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Equality crossroads
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Get the variable names of the base data frame and the one to be joined
        to_join_df_name     <- names(data_frames)[i]
        joined_df_var_names <- names(joined_df)
        to_join_var_names   <- names(to_join_df)

        # On equal names there is just one combination which can be input directly
        if (!unequal_names){
            join_variables <- base_on
        }
        # On unequal names each individual variable combination has to be checked on the corresponding data frame
        else{
            to_join_on <- on[[i]]

            # Check if the same number of 'on' variables are provided
            if (length(base_on) != length(to_join_on)){
                print_message("ERROR", c("Unequal number of <on> variables provided: [on1] vs [on_i].",
                                         "Join will be aborted."), on1 = base_on, on_i = to_join_on)
                return(invisible(NULL))
            }

            join_variables  <- stats::setNames(to_join_on, base_on)
        }

        # Check for intersecting variable names and clear them before the join,
        # because otherwise it can happen that multiple variables with the exact
        # same name get created through the join below.
        joined_df_var_names <- joined_df_var_names[!joined_df_var_names %in% on]
        to_join_var_names   <- to_join_var_names[!to_join_var_names %in% on]

        duplicate_names <- intersect(joined_df_var_names, to_join_var_names)

        if (length(duplicate_names) > 0){
            # On other than right joins clear from to right data frame
            if (!join_method %in% c("right", "inner_right")){
                print_message("WARNING", c("Duplicate variable names found: [duplicates].",
                                           "These variables will be removed from '[data_frame]' before the join."),
                              duplicates = duplicate_names, data_frame = to_join_df_name)
            }
            # On right join clear from the left (base) data frame
            else{
                print_message("WARNING", c("Duplicate variable names found: [duplicates].",
                                           "These variables will be removed from '[data_frame]' before the join."),
                              duplicates = duplicate_names, data_frame = base_df_name)
            }

            to_join_df <- to_join_df |> dropp(duplicate_names)
        }

        # Perform the actual join
        joined_df <- collapse::join(joined_df, to_join_df,
                                    on      = join_variables,
                                    how     = "full",
                                    verbose = FALSE,
									overid  = 2)

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Subset data frame according to provided join methods
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next(paste0("Subset ", i - 1), "Join")
        #---------------------------------------------------------------------#

        if (tolower(join_method) == "left"){
            joined_df <- joined_df |> collapse::fsubset(joined_df[[join_keys[[1]]]] == 1)
        }
        else if (tolower(join_method) == "right"){
            joined_df <- joined_df |> collapse::fsubset(joined_df[[join_keys[[i]]]] == 1)
        }
        else if (tolower(join_method) == "inner"){
            joined_df <- joined_df |> collapse::fsubset(joined_df[[join_keys[[1]]]] == 1 & joined_df[[join_keys[[i]]]] == 1)
        }
        else if (tolower(join_method) == "outer"){
            joined_df <- joined_df |> collapse::fsubset(is.na(joined_df[[join_keys[[1]]]]) | is.na(joined_df[[join_keys[[i]]]]))
        }
        else if (tolower(join_method) == "left_inner"){
            joined_df <- joined_df |> collapse::fsubset(joined_df[[join_keys[[1]]]] == 1 & is.na(joined_df[[join_keys[[i]]]]))
        }
        else if (tolower(join_method) == "right_inner"){
            joined_df <- joined_df |> collapse::fsubset(is.na(joined_df[[join_keys[[1]]]]) & joined_df[[join_keys[[i]]]] == 1)
        }

        # Drop indicator of joined data frame
        if (!keep_indicators){
            key_to_drop <- as.character(join_keys[[i]])
            joined_df   <- joined_df |> dropp(key_to_drop)
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Clean up
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Finish join ", "Join")
    #-------------------------------------------------------------------------#

    # Drop indicator of base data frame
    if (!keep_indicators){
        key_to_drop <- as.character(join_keys[[1]])
        joined_df <- joined_df |> dropp(key_to_drop)
    }
    # If join indicators should stay in the data frame, sort them to the back
    else{
        joined_df <- joined_df |> setcolorder_by_pattern("^\\.")
    }

    print_closing()

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_end()
    monitor_df |> monitor_plot(draw_plot = monitor)
    #-------------------------------------------------------------------------#

    joined_df
}

#' Set Global Styling Options For Excel Workbooks
#'
#' @name style_options
#'
#' @description
#' Modify Styling options for Excel workbooks. Available parameters can be seen in
#' [excel_output_style()] or [number_format_style()].
#'
#' [set_style_options()] sets the styling options for Excel workbooks globally.These
#' options are used by all tabulation and output functions, which are capable of
#' exporting styled outputs.
#'
#' @param ... Put in any styling option from [excel_output_style()] or [number_format_style()]
#' with the new value.
#' @param save_file A full file path to an RDS file in which global style options
#' should be stored.
#'
#' @return
#' [set_style_options()]: Returns modified global styling options.
#'
#' @seealso
#' Functions that use global styling options: [any_table()], [frequencies()],
#' [crosstabs()], [export_with_style()].
#'
#' @examples
#' # This function can process any parameter from excel_output_style() or
#' # number_format_style() and sets the option globally.
#' set_style_options(save_path    = "C:/My Projects/",
#'                   sum_decimals = 8)
#'
#' @rdname style_options
#'
#' @export
set_style_options <- function(..., save_file = NULL){
    # Translate ... into a list if possible
    style_list <- tryCatch({
        # Force evaluation to see if it exists
        list(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(style_list)){
        print_message("ERROR", c("Unknown object found. See 'excel_output_style()' and 'number_format_style()'",
								 "for valid function parameters. Global style remains unchanged."))
        return(invisible(.qol_options[["excel_style"]]))
    }

    if (length(style_list) == 0){
        print_message("ERROR", c("Empty list found. See 'excel_output_style()' and 'number_format_style()'",
								 "for valid function parameters. Global style remains unchanged."))
        return(invisible(.qol_options[["excel_style"]]))
    }

    # Define expected types for validation
    number_numerics <- c("pct_decimals", "freq_decimals", "freq.g0_decimals", "sum_decimals", "sum.wgt_decimals",
                         "mean_decimals", "median_decimals", "mode_decimals", "min_decimals", "max_decimals",
                         "sd_decimals", "variance_decimals", "first_decimals", "last_decimals", "p_decimals",
                         "missing_decimals")

    number_characters <- c("pct_excel", "freq_excel", "freq.g0_excel", "sum_excel", "sum.wgt_excel", "mean_excel",
                           "median_excel", "mode_excel", "min_excel", "max_excel", "sd_excel", "variance_excel",
                           "first_excel", "last_excel", "p_excel", "missing_excel")

    logicals <- c("freeze_col_header", "freeze_row_header", "filters", "grid_lines", "by_as_subheaders", "header_font_bold",
                  "header_borders", "cat_col_font_bold", "cat_col_borders", "table_font_bold", "table_borders",
                  "box_font_bold", "box_borders", "title_font_bold", "footnote_font_bold", "as_heatmap")

    numerics <- c("start_row", "start_column", "header_font_size", "header_indent", "subheader_font_size", "subheader_indent",
                  "cat_col_font_size", "cat_col_indent", "table_font_size", "table_indent",
                  "box_font_size", "box_indent", "title_font_size", "footnote_font_size",
                  "title_heights", "header_heights", "subheader_heights", "table_heights", "footnote_heights",
                  "column_widths", "row_heights")

    characters  <- c("sheet_name", "font", "header_alignment", "header_wrap", "subheader_alignment", "subheader_wrap",
                     "cat_col_alignment", "cat_col_wrap", "table_alignment", "box_alignment", "box_wrap", "title_alignment",
                     "footnote_alignment", "na_symbol", "save_path", "file", "header_stat_merging")

    colors <- c("header_back_color", "header_font_color", "header_border_color", "subheader_back_color", "subheader_font_color",
                "subheader_border_color", "cat_col_back_color", "cat_col_font_color", "cat_col_border_color", "table_back_color",
                "table_font_color", "table_border_color", "box_back_color", "box_font_color", "box_border_color", "title_font_color",
                "footnote_font_color", "heatmap_low_color", "heatmap_middle_color", "heatmap_high_color", "background_color")

    # Loop through passed arguments and check if they are of valid type
    for (style_option in names(style_list)){
        value <- style_list[[style_option]]

        if (style_option %in% logicals && !all(is.logical(value))){
            print_message("WARNING", "'[style_option]' must be <logical>. Option will be omitted.", style_option = style_option)
            style_list[[style_option]] <- NULL
        }
        else if (style_option %in% c(numerics, number_numerics) && !all(is.numeric(value))){
            print_message("WARNING", "'[style_option]' must be <numeric>. Option will be omitted.", style_option = style_option)
            style_list[[style_option]] <- NULL
        }
        else if (style_option %in% c(characters, number_characters) && !all(is.character(value))){
            print_message("WARNING", "'[style_option]' must be <character>. Option will be omitted.", style_option = style_option)
            style_list[[style_option]] <- NULL
        }
        else if (style_option %in% colors && !all(grepl("^[A-Fa-f0-9]{6}$", value)) && value != ""){
            print_message("WARNING", "'[style_option]' must be a 6 character <hex code>. Option will be omitted.", style_option = style_option)
            style_list[[style_option]] <- NULL
        }
        else if (!style_option %in% c(number_numerics, number_characters, logicals, numerics, characters, colors)){
            print_message("WARNING", c("'[style_option]' is not a valid style option. See 'excel_output_style()' and 'number_format_style()'",
                                       "for valid function parameters. Option will be omitted."), style_option = style_option)
            style_list[[style_option]] <- NULL
        }
    }

    # Split up the list into the number formats and the rest. The number formats are a
    # list inside the list of style elements. So the number format list is added as full
    # list back again
    number_fmt_list <- style_list[names(style_list) %in% c(number_characters, number_numerics)]
    style_list      <- style_list[setdiff(names(style_list), names(number_fmt_list))]

    if (!length(number_fmt_list) == 0){
        style_list[["number_formats"]] <- number_fmt_list
    }

    # Update the internal state
    if (length(style_list) > 0){
        .qol_options[["excel_style"]] <- utils::modifyList(.qol_options[["excel_style"]], style_list)

        print_message("NOTE", "Global style options successfully changed.")
    }

    # Save global style options as physical file
    if (!is.null(save_file)){
        if (!file.exists(dirname(save_file))){
            print_message("ERROR", "Path does not exist: [path]", path = save_file)
        }
        else{
            extension <- tolower(tools::file_ext(save_file))

            if (extension == ""){
                save_file <- paste0(save_file, ".rds")
                extension <- "rds"
            }

            if (!extension == "rds"){
                save_file <- gsub(extension, "rds", save_file)
            }

            saveRDS(.qol_options[["excel_style"]], file = save_file)

            print_message("NOTE", "Global style has been saved to: [save_file]",
                          save_file = save_file)
        }
    }

    invisible(.qol_options[["excel_style"]])
}


#' Get Global Styling Options For Excel Workbooks
#'
#' @description
#' [get_style_options()] Prints out the currently set global styling options.
#'
#' @param from_file A full file path to an RDS file in which global style options
#' are stored.
#'
#' @return
#' [get_style_options()]: List of global styling options.
#'
#' @examples
#' get_style_options()
#'
#' @rdname style_options
#'
#' @export
get_style_options <- function(from_file = NULL){
    # If a saved file is specified, check if it exists and read it in
    if (!is.null(from_file)){
        if (!file.exists(from_file)){
            print_message("ERROR", "File does not exist: [file]", file = from_file)
        }
        else{
            style_list <- readRDS(file = from_file)
            set_style_options(style_list)
        }
    }

    # To make the console output more readable the list will be formatted before
    # outputting it. First expand the number format list inside the excel style
    # list, to be able to handle all formatting in one go.
    style_list <- c(.qol_options[["excel_style"]], unlist(.qol_options[["excel_style"]]["number_formats"]))
    style_list[["number_formats"]] <- NULL
    style_names <- gsub("\\.", " - ", names(style_list))

    # Add padding to names to make the values display in one column
    padded_names <- sprintf(paste0("%-", max(nchar(style_names)), "s"), style_names)

    # Join complete lines and print
    output_vector <- paste0(padded_names, " : ", style_list)

    print_message("NEUTRAL", output_vector)

    # Return style list
    invisible(.qol_options[["excel_style"]])
}


#' Reset Global Styling Options For Excel Workbooks
#'
#' @description
#' [reset_style_options()] Resets global style options to the default parameters.
#' This includes all options set with [set_style_options()], [set_labels()],
#' [set_titles()] and [set_footnotes()].
#'
#' @return
#' [reset_style_options()]: Returns default global styling options.
#'
#' @examples
#' # Reset all the style options including variable and statistic labels as well
#' # as titles and footnotes.
#' reset_style_options()
#'
#' @rdname style_options
#'
#' @export
reset_style_options <- function(){
    .qol_options[["excel_style"]] <- excel_output_style()
    .qol_options[["var_labels"]]  <- list()
    .qol_options[["stat_labels"]] <- list()
    .qol_options[["titles"]]      <- c()
    .qol_options[["footnotes"]]   <- c()

    print_message("NOTE", c("Style options have been reset, this includes:",
                            "Everything from excel_output_style(), variable and statistic labels,",
                            "as well as titles and footnotes"))

    invisible(.qol_options)
}


#' Set File To NULL In Global Styling Options For Excel Workbooks
#'
#' @description
#' [close_file()] is a simple, more readable wrapper for setting file parameter to NULL.
#'
#' @return
#' [close_file()]: List of global styling options with file = NULL.
#'
#' @examples
#' # Reset the file parameter
#' close_file()
#'
#' @rdname style_options
#'
#' @export
close_file <- function(){
    print_message("NOTE", "File '[file]' has been closed.", file = .qol_options[["excel_style"]][["file"]])

    .qol_options[["excel_style"]][["file"]] <- NULL

    invisible(.qol_options[["excel_style"]][["file"]])
}


#' Set Global Variable Labels
#'
#' @description
#' [set_labels()]: Can set variable and statistic labels globally so that they don't
#' have to be provided in every output function separately.
#'
#' @param ... [set_labels()]: Put in the variable names and their respective labels.
#'
#' @return
#' [set_labels()]: List of variable and statistic labels.
#'
#' @seealso
#' Functions that use global variable and statistic labels: [any_table()], [frequencies()],
#' [crosstabs()].
#'
#' Functions that also use global variable labels: [export_with_style()].
#'
#' @examples
#' # Set variable and statistic labels globally, retrieve and reset them
#' set_labels(age_gr = "Group of ages",
#'            status = "Current status",
#'            pct    = "%")
#'
#' # To reset labels do this
#' set_labels(NULL)
#'
#' @rdname style_options
#'
#' @export
set_labels <- function(...){
    # Translate ... into a list if possible
    label_list <- tryCatch({
        # Force evaluation to see if it exists
        list(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(label_list)){
        .qol_options[["var_labels"]]  <- list()
        .qol_options[["stat_labels"]] <- list()

        print_message("NOTE", "Global labels have been removed.")

        return(invisible(list("var_labels"  = .qol_options[["var_labels"]],
                              "stat_labels" = .qol_options[["stat_labels"]])))
    }

    if (is.null(names(label_list))){
        print_message("ERROR", "List must have names to set labels. Global labels remain unchanged.")

        return(invisible(list("var_labels"  = .qol_options[["var_labels"]],
                              "stat_labels" = .qol_options[["stat_labels"]])))
    }

    if (length(label_list) == 0){
        print_message("ERROR", "Empty list found. Global labels remain unchanged.")
        return(invisible(list("var_labels"  = .qol_options[["var_labels"]],
                              "stat_labels" = .qol_options[["stat_labels"]])))
    }

    # Define explicit list of statistic names. Percentile names will be checked
    # inside the loop on the fly.
    statistics <- c("sum", "sum_wgt", "freq", "freq_g0", "pct", "mean", "median",
                    "mode", "min", "max", "sd", "variance", "first", "last", "pn",
                    "missing")

    # Check which label type is passed in the list and sort it to the according
    # label list.
    if (length(.qol_options[["var_labels"]]) == 0){
        .qol_options[["var_labels"]] <- list()
    }
    if (length(.qol_options[["stat_labels"]]) == 0){
        .qol_options[["stat_labels"]] <- list()
    }

    for (i in seq_len(length(label_list))){
        label <- names(label_list)[i]

        # Check if the name is in the explicit list OR matches the pattern p1 to p99
        is_statistic_label <- (label %in% statistics) || grepl("^p[1-9][0-9]?$", label)

        if (is_statistic_label){
            .qol_options[["stat_labels"]][[label]] <- label_list[[i]]
        }
        else{
            .qol_options[["var_labels"]][[label]] <- label_list[[i]]
        }
    }

    print_message("NOTE", "Labels where succesfully added to global label lists.")

    invisible(list("var_labels"  = .qol_options[["var_labels"]],
                   "stat_labels" = .qol_options[["stat_labels"]]))
}


#' Get Global Variable Labels
#'
#' @description
#' [get_labels()]: Get the globally stored variable and statistic labels as a list
#' and print the contents to the console.
#'
#' @return
#' [get_labels()]: List of variable and statistic labels.
#'
#' @examples
#' get_labels()
#'
#' @rdname style_options
#'
#' @export
get_labels <- function(){
    # To make the console output more readable the list will be formatted before
    # outputting it. First expand the number format list inside the excel style
    # list, to be able to handle all formatting in one go.
    label_list <- c(.qol_options[["var_labels"]], unlist(.qol_options[["stat_labels"]]))

    if (length(label_list) > 0){
        label_names <- names(label_list)

        # Add padding to names to make the values display in one column
        padded_names <- sprintf(paste0("%-", max(nchar(label_names)), "s"), label_names)

        # Join complete lines and print
        output_vector <- paste0(padded_names, " : ", label_list)

        print_message("NEUTRAL", output_vector)
    }

    # Return style list
    invisible(list("var_labels"  = .qol_options[["var_labels"]],
                   "stat_labels" = .qol_options[["stat_labels"]]))
}


#' Set Global Print Option
#'
#' @name qol_options
#'
#' @description
#' [set_print()]: Set the print option globally for the tabulation and export to
#' Excel functions.
#'
#' @param ... [set_print()]: Put in TRUE or FALSE to activate or deactivate the option.
#'
#' @return
#' [set_print()]: Changed global print option.
#'
#' @examples
#' # Set and get the global print option
#' set_print(FALSE)
#' set_print(TRUE)
#'
#' @rdname qol_options
#'
#' @export
set_print <- function(...){
    # Translate ... into a list if possible
    print_option <- tryCatch({
        # Force evaluation to see if it exists
        unlist(list(...))
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(print_option)){
        print_message("ERROR", "Unknown object found. Global option remains unchanged.")
        return(invisible(.qol_options[["print"]]))
    }

    if (!is.logical(print_option)){
        print_message("ERROR", "Print option can only be TRUE or FALSE. Global option remains unchanged.")
        return(invisible(.qol_options[["print"]]))
    }

    .qol_options[["print"]] <- print_option

    print_message("NOTE", "Global print option is now [option].", option = print_option)

    invisible(.qol_options[["print"]])
}


#' Get Global Print Option
#'
#' @description
#' [get_print()]: Get the globally stored print option.
#'
#' @return
#' [get_print()]: TRUE or FALSE.
#'
#' @examples
#' get_print()
#'
#' @rdname qol_options
#'
#' @export
get_print <- function(){
    .qol_options[["print"]]
}


#' Set Global Monitor Option
#'
#' @description
#' [set_monitor()]: Set the monitor option globally for the heavier functions which are
#' able to show how they work internally.
#'
#' @param ... [set_monitor()]: Put in TRUE or FALSE to activate or deactivate the option.
#'
#' @return
#' [set_monitor()]: Changed global monitor option.
#'
#' @examples
#' # Set and get the global monitor option
#' set_monitor(TRUE)
#' set_monitor(FALSE)
#'
#' @rdname qol_options
#'
#' @export
set_monitor <- function(...){
    # Translate ... into a list if possible
    monitor_option <- tryCatch({
        # Force evaluation to see if it exists
        unlist(list(...))
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(monitor_option)){
        print_message("ERROR", "Unknown object found. Global option remains unchanged.")
        return(invisible(.qol_options[["monitor"]]))
    }

    if (!is.logical(monitor_option)){
        print_message("ERROR", "Monitor option can only be TRUE or FALSE. Global option remains unchanged.")
        return(invisible(.qol_options[["monitor"]]))
    }

    .qol_options[["monitor"]] <- monitor_option

    print_message("NOTE", "Global monitor option is now [option].", option = monitor_option)

    invisible(.qol_options[["monitor"]])
}


#' Get Global Monitor Option
#'
#' @description
#' [get_monitor()]: Get the globally stored monitor option.
#'
#' @return
#' [get_monitor()]: TRUE or FALSE.
#'
#' @examples
#' get_monitor()
#'
#' @rdname qol_options
#'
#' @export
get_monitor <- function(){
    .qol_options[["monitor"]]
}


#' Set Global NA Removal Option
#'
#' @description
#' [set_na.rm()]: Set the na.rm option globally for each function which can remove
#' NA values.
#'
#' @param ... [set_na.rm()]: Put in TRUE or FALSE to activate or deactivate the option.
#'
#' @return
#' [set_na.rm()]: Changed global na.rm option.
#'
#' @examples
#' # Set and get the global NA removal option
#' set_na.rm(TRUE)
#' set_na.rm(FALSE)
#'
#' @rdname qol_options
#'
#' @export
set_na.rm <- function(...){
    # Translate ... into a list if possible
    na_option <- tryCatch({
        # Force evaluation to see if it exists
        unlist(list(...))
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(na_option)){
        print_message("ERROR", "Unknown object found. Global option remains unchanged.")
        return(invisible(.qol_options[["na.rm"]]))
    }

    if (!is.logical(na_option)){
        print_message("ERROR", "NA removal option can only be TRUE or FALSE. Global option remains unchanged.")
        return(invisible(.qol_options[["na.rm"]]))
    }

    .qol_options[["na.rm"]] <- na_option

    print_message("NOTE", "Global NA removal option is now [option].", option = na_option)

    invisible(.qol_options[["na.rm"]])
}


#' Get Global NA Removal Option
#'
#' @description
#' [get_na.rm()]: Get the globally stored na.rm option.
#'
#' @return
#' [get_na.rm()]: TRUE or FALSE.
#'
#' @examples
#' get_na.rm()
#'
#' @rdname qol_options
#'
#' @export
get_na.rm <- function(){
    .qol_options[["na.rm"]]
}


#' Set Global Print Missing Categories Option
#'
#' @description
#' [set_print_miss()]: Set the print_miss option globally for each function which can display
#' missing categories.
#'
#' @param ... [set_print_miss()]: Put in TRUE or FALSE to activate or deactivate the option.
#'
#' @return
#' [set_print_miss()]: Changed global print_miss option.
#'
#' @examples
#' # Set and get the global print missing option
#' set_print_miss(TRUE)
#' set_print_miss(FALSE)
#'
#' @rdname qol_options
#'
#' @export
set_print_miss <- function(...){
    # Translate ... into a list if possible
    print_miss_option <- tryCatch({
        # Force evaluation to see if it exists
        unlist(list(...))
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(print_miss_option)){
        print_message("ERROR", "Unknown object found. Global option remains unchanged.")
        return(invisible(.qol_options[["print_miss"]]))
    }

    if (!is.logical(print_miss_option)){
        print_message("ERROR", "Print missing categories option can only be TRUE or FALSE. Global option remains unchanged.")
        return(invisible(.qol_options[["print_miss"]]))
    }

    .qol_options[["print_miss"]] <- print_miss_option

    print_message("NOTE", "Global print missing option is now [option].", option = print_miss_option)

    invisible(.qol_options[["print_miss"]])
}


#' Get Global Print Missing Categories Option
#'
#' @description
#' [get_print_miss()]: Get the globally stored print_miss option.
#'
#' @return
#' [get_print_miss()]: TRUE or FALSE.
#'
#' @examples
#' get_print_miss()
#'
#' @rdname qol_options
#'
#' @export
get_print_miss <- function(){
    .qol_options[["print_miss"]]
}


#' Set Global Output Option
#'
#' @description
#' [set_output()]: Set the output option globally for each function that can output
#' results to "console", "text", "excel" or "excel_nostyle".
#'
#' @param ... [set_print_miss()]: Input option to set global output style.
#'
#' @return
#' [set_output()]: Changed global output option.
#'
#' @examples
#' # Set and get the global output option
#' set_output("excel")
#'
#' @rdname qol_options
#'
#' @export
set_output <- function(...){
    # Translate ... into a list if possible
    output_option <- tryCatch({
        # Force evaluation to see if it exists
        unlist(list(...))
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(output_option)){
        print_message("ERROR", "Unknown object found. Global option remains unchanged.")
        return(invisible(.qol_options[["output"]]))
    }

    if (!is.character(output_option)){
        print_message("ERROR", "Output can only be 'console', 'text', 'excel' or 'excel_nostyle'. Global option remains unchanged.")
        return(invisible(.qol_options[["output"]]))
    }

    if (!tolower(output_option) %in% c("console", "text", "excel", "excel_nostyle")){
        print_message("ERROR", "Output can only be 'console', 'text', 'excel' or 'excel_nostyle'. Global option remains unchanged.")
        return(invisible(.qol_options[["output"]]))
    }

    .qol_options[["output"]] <- tolower(output_option)

    print_message("NOTE", "Global output option is now [option].", option = .qol_options[["output"]])

    invisible(.qol_options[["output"]])
}


#' Get Global Output Option
#'
#' @description
#' [get_output()]: Get the globally stored output option.
#'
#' @return
#' [get_output()]: Current output option as character.
#'
#' @examples
#' get_output()
#'
#' @rdname qol_options
#'
#' @export
get_output <- function(){
    .qol_options[["output"]]
}


#' Set Global Titles
#'
#' @description
#' [set_titles()]: Set the titles globally for each function that can print titles
#' above the output table.
#'
#' @param ... [set_titles()]: Put in the titles that should appear above tables.
#'
#' @return
#' [set_titles()]: Changed global titles.
#'
#' @examples
#' # Set and get table titles globally
#' set_titles("This is title number 1 link: https://cran.r-project.org/",
#'            "This is title number 2 cell: W22",
#'            "This is title number 3 file: C:/MyFolder/MyFile.docx",
#'            "This is title number 4")
#'
#' # To reset titles do
#' set_titles(NULL)
#'
#' @rdname qol_options
#'
#' @export
set_titles <- function(...){
    # Translate ... into a list if possible
    titles_option <- tryCatch({
        # Force evaluation to see if it exists
        unlist(list(...))
    }, error = function(e){
        # Evaluation failed
        "ERROR"
    })

    if (is.null(titles_option)){
        .qol_options[["titles"]] <- c()

        print_message("NOTE", "Global titles have been removed.")

        return(invisible(.qol_options[["titles"]]))
    }

    if (length(titles_option) == 1 && titles_option == "ERROR"){
        print_message("ERROR", "Unknown object found. Global titles remain unchanged.")
        return(invisible(.qol_options[["titles"]]))
    }

    if (!is.character(titles_option)){
        print_message("ERROR", "Titles must be provided as character. Global titles remain unchanged.")
        return(invisible(.qol_options[["titles"]]))
    }

    .qol_options[["titles"]] <- titles_option

    print_message("NOTE", "Global titles have been successfully set.")

    invisible(.qol_options[["titles"]])
}


#' Get Global Titles
#'
#' @description
#' [get_titles()]: Get the globally stored titles.
#'
#' @return
#' [get_titles()]: Current titles as character.
#'
#' @examples
#' get_titles()
#'
#' @rdname qol_options
#'
#' @export
get_titles <- function(){
    print_message("NEUTRAL", c("", .qol_options[["titles"]], ""))

    invisible(.qol_options[["titles"]])
}


#' Set Global Footnotes
#'
#' @description
#' [set_footnotes()]: Set the footnotes globally for each function that can print footnotes
#' above the output table.
#'
#' @param ... [set_footnotes()]: Put in the footnotes that should appear below tables.
#'
#' @return
#' [set_footnotes()]: Changed global footnotes.
#'
#' @examples
#' # Set and get table footnotes globally
#' set_footnotes("This is footnote number 1 link: https://cran.r-project.org/",
#'               "This is footnote number 2 cell: W22",
#'               "This is footnote number 3 file: C:/MyFolder/MyFile.docx",
#'               "This is footnote number 4")
#'
#' # To reset footnotes do
#' set_footnotes(NULL)
#'
#' @rdname qol_options
#'
#' @export
set_footnotes <- function(...){
    # Translate ... into a list if possible
    footnotes_option <- tryCatch({
        # Force evaluation to see if it exists
        unlist(list(...))
    }, error = function(e){
        # Evaluation failed
        "ERROR"
    })

    if (is.null(footnotes_option)){
        .qol_options[["footnotes"]] <- c()

        print_message("NOTE", "Global footnotes have been removed.")

        return(invisible(.qol_options[["footnotes"]]))
    }

    if (length(footnotes_option) == 1 && footnotes_option == "ERROR"){
        print_message("ERROR", "Unknown object found. Global footnotes remain unchanged.")
        return(invisible(.qol_options[["footnotes"]]))
    }

    if (!is.character(footnotes_option)){
        print_message("ERROR", "Footnotes must be provided as character. Global footnotes remain unchanged.")
        return(invisible(.qol_options[["footnotes"]]))
    }

    .qol_options[["footnotes"]] <- footnotes_option

    print_message("NOTE", "Global footnotes have been successfully set.")

    invisible(.qol_options[["footnotes"]])
}


#' Get Global Footnotes
#'
#' @description
#' [get_footnotes()]: Get the globally stored footnotes.
#'
#' @return
#' [get_footnotes()]: Current footnotes as character.
#'
#' @examples
#' get_footnotes()
#'
#' @rdname qol_options
#'
#' @export
get_footnotes <- function(){
    print_message("NEUTRAL", c("", .qol_options[["footnotes"]], ""))

    invisible(.qol_options[["footnotes"]])
}


#' Set Global Number Of Used Threads
#'
#' @description
#' [set_threads()]: Globally sets the number of used threads for the save and load
#' file functions.
#'
#' @param ... [set_threads()]: Put in the number of threads to use or NULL to reset.
#'
#' @return
#' [set_threads()]: Changed global number of used threads.
#'
#' @examples
#' # Set and get used threads globally
#' set_threads(8)
#'
#' @rdname qol_options
#'
#' @export
set_threads <- function(...){
    # Translate ... into a list if possible
    threads_option <- tryCatch({
        # Force evaluation to see if it exists
        unlist(list(...))
    }, error = function(e){
        # Evaluation failed
        "ERROR"
    })

    if (is.null(threads_option)){
        .qol_options[["threads"]] <- fst::threads_fst(NULL)
        return(invisible(.qol_options[["threads"]]))
    }

    if (length(threads_option) == 1 && threads_option == "ERROR"){
        print_message("ERROR", "Unknown object found. Global number of used threads remains unchanged.")
        return(invisible(.qol_options[["threads"]]))
    }

    if (!is.numeric(threads_option)){
        print_message("ERROR", "Number of used threads must be provided as integer value. Global number of used threads remains unchanged.")
        return(invisible(.qol_options[["threads"]]))
    }

    .qol_options[["threads"]] <- as.integer(threads_option)

    print_message("NOTE", "Global number of used threads is now [option].", option = .qol_options[["threads"]])

    invisible(.qol_options[["threads"]])
}


#' Get Global Number Of Used Threads
#'
#' @description
#' [get_threads()]: Get the globally stored number of used threads.
#'
#' @return
#' [get_threads()]: Current number of used threads.
#'
#' @examples
#' get_threads()
#'
#' @rdname qol_options
#'
#' @export
get_threads <- function(){
    .qol_options[["threads"]]
}


#' Reset Global Options
#'
#' @description
#' [reset_qol_options()] Resets global options to the default parameters.
#'
#' @return
#' [reset_qol_options()]: Returns default global options.
#'
#' @examples
#' # Reset all globally set options
#' reset_qol_options()
#'
#' @rdname style_options
#'
#' @export
reset_qol_options <- function(){
    .qol_options[["print"]]       <- TRUE
    .qol_options[["monitor"]]     <- FALSE
    .qol_options[["na.rm"]]       <- FALSE
    .qol_options[["print_miss"]]  <- FALSE
    .qol_options[["output"]]      <- "console"
    .qol_options[["excel_style"]] <- excel_output_style()
    .qol_options[["var_labels"]]  <- list()
    .qol_options[["stat_labels"]] <- list()
    .qol_options[["titles"]]      <- c()
    .qol_options[["footnotes"]]   <- c()
    .qol_options[["threads"]]     <- suppressMessages(fst::threads_fst(NULL))

    print_message("NOTE", c("All qol package options have been reset."))

    invisible(.qol_options)
}


#' Go To GitHub NEWS Page
#'
#' @description
#' Opens browser and goes to the Github NEWS page
#'
#' @return
#' URL.
#'
#' @export
qol_news <- function(){
    utils::browseURL("https://s3rdia.github.io/qol/news/index.html")
}


#' Go To GitHub NEWS Page
#'
#' @description
#' Opens browser and goes to the DeepWiki page
#'
#' @return
#' URL.
#'
#' @export
qol_chat <- function(){
    utils::browseURL("https://deepwiki.com/s3rdia/qol")
}

set_no_print(TRUE)

dummy_df  <- dummy_data(3000)
dummy_big <- dummy_data(10000)

dummy_df[["binary"]] <- replicate(nrow(dummy_df), {
    paste0(sample(0:1, 2, replace = TRUE), collapse = "")
})

sum_df   <- dummy_df |>
    summarise_plus(class      = c(year, sex),
                   values     = weight,
                   statistics = c("sum"),
                   nesting    = "deepest",
                   na.rm      = TRUE)

sum_df2  <- dummy_df |>
    summarise_plus(class      = c(year, sex, age),
                   values     = weight,
                   statistics = c("sum"),
                   nesting    = "deepest",
                   na.rm      = TRUE)

age. <- discrete_format(
    "under 50"    = 0:49,
    "50 and more" = 50:100)

age2. <- discrete_format(
    "Total"       = 0:100,
    "under 50"    = 0:49,
    "50 and more" = 50:100)

state. <- discrete_format(
    "West" = 1:10,
    "East" = 11:16)

sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

education. <- discrete_format(
    "low"    = "low",
    "middle" = "middle",
    "high"   = "high")


# Simplest form of any_table
result_list <- dummy_big |>
      any_table(rows    = "age",
                columns = "sex",
                values  = weight,
                print   = FALSE)

expect_inherits(result_list, "qol_table", info = "Simplest form of any_table")
expect_equal(length(result_list), 4, info = "Simplest form of any_table")
expect_equal(names(result_list[[1]]), c("row.label", "var1", "weight_sum_1",
                                        "weight_sum_2", "weight_sum_NA"), info = "Simplest form of any_table")
expect_equal(result_list[[1]][["var1"]][1:90], as.character(0:89), info = "Simplest form of any_table")
expect_equal(result_list[[1]][["var1"]][1:90], as.character(0:89), info = "Simplest form of any_table")


# any_table with combinations
result_list <- dummy_df |>
      any_table(rows    = "age + education",
                columns = "sex + year",
                values  = income,
                weight  = weight,
                full_precision = TRUE,
                output  = "excel_nostyle",
                print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with combinations")
expect_equal(length(result_list), 4, info = "any_table with combinations")


# any_table with multiple combinations
result_list <- dummy_df |>
      any_table(rows    = c("age", "age + education"),
                columns = c("sex + year", "sex"),
                values  = weight,
                output  = "excel_nostyle",
                print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with multiple combinations")
expect_equal(length(result_list), 4, info = "any_table with multiple combinations")


# any_table many combinations don't break
result_list <- dummy_df |>
        any_table(rows    = c("age", "age + education", "state",
                              "state + age", "education + age"),
                  columns = c("year", "sex + year", "sex"),
                  values  = weight,
                  output  = "excel_nostyle",
                  print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table many combinations don't break")
expect_equal(length(result_list), 4, info = "any_table many combinations don't break")


# any_table with linked titles and footnotes
txt_file  <- system.file("extdata", "qol_example_data_txt.txt",   package = "qol")

set_titles("Hello world1",
           "Hello world2 link: https://cran.r-project.org/",
           "Hello world3 cell: A8",
           "Hello world4 file: txt_file")
set_footnotes("This is a footnote1",
              "This is a footnote2 link: https://cran.r-project.org/",
              "This is a footnote3 cell: A8",
              "This is a footnote4 file: txt_file")

set_style_options(as_heatmap          = TRUE,
                  header_stat_merging = "all")

result_list <- dummy_df |>
      any_table(rows      = "age",
                columns   = "sex",
                values    = weight,
                titles    = "Hello world link: https://cran.r-project.org/",
                footnotes = "This is a footnote link: https://cran.r-project.org/",
                print     = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with linked titles and footnotes")
expect_equal(length(result_list), 4, info = "any_table with linked titles and footnotes")

# any_table with multiple titles and footnotes
set_titles("Hello world1", "Hello world2")
set_footnotes("This is a footnote1", "This is a footnote2")

set_style_options(header_stat_merging = "none",
                  title_font_color    = c("FF00FF", "00FF00"),
                  title_font_size     = c(10, 11),
                  title_font_bold     = c(TRUE, FALSE),
                  footnote_font_color = c("FF00FF", "00FF00"),
                  footnote_font_size  = c(10, 11),
                  footnote_font_bold  = c(TRUE, FALSE),
                  as_heatmap          = FALSE)

result_list <- dummy_df |>
    any_table(rows      = "age",
              columns   = "sex",
              values    = weight,
              print     = FALSE)

reset_style_options()

expect_inherits(result_list, "qol_table", info = "any_table with multiple titles and footnotes")
expect_equal(length(result_list), 4, info = "any_table with multiple titles and footnotes")


# any_table with multiple titles and footnotes makes style options scale accordingly
set_titles("Hello world1", "Hello world2", "Hello world3")
set_footnotes("This is a footnote1", "This is a footnote2", "This is a footnote3")

result_list <- dummy_df |>
    any_table(rows      = "age",
              columns   = "sex",
              values    = weight,
              print     = FALSE)

style <- result_list[["meta"]][["style"]]

reset_style_options()

expect_equal(length(style[["title_font_color"]]),    3, info = "any_table with multiple titles and footnotes makes style options scale accordingly")
expect_equal(length(style[["title_font_size"]]),     3, info = "any_table with multiple titles and footnotes makes style options scale accordingly")
expect_equal(length(style[["title_font_bold"]]),     3, info = "any_table with multiple titles and footnotes makes style options scale accordingly")
expect_equal(length(style[["title_alignment"]]),     3, info = "any_table with multiple titles and footnotes makes style options scale accordingly")
expect_equal(length(style[["footnote_font_color"]]), 3, info = "any_table with multiple titles and footnotes makes style options scale accordingly")
expect_equal(length(style[["footnote_font_size"]]),  3, info = "any_table with multiple titles and footnotes makes style options scale accordingly")
expect_equal(length(style[["footnote_font_bold"]]),  3, info = "any_table with multiple titles and footnotes makes style options scale accordingly")
expect_equal(length(style[["footnote_alignment"]]),  3, info = "any_table with multiple titles and footnotes makes style options scale accordingly")


# any_table with variable and stat labels
result_list <- dummy_df |>
      any_table(rows        = "age",
                columns     = "sex",
                values      = weight,
                var_labels  = list(age = "Single ages", sex = "Sex", weight = "Population"),
                stat_labels = list(sum = "Counts"),
                box         = "Test",
                print       = FALSE)

expect_true("Single ages" %in% result_list[["table"]][["row.label"]], info = "any_table with variable and stat labels")


# any_table with removed variable and stat labels
result_list <- dummy_df |>
      any_table(rows        = "age",
                columns     = "sex",
                values      = weight,
                var_labels  = list(age = "", sex = "", weight = ""),
                stat_labels = list(sum = ""),
                print       = FALSE)

expect_true(!"row.label" %in% names(result_list[["table"]]), info = "any_table with removed variable and stat labels")


# any_table with different percentages
result_list <- dummy_df |>
      any_table(rows       = "age",
                columns    = "sex",
                values     = c(probability, weight),
                statistics = c("sum", "pct_total"),
                pct_group  = c("age", "sex"),
                pct_value  = list(rate = "probability / weight",
                                  sex  = 1),
                output     = "excel_nostyle",
                print      = FALSE)

expect_true(all(c("weight_pct_group_age_1", "weight_pct_total_1",
                  "rate_pct_value_1", "sex_pct_value_1") %in% names(result_list[["table"]])),
            info = "any_table with different percentages")


# any_table with keywords for row and column percentages
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = "sex",
                  values     = c(probability, weight),
                  statistics = c("pct_group"),
                  pct_group  = c("row_pct", "col_pct"),
                  output     = "excel_nostyle",
                  print      = FALSE)

expect_true(all(c("weight_pct_group_row_1", "probability_pct_group_row_2",
                  "weight_pct_group_col_1", "probability_pct_group_col_2")
                %in% names(result_list[["table"]])), info = "any_table with keywords for row and column percentages")


# any_table with block row percentages
result_list <- dummy_df |>
    any_table(rows      = "age + (sex, education)",
              columns   = "state",
              values    = weight,
              pct_block = "rows",
              output    = "excel_nostyle",
              formats   = list(state = state., age = age., sex = sex., education = education.),
              print     = FALSE,
              na.rm     = TRUE)

result_df <- result_list[["table"]] |> if.(var2 == "Total")
result_df[["weight_pct_block_rows_West"]] <- round(result_df[["weight_pct_block_rows_West"]])
result_df[["weight_pct_block_rows_East"]] <- round(result_df[["weight_pct_block_rows_East"]])

expect_equal(collapse::funique(result_df[["weight_pct_block_rows_West"]]), 100, info = "any_table with block row percentages")
expect_equal(collapse::funique(result_df[["weight_pct_block_rows_East"]]), 100, info = "any_table with block row percentages")


# any_table with block column percentages
result_list <- dummy_df |>
    any_table(rows      = "state",
              columns   = "age + (sex, education)",
              values    = weight,
              pct_block = "columns",
              output    = "excel_nostyle",
              formats   = list(state = state., age = age., sex = sex., education = education.),
              order_by  = "blocks",
              print     = FALSE,
              na.rm     = TRUE)

result_df <- result_list[["table"]]
result_df[["weight_pct_block_columns_under 50_Total"]] <- round(result_df[["weight_pct_block_columns_under 50_Total"]])

expect_equal(collapse::funique(result_df[["weight_pct_block_columns_under 50_Total"]]), 100, info = "any_table with block column percentages")
expect_equal(names(result_df),
             c("row.label", "var1", "weight_pct_block_columns_under 50_Total",
               "weight_pct_block_columns_under 50_Male", "weight_pct_block_columns_under 50_Female", "weight_pct_block_columns_under 50_low",
               "weight_pct_block_columns_under 50_middle", "weight_pct_block_columns_under 50_high", "weight_pct_block_columns_50 and more_Total",
               "weight_pct_block_columns_50 and more_Male", "weight_pct_block_columns_50 and more_Female", "weight_pct_block_columns_50 and more_low",
               "weight_pct_block_columns_50 and more_middle", "weight_pct_block_columns_50 and more_high"), info = "any_table with block order")


# any_table can expand combinations
result_list <- dummy_df |>
    any_table(rows       = "age + (state, first_person + education)",
              columns    = "sex",
              values     = weight,
              output     = "excel_nostyle",
              formats    = list(state = state., age = age., sex = sex., education = education.),
              print      = FALSE)

expect_true(all(c("West", "East", "0", "1") %in% result_list[["table"]][["var2"]]), info = "any_table can expand combinations")
expect_true(all(c("low", "middle", "high", "") %in% result_list[["table"]][["var3"]]), info = "any_table can expand combinations")


# any_table with a lot of statistics doesn't break
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = "sex",
                  values     = weight,
                  statistics = c("freq", "freq_g0", "mean", "median", "mode", "min", "max",
                                 "first", "last", "sum_wgt", "p1", "p99", "sd", "variance",
                                 "missing"),
                  output     = "excel_nostyle",
                  print      = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with a lot of statistics doesn't break")
expect_equal(length(result_list), 4, info = "any_table with a lot of statistics doesn't break")


# any_table with interleaved order
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = "sex",
                  values     = weight,
                  statistics = c("sum", "freq", "missing"),
                  order_by   = "interleaved",
                  output     = "excel_nostyle",
                  formats    = list(age = age., sex = sex.),
                  print      = FALSE,
                  na.rm      = TRUE)

expect_equal(names(result_list[[1]]),
               c("row.label", "var1", "weight_sum_Total", "weight_freq_Total", "weight_missing_Total",
                 "weight_sum_Male", "weight_freq_Male", "weight_missing_Male",
                 "weight_sum_Female", "weight_freq_Female", "weight_missing_Female"),
             info = "any_table with interleaved order")


# any_table with values order
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = c("sex", "state"),
                  values     = c(weight, income),
                  statistics = "sum",
                  order_by   = "values",
                  output     = "excel_nostyle",
                  formats    = list(age = age., sex = sex., state = state.),
                  print      = FALSE,
                  na.rm      = TRUE)

expect_equal(names(result_list[[1]]),
             c("row.label", "var1", "weight_sum_Total", "weight_sum_Male", "weight_sum_Female",
               "weight_sum_West", "weight_sum_East",
               "income_sum_Total", "income_sum_Male", "income_sum_Female", "income_sum_West",
               "income_sum_East"), info = "any_table with values order")


# any_table with columns order
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = c("sex", "state"),
                  values     = c(weight, income),
                  statistics = "sum",
                  order_by   = "columns",
                  output     = "excel_nostyle",
                  formats    = list(age = age., sex = sex., state = state.),
                  print      = FALSE,
                  na.rm      = TRUE)

expect_equal(names(result_list[[1]]),
             c("row.label", "var1", "weight_sum_Total", "weight_sum_Male", "weight_sum_Female",
               "income_sum_Total", "income_sum_Male", "income_sum_Female",
               "weight_sum_West", "weight_sum_East", "income_sum_West",
               "income_sum_East"), info = "any_table with columns order")


# any_table with by variables
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = "sex",
                  values     = weight,
                  by         = education,
                  print_miss = TRUE,
                  print      = FALSE)

expect_true("BY" %in% names(result_list[[1]]), info = "any_table with by variables")
expect_equal(length(unique(result_list[[1]][["BY"]])), 1, info = "any_table with by variables")


# any_table with multiple by variables
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = "sex",
                  values     = weight,
                  by         = c(education, year),
                  output     = "excel_nostyle",
                  print_miss = TRUE,
                  print      = FALSE)

expect_true("BY" %in% names(result_list[["table"]]), info = "any_table with multiple by variables")
expect_equal(length(unique(result_list[["table"]][["BY"]])), 2, info = "any_table with multiple by variables")


# any_table with by variables and multiple row and column variables
result_list <- dummy_df |>
       any_table(rows    = c("first_person", "state", "first_person + state"),
                 columns = c("sex", "education", "sex + education"),
                 values  = weight,
                 by      = year,
                 print   = FALSE,
                 output  = "excel_nostyle",
                 na.rm   = TRUE)

expect_true("BY" %in% names(result_list[["table"]]), info = "any_table with by variables and multiple row and column variables")


# any_table with by variables as subheaders
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  by      = education,
                  style   = excel_output_style(by_as_subheaders = TRUE),
                  print   = FALSE)

expect_true("BY" %in% names(result_list[[1]]), info = "any_table with by variables as subheaders")
expect_equal(length(unique(result_list[[1]][["BY"]])), 1, info = "any_table with by variables as subheaders")


# any_table with by variables throws a warning if by is also part of rows or columns
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  by      = sex,
                  print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "The provided <by> variable 'sex' is also part of",
             info = "any_table with by variables throws a warning if by is also part of rows or columns")


# any_table with NAs removed
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  output  = "excel_nostyle",
                  na.rm   = TRUE,
                  print   = FALSE)

expect_true(sum(is.na(result_list[["table"]][["var1"]])) == 0, info = "any_table with NAs removed")


# any_table with applied single discrete labels
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  formats = list(age = discrete_format(
                      "under 18"       = 0:17,
                      "18 to under 25" = 18:24,
                      "25 to under 55" = 25:54,
                      "55 to under 65" = 55:64,
                      "65 and older"   = 65:100)),
                  output = "excel_nostyle",
                  print  = FALSE)

expect_true(all(c("under 18", "18 to under 25", "25 to under 55",
                  "55 to under 65", "65 and older")
                %in% result_list[["table"]][["var1"]]), info = "any_table with applied single discrete labels")


# any_table with applied discrete multilabels
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  formats = list(sex = discrete_format(
                      "Total"  = 1:2,
                      "Male"   = 1,
                      "Female" = 2),
                      age = discrete_format(
                      "Total"          = 0:100,
                      "under 18"       = 0:17,
                      "18 to under 25" = 18:24,
                      "25 to under 55" = 25:54,
                      "55 to under 65" = 55:64,
                      "65 and older"   = 65:100)),
                  output = "excel_nostyle",
                  print  = FALSE)

expect_true(all(c("under 18", "18 to under 25", "25 to under 55",
                  "55 to under 65", "65 and older")
                %in% result_list[["table"]][["var1"]]), info = "any_table with applied discrete multilabels")


# any_table with applied interval multilabels
result_list <- dummy_df |>
        any_table(rows    = "income",
                  columns = "sex",
                  values  = weight,
                  formats = list(sex = discrete_format(
                      "Total"  = 1:2,
                      "Male"   = 1,
                      "Female" = 2),
                      income = interval_format(
                          "Total"              =    0:100000,
                          "below 500"          =    0:500,
                          "500 to under 1000"  =  500:1000,
                          "1000 to under 2000" = 1000:2000,
                          "2000 and more"      = 2000:100000)),
                  output = "excel_nostyle",
                  print  = FALSE)

expect_true(all(c("Total", "below 500", "2000 and more")
                %in% result_list[["table"]][["var1"]]), info = "any_table with applied interval multilabels")


# any_table can silence format expressions
sex2. <- discrete_format(
    "!Total" = 1:2,
    "Male"   = 1,
    "Female" = 2)

result_list <- dummy_df |>
    any_table(rows       = "first_person",
              columns    = "sex",
              values     = weight,
              statistics = "sum",
              output     = "excel_nostyle",
              formats    = list(sex = sex2.),
              print      = FALSE)

expect_false(any(grepl("!", names(result_list[["table"]]))), info = "any_table can silence format expressions")


# any_table able to apply format on numeric values stored as character (short route)
binary. <- discrete_format(
    "binary1" = c("00", "01"),
    "binary2" = c("10", "11"))

result_list <- dummy_df |>
    any_table(rows    = "binary",
              columns = "sex",
              values  = weight,
              formats = list(binary = binary.),
              output  = "excel_nostyle",
              print   = FALSE)

expect_equal(result_list[[1]][["var1"]], c("binary1", "binary2"), info = "any_table able to apply format on numeric values stored as character (short route)")


# any_table doesn't convert numeric values stored as character (short route)
result_list <- dummy_df |>
        any_table(rows    = "binary",
                  columns = "sex",
                  values  = weight,
                  output  = "excel_nostyle",
                  print   = FALSE)

expect_equal(result_list[[1]][["var1"]], c("00", "01", "10", "11"), info = "any_table doesn't convert numeric values stored as character (short route)")


# any_table able to apply format on numeric values stored as character (long route)
binary. <- discrete_format(
    "binary1" = c("00", "01"),
    "binary2" = c("10", "11"))

result_list <- dummy_df |>
        any_table(rows       = "binary",
                  columns    = "sex",
                  statistics = "mean",
                  values     = weight,
                  formats    = list(binary = binary.),
                  output     = "excel_nostyle",
                  print      = FALSE)

expect_equal(result_list[[1]][["var1"]], c("binary1", "binary2"), info = "any_table able to apply format on numeric values stored as character (long route)")


# any_table doesn't convert numeric values stored as character (long route)
result_list <- dummy_df |>
        any_table(rows       = "binary",
                  columns    = "sex",
                  statistics = "mean",
                  values     = weight,
                  output     = "excel_nostyle",
                  print      = FALSE)

expect_equal(result_list[[1]][["var1"]], c("00", "01", "10", "11"), info = "any_table doesn't convert numeric values stored as character (long route)")


# any_table with fixed column headers
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  style   = excel_output_style(freeze_col_header = TRUE),
                  na.rm   = TRUE,
                  output  = "excel_nostyle",
                  print   = FALSE)

expect_true(sum(is.na(result_list[["table"]][["var1"]])) == 0, info = "any_table with fixed column headers")


# any_table with fixed row headers
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  style   = excel_output_style(freeze_row_header = TRUE),
                  na.rm   = TRUE,
                  output  = "excel_nostyle",
                  print   = FALSE)

expect_true(sum(is.na(result_list[["table"]][["var1"]])) == 0, info = "any_table with fixed row headers")


# any_table with fixed column and row headers
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  style   = excel_output_style(freeze_col_header = TRUE,
                                               freeze_row_header = TRUE,
                                               background_color  = "FF00FF"),
                  na.rm   = TRUE,
                  output  = "excel_nostyle",
                  print   = FALSE)

expect_true(sum(is.na(result_list[["table"]][["var1"]])) == 0, info = "any_table with fixed column and row headers")


# any_table warning with wrong output format
result_list <- dummy_df |>
                   any_table(rows    = "age",
                             columns = "sex",
                             values  = weight,
                             output  = "Test",
                             print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "<Output> format 'Test' not available.", info = "any_table warning with wrong output format")


# any_table warning with wrong output format
result_list <- dummy_df |>
                   any_table(rows     = "age",
                             columns  = "sex",
                             values   = weight,
                             order_by = "test",
                             output   = "excel_nostyle",
                             print    = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "<Order by> option 'test' doesn't exist", info = "any_table warning with wrong output format")


# any_table with pre summarised data
result_list <- sum_df |>
   any_table(rows       = "year",
             columns    = "sex",
             values     = weight_sum,
             output     = "excel_nostyle",
             print      = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with pre summarised data")
expect_equal(length(result_list), 4, info = "any_table with pre summarised data")


# any_table with pre summarised data and by variables
result_list <- sum_df2 |>
                any_table(rows       = "age",
                          columns    = "year",
                          by         = "sex",
                          values     = weight_sum,
                          output     = "excel_nostyle",
                          print_miss = TRUE,
                          print      = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with pre summarised data and by variables")
expect_equal(length(result_list), 4, info = "any_table with pre summarised data and by variables")


# any_table with pre summarised data can use formats
result_list <- sum_df2 |>
    any_table(rows       = "year + age",
              columns    = "sex",
              values     = weight_sum,
              formats    = list(sex = sex., age = age.),
              output     = "excel_nostyle",
              print      = FALSE)

expect_equal(collapse::funique(result_list[["table"]][["var2"]]), c("under 50", "50 and more"), info = "any_table with pre summarised data can use formats")
expect_true(all(c("weight_sum_Total", "weight_sum_Male", "weight_sum_Female") %in% names(result_list[["table"]])),
            info = "any_table with pre summarised data can use formats")


# any_table with no column variables
result_list <- dummy_df |>
       any_table(rows    = "age",
                 values  = weight,
                 output  = "excel_nostyle",
                 print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with no column variables")
expect_equal(length(result_list), 4, info = "any_table with no column variables")


# any_table throws a warning, if invalid statistic specified
result_list <- dummy_df |>
                   any_table(rows       = "age",
                             columns    = "sex",
                             values     = weight,
                             statistics = c("test", "sum"),
                             output     = "excel_nostyle",
                             print      = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "<Statistic> 'test' is invalid and will be omitted.",
               info = "any_table throws a warning, if invalid statistic specified")


# any_table throws a warning, if no valid statistic specified
result_list <- dummy_df |>
                   any_table(rows       = "age",
                             columns    = "sex",
                             values     = weight,
                             statistics = "test",
                             output     = "excel_nostyle",
                             print      = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "No valid <statistic> selected. 'sum' will be used.",
               info = "any_table throws a warning, if no valid statistic specified")


# Save any_table as Excel file
temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              values  = weight,
              output  = "excel_nostyle",
              style   = excel_output_style(save_path = dirname(temp_file),
                                           file      = basename(temp_file)))

expect_true(file.exists(temp_file), info = "Save any_table as Excel file")


# Combine tables into a single workbook
my_style <- excel_output_style(sheet_name = "tab1")

tab1 <- dummy_df |>
     any_table(rows    = "age",
               columns = "sex",
               values  = weight,
               output  = "excel_nostyle",
               style   = my_style,
               print   = FALSE)

my_style <- my_style |> modify_output_style(sheet_name = "tab2")

tab2 <- dummy_df |>
     any_table(rows    = "age",
               columns = "sex",
               values  = weight,
               by      = education,
               output  = "excel_nostyle",
               style   = my_style,
               print   = FALSE)

temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

result <- combine_into_workbook(tab1, tab2, style = excel_output_style(save_path = dirname(temp_file),
                                                                       file      = basename(temp_file)))

expect_inherits(result, c("wbWorkbook", "R6"), info = "Combine tables into a single workbook")
expect_true(file.exists(temp_file), info = "Combine tables into a single workbook")

result_list <- dummy_df |>
    any_table(rows     = "age",
              columns  = "sex",
              values   = weight,
              output   = "excel_nostyle",
              workbook = tab1,
              style    = my_style,
              print    = FALSE)

expect_inherits(result_list, "qol_table", info = "Combine tables into a single workbook")
expect_equal(length(result_list), 4, info = "Combine tables into a single workbook")


# Combine tables into a single workbook with table of contents
my_style <- excel_output_style(sheet_name = "tab1")

tab1 <- dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              values  = weight,
              output  = "excel_nostyle",
              style   = my_style,
              print   = FALSE)

my_style <- my_style |> modify_output_style(sheet_name = "tab2")

tab2 <- dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              values  = weight,
              by      = education,
              output  = "excel_nostyle",
              style   = my_style,
              print   = FALSE)

result <- combine_into_workbook(tab1, tab2,
                                table_of_contents = TRUE,
                                subheaders        = list("First Subheader"  = "tab1",
                                                         "Second Subheader" = "tab21"),
                                subheader_colors  = c("FF0000", "00FF00", "0000FF"),
                                colored_tabs      = TRUE,
                                print             = FALSE)

expect_inherits(result, c("wbWorkbook", "R6"), info = "Combine tables into a single workbook with table of contents")
expect_true("Contents" %in% openxlsx2::wb_get_sheet_names(result),
            info = "Combine tables into a single workbook with table of contents")

result <- combine_into_workbook(tab1, tab2,
                                table_of_contents = TRUE,
                                subheaders        = list("First Subheader"  = "tab1",
                                                         "Second Subheader" = "tab2"),
                                subheader_colors  = c("FF0000", "00FF00", "0000FF"),
                                colored_tabs      = TRUE,
                                print             = FALSE)

expect_error(print_stack_as_messages("ERROR"), "The following sheet name provided in the subheaders is invalid:",
             info = "Combine tables into a single workbook with table of contents")
expect_true(!"Contents" %in% openxlsx2::wb_get_sheet_names(result),
            info = "Combine tables into a single workbook with table of contents")


# Create and save table of contents standalone
temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

my_style <- excel_output_style(sheet_name = "tab1")

tab1 <- dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              values  = weight,
              output  = "excel_nostyle",
              style   = my_style,
              print   = FALSE)

my_style <- my_style |> modify_output_style(sheet_name = "tab2")

tab2 <- dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              values  = weight,
              by      = education,
              output  = "excel_nostyle",
              style   = my_style,
              print   = FALSE)

result <- combine_into_workbook(tab1, tab2, print = FALSE)

create_table_of_contents(result,
                         subheaders        = list("First Subheader"  = "tab1",
                                                  "Second Subheader" = "tab21"),
                         subheader_colors  = c("FF0000", "00FF00", "0000FF"),
                         colored_tabs      = TRUE,
                         style             = excel_output_style(save_path = dirname(temp_file),
                                                                file      = basename(temp_file)))

expect_true(file.exists(temp_file), info = "Create and save table of contents standalone")


# any_table throws a warning with missing statistic extension in pre summarised data
result_list <- sum_df |>
       any_table(rows    = "year",
                 columns = "sex",
                 values  = DEPTH,
                 output  = "excel_nostyle",
                 print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "All <values> variables need to have the <statistic> extension in their variable name.",
               info = "any_table throws a warning with missing statistic extension in pre summarised data")

expect_inherits(result_list, "qol_table", info = "any_table throws a warning with missing statistic extension in pre summarised data")
expect_equal(length(result_list), 4, info = "any_table throws a warning with missing statistic extension in pre summarised data")


# any_table auto generates missing TYPE variable in pre summarised data
result_list <- sum_df |>
       any_table(rows    = "year",
                 columns = "sex",
                 values  = weight_sum,
                 output  = "excel_nostyle",
                 print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table auto generates missing TYPE variable in pre summarised data")
expect_equal(length(result_list), 4, info = "any_table auto generates missing TYPE variable in pre summarised data")


# any_table outputs unweighted results without values variable
result_list <- dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              output  = "excel_nostyle",
              print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table outputs unweighted results without values variable")
expect_equal(length(result_list), 4, info = "any_table outputs unweighted results without values variable")


# any_table throws a warning, if invalid format is used
result_list <- dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              values  = weight,
              output  = "excel_nostyle",
              formats = list(age = age., sex = test),
              print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Format for variable 'sex' does not exist and can't be applied.",
               info = "any_table throws a warning, if invalid format is used")


# any_table can compute values on the fly
result_list <- dummy_df |>
    any_table(rows       = "first_person",
              columns    = "sex",
              values     = "probability",
              statistics = c("sum", "sum_wgt"),
              compute    = list(percent    = probability_sum * 100 / sum_wgt,
                                pct_square = percent ^ 2),
              weight     = weight,
              output     = "excel_nostyle",
              na.rm      = TRUE,
              print      = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table can compute values on the fly")
expect_equal(length(result_list), 4, info = "any_table can compute values on the fly")
expect_equal(names(result_list[[1]]), c("row.label", "var1", "probability_sum_1",
                                        "probability_sum_2", "sum_wgt_1", "sum_wgt_2",
                                        "percent_sum_1", "percent_sum_2", "pct!!!square_sum_1", "pct!!!square_sum_2"),
             info = "any_table can compute values on the fly")


# any_table orders individual variables by name
result_list <- dummy_df |>
    any_table(rows       = "first_person",
              columns    = "sex",
              values     = "probability",
              statistics = c("sum", "sum_wgt"),
              compute    = list(percent    = probability_sum * 100 / sum_wgt,
                                pct_square = percent ^ 2),
              weight     = weight,
              order_by   = c("pct_square", "percent", "sum_wgt"),
              output     = "excel_nostyle",
              na.rm      = TRUE,
              print      = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table orders individual variables by name")
expect_equal(length(result_list), 4, info = "any_table orders individual variables by name")
expect_equal(names(result_list[[1]]), c("row.label", "var1", "pct!!!square_sum_1", "pct!!!square_sum_2",
                                        "percent_sum_1", "percent_sum_2", "sum_wgt_1", "sum_wgt_2",
                                        "probability_sum_1",  "probability_sum_2"),
             info = "any_table orders individual variables by name")
expect_message(print_stack_as_messages("NOTE"), "The following variables generated by compute are missing",
               info = "any_table orders individual variables by name")


# any_table is able to output specific statistics per variable
result_list <- dummy_df |>
    any_table(rows       = "age",
              columns    = "sex",
              statistics = list("sum"       = income,
                                "pct_group" = weight),
              na.rm      = TRUE,
              print      = FALSE)

expect_equal(collapse::fncol(result_list[[1]]), 6, info = "any_table is able to output specific statistics per variable")
expect_true(all(c("income_sum_1", "income_sum_2", "weight_pct_group_sex_1", "weight_pct_group_sex_2") %in% names(result_list[[1]])),
            info = "any_table is able to output specific statistics per variable")
expect_true(!all(c("income_pct_group_1", "income_pct_group_2", "weight_sum_1", "weight_sum_2") %in% names(result_list[[1]])),
            info = "any_table is able to output specific statistics per variable")

result_list <- dummy_df |>
    any_table(rows       = "age",
              columns    = "sex",
              statistics = list("sum"       = c(income, expenses),
                                "pct_group" = weight),
              na.rm      = TRUE,
              print      = FALSE)

expect_equal(collapse::fncol(result_list[[1]]), 8, info = "any_table is able to output specific statistics per variable")
expect_true(all(c("income_sum_1", "income_sum_2", "expenses_sum_1", "expenses_sum_2",
                  "weight_pct_group_sex_1", "weight_pct_group_sex_2") %in% names(result_list[[1]])),
            info = "any_table is able to output specific statistics per variable")
expect_true(!all(c("income_pct_group_1", "income_pct_group_2", "expenses_pct_group_1",
                   "expenses_pct_group_2", "weight_sum_1", "weight_sum_2") %in% names(result_list[[1]])),
            info = "any_table is able to output specific statistics per variable")


# any_table can handle duplicate column names without NA values
result_list <- dummy_df |>
    any_table(rows    = "education",
              columns = c("sex", "age"),
              values  = weight,
              na.rm   = TRUE,
              print   = FALSE)

expect_true(all(c("weight_sum_1.dup1", "weight_sum_2.dup1",
                  "weight_sum_1.dup2", "weight_sum_1.dup2") %in% names(result_list[[1]])),
            info = "any_table can handle duplicate column names without NA values")


result_list <- dummy_df |>
    any_table(rows    = "education",
              columns = c("sex", "age"),
              values  = weight,
              formats = list(sex = sex., age = age2.),
              na.rm   = TRUE,
              print   = FALSE)

expect_true(all(c("weight_sum_Total.dup1", "weight_sum_Total.dup2") %in% names(result_list[[1]])),
            info = "any_table can handle duplicate column names without NA values")


# any_table throws a warning with invalid by variable
result_list <- dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              by      = "test",
              values  = weight,
              print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "The provided <by> variable 'test' is not part of",
               info = "any_table throws a warning with invalid by variable")


# any_table throws a warning with values stored as character
result_list <- dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              values  = education,
              print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "The following <values> are stored as character variables in the data frame",
               info = "any_table throws a warning with values stored as character")

result_list <- dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              values  = c(education, income_class),
              print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "The following <values> are stored as character variables in the data frame",
               info = "any_table throws a warning with values stored as character")

###############################################################################
# Html checks
###############################################################################

# any_table can render output in html format
result_html <- dummy_df |>
    any_table(rows       = c("sex", "age"),
              columns    = "year",
              values     = weight,
              statistics = c("sum", "pct_group"),
              pct_group  = "sex",
              formats    = list(sex = sex., age = age.),
              by         = "education",
              titles     = "My Title",
              footnotes  = "My Footnote",
              na.rm      = TRUE,
              print      = FALSE,
              output     = "html",
              style      = excel_output_style(by_as_subheaders = TRUE,
                                              as_heatmap       = TRUE))

html_out <- result_html[["html"]]

expect_inherits(result_html, "qol_table",                         info = "any_table can render output in html format")
expect_true(grepl("<!DOCTYPE html>",     html_out, fixed = TRUE), info = "any_table html render returns a html document")
expect_true(grepl("<thead>",             html_out, fixed = TRUE), info = "any_table html render contains a table header")
expect_true(grepl("<tbody>",             html_out, fixed = TRUE), info = "any_table html render contains a table body")
expect_true(grepl("class=\"box\"",       html_out, fixed = TRUE), info = "any_table html render renders the top left box")
expect_true(grepl("class=\"subheader\"", html_out, fixed = TRUE), info = "any_table html render renders by variable blocks as subheaders")
expect_true(grepl("My Title",            html_out, fixed = TRUE), info = "any_table html render renders the titles")
expect_true(grepl("My Footnote",         html_out, fixed = TRUE), info = "any_table html render renders the footnotes")
expect_true(grepl("</div>\n<div class=\"qol-spacer\"></div>\n               <div class=\"qol-wrap\">", html_out),
            info = "any_table html render puts the blank row between the titles and the table")
expect_true(grepl("</div>\n               <div class=\"qol-spacer\"></div>\n<div class=\"qol-footnotes qol-footnotes-first\"", html_out),
            info = "any_table html render puts the blank row between the table and the footnotes")
expect_true(grepl(".qol-spacer {", html_out, fixed = TRUE), info = "any_table html render styles the blank row")
expect_true(grepl('class="qol-footnotes qol-footnotes-first"', html_out, fixed = TRUE),
            info = "any_table html render marks the top footnote for the separating line")
expect_true(grepl('.qol-footnotes-first::before', html_out, fixed = TRUE),
            info = "any_table html render draws the footnote line via a pseudo element")
expect_true(grepl('width: var(--qol-fn-line, 0px)', html_out, fixed = TRUE),
            info = "any_table html render sizes the footnote line via a css variable")
expect_true(grepl('--qol-fn-line", catWidth + "px"', html_out, fixed = TRUE),
            info = "any_table html render measures the row header width for the footnote line")
expect_true(grepl('class="qol-sheet" style="margin-left: 8ch"', html_out, fixed = TRUE),
            info = "any_table html render offsets the sheet by one empty column with start_column = 2")
expect_true(grepl("<body>\n               <div class=\"qol-spacer\"></div>\n<div class=\"qol-sheet\"", html_out),
            info = "any_table html render inserts one empty row before the titles with start_row = 2")
expect_false(grepl('style="height:', html_out, fixed = TRUE), info = "any_table html render writes no inline row heights with row_heights = auto")


# any_table html render inserts no empty columns or rows on start at 1,1
result_offset <- dummy_df |>
    any_table(rows       = "sex",
              columns    = "year",
              output     = "html",
              print      = FALSE,
              style      = excel_output_style(start_row = 1, start_column = 1))

html_offset <- result_offset[["html"]]

expect_equal(lengths(regmatches(html_offset, gregexpr("class=\"qol-spacer\"", html_offset))), 0L,
             info = "any_table html render inserts no empty columns or rows on start at 1,1")
expect_false(grepl("margin-left:", html_offset),
             info = "any_table html render inserts no empty columns or rows on start at 1,1")


# any_table html render inserts custom statistic labels
result_stats <- dummy_df |>
    any_table(rows        = "sex",
              columns     = "year",
              values      = weight,
              statistics  = c("sum", "mean"),
              formats     = list(sex = sex.),
              stat_labels = list(sum = "1000", mean = "Average"),
              output      = "html",
              print       = FALSE)

html_stats <- result_stats[["html"]]

expect_true(grepl(">1000<",    html_stats, fixed = TRUE), info = "any_table html render inserts custom labels")
expect_true(grepl(">Average<", html_stats, fixed = TRUE), info = "any_table html render inserts custom labels")
expect_false(grepl(">sum<",    html_stats),               info = "any_table html render inserts custom labels")
expect_false(grepl(">mean<",   html_stats),               info = "any_table html render inserts custom labels")


# any_table html render sets individual column widths
result_widths <- dummy_df |>
    any_table(rows       = c("sex", "age"),
              columns    = "year",
              values     = weight,
              statistics = "sum",
              formats    = list(sex = sex., age = age.),
              output     = "html",
              style      = excel_output_style(column_widths = c(2, 20, 20, 20, 10)),
              print      = FALSE)

html_widths <- result_widths[["html"]]

colgroup <- regmatches(html_widths, regexpr("<colgroup>.*?</colgroup>", html_widths))

expect_true(grepl("width: 20ch",          colgroup,    fixed = TRUE), info = "any_table html render sets individual column widths")
expect_true(grepl('style="width: 100ch"', html_widths, fixed = TRUE), info = "any_table html render sets individual column widths")
expect_true(grepl('class="qol-sheet" style="margin-left: 2ch"', html_widths, fixed = TRUE), info = "any_table html render sets individual column widths")
expect_true(grepl("table-layout: fixed",  html_widths, fixed = TRUE), info = "any_table html render sets individual column widths")


# any_table html render sets individual row heights
result_heights <- dummy_df |>
    any_table(rows       = c("sex", "age"),
              columns    = "year",
              values     = weight,
              statistics = "sum",
              formats    = list(sex = sex., age = age.),
              titles     = "My Title",
              footnotes  = "My Footnote",
              output     = "html",
              style      = excel_output_style(row_heights = c(10, 20, 30, 40, 50, 60, 70)),
              print      = FALSE)

html_heights <- result_heights[["html"]]

expect_true(grepl('<div class="qol-spacer" style="height: 10pt"></div>', html_heights, fixed = TRUE),
            info = "aany_table html render sets individual row heights")
expect_true(grepl('<div class="qol-titles" style="height: 20pt[^"]*">My Title</div>', html_heights),
            info = "any_table html render applies the second row height to the title")
expect_true(grepl('<div class="qol-spacer" style="height: 30pt"></div>', html_heights, fixed = TRUE),
            info = "any_table html render applies the third row height to the blank row below the title")
expect_true(grepl('<tr style="height: 40pt">', html_heights, fixed = TRUE),
            info = "any_table html render applies a row height to the first column header row")
expect_true(grepl('<tr style="height: 50pt">', html_heights, fixed = TRUE),
            info = "any_table html render applies a row height to the second column header row")
expect_true(grepl('<tr style="height: 60pt">', html_heights, fixed = TRUE),
            info = "any_table html render applies a row height to the table body rows")
expect_true(grepl('<div class="qol-spacer" style="height: 70pt"></div>', html_heights, fixed = TRUE),
            info = "any_table html render applies a row height to the blank row above the footnotes")
expect_true(grepl('<div class="qol-footnotes qol-footnotes-first" style="height: 70pt[^"]*">My Footnote</div>', html_heights),
            info = "any_table html render applies the last row height to the footnote")


# any_table html render part-specific heights apply only to the parts that exist
result_parts <- dummy_df |>
    any_table(rows       = c("sex", "age"),
              columns    = "year",
              values     = weight,
              statistics = "sum",
              formats    = list(sex = sex., age = age.),
              titles     = c("Title One", "Title Two"),
              footnotes  = c("Note One", "Note Two"),
              output     = "html",
              style      = excel_output_style(title_heights    = c(20, 30),
                                              footnote_heights = c(12, 14)),
              print      = FALSE)

html_parts <- result_parts[["html"]]

expect_true(grepl('<div class="qol-titles" style="height: 20pt[^"]*">Title One</div>', html_parts),
            info = "any_table html render applies the first title height to the first title")
expect_true(grepl('<div class="qol-titles" style="height: 30pt[^"]*">Title Two</div>', html_parts),
            info = "any_table html render applies the second title height to the second title")
expect_true(grepl('<div class="qol-footnotes qol-footnotes-first" style="height: 12pt[^"]*">Note One</div>', html_parts),
            info = "any_table html render applies the first footnote height to the first footnote")
expect_true(grepl('<div class="qol-footnotes" style="height: 14pt[^"]*">Note Two</div>', html_parts),
            info = "any_table html render applies the second footnote height to the second footnote")
expect_false(grepl('style="height:', gsub("<div class=\"qol-titles\"[^>]*>|<div class=\"qol-footnotes[^\"]*\"[^>]*>", "", html_parts), fixed = TRUE),
             info = "any_table html render part-specific heights leave header and body rows untouched")


# any_table html render part-specific heights override the global row_heights for their own part only
result_override <- dummy_df |>
    any_table(rows       = c("sex", "age"),
              columns    = "year",
              values     = weight,
              statistics = "sum",
              formats    = list(sex = sex., age = age.),
              titles     = "My Title",
              output     = "html",
              style      = excel_output_style(row_heights    = c(10, 20, 30, 40, 50, 60),
                                              title_heights  = 99,
                                              header_heights = 25,
                                              table_heights  = 18),
              print      = FALSE)

html_override <- result_override[["html"]]

expect_true(grepl('<div class="qol-titles" style="height: 99pt[^"]*">My Title</div>', html_override),
            info = "any_table html render title_heights override row_heights for the title")
expect_true(grepl('<tr style="height: 25pt">', html_override, fixed = TRUE),
            info = "any_table html render header_heights override row_heights for the header rows")
expect_true(grepl('<tr style="height: 18pt">', html_override, fixed = TRUE),
            info = "any_table html render table_heights override row_heights for the body rows")
expect_false(grepl('<tr style="height: 40pt">|<tr style="height: 50pt">|<tr style="height: 60pt">', html_override, fixed = TRUE),
             info = "any_table html render part-specific heights suppress the matching row_heights slices")


# any_table html render subheader_heights apply to the by block banners
result_subheader <- dummy_df |>
    any_table(rows       = c("sex", "age"),
              columns    = "year",
              values     = weight,
              statistics = "sum",
              formats    = list(sex = sex., age = age.),
              by         = "education",
              output     = "html",
              style      = excel_output_style(by_as_subheaders  = TRUE,
                                              subheader_heights = 22,
                                              table_heights    = 15),
              print      = FALSE)

html_subheader <- result_subheader[["html"]]

expect_true(grepl('<tr style="height: 22pt">', html_subheader, fixed = TRUE),
            info = "any_table html render subheader_heights apply to the by block banners")
expect_true(grepl('<tr style="height: 15pt">', html_subheader, fixed = TRUE),
            info = "any_table html render table_heights apply to the data rows of by blocks")


# any_table html render renders every by expression as its own table
result_by_blocks <- dummy_df |>
    any_table(rows       = c("sex", "age"),
              columns    = "year",
              values     = weight,
              statistics = "sum",
              formats    = list(sex = sex., age = age.),
              by         = "education",
              titles     = "My Title",
              footnotes  = "My Footnote [by_var]",
              output     = "html",
              print      = FALSE)

html_by_blocks <- result_by_blocks[["html"]]

n_sheets <- lengths(regmatches(html_by_blocks, gregexpr("class=\"qol-sheet\"",        html_by_blocks)))
n_tables <- lengths(regmatches(html_by_blocks, gregexpr("<table class=\"qol-table\"", html_by_blocks)))

expect_true(n_sheets == n_tables && n_tables > 1, info = "any_table html render renders every by expression as its own table")
expect_false(grepl("class=\"subheader\"", html_by_blocks, fixed = TRUE), info = "any_table html render renders no subheaders when by expressions are individual tables")
expect_true(all(vapply(c("education = low", "education = middle", "education = high", "education = ."),
                       function(by_info) grepl(by_info, html_by_blocks, fixed = TRUE),
                       logical(1))), info = "any_table html render adds the by expression to the titles of every table")
expect_true(grepl("My Footnote low",    html_by_blocks, fixed = TRUE) &&
            grepl("My Footnote middle", html_by_blocks, fixed = TRUE) &&
            grepl("My Footnote high",   html_by_blocks, fixed = TRUE), info = "any_table html render replaces the [by_var] placeholder in the footnotes")
expect_true(grepl("</div>\n<div class=\"qol-spacer\"></div>\n<div class=\"qol-sheet\"", html_by_blocks),
            info = "any_table html render separates the individual tables by a blank row")


# any_table html render adds no width measurement script without titles or footnotes
result_plain <- dummy_df |>
    any_table(rows       = "sex",
              columns    = "year",
              values     = weight,
              statistics = "sum",
              formats    = list(sex = sex.),
              output     = "html",
              style      = excel_output_style(),
              print      = FALSE)

html_plain <- result_plain[["html"]]

expect_false(grepl("<script>", html_plain, fixed = TRUE), info = "any_table html render adds no width measurement script without titles or footnotes")


# any_table html render displays the decimals from the number format style
result_decimals <- dummy_df |>
    any_table(rows       = "sex",
              columns    = "year",
              values     = income,
              statistics = "sum",
              formats    = list(sex = sex.),
              output     = "html",
              style      = excel_output_style(
                  number_formats = number_format_style(
                      sum_excel    = "#,###,##0", # 0 decimals in the excel format
                      sum_decimals = 2)),         # but 2 decimals to display
              print      = FALSE)

html_decimals  <- result_decimals[["html"]]
cells_decimals <- regmatches(html_decimals, gregexpr('<td class="data"[^>]*>[^<]+</td>', html_decimals))[[1]]

expect_true(any(grepl(",[0-9]{2}</td>$",  cells_decimals)), info = "any_table html render displays the decimals from the number format style")
expect_false(any(grepl(",[0-9]{3}</td>$", cells_decimals)), info = "any_table html render displays the decimals from the number format style")
expect_true(any(grepl("^<td[^>]*>[0-9]{2,3}([.][0-9]{3})+,[0-9]{2}</td>$", cells_decimals)), info = "any_table html render uses a dot as thousand separator")
expect_false(any(grepl("^<td[^>]*>[0-9]{1,3},[0-9]{3}", cells_decimals)), info = "any_table html render does not use the separators of the excel formats")


# The outer borders stay open even when all inner borders are switched on
result_full_borders <- dummy_df |>
    any_table(rows       = "sex",
              columns    = "year",
              values     = weight,
              statistics = "sum",
              formats    = list(sex = sex.),
              output     = "html",
              style      = excel_output_style(header_borders    = TRUE,
                                              box_borders       = TRUE,
                                              cat_col_borders   = TRUE,
                                              table_borders     = TRUE,
                                              subheader_borders = TRUE),
              print      = FALSE)

html_full_borders <- result_full_borders[["html"]]

expect_true(grepl("table.qol-table th.box { border-left: none; }", html_full_borders, fixed = TRUE),
            info = "any_table html render keeps the outer left edge open when all borders are on")
expect_true(grepl("table.qol-table td.data { border-right: none; border-bottom: none; }", html_full_borders, fixed = TRUE),
            info = "any_table html render keeps the outer right and bottom edges open when all borders are on")


# Hyperlink keywords are translated into html links. "link:" points to a webpage,
# "file:" to a file on disk and "cell:" is a pseudo hyperlink that leads nowhere.
# All links are styled like the excel links with a blue font color and underline.
result_html_links <- dummy_df |>
    any_table(rows      = "age",
              columns   = "sex",
              values    = weight,
              titles    = c("Hello world1 link: https://cran.r-project.org/",
                            "Hello world2 cell: A8",
                            "Hello world3 file: test.txt"),
              footnotes = c("This is a footnote1 link: https://cran.r-project.org/",
                            "This is a footnote2 cell: A8",
                            "This is a footnote3 file: test.txt"),
              output    = "html",
              print     = FALSE)

html_links    <- result_html_links[["html"]]
link_file_url <- paste0("file:///", gsub("\\\\", "/", "test.txt"))

expect_true(grepl('<a href="https://cran.r-project.org/" target="_blank">Hello world1</a>', html_links, fixed = TRUE),
            info = "any_table html render renders a link: keyword in a title as a hyperlink")
expect_true(grepl('<a href="#" onclick="return false;">Hello world2</a>', html_links, fixed = TRUE),
            info = "any_table html render renders a cell: keyword in a title as a pseudo hyperlink")
expect_true(grepl(paste0('<a href="', link_file_url, '" target="_blank">Hello world3</a>'), html_links, fixed = TRUE),
            info = "any_table html render renders a file: keyword in a title as a hyperlink to the file")
expect_true(grepl('<a href="https://cran.r-project.org/" target="_blank">This is a footnote1</a>', html_links, fixed = TRUE),
            info = "any_table html render renders a link: keyword in a footnote as a hyperlink")
expect_true(grepl('<a href="#" onclick="return false;">This is a footnote2</a>', html_links, fixed = TRUE),
            info = "any_table html render renders a cell: keyword in a footnote as a pseudo hyperlink")
expect_true(grepl(paste0('<a href="', link_file_url, '" target="_blank">This is a footnote3</a>'), html_links, fixed = TRUE),
            info = "any_table html render renders a file: keyword in a footnote as a hyperlink to the file")
expect_true(grepl('Hello world1', html_links, fixed = TRUE),
            info = "any_table html render leaves titles without a hyperlink keyword unmodified")
expect_true(grepl(".qol-titles a { color: #0000FF; text-decoration: underline; }", html_links, fixed = TRUE),
            info = "any_table html render styles title links like the excel links")
expect_true(grepl(".qol-footnotes a { color: #0000FF; text-decoration: underline; }", html_links, fixed = TRUE),
            info = "any_table html render styles footnote links like the excel links")


# any_table html render styles titles and footnotes
result_styled <- dummy_df |>
    any_table(rows      = "sex",
              columns   = "year",
              values    = weight,
              statistics = "sum",
              formats   = list(sex = sex.),
              titles    = c("Title A", "Title B"),
              footnotes = c("Note A", "Note B"),
              output    = "html",
              style     = excel_output_style(),
              print     = FALSE)

html_styled <- result_styled[["html"]]

expect_true(grepl('<div class="qol-titles" style="font-size: 10pt; color: #000000; font-weight: bold; text-align: left">Title A</div>',
                  html_styled, fixed = TRUE), info = "any_table html render bolds the titles by default")
expect_true(grepl('<div class="qol-titles" style="font-size: 10pt; color: #000000; font-weight: bold; text-align: left">Title B</div>',
                  html_styled, fixed = TRUE), info = "any_table html render bolds every title with the default title_font_bold")
expect_true(grepl('<div class="qol-footnotes qol-footnotes-first" style="font-size: 8pt; color: #000000; font-weight: normal; text-align: left">Note A</div>',
                  html_styled, fixed = TRUE), info = "any_table html render does not bold the footnotes by default")

result_per_title <- dummy_df |>
    any_table(rows      = "sex",
              columns   = "year",
              values    = weight,
              statistics = "sum",
              formats   = list(sex = sex.),
              titles    = c("Title A", "Title B", "Title C"),
              footnotes = c("Note A", "Note B"),
              output    = "html",
              print     = FALSE,
              style     = excel_output_style(title_font_color    = c("FF00FF", "00FF00", "0000FF"),
                                             title_font_size     = c(10, 11, 12),
                                             title_font_bold     = c(TRUE, FALSE, TRUE),
                                             title_alignment     = c("left", "center", "right"),
                                             footnote_font_color = c("FF0000", "000000"),
                                             footnote_font_bold  = c(TRUE, FALSE)))

html_per_title <- result_per_title[["html"]]

expect_true(grepl('<div class="qol-titles" style="font-size: 10pt; color: #FF00FF; font-weight: bold; text-align: left">Title A</div>',
                  html_per_title, fixed = TRUE), info = "any_table html render applies the first title font attributes to the first title")
expect_true(grepl('<div class="qol-titles" style="font-size: 11pt; color: #00FF00; font-weight: normal; text-align: center">Title B</div>',
                  html_per_title, fixed = TRUE), info = "any_table html render applies the second title font attributes to the second title")
expect_true(grepl('<div class="qol-titles" style="font-size: 12pt; color: #0000FF; font-weight: bold; text-align: right">Title C</div>',
                  html_per_title, fixed = TRUE), info = "any_table html render applies the third title font attributes to the third title")
expect_true(grepl('<div class="qol-footnotes qol-footnotes-first" style="font-size: 8pt; color: #FF0000; font-weight: bold; text-align: left">Note A</div>',
                  html_per_title, fixed = TRUE), info = "any_table html render applies the first footnote font attributes to the first footnote")
expect_true(grepl('<div class="qol-footnotes" style="font-size: 8pt; color: #000000; font-weight: normal; text-align: left">Note B</div>',
                  html_per_title, fixed = TRUE), info = "any_table html render applies the second footnote font attributes to the second footnote")


# Subheader wrap and cell indents are translated into css
expect_true(grepl('td.subheader { background: #FFFFFF; color: #000000; font-size: 10pt; font-weight: bold; text-align: center; border: 1px solid #000000; white-space: normal; }',
                  html_styled, fixed = TRUE),
            info = "any_table html render wraps the subheader text by default like excel")
expect_true(grepl('th.cat { background: #FFFFFF; color: #000000; font-size: 10pt; font-weight: normal; text-align: left; border: 1px solid #000000; white-space: normal; padding-left: 0.6em; }',
                  html_styled, fixed = TRUE),
            info = "any_table html render indents the category column cells by default like excel")
expect_true(grepl('td.data { background: #FFFFFF; color: #000000; font-size: 10pt; font-weight: normal; text-align: right; border: none; padding-left: 0.6em; }',
                  html_styled, fixed = TRUE),
            info = "any_table html render indents the table cells by default like excel")

result_no_indent <- dummy_df |>
    any_table(rows      = "sex",
              columns   = "year",
              values    = weight,
              statistics = "sum",
              formats   = list(sex = sex.),
              titles    = "T",
              output    = "html",
              print     = FALSE,
              style     = excel_output_style(cat_col_indent = 0, table_indent = 0))

html_no_indent <- result_no_indent[["html"]]

expect_false(grepl("padding-left", html_no_indent, fixed = TRUE),
            info = "any_table writes no indents with indent levels of 0")

###############################################################################
# Abort checks
###############################################################################

# any_table aborts, if column contains a row variable
result_list <- dummy_df |>
                   any_table(rows       = "sex",
                             columns    = "sex",
                             statistics = c("sum"),
                             print      = FALSE)

expect_error(print_stack_as_messages("ERROR"), "The provided <columns> variable '",
             info = "any_table aborts, if column contains a row variable")


# any_table aborts with duplicate column names because of NA values
result_list <- dummy_df |>
    any_table(rows    = "education",
              columns = c("sex", "age"),
              values  = weight,
              print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "Duplicate <columns> names found",
             info = "any_table aborts with duplicate column names because of NA values")


# any_table aborts with none existent row variable
result_list <- dummy_df |>
      any_table(rows    = c("age", "age + test"),
                columns = "sex",
                values  = weight,
                print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "The provided <rows> variable 'test' is not part of",
             info = "any_table aborts with none existent row variable")


# any_table aborts with none existent column variable
result_list <- dummy_df |>
      any_table(rows    = "age",
                columns = c("sex + test"),
                values  = weight,
                print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "The provided <columns> variable 'test' is not part of",
             info = "any_table aborts with none existent column variable")


# any_table aborts with no valid row variables
result_list <- dummy_df |>
       any_table(rows    = "",
                 columns = "sex",
                 values  = weight,
                 print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "No valid <rows> variables provided",
             info = "any_table aborts with no valid row variables")


# any_table aborts with row/column variable part of values
result_list <- dummy_df |>
       any_table(rows    = "age",
                 columns = "sex",
                 values  = "sex",
                 print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "The provided <rows>/<columns> variable 'sex' is also part of",
             info = "any_table aborts with row/column variable part of values")


# any_table outputs sum values with only invalid pct_value statistic and throws a warning
result_list <- dummy_df |>
       any_table(rows       = "age",
                 columns    = "sex",
                 values     = weight,
                 statistics = c("pct_value"),
                 pct_value  = list(rate = "Test1 / Test2"),
                 formats    = list(age = age.),
                 print      = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Variable 'Test1' not found in the data frame.",
               info = "any_table outputs sum values with only invalid pct_value statistic and throws a warning")

expect_equal(names(result_list[[1]]), c("row.label", "var1", "weight_sum_1",
                                        "weight_sum_2", "weight_sum_NA"), info = "any_table outputs sum values with only invalid pct_value statistic and throws a warning")


# any_table aborts with missing variable combination in pre summarised data
result_list <- sum_df2 |>
       any_table(rows    = c("year", "age"),
                 columns = "sex",
                 values  = weight_sum,
                 print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "The variable combination of '",
             info = "any_table aborts with missing variable combination in pre summarised data")


# Combine tables into a single workbook aborts, if no any_table or export_with_style object was found
temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

result <- combine_into_workbook(1, style = excel_output_style(save_path = dirname(temp_file),
                                                              file      = basename(temp_file)))
expect_error(print_stack_as_messages("ERROR"), "Unknown object found. Provide <any_table> or <export_with_style> results.",
             info = "Combine tables into a single workbook aborts, if no any_table object was found")

result <- combine_into_workbook(list(1), style = excel_output_style(save_path = dirname(temp_file),
                                                                    file      = basename(temp_file)))
expect_error(print_stack_as_messages("ERROR"), "Unknown object found. Provide <any_table> or <export_with_style> results.",
             info = "Combine tables into a single workbook aborts, if no any_table or export_with_style object was found")

expect_true(!file.exists(temp_file), info = "Combine tables into a single workbook aborts, if no any_table or export_with_style object was found")


# any_table aborts with no valid values after calculating the results
dummy_df |>
    any_table(rows       = "year",
              columns    = "sex",
              values     = weight,
              statistics = "pct_value",
              pct_value  = list(sex = "test",
                                age = "test"),
              print      = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Variable 'age' not found in the data frame",
            info = "any_table aborts with no valid values after calculating the results")
expect_warning(print_stack_as_messages("WARNING"), "Subsetting variable 'sex' by 'test' results in an empty data frame",
            info = "any_table aborts with no valid values after calculating the results")
expect_error(print_stack_as_messages("ERROR"), "After calculating the results, there are no valid values",
            info = "any_table aborts with no valid values after calculating the results")


# any_table aborts, if invalid workbook is passed
result_list <- dummy_df |>
    any_table(rows     = "age",
              columns  = "sex",
              workbook = list("test" = "test"),
              print    = FALSE)

expect_error(print_stack_as_messages("ERROR"), "Workbook object is invalid. You have to provide a workbook object",
             info = "any_table aborts, if invalid workbook is passed")


set_style_options(as_heatmap = FALSE)
set_no_print()

set_no_print(TRUE)

reset_style_options()
default_style_options <- get_style_options()
default_labels        <- get_labels()
default_print         <- get_print()
default_monitor       <- get_monitor()
default_na            <- get_na.rm()
default_print_miss    <- get_print_miss()
default_output        <- get_output()
default_titles        <- get_titles()
default_footnotes     <- get_footnotes()
default_threads       <- get_threads()


# Get global style options
expect_equal(default_style_options, get_style_options(), info = "Get global style options")


# Set global style options
set_style_options(save_path = "C:/MyPath/")

new_options <- get_style_options()

expect_equal(default_style_options[["save_path"]], NULL, info = "Set global style options")
expect_equal(new_options[["save_path"]], "C:/MyPath/", info = "Set global style options")

set_style_options(save_path = NULL)

new_options <- get_style_options()

expect_equal(new_options[["save_path"]], NULL, info = "Set global style options")


# Reset global style options
set_style_options(save_path = "C:/MyPath/")

new_options <- get_style_options()

expect_equal(new_options[["save_path"]], "C:/MyPath/", info = "Reset global style options")

reset_style_options()

reset_options <- get_style_options()
expect_equal(reset_options[["save_path"]], NULL, info = "Reset global style options")


# Close file in global options
set_style_options(file = "MyFile.xlsx")

new_options <- get_style_options()

expect_equal(new_options[["file"]], "MyFile.xlsx", info = "Close file in global options")

close_file()

closed_file <- get_style_options()
expect_equal(closed_file[["file"]], NULL, info = "Close file in global options")


# Save global style options to file
temp_file <- tempfile(fileext = ".rds")
on.exit(unlink(temp_file), add = TRUE)

set_style_options(title_heights = 1, save_file = temp_file)

expect_true(file.exists(temp_file), info = "Save global style options to file")

new_options <- get_style_options(from_file = temp_file)

expect_equal(new_options[["title_heights"]], 1, info = "Save global style options to file")


# Get global labels
expect_equal(length(default_labels), 2, info = "Get global labels")
expect_true(all(c("var_labels", "stat_labels") %in% names(default_labels)), info = "Get global labels")


# Get global print option
expect_equal(default_print, TRUE, info = "Get global print option")


# Get global monitor option
expect_equal(default_monitor, FALSE, info = "Get global monitor option")


# Get global na.rm option
expect_equal(default_na, FALSE, info = "Get global na.rm option")


# Set global variable labels
set_labels(var1 = "Variable 1",
           pct  = "Percent")

new_options <- get_labels()

expect_equal(new_options[["var_labels"]][["var1"]], "Variable 1", info = "Set global labels")
expect_equal(new_options[["stat_labels"]][["pct"]], "Percent",    info = "Set global labels")


# Set global print option
set_print(FALSE)

new_options <- get_print()

expect_equal(new_options, FALSE, info = "Set global print option")


# Set global monitor option
set_monitor(TRUE)

new_options <- get_monitor()

expect_equal(new_options, TRUE, info = "Set global monitor option")


# Set global na.rm option
set_na.rm(TRUE)

new_options <- get_na.rm()

expect_equal(new_options, TRUE, info = "Set global na.rm option")


# Set global print_miss option
set_print_miss(TRUE)

new_options <- get_print_miss()

expect_equal(new_options, TRUE, info = "Set global print_miss option")


# Set global output option
set_output("excel")

new_options <- get_output()

expect_equal(new_options, "excel", info = "Set global output option")


# Set global titles
set_titles("Title1", "Title2")

new_options <- get_titles()

expect_equal(new_options, c("Title1", "Title2"), info = "Set global titles")


# Set global footnotes
set_footnotes("Footnote1", "Footnote2")

new_options <- get_footnotes()

expect_equal(new_options, c("Footnote1", "Footnote2"), info = "Set global footnotes")


# Set global threads
set_threads(1)

new_options <- get_threads()

expect_equal(new_options, 1, info = "Set global threads")

set_threads(NULL)

new_options <- get_threads()

expect_equal(new_options, suppressMessages(fst::threads_fst()), info = "Set global threads")


# Reset global options
reset_qol_options()

new_print      <- get_print()
new_monitor    <- get_monitor()
new_na         <- get_na.rm()
new_print_miss <- get_print_miss()
new_output     <- get_output()
new_titles     <- get_titles()
new_footnotes  <- get_footnotes()
new_threads    <- get_threads()

expect_true(default_print      == new_print, info = "Reset global options")
expect_true(default_monitor    == new_monitor, info = "Reset global options")
expect_true(default_na         == new_na, info = "Reset global options")
expect_true(default_print_miss == new_print_miss, info = "Reset global options")
expect_true(default_output     == new_output, info = "Reset global options")
expect_equal(new_titles, NULL, info = "Reset global options")
expect_equal(new_footnotes, NULL, info = "Reset global options")
expect_equal(new_threads, suppressMessages(fst::threads_fst()), info = "Reset global options")

###############################################################################
# Warning checks
###############################################################################


# Warning on setting wrong logical style in global options
set_style_options(filters = 1)

expect_warning(print_stack_as_messages("WARNING"), "'filters' must be <logical>. Option will be omitted.", info = "Warning on setting wrong logical style in global options")


# Warning on setting wrong numeric style in global options
set_style_options(start_row = "1")

expect_warning(print_stack_as_messages("WARNING"), "'start_row' must be <numeric>. Option will be omitted.", info = "Warning on setting wrong numeric style in global options")


# Warning on setting wrong character style in global options
set_style_options(sheet_name = 1)

expect_warning(print_stack_as_messages("WARNING"), "'sheet_name' must be <character>. Option will be omitted.", info = "Warning on setting wrong character style in global options")


# Warning on setting wrong color style in global options
set_style_options(header_back_color = 1)

expect_warning(print_stack_as_messages("WARNING"), "'header_back_color' must be a 6 character <hex code>. Option will be omitted", info = "Warning on setting wrong color style in global options")


# Warning on setting non existent style in global options
set_style_options(test = 1)

expect_warning(print_stack_as_messages("WARNING"), "'test' is not a valid style option. See", info = "Warning on setting non existent style in global options")

###############################################################################
# Abort checks
###############################################################################


# Abort setting global style options on empty list
set_style_options()
expect_error(print_stack_as_messages("ERROR"), "Empty list found.", info = "Abort setting global style options on empty list")


# Abort setting global labels on list without names.
set_labels(1)
expect_error(print_stack_as_messages("ERROR"), "List must have names to set labels.", info = "Abort setting global labels on list without names.")


# Abort setting global options on empty list
set_print(1)
expect_error(print_stack_as_messages("ERROR"), "Print option can only be TRUE or FALSE. Global option remains unchanged.", info = "Abort setting global options on empty list")

set_monitor(1)
expect_error(print_stack_as_messages("ERROR"), "Monitor option can only be TRUE or FALSE. Global option remains unchanged.", info = "Abort setting global options on empty list")

set_na.rm(1)
expect_error(print_stack_as_messages("ERROR"), "NA removal option can only be TRUE or FALSE. Global option remains unchanged.", info = "Abort setting global options on empty list")

set_print_miss(1)
expect_error(print_stack_as_messages("ERROR"), "Print missing categories option can only be TRUE or FALSE. Global option remains unchanged.", info = "Abort setting global options on empty list")

set_output(1)
expect_error(print_stack_as_messages("ERROR"), "Output can only be 'console', 'text', 'excel', 'excel_nostyle', 'html' or 'excel_html'. Global option remains unchanged.", info = "Abort setting global options on empty list")

set_titles(1)
expect_error(print_stack_as_messages("ERROR"), "Titles must be provided as character. Global titles remain unchanged.", info = "Abort setting global options on empty list")

set_footnotes(1)
expect_error(print_stack_as_messages("ERROR"), "Footnotes must be provided as character. Global footnotes remain unchanged.", info = "Abort setting global options on empty list")

set_threads("")
expect_error(print_stack_as_messages("ERROR"), "Number of used threads must be provided as integer value. Global number of used threads remains unchanged.", info = "Abort setting global options on empty list")


set_no_print()

set_no_print(TRUE)

reset_style_options()
default_style_options <- get_style_options()
default_labels        <- get_labels()
default_print         <- get_print()
default_monitor       <- get_monitor()
default_na            <- get_na.rm()
default_print_miss    <- get_print_miss()
default_output        <- get_output()
default_titles        <- get_titles()
default_footnotes     <- get_footnotes()
default_threads       <- get_threads()


# Get global style options
expect_equal(default_style_options, get_style_options(), info = "Get global style options")


# Set global style options
set_style_options(save_path = "C:/MyPath/")

new_options <- get_style_options()

expect_equal(default_style_options[["save_path"]], NULL, info = "Set global style options")
expect_equal(new_options[["save_path"]], "C:/MyPath/", info = "Set global style options")


# Reset global style options
set_style_options(save_path = "C:/MyPath/")

new_options <- get_style_options()

expect_equal(new_options[["save_path"]], "C:/MyPath/", info = "Reset global style options")

reset_style_options()

reset_options <- get_style_options()
expect_equal(reset_options[["save_path"]], NULL, info = "Reset global style options")


# Close file in global options
set_style_options(file = "MyFile.xlsx")

new_options <- get_style_options()

expect_equal(new_options[["file"]], "MyFile.xlsx", info = "Close file in global options")

close_file()

closed_file <- get_style_options()
expect_equal(closed_file[["file"]], NULL, info = "Close file in global options")


# Save global style options to file
temp_file <- tempfile(fileext = ".rds")
on.exit(unlink(temp_file), add = TRUE)

set_style_options(title_heights = 1, save_file = temp_file)

expect_true(file.exists(temp_file), info = "Save global style options to file")

new_options <- get_style_options(from_file = temp_file)

expect_equal(new_options[["title_heights"]], 1, info = "Save global style options to file")



# Get global labels
expect_equal(length(default_labels), 2, info = "Get global labels")
expect_true(all(c("var_labels", "stat_labels") %in% names(default_labels)), info = "Get global labels")


# Get global print option
expect_equal(default_print, TRUE, info = "Get global print option")


# Get global monitor option
expect_equal(default_monitor, FALSE, info = "Get global monitor option")


# Get global na.rm option
expect_equal(default_na, FALSE, info = "Get global na.rm option")


# Set global variable labels
set_labels(var1 = "Variable 1",
           pct  = "Percent")

new_options <- get_labels()

expect_equal(new_options[["var_labels"]][["var1"]], "Variable 1", info = "Set global labels")
expect_equal(new_options[["stat_labels"]][["pct"]], "Percent",    info = "Set global labels")


# Set global print option
set_print(FALSE)

new_options <- get_print()

expect_equal(new_options, FALSE, info = "Set global print option")


# Set global monitor option
set_monitor(TRUE)

new_options <- get_monitor()

expect_equal(new_options, TRUE, info = "Set global monitor option")


# Set global na.rm option
set_na.rm(TRUE)

new_options <- get_na.rm()

expect_equal(new_options, TRUE, info = "Set global na.rm option")


# Set global print_miss option
set_print_miss(TRUE)

new_options <- get_print_miss()

expect_equal(new_options, TRUE, info = "Set global print_miss option")


# Set global output option
set_output("excel")

new_options <- get_output()

expect_equal(new_options, "excel", info = "Set global output option")


# Set global titles
set_titles("Title1", "Title2")

new_options <- get_titles()

expect_equal(new_options, c("Title1", "Title2"), info = "Set global titles")


# Set global footnotes
set_footnotes("Footnote1", "Footnote2")

new_options <- get_footnotes()

expect_equal(new_options, c("Footnote1", "Footnote2"), info = "Set global footnotes")


# Set global threads
set_threads(1)

new_options <- get_threads()

expect_equal(new_options, 1, info = "Set global threads")

set_threads(NULL)

new_options <- get_threads()

expect_equal(new_options, suppressMessages(fst::threads_fst()), info = "Set global threads")


# Reset global options
reset_qol_options()

new_print      <- get_print()
new_monitor    <- get_monitor()
new_na         <- get_na.rm()
new_print_miss <- get_print_miss()
new_output     <- get_output()
new_titles     <- get_titles()
new_footnotes  <- get_footnotes()
new_threads    <- get_threads()

expect_true(default_print      == new_print, info = "Reset global options")
expect_true(default_monitor    == new_monitor, info = "Reset global options")
expect_true(default_na         == new_na, info = "Reset global options")
expect_true(default_print_miss == new_print_miss, info = "Reset global options")
expect_true(default_output     == new_output, info = "Reset global options")
expect_equal(new_titles, NULL, info = "Reset global options")
expect_equal(new_footnotes, NULL, info = "Reset global options")
expect_equal(new_threads, suppressMessages(fst::threads_fst()), info = "Reset global options")

###############################################################################
# Warning checks
###############################################################################


# Warning on setting wrong logical style in global options
set_style_options(filters = 1)

expect_warning(print_stack_as_messages("WARNING"), "'filters' must be <logical>. Option will be omitted.", info = "Warning on setting wrong logical style in global options")


# Warning on setting wrong numeric style in global options
set_style_options(start_row = "1")

expect_warning(print_stack_as_messages("WARNING"), "'start_row' must be <numeric>. Option will be omitted.", info = "Warning on setting wrong numeric style in global options")


# Warning on setting wrong character style in global options
set_style_options(sheet_name = 1)

expect_warning(print_stack_as_messages("WARNING"), "'sheet_name' must be <character>. Option will be omitted.", info = "Warning on setting wrong character style in global options")


# Warning on setting wrong color style in global options
set_style_options(header_back_color = 1)

expect_warning(print_stack_as_messages("WARNING"), "'header_back_color' must be a 6 character <hex code>. Option will be omitted", info = "Warning on setting wrong color style in global options")


# Warning on setting non existent style in global options
set_style_options(test = 1)

expect_warning(print_stack_as_messages("WARNING"), "'test' is not a valid style option. See", info = "Warning on setting non existent style in global options")

###############################################################################
# Abort checks
###############################################################################


# Abort setting global style options on empty list
set_style_options()
expect_error(print_stack_as_messages("ERROR"), "Empty list found.", info = "Abort setting global style options on empty list")


# Abort setting global labels on list without names.
set_labels(1)
expect_error(print_stack_as_messages("ERROR"), "List must have names to set labels.", info = "Abort setting global labels on list without names.")


# Abort setting global options on empty list
set_print(1)
expect_error(print_stack_as_messages("ERROR"), "Print option can only be TRUE or FALSE. Global option remains unchanged.", info = "Abort setting global options on empty list")

set_monitor(1)
expect_error(print_stack_as_messages("ERROR"), "Monitor option can only be TRUE or FALSE. Global option remains unchanged.", info = "Abort setting global options on empty list")

set_na.rm(1)
expect_error(print_stack_as_messages("ERROR"), "NA removal option can only be TRUE or FALSE. Global option remains unchanged.", info = "Abort setting global options on empty list")

set_print_miss(1)
expect_error(print_stack_as_messages("ERROR"), "Print missing categories option can only be TRUE or FALSE. Global option remains unchanged.", info = "Abort setting global options on empty list")

set_output(1)
expect_error(print_stack_as_messages("ERROR"), "Output can only be 'console', 'text', 'excel' or 'excel_nostyle'. Global option remains unchanged.", info = "Abort setting global options on empty list")

set_titles(1)
expect_error(print_stack_as_messages("ERROR"), "Titles must be provided as character. Global titles remain unchanged.", info = "Abort setting global options on empty list")

set_footnotes(1)
expect_error(print_stack_as_messages("ERROR"), "Footnotes must be provided as character. Global footnotes remain unchanged.", info = "Abort setting global options on empty list")

set_threads("")
expect_error(print_stack_as_messages("ERROR"), "Number of used threads must be provided as integer value. Global number of used threads remains unchanged.", info = "Abort setting global options on empty list")


set_no_print()

set_no_print(TRUE)

dummy_df  <- dummy_data(3000)
dummy_big <- dummy_data(10000)

dummy_df[["binary"]] <- replicate(nrow(dummy_df), {
    paste0(sample(0:1, 2, replace = TRUE), collapse = "")
})

sum_df   <- dummy_df |>
    summarise_plus(class      = c(year, sex),
                   values     = weight,
                   statistics = c("sum"),
                   nesting    = "deepest",
                   na.rm      = TRUE)

sum_df2  <- dummy_df |>
    summarise_plus(class      = c(year, sex, age),
                   values     = weight,
                   statistics = c("sum"),
                   nesting    = "deepest",
                   na.rm      = TRUE)

age. <- discrete_format(
    "under 50"    = 0:49,
    "50 and more" = 50:100)

state. <- discrete_format(
    "West" = 1:10,
    "East" = 11:16)

sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

education. <- discrete_format(
    "low"    = "low",
    "middle" = "middle",
    "high"   = "high")


# Simplest form of any_table
result_list <- dummy_big |>
      any_table(rows    = "age",
                columns = "sex",
                values  = weight,
                print   = FALSE)

expect_inherits(result_list, "qol_table", info = "Simplest form of any_table")
expect_equal(length(result_list), 3, info = "Simplest form of any_table")
expect_equal(names(result_list[[1]]), c("row.label", "var1", "weight_sum_1",
                                        "weight_sum_2", "weight_sum_NA"), info = "Simplest form of any_table")
expect_equal(result_list[[1]][["var1"]][1:90], as.character(0:89), info = "Simplest form of any_table")
expect_equal(result_list[[1]][["var1"]][1:90], as.character(0:89), info = "Simplest form of any_table")


# any_table with combinations
result_list <- dummy_df |>
      any_table(rows    = "age + education",
                columns = "sex + year",
                values  = income,
                weight  = weight,
                output  = "excel_nostyle",
                print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with combinations")
expect_equal(length(result_list), 3, info = "any_table with combinations")


# any_table with multiple combinations
result_list <- dummy_df |>
      any_table(rows    = c("age", "age + education"),
                columns = c("sex + year", "sex"),
                values  = weight,
                output  = "excel_nostyle",
                print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with multiple combinations")
expect_equal(length(result_list), 3, info = "any_table with multiple combinations")


# any_table many combinations don't break
result_list <- dummy_df |>
        any_table(rows    = c("age", "age + education", "state",
                              "state + age", "education + age"),
                  columns = c("year", "sex + year", "sex"),
                  values  = weight,
                  output  = "excel_nostyle",
                  print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table many combinations don't break")
expect_equal(length(result_list), 3, info = "any_table many combinations don't break")


# any_table with linked titles and footnotes
txt_file  <- system.file("extdata", "qol_example_data_txt.txt",   package = "qol")

set_titles("Hello world1",
           "Hello world2 link: https://cran.r-project.org/",
           "Hello world3 cell: A8",
           "Hello world4 file: txt_file")
set_footnotes("This is a footnote1",
              "This is a footnote2 link: https://cran.r-project.org/",
              "This is a footnote3 cell: A8",
              "This is a footnote4 file: txt_file")

set_style_options(as_heatmap = TRUE)

set_style_options(header_stat_merging = "all")
result_list <- dummy_df |>
      any_table(rows      = "age",
                columns   = "sex",
                values    = weight,
                titles    = "Hello world link: https://cran.r-project.org/",
                footnotes = "This is a footnote link: https://cran.r-project.org/",
                print     = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with linked titles and footnotes")
expect_equal(length(result_list), 3, info = "any_table with linked titles and footnotes")


# any_table with multiple titles and footnotes
set_titles("Hello world1", "Hello world2")
set_footnotes("This is a footnote1", "This is a footnote2")

set_style_options(header_stat_merging = "none",
                  title_font_color    = c("FF00FF", "00FF00"),
                  title_font_size     = c(10, 11),
                  title_font_bold     = c(TRUE, FALSE),
                  footnote_font_color = c("FF00FF", "00FF00"),
                  footnote_font_size  = c(10, 11),
                  footnote_font_bold  = c(TRUE, FALSE))

result_list <- dummy_df |>
    any_table(rows      = "age",
              columns   = "sex",
              values    = weight,
              print     = FALSE)

reset_style_options()

expect_inherits(result_list, "qol_table", info = "any_table with multiple titles and footnotes")
expect_equal(length(result_list), 3, info = "any_table with multiple titles and footnotes")


# any_table with variable and stat labels
result_list <- dummy_df |>
      any_table(rows        = "age",
                columns     = "sex",
                values      = weight,
                var_labels  = list(age = "Single ages", sex = "Sex", weight = "Population"),
                stat_labels = list(sum = "Counts"),
                box         = "Test",
                print       = FALSE)

expect_true("Single ages" %in% result_list[["table"]][["row.label"]], info = "any_table with variable and stat labels")


# any_table with removed variable and stat labels
result_list <- dummy_df |>
      any_table(rows        = "age",
                columns     = "sex",
                values      = weight,
                var_labels  = list(age = "", sex = "", weight = ""),
                stat_labels = list(sum = ""),
                print       = FALSE)

expect_true(!"row.label" %in% names(result_list[["table"]]), info = "any_table with removed variable and stat labels")


# any_table with different percentages
result_list <- dummy_df |>
      any_table(rows       = "age",
                columns    = "sex",
                values     = c(probability, weight),
                statistics = c("sum", "pct_total"),
                pct_group  = c("age", "sex"),
                pct_value  = list(rate = "probability / weight",
                                  sex  = 1),
                output     = "excel_nostyle",
                print      = FALSE)

expect_true(all(c("weight_pct_group_age_1", "weight_pct_total_1",
                  "rate_pct_value_1", "sex_pct_value_1") %in% names(result_list[["table"]])),
            info = "any_table with different percentages")


# any_table with keywords for row and column percentages
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = "sex",
                  values     = c(probability, weight),
                  statistics = c("pct_group"),
                  pct_group  = c("row_pct", "col_pct"),
                  output     = "excel_nostyle",
                  print      = FALSE)

expect_true(all(c("weight_pct_group_row_1", "probability_pct_group_row_2",
                  "weight_pct_group_col_1", "probability_pct_group_col_2")
                %in% names(result_list[["table"]])), info = "any_table with keywords for row and column percentages")


# any_table with block row percentages
result_list <- dummy_df |>
    any_table(rows      = "age + (sex, education)",
              columns   = "state",
              values    = weight,
              pct_block = "rows",
              output    = "excel_nostyle",
              formats   = list(state = state., age = age., sex = sex., education = education.),
              print     = FALSE,
              na.rm     = TRUE)

result_df <- result_list[["table"]] |> if.(var2 == "Total")
result_df[["weight_pct_block_rows_West"]] <- round(result_df[["weight_pct_block_rows_West"]])
result_df[["weight_pct_block_rows_East"]] <- round(result_df[["weight_pct_block_rows_East"]])

expect_equal(collapse::funique(result_df[["weight_pct_block_rows_West"]]), 100, info = "any_table with block row percentages")
expect_equal(collapse::funique(result_df[["weight_pct_block_rows_East"]]), 100, info = "any_table with block row percentages")


# any_table with block column percentages
result_list <- dummy_df |>
    any_table(rows      = "state",
              columns   = "age + (sex, education)",
              values    = weight,
              pct_block = "columns",
              output    = "excel_nostyle",
              formats   = list(state = state., age = age., sex = sex., education = education.),
              order_by  = "blocks",
              print     = FALSE,
              na.rm     = TRUE)

result_df <- result_list[["table"]]
result_df[["weight_pct_block_columns_under 50_Total"]] <- round(result_df[["weight_pct_block_columns_under 50_Total"]])

expect_equal(collapse::funique(result_df[["weight_pct_block_columns_under 50_Total"]]), 100, info = "any_table with block column percentages")
expect_equal(names(result_df),
             c("row.label", "var1", "weight_pct_block_columns_under 50_Total",
               "weight_pct_block_columns_under 50_Male", "weight_pct_block_columns_under 50_Female", "weight_pct_block_columns_under 50_low",
               "weight_pct_block_columns_under 50_middle", "weight_pct_block_columns_under 50_high", "weight_pct_block_columns_50 and more_Total",
               "weight_pct_block_columns_50 and more_Male", "weight_pct_block_columns_50 and more_Female", "weight_pct_block_columns_50 and more_low",
               "weight_pct_block_columns_50 and more_middle", "weight_pct_block_columns_50 and more_high"), info = "any_table with block order")


# any_table can expand combinations
result_list <- dummy_df |>
    any_table(rows       = "age + (state, first_person + education)",
              columns    = "sex",
              values     = weight,
              output     = "excel_nostyle",
              formats    = list(state = state., age = age., sex = sex., education = education.),
              print      = FALSE)

expect_true(all(c("West", "East", "0", "1") %in% result_list[["table"]][["var2"]]), info = "any_table can expand combinations")
expect_true(all(c("low", "middle", "high", "") %in% result_list[["table"]][["var3"]]), info = "any_table can expand combinations")


# any_table with a lot of statistics doesn't break
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = "sex",
                  values     = weight,
                  statistics = c("freq", "freq_g0", "mean", "median", "mode", "min", "max",
                                 "first", "last", "sum_wgt", "p1", "p99", "sd", "variance",
                                 "missing"),
                  output     = "excel_nostyle",
                  print      = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with a lot of statistics doesn't break")
expect_equal(length(result_list), 3, info = "any_table with a lot of statistics doesn't break")


# any_table with interleaved order
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = "sex",
                  values     = weight,
                  statistics = c("sum", "freq", "missing"),
                  order_by   = "interleaved",
                  output     = "excel_nostyle",
                  formats    = list(age = age., sex = sex.),
                  print      = FALSE,
                  na.rm      = TRUE)

expect_equal(names(result_list[[1]]),
               c("row.label", "var1", "weight_sum_Total", "weight_freq_Total", "weight_missing_Total",
                 "weight_sum_Male", "weight_freq_Male", "weight_missing_Male",
                 "weight_sum_Female", "weight_freq_Female", "weight_missing_Female"),
             info = "any_table with interleaved order")


# any_table with values order
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = c("sex", "state"),
                  values     = c(weight, income),
                  statistics = "sum",
                  order_by   = "values",
                  output     = "excel_nostyle",
                  formats    = list(age = age., sex = sex., state = state.),
                  print      = FALSE,
                  na.rm      = TRUE)

expect_equal(names(result_list[[1]]),
             c("row.label", "var1", "weight_sum_Total", "weight_sum_Male", "weight_sum_Female",
               "weight_sum_West", "weight_sum_East",
               "income_sum_Total", "income_sum_Male", "income_sum_Female", "income_sum_West",
               "income_sum_East"), info = "any_table with values order")


# any_table with columns order
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = c("sex", "state"),
                  values     = c(weight, income),
                  statistics = "sum",
                  order_by   = "columns",
                  output     = "excel_nostyle",
                  formats    = list(age = age., sex = sex., state = state.),
                  print      = FALSE,
                  na.rm      = TRUE)

expect_equal(names(result_list[[1]]),
             c("row.label", "var1", "weight_sum_Total", "weight_sum_Male", "weight_sum_Female",
               "income_sum_Total", "income_sum_Male", "income_sum_Female",
               "weight_sum_West", "weight_sum_East", "income_sum_West",
               "income_sum_East"), info = "any_table with columns order")


# any_table with by variables
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = "sex",
                  values     = weight,
                  by         = education,
                  print_miss = TRUE,
                  print      = FALSE)

expect_true("BY" %in% names(result_list[[1]]), info = "any_table with by variables")
expect_equal(length(unique(result_list[[1]][["BY"]])), 1, info = "any_table with by variables")


# any_table with multiple by variables
result_list <- dummy_df |>
        any_table(rows       = "age",
                  columns    = "sex",
                  values     = weight,
                  by         = c(education, year),
                  output     = "excel_nostyle",
                  print_miss = TRUE,
                  print      = FALSE)

expect_true("BY" %in% names(result_list[["table"]]), info = "any_table with multiple by variables")
expect_equal(length(unique(result_list[["table"]][["BY"]])), 2, info = "any_table with multiple by variables")


# any_table with by variables and multiple row and column variables
result_list <- dummy_df |>
       any_table(rows    = c("first_person", "state", "first_person + state"),
                 columns = c("sex", "education", "sex + education"),
                 values  = weight,
                 by      = year,
                 print   = FALSE,
                 output  = "excel_nostyle",
                 na.rm   = TRUE)

expect_true("BY" %in% names(result_list[["table"]]), info = "any_table with by variables and multiple row and column variables")


# any_table with by variables as subheaders
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  by      = education,
                  style   = excel_output_style(by_as_subheaders = TRUE),
                  print   = FALSE)

expect_true("BY" %in% names(result_list[[1]]), info = "any_table with by variables as subheaders")
expect_equal(length(unique(result_list[[1]][["BY"]])), 1, info = "any_table with by variables as subheaders")


# any_table with by variables throws a warning if by is also part of rows or columns
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  by      = sex,
                  print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "The provided <by> variable 'sex' is also part of",
             info = "any_table with by variables throws a warning if by is also part of rows or columns")


# any_table with NAs removed
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  output  = "excel_nostyle",
                  na.rm   = TRUE,
                  print   = FALSE)

expect_true(sum(is.na(result_list[["table"]][["var1"]])) == 0, info = "any_table with NAs removed")


# any_table with applied single discrete labels
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  formats = list(age = discrete_format(
                      "under 18"       = 0:17,
                      "18 to under 25" = 18:24,
                      "25 to under 55" = 25:54,
                      "55 to under 65" = 55:64,
                      "65 and older"   = 65:100)),
                  output = "excel_nostyle",
                  print  = FALSE)

expect_true(all(c("under 18", "18 to under 25", "25 to under 55",
                  "55 to under 65", "65 and older")
                %in% result_list[["table"]][["var1"]]), info = "any_table with applied single discrete labels")


# any_table with applied discrete multilabels
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  formats = list(sex = discrete_format(
                      "Total"  = 1:2,
                      "Male"   = 1,
                      "Female" = 2),
                      age = discrete_format(
                      "Total"          = 0:100,
                      "under 18"       = 0:17,
                      "18 to under 25" = 18:24,
                      "25 to under 55" = 25:54,
                      "55 to under 65" = 55:64,
                      "65 and older"   = 65:100)),
                  output = "excel_nostyle",
                  print  = FALSE)

expect_true(all(c("under 18", "18 to under 25", "25 to under 55",
                  "55 to under 65", "65 and older")
                %in% result_list[["table"]][["var1"]]), info = "any_table with applied discrete multilabels")


# any_table with applied interval multilabels
result_list <- dummy_df |>
        any_table(rows    = "income",
                  columns = "sex",
                  values  = weight,
                  formats = list(sex = discrete_format(
                      "Total"  = 1:2,
                      "Male"   = 1,
                      "Female" = 2),
                      income = interval_format(
                          "Total"              =    0:100000,
                          "below 500"          =    0:500,
                          "500 to under 1000"  =  500:1000,
                          "1000 to under 2000" = 1000:2000,
                          "2000 and more"      = 2000:100000)),
                  output = "excel_nostyle",
                  print  = FALSE)

expect_true(all(c("Total", "below 500", "2000 and more")
                %in% result_list[["table"]][["var1"]]), info = "any_table with applied interval multilabels")


# any_table can silence format expressions
sex. <- discrete_format(
    "!Total" = 1:2,
    "Male"   = 1,
    "Female" = 2)

result_list <- dummy_df |>
    any_table(rows       = "first_person",
              columns    = "sex",
              values     = weight,
              statistics = "sum",
              output     = "excel_nostyle",
              formats    = list(sex = sex.),
              print      = FALSE)

expect_false(any(grepl("!", names(result_list[["table"]]))), info = "any_table can silence format expressions")


# any_table able to apply format on numeric values stored as character (short route)
binary. <- discrete_format(
    "binary1" = c("00", "01"),
    "binary2" = c("10", "11"))

result_list <- dummy_df |>
    any_table(rows    = "binary",
              columns = "sex",
              values  = weight,
              formats = list(binary = binary.),
              output  = "excel_nostyle",
              print   = FALSE)

expect_equal(result_list[[1]][["var1"]], c("binary1", "binary2"), info = "any_table able to apply format on numeric values stored as character (short route)")


# any_table doesn't convert numeric values stored as character (short route)
result_list <- dummy_df |>
        any_table(rows    = "binary",
                  columns = "sex",
                  values  = weight,
                  output  = "excel_nostyle",
                  print   = FALSE)

expect_equal(result_list[[1]][["var1"]], c("00", "01", "10", "11"), info = "any_table doesn't convert numeric values stored as character (short route)")


# any_table able to apply format on numeric values stored as character (long route)
binary. <- discrete_format(
    "binary1" = c("00", "01"),
    "binary2" = c("10", "11"))

result_list <- dummy_df |>
        any_table(rows       = "binary",
                  columns    = "sex",
                  statistics = "mean",
                  values     = weight,
                  formats    = list(binary = binary.),
                  output     = "excel_nostyle",
                  print      = FALSE)

expect_equal(result_list[[1]][["var1"]], c("binary1", "binary2"), info = "any_table able to apply format on numeric values stored as character (long route)")


# any_table doesn't convert numeric values stored as character (long route)
result_list <- dummy_df |>
        any_table(rows       = "binary",
                  columns    = "sex",
                  statistics = "mean",
                  values     = weight,
                  output     = "excel_nostyle",
                  print      = FALSE)

expect_equal(result_list[[1]][["var1"]], c("00", "01", "10", "11"), info = "any_table doesn't convert numeric values stored as character (long route)")


# any_table with fixed column headers
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  style   = excel_output_style(freeze_col_header = TRUE),
                  na.rm   = TRUE,
                  output  = "excel_nostyle",
                  print   = FALSE)

expect_true(sum(is.na(result_list[["table"]][["var1"]])) == 0, info = "any_table with fixed column headers")


# any_table with fixed row headers
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  style   = excel_output_style(freeze_row_header = TRUE),
                  na.rm   = TRUE,
                  output  = "excel_nostyle",
                  print   = FALSE)

expect_true(sum(is.na(result_list[["table"]][["var1"]])) == 0, info = "any_table with fixed row headers")


# any_table with fixed column and row headers
result_list <- dummy_df |>
        any_table(rows    = "age",
                  columns = "sex",
                  values  = weight,
                  style   = excel_output_style(freeze_col_header = TRUE,
                                               freeze_row_header = TRUE,
                                               background_color  = "FF00FF"),
                  na.rm   = TRUE,
                  output  = "excel_nostyle",
                  print   = FALSE)

expect_true(sum(is.na(result_list[["table"]][["var1"]])) == 0, info = "any_table with fixed column and row headers")


# any_table warning with wrong output format
result_list <- dummy_df |>
                   any_table(rows    = "age",
                             columns = "sex",
                             values  = weight,
                             output  = "Test",
                             print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "<Output> format 'Test' not available.", info = "any_table warning with wrong output format")


# any_table warning with wrong output format
result_list <- dummy_df |>
                   any_table(rows     = "age",
                             columns  = "sex",
                             values   = weight,
                             order_by = "test",
                             output   = "excel_nostyle",
                             print    = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "<Order by> option 'test' doesn't exist", info = "any_table warning with wrong output format")


# any_table with pre summarised data
result_list <- sum_df |>
   any_table(rows       = "year",
             columns    = "sex",
             values     = weight_sum,
             output     = "excel_nostyle",
             print      = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with pre summarised data")
expect_equal(length(result_list), 3, info = "any_table with pre summarised data")


# any_table with pre summarised data and by variables
result_list <- sum_df2 |>
                any_table(rows       = "age",
                          columns    = "year",
                          by         = "sex",
                          values     = weight_sum,
                          output     = "excel_nostyle",
                          print_miss = TRUE,
                          print      = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with pre summarised data and by variables")
expect_equal(length(result_list), 3, info = "any_table with pre summarised data and by variables")


# any_table with no column variables
result_list <- dummy_df |>
       any_table(rows    = "age",
                 values  = weight,
                 output  = "excel_nostyle",
                 print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table with no column variables")
expect_equal(length(result_list), 3, info = "any_table with no column variables")


# any_table throws a warning, if invalid statistic specified
result_list <- dummy_df |>
                   any_table(rows       = "age",
                             columns    = "sex",
                             values     = weight,
                             statistics = c("test", "sum"),
                             output     = "excel_nostyle",
                             print      = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "<Statistic> 'test' is invalid and will be omitted.",
               info = "any_table throws a warning, if invalid statistic specified")


# any_table throws a warning, if no valid statistic specified
result_list <- dummy_df |>
                   any_table(rows       = "age",
                             columns    = "sex",
                             values     = weight,
                             statistics = "test",
                             output     = "excel_nostyle",
                             print      = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "No valid <statistic> selected. 'sum' will be used.",
               info = "any_table throws a warning, if no valid statistic specified")


# Save any_table as Excel file
temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              values  = weight,
              output  = "excel_nostyle",
              style   = excel_output_style(save_path = dirname(temp_file),
                                           file      = basename(temp_file)))

expect_true(file.exists(temp_file), info = "Save any_table as Excel file")


# Combine tables into a single workbook
my_style <- excel_output_style(sheet_name = "tab1")

tab1 <- dummy_df |>
     any_table(rows    = "age",
               columns = "sex",
               values  = weight,
               output  = "excel_nostyle",
               style   = my_style,
               print   = FALSE)

my_style <- my_style |> modify_output_style(sheet_name = "tab2")

tab2 <- dummy_df |>
     any_table(rows    = "age",
               columns = "sex",
               values  = weight,
               by      = education,
               output  = "excel_nostyle",
               style   = my_style,
               print   = FALSE)

temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

result <- combine_into_workbook(tab1, tab2, file = temp_file)

expect_inherits(result, c("wbWorkbook", "R6"), info = "Combine tables into a single workbook")
expect_true(file.exists(temp_file), info = "Combine tables into a single workbook")

result_list <- dummy_df |>
    any_table(rows     = "age",
              columns  = "sex",
              values   = weight,
              output   = "excel_nostyle",
              workbook = tab1,
              style    = my_style,
              print    = FALSE)

expect_inherits(result_list, "qol_table", info = "Combine tables into a single workbook")
expect_equal(length(result_list), 3, info = "Combine tables into a single workbook")


# any_table throws a warning with missing statistic extension in pre summarised data
result_list <- sum_df |>
       any_table(rows    = "year",
                 columns = "sex",
                 values  = DEPTH,
                 output  = "excel_nostyle",
                 print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "All <values> variables need to have the <statistic> extension in their variable name.",
               info = "any_table throws a warning with missing statistic extension in pre summarised data")

expect_inherits(result_list, "qol_table", info = "any_table throws a warning with missing statistic extension in pre summarised data")
expect_equal(length(result_list), 3, info = "any_table throws a warning with missing statistic extension in pre summarised data")


# any_table auto generates missing TYPE variable in pre summarised data
result_list <- sum_df |>
       any_table(rows    = "year",
                 columns = "sex",
                 values  = weight_sum,
                 output  = "excel_nostyle",
                 print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table auto generates missing TYPE variable in pre summarised data")
expect_equal(length(result_list), 3, info = "any_table auto generates missing TYPE variable in pre summarised data")


# any_table outputs unweighted results without values variable
result_list <- dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              output  = "excel_nostyle",
              print   = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table outputs unweighted results without values variable")
expect_equal(length(result_list), 3, info = "any_table outputs unweighted results without values variable")


# any_table throws a warning, if invalid format is used
result_list <- dummy_df |>
    any_table(rows    = "age",
              columns = "sex",
              values  = weight,
              output  = "excel_nostyle",
              formats = list(age = age., sex = test),
              print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Format for variable 'sex' does not exist and can't be applied.",
               info = "any_table throws a warning, if invalid format is used")


# any_table can compute values on the fly
result_list <- dummy_df |>
    any_table(rows       = "first_person",
              columns    = "sex",
              values     = "probability",
              statistics = c("sum", "sum_wgt"),
              compute    = list(percent    = probability_sum * 100 / sum_wgt,
                                pct_square = percent ^ 2),
              weight     = weight,
              output     = "excel_nostyle",
              na.rm      = TRUE,
              print      = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table can compute values on the fly")
expect_equal(length(result_list), 3, info = "any_table can compute values on the fly")
expect_equal(names(result_list[[1]]), c("row.label", "var1", "probability_sum_1",
                                        "probability_sum_2", "sum_wgt_1", "sum_wgt_2",
                                        "percent_1", "percent_2", "pct!!!square_1", "pct!!!square_2"),
             info = "any_table can compute values on the fly")


# any_table orders individual variables by name
result_list <- dummy_df |>
    any_table(rows       = "first_person",
              columns    = "sex",
              values     = "probability",
              statistics = c("sum", "sum_wgt"),
              compute    = list(percent    = probability_sum * 100 / sum_wgt,
                                pct_square = percent ^ 2),
              weight     = weight,
              order_by   = c("pct_square", "percent", "sum_wgt"),
              output     = "excel_nostyle",
              na.rm      = TRUE,
              print      = FALSE)

expect_inherits(result_list, "qol_table", info = "any_table orders individual variables by name")
expect_equal(length(result_list), 3, info = "any_table orders individual variables by name")
expect_equal(names(result_list[[1]]), c("row.label", "var1", "pct!!!square_1", "pct!!!square_2",
                                        "percent_1", "percent_2", "sum_wgt_1", "sum_wgt_2",
                                        "probability_sum_1",  "probability_sum_2"),
             info = "any_table orders individual variables by name")


# any_table is able to output specific statistics per variable
result_list <- dummy_df |>
    any_table(rows       = "age",
              columns    = "sex",
              statistics = list("sum"       = c(income, expenses),
                                "pct_group" = weight),
              na.rm      = TRUE,
              print      = FALSE)

expect_equal(collapse::fncol(result_list[[1]]), 8, info = "any_table is able to output specific statistics per variable")
expect_true(all(c("income_sum_1", "income_sum_2", "expenses_sum_1", "expenses_sum_2",
                  "weight_pct_group_sex_1", "weight_pct_group_sex_2") %in% names(result_list[[1]])),
            info = "any_table is able to output specific statistics per variable")
expect_true(!all(c("income_pct_group_1", "income_pct_group_2", "expenses_pct_group_1",
                   "expenses_pct_group_2", "weight_sum_1", "weight_sum_2") %in% names(result_list[[1]])),
            info = "any_table is able to output specific statistics per variable")


###############################################################################
# Abort checks
###############################################################################

# any_table aborts, if column contains a row variable
result_list <- dummy_df |>
                   any_table(rows       = "sex",
                             columns    = "sex",
                             statistics = c("sum"),
                             print      = FALSE)

expect_error(print_stack_as_messages("ERROR"), "The provided <columns> variable '",
             info = "any_table aborts, if column contains a row variable")


# any_table aborts with duplicate column names after pivot
result_list <- dummy_df |>
    any_table(rows    = "education",
              columns = c("sex + age", "age + sex"),
              values  = weight,
              print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "Duplicate <columns> names found",
             info = "any_table aborts with duplicate column names after pivot")


# any_table aborts with none existent row variable
result_list <- dummy_df |>
      any_table(rows    = c("age", "age + test"),
                columns = "sex",
                values  = weight,
                print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "The provided <rows> variable 'test' is not part of",
             info = "any_table aborts with none existent row variable")


# any_table aborts with none existent column variable
result_list <- dummy_df |>
      any_table(rows    = "age",
                columns = c("sex + test"),
                values  = weight,
                print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "The provided <columns> variable 'test' is not part of",
             info = "any_table aborts with none existent column variable")


# any_table aborts with no valid row variables
result_list <- dummy_df |>
       any_table(rows    = "",
                 columns = "sex",
                 values  = weight,
                 print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "No valid <rows> variables provided",
             info = "any_table aborts with no valid row variables")


# any_table aborts with invalid by variable
result_list <- dummy_df |>
       any_table(rows    = "age",
                 columns = "sex",
                 by      = "test",
                 values  = weight,
                 print   = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "The provided <by> variable 'test' is not part of",
               info = "any_table aborts with invalid by variable")


# any_table aborts with row/column variable part of values
result_list <- dummy_df |>
       any_table(rows    = "age",
                 columns = "sex",
                 values  = "sex",
                 print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "The provided <rows>/<columns> variable 'sex' is also part of",
             info = "any_table aborts with row/column variable part of values")


# any_table outputs sum values with only invalid pct_value statistic and throws a warning
result_list <- dummy_df |>
       any_table(rows       = "age",
                 columns    = "sex",
                 values     = weight,
                 statistics = c("pct_value"),
                 pct_value  = list(rate = "Test1 / Test2"),
                 formats    = list(age = age.),
                 print      = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Variable 'Test1' not found in the data frame.",
               info = "any_table outputs sum values with only invalid pct_value statistic and throws a warning")

expect_equal(names(result_list[[1]]), c("row.label", "var1", "weight_sum_1",
                                        "weight_sum_2", "weight_sum_NA"), info = "any_table outputs sum values with only invalid pct_value statistic and throws a warning")


# any_table aborts with missing variable combination in pre summarised data
result_list <- sum_df2 |>
       any_table(rows    = c("year", "age"),
                 columns = "sex",
                 values  = weight_sum,
                 print   = FALSE)

expect_error(print_stack_as_messages("ERROR"), "The variable combination of '",
             info = "any_table aborts with missing variable combination in pre summarised data")


# Combine tables into a single workbook aborts, if no any_table or export_with_style object was found
temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

result <- combine_into_workbook(1, file = temp_file)
expect_error(print_stack_as_messages("ERROR"), "Unknown object found. Provide <any_table> or <export_with_style> results.",
             info = "Combine tables into a single workbook aborts, if no any_table object was found")

result <- combine_into_workbook(list(1), file = temp_file)
expect_error(print_stack_as_messages("ERROR"), "Unknown object found. Provide <any_table> or <export_with_style> results.",
             info = "Combine tables into a single workbook aborts, if no any_table or export_with_style object was found")

expect_true(!file.exists(temp_file), info = "Combine tables into a single workbook aborts, if no any_table or export_with_style object was found")


# any_table aborts with no valid values after calculating the results
dummy_df |>
    any_table(rows       = "year",
              columns    = "sex",
              values     = weight,
              statistics = "pct_value",
              pct_value  = list(sex = "test",
                                age = "test"),
              print      = FALSE)

expect_warning(print_stack_as_messages("WARNING"), "Variable 'age' not found in the data frame",
            info = "any_table aborts with no valid values after calculating the results")
expect_warning(print_stack_as_messages("WARNING"), "Subsetting variable 'sex' by 'test' results in an empty data frame",
            info = "any_table aborts with no valid values after calculating the results")
expect_error(print_stack_as_messages("ERROR"), "After calculating the results, there are no valid values",
            info = "any_table aborts with no valid values after calculating the results")


# any_table aborts, if invalid workbook is passed
result_list <- dummy_df |>
    any_table(rows     = "age",
              columns  = "sex",
              workbook = list("test" = "test"),
              print    = FALSE)

expect_error(print_stack_as_messages("ERROR"), "Workbook object is invalid. You have to provide a workbook object",
             info = "any_table aborts, if invalid workbook is passed")


set_style_options(as_heatmap = FALSE)
set_no_print()

#' Fast And Powerful Yet Simple To Use Summarise
#'
#' @description
#' [summarise_plus()] creates a new aggregated data table with the desired grouping.
#' It can output only the deepest nested combination of the grouping variables (default)
#' or you can also output every possible combination of the grouping variables at once,
#' with just one small change. Besides the normal summary functions like sum, mean
#' or median, you can also calculate their respective weighted version by just
#' setting a weight variable.
#'
#' @param data_frame A data frame to summarise.
#' @param class A vector containing all grouping variables.
#' @param values A vector containing all variables that should be summarised.
#' @param statistics Available functions:
#' - "sum"       -> Weighted and unweighted sum
#' - "sum_wgt"   -> Sum of all weights
#' - "freq"      -> Unweighted frequency
#' - "freq_g0"   -> Unweighted frequency of all values greater than zero
#' - "pct_group" -> Weighted and unweighted percentages within the respective group
#' - "pct_total" -> Weighted and unweighted percentages compared to the grand total
#' - "mean"      -> Weighted and unweighted mean
#' - "median"    -> Weighted and unweighted median
#' - "mode"      -> Weighted and unweighted mode
#' - "min"       -> Minimum
#' - "max"       -> Maximum
#' - "sd"        -> Weighted and unweighted standard deviation
#' - "variance"  -> Weighted and unweighted standard variance
#' - "first"     -> First value
#' - "last"      -> Last value
#' - "pn"        -> Weighted and unweighted percentiles (any p1, p2, p3, ... possible)
#' Uses the SAS default way of calculating percentiles, which is Type 2 described in [quantile()].
#' - "missing"   -> Missings generated by the value variables
#' @param formats A list in which is specified which formats should be applied to which
#' class variables.
#' @param types A character vector specifying the different combinations of group
#' variables which should be computed when using nesting = "all". If left empty all
#' possible combinations will be computed.
#' @param weight Put in a weight variable to compute weighted results.
#' @param nesting The predefined value is "deepest" meaning that only the fully
#' nested version of all class variables will be computed. If set to "all", all
#' possible combinations will be computed in one data table. The option "single"
#' only outputs the ungrouped summary of all class variables in one data table.
#' @param merge_back Newly summarised variables can be merged back to the original
#' data frame if TRUE. Only works if nested = "deepest" and no formats are defined.
#' @param na.rm FALSE by default. If TRUE removes all NA values from the class variables.
#' @param print_miss FALSE by default. If TRUE outputs all possible categories of the
#' grouping variables based on the provided formats, even if there are no observations
#' for a combination.
#' @param monitor FALSE by default. If TRUE, outputs two charts to visualize the
#' functions time consumption.
#' @param notes TRUE by default. Prints notifications about NA values produced by
#' class variables during summarise.
#'
#' @details
#' [summarise_plus()] is based on the 'SAS' procedure Proc Summary, which provides
#' efficient and readable ways to perform complex aggregations.
#'
#' Normally you would compute new categorical variables beforehand - probably even in
#' different forms, if you wanted to have different categorizations - and bloat up
#' the data set. After all this recoding footwork you could finally use multiple
#' summaries to compute all the stats you need to then put them back together. With this
#' function this is no more necessary.
#'
#' In [summarise_plus()] you put in the original data frame and let the recoding happen
#' via format containers. This is very efficient, since new variables and categories
#' are only created just before the summarise happens.
#'
#' Additionally you can specify whether you only want to produce the all nested version
#' of all group variables or whether you want to produce every possible combination in
#' one go. All with a single option.
#'
#' The function is optimized to always take the fastest route, depending on the options
#' specified.
#'
#' @return
#' Returns a summarised `data.table`.
#'
#' If `nesting` is set to `"all"` or `"single"`, the returned table automatically includes three hierarchical metadata variables:
#' - **`TYPE`** (character): The active grouping combination for the row (e.g., `"year+sex"`), or `"total"` for the grand total.
#' - **`TYPE_NR`** (integer): A unique sequential identifier assigned to each distinct grouping combination (`TYPE`).
#' - **`DEPTH`** (integer): The number of active grouping variables in the row's combination (e.g., `0` for `"total"`, `1` for single variables, etc.).
#'
#' @seealso
#' Creating formats: [discrete_format()] and [interval_format()].
#'
#' Functions that also make use of formats: [frequencies()], [crosstabs()],
#' [any_table()], [recode.()], [recode_multi()], [transpose_plus()], [sort_plus()].
#'
#' @examples
#' # NOTE: The used threads are set here to prevent a NOTE from the CRAN checks,
#' #       otherwise this is not necessary.
#' old_threads <- data.table::getDTthreads()
#' data.table::setDTthreads(1)
#'
#' # Example formats
#' age. <- discrete_format(
#'     "Total"          = 0:100,
#'     "under 18"       = 0:17,
#'     "18 to under 25" = 18:24,
#'     "25 to under 55" = 25:54,
#'     "55 to under 65" = 55:64,
#'     "65 and older"   = 65:100)
#'
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' income. <- interval_format(
#'     "Total"              =    0:100000,
#'     "below 500"          =    0:500,
#'     "500 to under 1000"  =  500:1000,
#'     "1000 to under 2000" = 1000:2000,
#'     "2000 and more"      = 2000:100000)
#'
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Summarise the data  with all class variables nested
#' all_nested <- my_data |>
#'     summarise_plus(class      = c(year, sex, age),
#'                    values     = income,
#'                    statistics = c("sum", "pct_group", "pct_total", "sum_wgt", "freq"),
#'                    formats    = list(sex = sex., age = age.),
#'                    weight     = weight,
#'                    nesting    = "deepest",
#'                    na.rm      = TRUE)
#'
#' # Output all possible combinations of the provided class variables. Meaning:
#' # Total, combinations of 2, combinations of 3, ...
#' all_possible <- my_data |>
#'     summarise_plus(class      = c(year, sex, age, income),
#'                    values     = c(probability),
#'                    statistics = c("sum", "p1", "p99", "min", "max", "freq", "freq_g0"),
#'                    formats    = list(sex    = sex.,
#'                                      age    = age.,
#'                                      income = income.),
#'                    weight     = weight,
#'                    nesting    = "all",
#'                    na.rm      = TRUE)
#'
#' # Formats can also be passed as characters
#' single <- my_data |>
#'     summarise_plus(class      = c(year, age, sex),
#'                    values     = weight,
#'                    statistics = c("sum", "mean"),
#'                    formats    = list(sex = "sex.", age = "age."),
#'                    nesting    = "single")
#'
#' # Summarise and merge results back to the original data frame
#' merge_back <- my_data |>
#'     summarise_plus(class      = c(year, age, sex),
#'                    values     = weight,
#'                    statistics = c("sum", "mean"),
#'                    nesting    = "deepest",
#'                    merge_back = TRUE)
#'
#' # Select only certain variable combinations of all the possible combinations
#' certain_types <- my_data |>
#'     summarise_plus(class      = c(year, sex, age),
#'                    values     = c(probability),
#'                    statistics = c("sum", "mean", "freq"),
#'                    formats    = list(sex = sex.,
#'                                      age = age.),
#'                    types      = c("year", "year + age", "age + sex"),
#'                    weight     = weight,
#'                    nesting    = "all",
#'                    na.rm      = TRUE)
#'
#' # You can also select repeating combinations faster like this
#' combi_types <- my_data |>
#'     summarise_plus(class      = c(year, sex, age, education),
#'                    values     = weight,
#'                    statistics = "sum",
#'                    formats    = list(sex = sex.,
#'                                      age = age.),
#'                    types      = "year + sex + (age, age + education, education)",
#'                    nesting    = "all",
#'                    na.rm      = TRUE)
#'
#' # Select specific statistics for specific variables
#' specific_stats <- my_data |>
#'     summarise_plus(class      = c(year, sex),
#'                    statistics = list("sum"       = c(weight, income),
#'                                      "mean"      = expenses,
#'                                      "pct_group" = balance),
#'                    formats    = list(sex = sex.),
#'                    na.rm      = TRUE)
#'
#' # Works without grouping
#' no_group <- my_data |> summarise_plus(values = income)
#'
#' # And also without values
#' no_values  <- my_data |> summarise_plus(class = c(year, sex, age))
#' no_nothing <- my_data |> summarise_plus()
#'
#' # NOTE: The used threads are set here to prevent a NOTE from the CRAN checks,
#' #       otherwise this is not necessary.
#' data.table::setDTthreads(old_threads)
#'
#' @export
summarise_plus <- function(data_frame,
                           class      = NULL,
                           values,
                           statistics = c("sum", "freq"),
                           formats    = list(),
                           types      = "",
                           weight     = NULL,
                           nesting    = "deepest",
                           merge_back = FALSE,
                           na.rm      = .qol_options[["na.rm"]],
                           print_miss = .qol_options[["print_miss"]],
                           monitor    = .qol_options[["monitor"]],
                           notes      = TRUE){
    # Measure the time
    print_start_message()
    print_step("GREY", "Error handling")

    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Error handling", "Preparation")
    #-------------------------------------------------------------------------#

    ###########################################################################
    # Early evaluations
    ###########################################################################

    # First convert data frame to data table
    if (!data.table::is.data.table(data_frame)){
        data_frame <- data.table::as.data.table(data_frame)
    }

    # Evaluate formats early
    if (!is_list_of_dfs(formats)){
        formats_list <- as.list(substitute(formats))[-1]
        formats      <- evaluate_formats(formats_list)
    }

    # Remove empty formats and throw a warning. This can happen if there is e.g.
    # a typo in the format.
    for (variable in names(formats)){
        if (is.null(formats[[variable]])){
            formats[[variable]] <- NULL
            print_message("WARNING", "Format for variable '[variable]' does not exist and can't be applied.", variable = variable)
        }
    }

    # Look up variable names in format data frame to check whether there is an
    # interval or discrete format
    flag_interval <- FALSE

    for (current_var in names(formats)){
        format_df          <- formats[[current_var]]
        interval_variables <- c("from", "to")
        actual_variables   <- names(format_df)[1:2]

        if (identical(interval_variables, actual_variables)){
            flag_interval <- TRUE
            break
        }
    }

    ###########################################################################
    # Error handling
    ###########################################################################

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Nesting
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Correct nesting option if not set right
    if (!nesting %in% c("deepest", "all", "single")){
        print_message("WARNING", c("<Nested> option '[nesting]' doesn't exist. Options 'deepest', 'all' and 'single'",
								   "are available. <Nested> will be set to 'deepest'."), nesting = nesting)
        nesting <- "deepest"
    }

    # Setup of the monitoring results should be displayed depending on the nesting
    monitor_grouping<- "group"

    if (nesting == "deepest"){
        monitor_grouping <- "section"
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Weight
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    weight     <- get_origin_as_char(weight, substitute(weight))
    data_frame <- data_frame |> check_weight(weight)
    weight_var <- ".temp_weight"

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Merge back
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Setup necessary options per code if merge_back so that there are no errors
    # even if the user didn't set them up correctly
    if (merge_back){
        if (nesting != "deepest" || length(formats) > 0){
            print_message("WARNING", c("Merging variables back only works with nesting = 'deepest' and only without formats.",
									   "Options will be set accordingly."))
        }

        nesting <- "deepest"
        formats <- list()
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Statistics
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Translate statistics if possible
    temp_statistics <- tryCatch({
        # Force evaluation to see if it exists
        get_origin_as_char(statistics, substitute(statistics))
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    # In case statistics couldn't be evaluated, this means a list was passed,
    # which provides specific variables for specific statistics. This list needs
    # to be evaluated differently.
    vars_per_stat_list <- NULL

    if (is.null(temp_statistics)){
        vars_per_stat_list <- lapply(as.list(substitute(statistics)), function(element){
            element <- tolower(as.character(element))

            element[!element %in% "c"]
        })[-1]

        statistics <- names(vars_per_stat_list)
    }
    # Prepare statistics vector as normal
    else{
        # Capture the original statistics in case it is a list of characters
        # passed down by another function for reevaluation below.
        original_statistics <- statistics
        statistics          <- tolower(temp_statistics)
    }

    list_of_statistics <- get_complete_statistics_list(statistics)

    # Get the intersection of the requested statistics to make sure
    # only valid actions are passed down
    requested   <- collapse::funique(unlist(list(statistics)))
    valid_stats <- requested[requested %in% names(list_of_statistics)]

    # In case of another function passing a list of characters, this list is evaluated
    # to the passed symbol instead of the actual list. The statistic validation will
    # therefore result in an empty vector. This case is captured here to reevaluate
    # the statistics based on the original passed list of characters.
    if (length(valid_stats) == 0 && is.list(original_statistics)){
        vars_per_stat_list <- original_statistics
        statistics         <- names(vars_per_stat_list)

        requested   <- collapse::funique(unlist(list(statistics)))
        valid_stats <- requested[requested %in% names(list_of_statistics)]
    }

    # If percentages are selected, make sure sum and sum_wgt are also part of statistics
    flag_remove_sum     <- FALSE
    flag_remove_sum_wgt <- FALSE

    if (any(c("pct_group", "pct_total") %in% tolower(requested))){
        if (!"sum" %in% valid_stats){
            valid_stats     <- c(valid_stats, "sum")
            flag_remove_sum <- TRUE
        }

        if (!"sum_wgt" %in% valid_stats){
            valid_stats         <- c(valid_stats, "sum_wgt")
            flag_remove_sum_wgt <- TRUE
        }
    }

    selected_stats <- list_of_statistics[valid_stats]

    # Check if there are any statistics selected which aren't any kind of sums.
    # If only sums are selected then it is faster to summarise first without formats
    # and then apply the formats to a much smaller data frame.
    only_sums <- valid_stats[!valid_stats %in% c("sum", "sum_wgt", "freq", "freq_g0",
                                                 "missing", "pct_group", "pct_total")]

    # Check for invalid statistics
    invalid_stats <- requested[!requested %in% c(names(list_of_statistics), "sum_wgt",
                                                 "pct_group", "pct_total", paste0("p", 1:100))]

    if (length(invalid_stats) > 0){
        print_message("WARNING", "<Statistic> '[invalid]' is invalid and will be omitted.", invalid = invalid_stats)

        if (length(selected_stats) == 0){
            print_message("WARNING", "No valid <statistic> selected. 'sum' will be used.")

            statistics     <- "sum"
            requested      <- "sum"
            valid_stats    <- "sum"
            selected_stats <- list_of_statistics[valid_stats]
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Class
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    class <- get_origin_as_char(class, substitute(class))

    # Make sure that the variables provided are part of the data frame.
    class <- data_frame |> part_of_df(class)

    # If no grouping variables are provided create a pseudo grouping variable
    if (length(class) == 0){
        class <- "pseudo_class"
        data_frame[["pseudo_class"]] <- 1
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Values
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    values <- get_origin_as_char(values, substitute(values))

    # If no value variables are provided generate them automatically
    if (length(values) <= 1){
        if (length(values) == 0 || values == ""){
            # If a statistics list with variable names is given, extract them an set
            # them as value variables.
            if (!is.null(vars_per_stat_list)){
                values       <- collapse::funique(unlist(vars_per_stat_list))
                names(values) <- NULL
            }
            # Otherwise create a pseudo grouping variable
            else{
                values               <- ".temp_values"
                data_frame[[values]] <- 1
            }
        }
    }

    # Make sure there is no class variable that is also a value variable.
    values <- resolve_intersection(values, class)

    # Make sure that the variables provided are part of the data frame.
    values <- data_frame |> part_of_df(values)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Types
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Enable the use of macro variables
    types <- apply_macro(types)

    # Expand brackets first to be able to identify the unique variables used afterwards
    types_expand <- expand_brackets(types)

    if (types_expand[[2]]){
        types <- names(types_expand[[1]])
    }
    else{
        types_expand <- NULL
    }

    # Get row variables from provided combinations
    type_vars <- unlist_variables(types)

    # If there are type provided
    if (is.null(type_vars)){
        print_message("WARNING", "<Types> must be provided in quotation marks. Types will be ignored.")
        types <- NULL
    }
    else if (length(type_vars) > 0){
        # Exclude total because it will most likely not be part of the data frame
        type_vars <- type_vars[type_vars != "total"]
        type_vars <- data_frame |> part_of_df(type_vars, check_only = TRUE)

        if (is.list(type_vars)){
            print_message("WARNING", c("The provided <type> '[type_vars]' [?is/are] not part of",
									   "the data frame. Types will be ignored."), type_vars = type_vars[[1]])
            types <- NULL
        }

        invalid_types <- type_vars[!type_vars %in% class]

        if (length(invalid_types) > 0){
            print_message("WARNING", c("The provided <type> '[invalid]' [?is/are] not part of",
									   "the <class> variables. Types will be ignored."), invalid = invalid_types)
            types <- NULL
        }

        rm(invalid_types)
    }
    else{
        types <- NULL
    }

    # Get the maximum combination depth to be able to break a loop later on earlier
    max_combination_depth <- 9999

    if (!is.null(types)){
        max_combination_depth <- max(nchar(gsub("[^+]", "", types))) + 1
    }

    rm(type_vars)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Double entries
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    class  <- remove_doubled_values(class)
    values <- remove_doubled_values(values)

    ###########################################################################
    # Summarisation starts
    ###########################################################################

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Pre compute", "Preparation")
    #-------------------------------------------------------------------------#

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Some preparations beforehand
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # If types are specified reorder them alphabetically
    if (length(types) > 0){
        reordered_types <- reorder_combination(types)
    }

    # Get the group vars first
    group_vars <- class

    # Keep hold of original data frame to calculate correct missings for messages and
    # for merge back option
    original_df <- data_frame

    result_df <- NULL

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # If only statistics are provided, which rely on summarisation, the function
    # can take a shortcut by pre summarising the data without any formats.
    # Any additional summarise has way less cases to work with and therefore runs
    # way faster.
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Determine whether shortcut is possible
    flag_shortcut <- FALSE

    # Determine when a faster route can be taken to compute the results. Basically
    # this is possible with operations based on sums.
    if (length(only_sums) == 0 && !flag_interval){
        flag_shortcut <- TRUE
    }

    # Pre summarise data frame if only sums as statistics selected
    if (flag_shortcut){
        print_step("MAJOR", "Pre summarise")

        monitor_df <- monitor_df |> monitor_end()

        # Summarise first
        result_list <- data_frame |>
            matrix_summarise(values,
                             group_vars,
                             data_frame[[weight_var]],
                             valid_stats,
                             list_of_statistics,
                             monitor_df,
                             !flag_shortcut)

        data_frame  <- result_list[[1]]
        monitor_df  <- result_list[[2]]
        rm(result_list)

        # Catch new variable names
        values <- data_frame |> inverse(group_vars)
    }

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Apply formats", "Apply formats")
    #-------------------------------------------------------------------------#

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Nesting crossroad
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # If only the combination of every grouping variable should be evaluated
    if (tolower(nesting) == "deepest"){
        # Remove NAs from grouping variables
        if (na.rm){
            data_frame <- collapse::na_omit(data_frame, cols = class)
        }

        print_step("MAJOR", "Executing nested merge")
        get_group_missings(original_df[group_vars], notes, na.rm)

        if (!merge_back){
            rm(original_df)
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # There are different routes for the shortcut and no shortcut summarise
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        if (flag_shortcut){
            result_df <- data_frame |> apply_format(formats, group_vars)

            # Final summarise with formatted data frame
            result_df <- result_df |>
                collapse::fgroup_by(group_vars) |>
                collapse::fsummarise(across(values, collapse::fsum))

            monitor_df <- monitor_df |> monitor_end()

            result_list <- list(result_df, monitor_df)

            fast_pct <- TRUE
        }
        else{
            # Apply formats first
            result_df <- data_frame |> apply_format(formats, group_vars)

            monitor_df <- monitor_df |> monitor_end()

            # If there are multiple evaluation types given, every variable gets an
            # extension to it's variable name
            # If there is only one evaluation type given, the original variable names
            # are kept without extension
            result_list <- result_df |>
                matrix_summarise(values,
                                 group_vars,
                                 result_df[[weight_var]],
                                 valid_stats,
                                 list_of_statistics,
                                 monitor_df,
                                 !flag_shortcut)

            fast_pct <- FALSE
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        #  Compute percentages
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        result_list <- compute_group_percentages(data_frame,
                                                 result_list[[1]],
                                                 statistics,
                                                 group_vars,
                                                 formats,
                                                 values,
                                                 weight_var,
                                                 list_of_statistics,
                                                 result_list[[2]],
                                                 fast_pct)

        result_list <- compute_total_percentages(data_frame,
                                                 result_list[[1]],
                                                 statistics,
                                                 group_vars,
                                                 values,
                                                 weight_var,
                                                 list_of_statistics,
                                                 result_list[[2]],
                                                 fast_pct)

        # Split results and monitor
        result_df  <- result_list[[1]]
        monitor_df <- result_list[[2]]

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_start("Clean up", "Clean up")
        #---------------------------------------------------------------------#

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        #  Clean up
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Generate TYPE variables
        type <- paste(group_vars, collapse = "+")
        result_df[["TYPE"]]    <- type
        result_df[["TYPE_NR"]] <- 1
        result_df[["DEPTH"]]   <- length(group_vars)

        # Reorder new variables in order of requested statistics
        result_df <- result_df |>
            handle_sum_drops(flag_remove_sum, flag_remove_sum_wgt) |>
            reorder_summarised_columns(requested)

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # If no formats are used (meaning only pre defined variables) it is
        # possible to merge the summarized variables back to the original
        # data frame in one go.
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        if (merge_back){
            print_step("MAJOR", "Merging back")
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Merge back", "Merge back")
            #-----------------------------------------------------------------#

            # Don't merge back type variables, only summarised variable
            result_df <- result_df |> drop_type_vars()

            result_df <- original_df |>
                collapse::fungroup() |>
                collapse::join(result_df,
                               on      = group_vars,
                               how     = "left",
                               verbose = FALSE,
                               overid  = 2) |>
                dropp(".temp_weight")
        }
        # If formats are applied and all format categories schould be output
        else if (print_miss && !is.null(formats)){
            result_df <- result_df |> print_missing(formats, group_vars)
        }

    }
    # If every possible combination of the given grouping variables should be evaluated
    else if (tolower(nesting) %in% c("all", "single")){
        all_results <- list()
        index <- 1

        print_step("MAJOR", "Executing merge:")
        print_step("MINOR", "total")

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # The results of all the possible combinations are computed one after
        # another starting with the grand total (ungrouped)
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        if (flag_shortcut){
            # Create a new original data frame to be able to remove NA values
            original_df_shortcut <- data_frame

            # Final summarise
            total_df <- data_frame |>
                collapse::fsummarise(across(values, collapse::fsum))

            sum_columns   <- values[grepl("_sum$", values)]
            new_group_pct <- paste0(gsub("_sum$", "", sum_columns), "_pct_group")
            new_total_pct <- paste0(gsub("_sum$", "", sum_columns), "_pct_total")

            monitor_df <- monitor_df |> monitor_end()

            total_list <- list(total_df, monitor_df)
        }
        else{
            monitor_df <- monitor_df |> monitor_end()

            total_list <- data_frame |>
                collapse::fungroup() |>
                dropp(class) |>
                matrix_summarise(values,
                                 NULL,
                                 data_frame[[weight_var]],
                                 valid_stats,
                                 list_of_statistics,
                                 monitor_df,
                                 !flag_shortcut)

            sum_columns   <- paste0(values, "_sum")
            new_group_pct <- paste0(values, "_pct_group")
            new_total_pct <- paste0(values, "_pct_total")
        }

        # Split results and monitor
        total_df   <- total_list[[1]]
        monitor_df <- total_list[[2]]

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_start("pct_group(total)", "Calc(total)")
        #---------------------------------------------------------------------#

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Total percentages
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        if (any(c("pct_group", "pct_total") %in% statistics)){
            # Copy total sums to have them ready for the group data frames. This way
            # they can be easily joined to compute total percentages faster.
            new_columns <- paste0(sum_columns, "_qol")

            total_df_copy <- data.table::copy(total_df) |>
                keep(sum_columns) |>
                collapse::frename(stats::setNames(new_columns, sum_columns))
        }

        # Compute percentages
        if ("pct_group" %in% statistics){
            # For compute percentages for every variable
            for (i in seq_along(new_group_pct)){
                current_new_var <- new_group_pct[i]

                total_df[[current_new_var]] <- 100
            }
        }

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("pct_total(total)", "Calc(total)")
        #---------------------------------------------------------------------#

        if ("pct_total" %in% statistics){
            # For compute percentages for every variable
            for (i in seq_along(new_total_pct)){
                current_new_var <- new_total_pct[i]

                total_df[[current_new_var]] <- 100
            }
        }

        # Every grouping variable which was not part of the current grouping
        # gets set to a missing value
        total_df[group_vars]  <- NA
        total_df[["TYPE"]]    <- "total"
        total_df[["TYPE_NR"]] <- as.integer(index)
        total_df[["DEPTH"]]   <- as.integer(0)

        # Add data frame to list to add them together at the end
        all_results[["total"]] <- total_df

        monitor_df <- monitor_df |> monitor_end()

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # If grouping variables where defined: compute every possible combination
        # of these variables. Starting with single combinations, then double,
        # triple, and so on.
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        if (length(group_vars) > 0){

            index <- index + 1

            print_step("MAJOR", "Executing combination merge:")

            for (i in seq_along(group_vars)){
                # Break loop early if the maximum depth of combinations is surpassed.
                # This only takes effect if types are selected.
                if (i > max_combination_depth){
                    break
                }

                # Generate every possible combination
                combinations <- utils::combn(group_vars, i, simplify = FALSE)

                # If only combinations of depth 0 and 1 should be generated break out of the loop
                if (tolower(nesting) == "single" && i > 1){
                    break
                }

                #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                # Loop through combinations
                #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

                for (combination in combinations){
                    # If types are specified check first if combination is of the right type
                    if (length(types) > 0){
                        #-----------------------------------------------------#
                        monitor_df <- monitor_df |>
                            monitor_start(paste0("Drop Combo(", paste(combination, collapse = " + "), ")"),
                                          "Drop Combo")
                        #-----------------------------------------------------#

                        # Order current combination alphabetically
                        reordered_combo <- paste(sort(combination), collapse = "+")

                        # Check if current combination is part of the specified types.
                        # If not jump to next combination.
                        if (!reordered_combo %in% reordered_types){
                            next
                        }

                        monitor_df <- monitor_df |> monitor_end()
                    }

                    #---------------------------------------------------------#
                    monitor_df <- monitor_df |>
                        monitor_start(paste0("Formats(", paste(combination, collapse = " + "), ")"),
                                      paste0("Formats(", paste(combination, collapse = " + "), ")"))
                    #---------------------------------------------------------#

                    # Remove NAs from grouping variables. This has to be done individually, because
                    # if otherwise done before in general, too many observations would be removed.
                    # E.g. two of three variables don't have NA values, then the removal of NA values in
                    # the third variable would also effect the first two variables. To get the maximum
                    # out of each combination this is done for every combination separately.
                    if (na.rm){
                        if (flag_shortcut){
                            data_frame <- original_df_shortcut[!collapse::missing_cases(original_df_shortcut[combination]), ]
                        }
                        else{
                            data_frame <- original_df[!collapse::missing_cases(original_df[combination]), ]
                        }
                    }

                    print_step("MINOR", paste(combination, collapse = " + "))

                    get_group_missings(original_df[combination], notes, na.rm)

                    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                    # Shortcut crossroad
                    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

                    if (flag_shortcut){
                        # Final summarise with formatted data frame
                        group_df <- data_frame |>
                            collapse::fgroup_by(combination) |>
                            collapse::fsummarise(across(values, collapse::fsum)) |>
                            apply_format(formats, combination)

                        # Final summarise with formatted data frame
                        group_df <- group_df |>
                            collapse::fgroup_by(combination) |>
                            collapse::fsummarise(across(values, collapse::fsum))

                        monitor_df <- monitor_df |> monitor_end()

                        group_list <- list(group_df, monitor_df)

                        fast_pct <- TRUE
                    }
                    else{
                        # Apply formats first
                        group_df <- data_frame |>
                            collapse::fselect(combination, values, weight_var) |>
                            apply_format(formats, combination)

                        monitor_df <- monitor_df |> monitor_end()

                        # If there are multiple evaluation types given, every variable gets an
                        # extension to it's variable name
                        # If there is only one evaluation type given, the original variable names
                        # are kept without extension
                        group_list <- group_df |>
                            matrix_summarise(values,
                                             combination,
                                             group_df[[weight_var]],
                                             valid_stats,
                                             list_of_statistics,
                                             monitor_df,
                                             !flag_shortcut)

                        fast_pct <- FALSE
                    }

                    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                    # Compute percentages
                    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

                    if (length(types) == 0){
                        # Single variable combinations
                        if (length(combination) == 1){
                            if (!na.rm){
                                # If there is only one grouping variable the super group will be
                                # NULL and therefore can't be grouped. The super total therefore
                                # is simply the grand total.
                                group_list <- compute_total_percentages_short(group_list[[1]],
                                                                              total_df_copy,
                                                                              statistics,
                                                                              sum_columns,
                                                                              combination,
                                                                              "pct_group",
                                                                              group_list[[2]])
                            }
                            else{
                                # If NA values should be removed, the group percentages are computed
                                # as usual so that each variable can be 100 % individually.
                                group_list <- compute_group_percentages(data_frame,
                                                                        group_list[[1]],
                                                                        statistics,
                                                                        combination,
                                                                        formats,
                                                                        values,
                                                                        weight_var,
                                                                        list_of_statistics,
                                                                        group_list[[2]],
                                                                        fast_pct)
                            }
                        }
                        # Combinations of more than one variable
                        else{
                            if (!na.rm){
                                # Evaluate the group percentages based on the super group.
                                # Since the depending super group is already computed
                                # in one of the previous iterations we can make use of
                                # that and simply join the super_df accordingly.
                                super_group <- paste(combination[-length(combination)], collapse = "+")

                                group_list <- compute_group_percentages_short(group_list[[1]],
                                                                              all_results[[super_group]],
                                                                              statistics,
                                                                              combination,
                                                                              sum_columns,
                                                                              group_list[[2]],
                                                                              fast_pct)
                            }
                            else{
                                # If NA values should be removed, the group percentages are computed
                                # as usual so that each variable combination can be 100 % individually.
                                group_list <- compute_group_percentages(data_frame,
                                                                        group_list[[1]],
                                                                        statistics,
                                                                        combination,
                                                                        formats,
                                                                        values,
                                                                        weight_var,
                                                                        list_of_statistics,
                                                                        group_list[[2]],
                                                                        fast_pct)
                            }
                        }
                    }
                    # When types are specified
                    else{
                        group_list <- compute_group_percentages(data_frame,
                                                                group_list[[1]],
                                                                statistics,
                                                                combination,
                                                                formats,
                                                                values,
                                                                weight_var,
                                                                list_of_statistics,
                                                                group_list[[2]],
                                                                fast_pct)
                    }

                    group_list <- compute_total_percentages_short(group_list[[1]],
                                                                  total_df_copy,
                                                                  statistics,
                                                                  sum_columns,
                                                                  combination,
                                                                  "pct_total",
                                                                  group_list[[2]])

                    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                    # Clean up
                    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

                    # Split results and monitor
                    group_df   <- group_list[[1]]
                    monitor_df <- group_list[[2]]

                    #---------------------------------------------------------#
                    monitor_df <- monitor_df |>
                        monitor_start(paste0("Finish(", paste(combination, collapse = " + "), ")"),
                                      paste0("Finish(", paste(combination, collapse = " + "), ")"))
                    #---------------------------------------------------------#

                    # Every grouping variable which was not part of the current grouping
                    # gets set to a missing value
                    missing_vars <- setdiff(group_vars, combination)
                    group_df[missing_vars] <- NA
                    type <- paste(combination, collapse = "+")
                    group_df[["TYPE"]]    <- type
                    group_df[["TYPE_NR"]] <- as.integer(index)
                    group_df[["DEPTH"]]   <- as.integer(i)

                    # If formats are applied and all format categories schould be output
                    if (print_miss && !is.null(formats)){
                        group_df <- group_df |> print_missing(formats, combination)
                    }

                    # Add data frame to list to add them together at the end
                    all_results[[type]] <- group_df

                    index <- index + 1

                    monitor_df <- monitor_df |> monitor_end()
                }
            }
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Additional clean up
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        print_step("MAJOR", "Putting results together")

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_start("Clean up", "Clean up")
        #---------------------------------------------------------------------#

        # Put all computed data frames one below the other and sort the variables
        # in the order: groups -> types -> results
        result_df <- data.table::rbindlist(all_results, fill = TRUE)

        # Reorder new variables in order of requested statistics
        result_df <- result_df |>
            data.table::setcolorder(c(group_vars, "TYPE", "TYPE_NR", "DEPTH")) |>
            handle_sum_drops(flag_remove_sum, flag_remove_sum_wgt) |>
            reorder_summarised_columns(requested)

        # Automatically fuse variables of depth 0 and 1 if only single nested combinations
        # were generated
        if (tolower(nesting) == "single"){
            result_df <- result_df |>
                fuse_variables("fused_vars", group_vars)
        }

        # If types are defined, remove total if not defined in types
        if (length(types) > 0 && !"total" %in% types){
            result_df <- result_df |> collapse::fsubset(TYPE != "total")
        }
    }

    #---------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_start("Clean up", "Clean up")
    #---------------------------------------------------------------------#

    # If only specific variables should be kept per statistic, clean up the additional
    # not needed variables.
    if (!is.null(vars_per_stat_list)){
        vars_to_keep <- c()

        # Transform statistics list into a vector of variable names
        for (name in names(vars_per_stat_list)){
            vars_to_keep <- c(vars_to_keep, paste0(vars_per_stat_list[[name]], "_", name))
        }

        # Get all produced statistic extensions
        stat_extensions <- paste0("_", names(vars_per_stat_list), "$", collapse = "|")

        # Get all statistics variables and exclude the ones to keep to get a
        # vector of variable names to drop.
        statistic_variables <- grep(stat_extensions, names(result_df), value = TRUE)
        vars_to_drop        <- statistic_variables[!statistic_variables %in% vars_to_keep]

        result_df <- suppressMessages(result_df |> dropp(vars_to_drop))
    }

    # Drop pseudo group variable if there is one
    if (any(class == "pseudo_class")){
        result_df <- result_df |> dropp(class)
    }

    #---------------------------------------------------------------------#
    monitor_df |> monitor_plot(by = monitor_grouping, draw_plot = monitor)
    #---------------------------------------------------------------------#

    if (flag_shortcut){
        print_closing(3)
    }
    else{
        print_closing(10)
    }

    result_df
}

###############################################################################
# List of all possible actions
###############################################################################
static_statistics <- list(sum      = collapse::fsum,
                          sum_wgt  = collapse::fsum,
                          freq     = function(x, w, g) collapse::fnobs(x, g = g),
                          freq_g0  = function(x, w, g) freq_g0_qol(x, group = g),
                          mean     = collapse::fmean,
                          median   = collapse::fmedian,
                          mode     = collapse::fmode,
                          min      = function(x, w, g) collapse::fmin(x, g = g),
                          max      = function(x, w, g) collapse::fmax(x, g = g),
                          sd       = collapse::fsd,
                          variance = collapse::fvar,
                          first    = function(x, w, g) collapse::ffirst(x, g = g),
                          last     = function(x, w, g) collapse::flast(x, g = g),
                          missing  = function(x, w, g) collapse::fsum(data.table::as.data.table(is.na(x)), g = g))


#' Compute Percentile Functions
#'
#' @description
#' Computes percentile functions and adds them to the list of statistical functions.
#'
#' @param statistics User provided statistics.
#'
#' @return
#' Returns a complete list of named statistical functions.
#'
#' @noRd
get_complete_statistics_list <- function(statistics){
    all_stats <- static_statistics

    for (stat_name in statistics){
        # Match pattern p<number>
        if (grepl("^p\\d+$", stat_name)){
            prob <- as.numeric(sub("^p", "", stat_name)) / 100

            if (prob > 1){
                print_message("WARNING", "Percentiles are only possible from p0 to p100. [stat_name] will be omitted.",
                              stat_name = stat_name, always_print = TRUE)
                next
            }

            # Define the function
            all_stats[[stat_name]] <- (function(prob){
                force(prob)
                function(x, w = NULL, g = NULL){
                    percentiles_qol(x, w, g, probs = prob)
                }
            })(prob)
        }
    }

    all_stats
}


###############################################################################
# Turns data frame into matrix for fast computation. Evaluates statistics and
# turns everything back into a data frame.
###############################################################################
#' Core Summarise
#'
#' @description
#' This is the core summarisation process. The input data frame is split up in a
#' matrix of values and the grouping variables to make summarisation faster. At
#' the end everything is put back together as a data table.
#'
#' @param data_frame The data frame to be summarised.
#' @param values A vector containing all variables that should be summarised.
#' @param group_vars A vector containing all grouping variables.
#' @param weight Weight variable for weighted results.
#' @param statistics User specified statistics.
#' @param list_of_statistics A list containing all possible statistic functions.
#' @param monitor_df Data frame which stores the monitoring values.
#'
#' @return
#' Returns a list with the summarised data table and the monitoring data frame.
#'
#' @noRd
matrix_summarise <- function(data_frame,
                             values,
                             group_vars,
                             weight,
                             statistics,
                             list_of_statistics,
                             monitor_df,
                             flag_shortcut = TRUE){

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Preparation
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_start("Matrix conversion", paste0("Calc(", paste(group_vars, collapse = " + "), ")"))
    #-------------------------------------------------------------------------#

    # Store original types of grouping variables to later reapply them.
    original_types <- lapply(data_frame[group_vars], class)

    # Create group
    if (is.null(group_vars)){
        # In case there is no grouping (e.g. total percentages) create pseudo group
        group_vars <- "pseudo_group"
        data_frame[[group_vars]] <- 1
        grouping <- collapse::GRP(data_frame, group_vars, sort = flag_shortcut)

    }
    else{
        grouping <- collapse::GRP(data_frame, group_vars, sort = flag_shortcut)
    }

    # Extract group and value columns
    result_df <- data.table::as.data.table(grouping[["groups"]])
    value_df  <- data_frame |> collapse::fselect(values)

    monitor_df <- monitor_df |> monitor_end()

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Compute statistics
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Remove percentages from the statistics vector because they are handled
    # differently than the other single statistics.
    statistics <- grep("^pct", statistics, value = TRUE, invert = TRUE)

    for (single_stat in statistics){
        #-----------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_start(single_stat, paste0("Calc(", paste(group_vars, collapse = " + "), ")"))
        #-----------------------------------------------------------------#

        stat_function <- list_of_statistics[[single_stat]]

        # All functions are handled the same and called from the statistic function
        # list.
        if (single_stat != "sum_wgt"){
            stat_result <- stat_function(value_df, w = weight, g = grouping)

            result_df[, c(paste0(values, "_", single_stat))] <- stat_result
        }
        # With one exception. The sum of weights.
        else{
            stat_result <- data.table::as.data.table(weight) |>
                collapse::fsum(g = grouping) |>
                collapse::frename("weight" = "sum_wgt")

            result_df <- cbind(result_df, stat_result)
        }

        monitor_df <- monitor_df |> monitor_end()
    }

    #-----------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_start("Restore types", "Restore types")
    #-----------------------------------------------------------------#

    # Drop pseudo group, if no grouping variables where specified
    if ("pseudo_group" %in% names(result_df)){
        result_df <- result_df |> collapse::fselect(-pseudo_group)
    }

    # Restore original variable types of numeric variables
    for (variable in names(original_types)){
        var_type <- original_types[[variable]]

        if (length(var_type) > 1){
            next
        }

        if (var_type == "integer"){
            result_df[[variable]] <- as.integer(result_df[[variable]])
        }
        else if (var_type == "double"){
            result_df[[variable]] <- as.double(result_df[[variable]])
        }
        else if (var_type == "numeric"){
            result_df[[variable]] <- as.numeric(result_df[[variable]])
        }
    }

    monitor_df <- monitor_df |> monitor_end()

    list(result_df, monitor_df)
}

###############################################################################
# Handle what happens with automatically generated sum variables
###############################################################################
#' Drop Sum Stats
#'
#' @description
#' Drops sum and sum_wgt variables if not specified by the user.
#'
#' @param result_df The result data frame which contains the sum variables.
#' @param statistics The user provided statistics to check whether sum stats are
#' provided.
#'
#' @return
#' Returns an adjusted result data frame.
#'
#' @noRd
handle_sum_drops <- function(result_df, flag_sum, flag_sum_wgt){
    # Handle sum and sum of weights
    if (flag_sum_wgt){
        result_df <- result_df |> dropp("sum_wgt")
    }

    sum_columns <- grep("_sum$", names(result_df), value = TRUE)

    if (flag_sum){
        result_df <- result_df |> dropp(sum_columns)
    }

    result_df
}

###############################################################################
# After summarise the new variables are ordered like the list_of_statistics.
# This function reorders them in the order which the user has entered on
# summarise_plus function call.
###############################################################################
#' Order Columns by Stats
#'
#' @description
#' Order the value columns of the final result data frame by stats as provided
#' by the user.
#'
#' @param data_frame The result data frame which contains all variables.
#' @param requested_stats The user provided statistics to get the order from.
#'
#' @return
#' Returns a reordered data frame.
#'
#' @noRd
reorder_summarised_columns <- function(data_frame, requested_stats){
    # Select all summarised variables in blocks of the requested statistics
    ordered_cols <- unlist(lapply(requested_stats, function(single_stat){
        # Concatenates the variable ending like "_sum", "_mean", etc. and selects
        # all variables within the given data frame and returns them
        grep(paste0(single_stat, "$"),
             names(data_frame),
             value = TRUE)
    }))

    ordered_cols <- collapse::funique(ordered_cols)

    # Put the ordered columns at the end of the data frame
    data_frame |> data.table::setcolorder(ordered_cols, after = collapse::fncol(data_frame))
}

###############################################################################
# Handle what happens with missing categories, if they appear in a format
###############################################################################
#' Output Missing Values If They Are Present In A Format
#'
#' @description
#' Sometimes it can be handy to not only output all categories which are present
#' in a data frame but also additionally all categories which could have been
#' potentialy present. A format can provide more categories which are NA in the
#' data frame, this functions outputs them as NA rows.
#'
#' @param data_frame The result data frame which contains all formatte variables.
#' @param formats The provided formats.
#' @param group_vars All grouping variables.
#'
#' @return
#' Returns a data frame with added missing categories.
#'
#' @noRd
print_missing <- function(data_frame,
                          formats,
                          group_vars){
    format_list <- formats

    # First check if given variables to format are part of the data frame. This only
    # can be the case if "formats" is provided but the variables are missing in the
    # "class" parameter. In this case abort, because there shouldn't be extra categories,
    # which shouldn't be there in the first place.
    format_names <- names(formats)
    format_vars  <- group_vars[group_vars %in% format_names]

    if (length(format_vars) == 0){
        return(data_frame)
    }

    # Determine which variables are actually formatted in the data frame and which
    # have to be added unformatted.
    no_format_vars <- group_vars[!group_vars %in% format_names]

    # Make sure only variables stay in the format list which are part of the group variables.
    # There is a missmatch, when generating all combinations. Some combinations have lesser
    # grouping variables than formats are provided.
    format_select <- group_vars

    if (length(no_format_vars) > 0){
        format_select <- format_select[!group_vars %in% no_format_vars]
    }

    format_list <- format_list[format_select]

    # Normally I would check if there are missing categories at all and if not abort.
    # I think this has no benefit here and only costs time, since I am only using this
    # function as intended. I leave this comment here, to know what is missing, if a
    # case comes up for which this check would be neccessary.

    # Every unformatted variable is put into a list as single data frame with only
    # its unique values. Only if unformatted variables are present.
    if (length(no_format_vars) > 0){
        no_format_list <- lapply(data_frame |> keep(no_format_vars),
                                 function(variable){
                                     no_format_df        <- data.table::as.data.table(collapse::funique(variable))
                                     names(no_format_df) <- "label"

                                     no_format_df
                                 })

        format_list <- c(format_list, no_format_list)
    }

    # Create the cartesian product of all given formats to generate all possible
    # categories.
    format_df <- suppressMessages(expand_formats(format_list)) |>
        data.table::setcolorder(group_vars) |>
        data.table::setorderv(group_vars)

    # Generate pseudo variable to join
    format_df[[".pseudo_join"]] <- 1

    # Get values from automatically generated variables to add them back for NA
    # values after the join
    type_value    <- data_frame[1, "TYPE"]
    type_nr_value <- data_frame[1, "TYPE_NR"]
    depth_value   <- data_frame[1, "DEPTH"]

    # Perform left join on all format categories, so that each possible category
    # appears in the final data frame. Additionally take over the values from the
    data_frame <- suppressMessages(
        collapse::join(format_df, data_frame,
                       on       = group_vars,
                       how      = "left",
                       multiple = TRUE,
                       verbose  = FALSE,
                       overid   = 2) |>
            dropp(".pseudo_join"))

    # Re apply automatically generated variables to get rid of NA values
    data_frame[["TYPE"]]    <- type_value
    data_frame[["TYPE_NR"]] <- type_nr_value
    data_frame[["DEPTH"]]   <- depth_value

    data_frame
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

test_df <- data.frame(
    group  = c("A", "A", "A", "A", "B", "B", "B", "B", "B"),
    value  = c(0, 1, 2, 2, 3, 4, 4, 100, NA),
    weight = c(7, 8, 6, 5, 3, 7, 9, 2, NA))

dummy_df <- dummy_data(1000)

dummy_df[["binary"]] <- replicate(nrow(dummy_df), {
    paste0(sample(0:1, 2, replace = TRUE), collapse = "")
})

###############################################################################
# Do statistics produce correct output (unweighted)
###############################################################################

# Unweighted sum is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "sum") |>
    remove_stat_extension("sum")

expect_equal(result_df[["value"]][1], collapse::fsum(c(0, 1, 2, 2)),       info = "Unweighted sum is correct")
expect_equal(result_df[["value"]][2], collapse::fsum(c(3, 4, 4, 100, NA)), info = "Unweighted sum is correct")


# Unweighted sum of weights is correct (every observation gets weight of 1)
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "sum_wgt")

expect_equal(result_df[["sum_wgt"]][1], 4, info = "Unweighted sum of weights is correct (every observation gets weight of 1)")
expect_equal(result_df[["sum_wgt"]][2], 5, info = "Unweighted sum of weights is correct (every observation gets weight of 1)")


# Unweighted frequency is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "freq") |>
    remove_stat_extension("freq")

expect_equal(result_df[["value"]][1], 4, info = "Unweighted frequency is correct")
expect_equal(result_df[["value"]][2], 4, info = "Unweighted frequency is correct")


# Unweighted frequency greater zero is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "freq_g0") |>
    remove_stat_extension("freq_g0")

expect_equal(result_df[["value"]][1], 3, info = "Unweighted frequency greater zero is correct")
expect_equal(result_df[["value"]][2], 4, info = "Unweighted frequency greater zero is correct")


# Unweighted missing is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "missing") |>
    remove_stat_extension("missing")

expect_equal(result_df[["value"]][1], 0, info = "Unweighted missing is correct")
expect_equal(result_df[["value"]][2], 1, info = "Unweighted missing is correct")


# Unweighted mean is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "mean") |>
    remove_stat_extension("mean")

expect_equal(result_df[["value"]][1], collapse::fmean(c(0, 1, 2, 2)), info = "Unweighted mean is correct")
expect_equal(result_df[["value"]][2], collapse::fmean(c(3, 4, 4, 100, NA)), info = "Unweighted mean is correct")


# Unweighted median is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "median") |>
    remove_stat_extension("median")

expect_equal(result_df[["value"]][1], collapse::fmedian(c(0, 1, 2, 2)), info = "Unweighted median is correct")
expect_equal(result_df[["value"]][2], collapse::fmedian(c(3, 4, 4, 100, NA)), info = "Unweighted median is correct")


# Unweighted mode is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "mode") |>
    remove_stat_extension("mode")

expect_equal(result_df[["value"]][1], 2, info = "Unweighted mode is correct")
expect_equal(result_df[["value"]][2], 4, info = "Unweighted mode is correct")


# Unweighted min and max are correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = c("min", "max"))

expect_equal(result_df[["value_min"]][1], collapse::fmin(c(0, 1, 2, 2)), info = "Unweighted min and max are correct")
expect_equal(result_df[["value_max"]][1], collapse::fmax(c(0, 1, 2, 2)), info = "Unweighted min and max are correct")
expect_equal(result_df[["value_min"]][2], collapse::fmin(c(3, 4, 4, 100, NA)), info = "Unweighted min and max are correct")
expect_equal(result_df[["value_max"]][2], collapse::fmax(c(3, 4, 4, 100, NA)), info = "Unweighted min and max are correct")


# Unweighted first and last are correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = c("first", "last"))

expect_equal(result_df[["value_first"]][1],  0, info = "Unweighted first and last are correct")
expect_equal(result_df[["value_last"]][1],   2, info = "Unweighted first and last are correct")
expect_equal(result_df[["value_first"]][2],  3, info = "Unweighted first and last are correct")
expect_equal(result_df[["value_last"]][2], 100, info = "Unweighted first and last are correct")


# Unweighted percentiles are correct
result_df <- test_df |>
    collapse::fsubset(!is.na(value)) |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = c("p1", "p50", "median", "p99"))

expect_equal(result_df[["value_p1"]][1],  collapse::fquantile(c(0, 1, 2, 2),   probs = 0.01, names = FALSE), info = "Unweighted percentiles are correct")
expect_equal(result_df[["value_p50"]][1], collapse::fquantile(c(0, 1, 2, 2),   probs = 0.5,  names = FALSE), info = "Unweighted percentiles are correct")
expect_equal(result_df[["value_p99"]][1], collapse::fquantile(c(0, 1, 2, 2),   probs = 0.99, names = FALSE), info = "Unweighted percentiles are correct")
expect_equal(result_df[["value_p1"]][2],  collapse::fquantile(c(3, 4, 4, 100), probs = 0.01, names = FALSE), info = "Unweighted percentiles are correct")
expect_equal(result_df[["value_p50"]][2], collapse::fquantile(c(3, 4, 4, 100), probs = 0.5,  names = FALSE), info = "Unweighted percentiles are correct")
expect_equal(result_df[["value_p99"]][2], collapse::fquantile(c(3, 4, 4, 100), probs = 0.99, names = FALSE), info = "Unweighted percentiles are correct")
expect_equal(result_df[["value_p50"]][1], result_df[["value_median"]][1], info = "Unweighted percentiles are correct")
expect_equal(result_df[["value_p50"]][2], result_df[["value_median"]][2], info = "Unweighted percentiles are correct")


# Unweighted sd and variance are correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = c("sd", "variance"))

expect_equal(result_df[["value_sd"]][1],       collapse::fsd(c(0, 1, 2, 2)), info = "Unweighted sd and variance are correct")
expect_equal(result_df[["value_variance"]][1], collapse::fvar(c(0, 1, 2, 2)), info = "Unweighted sd and variance are correct")
expect_equal(result_df[["value_sd"]][2],       collapse::fsd(c(3, 4, 4, 100, NA)), info = "Unweighted sd and variance are correct")
expect_equal(result_df[["value_variance"]][2], collapse::fvar(c(3, 4, 4, 100, NA)), info = "Unweighted sd and variance are correct")


# Unweighted percentages are correct
result_df <- dummy_df |>
    summarise_plus(class      = c(education, sex),
                   values     = income,
                   statistics = c("pct_group", "pct_total"),
                   na.rm      = TRUE)

sum_pct_group    <- collapse::fsum(result_df[["income_pct_group"]])
expect_pct_group <- length(unique(result_df[["education"]])) * 100

sum_pct_total    <- collapse::fsum(result_df[["income_pct_total"]])

expect_equal(sum_pct_group, expect_pct_group, info = "Unweighted percentages are correct")
expect_equal(sum_pct_total, 100, info = "Unweighted percentages are correct")

###############################################################################
# Do statistics produce correct output (weighted)
###############################################################################

# Weighted sum is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "sum",
                   weight     = weight) |>
        remove_stat_extension("sum")

expect_message(print_stack_as_messages("NOTE"), "Missing values in weight variable 'weight' will be converted to 0", info = "Weighted sum is correct")

expect_equal(result_df[["value"]][1], collapse::fsum(c(0, 1, 2, 2),       w = c(7, 8, 6, 5)), info = "Weighted sum is correct")
expect_equal(result_df[["value"]][2], collapse::fsum(c(3, 4, 4, 100, NA), w = c(3, 7, 9, 2, NA)), info = "Weighted sum is correct")


# sum of weights is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "sum_wgt",
                   weight     = weight)

expect_message(print_stack_as_messages("NOTE"), "Missing values in weight variable 'weight' will be converted to 0", info = "sum of weights is correct")

expect_equal(result_df[["sum_wgt"]][1], collapse::fsum(c(7, 8, 6, 5)), info = "sum of weights is correct")
expect_equal(result_df[["sum_wgt"]][2], collapse::fsum(c(3, 7, 9, 2, NA)), info = "sum of weights is correct")


# Weighted frequency is correct (unweighted count)
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "freq",
                   weight     = weight) |>
        remove_stat_extension("freq")

expect_message(print_stack_as_messages("NOTE"), "Missing values in weight variable 'weight' will be converted to 0", info = "Weighted frequency is correct (unweighted count)")

expect_equal(result_df[["value"]][1], 4, info = "Weighted frequency is correct (unweighted count)")
expect_equal(result_df[["value"]][2], 4, info = "Weighted frequency is correct (unweighted count)")


# Weighted frequency greater zero is correct (unweighted count > 0)
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "freq_g0",
                   weight     = weight) |>
        remove_stat_extension("freq_g0")

expect_message(print_stack_as_messages("NOTE"), "Missing values in weight variable 'weight' will be converted to 0", info = "Weighted frequency greater zero is correct (unweighted count > 0)")

expect_equal(result_df[["value"]][1], 3, info = "Weighted frequency greater zero is correct (unweighted count > 0)")
expect_equal(result_df[["value"]][2], 4, info = "Weighted frequency greater zero is correct (unweighted count > 0)")


# Weighted missing is correct (unweighted missing count)
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "missing",
                   weight     = weight) |>
        remove_stat_extension("missing")

expect_message(print_stack_as_messages("NOTE"), "Missing values in weight variable 'weight' will be converted to 0", info = "Weighted missing is correct (unweighted missing count)")

expect_equal(result_df[["value"]][1], 0, info = "Weighted missing is correct (unweighted missing count)")
expect_equal(result_df[["value"]][2], 1, info = "Weighted missing is correct (unweighted missing count)")


# Weighted mean is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "mean",
                   weight     = weight) |>
        remove_stat_extension("mean")

expect_message(print_stack_as_messages("NOTE"), "Missing values in weight variable 'weight' will be converted to 0", info = "Weighted mean is correct")

expect_equal(result_df[["value"]][1], collapse::fmean(c(0, 1, 2, 2),       w = c(7, 8, 6, 5)), info = "Weighted mean is correct")
expect_equal(result_df[["value"]][2], collapse::fmean(c(3, 4, 4, 100, NA), w = c(3, 7, 9, 2, NA)), info = "Weighted mean is correct")


# Weighted median is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "median",
                   weight     = weight) |>
        remove_stat_extension("median")

expect_message(print_stack_as_messages("NOTE"), "Missing values in weight variable 'weight' will be converted to 0", info = "Weighted median is correct")

expect_equal(result_df[["value"]][1], collapse::fmedian(c(0, 1, 2, 2),       w = c(7, 8, 6, 5)), info = "Weighted median is correct")
expect_equal(result_df[["value"]][2], collapse::fmedian(c(3, 4, 4, 100, NA), w = c(3, 7, 9, 2, NA)), info = "Weighted median is correct")


# Weighted mode is correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = "mode",
                   weight     = weight) |>
        remove_stat_extension("mode")

expect_message(print_stack_as_messages("NOTE"), "Missing values in weight variable 'weight' will be converted to 0", info = "Weighted mode is correct")

expect_equal(result_df[["value"]][1], 2, info = "Weighted mode is correct")
expect_equal(result_df[["value"]][2], 4, info = "Weighted mode is correct")


# Weighted min and max are correct (same as unweighted)
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = c("min", "max"),
                   weight     = weight)

expect_message(print_stack_as_messages("NOTE"), "Missing values in weight variable 'weight' will be converted to 0", info = "Weighted min and max are correct (same as unweighted)")

expect_equal(result_df[["value_min"]][1], collapse::fmin(c(0, 1, 2, 2)), info = "Weighted min and max are correct (same as unweighted)")
expect_equal(result_df[["value_max"]][1], collapse::fmax(c(0, 1, 2, 2)), info = "Weighted min and max are correct (same as unweighted)")
expect_equal(result_df[["value_min"]][2], collapse::fmin(c(3, 4, 4, 100, NA)), info = "Weighted min and max are correct (same as unweighted)")
expect_equal(result_df[["value_max"]][2], collapse::fmax(c(3, 4, 4, 100, NA)), info = "Weighted min and max are correct (same as unweighted)")


# Weighted first and last are correct (same as unweighted)
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = c("first", "last"),
                   weight     = weight)

expect_message(print_stack_as_messages("NOTE"), "Missing values in weight variable 'weight' will be converted to 0", info = "Weighted first and last are correct (same as unweighted)")

expect_equal(result_df[["value_first"]][1], 0, info = "Weighted first and last are correct (same as unweighted)")
expect_equal(result_df[["value_last"]][1], 2, info = "Weighted first and last are correct (same as unweighted)")
expect_equal(result_df[["value_first"]][2], 3, info = "Weighted first and last are correct (same as unweighted)")
expect_equal(result_df[["value_last"]][2], 100, info = "Weighted first and last are correct (same as unweighted)")


# Weighted percentiles are correct
result_df <- test_df |>
    collapse::fsubset(!is.na(value)) |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = c("p50", "median"),
                   weight     = weight)

test_df2 <- test_df |>
    collapse::fsubset(!is.na(value))

result_collapse <- collapse::fmedian(x = test_df2[["value"]],
                                     w = test_df2[["weight"]],
                                     g = collapse::GRP(test_df2, ~group))

expect_true(result_df[["value_p50"]][1] == result_collapse[1], info = "Weighted percentiles are correct")
expect_true(result_df[["value_p50"]][2] == result_collapse[2], info = "Weighted percentiles are correct")
expect_equal(result_df[["value_p50"]][1], result_df[["value_median"]][1], info = "Weighted percentiles are correct")
expect_equal(result_df[["value_p50"]][2], result_df[["value_median"]][2], info = "Weighted percentiles are correct")


# Weighted sd and variance are correct
result_df <- test_df |>
    summarise_plus(class      = group,
                   values     = value,
                   statistics = c("sd", "variance"),
                   weight     = weight)

expect_message(print_stack_as_messages("NOTE"), "Missing values in weight variable 'weight' will be converted to 0", info = "Weighted sd and variance are correct")

expect_equal(result_df[["value_sd"]][1],       collapse::fsd(c(0, 1, 2, 2),        w = c(7, 8, 6, 5)), info = "Weighted sd and variance are correct")
expect_equal(result_df[["value_variance"]][1], collapse::fvar(c(0, 1, 2, 2),       w = c(7, 8, 6, 5)), info = "Weighted sd and variance are correct")
expect_equal(result_df[["value_sd"]][2],       collapse::fsd(c(3, 4, 4, 100, NA),  w = c(3, 7, 9, 2, NA)), info = "Weighted sd and variance are correct")
expect_equal(result_df[["value_variance"]][2], collapse::fvar(c(3, 4, 4, 100, NA), w = c(3, 7, 9, 2, NA)), info = "Weighted sd and variance are correct")


# Weighted percentages are correct
result_df <- dummy_df |>
    summarise_plus(class      = c(education, sex),
                   values     = income,
                   statistics = c("pct_group", "pct_total"),
                   weight     = weight,
                   na.rm      = TRUE)

sum_pct_group    <- collapse::fsum(result_df[["income_pct_group"]])
expect_pct_group <- length(unique(result_df[["education"]])) * 100

sum_pct_total    <- collapse::fsum(result_df[["income_pct_total"]])

expect_equal(sum_pct_group, expect_pct_group, info = "Weighted percentages are correct")
expect_equal(sum_pct_total, 100, info = "Weighted percentages are correct")

###############################################################################
# Other checks
###############################################################################

# Different way of passing variables in (single variables)
result_df1 <- dummy_df |>
    summarise_plus(class  = year,
                   values = income,
                   weight = weight)

result_df2 <- dummy_df |>
    summarise_plus(class  = "year",
                   values = "income",
                   weight = "weight")

expect_identical(result_df1, result_df2, info = "Different way of passing variables in (single variables)")


# Different way of passing variables in (multiple variables)
result_df1 <- dummy_df |>
    summarise_plus(class  = c(year, sex),
                   values = c(income, weight),
                   weight = weight)

result_df2 <- dummy_df |>
    summarise_plus(class  = c("year", "sex"),
                   values = c("income", "weight"),
                   weight = "weight")

result_df3 <- dummy_df |>
    summarise_plus(class  = list(year, sex),
                   values = list(income, weight),
                   weight = weight)

result_df4 <- dummy_df |>
    summarise_plus(class  = list("year", "sex"),
                   values = list("income", "weight"),
                   weight = "weight")

result_df5 <- dummy_df |>
    summarise_plus(class  = year,
                   values = income,
                   weight = weight)

result_df6 <- dummy_df |>
    summarise_plus(class  = "year",
                   values = "income",
                   weight = "weight")

expect_identical(result_df1, result_df2, info = "Different way of passing variables in (multiple variables)")
expect_identical(result_df1, result_df3, info = "Different way of passing variables in (multiple variables)")
expect_identical(result_df1, result_df4, info = "Different way of passing variables in (multiple variables)")


# Simplest form without specifying statistics leads to sum and freq output
result_df <- dummy_df |>
    summarise_plus(class  = year,
                   values = income)

expect_equal(collapse::fncol(result_df), 6, info = "Simplest form without specifying statistics leads to sum and freq output")
expect_true(all(c("year", "income_sum", "income_freq") %in% names(result_df)), info = "Simplest form without specifying statistics leads to sum and freq output")


# Weighted vs. unweighted output
result_df1 <- dummy_df |>
    summarise_plus(class  = year,
                   values = income)

result_df2 <- dummy_df |>
    summarise_plus(class  = year,
                   values = income,
                   weight = weight)

expect_equal(collapse::fncol(result_df1), 6, info = "Weighted vs. unweighted output")
expect_equal(collapse::fncol(result_df2), 6, info = "Weighted vs. unweighted output")

expect_true(all(c("year", "income_sum", "income_freq") %in% names(result_df1)), info = "Weighted vs. unweighted output")
expect_true(all(c("year", "income_sum", "income_freq") %in% names(result_df2)), info = "Weighted vs. unweighted output")

expect_identical(result_df1[["year"]], result_df2[["year"]], info = "Weighted vs. unweighted output")
expect_identical(result_df1[["income_freq"]], result_df2[["income_freq"]], info = "Weighted vs. unweighted output")
expect_false(identical(result_df1[["income_sum"]], result_df2[["income_sum"]]), info = "Weighted vs. unweighted output")


# Specifying many statistics doesn't break function
result_df <- dummy_df |>
    collapse::fsubset(!is.na(income) & !is.na(probability)) |>
    summarise_plus(class      = c(year, sex),
                   values     = c(income, probability),
                   statistics = c("sum", "freq", "freq_g0", "mean", "median", "mode", "min", "max",
                                  "first", "last", "pct_group", "pct_total", "sum_wgt", "p1", "p99",
                                  "sd", "variance", "missing"),
                   weight     = weight)

variable_count <- 2 + (2 * 17) + 1 + 3
expect_equal(collapse::fncol(result_df), variable_count, info = "Specifying many statistics doesn't break function")


# Specifying only one statistic puts out variable names without extension
result_df1 <- dummy_df |>
    summarise_plus(class      = c(year, sex),
                   values     = weight,
                   statistics = "sum")

result_df2 <- dummy_df |>
    summarise_plus(class      = c(year, sex),
                   values     = weight,
                   statistics = c("mean", "sum_wgt"))

expect_equal(names(result_df1), c("year", "sex", "TYPE", "TYPE_NR", "DEPTH", "weight_sum"), info = "Specifying only one statistic puts out variable names without extension")
expect_equal(names(result_df2), c("year", "sex", "TYPE", "TYPE_NR", "DEPTH", "weight_mean", "sum_wgt"), info = "Specifying only one statistic puts out variable names without extension")


# None existent statistics will be omitted
result_df <- dummy_df |>
    summarise_plus(class      = c(year, sex),
                   values     = income,
                   statistics = c("sum", "test"),
                   weight     = weight)

expect_equal(collapse::fncol(result_df), 6, info = "None existent statistics will be omitted")


# Merging variables back to original data frame creates new column
result_df <- dummy_df |>
    summarise_plus(class      = c(year, sex),
                   values     = c(income),
                   statistics = "sum",
                   weight     = weight,
                   merge_back = TRUE)

expect_equal(collapse::fncol(result_df), collapse::fncol(dummy_df) + 1, info = "Merging variables back to original data frame creates new column")
expect_equal(collapse::fnrow(result_df), collapse::fnrow(dummy_df), info = "Merging variables back to original data frame creates new column")


# Don't nest class variables but compute them separately and fuse variables into one super variable
result_df <- dummy_df |>
    summarise_plus(class      = c(year, sex),
                   values     = c(income),
                   statistics = "sum",
                   weight     = weight,
                   nesting    = "single")

expect_equal(collapse::fncol(result_df), 5, info = "Don't nest class variables but compute them separately and fuse variables into one super variable")

expect_true(all(c("TYPE", "TYPE_NR", "DEPTH") %in% names(result_df)), info = "Don't nest class variables but compute them separately and fuse variables into one super variable")

expect_equal(max(result_df[["DEPTH"]]), 1, info = "Don't nest class variables but compute them separately and fuse variables into one super variable")


# Drop auto generated variables after summarise
result_df <- dummy_df |>
    summarise_plus(class      = c(year, sex),
                   values     = c(income),
                   statistics = "sum",
                   weight     = weight,
                   nesting    = "single") |>
    drop_type_vars()

expect_equal(collapse::fncol(result_df), 2, info = "Drop auto generated variables after summarise")

expect_true(!all(c("TYPE", "TYPE_NR", "DEPTH") %in% names(result_df)), info = "Drop auto generated variables after summarise")


# Generate all possible combinations of class variables
result_df <- dummy_df |>
    summarise_plus(class      = c(year, sex, age, education),
                   values     = c(income),
                   statistics = c("sum", "mean", "pct_group", "pct_total"),
                   weight     = weight,
                   nesting    = "all")

expect_equal(collapse::fncol(result_df), 11, info = "Generate all possible combinations of class variables")

expect_true(all(c("TYPE", "TYPE_NR", "DEPTH") %in% names(result_df)), info = "Generate all possible combinations of class variables")

expect_equal(min(result_df[["DEPTH"]]), 0, info = "Generate all possible combinations of class variables")
expect_equal(max(result_df[["DEPTH"]]), 4, info = "Generate all possible combinations of class variables")

expect_true(is.na(result_df[["year"]][1]) & is.na(result_df[["sex"]][1]) & is.na(result_df[["age"]][1]) & is.na(result_df[["education"]][1]), info = "Generate all possible combinations of class variables")


# Generate only chosen combinations of class variables
result_df <- dummy_df |>
    summarise_plus(class      = c(year, sex, age),
                   values     = c(income),
                   statistics = c("sum", "mean", "pct_group", "pct_total"),
                   weight     = weight,
                   types      = c("total", "year + sex", "sex + age", "age"),
                   nesting    = "all")

expect_equal(min(result_df[["DEPTH"]]), 0, info = "Generate only chosen combinations of class variables")
expect_equal(max(result_df[["DEPTH"]]), 2, info = "Generate only chosen combinations of class variables")

expect_equal(length(unique(result_df[["TYPE"]])), 4, info = "Generate only chosen combinations of class variables")
expect_true("year+sex" %in% result_df[["TYPE"]], info = "Generate only chosen combinations of class variables")
expect_true("sex+age" %in% result_df[["TYPE"]], info = "Generate only chosen combinations of class variables")
expect_true("age" %in% result_df[["TYPE"]], info = "Generate only chosen combinations of class variables")
expect_true("total" %in% result_df[["TYPE"]], info = "Generate only chosen combinations of class variables")


# Summarise possible with empty class vector
result_df <- dummy_df |>
    summarise_plus(class  = c(),
                   values = income,
                   statistics = "sum")

expect_equal(collapse::fncol(result_df), 4, info = "Summarise possible with empty class vector")
expect_equal(collapse::fnrow(result_df), 1, info = "Summarise possible with empty class vector")


# Summarise possible with no class variables provided
result_df <- dummy_df |>
    summarise_plus(values = income,
                   statistics = "sum")

expect_equal(collapse::fncol(result_df), 4, info = "Summarise possible with no class variables provided")
expect_equal(collapse::fnrow(result_df), 1, info = "Summarise possible with no class variables provided")


# Summarise uses temporary variable, if no analysis variable is provided
result_df <- dummy_df |> summarise_plus()

expect_equal(as.character(result_df[1, ]), c("pseudo_class", "1", "1", "1000", "1000"),
             info = "Summarise uses temporary variable, if no analysis variable is provided")


# Summarise is able to output specific statistics per variable
result_df <- dummy_df |>
    summarise_plus(class      = sex,
                   statistics = list("sum"       = c(income, expenses),
                                     "pct_group" = weight))

expect_equal(collapse::fncol(result_df), 7, info = "Summarise is able to output specific statistics per variable")
expect_true(all(c("income_sum", "expenses_sum", "weight_pct_group") %in% names(result_df)),
            info = "Summarise is able to output specific statistics per variable")
expect_true(!all(c("income_pct_group", "expenses_pct_group", "weight_sum") %in% names(result_df)),
            info = "Summarise is able to output specific statistics per variable")

###############################################################################
# Format checks
###############################################################################

# Apply single discrete labels
sex. <- discrete_format(
    "Male"   = 1,
    "Female" = 2)
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

format_df <- dummy_df |>
    summarise_plus(class      = c(sex, age),
                   values     = income,
                   statistics = "sum",
                   formats    = list(sex = sex., age = age.),
                   weight     = weight,
                   nesting    = "deepest")

no_format_df <- dummy_df |>
    summarise_plus(class      = c(sex, age),
                   values     = income,
                   statistics = "sum",
                   weight     = weight,
                   nesting    = "deepest")

expect_equal(collapse::fsum(format_df[["income"]]),
             collapse::fsum(no_format_df[["income"]]), info = "Apply single discrete labels")
expect_true(all(c("Male", "Female") %in% format_df[["sex"]]), info = "Apply single discrete labels")
expect_true(all(c("under 18",
                  "18 to under 25",
                  "25 to under 55",
                  "55 to under 65",
                  "65 and older") %in% format_df[["age"]]), info = "Apply single discrete labels")


# Apply discrete multilabels
sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

format_df <- dummy_df |>
    summarise_plus(class      = c(sex, age),
                   values     = weight,
                   statistics = "sum",
                   formats    = list(sex = sex., age = age.),
                   nesting    = "deepest",
                   na.rm      = TRUE) |>
    remove_stat_extension("sum")

no_format_df <- dummy_df |>
    summarise_plus(class      = c(sex, age),
                   values     = weight,
                   statistics = "sum",
                   nesting    = "deepest",
                   na.rm      = TRUE) |>
    remove_stat_extension("sum")

expect_equal(collapse::fsum(format_df[["weight"]]),
             collapse::fsum(no_format_df[["weight"]]) * 4, info = "Apply discrete multilabels")
expect_true(all(c("Total", "Male", "Female") %in% format_df[["sex"]]), info = "Apply discrete multilabels")
expect_true(all(c("Total",
                  "under 18",
                  "18 to under 25",
                  "25 to under 55",
                  "55 to under 65",
                  "65 and older") %in% format_df[["age"]]), info = "Apply discrete multilabels")


# Apply single interval label
income. <- interval_format(
    "below 500"          =    0:500,
    "500 to under 1000"  =  500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

format_df <- dummy_df |>
    summarise_plus(class      = c(income),
                   values     = weight,
                   statistics = "sum",
                   formats    = list(income = income.),
                   nesting    = "deepest") |>
    remove_stat_extension("sum")

expect_equal(collapse::fsum(format_df[["weight"]]),
             collapse::fsum(dummy_df[["weight"]]), info = "Apply single interval label")
expect_true(all(c("below 500",
                  "500 to under 1000",
                  "1000 to under 2000",
                  "2000 and more") %in% format_df[["income"]]), info = "Apply single interval label")


# Apply interval multilabel
income. <- interval_format(
    "Total"              =    0:100000,
    "below 500"          =    0:500,
    "500 to under 1000"  =  500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

format_df <- dummy_df |>
    summarise_plus(class      = c(income),
                   values     = weight,
                   statistics = "sum",
                   formats    = list(income = income.),
                   nesting    = "deepest",
                   na.rm      = TRUE) |>
    remove_stat_extension("sum")

no_format_df <- dummy_df |>
    collapse::fsubset(!is.na(income))

expect_equal(collapse::fsum(format_df[["weight"]]),
             collapse::fsum(no_format_df[["weight"]]) * 2, info = "Apply interval multilabel")
expect_true(all(c("Total",
                  "below 500",
                  "500 to under 1000",
                  "1000 to under 2000",
                  "2000 and more") %in% format_df[["income"]]), info = "Apply interval multilabel")

# Output missing categories in summarise_plus with deepest nesting
sex. <- discrete_format(
    "Male"   = 1,
    "Female" = 2)
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

format_df <- dummy_df |>
    if.(sex == 1 & (age < 25 | age >= 55)) |>
    summarise_plus(class      = c(year, sex, age),
                   values     = weight,
                   formats    = list(sex = sex., age = age.),
                   print_miss = TRUE)

expect_true("Female" %in% format_df[["sex"]], info = "Output missing categories in summarise_plus with deepest nesting")
expect_true("25 to under 55" %in% format_df[["age"]], info = "Output missing categories in summarise_plus with deepest nesting")


# Output missing categories in summarise_plus with all combnations
sex. <- discrete_format(
    "Male"   = 1,
    "Female" = 2)
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

format_df <- dummy_df |>
    if.(sex == 1 & (age < 25 | age >= 55)) |>
    summarise_plus(class      = c(year, sex, age),
                   values     = weight,
                   formats    = list(sex = sex., age = age.),
                   nesting    = "all",
                   print_miss = TRUE)

expect_true("Female" %in% format_df[["sex"]], info = "Output missing categories in summarise_plus with all combnations")
expect_true("25 to under 55" %in% format_df[["age"]], info = "Output missing categories in summarise_plus with all combnations")


# Output missing categories in summarise_plus also works without formats
missing_df <- dummy_df |>
    if.(sex == 1, education = "low") |>
    summarise_plus(class      = c(sex, education),
                   values     = weight,
                   print_miss = TRUE)

expect_true(all(c("middle", "high") %in% missing_df[sex == 1, education]), info = "Output missing categories in summarise_plus also works without formats")


# Use the 'other' format keyword with summarise_plus
age. <- discrete_format(
    "under 18"     = 0:17,
    "18 and older" = "other")

format_df <- dummy_df |>
    summarise_plus(class   = age,
                   values  = weight,
                   formats = list(age = age.))

unique_values <- as.character(format_df[["age"]] |> collapse::funique())

expect_equal(unique_values,
             c("under 18", "18 and older", NA), info = "Use the 'other' format keyword with summarise_plus")


# summarise_plus doesn't convert numeric values stored as character (short route)
result_df <- dummy_df |>
        summarise_plus(class  = binary,
                       values = weight)

expect_equal(result_df[["binary"]], c("00", "01", "10", "11"), info = "summarise_plus doesn't convert numeric values stored as character (short route)")


# summarise_plus doesn't convert numeric values stored as character (long route)
result_df <- dummy_df |>
      summarise_plus(class  = binary,
                     values = weight,
                     statistics = "mean")

expect_equal(result_df[["binary"]], c("00", "01", "10", "11"), info = "summarise_plus doesn't convert numeric values stored as character (long route)")


# summarise_plus converts numeric values back to numeric (short route)
result_df <- dummy_df |>
      summarise_plus(class  = age,
                     values = weight)

expect_equal(result_df[["age"]][1:5], c(0, 1, 2, 3, 4), info = "summarise_plus converts numeric values back to numeric (short route)")


# summarise_plus converts numeric values back to numeric (long route)
result_df <- dummy_df |>
      summarise_plus(class  = age,
                     values = weight,
                     statistics = "mean")

expect_equal(result_df[["age"]][1:5], c(0, 1, 2, 3, 4), info = "summarise_plus converts numeric values back to numeric (long route)")

###############################################################################
# Warning checks
###############################################################################

# Entering none existing variable as weight leads to unweighted results
result_df1 <- dummy_df |> summarise_plus(class  = year,
                                         values = income)

result_df2 <- dummy_df |>
           summarise_plus(class  = year,
                          values = income,
                          weight = abc)

expect_warning(print_stack_as_messages("WARNING"), "Provided weight variable is not part of the data frame",
               info = "Entering none existing variable as weight leads to unweighted results")

expect_true(all(c("year", "income_sum", "income_freq") %in% names(result_df1)),
            info = "Entering none existing variable as weight leads to unweighted results")
expect_true(all(c("year", "income_sum", "income_freq") %in% names(result_df2)),
            info = "Entering none existing variable as weight leads to unweighted results")

expect_identical(result_df1, result_df2, info = "Entering none existing variable as weight leads to unweighted results")


# Entering none numeric variable as weight leads to unweighted results
result_df1 <- dummy_df |> summarise_plus(class  = year,
                                         values = income)

result_df2 <- dummy_df |>
           summarise_plus(class  = year,
                          values = income,
                          weight = education)

expect_warning(print_stack_as_messages("WARNING"), "Provided weight variable is not numeric",
               info = "Entering none numeric variable as weight leads to unweighted results")

expect_true(all(c("year", "income_sum", "income_freq") %in% names(result_df1)),
            info = "Entering none numeric variable as weight leads to unweighted results")
expect_true(all(c("year", "income_sum", "income_freq") %in% names(result_df2)),
            info = "Entering none numeric variable as weight leads to unweighted results")

expect_identical(result_df1, result_df2, info = "Entering none numeric variable as weight leads to unweighted results")


# Percentiles won't be calculated if value variable has NA values
result_df <- dummy_df |>
            summarise_plus(class      = c(year, sex),
                           values     = c(income, probability),
                           statistics = c("p1", "p99"),
                           weight     = weight)

expect_warning(print_stack_as_messages("WARNING"), "To calculate percentiles there may be no NA in the value variables",
               info = "Percentiles won't be calculated if value variable has NA values")
expect_equal(collapse::fncol(result_df), 5, info = "Percentiles won't be calculated if value variable has NA values")


# Percentiles above 100 not allowed
result_df <- dummy_df |>
           summarise_plus(class      = c(year, sex),
                          values     = c(income, probability),
                          statistics = c("p101"),
                          weight     = weight)

expect_warning(print_stack_as_messages("WARNING"), "<Statistic> 'p101' is invalid and will be omitted",
               info = "Percentiles above 100 not allowed")

expect_equal(collapse::fncol(result_df), 7, info = "Percentiles above 100 not allowed")


# None existent class variables will be omitted
result_df <- dummy_df |>
           summarise_plus(class      = c(year, sex, test),
                          values     = c(income, probability),
                          statistics = "sum",
                          weight     = weight)

expect_warning(print_stack_as_messages("WARNING"), "This variable will be omitted during computation",
               info = "None existent class variables will be omitted")

expect_equal(collapse::fncol(result_df), 7, info = "None existent class variables will be omitted")


# None existent analysis variable will be omitted
result_df <- dummy_df |>
           summarise_plus(class      = c(year, sex),
                          values     = c(income, probability, test),
                          statistics = "sum",
                          weight     = weight)

expect_warning(print_stack_as_messages("WARNING"), "This variable will be omitted during computation",
               info = "None existent analysis variable will be omitted")

expect_equal(collapse::fncol(result_df), 7, info = "None existent analysis variable will be omitted")


# Double class variables will be omitted
result_df <- dummy_df |>
           summarise_plus(class      = c(year, sex, sex),
                          values     = c(income, probability),
                          statistics = "sum",
                          weight     = weight)

expect_warning(print_stack_as_messages("WARNING"), "Some <class> variables are provided more than once",
               info = "Double class variables will be omitted")

expect_equal(collapse::fncol(result_df), 7, info = "Double class variables will be omitted")


# Double analysis variables will be omitted
result_df <- dummy_df |>
           summarise_plus(class      = c(year, sex),
                          values     = c(income, probability, income),
                          statistics = "sum",
                          weight     = weight)

expect_warning(print_stack_as_messages("WARNING"), "Some <values> variables are provided more than once",
               info = "Double analysis variables will be omitted")

expect_equal(collapse::fncol(result_df), 7, info = "Double analysis variables will be omitted")


# Analysis variable will be omitted if also passed as class variable
result_df <- dummy_df |>
           summarise_plus(class      = c(year, sex, age),
                          values     = c(age, probability),
                          statistics = "sum",
                          weight     = weight)

expect_warning(print_stack_as_messages("WARNING"), "This variable will be omitted as <values> variable during computation",
               info = "Analysis variable will be omitted if also passed as class variable")

expect_equal(collapse::fncol(result_df), 7, info = "Analysis variable will be omitted if also passed as class variable")


# Merging variables back works if wrong nesting option ist provided
result_df <- dummy_df |>
           summarise_plus(class      = c(year, sex),
                          values     = c(income),
                          statistics = "sum",
                          weight     = weight,
                          nesting    = "all",
                          merge_back = TRUE)

expect_warning(print_stack_as_messages("WARNING"), "Merging variables back only works with nesting",
               info = "Merging variables back works if wrong nesting option ist provided")

expect_equal(collapse::fncol(result_df), collapse::fncol(dummy_df) + 1, info = "Merging variables back works if wrong nesting option ist provided")
expect_equal(collapse::fnrow(result_df), collapse::fnrow(dummy_df), info = "Merging variables back works if wrong nesting option ist provided")


# Drop auto generated variables will be omitted if not in data frame
result_df <- dummy_df |> drop_type_vars()

expect_warning(print_stack_as_messages("WARNING"), "The provided variable to drop", info = "Drop auto generated variables will be omitted if not in data frame")


# Invalid statistic will be omitted
result_df <- dummy_df |>
           summarise_plus(class      = year,
                          values     = income,
                          statistics = c("test", "sum"))

expect_warning(print_stack_as_messages("WARNING"), "<Statistic> 'test' is invalid and will be omitted.", info = "Invalid statistic will be omitted")

expect_equal(collapse::fncol(result_df), 5, info = "Invalid statistic will be omitted")


# Invalid statistic will be omitted 'sum' will be chosen as statistic if no valid one is provided
result_df <- dummy_df |>
           summarise_plus(class      = year,
                          values     = income,
                          statistics = "test")

expect_warning(print_stack_as_messages("WARNING"), "No valid <statistic> selected. 'sum' will be used.",
           info = "'Invalid statistic will be omitted 'sum' will be chosen as statistic if no valid one is provided")

expect_equal(collapse::fncol(result_df), 5, info = "'Invalid statistic will be omitted 'sum' will be chosen as statistic if no valid one is provided")


# summarise_plus throws a warning, if invalid format is used
result_list <- dummy_df |>
    summarise_plus(class   = "sex",
                   formats = list(sex = test))

expect_warning(print_stack_as_messages("WARNING"), "Format for variable 'sex' does not exist and can't be applied.",
               info = "summarise_plus throws a warning, if invalid format is used")


# summarise_plus throws a warning with values stored as character
result_list <- dummy_df |>
    summarise_plus(class  = "sex",
                   values = education)

expect_warning(print_stack_as_messages("WARNING"), "The following <values> are stored as character variables in the data frame",
               info = "summarise_plus throws a warning with values stored as character")

result_list <- dummy_df |>
    summarise_plus(class  = "sex",
                   values = c(education, income_class))

expect_warning(print_stack_as_messages("WARNING"), "The following <values> are stored as character variables in the data frame",
               info = "summarise_plus throws a warning with values stored as character")


# summarise_plus keeps original values which are not covered by a format
sex. <- discrete_format(
    "Male"   = 1,
    "Female" = 2)

sum_df <- dummy_df |>
    summarise_plus(class   = "number_of_persons",
                   formats = list(number_of_persons = sex.))

expect_true(all(c("Male", "Female", as.character(3:5)) %in% sum_df[["number_of_persons"]]),
            info = "summarise_plus keeps original values which are not covered by a format")


set_no_print()

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

dummy_df <- dummy_data(1000)


# Different way of passing variables in
split_list1 <- dummy_df |> split_by(sex)
split_list2 <- dummy_df |> split_by("sex")

expect_identical(split_list1, split_list2, info = "Different way of passing variables in")


# Split data frame by variable
split_list <- dummy_df |> split_by(sex)

expect_inherits(split_list, "list", info = "Split data frame by variable")
expect_equal(length(split_list), 3, info = "Split data frame by variable")

expect_true(!2 %in% split_list[["1"]][["sex"]], info = "Split data frame by variable")
expect_true(!1 %in% split_list[["2"]][["sex"]], info = "Split data frame by variable")
expect_true(!all(c(1, 2) %in% split_list[["NA_sex"]][["sex"]]), info = "Split data frame by variable")


# Split data frame by condition
split_list <- dummy_df |> split_by(sex == 1)

expect_inherits(split_list, "list", info = "Split data frame by condition")
expect_equal(length(split_list), 1, info = "Split data frame by condition")
expect_true(!2 %in% split_list[["sex == 1"]][["sex"]], info = "Split data frame by condition")


# Split data frame by multiple conditions
split_list <- dummy_df |>
       split_by(sex == 1 & income > 2000,
                age < 18)

expect_inherits(split_list, "list", info = "Split data frame by multiple conditions")
expect_equal(length(split_list), 2, info = "Split data frame by multiple conditions")
expect_true(min(split_list[[1]][["income"]]) > 2000, info = "Split data frame by multiple conditions")
expect_true(max(split_list[[2]][["age"]]) < 18, info = "Split data frame by multiple conditions")


# Split with inverse group
split_list <- dummy_df |>
       split_by(sex == 1, inverse = TRUE)

expect_inherits(split_list, "list", info = "Split with inverse group")
expect_equal(length(split_list), 1, info = "Split with inverse group")
expect_true(!1 %in% split_list[["not (sex == 1)"]][["sex"]], info = "Split with inverse group")


# Split data frame by variable and condition
split_list <- dummy_df |>
       split_by(sex,
                age < 18)

expect_inherits(split_list, "list", info = "Split data frame by variable and condition")
expect_equal(length(split_list), 4, info = "Split data frame by variable and condition")

expect_true(!2 %in% split_list[["1"]][["sex"]], info = "Split data frame by variable and condition")
expect_true(!1 %in% split_list[["2"]][["sex"]], info = "Split data frame by variable and condition")
expect_true(!all(c(1, 2) %in% split_list[["NA_sex"]][["sex"]]), info = "Split data frame by variable and condition")
expect_true(max(split_list[["age < 18"]][["age"]]) < 18, info = "Split data frame by variable and condition")


# Split data frame by variable using formats
split_list <- dummy_df |>
       split_by(sex,
                formats = list(sex = discrete_format(
                    "Total"  = 1:2,
                    "Male"   = 1,
                    "Female" = 2)))

expect_inherits(split_list, "list", info = "Split data frame by variable using formats")
expect_equal(length(split_list), 4, info = "Split data frame by variable using formats")

expect_equal(names(split_list), c("Total", "Male", "Female", "NA_sex"), info = "Split data frame by variable using formats")

###############################################################################
# Warning checks
###############################################################################


# Split data frame throws a warning, if variable is not part of the data frame
split_list <- dummy_df |> split_by("test")

expect_warning(print_stack_as_messages("WARNING"), "The provided <conditions> variable 'test' is not part of", info = "Split data frame throws a warning, if variable is not part of the data frame")


###############################################################################
# Abort checks
###############################################################################


# Split data frame aborts, if duplicate list entry names are produced
split_list <- dummy_df |> split_by("state", "age")

expect_error(print_stack_as_messages("ERROR"), "Variable 'age' caused duplicate list entry names.", info = "Split data frame aborts, if duplicate list entry names are produced")


# Split data frame aborts, if no valid split element is passed
split_list <- dummy_df |> split_by(1)

expect_error(print_stack_as_messages("ERROR"), "Only single variables or conditions allowed. Splitting will be aborted.", info = "Split data frame aborts, if no valid split element is passed")


set_no_print()

#' Get Variable Names Which Are Not Part Of The Given Vector
#'
#' @description
#' If you have stored variable names inside a character vector, this function gives you
#' the inverse variable name vector.
#'
#' @param data_frame The data frame from which to take the variable names.
#' @param var_names A character vector of variable names.
#'
#' @return
#' Returns the inverse vector of variable names compared to the given vector.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Get variable names
#' var_names <- c("year", "age", "sex")
#' other_names <- my_data |> inverse(var_names)
#'
#' # Can also be used to just get all variable names
#' all_names <- my_data |> inverse()
#'
#' @export
inverse <- function(data_frame, var_names){
    # Convert to character vectors
    var_names <- get_origin_as_char(var_names, substitute(var_names))

    names(data_frame)[!names(data_frame) %in% var_names]
}


#' Get All Variable Names Between Two Variables
#'
#' @description
#' Get all the variable names inside a data frame between two variables (including the provided ones)
#' as a character vector
#'
#' @param data_frame The data frame from which to take the variable names.
#' @param from Starting variable of variable range.
#' @param to Ending variable of variable range.
#'
#' @return
#' Returns a character vector of variable names.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Get variable names
#' var_names <- my_data |> vars_between(state, income)
#'
#' # Get variable names in reverse order
#' vars_reverse <- my_data |> vars_between(income, state)
#'
#' # If you only provide "from" or "to" you get all variable names from a point to
#' # the end or from the beginning to a given point.
#' vars_from <- my_data |> vars_between(state)
#' vars_to   <- my_data |> vars_between(to = state)
#'
#' # Or just get all variable names
#' vars_all <- my_data |> vars_between()
#'
#' @export
vars_between <- function(data_frame, from, to){
    # Convert to character vectors
    from <- get_origin_as_char(from, substitute(from))
    to   <- get_origin_as_char(to, substitute(to))

    # Check if a vector and not a single variable name was given. If a vector was
    # given, only consider the first element.
    if (length(from) > 1 || length(to) > 1){
        print_message("NOTE", c("Only single variable names allowed for 'from' and 'to'. The respective",
								"first element will be considered."))

        from <- from[1]
        to   <- to[1]
    }

    # Check if "from" and "to" variable names are part of the data frame.
    # If not, adjust accordingly.
    var_names <- names(data_frame)

    if (from != "" && !from %in% var_names){
        print_message("WARNING", c("'from' variable '[from]' is not part of the data frame.",
								   "Selection will start from the first variable."), from = from)

        from = ""
    }
    if (to != "" && !to %in% var_names){
        print_message("WARNING", c("'to' variable '[to]' is not part of the data frame.",
								   "Selection will go until the last variable."), to = to)

        to = ""
    }

    # If "from" is not entered, return all names from the first position to the end
    if (from == ""){
        start <- 1L
    }
    # Else find position in data frame names
    else{
        start <- match(from, var_names)
    }

    # If "to" is not entered, return all names up to the last position from the start
    if (to == ""){
        end <- length(var_names)
    }
    # Else find position in data frame names
    else{
        end <- match(to, var_names)
    }

    # Return
    columns <- start:end
    names(data_frame[columns])
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

dummy_df <- dummy_data(1000)


# Get inverse variable names from data frame works correct
var_names     <- c("state", "age", "sex", "education")
inverse_names <- dummy_df |> inverse(var_names)

var_count <- length(names(dummy_df))

expect_equal(length(var_names) + length((inverse_names)), var_count, info = "Get inverse variable names from data frame works correct")
expect_true(!all(var_names %in% inverse_names), info = "Get inverse variable names from data frame works correct")


# Get variable names between two variables
var_names <- dummy_df |> vars_between(state, person_id)

expect_equal(length(var_names), 5, info = "Get variable names between two variables")


# Get variable names from variable to the end
var_names <- dummy_df |> vars_between(body_weight)

expect_equal(length(var_names), 8, info = "Get variable names from variable to the end")


# Get variable names from beginning to position
var_names <- dummy_df |> vars_between(to = person_id)

expect_equal(length(var_names), 6, info = "Get variable names from beginning to position")

###############################################################################
# Warning checks
###############################################################################

# Providing not existing from-variable throws a warning
var_names <- dummy_df |> vars_between(test, person_id)

expect_warning(print_stack_as_messages("WARNING"), "'from' variable '", info = "Providing not existing from-variable throws a warning")

expect_equal(length(var_names), 6, info = "Providing not existing from-variable throws a warning")


# Providing not existing to-variable throws a warning
var_names <- dummy_df |> vars_between(state, test)

expect_warning(print_stack_as_messages("WARNING"), "'to' variable '", info = "Providing not existing to-variable throws a warning")

expect_equal(length(var_names), 19, info = "Providing not existing to-variable throws a warning")


# Providing more than one from or to variable throws a note
var_names <- dummy_df |> vars_between(c(state, age), income)

expect_message(print_stack_as_messages("NOTE"), "Only single variable names allowed for",
               info = "Providing more than one from or to variable throws a note")

expect_equal(length(var_names), 14, info = "Providing more than one from or to variable throws a note")


set_no_print()

###############################################################################
# Conversion
###############################################################################

#' Converts Numbers into 'Excel' Ranges
#'
#' @description
#' Converts a column number into the according letter to form a cell reference
#' like it is used in 'Excel' (e.g "A1"). Also can compute a range from cell to
#' cell (e.g. "A1:BY22").
#'
#' @param row Single row number.
#' @param column Single column number.
#' @param from_row Range start row.
#' @param from_column Range start column.
#' @param to_row Range end row.
#' @param to_column Range end column.
#'
#' @return
#' Returns a character with an 'Excel' range.
#'
#' @examples
#' single_cell <- get_excel_range(row = 1, column = 6)
#' range       <- get_excel_range(from_row = 1, from_column = 6,
#'                                  to_row = 5,   to_column = 35)
#'
#' @export
get_excel_range <- function(row      = NULL, column      = NULL,
                            from_row = NULL, from_column = NULL,
                            to_row   = NULL, to_column   = NULL){

    # Get single cell
    if (!is.null(row) && !is.null(column)){
        if (row <= 0 || column <= 0){
            print_message("ERROR", "Row and column must be greater than 0.")
            return(NULL)
        }

        return(openxlsx2::wb_dims(rows = row, cols = column))
    }

    # Else get cell range
    if (!is.null(from_row) && !is.null(from_column) &&
        !is.null(to_row) && !is.null(to_column)){
            if (from_column <= 0 || from_row <= 0 || to_row <= 0 || to_column <= 0){
                # No error message here because any_table runs into this regularly if
                # e.g. there are no titles set.
                return(NULL)
            }

        return(openxlsx2::wb_dims(rows = seq.int(from_row, to_row),
                                  cols = seq.int(from_column, to_column)))
    }

    NULL
}

###############################################################################
# Getting table parts for any_table
###############################################################################
#' Get Different Table Parts as Excel Ranges
#'
#' @description
#' Identifies the different table parts (like titles, header, inner table cells)
#' and returns their Ranges in Excel format to be used by openxlsx2.
#'
#' @param table The data frame which holds the table information.
#' @param multi_header The multi layered column header produced by any_table.
#' @param titles Titles if there are any.
#' @param footnotes Footnotes if there are any.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a list with table ranges.
#'
#' @noRd
get_any_table_ranges <- function(table,
                                 multi_header,
                                 stats_row,
                                 titles     = NULL,
                                 footnotes  = NULL,
                                 style      = excel_output_style()){
    # If titles are provided put them in the starting row
    if (!is.null(titles)){
        title.row     <- style[["start_row"]]
        title.column  <- style[["start_column"]]
        title.length  <- length(titles)

        header.row    <- title.row + length(titles) + 1
        header.column <- title.column
    }
    # If no titles are provided the table headers are in the starting row
    else{
        title.row     <- 0
        title.column  <- 0
        title.length  <- 0

        header.row    <- style[["start_row"]]
        header.column <- style[["start_column"]]
    }

    # Get the basic starting cells, as well as table dimensions
    header.width  <- collapse::fncol(multi_header)
    header.length <- collapse::fnrow(multi_header)

    table.row     <- header.row + header.length
    table.length  <- collapse::fnrow(table)
    table.width   <- collapse::fncol(table)
    table.end     <- header.column + table.width - 1
    cat_col.width <- collapse::fncol(table) - collapse::fncol(multi_header)

    footnote.row <- table.row + table.length + 1

    if (!is.null(footnotes)){
        footnote.length <- length(footnotes)
    }
    else{
        footnote.length <- 0
    }

    # Individually compute cell ranges in Excel format
    title_range <- get_excel_range(from_row  = title.row, from_column = title.column,
                                   to_row    = title.row + (title.length - 1),
                                   to_column = title.column)

    whole_tab_range <- get_excel_range(from_row  = header.row, from_column = header.column,
                                       to_row    = table.row     + table.length - 1,
                                       to_column = header.column + (cat_col.width - 1) + header.width)

    header_range <- get_excel_range(from_row  = header.row, from_column = header.column,
                                    to_row    = header.row    + (header.length - 1),
                                    to_column = header.column + (cat_col.width - 1) + header.width)

    box_range <- get_excel_range(from_row  = header.row, from_column = header.column,
                                 to_row    = header.row    + (header.length - 1),
                                 to_column = header.column + (cat_col.width - 1))

    cat_col_range <- get_excel_range(from_row  = table.row, from_column = header.column,
                                     to_row    = table.row     + (table.length - 1),
                                     to_column = header.column + (cat_col.width - 1))

    table_range <- get_excel_range(from_row    = table.row,
                                   from_column = header.column + cat_col.width,
                                   to_row      = table.row     + (table.length - 1),
                                   to_column   = header.column + (cat_col.width - 1) + header.width)

    footnote_range <- get_excel_range(from_row    = footnote.row,
                                      from_column = header.column,
                                      to_row      = footnote.row  + (footnote.length - 1),
                                      to_column   = header.column)

    # Get subheader ranges, if present
    if ("subheaders" %in% names(style)){
        subheader_range     <- list()
        pre_subheader_range <- list()
        subheader_rows      <- style[["subheader_rows"]][style[["subheader_rows"]] != 0]

        # Loop through all subheader rows and extract excel ranges
        for(row in seq_along(subheader_rows)){
            current_row <- subheader_rows[[row]]

            subheader_range[[row]] <- get_excel_range(from_row    = table.row + (current_row - 1),
                                                      from_column = header.column + cat_col.width,
                                                      to_row      = table.row + (current_row - 1),
                                                      to_column   = header.column + (cat_col.width - 1) + header.width)

            pre_subheader_range[[row]] <- get_excel_range(from_row    = table.row + (current_row - 1),
                                                          from_column = header.column,
                                                          to_row      = table.row + (current_row - 1),
                                                          to_column   = header.column + (cat_col.width - 1))
        }
    }

    # Output ranges for further use
    all    <- as.list(environment())
    ranges <- names(formals(sys.function()))

    all[setdiff(names(all), ranges)]
}


#' Get All The Table Ranges of Any Table as Excel Ranges
#'
#' @description
#' Identifies the different table ranges and returns them as ranges in Excel format
#' to be used by 'openxlsx2'.
#'
#' @param any_tab The data frame for any table.
#' @param multi_header The multi layered column header produced by any_table.
#' @param col_header_dimensions A list containing the column variable names and the
#' column header dimensions.
#' @param row_header_dimensions A list containing the row variable names and the
#' row header dimensions.
#' @param titles Titles if there are any.
#' @param footnotes Footnotes if there are any.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a list with table ranges.
#'
#' @noRd
get_any_tab_ranges <- function(any_tab,
                               multi_header,
                               stats_row,
                               titles     = NULL,
                               footnotes  = NULL,
                               style      = excel_output_style()){
    # Get basic table ranges
    table_ranges <- get_any_table_ranges(any_tab, multi_header, stats_row,
                                         titles, footnotes, style)

    # Get specific parts of the table
    any_col_ranges <- list()

    chunks        <- rle(stats_row)
    chunk_lengths <- chunks[["lengths"]]

    col_start <- table_ranges[["header.column"]] + table_ranges[["cat_col.width"]]

    # Get the more detailed blocks
    if (style[["header_stat_merging"]] == "block"){
        start_var <- (collapse::fncol(any_tab) - collapse::fncol(multi_header)) + 1

        # Get the root variable names to count the number of group elements
        root_names <- sub("^([^_]+_[^_]+)_.*$", "\\1", names(any_tab)[start_var:collapse::fncol(any_tab)])

        any_col_ranges[["block_lengths"]] <- rle(root_names)[["lengths"]]

        # Get the statistic symbols or texts to print
        any_col_ranges[["block_values"]] <- c()
        current_column <- 0

        for (block_length in any_col_ranges[["block_lengths"]]){
            current_column <- current_column + block_length

            any_col_ranges[["block_values"]] <- c(any_col_ranges[["block_values"]],
                                                  multi_header[collapse::fnrow(multi_header), current_column])
        }
    }

    # Get the wider blocks per statistic
    for (i in seq_along(chunk_lengths)){
        from_col <- col_start + sum(chunk_lengths[seq_len(i - 1)])
        to_col   <- col_start + sum(chunk_lengths[seq_len(i)])

        range_name <- paste0("any_col_ranges", i)

        any_col_ranges[[range_name]] <-
                get_excel_range(from_row    = table_ranges[["table.row"]],
                                from_column = from_col,
                                to_row      = table_ranges[["table.row"]] +
                                             (table_ranges[["table.length"]] - 1),
                                to_column   = to_col - 1)
    }

    # Output ranges for further use
    c(table_ranges, any_col_ranges)
}


#' Merge Column Header Labels With The Same Name
#'
#' @description
#' If columns have the same name or if there are empty header cells, merge them
#' together to enhance the visual appeal of the table header.
#'
#' @param wb The currently processed workbook.
#' @param column_header The multi layered column header.
#' @param ranges Table ranges.
#'
#' @return
#' Returns a workbook with merged column headers.
#'
#' @noRd
handle_col_header_merge <- function(wb, column_header, ranges, style){
    # Get all values in order of appearance with their respective lengths
    row_values <- lapply(seq_len(collapse::fnrow(column_header)), function(row){
        rle(as.character(column_header[row, ]))
    })

    # Merges the statistic labels block wise. E.g. if multiple percentages are side
    # by side, then the individual percentage blocks get merged, so that one can still
    # tell them apart.
    if (style[["header_stat_merging"]] == "block"){
        row <- length(row_values)
        row_values[[row]] <- list(lengths = ranges[["block_lengths"]],
                                  values  = ranges[["block_values"]])
    }
    # No merging, every statistic is written over every column. This is identical
    # behavior in comparison to SAS.
    else if (style[["header_stat_merging"]] == "none"){
        row <- length(row_values)
        row_values[[row]] <- list(lengths = rep(1, collapse::fncol(column_header)),
                                  values  = as.character(column_header[row, ]))
    }

    # Define offsets in Excel table
    row_offset <- ranges[["header.row"]] - 1
    col_offset <- ranges[["header.column"]] + (ranges[["cat_col.width"]] - 1)

    # Loop through all rows
    number_of_rows <- length(row_values)
    merge_dims     <- character()

    for (row in seq_along(row_values)){
        current_row <- row_values[[row]]
        start_col <- 1

        # Loop through all column chunks
        for (i in seq_along(current_row[["values"]])){
            value <- current_row[["values"]][i]
            space <- current_row[["lengths"]][i]

            # If no empty value and there is an actual need for merging because more
            # than one column has the same value
            if (value != ""){
                from_row <- row
                from_col <- start_col
                to_row   <- row
                to_col   <- start_col + (space - 1)

                # Expand down if upcoming rows are empty in the same part
                if (row < number_of_rows){
                    for (next_row in (row + 1):length(row_values)){
                        slice <- as.character(column_header[next_row, from_col:to_col])

                        # If upcoming row in the same part is empty, expand cell merging
                        # further down
                        if (all(slice == "")){
                            to_row <- next_row
                        }
                        # If row part is not empty exit loop. No further expand of
                        # cell merging.
                        else{
                            break
                        }
                    }
                }

                # Merge cells
                if (from_row == to_row && from_col == to_col){
                    start_col <- start_col + space
                    next
                }

                # Collect all the dims that need to be merged in a vector
                merge_dims <- c(merge_dims, get_excel_range(from_row    = row_offset + from_row,
                                                            from_column = col_offset + from_col,
                                                            to_row      = row_offset + to_row,
                                                            to_column   = col_offset + to_col))
            }

            start_col <- start_col + max(1, space)
        }
    }

    # Merge all dims in one go
    wb$merge_cells(dims = merge_dims)

    wb
}


#' Merge Row Header Labels With The Same Name
#'
#' @description
#' If rows have the same name or if there are empty header cells, merge them
#' together to enhance the visual appeal of the table header.
#'
#' @param wb The currently processed workbook.
#' @param row_header The multi layered row header.
#' @param ranges Table ranges.
#'
#' @return
#' Returns a workbook with merged column headers.
#'
#' @noRd
handle_row_header_merge <- function(wb, row_header, ranges){
    # Make sure row_header is treated as a matrix, even though there can be only one column
    if (is.null(dim(row_header))){
        dim(row_header) <- c(length(row_header), 1)
    }

    # Get all values in order of appearance with their respective lengths (per column)
    col_values <- lapply(seq_len(collapse::fncol(row_header)), function(column){
        rle(as.character(row_header[, column]))
    })

    # Define offsets in Excel table
    row_offset <- ranges[["table.row"]] - 1
    col_offset <- ranges[["header.column"]] - 1

    # Loop through all columns
    number_of_columns <- length(col_values)
    merge_dims <- character()

    for (column in seq_along(col_values)){
        current_col <- col_values[[column]]
        start_row   <- 1

        # Loop through all row chunks
        for (i in seq_along(current_col[["values"]])){
            value <- current_col[["values"]][i]
            space <- current_col[["lengths"]][i]

            # If no empty value and there is an actual need for merging because more
            # than one column has the same value
            if (value != ""){
                from_col <- column
                from_row <- start_row
                to_col   <- column
                to_row   <- start_row + (space - 1)

                # Expand right if upcoming columns are empty in the same block
                if (column < number_of_columns){
                    for (next_column in seq.int(column + 1, number_of_columns)){
                        slice <- as.character(row_header[from_row:to_row, next_column])

                        # If upcoming column in the same part is empty, expand cell merging
                        # further right
                        if (all(slice == "")){
                            to_col <- next_column
                        }
                        # If column part is not empty exit loop. No further expand of
                        # cell merging.
                        else{
                            break
                        }
                    }
                }

                # Merge cells
                if (from_row == to_row && from_col == to_col){
                    start_row <- start_row + space
                    next
                }

                # Collect all the dims that need to be merged in a vector
                merge_dims <- c(merge_dims, get_excel_range(from_row    = row_offset + from_row,
                                                            from_column = col_offset + from_col,
                                                            to_row      = row_offset + to_row,
                                                            to_column   = col_offset + to_col))
            }

            start_row <- start_row + space
        }
    }

    # Merge all dims in one go
    wb$merge_cells(dims = merge_dims)

    wb
}


###############################################################################
# Get ranges for export with style
###############################################################################
#' Get Data Frame Ranges
#'
#' @description
#' Identifies the different table ranges of a data frame and returns them as ranges
#' in Excel format to be used by 'openxlsx2'.
#'
#' @param data_frame The data frame to get the ranges from.
#' @param titles Titles if there are any.
#' @param footnotes Footnotes if there are any.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a list with table ranges.
#'
#' @noRd
get_df_ranges <- function(data_frame,
                          titles     = NULL,
                          footnotes  = NULL,
                          style      = excel_output_style()){
    # Get basic table ranges
    table_ranges <- get_table_ranges(data_frame, titles, footnotes, style)

    # Get specific parts of mean table
    df_col_ranges <- list()
    df_col_types  <- list()

    col_start <- table_ranges[["header.column"]] - 1

    extensions <- c("sum", "pct", "group", "total", "value", "freq",
                    "g0", "mean", "median", "mode", "min", "max", "first",
                    "last", "wgt", "p", "sd", "variance", "missing")

    format_index <- 1

    for (i in seq_len(collapse::fncol(data_frame))){
        # If a variable doesn't have a statistics extension, it is likely not
        # a variable that needs a number format.
        var_end <- sub("p[0-9]+$", "p", utils::tail(strsplit(names(data_frame)[[i]], "_")[[1]], 1))

        if (!var_end %in% extensions){
            next
        }

        # Add number format
        col_to_format <- col_start + i

        range_name <- paste0("df_col_ranges", format_index)
        type_name  <- paste0("df_col_types", format_index)

        df_col_ranges[[range_name]] <-
            get_excel_range(from_row    = table_ranges[["table.row"]],
                            from_column = col_to_format,
                            to_row      = table_ranges[["table.row"]] +
                                (table_ranges[["table.length"]] - 1),
                            to_column   = col_to_format)

        df_col_types[[type_name]] <- var_end

        format_index <- format_index + 1
    }

    table_ranges[["num_format.length"]] <- format_index - 1

    # Output ranges for further use
    c(table_ranges, df_col_ranges, df_col_types)
}


###############################################################################
# Getting table parts
###############################################################################
#' Get Different Table Parts as Excel Ranges
#'
#' @description
#' Identifies the different table parts (like titles, header, inner table cells)
#' and returns their Ranges in Excel format to be used by openxlsx2.
#'
#' @param table The data frame which holds the table information.
#' @param titles Titles if there are any.
#' @param footnotes Footnotes if there are any.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a list with table ranges.
#'
#' @noRd
get_table_ranges <- function(table,
                             titles     = NULL,
                             footnotes  = NULL,
                             style      = excel_output_style()){
    # If titles are provided put them in the starting row
    if (!is.null(titles)){
        title.row     <- style[["start_row"]]
        title.column  <- style[["start_column"]]
        title.length  <- length(titles)

        header.row    <- title.row + length(titles) + 1
        header.column <- title.column
    }
    # If no titles are provided the table headers are in the starting row
    else{
        title.row     <- 0
        title.column  <- 0
        title.length  <- 0

        header.row    <- style[["start_row"]]
        header.column <- style[["start_column"]]
    }

    # Get the basic starting cells, as well as table dimensions
    header.width  <- collapse::fncol(table) - 1

    table.row     <- header.row + 1
    table.length  <- collapse::fnrow(table)
    table.width   <- collapse::fncol(table)
    table.end     <- header.column + table.width - 1
    cat_col.width <- 1

    footnote.row <- table.row + table.length + 1

    if (!is.null(footnotes)){
        footnote.length <- length(footnotes)
    }
    else{
        footnote.length <- 0
    }

    # Individually compute cell ranges in Excel format
    title_range <- get_excel_range(from_row  = title.row, from_column = title.column,
                                   to_row    = title.row + (title.length - 1),
                                   to_column = title.column)

    whole_tab_range <- get_excel_range(from_row  = header.row, from_column = header.column,
                                       to_row    = header.row    + table.length,
                                       to_column = header.column + header.width)

    header_range <- get_excel_range(from_row  = header.row, from_column = header.column,
                                    to_row    = header.row,
                                    to_column = header.column + header.width)

    box_range <- get_excel_range(from_row  = header.row, from_column = header.column,
                                 to_row    = header.row,
                                 to_column = header.column)

    cat_col_range <- get_excel_range(from_row  = table.row, from_column = header.column,
                                     to_row    = header.row    + table.length,
                                     to_column = header.column)

    table_range <- get_excel_range(from_row    = table.row,
                                   from_column = header.column + 1,
                                   to_row      = header.row    + table.length,
                                   to_column   = header.column + header.width)

    footnote_range <- get_excel_range(from_row    = footnote.row,
                                      from_column = header.column,
                                      to_row      = footnote.row  + (footnote.length - 1),
                                      to_column   = header.column)

    # Output ranges for further use
    all    <- as.list(environment())
    ranges <- names(formals(sys.function()))

    all[setdiff(names(all), ranges)]
}


#' Get Columns of Mean Table as Excel Ranges
#'
#' @description
#' Identifies the different table columns and returns their ranges in Excel format
#' to be used by 'openxlsx2'.
#'
#' @param mean_tab The data frame for the mean table produced by frequencies.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a list with table ranges.
#'
#' @noRd
get_mean_tab_ranges <- function(mean_tab,
                                style = excel_output_style()){
    # Get basic table ranges
    table_ranges <- get_table_ranges(mean_tab, style = style)

    # Get specific parts of mean table
    mean_col_ranges <- c()

    for (column in 1:6){
        mean_col_ranges <- c(mean_col_ranges,
                             get_excel_range(from_row    = table_ranges[["header.row"]] + 1,
                                             from_column = table_ranges[["header.column"]] + column,
                                             to_row      = table_ranges[["header.row"]] + table_ranges[["table.length"]],
                                             to_column   = table_ranges[["header.column"]] + column))
    }

    # Output ranges for further use
    c(table_ranges,
      mean_col_ranges = mean_col_ranges)
}


#' Get Columns of Freq Table as Excel Ranges
#'
#' @description
#' Identifies the different table columns and returns their ranges in Excel format
#' to be used by 'openxlsx2'.
#'
#' @param freq_tab The data frame for the freq table produced by frequencies.
#' @param titles Titles if there are any.
#' @param footnotes Footnotes if there are any.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a list with table ranges.
#'
#' @noRd
get_freq_tab_ranges <- function(freq_tab,
                                titles     = NULL,
                                footnotes  = NULL,
                                style      = excel_output_style()){
    # Get basic table ranges
    table_ranges <- get_table_ranges(freq_tab, titles, footnotes, style)

    # Get specific parts of mean table
    freq_col_ranges <- c()

    for (column in 1:6){
        freq_col_ranges <- c(freq_col_ranges,
                             get_excel_range(from_row    = table_ranges[["header.row"]] + 1,
                                             from_column = table_ranges[["header.column"]] + column,
                                             to_row      = table_ranges[["header.row"]] + table_ranges[["table.length"]],
                                             to_column   = table_ranges[["header.column"]] + column))
    }

    # Output ranges for further use
    c(table_ranges,
      freq_col_ranges = freq_col_ranges)
}

###############################################################################
# Format titles and footnotes if there are any
###############################################################################
#' Format Titles and Footnotes
#'
#' @description
#' Format titles and footnotes if there are any.
#'
#' @param wb The workbook to modify.
#' @param ranges Table ranges to format.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a further formatted workbook.
#'
#' @noRd
format_titles_foot_excel <- function(wb, titles, footnotes, ranges, style, output){
    if (output == "excel"){
        pattern <- "(link|cell|file):"

        if (length(titles) > 0){
            # Format
            if (length(style[["title_font_color"]]) == 1){
                for (i in seq_len(length(style[["title_alignment"]]))){
                    # Apply cell styles
                    wb$add_cell_style(dims       = ranges[["title_range"]],
                                      horizontal = style[["title_alignment"]][i],
                                      vertical   = "center",
                                      wrap_text  = "1",
                                      apply_font = TRUE,
                                      font_id    = wb$styles_mgr$get_font_id("title_font"))
                }
            }
            else{
                # Apply cell styles
                full_range <- openxlsx2::dims_to_rowcol(ranges[["title_range"]])

                # Adjust alignment vector to fit the rest
                style[["title_alignment"]] <- fill_or_trim(style[["title_alignment"]],
                                                           length(titles))

                for (i in seq_along(titles)){
                    specific_range <- openxlsx2::wb_dims(as.integer(full_range[["row"]][i]),
                                                         full_range[["col"]])

                    wb$add_cell_style(dims       = specific_range,
                                      horizontal = style[["title_alignment"]][i],
                                      vertical   = "center",
                                      wrap_text  = "1",
                                      apply_font = TRUE,
                                      font_id    = wb$styles_mgr$get_font_id(paste0("title", i, "_font")))
                }
            }

            # Merge the titles over the span of the table
            for (title in seq_along(titles)){
                wb$merge_cells(dims = openxlsx2::wb_dims(
                    cols = style[["start_column"]]:ranges[["table.end"]],
                    rows = style[["start_row"]] + (title - 1)))
            }

            # Set row heights for titles if specified
            title_heights <- style[["title_heights"]]

            if (!is.null(title_heights)){
                number_of_rows <- length(titles)
                start_row      <- ranges[["title.row"]]
                end_row        <- start_row + number_of_rows - 1

                title_heights <- fill_or_trim(title_heights,
                                              number_of_rows)

                wb$set_row_heights(rows    = start_row:end_row,
                                   heights = title_heights)
            }

            # Shortcut: if no links found, just add all titles at once
            if (!any(grepl(pattern, titles))){
                wb$add_data(x         = titles,
                            start_col = style[["start_column"]],
                            start_row = style[["start_row"]])
            }
            # Otherwise print title by title
            else{
                # Adjust alignment vector to fit the rest
                style[["title_alignment"]] <- fill_or_trim(style[["title_alignment"]],
                                                           length(titles))

                for (i in seq_along(titles)){
                    title <- titles[[i]]

                    # If there is a hyperlink in the text
                    if (grepl(pattern, title)){
                        # Identify which of the patterns was found
                        matched_pattern <- regexec(pattern, title)
                        found_pattern   <- regmatches(title, matched_pattern)[[1]][2]

                        # Get the link and the title range
                        current_link <- sub(paste0(".*", pattern), "", title)
                        current_pos  <- get_excel_range(row    = ranges[["title.row"]] + (i - 1),
                                                        column = ranges[["title.column"]])

                        # Remove link from the text
                        titles[[i]] <- sub(paste0(pattern, ".*"), "", title)

                        # Add hyperlink
                        if (found_pattern == "link"){
                            formula <- openxlsx2::create_hyperlink(text = titles[[i]], file = current_link)

                            wb$add_formula(x = formula, dims = current_pos)
                        }
                        # Add link to cell
                        else if (found_pattern == "cell"){
                            formula <- paste0('=HYPERLINK(\"#', current_link, '\", "', titles[[i]], '")')

                            wb$add_formula(x = formula, dims = current_pos)
                        }
                        # Add file link
                        else if (found_pattern == "file"){
                            formula <- paste0('=HYPERLINK("[', current_link, ']", "', titles[[i]], '")')

                            wb$add_formula(x = formula, dims = current_pos)
                        }

                        if (length(style[["title_font_color"]]) == 1){
                            wb$add_cell_style(dims       = current_pos,
                                              horizontal = style[["title_alignment"]][i],
                                              vertical   = "center",
                                              wrap_text  = "1",
                                              apply_font = TRUE,
                                              font_id    = wb$styles_mgr$get_font_id("title_link_font"))
                        }
                        else{
                            wb$add_cell_style(dims       = current_pos,
                                              horizontal = style[["title_alignment"]][i],
                                              vertical   = "center",
                                              wrap_text  = "1",
                                              apply_font = TRUE,
                                              font_id    = wb$styles_mgr$get_font_id(paste0("title", i, "_link_font")))
                        }
                    }
                    # Without links just insert plain text
                    else{
                        current_pos <- get_excel_range(row    = ranges[["title.row"]] + (i - 1),
                                                       column = ranges[["title.column"]])

                        wb$add_data(x = title, dims = current_pos)
                    }
                }
            }
        }

        # Format footnotes if there are any
        if (length(footnotes) > 0){
            # Format
            if (length(style[["footnote_font_color"]]) == 1){
                for (i in seq_len(length(style[["footnote_alignment"]]))){
                    # Apply cell styles
                    wb$add_cell_style(dims       = ranges[["footnote_range"]],
                                      horizontal = style[["footnote_alignment"]][i],
                                      vertical   = "center",
                                      wrap_text  = "1",
                                      apply_font = TRUE,
                                      font_id    = wb$styles_mgr$get_font_id("footnote_font"))
                }

                # Format first footnote row special with a separating line
                full_range    <- openxlsx2::dims_to_rowcol(ranges[["footnote_range"]])
                cat_col_range <- openxlsx2::dims_to_rowcol(ranges[["cat_col_range"]])

                first_foot <- openxlsx2::wb_dims(ranges[["footnote.row"]],
                                                 cat_col_range[["col"]])

                wb$add_cell_style(dims         = first_foot,
                                  horizontal   = style[["footnote_alignment"]][1],
                                  vertical     = "center",
                                  wrap_text    = "1",
                                  apply_font   = TRUE,
                                  font_id      = wb$styles_mgr$get_font_id("footnote_font"),
                                  apply_border = TRUE,
                                  border_id    = wb$styles_mgr$get_border_id("footnote_borders"))
            }
            else{
                # Apply cell styles
                full_range    <- openxlsx2::dims_to_rowcol(ranges[["footnote_range"]])
                cat_col_range <- openxlsx2::dims_to_rowcol(ranges[["cat_col_range"]])

                # Adjust alignment vector to fit the rest
                style[["footnote_alignment"]] <- fill_or_trim(style[["footnote_alignment"]],
                                                           length(footnotes))

                for (i in seq_along(footnotes)){
                    if (i > 1){
                        specific_range <- openxlsx2::wb_dims(as.integer(full_range[["row"]][i]),
                                                             full_range[["col"]])

                        wb$add_cell_style(dims       = specific_range,
                                          horizontal = style[["footnote_alignment"]][i],
                                          vertical   = "center",
                                          wrap_text  = "1",
                                          apply_font = TRUE,
                                          font_id    = wb$styles_mgr$get_font_id(paste0("footnote", i, "_font")))
                    }
                    # Format first footnote row special with a separating line
                    else{
                        specific_range <- openxlsx2::wb_dims(as.integer(full_range[["row"]][i]),
                                                             cat_col_range[["col"]])

                        wb$add_cell_style(dims       = specific_range,
                                          horizontal = style[["footnote_alignment"]][i],
                                          vertical   = "center",
                                          wrap_text  = "1",
                                          apply_font = TRUE,
                                          font_id    = wb$styles_mgr$get_font_id(paste0("footnote", i, "_font")),
                                          apply_border = TRUE,
                                          border_id    = wb$styles_mgr$get_border_id("footnote_borders"))
                    }
                }
            }

            # Merge the footnotes over the span of the table
            for (footnote in seq_along(footnotes)){
                wb$merge_cells(dims = openxlsx2::wb_dims(
                    cols = style[["start_column"]]:ranges[["table.end"]],
                    rows = ranges[["footnote.row"]] + (footnote - 1)))
            }

            # Set row heights for footnotes if specified
            footnote_heights <- style[["footnote_heights"]]

            if (!is.null(footnote_heights)){
                number_of_rows   <- length(footnotes)
                start_row        <- ranges[["footnote.row"]]
                end_row          <- start_row + number_of_rows - 1

                footnote_heights <- fill_or_trim(footnote_heights,
                                                 number_of_rows)

                wb$set_row_heights(rows    = start_row:end_row,
                                   heights = footnote_heights)
            }

            # Shortcut: if no links found, just add all titles at once
            if (!any(grepl(pattern, footnotes))){
                wb$add_data(x         = footnotes,
                            start_col = style[["start_column"]],
                            start_row = ranges[["footnote.row"]])
            }
            # Otherwise print footnote by footnote
            else{
                # Adjust alignment vector to fit the rest
                style[["footnote_alignment"]] <- fill_or_trim(style[["footnote_alignment"]],
                                                              length(footnotes))

                for (i in seq_along(footnotes)){
                    footnote <- footnotes[[i]]

                    # If there is a hyperlink in the text
                    if (grepl(pattern, footnote)){
                        # Identify which of the patterns was found
                        matched_pattern <- regexec(pattern, footnote)
                        found_pattern   <- regmatches(footnote, matched_pattern)[[1]][2]

                        # Get the link and the footnote range
                        current_link <- sub(paste0(".*", pattern), "", footnote)
                        current_pos  <- get_excel_range(row    = ranges[["footnote.row"]] + (i - 1),
                                                        column = ranges[["title.column"]])

                        # Remove link from the text
                        footnote[[i]] <- sub(paste0(pattern, ".*"), "", footnote)

                        # Add hyperlink
                        if (found_pattern == "link"){
                            formula <- openxlsx2::create_hyperlink(text = footnote[[i]], file = current_link)

                            wb$add_formula(x = formula, dims = current_pos)
                        }
                        # Add link to cell
                        else if (found_pattern == "cell"){
                            formula <- paste0('=HYPERLINK(\"#', current_link, '\", "', footnote[[i]], '")')

                            wb$add_formula(x = formula, dims = current_pos)
                        }
                        # Add file link
                        else if (found_pattern == "file"){
                            formula <- paste0('=HYPERLINK("[', current_link, ']", "', footnote[[i]], '")')

                            wb$add_formula(x = formula, dims = current_pos)
                        }

                        if (length(style[["footnote_font_color"]]) == 1){
                            wb$add_cell_style(dims       = current_pos,
                                              horizontal = style[["footnote_alignment"]][i],
                                              vertical   = "center",
                                              wrap_text  = "1",
                                              apply_font = TRUE,
                                              font_id    = wb$styles_mgr$get_font_id("footnote_link_font"))
                        }
                        else{
                            wb$add_cell_style(dims       = current_pos,
                                              horizontal = style[["footnote_alignment"]][i],
                                              vertical   = "center",
                                              wrap_text  = "1",
                                              apply_font = TRUE,
                                              font_id    = wb$styles_mgr$get_font_id(paste0("footnote", i, "_link_font")))
                        }
                    }
                    # Without links just insert plain text
                    else{
                        current_pos <- get_excel_range(row    = ranges[["footnote.row"]] + (i - 1),
                                                       column = ranges[["title.column"]])

                        wb$add_data(x = footnote, dims = current_pos)
                    }
                }
            }
        }
    }
    # If not a styled export just print titles and footnotes as they are
    else{
        if (length(titles) > 0){
            wb$add_data(x         = titles,
                        start_col = style[["start_column"]],
                        start_row = style[["start_row"]])
        }

        if (length(footnotes) > 0){
            wb$add_data(x         = footnotes,
                        start_col = style[["start_column"]],
                        start_row = ranges[["footnote.row"]])
        }
    }

    wb
}

###############################################################################
# Handle style elements
###############################################################################
#' Handle Cell Styles
#'
#' @description
#' Apply cell styles to a range with all style elements in one go.
#'
#' @param wb The workbook to modify.
#' @param ranges Table ranges to format.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a further formatted workbook.
#'
#' @noRd
handle_cell_styles <- function(wb,
                               ranges,
                               style = excel_output_style()){
    # Apply individual styles for each table part
    for (type in c("header", "box", "cat_col", "table", "subheader")){

        apply_font <- NULL
        font_id    <- NULL
        if (paste0(type, "_font") %in% wb$styles_mgr$font$name){
            apply_font <- TRUE
            font_id    <- wb$styles_mgr$get_font_id(paste0(type, "_font"))
        }

        apply_border <- NULL
        border_id    <- NULL
        if (paste0(type, "_borders") %in% wb$styles_mgr$border$name){
            apply_border <- TRUE
            border_id    <- wb$styles_mgr$get_border_id(paste0(type, "_borders"))
        }

        apply_fill <- NULL
        fill_id    <- NULL
        if (paste0(type, "_fill") %in% wb$styles_mgr$fill$name){
            apply_fill <- TRUE
            fill_id    <- wb$styles_mgr$get_fill_id(paste0(type, "_fill"))
        }

        # Subheader ranges are stored inside a list which has to be looped over
        if (type == "subheader"){
            for (sub_range in seq_along(ranges[["subheader_range"]])){
                wb$add_cell_style(dims         = ranges[["subheader_range"]][[sub_range]],
                                  horizontal   = style[[paste0(type, "_alignment")]],
                                  vertical     = "center",
                                  wrap_text    = style[[paste0(type, "_wrap")]],
                                  indent       = style[[paste0(type, "_indent")]],
                                  apply_font   = apply_font,
                                  font_id      = font_id,
                                  apply_border = apply_border,
                                  border_id    = border_id,
                                  apply_fill   = apply_fill,
                                  fill_id      = fill_id)
            }
        }
        # All other ranges are stored as single entries
        else{
            wb$add_cell_style(dims         = ranges[[paste0(type, "_range")]],
                              horizontal   = style[[paste0(type, "_alignment")]],
                              vertical     = "center",
                              wrap_text    = style[[paste0(type, "_wrap")]],
                              indent       = style[[paste0(type, "_indent")]],
                              apply_font   = apply_font,
                              font_id      = font_id,
                              apply_border = apply_border,
                              border_id    = border_id,
                              apply_fill   = apply_fill,
                              fill_id      = fill_id)
        }
    }

    wb
}


#' Set Up Different Fill Styles
#'
#' @description
#' Pass fill styles for the different parts of the table into the style manager
#' to apply them more efficiently later.
#'
#' @param wb The workbook to modify.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a workbook with added style elements.
#'
#' @noRd
handle_fill_styles <- function(wb, style = excel_output_style()){
    # Add individual fill styles for each table part
    for (type in c("header", "subheader", "box", "cat_col", "table")){
        # With color fill
        if (style[[paste0(type, "_back_color")]] != ""){
            wb$styles_mgr$add(
                openxlsx2::create_fill(pattern_type = "solid",
                                       fg_color     = openxlsx2::wb_color(style[[paste0(type, "_back_color")]])),
                paste0(type, "_fill"))
        }
        # With transparent background
        else{
            wb$styles_mgr$add(
                openxlsx2::create_fill(pattern_type = "",
                                       fg_color     = NULL),
                paste0(type, "_fill"))
        }
    }

    wb
}


#' Set Up Different Font Styles
#'
#' @description
#' Pass font styles for the different parts of the table into the style manager
#' to apply them more efficiently later.
#'
#' @param wb The workbook to modify.
#' @param texts A list of the actual titles and footnotes.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a workbook with added style elements.
#'
#' @noRd
handle_font_styles <- function(wb,
                               texts = list("title" = c(), "footnote" = c()),
                               style = excel_output_style()){
    # Font can only be set globally here. Below in create_font seems to be bugged.
    wb$set_base_font(font_name = style[["font"]])

    # Add individual font styles for each table part
    for (type in c("title", "footnote", "header", "subheader", "box", "cat_col", "table")){
        if (type %in% c("title", "footnote")){
            multi_style <- max(length(style[[paste0(type, "_font_color")]]),
                               length(style[[paste0(type, "_font_size")]]),
                               length(style[[paste0(type, "_font_bold")]]))

            # For titles and footnotes multiple font colors per line can be specified
            if (multi_style > 1){
                # If there are multiple titles or footnotes, bring the formatting on
                # the same amount as the actual titles and footnotes.
                if (length(texts[[type]]) > 0){
                    multi_style <- length(texts[[type]])
                }

                style[[paste0(type, "_font_color")]] <- fill_or_trim(style[[paste0(type, "_font_color")]], multi_style)
                style[[paste0(type, "_font_size")]]  <- fill_or_trim(style[[paste0(type, "_font_size")]],  multi_style)
                style[[paste0(type, "_font_bold")]]  <- fill_or_trim(style[[paste0(type, "_font_bold")]],  multi_style)

                # Look up title font colors
                for (i in seq_len(length(style[[paste0(type, "_font_color")]]))){
                    # Normal style
                    wb$styles_mgr$add(
                        openxlsx2::create_font(sz    = style[[paste0(type, "_font_size")]][i],
                                               color = openxlsx2::wb_color(hex = style[[paste0(type, "_font_color")]][i]),
                                               b     = style[[paste0(type, "_font_bold")]][i]),
                        paste0(type, i, "_font"))

                    # Hyperlink style
                    wb$styles_mgr$add(
                        openxlsx2::create_font(sz    = style[[paste0(type, "_font_size")]][i],
                                               color = openxlsx2::wb_color(hex = "0000FF"),
                                               u     = "single"),
                        paste0(type, i, "_link_font"))
                }
            }
            else{
                # Normal style
                wb$styles_mgr$add(
                    openxlsx2::create_font(sz    = style[[paste0(type, "_font_size")]],
                                           color = openxlsx2::wb_color(hex = style[[paste0(type, "_font_color")]]),
                                           b     = style[[paste0(type, "_font_bold")]]),
                    paste0(type, "_font"))

                # Hyperlink style
                wb$styles_mgr$add(
                    openxlsx2::create_font(sz    = style[[paste0(type, "_font_size")]],
                                           color = openxlsx2::wb_color(hex = "0000FF"),
                                           u     = "single"),
                    paste0(type, "_link_font"))
            }
        }
        else{
            wb$styles_mgr$add(
                openxlsx2::create_font(sz    = style[[paste0(type, "_font_size")]],
                                       color = openxlsx2::wb_color(hex = style[[paste0(type, "_font_color")]]),
                                       b     = style[[paste0(type, "_font_bold")]]),
                paste0(type, "_font"))
        }
    }

    wb
}


#' Set Up Different Border Styles
#'
#' @description
#' Pass border styles for the different parts of the table into the style manager
#' to apply them more efficiently later.
#'
#' @param wb The workbook to modify.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a workbook with added style elements.
#'
#' @noRd
handle_border_styles <- function(wb, style = excel_output_style()){
    if (style[["header_borders"]]){
        wb$styles_mgr$add(
            openxlsx2::create_border(
                      bottom = "thin",
                      top    = "thin",
                      left   = "thin",
                      bottom_color = openxlsx2::wb_color(hex = style[["header_border_color"]]),
                      top_color    = openxlsx2::wb_color(hex = style[["header_border_color"]]),
                      left_color   = openxlsx2::wb_color(hex = style[["header_border_color"]])),
            "header_borders")
    }
    if (style[["subheader_borders"]]){
        wb$styles_mgr$add(
            openxlsx2::create_border(
                bottom = "thin",
                top    = "thin",
                bottom_color = openxlsx2::wb_color(hex = style[["subheader_border_color"]]),
                top_color    = openxlsx2::wb_color(hex = style[["subheader_border_color"]])),
            "subheader_borders")
    }
    if (style[["box_borders"]]){
        wb$styles_mgr$add(
            openxlsx2::create_border(
                      bottom = "thin",
                      top    = "thin",
                      right  = "thin",
                      bottom_color = openxlsx2::wb_color(hex = style[["box_border_color"]]),
                      top_color    = openxlsx2::wb_color(hex = style[["box_border_color"]]),
                      right_color  = openxlsx2::wb_color(hex = style[["box_border_color"]])),
            "box_borders")
    }
    if (style[["cat_col_borders"]]){
        wb$styles_mgr$add(
            openxlsx2::create_border(
                      top   = "thin",
                      right = "thin",
                      top_color   = openxlsx2::wb_color(hex = style[["cat_col_border_color"]]),
                      right_color = openxlsx2::wb_color(hex = style[["cat_col_border_color"]])),
            "cat_col_borders")
    }
    if (style[["table_borders"]]){
        wb$styles_mgr$add(
            openxlsx2::create_border(
                      top  = "thin",
                      left = "thin",
                      top_color  = openxlsx2::wb_color(hex = style[["table_border_color"]]),
                      left_color = openxlsx2::wb_color(hex = style[["table_border_color"]])),
            "table_borders")
    }

    wb$styles_mgr$add(
        openxlsx2::create_border(
            top  = "thin",
            top_color  = openxlsx2::wb_color(hex = style[["table_border_color"]])),
        "footnote_borders")

    wb
}

#' Set Up Different Number Format Styles
#'
#' @description
#' Pass number format styles for the different statistics into the style manager
#' to apply them more efficiently later.
#'
#' @param wb The workbook to modify.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a workbook with added style elements.
#'
#' @noRd
handle_number_styles <- function(wb, style = excel_output_style()){
    # Add individual font styles for each table part
    index <- 900

    for (type in c("pct", "freq", "freq.g0", "sum", "sum.wgt", "mean", "median",
                   "mode", "min", "max", "sd", "variance", "first", "last", "p",
                   "missing")){
        wb$styles_mgr$add(
            openxlsx2::create_numfmt(numFmtId   = index,
                                     formatCode = style[["number_formats"]][[paste0(type, "_excel")]]),
            paste0(type, "_numfmt"))

        index <- index + 1
    }

    wb
}


#' Set Up Styles
#'
#' @description
#' Set up all styles in the style manager to apply them more efficiently later.
#'
#' @param wb The workbook to modify.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a workbook with added style elements.
#'
#' @noRd
prepare_styles <- function(wb,
                           texts = list("title" = c(), "footnote" = c()),
                           style = excel_output_style()){
    wb |>
        handle_fill_styles(style) |>
        handle_font_styles(texts, style) |>
        handle_border_styles(style) |>
        handle_number_styles(style)
}

###############################################################################
# Handle cell dimensions
###############################################################################
#' Handle Column Widths and Row Heights
#'
#' @description
#' Handle column width and row heights depending on specified options.
#'
#' @param wb The workbook to modify.
#' @param ranges Table ranges to format.
#' @param number_of_columns The number of columns that should be formatted.
#' @param number_of_rows The number of rows that should be formatted.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a further formatted workbook.
#'
#' @noRd
handle_col_row_dimensions <- function(wb,
                                      ranges,
                                      number_of_columns,
                                      number_of_rows,
                                      style = excel_output_style()){
    column_widths <- style[["column_widths"]]
    start_column  <- 1
    end_column    <- style[["start_column"]] + (number_of_columns - 1)

    # If specific column widths are specified
    if (length(column_widths) > 1){
        column_widths <- fill_or_trim(column_widths,
                                      end_column)
    }
    # If only one column width is specified
    else if (column_widths != "auto"){
        column_widths <- fill_or_trim(column_widths,
                                      end_column)
    }
    # On auto format only format row header columns. On large workbooks this
    # is very slow so only make sure the text in front are readable.
    else{
        start_column      <- ranges[["header.column"]]
        number_of_columns <- ranges[["header.column"]] + (ranges[["cat_col.width"]] - 1)
    }

    row_heights <- style[["row_heights"]]
    start_row   <- 1
    end_row     <- style[["start_row"]] + (number_of_rows - 1)

    # If specific column widths are specified
    if (length(row_heights) > 1){
        row_heights <- fill_or_trim(row_heights,
                                    end_row)
    }
    # If only one row height is specified
    else if (row_heights != "auto"){
        row_heights <- fill_or_trim(row_heights,
                                    end_row)
    }
    else{
        row_heights <- NULL
    }

    wb$set_col_widths(cols   = start_column:end_column,
                      widths = column_widths)
    wb$set_row_heights(rows    = start_row:end_row,
                       heights = row_heights)

    wb
}


#' Handle Automatic Dimensions
#'
#' @description
#' Handle column width and row heights if they should be automatically adjusted.
#'
#' @param wb The workbook to modify.
#' @param ranges Table ranges to format.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a further formatted workbook.
#'
#' @noRd
handle_auto_dimensions <- function(wb,
                                   ranges,
                                   style = excel_output_style()){
    # If autofit columns is selected, set a manual size after the first table
    # column which is sufficient in most cases. The integrated autofit often
    # gives to narrow columns.
    column_width <- style[["column_widths"]]

    if (length(column_width) == 1){
        if (column_width == "auto"){
            start_col <- style[["start_column"]] + 1

            wb$set_col_widths(cols   = start_col:ranges[["table.end"]],
                              widths = 10)
        }
    }

    # If autofit heights is selected, set a manual size for the table header,
    # since this is the only row to safely say that it has multi row text.
    # There is no good solution to handle the other row. One could only make
    # a vague guess because the actual space each individual font takes at different
    # sizes is unclear. For now the user has to handle multi row texts manually.
    row_heights <- style[["row_heights"]]

    if (length(row_heights) == 1){
        if (row_heights == "auto"){
            wb$set_row_heights(rows    = ranges[["header.row"]],
                               heights = 30)
        }
    }

    wb
}


#' Handle Automatic Dimensions For Any_Table
#'
#' @description
#' Handle column width and row heights if they should be automatically adjusted in
#' an any_table.
#'
#' @param wb The workbook to modify.
#' @param ranges Table ranges to format.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a further formatted workbook.
#'
#' @noRd
handle_any_auto_dimensions <- function(wb,
                                       ranges,
                                       style = excel_output_style()){
    # If autofit columns is selected, set a manual size after the table row header
    # columns which is sufficient in most cases. The integrated autofit often
    # gives to narrow columns.
    column_width <- style[["column_widths"]]

    if (length(column_width) == 1){
        if (column_width == "auto"){
            start_col <- style[["start_column"]] + ranges[["cat_col.width"]]

            wb$set_col_widths(cols   = start_col:ranges[["table.end"]],
                              widths = 9)
        }
    }

    wb
}


#' Handle Header and Table Heights
#'
#' @description
#' Handle header and table row heights if they should be adjusted individually.
#'
#' @param wb The workbook to modify.
#' @param ranges Table ranges to format.
#' @param style A list of style elements to format the table.
#'
#' @return
#' Returns a further formatted workbook.
#'
#' @noRd
handle_header_table_dim <- function(wb,
                                    ranges,
                                    style = excel_output_style()){
    # Set row heights for header if specified
    header_heights <- style[["header_heights"]]

    if (!is.null(header_heights)){
        number_of_rows <- max(ranges[["header.length"]], 1)
        start_row      <- ranges[["header.row"]]
        end_row        <- start_row + max((number_of_rows - 1), 0)

        header_heights <- fill_or_trim(header_heights,
                                       number_of_rows)

        wb$set_row_heights(rows    = start_row:end_row,
                           heights = header_heights)
    }

    # Set row heights for table if specified
    table_heights <- style[["table_heights"]]

    if (!is.null(table_heights)){
        number_of_rows <- ranges[["table.length"]]
        start_row      <- ranges[["table.row"]]
        end_row        <- start_row + (number_of_rows - 1)

        table_heights <- fill_or_trim(table_heights,
                                      number_of_rows)

        wb$set_row_heights(rows    = start_row:end_row,
                           heights = table_heights)
    }

    wb
}


#' Adjust Format Vector to Needed Length
#'
#' @description
#' Either fill up or trim down the provided format vector to the needed number
#' of columns/rows which have to be formatted.
#'
#' @param format_vector Format vector to be adjusted.
#' @param number_to_format Needed length to be adjusted.
#'
#' @return
#' Returns an adjusted format vector.
#'
#' @noRd
fill_or_trim <- function(format_vector,
                         number_to_format){
    # If length of provided vector is lesser than there are places to format
    if (length(format_vector) < number_to_format){
        # Repeat the last vector element until the number to format is reached
        format_vector <- c(format_vector,
                           rep(utils::tail(format_vector, 1),
                               number_to_format - length(format_vector)))
    }
    # If length of provided vector is equal or greater than there are places to format
    else{
        # Cut elements down to number of places to format
        format_vector <- utils::head(format_vector, number_to_format)
    }

    format_vector
}


#' Split Up 'Excel' Ranges
#'
#' @description
#' Split up a range into individual column ranges.
#'
#' @param range Range to be split up.
#'
#' @return
#' Returns vector of column ranges.
#'
#' @noRd
split_up_ranges <- function(range){
    # Extract the row numbers from range
    rows <- unlist(regmatches(range, gregexpr("[0-9]+", range)))

    start_row <- rows[1]
    end_row   <- rows[2]

    # Extract the column letters from range
    columns <- unlist(regmatches(range, gregexpr("[A-Z]+", range)))

    # Build a sequence of letters for the provided range
    column_numbers  <- openxlsx2::col2int(paste(columns[1], columns[2], sep = ":"))
    column_sequence <- openxlsx2::int2col(column_numbers)

    # Combine everything back together to individual columns
    paste0(column_sequence, start_row, ":", column_sequence, end_row)
}

###############################################################################
# Style options
###############################################################################
#' Style for 'Excel' Table Outputs
#'
#' @description
#' Set different options which define the visual output of 'Excel' tables produced
#' by [frequencies()], [crosstabs()] and [any_table()].
#'
#' @param save_path If NULL, opens the output as temporary file. Otherwise specify an output path.
#' @param file If NULL, opens the output as temporary file. Otherwise specify a filename with extension.
#' @param sheet_name Name of the sheet inside the workbook to which the output shall be written.
#' If multiple outputs are produced in one go, the sheet name additionally receives a running number.
#' @param font Set the font to be used for the entire output.
#' @param column_widths Specify whether column widths should be set automatically and individually or
#' if a numeric vector is passed each column width can be specified manually. If a table has more
#' columns than column widths are provided, the last given column width will be repeated until the
#' end of the table.
#' @param row_heights Specify whether row heights should be set automatically and individually or
#' if a numeric vector is passed each row height can be specified manually. If a table has more
#' rows than row heights are provided, the last given row height will be repeated until the
#' end of the table.
#' @param title_heights Set individual row heights for the titles only.
#' @param header_heights Set individual row heights for the table header only.
#' @param subheader_heights Set individual row heights for the table subheader only.
#' @param table_heights Set individual row heights for the table body only.
#' @param footnote_heights Set individual row heights for the footnotes only.
#' @param start_row The row in which the table starts.
#' @param start_column The column in which the table starts.
#' @param freeze_col_header Whether to freeze the column header so that it is always visible
#' while scrolling down the document.
#' @param freeze_row_header Whether to freeze the row header so that it is always visible
#' while scrolling sideways in the document.
#' @param filters Whether to set filters in the column header, when exporting a data frame.
#' @param grid_lines Whether to show grid lines or not.
#' @param by_as_subheaders Whether to format by variables as subheaders in one table instead
#' of single tables on multiple sheets.
#' @param background_color Background cell color for any cell which isn't covered by the other
#' background color options.
#' @param header_back_color Background cell color of the table header.
#' @param header_font_color Font color of the table header.
#' @param header_font_size Font size of the table header.
#' @param header_font_bold Whether to print the table header in bold letters.
#' @param header_alignment Set the text alignment of the table header.
#' @param header_stat_merging Set how far the statistics row is merged. Allowed are "all",
#' "block" (default), "none".
#' @param header_wrap Whether to wrap the texts in the table header.
#' @param header_indent Indentation level of the table header.
#' @param header_borders Whether to draw borders around the table header cells.
#' @param header_border_color Borders colors of the table header cells.
#' @param subheader_back_color Background cell color of the table subheader.
#' @param subheader_font_color Font color of the table subheader.
#' @param subheader_font_size Font size of the table subheader.
#' @param subheader_font_bold Whether to print the table subheader in bold letters.
#' @param subheader_alignment Set the text alignment of the table subheader.
#' @param subheader_wrap Whether to wrap the texts in the table subheader.
#' @param subheader_indent Indentation level of the table subheader.
#' @param subheader_borders Whether to draw borders around the table subheader cells.
#' @param subheader_border_color Borders colors of the table subheader cells.
#' @param cat_col_back_color Background cell color of the category columns inside the table.
#' @param cat_col_font_color Font color of the category columns inside the table.
#' @param cat_col_font_size Font size of the category columns inside the table.
#' @param cat_col_font_bold Whether to print the category columns inside the table in bold letters.
#' @param cat_col_alignment Set the text alignment of the category columns inside the table.
#' @param cat_col_wrap Whether to wrap the texts in the category columns inside the table.
#' @param cat_col_indent Indentation level of the category columns inside the table.
#' @param cat_col_borders Whether to draw borders around the category columns inside the table.
#' @param cat_col_border_color Borders colors of the category columns inside the table.
#' @param table_back_color Background color of the inner table cells.
#' @param table_font_color Font color of the inner table cells.
#' @param table_font_size Font size of the inner table cells.
#' @param table_font_bold Whether to print the inner table cells in bold numbers
#' @param table_alignment Set the text alignment of the inner table cells.
#' @param table_indent Indentation level of the inner table cells.
#' @param table_borders Whether to draw borders around the inner table cells.
#' @param table_border_color Borders colors of the inner table cells.
#' @param as_heatmap Whether to lay a conditional formatting over the values.
#' @param heatmap_low_color The color for lower values in the conditional formatting.
#' @param heatmap_middle_color The color for middle values in the conditional formatting.
#' @param heatmap_high_color The color for high values in the conditional formatting.
#' @param box_back_color Background color of the left box in table header.
#' @param box_font_color Font color of the left box in table header.
#' @param box_font_size Font size of the left box in table header.
#' @param box_font_bold Whether to print the left box in table header in bold letters.
#' @param box_alignment Set the text alignment of the left box in table header.
#' @param box_wrap Whether to wrap the texts in the left box in table header.
#' @param box_indent Indentation level of the left box in table header.
#' @param box_borders Whether to draw borders around the left box in table header.
#' @param box_border_color Borders colors of the left box in table header.
#' @param number_formats Put in a list of number formats which should be assigned to
#' the different stats. Number formats can be created with [number_format_style()].
#' @param title_font_color Font color of the titles.
#' @param title_font_size Font size of the tables titles.
#' @param title_font_bold Whether to print the tables titles in bold letters.
#' @param title_alignment Set the text alignment of the titles.
#' @param footnote_font_color Font color of the footnotes
#' @param footnote_font_size Font size of the tables footnotes
#' @param footnote_font_bold Whether to print the tables footnotes in bold letters.
#' @param footnote_alignment Set the text alignment of the footnotes.
#' @param na_symbol Define the symbol that should be used for NA values.
#'
#' @details
#' [excel_output_style()] is based on the Output Delivery System (ODS) in 'SAS',
#' which provides efficient and readable ways to set up different table styles.
#'
#' With the output style you have full control over the table design. There is no
#' need to think about calculating the right place to input a background color or a
#' border of a certain type and how to do this in a loop for multiple cells. Just
#' input colors, borders, font styles, etc. for the different table parts and everything
#' else is handled by the functions capable of using styles.
#'
#' The concept basically is: design over complex calculations.
#'
#' @return
#' Returns a list of named style options.
#'
#' @seealso
#' Creating a custom table style: [modify_output_style()],
#' [number_format_style()], [modify_number_formats()].
#'
#' Global style options: [set_style_options()], [set_labels()].
#'
#' Functions that can handle styles: [frequencies()], [crosstabs()], [any_table()],
#' [export_with_style()]
#'
#' @examples
#' # For default values
#' excel_style <- excel_output_style()
#'
#' # Set specific options, the rest will be set to default values
#' excel_style <- excel_output_style(font       = "Calibri",
#'                                   sheet_name = "My_Output")
#'
#' # For cells with no background color pass an empty string
#' excel_style <- excel_output_style(table_back_color = "")
#'
#' @export
excel_output_style <- function(save_path              = NULL,
                               file                   = NULL,
                               sheet_name             = "Table",
                               font                   = "Arial",
                               column_widths          = "auto",
                               row_heights            = "auto",
                               title_heights          = NULL,
                               header_heights         = NULL,
                               subheader_heights      = NULL,
                               table_heights          = NULL,
                               footnote_heights       = NULL,
                               start_row              = 2,
                               start_column           = 2,
                               freeze_col_header      = FALSE,
                               freeze_row_header      = FALSE,
                               filters                = TRUE,
                               grid_lines             = TRUE,
                               by_as_subheaders       = FALSE,
                               background_color       = "FFFFFF",
                               header_back_color      = "FFFFFF",
                               header_font_color      = "000000",
                               header_font_size       = 10,
                               header_font_bold       = TRUE,
                               header_alignment       = "center",
                               header_stat_merging    = "block",
                               header_wrap            = "1",
                               header_indent          = 0,
                               header_borders         = TRUE,
                               header_border_color    = "000000",
                               subheader_back_color   = "FFFFFF",
                               subheader_font_color   = "000000",
                               subheader_font_size    = 10,
                               subheader_font_bold    = TRUE,
                               subheader_alignment    = "center",
                               subheader_wrap         = "1",
                               subheader_indent       = 0,
                               subheader_borders      = TRUE,
                               subheader_border_color = "000000",
                               cat_col_back_color     = "FFFFFF",
                               cat_col_font_color     = "000000",
                               cat_col_font_size      = 10,
                               cat_col_font_bold      = FALSE,
                               cat_col_alignment      = "left",
                               cat_col_wrap           = "1",
                               cat_col_indent         = 1,
                               cat_col_borders        = TRUE,
                               cat_col_border_color   = "000000",
                               table_back_color       = "FFFFFF",
                               table_font_color       = "000000",
                               table_font_size        = 10,
                               table_font_bold        = FALSE,
                               table_alignment        = "right",
                               table_indent           = 1,
                               table_borders          = FALSE,
                               table_border_color     = "000000",
                               as_heatmap             = FALSE,
                               heatmap_low_color      = "F8696B",
                               heatmap_middle_color   = "FFFFFF",
                               heatmap_high_color     = "63BE7B",
                               box_back_color         = "FFFFFF",
                               box_font_color         = "000000",
                               box_font_size          = 10,
                               box_font_bold          = TRUE,
                               box_alignment          = "center",
                               box_wrap               = "1",
                               box_indent             = 0,
                               box_borders            = TRUE,
                               box_border_color       = "000000",
                               number_formats         = number_format_style(),
                               title_font_color       = "000000",
                               title_font_size        = 10,
                               title_font_bold        = TRUE,
                               title_alignment        = "left",
                               footnote_font_color    = "000000",
                               footnote_font_size     = 8,
                               footnote_font_bold     = FALSE,
                               footnote_alignment     = "left",
                               na_symbol              = "."){

    as.list(environment())
}


#' Modify Style for 'Excel' Table Outputs
#'
#' @description
#' Modify a previously created style with [excel_output_style()].
#'
#' @param style_to_modify A pre created style where only certain elements should be
#' modified while the rest is kept as is.
#' @param ... Pass in names and corresponding new values for existing style elements.
#'
#' @details
#' [modify_output_style()] is based on the Output Delivery System (ODS) in 'SAS',
#' which provides efficient and readable ways to set up different table styles.
#'
#' With the output style you have full control over the table design. There is no
#' need to think about calculating the right place to input a background color or a
#' border of a certain type and how to do this in a loop for multiple cells. Just
#' input colors, borders, font styles, etc. for the different table parts and everything
#' else is handled by the functions capable of using styles.
#'
#' The concept basically is: design over complex calculations.
#'
#' @return
#' Returns a modified list of named style options.
#'
#' @seealso
#' Creating a custom table style: [excel_output_style()],
#' [number_format_style()], [modify_number_formats()].
#'
#' Global style options: [set_style_options()], [set_labels()].
#'
#' Functions that can handle styles: [frequencies()], [crosstabs()], [any_table()],
#' [export_with_style()]
#'
#' @examples
#' # For default values
#' excel_style <- excel_output_style()
#'
#' # Set specific options, the rest will be kept as is
#' excel_style <- excel_style |> modify_output_style(sheet_name      = "Sheet",
#'                                                   title_font_bold = FALSE)
#'
#' # For cells with no background color pass an empty string
#' excel_style <- excel_style |> modify_output_style(table_back_color = "")
#'
#' @export
modify_output_style <- function(style_to_modify, ...){
    style_elements <- list(...)

    # Loop through all elements to modify and if exists then set new value
    for (element in seq_along(style_elements)){
        name <- names(style_elements)[element]

        if (!name %in% names(style_to_modify)){
            print_message("WARNING", "Style element '{name}' is invalid and will be omitted.", name = name)
        }

        style_to_modify[[name]] <- style_elements[[element]]
    }

    style_to_modify
}


###############################################################################
# Number formats for any_table
###############################################################################
#' Number Formats Used by [any_table()]
#'
#' @description
#' Set individual number formats for the different statistics in tables produced
#' with [any_table()].
#'
#' @param pct_excel Number format for percentage applied in Excel workbook.
#' @param freq_excel Number format for frequency applied in Excel workbook.
#' @param freq.g0_excel Number format for frequency greater zero applied in Excel workbook.
#' @param sum_excel Number format for sum applied in Excel workbook.
#' @param sum.wgt_excel Number format for sum of weights applied in Excel workbook.
#' @param mean_excel Number format for mean applied in Excel workbook.
#' @param median_excel Number format for median applied in Excel workbook.
#' @param mode_excel Number format for mode applied in Excel workbook.
#' @param min_excel Number format for min applied in Excel workbook.
#' @param max_excel Number format for max applied in Excel workbook.
#' @param sd_excel Number format for sd applied in Excel workbook.
#' @param variance_excel Number format for variance applied in Excel workbook.
#' @param first_excel Number format for first applied in Excel workbook.
#' @param last_excel Number format for last applied in Excel workbook.
#' @param missing_excel Number format for missing applied in Excel workbook.
#' @param p_excel Number format for percentile applied in Excel workbook.
#' @param pct_decimals Number of decimals for percentage.
#' @param freq_decimals Number of decimals for frequency.
#' @param freq.g0_decimals Number of decimals for frequency greater zero.
#' @param sum_decimals Number of decimals for sum.
#' @param sum.wgt_decimals Number of decimals for sum of weights.
#' @param mean_decimals Number of decimals for mean.
#' @param median_decimals Number of decimals for median.
#' @param mode_decimals Number of decimals for mode.
#' @param min_decimals Number of decimals for min.
#' @param max_decimals Number of decimals for max.
#' @param sd_decimals Number of decimals for sd.
#' @param variance_decimals Number of decimals for variance.
#' @param first_decimals Number of decimals for first.
#' @param last_decimals Number of decimals for last.
#' @param missing_decimals Number of decimals for missing.
#' @param p_decimals Number of decimals for percentile.
#'
#' @details
#' [number_format_style()] is based on 'SAS' number formats and the Output Delivery
#' System (ODS), which provides efficient and readable ways to set up different table
#' styles.
#'
#' With the number format style you have full control over formatting numbers according
#' to the different statistics. There is no need to think about calculating the right
#' place to input the number formats and how to do this in a loop for multiple cells.
#' Just input the different number formats and decimals for the different statistics
#' and everything else is handled by the functions capable of using number styles.
#'
#' The concept basically is: design over complex calculations.
#'
#' @return
#' Returns a list of named number format options.
#'
#' @seealso
#' Creating a custom table style: [excel_output_style()], [modify_output_style()],
#' [modify_number_formats()].
#'
#' Global style options: [set_style_options()], [set_labels()].
#'
#' Functions that can handle styles: [frequencies()], [crosstabs()], [any_table()],
#' [export_with_style()]
#'
#' @examples
#' # For default values
#' format_list <- number_format_style()
#'
#' # Set specific options, the rest will be set to default values
#' format_list <- number_format_style(pct_excel    = "0.00000000",
#'                                    pct_decimals = 8)
#'
#' # IMPORTANT: Don't forget to add individual formats to an excel style, otherwise
#' # they won't come into affect.
#' excel_style <- excel_output_style(number_formats = format_list)
#'
#' @export
number_format_style <- function(pct_excel         = "0.0",
                                freq_excel        = "#,###,##0",
                                freq.g0_excel     = "#,###,##0",
                                sum_excel         = "#,###,##0",
                                sum.wgt_excel     = "#,###,##0",
                                mean_excel        = "#,###,##0",
                                median_excel      = "#,###,##0",
                                mode_excel        = "#,###,##0",
                                min_excel         = "#,###,##0",
                                max_excel         = "#,###,##0",
                                sd_excel          = "#,###,##0.000",
                                variance_excel    = "#,###,##0.000",
                                first_excel       = "#,###,##0",
                                last_excel        = "#,###,##0",
                                p_excel           = "#,###,##0",
                                missing_excel     = "#,###,##0",
                                pct_decimals      = 1,
                                freq_decimals     = 0,
                                freq.g0_decimals  = 0,
                                sum_decimals      = 3,
                                sum.wgt_decimals  = 3,
                                mean_decimals     = 2,
                                median_decimals   = 2,
                                mode_decimals     = 2,
                                min_decimals      = 2,
                                max_decimals      = 2,
                                sd_decimals       = 3,
                                variance_decimals = 3,
                                first_decimals    = 0,
                                last_decimals     = 0,
                                p_decimals        = 2,
                                missing_decimals  = 0){

    as.list(environment())
}


#' Modify Number Formats Used by [any_table()]
#'
#' @description
#' Modify previously created number formats with [number_format_style()].
#'
#' @param formats_to_modify Pre created number formats where only certain elements
#' should be modified while the rest is kept as is.
#' @param ... Pass in names and corresponding new values for existing number formats.
#'
#' @details
#' [modify_number_formats()] is based on 'SAS' number formats and the Output Delivery
#' System (ODS), which provides efficient and readable ways to set up different table
#' styles.
#'
#' With the number format style you have full control over formatting numbers according
#' to the different statistics. There is no need to think about calculating the right
#' place to input the number formats and how to do this in a loop for multiple cells.
#' Just input the different number formats and decimals for the different statistics
#' and everything else is handled by the functions capable of using number styles.
#'
#' The concept basically is: design over complex calculations.
#'
#' @return
#' Returns a modified list of number format options.
#'
#' @seealso
#' Creating a custom table style: [excel_output_style()], [modify_output_style()],
#' [number_format_style()].
#'
#' Global style options: [set_style_options()], [set_labels()].
#'
#' Functions that can handle styles: [frequencies()], [crosstabs()], [any_table()],
#' [export_with_style()].
#'
#' @examples
#' # For default values
#' format_list <- number_format_style(pct_excel    = "0.00000000",
#'                                    pct_decimals = 8)
#'
#' # Set specific options, the rest will be kept as is
#' format_list <- format_list |> modify_number_formats(sum_excel = "#,###,##0.000")
#'
#' # IMPORTANT: Don't forget to add individual formats to an excel style, otherwise
#' # they won't come into affect.
#' excel_style <- excel_output_style(number_formats = format_list)
#'
#' @export
modify_number_formats <- function(formats_to_modify, ...){
    format_elements <- list(...)

    # Loop through all elements to modify and if exists then set new value
    for (element in seq_along(format_elements)){
        name <- names(format_elements)[element]

        if (!name %in% names(formats_to_modify)){
            print_message("WARNING", "Number format '[name]' is invalid and will be omitted.", name = name)
        }

        formats_to_modify[[name]] <- format_elements[[element]]
    }

    formats_to_modify
}

set_no_print(TRUE)

# Modfiy Excel style
my_style <- excel_output_style()
my_style <- my_style |> modify_output_style(font = "Calibri")

expect_equal(my_style[["font"]], "Calibri", info = "Modfiy Excel style")


# Add none existent element to style
my_style <- excel_output_style()
my_style <- my_style |> modify_output_style(test = "Calibri")

expect_warning(print_stack_as_messages("WARNING"), "Style element 'test' is invalid and will", info = "Add none existent element to style")


# Modfiy number format style
my_num_fmt <- number_format_style()
my_num_fmt <- my_num_fmt |> modify_number_formats(pct_decimals = 8)

expect_equal(my_num_fmt[["pct_decimals"]], 8, info = "Modfiy number format style")


# Modfiy number format style
my_num_fmt <- number_format_style()
my_num_fmt <- my_num_fmt |> modify_number_formats(test = 8)

expect_warning(print_stack_as_messages("WARNING"), "Number format 'test' is invalid and will", info = "Modfiy number format style")


set_no_print()

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
# Printing the output is always set to FALSE, so the overall code can be tested
# but without drawing the whole outputs on screen.
###############################################################################

set_style_options(as_heatmap = TRUE)

dummy_df <- dummy_data(100)
dummy_df[["weight_sum"]] <- dummy_df[["weight"]]


# Simplest form of export_with_style
result_df <- dummy_df |> export_with_style(print = FALSE)

expect_inherits(result_df, "qol_styled_data", info = "Simplest form of export_with_style")
expect_equal(length(result_df), 3, info = "Simplest form of export_with_style")


# export_with_style with full styling
result_df <- dummy_df |>
      export_with_style(titles     = "Hello world link: https://cran.r-project.org/",
                        footnotes  = "This is a footnote link: https://cran.r-project.org/",
                        var_labels = list(weight_sum = "Test label"),
                        style      = excel_output_style(freeze_col_header = TRUE,
                                                        freeze_row_header = TRUE,
                                                        background_color  = "FF00FF"),
                        print = FALSE)

expect_inherits(result_df, "qol_styled_data", info = "export_with_style with full styling")
expect_equal(length(result_df), 3, info = "Simplest form of export_with_style")


# export_with_style with individual column alignment
result_df <- dummy_df |> export_with_style(column_align = c("right", "left"), print = FALSE)

expect_inherits(result_df, "qol_styled_data", info = "export_with_style with individual column alignment")
expect_equal(length(result_df), 3, info = "export_with_style with individual column alignment")



# export_with_style with fixed row headers
result_df <- dummy_df |>
      export_with_style(style = excel_output_style(freeze_row_header = TRUE),
                        print = FALSE)

expect_inherits(result_df, "qol_styled_data", info = "export_with_style with fixed row headers")
expect_equal(length(result_df), 3, info = "Simplest form of export_with_style")



# export_with_style with fixed column headers
result_df <- dummy_df |>
      export_with_style(style = excel_output_style(freeze_col_header = TRUE),
                        print = FALSE)

expect_inherits(result_df, "qol_styled_data", info = "export_with_style with fixed column headers")
expect_equal(length(result_df), 3, info = "Simplest form of export_with_style")


# Save export_with_style as Excel file
temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

dummy_df |>
     export_with_style(style = excel_output_style(save_path = dirname(temp_file),
                                                  file      = basename(temp_file)))

expect_true(file.exists(temp_file), info = "Save export_with_style as Excel file")


# export_with_style can pass on workbooks
my_style <- excel_output_style(sheet_name = "tab1")

df1 <- dummy_df |>
    export_with_style(style = my_style,
                      print = FALSE)

my_style <- my_style |> modify_output_style(sheet_name = "tab2")

result_list <- dummy_df |>
    export_with_style(style    = my_style,
                      workbook = df1,
                      print    = FALSE)

expect_inherits(result_list, "qol_styled_data", info = "export_with_style can pass on workbooks")
expect_equal(length(result_list), 3, info = "export_with_style can pass on workbooks")


# export_with_style aborts, if invalid workbook is passed
result_list <- dummy_df |>
    export_with_style(workbook = list("test" = "test"),
                      print    = FALSE)

expect_error(print_stack_as_messages("ERROR"), "Workbook object is invalid. You have to provide a workbook object",
             info = "export_with_style aborts, if invalid workbook is passed")


set_style_options(as_heatmap = FALSE)
set_no_print()

#' Recode New Variables With Formats
#'
#' @name recode
#'
#' @description
#' Instead of writing multiple if-clauses to recode values into a new variable,
#' you can use formats to recode a variable into a new one.
#'
#' @param data_frame A data frame which contains the the original variables to recode.
#' @param ... [recode.()] Pass in the original variable name that should be recoded
#' along with the corresponding format container in the form: variable = format.
#'
#' In [recode_multi()] multiple variables can be recoded in one go and multilabels
#' can be applied. This overwrites the original variables and duplicates rows if
#' multilabels are applied. In occasions were you want to use format containers to
#' afterwards perform operations with other packages, you can make use of this
#' principle with this function.
#'
#' @details
#' [recode.()] is based on the 'SAS' function put(), which provides an efficient
#' and readable way, to generate new variables with the help of formats.
#'
#' When creating a format you can basically write code like you think: This new
#' category consists of these original values. And after that you just apply these
#' new categories to the original values to create a new variable. No need for multiple
#' if_else statements.
#'
#' @return
#' [recode.()]: Returns a vector with recoded values.
#'
#' @seealso
#' Creating formats: [discrete_format()] and [interval_format()].
#'
#' Functions that also make use of formats: [frequencies()], [crosstabs()],
#' [any_table()], [summarise_plus()], [transpose_plus()], [sort_plus()]
#'
#' @examples
#' # Example formats
#' age. <- discrete_format(
#'     "under 18"       = 0:17,
#'     "18 to under 25" = 18:24,
#'     "25 to under 55" = 25:54,
#'     "55 to under 65" = 55:64,
#'     "65 and older"   = 65:100)
#'
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Call function
#' my_data[["age_group1"]] <- my_data |> recode.(age = age.)
#'
#' # Formats can also be passed as characters
#' my_data[["age_group2"]] <- my_data |> recode.(age = "age.")
#'
#' # Multilabel recode
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' income. <- interval_format(
#'     "Total"              =    0:100000,
#'     "below 500"          =    0:500,
#'     "500 to under 1000"  =  500:1000,
#'     "1000 to under 2000" = 1000:2000,
#'     "2000 and more"      = 2000:100000)
#'
#' # recode_multi() can not only apply multiple recodings, but it can also
#' # apply multilabels.
#' # NOTE: Recoding will always be in place. When applying multilabels the
#' #       result data frame will have more observations than before.
#' multi_data <- my_data |> recode_multi(sex = sex., income = income.)
#'
#' @rdname recode
#'
#' @export
recode. <- function(data_frame,
                    ...){
    # Measure the time
    print_start_message(suppress = TRUE)

    ###########################################################################
    # Early evaluations
    ###########################################################################

    # Translate ... into separately controllable arguments
    formats <- tryCatch({
        # Force evaluation to see if it exists
        list(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(formats)){
        print_message("ERROR", "Unknown object found. Recode will be aborted.")
        return(invisible(NULL))
    }

    # Evaluate formats early
    if (!is_list_of_dfs(formats)){
        formats <- evaluate_formats(formats)
    }

    ###########################################################################
    # Error handling
    ###########################################################################

    # Get information from ... list
    current_var <- names(formats)[1]
    format_df   <- formats[[current_var]]

    if (!current_var %in% names(data_frame)){
        print_message("ERROR", "Variable '[var]' not found in the input data frame. No format will be applied.", var = current_var)
        return(invisible(NULL))
    }

    if (!data.table::is.data.table(format_df)){
        print_message("ERROR", "The format for '[var]' must be a data table. No format will be applied.", var = current_var)
        return(invisible(NULL))
    }

    if (names(format_df)[1] == "value" && collapse::any_duplicated(format_df[["value"]])){
        print_message("WARNING", c("The format for '[var]' is a multilabel. A multilabel can't be fully applied in recode.",
								   "Only one of the matching categories will be applied."), var = current_var)

        format_df <- format_df |> unique(by = "value", fromLast = FALSE)
    }

    if (is.factor(data_frame[[current_var]])){
        print_message("NOTE", c("'[var]' is a factor variable. Formats only work if the visible character values",
								"are specified as input values and not the factor levels."), var = current_var)
    }

    ###########################################################################
    # Recode
    ###########################################################################

    # Only keep the variable to be recoded
    data_frame <- data_frame |> keep(current_var)

    # Look up variable names in format data frame to check whether it is an
    # interval or discrete format
    interval_variables <- c("from", "to")
    actual_variables   <- names(format_df)[1:2]

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # In case of interval format
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (identical(interval_variables, actual_variables)){
        # Reduce multilabel formats to unique ranges which don't overlap
        format_original <- format_df
        format_df       <- format_df |> collapse::fsubset(from >= data.table::shift(cummax(to), fill = -Inf))

        if (collapse::fnrow(format_original) > collapse::fnrow(format_df)){
            print_message("WARNING", c("The format for '[current_var]' is a multilabel. A multilabel can't be fully applied in recode.",
                                       "Only one of the matching categories will be applied."), current_var = current_var)
        }

        # Get number of rows from data frame to compare after the merge to check for multilabel
        original_rows <- collapse::fnrow(data_frame)

        # Separate NAs from rest of the data frame because the used join
        # can't handle them
        na_positions <- is.na(data_frame[[current_var]])
        data_frame   <- data_frame |> collapse::fsubset(!na_positions)

        # Generate pseudo variables for range merging
        data_frame[["qol_from"]] <- data_frame[[as.character(current_var)]]
        data_frame[["qol_to"]]   <- data_frame[[as.character(current_var)]]

        data_frame[["qol_ID"]] <- seq_len(collapse::fnrow(data_frame))

        # Set key variables
        temp_dt   <- data.table::as.data.table(data_frame)
        format_dt <- data.table::copy(format_df)

        data.table::setkey(format_dt, from, to)

        # Merge data frame with format by range
        temp_dt <- data.table::foverlaps(temp_dt, format_dt,
                                         by.x = c("qol_from", "qol_to"),
                                         by.y = c("from", "to")) |>
            keep("label")

        # NA values are now inserted in the same spots as they where before to
        # ensure that there will be no missmatch with the original data frame.
        data_frame                <- rep(NA, original_rows)
        data_frame[!na_positions] <- temp_dt[["label"]]
        data_frame                <- data.table::as.data.table(data_frame)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # In case of discrete format
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    else{
        # Rename label column to be specific to the variable
        format_df <- format_df |>
            collapse::frename(stats::setNames("value", current_var))

        # Join format with data frame
        data_frame <- data_frame |>
            collapse::join(format_df,
                           on      = current_var,
                           how     = "left",
                           verbose = FALSE,
						   overid  = 2) |>
            keep("label")
    }

    print_closing()

    as.vector(data_frame)[[1]]
}


#' @param convert FALSE by default. If TRUE converts recoded variables to numeric
#' or character depending on the input format instead of leaving them as factors.
#'
#' @return
#' [recode_multi()]: Returns a data frame with the newly recoded variable.
#'
#' @rdname recode
#'
#' @export
recode_multi <- function(data_frame, ..., convert = FALSE){
    # Measure the time
    print_start_message(suppress = TRUE)

    # Translate ... into a list if possible
    formats <- tryCatch({
        # Force evaluation to see if it exists
        list(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(formats)){
        print_message("ERROR", c("Unknown object found. Provide recode arguments in the form: variable_name = format_df.",
								 "Recoding will be aborted."))
        return(data_frame)
    }

    # Evaluate formats early
    if (!is_list_of_dfs(formats)){
        formats <- evaluate_formats(formats)
    }

    # Get information from ... list
    variables <- names(formats)
    var_order <- names(data_frame)

    # Apply formats
    data_frame <- data_frame |>
        apply_format(formats, variables) |>
        data.table::setcolorder(var_order)

    # Convert formatted values to intended type
    if (convert){
        for (variable in variables){
            if (is.numeric(formats[[variable]][["label"]])){
                data_frame[[variable]] <- as.numeric(data_frame[[variable]])
            }
            else{
                data_frame[[variable]] <- as.character(data_frame[[variable]])
            }
        }
    }

    print_closing(5)

    data_frame
}

#' Dummy Data
#'
#' @description
#' The dummy data frame contains a few randomly generated variables like year, sex,
#' age, income and weight to test out functionalities. It can be generated with the desired
#' number of observations.
#'
#' @param no_obs Number of observations.
#' @param insert_na TRUE by default. Inserts random NA values into variables.
#' @param monitor FALSE by default. If TRUE outputs two charts to visualize the functions time consumption.
#'
#' @return Returns a dummy data table.
#'
#' @examples
#' my_data <- dummy_data(1000)
#'
#' @export
dummy_data <- function(no_obs    = 25000,
                       insert_na = TRUE,
                       monitor   = .qol_options[["monitor"]]){
    # Measure the time
    print_start_message()

    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Preparation")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Generating dummy file. Computing ...")

    # Reduce the number of observations to the number of households, because different
    # household sizes will be generated which basically expands this number to the number
    # of observations again. Additionally the number of households is divided by three
    # because at the end the cases are doubled up for three years. Working with this
    # reduced number significantly speeds up the case generation.
    if (no_obs < 1000){
        number_of_households <- max(1, no_obs)
    }
    else if (no_obs < 100000){
        number_of_households <- as.integer(no_obs / (1.8 * 5))
    }
    else{
        number_of_households <- as.integer(no_obs / (2.2 * 4))
    }

    # Prepare probabilities so that each state has a different number of observations
    state_probs <- c(3.5, 2.2, 9.6, 0.8, 21.6, 7.6, 5.0, 13.5, 15.8, 1.2, 4.4, 3.0, 1.9, 4.9, 2.6, 2.5)

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate households")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "households and states")

    # Household sizes are generated with different probabilities, so that smaller numbers
    # are generated more frequently.
    persons_per_household <- sample(1:7, size = number_of_households, replace = TRUE,
                                    prob = c(0.4, 0.27, 0.15, 0.08, 0.06, 0.03, 0.01))

    # Generate unique household ids and a running number for each person inside the
    # household
    dummy_temp <- data.table::data.table(
        household_id = rep(1:number_of_households, times = persons_per_household))

    dummy_temp[["person_id"]] <- sequence(persons_per_household)

    # Randomly generate sixteen states
    dummy_temp[["state"]] <- as.integer(rep(sample(1:16, number_of_households,
                                                   replace = TRUE, prob = state_probs),
                                            times   = persons_per_household))

    # Save the original number of observations to later trim the data frame to this number
    orig_obs <- no_obs
    no_obs   <- collapse::fnrow(dummy_temp)

    # Some variables will be generated on the level of households. Prepare a group for this.
    group_hh <- collapse::GRP(dummy_temp[["household_id"]])

    # Prepare variable to identify the first person per household
    dummy_temp[["first_person"]] <- data.table::fifelse(dummy_temp[["person_id"]] == 1, 1, 0)

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate weights")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "weights")

    # Generate unique weights per person and also pass on the first weight per household
    # onto all other persons in the same household.
    scale_factor <- 25000 / orig_obs
    dummy_temp[["weight"]] <- stats::runif(no_obs, 2.9, 3.7) * scale_factor
    dummy_temp[["weight"]] <- dummy_temp[["weight"]] |> collapse::ffirst(g = group_hh, TRA = "fill")
    dummy_temp[["weight_per_year"]] <- dummy_temp[["weight"]] * 5

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate sex")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "sex")

    # Generate sex variable, but let each state have a different distribution
    sex_probs <- stats::runif(16, 0.3, 0.7)

    dummy_temp[["sex"]] <- 1 + stats::rbinom(no_obs, 1, sex_probs[dummy_temp[["state"]]])

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate age")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "age")

    # Generate random ages
    dummy_temp[["age"]] <- floor(stats::runif(no_obs, min = 25, max = 55))

    # Get the positions of certain persons depending on their person id
    id_1  <- which(dummy_temp[["person_id"]] == 1)
    id_34 <- which(dummy_temp[["person_id"]] %in% c(3, 4))
    id_56 <- which(dummy_temp[["person_id"]] %in% c(5, 6))

    # Bring a little variety to the age distribution by generating younger and older
    # persons in specific spots
    dummy_temp[id_1,  "age"] <- floor(stats::runif(length(id_1),  min = 18, max = 95))
    dummy_temp[id_34, "age"] <- floor(stats::runif(length(id_34), min =  0, max =  25))
    dummy_temp[id_56, "age"] <- floor(stats::runif(length(id_56), min = 50, max = 95))

    # Bring in more variety by making people younger or older depending on the state
    # they live in.
    dummy_temp[["age"]] <- data.table::fifelse(dummy_temp[["state"]] >= 11, dummy_temp[["age"]] + 5, dummy_temp[["age"]] - 5)
    dummy_temp[["age"]] <- data.table::fifelse(dummy_temp[["sex"]] == 1,    dummy_temp[["age"]] + 3, dummy_temp[["age"]] - 2)
    dummy_temp[["age"]] <- data.table::fifelse(dummy_temp[["age"]] < 0,     0,                       dummy_temp[["age"]])
    dummy_temp[["age"]] <- data.table::fifelse(dummy_temp[["age"]] > 95,    95,                      dummy_temp[["age"]])

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate income")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "income")

    # Generate random incomes and bring a bit variety to the distribution depending
    # on certain aspects like age, sex, and state.
    dummy_temp[["income"]] <- stats::runif(no_obs, 0, 5000)
    dummy_temp[["income"]] <- data.table::fifelse(dummy_temp[["age"]] < 18,                             0,                             dummy_temp[["income"]])
    dummy_temp[["income"]] <- data.table::fifelse(dummy_temp[["age"]] >= 30 & dummy_temp[["age"]] < 50, dummy_temp[["income"]] * 1.2,  dummy_temp[["income"]])
    dummy_temp[["income"]] <- data.table::fifelse(dummy_temp[["age"]] >= 50 & dummy_temp[["age"]] < 65, dummy_temp[["income"]] * 1.5,  dummy_temp[["income"]])
    dummy_temp[["income"]] <- data.table::fifelse(dummy_temp[["age"]] >= 65,                            dummy_temp[["income"]] * 0.75, dummy_temp[["income"]])
    dummy_temp[["income"]] <- data.table::fifelse(dummy_temp[["sex"]] == 1,                             dummy_temp[["income"]] * 0.8,  dummy_temp[["income"]])
    dummy_temp[["income"]] <- data.table::fifelse(dummy_temp[["state"]] >= 11,                          dummy_temp[["income"]] * 0.7,  dummy_temp[["income"]])

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate expenses and balance")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "expenses and balance")

    # Generate random expenses and bring a bit variety to the distribution depending
    # on certain aspects like age, sex, and state.
    dummy_temp[["expenses"]] <- stats::runif(no_obs, -5000, 0)
    dummy_temp[["expenses"]] <- data.table::fifelse(dummy_temp[["age"]] < 18,                             0,                               dummy_temp[["expenses"]])
    dummy_temp[["expenses"]] <- data.table::fifelse(dummy_temp[["age"]] >= 30 & dummy_temp[["age"]] < 50, dummy_temp[["expenses"]] * 1.2,  dummy_temp[["expenses"]])
    dummy_temp[["expenses"]] <- data.table::fifelse(dummy_temp[["age"]] >= 50 & dummy_temp[["age"]] < 65, dummy_temp[["expenses"]] * 1.5,  dummy_temp[["expenses"]])
    dummy_temp[["expenses"]] <- data.table::fifelse(dummy_temp[["age"]] >= 65,                            dummy_temp[["expenses"]] * 0.75, dummy_temp[["expenses"]])
    dummy_temp[["expenses"]] <- data.table::fifelse(dummy_temp[["sex"]] == 1,                             dummy_temp[["expenses"]] * 0.8,  dummy_temp[["expenses"]])
    dummy_temp[["expenses"]] <- data.table::fifelse(dummy_temp[["state"]] >= 11,                          dummy_temp[["expenses"]] * 0.7,  dummy_temp[["expenses"]])

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate probability")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "probability")

    # Generate unique probabilities per person and also pass on the first probability
    # per household onto all other persons in the same household.
    prob_temp <- sample(0:2, no_obs, replace = TRUE)

    # Probability can be 0, 1 or something in between
    id_2            <- which(prob_temp == 2)
    prob_temp[id_2] <- stats::runif(length(id_2), 0, 1)

    dummy_temp[["probability"]] <- prob_temp |> collapse::ffirst(g = group_hh, TRA = "fill")

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate education")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "education")

    male_id   <- which(dummy_temp[["sex"]] == 1)
    female_id <- which(dummy_temp[["sex"]] == 2)

    dummy_temp[male_id, "education"] <- sample(c("low", "middle", "high"),
                                               length(male_id),
                                               replace = TRUE,
                                               prob    = c(0.35, 0.4, 0.25))

    dummy_temp[female_id, "education"] <- sample(c("low", "middle", "high"),
                                                 length(female_id),
                                                 replace = TRUE,
                                                 prob    = c(0.2, 0.45, 0.35))

    # People of younger ages should all have low education
    dummy_temp[dummy_temp[["age"]] < 18, "education"] <- "low"

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate body height")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "body_height")

    # Younger people should be smaller and grow steadily
    growth_factor       <- rep(1, no_obs)
    id20                <- which(dummy_temp[["age"]] < 21)
    growth_factor[id20] <- pmin(1, ((dummy_temp[id20, "age"] + 1) / 20) ^ 0.6)

    # Generate basic body height around fixed "maximum" values
    dummy_temp[["body_height"]] <- stats::rnorm(no_obs, data.table::fifelse(dummy_temp[["sex"]] == 1,
                                                                            178, 165),
                                                7) * growth_factor

    # Generate smaller and bigger spikes in steps of five and ten
    probs <- stats::runif(no_obs)

    dummy_temp[["body_height"]] <- as.integer(
        data.table::fcase(probs < 0.5, round(dummy_temp[["body_height"]] / 10) * 10,
                          probs < 0.8, round(dummy_temp[["body_height"]] / 5) * 5,
                          default = dummy_temp[["body_height"]]))

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate body weight")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "body_weight")

    # Generate some variety in body weight
    bmi                         <- stats::rlnorm(no_obs, log(24), 0.25)
    dummy_temp[["body_weight"]] <- 5 + (bmi * (dummy_temp[["body_height"]] / 100) ^ 2)

    # Generate smaller and bigger spikes in steps of five and ten
    dummy_temp[["body_weight"]] <- as.integer(
        data.table::fcase(probs < 0.5, round(dummy_temp[["body_weight"]] / 10) * 10,
                          probs < 0.8, round(dummy_temp[["body_weight"]] / 5) * 5,
                          default = dummy_temp[["body_weight"]]))

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate NUTS3")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "NUTS2 and NUTS3")

    # Read in NUTS3 codes with individual probabilities
    nuts_file            <- suppressMessages(import_data(system.file("extdata", "qol_nuts.csv", package = "qol")))
    nuts_file[["NUTS3"]] <- nuts_file |> concat("NUTS3", padding_char = "0", padding_length = 5)

    # Split up codes by state
    nuts_by_state <- split(nuts_file[["NUTS3"]],    substr(nuts_file[["NUTS3"]], 1, 2))
    prob_by_state <- split(nuts_file[["nuts_prob"]], substr(nuts_file[["NUTS3"]], 1, 2))

    # Generate subset of dummy data which only contains the first person per household.
    # Regions will only be created for them.
    region_temp <- dummy_temp |>
        collapse::fselect(c("state", "household_id", "first_person")) |>
        collapse::fsubset(first_person == 1)

    # Select random regions for the different states based on probabilities
    for (state in seq_along(1:16)){
        ids <- region_temp[["state"]] == state
        region_temp[["NUTS3"]][ids] <- sample(nuts_by_state[[state]],
                                              size    = collapse::fsum(ids),
                                              replace = TRUE,
                                              prob    = prob_by_state[[state]])
    }

    # Join regions back to all cases. Regions are joined on households so that every
    # person within a household receives the same region.
    dummy_temp <- collapse::join(dummy_temp, region_temp,
                                 on       = c("state", "household_id"),
                                 how      = "left",
                                 verbose  = FALSE,
								 overid   = 2)

    # Extract NUTS2 region from NUTS3 code
    dummy_temp[["NUTS2"]] <- substr(dummy_temp[["NUTS3"]], 3, 3)

    if (insert_na){
        #-------------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Insert NA values")
        #-------------------------------------------------------------------------#
        print_step("MAJOR", "Insert NA values")

        dummy_temp[["age"]]          <- collapse::na_insert(dummy_temp[["age"]],    prop = 0.05)
        dummy_temp[["sex"]]          <- collapse::na_insert(dummy_temp[["sex"]],    prop = 0.05)
        dummy_temp[["income"]]       <- collapse::na_insert(dummy_temp[["income"]], prop = 0.05)
        dummy_temp[["expenses"]]     <- data.table::fifelse(is.na(dummy_temp[["income"]]), NA, dummy_temp[["expenses"]])
        dummy_temp[["education"]]    <- collapse::na_insert(dummy_temp[["education"]],   prop = 0.05)
        dummy_temp[["body_height"]]  <- collapse::na_insert(dummy_temp[["body_height"]], prop = 0.05)
        dummy_temp[["body_weight"]]  <- collapse::na_insert(dummy_temp[["body_weight"]], prop = 0.05)
    }

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate years")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Expand years")

    # Sort and renumber household ids inside states
    dummy_temp <- dummy_temp |> data.table::setorderv(c("state", "household_id", "person_id"))
    dummy_temp[["household_id"]] <- stats::ave(dummy_temp[["household_id"]],
                                               dummy_temp[["state"]],
                                               FUN = function(case) match(case, unique(case)))

    # Prepare years
    current_year <- as.numeric(format(Sys.Date(), "%Y"))
    start_year   <- current_year - 4

    years_seq <- data.table::data.table(temp. = 1,
                                        year  = start_year:current_year,
                                        age_factor      = 0:4,
                                        income_factor   = c(1, 1.05, 1.1, 1.2, 1.4),
                                        expenses_factor = c(1, 1.1, 1.15, 1.2, 1.3))

    # Create cartesian product with years to multiply cases
    dummy_temp[["temp."]] <- 1
    dummy_temp <- collapse::join(years_seq, dummy_temp,
                                 on       = "temp.",
                                 how      = "left",
                                 multiple = TRUE,
                                 verbose  = FALSE,
								 overid   = 2)

    print_step("MAJOR", "Readjust variables")

    # Reduce observations to desired number of observations
    random_sample <- sort(sample.int(collapse::fnrow(dummy_temp), orig_obs))
    dummy_temp    <- dummy_temp |> collapse::fsubset(random_sample)

    # Readjust person_id and first_person in household marker because of the sample above
    renumber                     <- data.table::rleid(dummy_temp[["household_id"]])
    dummy_temp[["person_id"]]    <- sequence(tabulate(renumber))
    dummy_temp[["first_person"]] <- data.table::fifelse(dummy_temp[["person_id"]] == 1, 1, 0)

    # Get the number of persons per household
    grp <- collapse::GRP(dummy_temp, c("year", "state", "household_id"), sort = FALSE)
    dummy_temp[["number_of_persons"]] <- collapse::fmax(dummy_temp[["person_id"]], g = grp, TRA = "fill")

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Advance values")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Advance values in later years")

    # Up the age and income/expenses
    dummy_temp <- dummy_temp |> collapse::fmutate(age      = age      + age_factor,
                                                  income   = income   * income_factor,
                                                  expenses = expenses * expenses_factor,
                                                  balance  = income   + expenses)

    # Up the body height for people under 18
    dummy_temp[["body_height"]] <- as.integer(data.table::fifelse(
        dummy_temp[["year"]] > start_year & dummy_temp[["age"]] < 18,
        dummy_temp[["body_height"]] + (dummy_temp[["age_factor"]] * dummy_temp[["income_factor"]] * 5),
        dummy_temp[["body_height"]]))

    # Bring random variety into body weights
    dummy_temp[["body_weight"]] <- as.integer(data.table::fifelse(
        dummy_temp[["year"]] > start_year & dummy_temp[["age"]] < 18,
        dummy_temp[["body_weight"]] + (dummy_temp[["age_factor"]] * dummy_temp[["income_factor"]] * 5),
        pmax(5, dummy_temp[["body_weight"]] + sample(-10:10, orig_obs, replace = TRUE))))

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Generate income classes")
    #-------------------------------------------------------------------------#
    print_step("MINOR", "income classes")

    dummy_temp[["income_class"]] <- as.character(cut(dummy_temp[["income"]],
                                                     breaks = c(0, 500, 1000, 1500, 2000, 2500, 3000, 3500, 4000, 4500, 5000, Inf),
                                                     labels = c("01.         under  500", "02.  500 to under 1000", "03. 1000 to under 1500", "04. 1500 to under 2000",
                                                                "05. 2000 to under 2500", "06. 2500 to under 3000", "07. 3000 to under 3500", "08. 3500 to under 4000",
                                                                "09. 4000 to under 4500", "10. 4500 to under 5000", "11. 5000 and more"), right = FALSE))

    dummy_temp[which(dummy_temp[["income"]] == 0), "income_class"] <- "00. no income"

    dummy_temp[["income_class"]] <- data.table::fifelse(is.na(dummy_temp[["income"]]), NA, dummy_temp[["income_class"]])

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Finish")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Finish")

    dummy_temp[["first_person"]] <- as.integer(dummy_temp[["first_person"]])
    dummy_temp[["sex"]]          <- as.integer(dummy_temp[["sex"]])
    dummy_temp[["age"]]          <- as.integer(dummy_temp[["age"]])

    dummy_temp <- dummy_temp |> keep("year", "state", "NUTS2", "NUTS3", "household_id", "person_id",
                                     "number_of_persons", "first_person", "age",
                                     "sex", "education", "body_height", "body_weight",
                                     "income_class", "income", "expenses", "balance", "probability",
                                     "weight", "weight_per_year", order_vars = TRUE)

    print_closing(5)

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_end()
    monitor_df |> monitor_plot(draw_plot = monitor)
    #-------------------------------------------------------------------------#

    dummy_temp
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

# Create dummy data with 1 observations
test_df <- dummy_data(1)
expect_equal(nrow(test_df), 1, info = "Create dummy data with 1 observations")


# Create dummy data with 10 observations
test_df <- dummy_data(10)
expect_equal(nrow(test_df), 10, info = "Create dummy data with 10 observations")


# Create dummy data with 100 observations
test_df <- dummy_data(100)
expect_equal(nrow(test_df), 100, info = "Create dummy data with 100 observations")


# Create dummy data with 100 observations
test_df <- dummy_data(1000)
expect_equal(nrow(test_df), 1000, info = "Create dummy data with 100 observations")


# Dummy data first_person has always person_id 1
first <- test_df |> collapse::fsubset(first_person == 1)
expect_equal(collapse::funique(first[["person_id"]]), 1, info = "Dummy data first_person has always person_id 1")


set_no_print()

# Print content report
dummy_df    <- dummy_data()
result_list <- content_report(dummy_df)

expect_equal(names(result_list), c("global", "variables"), info = "Print content report")


# Print tinytest custom report
result_file <- system.file("extdata", "qol_tinytest_results.fst", package = "qol")
results     <- load_file(dirname(result_file), basename(result_file))

output <- results |> report_test_results()

expect_equal(results, output, info = "Print content report")

#' Different Facets of Retain
#'
#' @name retain
#'
#' @description
#' These retain functions all have one thing in common: transferring a value from one case
#' to the next. What they make out of this functionality can be quiet different. Therefor
#' there is a function for each different use case.
#'
#' @param data_frame The data frame in which to compute retained variables.
#' @param by By group in which to compute the retained variable.
#' @param values One or multiple variables of which a value should be retained.
#'
#' @details
#' The functions listed here are based on the 'SAS' function retain. On a very basic level retain
#' can do two things, depending on the position in the 'SAS' code: It can either sort variables
#' column wise or it can - since it works row wise - remember a value from one row to the next.
#' The functions here concentrate on the second part.
#'
#' Remembering a value from a previous observation offers multiple use cases. E.g. always adding
#' +1 to the previous case creates a running number. Or if an observation knows the value of the
#' previous one, it can check whether it is of the same value or another, e.g. to mark first or last
#' cases within a group.
#'
#' In it's simplest form it can remember a value from the first observation and transfer it to all
#' other observations.
#'
#' All of these functions work on the whole data frame as well as on groups, e.g. to transfer
#' a value from the first person in a household to all other persons of the same household.
#'
#' @keywords internal
NULL


#' Compute Running Numbers
#'
#' @description
#' [running_number()] computes running numbers in a data frame. Without specifying
#' a by variable results in the row number. With by variable computes the running
#' number within each group of expressions.
#'
#' @return
#' [running_number()]: Returns a vector containing a running number.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Get row numbers
#' my_data[["run_nr"]] <- my_data |> running_number()
#'
#' # Running number per variable expression
#' my_data[["run_nr_by"]] <- my_data |> running_number(by = year)
#'
#' # Running number within a group of variables which need to be sorted beforehand
#' my_data[["run_nr_by_sort"]] <- my_data |>
#'     running_number(by   = c(sex, education, income_class),
#'                    sort = TRUE)
#'
#' # Running number within a group of variables which need to be sorted beforehand
#' my_data[["run_group"]] <- my_data |>
#'     running_number(by       = c(sex, education, income_class),
#'                    sort     = TRUE,
#'                    group_nr = TRUE)
#'
#' @rdname retain
#'
#' @export
running_number <- function(data_frame,
                           by       = NULL,
                           sort     = FALSE,
                           group_nr = FALSE){
    # Measure the time
    print_start_message(suppress = TRUE)

    ###########################################################################
    # Error handling
    ###########################################################################

    # Convert to character vectors
    by <- get_origin_as_char(by, substitute(by))

    # Make sure that the variables provided are part of the data frame.
    by <- data_frame |> part_of_df(by)

    ###########################################################################
    # Retain
    ###########################################################################

    # If the user specified a vector of by variables, only take the last one, as this is
    # most likely the group in which the running number should be generated.
    if (length(by) > 1 && !sort){
        print_message("NOTE", c("Running number is generated in current data frame order. Only last variable",
								"inside provided <by> vector '[by]' will be used. Use the <sort> parameter, if",
								"you want to create a new order before generating the running number."), by = by)

        by <- by[length(by)]
    }

    # In case by variables should be sorted
    if (length(by) > 0 && sort){
        data_frame <- data_frame |>
            data.table::setorderv(cols    = by,
                                  order   = 1,
                                  na.last = TRUE)

        by <- by[length(by)]
    }

    # In case of a running number inside a by variable group
    if (length(by) == 1){
        # Generate a running number where every element of the group has the same
        # number. Meaning the whole group receives a running number.
        variable <- data.table::rleid(data_frame[[by]])

        # Generate a running number within the group
        if (!group_nr){
            variable <- sequence(tabulate(variable))
        }
    }
    # In case of no by variable
    else{
        variable <- seq_len(collapse::fnrow(data_frame))
    }

    print_closing()

    variable
}


#' @description
#' [mark_case()] sets a flag for the first or last case within the provided by group.
#'
#' @param first [mark_case()]: If TRUE marks the first case within a group, otherwise
#' the last case.
#'
#' @return
#' [mark_case()]: Returns a vector containing a the marking for first or last cases.
#'
#' @examples
#' # Mark first and last cases
#' my_data[["first"]] <- my_data |> mark_case(by = household_id)
#' my_data[["last"]]  <- my_data |> mark_case(by    = household_id,
#'                                            first = FALSE)
#'
#' @rdname retain
#'
#' @export
mark_case <- function(data_frame,
                      by       = NULL,
                      first    = TRUE){
    # Measure the time
    print_start_message(suppress = TRUE)

    ###########################################################################
    # Error handling
    ###########################################################################

    # Convert to character vectors
    by <- get_origin_as_char(by, substitute(by))

    # Make sure that the variables provided are part of the data frame.
    by <- data_frame |> part_of_df(by)

    # If the user specified a vector of by variables, only take the last one, as this is
    # most likely the group in which the first and last cases should be marked.
    if (length(by) > 1){
        print_message("NOTE", c("Cases are marked in current data frame order. Only last variable",
								"'[by]' inside provided <by> vector will be used."), by = by[length(by)])

        by <- by[length(by)]
    }

    ###########################################################################
    # Retain
    ###########################################################################

    # In case by variable is specified, mark first/last case per group in current order
    if (length(by) == 1){
        # The trick in marking cases is, to offset check the same vector by one.
        # A TRUE offset check means, that the value has changed. FALSE means value is the
        # same, so we are still in the same group.
        # Mark first cases by placing the missing TRUE value at the front, meaning: Shift down the whole
        # vector by one.
        if (first){
            variable <- c(TRUE, data_frame[[by]][-1] != data_frame[[by]][-collapse::fnrow(data_frame)])
        }
        # Mark last cases by placing the missing TRUE value at the back, meaning: Shift up the whole
        # vector by one.
        else{
            variable <- c(data_frame[[by]][-1] != data_frame[[by]][-collapse::fnrow(data_frame)], TRUE)
        }
    }
    # In case no by variable is specified, mark first/last case of whole data frame
    else{
        number_of_obersvations <- collapse::fnrow(data_frame)

        # Create vector of all FALSE values with the length of the provided data frame
        variable <- logical(number_of_obersvations)

        if (first){
            variable[1] <- TRUE
        }
        else{
            variable[number_of_obersvations] <- TRUE
        }
    }

    # Output as 1/0 instead of TRUE/FALSE
    variable <- as.integer(variable)

    print_closing()

    variable
}


#' @description
#' [retain_value()] retains the first value for all cases of the same group and saves
#' it into a new variable.
#'
#' @return
#' [retain_value()]: Returns a vector containing a retained value.
#'
#' @examples
#' # Retain first value without grouping
#' my_data[["first_weight"]] <- my_data |> retain_value(value = weight)
#'
#' # Retain first value inside a group for multiple variables
#' my_data[, c("household_weight", "household_icome")] <- my_data |>
#'     retain_value(values = c(weight, income),
#'                  by     = c(state, household_id))
#'
#' @rdname retain
#'
#' @export
retain_value <- function(data_frame,
                         values,
                         by = NULL){
    # Measure the time
    print_start_message(suppress = TRUE)

    ###########################################################################
    # Error handling
    ###########################################################################

    # Convert to character vectors
    by <- get_origin_as_char(by, substitute(by))

    # Make sure that the variables provided are part of the data frame.
    by <- data_frame |> part_of_df(by)

    # Generate pseudo var if by is NULL
    if (is.null(by) || length(by) == 0){
        data_frame[[".pseudo_by"]] <- 1
        by <- ".pseudo_by"
    }

    # Convert to character vectors
    values <- get_origin_as_char(values, substitute(values))

    # Make sure that the variables provided are part of the data frame.
    values <- data_frame |> part_of_df(values)

    # Abort if no value provided
    if (length(values) <= 1){
        if (length(values) == 0 || values == ""){
            print_message("ERROR", "Must provide <values> to retain. Retain will be aborted.")
            return(invisible(NA))
        }
    }

    ###########################################################################
    # Retain
    ###########################################################################

    group <- collapse::GRP(data_frame, by)

    # If there are as much new variable names provided, as there are values given, then rename them
    data_frame <- data_frame[values] |> collapse::ffirst(g = group, TRA = "fill")

    print_closing()

    if (length(values) == 1){
        data_frame[[1]]
    }
    else{
        as.list(data_frame)
    }
}


#' @description
#' [retain_stat()] retains the chosen statistic for all cases of the same group and saves
#' it into a new variable.
#'
#' @param statistics Available functions:
#' - "sum"       -> Weighted and unweighted sum
#' - "sum_wgt"   -> Sum of all weights
#' - "freq"      -> Unweighted frequency
#' - "freq_g0"   -> Unweighted frequency of all values greater than zero
#' - "pct_group" -> Weighted and unweighted percentages within the respective group
#' - "pct_total" -> Weighted and unweighted percentages compared to the grand total
#' - "mean"      -> Weighted and unweighted mean
#' - "median"    -> Weighted and unweighted median
#' - "mode"      -> Weighted and unweighted mode
#' - "min"       -> Minimum
#' - "max"       -> Maximum
#' - "sd"        -> Weighted and unweighted standard deviation
#' - "variance"  -> Weighted and unweighted standard variance
#' - "first"     -> First value
#' - "last"      -> Last value
#' - "pn"        -> Weighted and unweighted percentiles (any p1, p2, p3, ... possible)
#' Uses the SAS default way of calculating percentiles, which is Type 2 described in [quantile()].
#' - "missing"   -> Missings generated by the value variables
#'
#' @return
#' [retain_stat()]: Returns a vector containing a retained sum.
#'
#' @examples
#' # Retain sum without grouping
#' my_data[["total_sum"]] <- my_data |> retain_stat(values = weight)
#'
#' # Retain sum inside a group for multiple variables
#' my_data[, c("weight_sum", "income_sum")] <- my_data |>
#'     retain_stat(values = c(weight, income),
#'                 by     = c(year, state, household_id))
#'
#' # Retain different statistics
#' my_data[, c("income_freq", "income_max")] <- my_data |>
#'     retain_stat(values     = income,
#'                 by         = c(year, state, household_id),
#'                 statistics = c("freq", "max"))
#'
#' @rdname retain
#'
#' @export
retain_stat <- function(data_frame,
                        values,
                        by         = NULL,
                        statistics = "sum"){
    # Measure the time
    print_start_message(suppress = TRUE)

    ###########################################################################
    # Error handling
    ###########################################################################

    # Convert to character vectors
    by <- get_origin_as_char(by, substitute(by))

    # Make sure that the variables provided are part of the data frame.
    by <- data_frame |> part_of_df(by)

    # Generate pseudo var if by is NULL
    if (is.null(by) || length(by) == 0){
        data_frame[[".pseudo_by"]] <- 1
        by <- ".pseudo_by"
    }

    # Convert to character vectors
    values <- get_origin_as_char(values, substitute(values))

    # Make sure that the variables provided are part of the data frame.
    values <- data_frame |> part_of_df(values)

    # Abort if no value provided
    if (length(values) <= 1){
        if (length(values) == 0 || values == ""){
            print_message("ERROR", "Must provide a <values> to retain. Retain will be aborted.")
            return(invisible(NA))
        }
    }

    # Abort if non-numeric value provided
    if (!all(collapse::vtypes(data_frame[values]) %in% c("numeric", "integer", "double"))){
        print_message("ERROR", "<Values> must be numeric. Retain will be aborted.")
        return(invisible(NA))
    }

    ###########################################################################
    # Retain
    ###########################################################################

    # In case by variable is specified, compute sum per group
    data_frame <- suppressMessages(data_frame |>
           summarise_plus(class      = by,
                          values     = values,
                          statistics = statistics,
                          nesting    = "deepest",
                          merge_back = TRUE))

    print_closing()

    if (length(values) == 1 && length(statistics) == 1){
        data_frame[[collapse::fncol(data_frame)]]
    }
    else{
        as.list(data_frame[-(1:(collapse::fncol(data_frame) - (length(values) * length(statistics))))])
    }
}


#' @description
#' [retain_variables()] orders the provided variables to the front or back of the
#' data frame. If a variable is not part of the data frame it will be added with
#' all NA values at the desired position.
#'
#' @param ... [retain_variables()]: Put in single variable names or variable ranges
#' (var_name1:var_name10) which should be ordered to the front or back of the data frame.
#' It is also possible to provide none existent variable names which will then be added
#' to the data frame.
#' @param retain_variables [retain_variables()]: FALSE by default. If TRUE puts the
#' variables at the end of the data frame instead of the beginning.
#'
#' @return
#' [retain_variables()]: Return the data frame with reordered variables.
#'
#' @examples
#' # Retain columns inside data frame, which orders them to the front
#' my_data <- my_data |> retain_variables(age, sex, income)
#'
#' # Retain columns inside data frame, but order them to the end.
#' # Variable ranges can also be used.
#' my_data <- my_data |> retain_variables(age:income, order_last = TRUE)
#'
#' # Retain columns inside data frame and add new variables with all NA values
#' my_data <- my_data |> retain_variables(age, sex, income, status1:status5)
#'
#' # You can also use the colon as a placeholder for any text
#' start1   <- my_data |> retain_variables("s:")   # Variable names start with "s"
#' end1     <- my_data |> retain_variables(":id")  # Variable names end with "id"
#' contain1 <- my_data |> retain_variables(":on:") # Variable names which contain "on"
#'
#' @rdname retain
#'
#' @export
retain_variables <- function(data_frame,
                             ...,
							 order_last = FALSE){
    # Measure the time
    print_start_message(suppress = TRUE)

    # Translate ... into a list if possible
    retain_list <- tryCatch({
        # Force evaluation to see if it exists
        dots_to_char(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(retain_list)){
        print_message("ERROR", "Unknown object found. Retaining will be aborted.")
        return(invisible(data_frame))
    }

    # Check if there are any colons in the selection and deparse variables accordingly
    var_names <- names(data_frame)
    variables <- character(0)

    for (variable in retain_list){
        if (grepl(":", variable, fixed = TRUE)){
            parts <- strsplit(variable, ":", fixed = TRUE)[[1]]

            # If there is only one part the selection is "text:" and if the first element
            # is empty the selection was either ":text" or ":text:". In these cases a selection
            # of variables inside the data frame happens.
            if (length(parts) == 1 || parts[[1]] == ""){
                variables <- c(variables, data_frame |> deparse_colon(variable))
            }
            # If both variables are not part of the data frame, add the variables as new ones
            else if (!any(c(parts[1], parts[2]) %in% var_names)){
                data_frame <- data_frame |> add_variable_range(variable)
                variables  <- c(variables, data_frame |> vars_between(parts[1], parts[2]))
            }
            # Add variables in range to retain vector
            else{
                variables <- c(variables, data_frame |> vars_between(parts[1], parts[2]))
            }
        }
        # If element is just a single variable
        else{
            # If variable is not part of the data frame add it as empty variable
            if (!variable %in% var_names){
                new_columns <- collapse::qDF(matrix(NA_integer_,
                                                    nrow = collapse::fnrow(data_frame),
                                                    ncol = 1))

                names(new_columns) <- variable

                # Add new empty variables to data frame
                data_frame <- data_frame |> collapse::add_vars(new_columns)
            }

            # Add variables in range to retain vector
            variables <- c(variables, variable)
        }
    }

    # Only keep unique values. It can happen that a colon selection at the end of the
    # provided variable list catches variables another time which already have been
    # added to the variable vector.
    variables <- variables |> collapse::funique()

    # Reorder variables
    if (!order_last){
        ordered_df <- data.table::copy(data_frame) |>
            data.table::setcolorder(variables,
                                    before = 1)
    }
    else{
        ordered_df <- data.table::copy(data_frame) |>
            data.table::setcolorder(variables,
                                    after = collapse::fncol(data_frame))
    }

    print_closing()

    ordered_df
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

test_df <- data.frame(
    var_by   = c(1, 1, 2, 3, 3),
    var_num  = c(1, NA, NA, 2, NA),
    var_num2 = c(1, 2, 3, 4, 5),
    var_char = c("a", NA, NA, "b", NA),
    var_sum  = c(1, 1, 1, 1, 1))

test_df2 <- data.frame(
    var1 = c(1, 1, 2, 2, 2, 3, 3, 3),
    var2 = c(3, 3, 2, 2, 2, 1, 1, 1))

dummy_df <- dummy_data(10)


# Generate running number without by
test_df[["run_nr"]] <- test_df |> running_number()

expect_equal(test_df[["run_nr"]], c(1, 2, 3, 4, 5), info = "Generate running number without by")


# Generate running number with by
test_df[["run_nr"]] <- test_df |> running_number(by = var_by)

expect_equal(test_df[["run_nr"]], c(1, 2, 1, 1, 2), info = "Generate running number with by")


# Generate running number with by and sort
test_df2[["run_nr"]] <- test_df2 |> running_number(by = c(var2, var1), sort = TRUE)

expect_equal(test_df2[["run_nr"]], c(1, 2, 3, 1, 2, 3, 1, 2), info = "Generate running number with by and sort")


# Generate running number with by and group
test_df2[["run_nr"]] <- test_df2 |> running_number(by = c(var2, var1), sort = TRUE, group_nr = TRUE)

expect_equal(test_df2[["run_nr"]], c(1, 1, 1, 2, 2, 2, 3, 3), info = "Generate running number with by and group")


# Mark first and last cases without by
test_df[["first"]] <- test_df |> mark_case()
test_df[["last"]]  <- test_df |> mark_case(first = FALSE)

expect_equal(test_df[["first"]], c(1, 0, 0, 0, 0), info = "Mark first and last cases without by")
expect_equal(test_df[["last"]],  c(0, 0, 0, 0, 1), info = "Mark first and last cases without by")


# Mark first and last cases with by
test_df[["first"]] <- test_df |> mark_case(by = var_by)
test_df[["last"]]  <- test_df |> mark_case(by = var_by, first = FALSE)

expect_equal(test_df[["first"]], c(1, 0, 1, 1, 0), info = "Mark first and last cases with by")
expect_equal(test_df[["last"]],  c(0, 1, 1, 0, 1), info = "Mark first and last cases with by")


# Retain value without by
test_df[["retain_value"]] <- test_df |>
    retain_value(values = var_num)

expect_equal(test_df[["retain_value"]], c(1, 1, 1, 2, 2), info = "Retain value without by")


# Retain value with by
test_df[["retain_value"]] <- test_df |>
      retain_value(values = var_num, by = var_by)

expect_equal(test_df[["retain_value"]], c(1, 1, NA, 2, 2), info = "Retain value with by")


# Retain character value
test_df[["retain_value"]] <- test_df |>
      retain_value(values = var_char)

expect_equal(test_df[["retain_value"]], c("a", "a", "a", "b", "b"), info = "Retain character value")


# Retain multiple values
test_df[, c("var_num_first", "var_char_first")] <-
    test_df |>
      retain_value(values = c(var_num, var_char),
                   by    = c(var_sum, var_by))

expect_equal(test_df[["var_num_first"]],  c(1, 1, NA, 2, 2), info = "Retain multiple values")
expect_equal(test_df[["var_char_first"]], c("a", "a", NA, "b", "b"), info = "Retain multiple values")


# Retain sum without by
test_df[["retain_stat"]] <- test_df |>
      retain_stat(values = var_sum)

expect_equal(test_df[["retain_stat"]], c(5, 5, 5, 5, 5), info = "Retain sum without by")


# Retain sum with by
test_df[["retain_stat"]] <- test_df |>
      retain_stat(values = var_sum, by = var_by)

expect_equal(test_df[["retain_stat"]], c(2, 2, 1, 2, 2), info = "Retain sum with by")


# Retain multiple sums
test_df[, c("var_num_sum", "var_num2_sum")] <-
    test_df |>
      retain_stat(values = c(var_num, var_num2),
                  by    = c(var_sum, var_by))

expect_equal(test_df[["var_num_sum"]],  c(1, 1, NA, 2, 2), info = "Retain multiple sums")
expect_equal(test_df[["var_num2_sum"]], c(3, 3, 3, 9, 9), info = "Retain multiple sums")


# Retain columns in a data frame and order them to the front
retain_df <- dummy_df |>
      retain_variables(age, sex, income)

expect_equal(names(retain_df)[1:3], c("age", "sex", "income"), info = "Retain columns in a data frame and order them to the front")


# Retain columns in a data frame and order them to the end
retain_df <- dummy_df |>
      retain_variables(age, sex, income, order_last = TRUE)

expect_equal(utils::tail(names(retain_df), 3), c("age", "sex", "income"), info = "Retain columns in a data frame and order them to the end")


# Retain column ranges in a data frame
retain_df <- dummy_df |>
      retain_variables(age:education)

expect_equal(names(retain_df)[1:3], c("age", "sex", "education"), info = "Retain column ranges in a data frame")


# Add new single NA columns with retain
retain_df <- dummy_df |>
      retain_variables(status, new_var, test_var)

expect_equal(names(retain_df)[1:3], c("status", "new_var", "test_var"), info = "Add new single NA columns with retain")


# Add new NA column range with retain
retain_df <- dummy_df |>
      retain_variables(status1:status3)

expect_equal(names(retain_df)[1:3], c("status1", "status2", "status3"), info = "Add new NA column range with retain")


# Retain variables starting with letter
retain_df1 <- dummy_df |>
      retain_variables("s:")

retain_df2 <- dummy_df |>
      retain_variables("s:", order_last = TRUE)

expect_equal(names(retain_df1)[1:2], c("state", "sex"), info = "Retain variables starting with letter")
expect_equal(utils::tail(names(retain_df2), 2), c("state", "sex"), info = "Retain variables starting with letter")


# Retain variables ending with letter
retain_df1 <- dummy_df |>
      retain_variables(":id")

retain_df2 <- dummy_df |>
      retain_variables(":id", order_last = TRUE)

expect_equal(names(retain_df1)[1:2], c("household_id", "person_id"), info = "Retain variables ending with letter")
expect_equal(utils::tail(names(retain_df2), 2), c("household_id", "person_id"), info = "Retain variables ending with letter")


# Retain variables containing letter
retain_df1 <- dummy_df |>
      retain_variables(":on:")

retain_df2 <- dummy_df |>
      retain_variables(":on:", order_last = TRUE)

expect_equal(names(retain_df1)[1:3], c("person_id", "number_of_persons", "first_person"), info = "Retain variables containing letter")
expect_equal(utils::tail(names(retain_df2), 3), c("number_of_persons", "first_person", "education"), info = "Retain variables containing letter")


# Retain variables with all actions together doesn't break
retain_df <- dummy_df |>
      retain_variables(age:education, "s:", ":id", ":on:", status1:status3, income)

expect_equal(names(retain_df)[1:12], c("age", "sex", "education", "state", "household_id", "person_id",
                                       "number_of_persons", "first_person", "status1", "status2", "status3",
                                       "income"), info = "Retain variables with all actions together doesn't break")

###############################################################################
# Note checks
###############################################################################

# Generate running number with multiple by variables
dummy_df[["run_nr"]] <- dummy_df |> running_number(by = c(year, sex))

expect_message(print_stack_as_messages("NOTE"), "Running number is generated in current data frame order.", info = "Generate running number with multiple by variables")


# Mark first and last cases with multiple by variables
dummy_df[["first"]] <- dummy_df |> mark_case(by = c(sex, household_id))

expect_message(print_stack_as_messages("NOTE"), "Cases are marked in current data frame order.", info = "Mark first and last cases with multiple by variables")


dummy_df[["last"]] <- dummy_df |> mark_case(by = c(sex, household_id), first = FALSE)

expect_message(print_stack_as_messages("NOTE"), "Cases are marked in current data frame order.", info = "Mark first and last cases with multiple by variables")

expect_true(all(c("first", "last") %in% names(dummy_df)), info = "Mark first and last cases with multiple by variables")

###############################################################################
# Abort checks
###############################################################################

# Retain value without providing a value
dummy_df[["retain_value"]] <- dummy_df |> retain_value()

expect_error(print_stack_as_messages("ERROR"), "Must provide <values> to retain. Retain will be aborted.", info = "Retain value without providing a value")


# Retain sum without providing sum_of
dummy_df[["retain_stat"]] <- dummy_df |> retain_stat()

expect_error(print_stack_as_messages("ERROR"), "Must provide a <values> to retain. Retain will be aborted.", info = "Retain sum without providing sum_of")


# Add new NA column range with wrong pattern returns original data frame
retain_df <- (dummy_df |>
     retain_variables(status1:age3))

expect_error(print_stack_as_messages("ERROR"), "Variable range has to be provided in the form 'var_name1:var_name10'.", info = "Add new NA column range with wrong pattern returns original data frame")

expect_equal(dummy_df, retain_df, info = "Add new NA column range with wrong pattern returns original data frame")


set_no_print()

#' Build a Master Script From Folder
#'
#' @description
#' [build_master()] reads a given folder structure, which contains scripts, and builds
#' a master script as a markdown file.
#'
#' @param dir The folder structure which contains the scripts to build upon.
#' @param master_name The file name which should be written.
#' @param author Authors name to be put in the header.
#' @param with_structure Whether the folder structure as tree should be written
#' to the master script.
#' @param with_run_all Whether a section, which let's the user run all scripts,
#' should be written to the master script.
#' @param with_run_folder Whether a section, which let's the user run all scripts from a
#' specific folder, should be written to the master script.
#' @param with_monitor FALSE by default. If TRUE, outputs two charts to visualize the
#' time consumption of the individual scripts.
#'
#' @details
#' The function works with folder structures that look like this:
#'
#' root/
#'
#'      subfolder1/
#'
#'          script1.R
#'
#'          script2.R
#'
#'          ....R
#'
#'      subfolder2/
#'
#'          script3.R
#'
#'          script4.R
#'
#'          ....R
#'
#'      .../
#'
#'          ....R
#'
#' @return
#' Returns the script as character vector and saves it as markdown file.
#'
#' @examples
#' # Example export file paths
#' # NOTE: These tempfiles are only for the examples. In reality you just call the
#' # main function and put in your desired path and name directly.
#' temp_file <- tempfile(fileext = ".rstheme")
#' file_name <- basename(tools::file_path_sans_ext(temp_file))
#'
#' # Example master
#' build_master(dir         = dirname(temp_file),
#'              master_name = file_name)
#'
#' # Manual cleanup for example
#' unlink(temp_file)
#'
#' @export
build_master <- function(dir,
                         master_name     = "Master",
                         author          = "",
                         with_structure  = TRUE,
                         with_run_all    = TRUE,
                         with_run_folder = TRUE,
                         with_monitor    = FALSE){
    # Measure the time
    print_start_message()

    # Apply macros to dir
    dir <- macro(dir)

    # Check if folder exists
    if (!dir.exists(dir) || dirname(dir) == "."){
        print_message("ERROR", "Directory '[dir]' does not exist.", dir = dir)
        return(invisible(NULL))
    }

    path <- ifelse(grepl("/$", dir), dir, paste0(dir, "/"))

    # Get folders in provided directory
    folders <- list.dirs(dir, recursive = TRUE, full.names = TRUE)

    # Remove root folder. Otherwise when rebuilding the master script, the root
    # folder would be captured as folder with files in it.
    if (length(folders) > 1){
        folders <- folders[-1]
    }

    # Get all .R scripts inside the folders
    scripts <- lapply(folders, function(folder){
        suppressMessages(libname(folder, get_files = TRUE, extensions = "R"))
    })
    names(scripts) <- folders
    scripts        <- Filter(Negate(is.null), scripts)

    # Setup header
    lines <- c(
        "################################################################################",
        paste0("# ", master_name),
        "#",
        paste0("# Author: ", author),
        "#",
        paste0("# Date: ", format(Sys.Date(), "%d.%m.%Y")),
        "################################################################################",
        "",
        "```{r load_package, echo = TRUE}",
        "library(qol)",
        "",
        "run_scripts <- function(scripts){",
        "    print_start_message()",
        '    monitor_df <- NULL |> monitor_start("Script start", "Script start")',
        "",
        "    for (file in scripts){",
        '        monitor_df <- monitor_df |> monitor_next(basename(file), basename(file))',
        '        print_step("MAJOR", file)',
        "",
        "        source(file, local = FALSE)",
        "",
        "        monitor_df <- monitor_df |> monitor_end()",
        "    }",
        "",
        "    print_closing()",
        "",
        paste0("    monitor_df |> monitor_plot(draw_plot = ", with_monitor, ")"),
        "}",
        "```",
        "")

    # Get the function call itself with all parameters and convert to character
    # so that it can be inserted in the file as rebuilt option.
    call_expr <- match.call()
    call_text <- paste(deparse(call_expr, width.cutoff = 500), collapse = "\n")

    # Rebuilt master section
    lines <- c(lines,
        "################################################################################",
        "# Rebuilt Master",
        "################################################################################",
        "",
        "```{r rebuilt_master, echo = TRUE}",
        paste0('master_file <- "', path, master_name, '.Rmd"'),
        "",
        "if (file.exists(master_file)){",
        "    file.remove(master_file)",
        "}",
        "",
        call_text,
        "",
        "rm(master_file)",
        "```",
        "")

    # Generate tree view folder structure
    if (with_structure){
        print_step("MAJOR", "Write folder structure")

        lines <- c(lines, print_folder_structure(scripts), "")
    }

    # Run all scripts in all folders
    if (with_run_all){
        print_step("MAJOR", "Write all scripts execution")

        # Get blanks for an even padding of the code
        sub_dirs        <- collapse::funique(sub(dir, "", names(scripts)))
        max_path_length <- collapse::fmax(nchar(sub_dirs))
        padding         <- substr(paste0(sub_dirs, strrep(" ", max_path_length)), nchar(sub_dirs) + 1, max_path_length)

        # Put together script block
        run_all_folders <- c(
            "",
            "################################################################################",
            "# Run All Scripts in All Folders",
            "################################################################################",
            "```{r run_all_scripts, echo = TRUE}",
            paste0('base_folder <- "', dir, '"'),
            paste0('scripts     <- c(', paste(paste0('libname(paste0(base_folder,"', sub_dirs, '"), ', padding, 'get_files = TRUE, extensions = "R")'), collapse = ',\n                 '), ')'),
            "",
            "run_scripts(scripts)",
            "",
            "rm(base_folder, scripts)",
            "```")

        lines <- c(lines, run_all_folders)
    }

    print_step("MAJOR", "Write script execution")

    # Run folders and files separate
    for (folder in names(scripts)){
        # Run all scripts in current folder
        folder_name <- gsub("[^a-zA-Z0-9_]", "_", paste0("run_", basename(folder)))
        all_scripts_in_folder <- unlist(scripts[[folder]])

        if (with_run_folder){
            print_step("MINOR", "folder: [folder]", folder = folder)

            # Put together script block
            lines <- c(lines, c("\n#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++",
                                paste0("#     Run All Scripts in Folder: ", basename(folder)),
                                "#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"),
                              paste0("```{r ", folder_name, ", echo = TRUE}"),
                              paste0('      scripts <- libname("', folder, '", get_files = TRUE, extensions = "R")'),
                                "",
                                "      run_scripts(scripts)",
                                "",
                                "      rm(scripts)",
                                "```")
        }

        # Add each script as its own chunk
        files <- scripts[[folder]]

        for (file in files){
            print_step("MINOR", "    file: [file]", file = file)

            file_name <- gsub("[^a-zA-Z0-9_]", "_", paste0("run_", basename(file)))

            # Put together script block
            lines <- c(lines, c("#-------------------------------------------------------------------------------#",
                                paste0("#         Run Script: ", basename(file)),
                                "#-------------------------------------------------------------------------------#"),
                              paste0("```{r ", file_name, ", echo = TRUE}"),
                              paste0('          run_scripts("', file, '")'),
                              "```")
        }
    }

    print_step("MAJOR", "Putting together master file")

    # Write master file
    writeLines(lines, con = paste0(path, master_name, ".Rmd"))

    print_closing(5)

    invisible(lines)
}


#' Convert List of Files Into a Hierarchical Text Representation
#'
#' @description
#' Writes a text based folder structure in a tree view.
#'
#' @param file_list A list, which contains the folder structure and filenames.
#'
#' @return
#' A formatted character vector.
#'
#' @noRd
print_folder_structure <- function(file_list){
    # Extract root folder
    root <- paste0(dirname(names(file_list)[1]), "/")

    # Root is always on top
    output <- c("################################################################################",
                "# Folder structure",
                "################################################################################",
                "",
                "# Root",
                "```{r open_root, echo = TRUE}",
                paste0('utils::browseURL("', root, '")'),
                "```")

    # Loop trough all list entries to build folder tree
    for (folder in names(file_list)){
        # Get folder and corresponding files first
        folder_path <- normalizePath(folder, winslash = "/")
        subfolder   <- basename(folder)
        files       <- file_list[[folder]]

        # Open folder section
        output <- c(output,
                    paste0("##......", subfolder),
                    paste0('```{r open_', subfolder, ', echo = TRUE}'),
                    paste0('        utils::browseURL("', folder_path, '")'),
                    "```")

        # Open files block
        for (file in files){
            file_path <- normalizePath(file, winslash = "/")

            output <- c(output,
                        paste0("###.............", basename(file)),
                        paste0('```{r open_', basename(file), ', echo = TRUE}'),
                        paste0('                file.edit("', file_path,'")'),
                        "```")
        }
    }

    output
}

#' Replace Patterns Inside Variable Names
#'
#' @description
#' Replace a certain pattern inside a variable name with a new one. This can be
#' used if there are multiple different variable names which have a pattern in
#' common (e.g. all end in "_sum" but start different), so that there don't have
#' to be multiple rename variable calls.
#'
#' @param data_frame The data frame in which there are variables to be renamed.
#' @param old_pattern The pattern which should be replaced in the variable names.
#' @param new_pattern The pattern which should be set in place for the old one.
#'
#' @return
#' Returns a data frame with renamed variables.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Summarise data
#' all_nested <- my_data |>
#'     summarise_plus(class      = c(year, sex),
#'                    values     = c(weight, income),
#'                    statistics = c("sum", "pct_group", "pct_total", "sum_wgt", "freq"),
#'                    weight     = weight,
#'                    nesting    = "deepest",
#'                    na.rm      = TRUE)
#'
#' # Rename variables by repacing patterns
#' new_names <- all_nested |>
#'     rename_pattern("pct", "percent") |>
#'     rename_pattern("_sum", "")
#'
#' @export
rename_pattern <- function(data_frame, old_pattern, new_pattern){
    if (length(old_pattern) > 1 || length(new_pattern) > 1){
        print_message("ERROR", "Only single pattern allowed. Rename pattern will be aborted.")
        return(data_frame)
    }

    # Replace old_pattern with new_pattern in all column names
    new_names <- gsub(old_pattern, new_pattern, names(data_frame))
    names(data_frame) <- new_names

    data_frame
}


#' Replace Statistic From Variable Names
#'
#' @description
#' Remove the statistic name from variable names, so that they get back their old
#' names without extension.
#'
#' @param data_frame The data frame in which there are variables to be renamed.
#' @param statistics Statistic extensions that should be removed from the variable names.
#'
#' @return
#' Returns a data frame with renamed variables.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Summarise data
#' all_nested <- my_data |>
#'     summarise_plus(class      = c(year, sex),
#'                    values     = c(weight, income),
#'                    statistics = c("sum", "pct_group", "pct_total", "sum_wgt", "freq"),
#'                    weight     = weight,
#'                    nesting    = "deepest",
#'                    na.rm      = TRUE)
#'
#' # Remove statistic extension
#' new_names <- all_nested |> remove_stat_extension("sum")
#'
#' @export
remove_stat_extension <- function(data_frame, statistics){
    var_names <- names(data_frame)

    # Remove statistic extensions
    for (stat in statistics){
        new_names <- sub(paste0("_", stat, "$"), "", var_names)
    }

    # Check if unique new names still have the same length as the original names.
    # Only if this is true the new names can be applied.
    if (length(var_names) == length(collapse::funique(new_names))){
        names(data_frame) <- new_names
    }
    # If there are duplicate names abort
    else{
        print_message("WARNING", "New variable names are not unique. Statistic extensions won't be removed.")
    }

    data_frame
}


#' Add Extensions to Variable Names
#'
#' @description
#' Renames variables in a data frame by adding the desired extensions to the original names.
#' This can be useful if you want to use pre summarised data with [any_table()], which needs
#' the value variables to have the statistic extensions.
#'
#' @param data_frame The data frame in which variables should gain extensions to their name.
#' @param from The position of the variable inside the data frame at which to start the renaming.
#' @param extensions The extensions to add.
#' @param reuse "none" by default, meaning only the provided extensions will be set. E.g. if
#' there are two extensions provided, two variables will be renamed. If "last", the last provided
#' extension will be used for every following variable until the end of the data frame. If "repeat",
#' the provided extensions will be repeated from the first one for every following variable until
#' the end of the data frame.
#'
#' @return
#' Returns a data frame with extended variable names.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(10)
#'
#' # Add extensions to variable names
#' new_names1 <- my_data |> add_extension(5, c("sum", "pct"))
#' new_names2 <- my_data |> add_extension(5, c("sum", "pct"), reuse = "last")
#' new_names3 <- my_data |> add_extension(5, c("sum", "pct"), reuse = "alternate")
#'
#' @export
add_extension <- function(data_frame,
                          from,
                          extensions,
                          reuse = "none"){
    if (!is.numeric(from)){
        print_message("ERROR", "From needs to be numeric. Adding extensions will be aborted.")
        return(data_frame)
    }

    if (!is.character(extensions)){
        print_message("ERROR", "Extensions need to be characters. Adding extensions will be aborted.")
        return(data_frame)
    }

    if (from > collapse::fncol(data_frame)){
        print_message("ERROR", "From is greater than number of columns in data frame. Adding extensions will be aborted.")
        return(data_frame)
    }

    if (!reuse %in% c("none", "last", "repeat")){
        print_message("WARNING", "Reuse must be one of 'none', 'last', 'repeat'. 'none' will be used.")
        reuse <- "none"
    }

    var_names <- names(data_frame)

    # Columns to modify
    target_columns <- from:length(var_names)
    n_target       <- length(target_columns)
    n_extensions   <- length(extensions)

    # Generally shorten extensions, if there are more than columns are left in the data frame
    extensions <- extensions[seq_len(min(n_extensions, n_target))]

    # Create the extended names
    if (reuse == "last" && n_target - n_extensions > 0){
        # Repeat the last extension for the remaining columns
        extensions <- c(extensions, rep(extensions[n_extensions], n_target - n_extensions))
    }
    else if(reuse == "none"){
        # Set new last column depending on number of extensions
        last_column    <- from + (length(extensions) - 1)
        target_columns <- from:last_column
    }

    # Apply extensions to variable names
    names(data_frame)[target_columns] <- paste0(var_names[target_columns], "_", extensions)

    data_frame
}


#' Replace Patterns While Protecting Exceptions
#'
#' @description
#' Replaces a provided pattern with another, while protecting exceptions. Exceptions can
#' contain the given pattern, but won't be changed during replacement.
#'
#' @param vector A vector containing the texts, where a pattern should be replaced.
#' @param pattern The pattern that should be replaced.
#' @param replacement The new pattern, which replaces the old one.
#' @param exceptions A character vector containing exceptions, which should not be altered.
#'
#' @return
#' Returns a vector with replaced pattern.
#'
#' @examples
#' # Vector, where underscores should be replaced
#' underscores <- c("my_variable", "var_with_underscores", "var_sum", "var_pct_total")
#'
#' # Extensions, where underscores shouldn't be replaced
#' extensions <- c("_sum", "_pct_group", "_pct_total", "_pct_value", "_pct", "_freq_g0",
#'                 "_freq", "_mean", "_median", "_mode", "_min", "_max", "_first",
#'                 "_last", "_p1", "_p2", "_p3", "_p4", "_p5", "_p6", "_p7", "_p8", "_p9",
#'                 "sum_wgt", "_sd", "_variance", "_missing")
#'
#' # Replace
#' new_vector <- underscores |> replace_except("_", ".", extensions)
#'
#' @export
replace_except <- function(vector,
                           pattern,
                           replacement,
                           exceptions = NULL){
    # If there are no exceptions just do a normal replace
    if (is.null(exceptions) || length(exceptions) == 0){
        return(gsub(pattern, replacement, vector))
    }

    # Replace pattern first in exceptions globally
    placeholder      <- "&%!"
    except_protected <- gsub(pattern, placeholder, exceptions, fixed = TRUE)

    # Protect exceptions in original vector
    for (i in seq_along(exceptions)){
        vector <- gsub(exceptions[i], except_protected[i], vector, fixed = TRUE)
    }

    # Replace pattern safely
    vector <- gsub(pattern, replacement, vector)

    # Reestablish protected pattern
    gsub(placeholder, pattern, vector, fixed = TRUE)
}


#' Rename One Or More Variables
#'
#' @description
#' Can rename one or more existing variable names into the corresponding new variable
#' names in one go.
#'
#' @param data_frame The data frame which contains the variable names to be renamed.
#' @param ... Pass in variables to be renamed in the form: "old_var" = "new_var".
#'
#' @return
#' Returns a data_frame with renamed variables.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(10)
#'
#' # Rename multiple variables at once
#' new_names_df <- my_data |> rename_multi(sex   = var1,
#'                                         age   = var2,
#'                                         state = var3)
#'
#' # Also works with variable names in quotation marks
#' new_names_df <- my_data |> rename_multi("sex"   = "var1",
#'                                         "age"   = "var2",
#'                                         "state" = "var3")
#'
#' @export
rename_multi <- function(data_frame, ...){
    # Measure the time
    print_start_message(suppress = TRUE)

    # Translate ... into a list if possible
    rename_list <- tryCatch({
        # Force evaluation to see if it exists
        list(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    # If capturing the ellipses failed, the variable names are probably passed
    # without quotation marks. If this is the case, capture them on another way.
    if (is.null(rename_list)){
        rename_list <- substitute(list(...))[-1]
    }

    # Get old and new names in separate vectors to rename them in one go
    old_names <- names(rename_list)

    # Depending on how the variable names were passed, thene wvariable names have
    # to be captured on a different way.
    if (is.list(rename_list)){
        new_names <- unlist(rename_list, use.names = FALSE)
    }
    else{
        new_names <- vapply(rename_list, deparse, character(1))
    }

    # Make sure that the variables provided are part of the data frame.
    old_names <- data_frame |> part_of_df(old_names, check_only = TRUE)

    if (is.list(old_names)){
        print_message("ERROR", c("The provided <old name> '[old]' is not part of",
								 "the data frame. Renaming will be aborted."), old = old_names[[1]])
        return(data_frame)
    }

    # If any of the new variable names is already part of the data frame abort
    invalid_new_names <- new_names[new_names %in% names(data_frame)]

    # If not all old names are part of the data frame abort
    if (length(invalid_new_names) > 0){
        print_message("ERROR", c("The provided <new name> '[new]' is already part of",
								 "the data frame. Renaming will be aborted."), new = invalid_new_names)
        return(data_frame)
    }

    # Rename all variables in one go
    data_frame <- data_frame |> collapse::frename(stats::setNames(old_names, new_names))

    print_closing()

    data_frame
}


#' Set First Data Frame Row As Variable Names
#'
#' @description
#' Sets the first row of a data frame as variable names and deletes it. In case
#' of NA, numeric values or empty characters in the first row, the old names are kept.
#'
#' @param data_frame A data frame for which to set new variable names.
#'
#' @return
#' Returns a data frame with renamed variables.
#'
#' @examples
#' # Example data frame
#' my_data <- data.frame(
#'               var1 = c("id", 1, 2, 3),
#'               var2 = c(NA, "a", "b", "c"),
#'               var3 = c("value", 1, 2, 3),
#'               var4 = c("", "a", "b", "c"),
#'               var5 = c(1, 2, 3, 4))
#'
#' my_data <- my_data |> first_row_as_names()
#'
#' @export
first_row_as_names <- function(data_frame){
    # Extract first row and current names
    new_names <- as.character(data_frame[1, ])
    old_names <- names(data_frame)

    # Set up condition on when to keep the old names
    keep_old <- is.na(new_names) |
                new_names == "" |
                !is.na(suppressWarnings(as.numeric(new_names)))

    # Rename conditionally
    names(data_frame) <- data.table::fifelse(keep_old, old_names, new_names)

    # Delete first row and return
    data_frame[-1, , drop = FALSE]
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

dummy_df <- dummy_data(10)


# Add extensions to limited variables
new_names_df <- dummy_df |> add_extension(9, c("sum", "pct"))

expect_true(all(c("age_sum", "sex_pct") %in% names(new_names_df)), info = "Add extensions to limited variables")


# Add extensions to all remaining variables
new_names_df <- dummy_df |> add_extension(5, c("sum", "pct"), reuse = "last")

expect_true("weight_pct" %in% names(new_names_df), info = "Add extensions to all remaining variables")


# Renaming multiple variables
new_names_df <- dummy_df |> rename_multi("sex" = "var1", "age" = "var2")

expect_true(all(c("var1", "var2") %in% names(new_names_df)), info = "Renaming multiple variables")
expect_true(!all(c("sex", "age")  %in% names(new_names_df)), info = "Renaming multiple variables")


# Renaming multiple variables without quotation marks
new_names_df <- dummy_df |> rename_multi(sex = var1, age = var2)

expect_true(all(c("var1", "var2") %in% names(new_names_df)), info = "Renaming multiple variables without quotation marks")
expect_true(!all(c("sex", "age")  %in% names(new_names_df)), info = "Renaming multiple variables without quotation marks")


# Renaming multiple variables using vectors
old_names <- c("sex",  "age")
new_names <- c("var1", "var2")

new_names_df <- dummy_df |> rename_multi(old_names = new_names)

expect_true(all(c("var1", "var2") %in% names(new_names_df)), info = "Renaming multiple variables without quotation marks")
expect_true(!all(c("sex", "age")  %in% names(new_names_df)), info = "Renaming multiple variables without quotation marks")


# Renaming based on first row in data frame
test_df <- data.frame(
              var1 = c("id", 1, 2, 3),
              var2 = c(NA, "a", "b", "c"),
              var3 = c("value", 1, 2, 3),
              var4 = c("", "a", "b", "c"),
              var5 = c(1, 2, 3, 4))

result_df <- test_df |> first_row_as_names()

expect_equal(names(result_df), c("id", "var2", "value", "var4", "var5"), info = "Renaming based on first row in data frame")
expect_equal(nrow(test_df) - 1, nrow(result_df), info = "Renaming based on first row in data frame")

###############################################################################
# Abort checks
###############################################################################

# Renaming aborts if old variable name not found in data frame
new_names_df <- dummy_df |> rename_multi("var1" = "var2")

expect_error(print_stack_as_messages("ERROR"), "The provided <old name> '", info = "Renaming aborts if old variable name not found in data frame")


# Renaming aborts if new variable name is already data frame
new_names_df <- dummy_df |> rename_multi("sex" = "age")

expect_error(print_stack_as_messages("ERROR"), "The provided <new name> '", info = "Renaming aborts if new variable name is already data frame")


set_no_print()

###############################################################################
# Notes, warnings and errors
###############################################################################

#' Print Styled Messages
#'
#' @name messages
#'
#' @description
#' Printing styled messages for different occasions. There are notifications,
#' warnings, errors, function call headlines, progress and function closing
#' messages. Or just a neutral one.
#'
#' @param type The message type, so that the function knows which symbol and coloring
#' to use.
#' @param text The message text to display.
#' @param ... Additional information to display like variable names. To use
#' these write &#91;NAME YOU PUT IN&#93; in the text.
#' @param always_print FALSE by default. If TRUE, prints headlines even in deeper
#' nested situations.
#' @param utf8 Whether to display complex characters or just plain text.
#' @param no_color Whether to display color codes or suppress them.
#'
#' @details
#' The message types in which you can enter custom texts, are capable of using
#' different styling operators. These are:
#'
#' - Insert list of elements: &#91;vector_name&#93;
#' - Adding conditional words, if list of elements has more than one element: &#91;?word&#93;
#' - Adding conditional singular/plural, depending on list of element length: &#91;?singular/plural&#93;
#' - Bold, italic and underline: &#91;b&#93;some text&#91;/b&#93;, &#91;i&#93;some text&#91;/i&#93;, &#91;u&#93;some text&#91;/u&#93;
#' - Coloring parts of the message: &#91;#FF00FF some text&#93;
#'
#' @seealso
#' Also have a look at the small helpers: [get_message_stack()], [set_no_print()],
#' [set_no_color()], [print_stack_as_messages()], [convert_square_brackets()]
#'
#' @return
#' Return text without styling or total running time.
#'
#' @examples
#' # Example messages
#' print_message("NOTE", c("Just a quick note that you can also insert e.g.[? a / ]variable",
#'                         "name[?s] like this: [listing].",
#'                         "Depending on the number of variables you can also alter the text."),
#'              listing = c("age", "state", "NUTS3"))
#'
#' print_message("WARNING", "Just a quick [#FF00FF colored warning]!")
#'
#' print_message("ERROR", "Or a [b]bold[/b], [i]italic[/i] and [u]underlined[/u] error.")
#'
#' print_message("NEUTRAL", c("You can also just output [u]plain text[/u] if you like and use",
#'                            "[#FFFF00 [b]all the different[/b] [i]formatting options.[/i]]"))
#'
#' @rdname messages
#'
#' @export
print_message <- function(type,
                          text,
                          ...,
                          always_print = FALSE,
                          utf8         = .qol_messages[["format"]][["utf8"]],
                          no_color     = .qol_messages[["no_color"]]){
    suppressed <- FALSE

    # Identify which function called the message
    caller <- sys.call(-1)

    if (is.call(caller) && is.name(caller[[1]])){
        # Get calling function and code line the message was called from
        caller <- ifelse(is.null(caller), "global", as.character(caller[[1]]))
    }
    else{
        caller <- "do.call"
    }

    if (!is.list(type) && !tolower(type) %in% c("note", "warning", "error", "neutral")){
        # Check if message type is a custom one and retrieve from global list
        if (!toupper(type) %in% names(.qol_messages[["custom_types"]])){
            type <- "neutral"
        }
    }

    depth     <- sys.nframe()
    sys_calls <- get_condensed_sys_calls()

    # Whenever a code block is executed, a unique token is generated. Based on this
    # token it can be decided, whether we are starting a new message stack (in case we
    # have a new token) or we keep the current stack (in case of multiple functions
    # running as a block)
    current_session_token   <- format(sys.frame(1))
    current_execution_token <- get_execution_token(depth, sys_calls)

    # In case it is the first starting message in the current run or if the stack
    # has a not suppressed end not. Meaning a function or pipe operation has finished.
    if (!identical(current_session_token, .qol_messages[["last_session"]]) ||
        !identical(current_execution_token, .qol_messages[["last_execution"]])){

        reset_messages()
    }

    # Add message to global stack
    if (is_stack() && !is_full_stack()){
        entry_nr <- length(.qol_messages[["stack"]]) + 1
    }
    # If no starting point is set, the message is called inside a function where no
    # starting message was called. In this case only capture the last message and reset
    # the message stack. Otherwise the stack would just grow and never reset.
    else{
        if (length(.qol_messages[["stack"]]) > 1){
            .qol_messages[["stack"]]        <- list()
            .qol_messages[["last_message"]] <- NULL
            .qol_messages[["last_type"]]    <- NULL
        }

        entry_nr <- 1
    }

    # Don't print message, if messages are suppressed
    if (!always_print && is_message_suppressed(sys_calls)){
        text       <- paste(text, collapse = " ")
        suppressed <- TRUE
    }
    # If messages are not suppressed
    else{
        last_message <- get_last_active_message()

        # If last message is a timed message, print a new line first, otherwise the
        # function would print on the same line.
        if (is_time_stamp_allowed(last_message) && !is_no_print_active()){
            cat("\n")
        }

        text <- print_to_console(type = type, text = text, ..., utf8 = utf8, no_color = no_color)
    }

    # Catch type for message stack on custom messages
    if (is.list(type)){
        type <- type[["type"]]
    }

    # Add message to global stack
    .qol_messages[["stack"]][[entry_nr]] <- list(type       = toupper(type),
                                                 text       = text,
                                                 suppressed = suppressed,
                                                 new_line   = TRUE,
                                                 print_time = FALSE,
                                                 in_place   = FALSE,
                                                 depth      = depth,
                                                 caller     = caller,
                                                 call_stack = sys_calls,
                                                 time       = Sys.time())

    invisible(text)
}


#' @description
#' This is the core printing function which formats the text and prints it to the
#' console.
#'
#' @param new_line TRUE by default. Whether to directly add a new line to the console
#' or not.
#'
#' @noRd
print_to_console <- function(type,
                             text,
                             ...,
                             utf8     = .qol_messages[["format"]][["utf8"]],
                             no_color = .qol_messages[["no_color"]],
                             new_line = TRUE){
    # Check if message type is a custom one and retrieve from global list
    if (!is.list(type) && toupper(type) %in% names(.qol_messages[["custom_types"]])){
        type <- .qol_messages[["custom_types"]][[toupper(type)]]
    }

    # If normal pre defined message
    if (!is.list(type)){
        .qol_messages[["last_type"]] <- toupper(type)

        type <- tolower(type)

        if (!type %in% c("note", "warning", "error", "neutral", "major", "minor", "grey")){
            type <- "neutral"
        }

        # Output message
        if (utf8){
            # First format all line breaks so that the next text lines are indented
            seperator <- paste0("\n", .qol_messages[["format"]][[paste0(type, "_indent_ansi")]])
            text      <- paste(text, collapse = seperator)

            # Insert special grey message format
            if (type == "grey"){
                text <- paste0("[i][", .qol_messages[["format"]][["time_color"]], " ", text, "][/i]")
            }

            # Insert formatting
            text_orig <- convert_square_brackets(text, ...)
            text      <- paste(.qol_messages[["format"]][[paste0(type, "_ansi")]], text_orig, sep = "")
        }
        else{
            # First format all line breaks so that the next text lines are indented
            seperator <- paste0("\n", .qol_messages[["format"]][[paste0(type, "_indent_pt")]])
            text      <- paste(text, collapse = seperator)

            # Insert special grey message format
            if (type == "grey"){
                text <- paste0("[i][", .qol_messages[["format"]][["time_color"]], " ", text, "][/i]")
            }

            # Insert formatting
            text_orig <- convert_square_brackets(text, ...)
            text      <- paste(.qol_messages[["format"]][[paste0(type, "_pt")]], text_orig, sep = "")
        }
    }
    # If custom message
    else{
        .qol_messages[["last_type"]] <- type[["type"]]

        # First format all line breaks so that the next text lines are indented
        if (utf8){
            if ("timed" %in% names(type)){
                message_start <- paste0(rep("\u00a0", type[["indent"]]), type[["ansi_wait_icon"]])
            }
            else{
                message_start <- paste0(rep("\u00a0", type[["indent"]]), type[["ansi_icon"]])
            }
        }
        else{
            if ("timed" %in% names(type)){
                message_start <- paste0(rep("\u00a0", type[["indent"]]), type[["text_wait_icon"]])
            }
            else{
                message_start <- paste0(rep("\u00a0", type[["indent"]]), type[["text_icon"]])
            }
        }

        # Build the line indentation for each line after the first one
        seperator <- paste0("\n[b][", type[["color"]], " ", message_start, "][/b] ", strrep("\u00a0", nchar(type[["type"]]) + 2))

        # Insert special message format
        if (!is.null(type[["text_color"]])){
            text <- paste0("[", type[["text_color"]], " ", text, "]")
        }
        if (type[["text_bold"]]){
            text <- paste0("[b]", text, "[/b]")
        }
        if (type[["text_italic"]]){
            text <- paste0("[i]", text, "[/i]")
        }
        if (type[["text_underline"]]){
            text <- paste0("[u]", text, "[/u]")
        }

        # Paste multiline text together
        text <- paste(text, collapse = seperator)

        # Insert formatting
        text_orig <- convert_square_brackets(text, ...)

        # Generate timed type and put final message together
        if ("timed" %in% names(type)){
            note_text <- hex_to_ansi(paste(message_start, " ", sep = ""), hex_color = type[["color"]], bold = TRUE)
            text      <- paste(note_text, text_orig, sep = "")
        }
        # Generate colored note type and put final message together
        else{
            note_text <- hex_to_ansi(paste(message_start, " ", type[["type"]], ": ", sep = ""), hex_color = type[["color"]], bold = TRUE)
            text      <- paste(note_text, text_orig, sep = "")
        }
    }

    # Print message to the console conditionally.
    prefix <- ifelse(is_new_line_necessary(), "\n", "")

    if (!is_no_print_active()){
        # If no color should be printed, remove individual styling
        if (no_color){
            ansi_regex <- "(?>\\x1B\\[|\\033\\[)[0-9;:]*[a-zA-Z]"
            text       <- gsub(ansi_regex, "", text, perl = TRUE)
        }

        # Put a line break afterwards, otherwise the next message would be printed on the same line.
        if (new_line){
            caller <- as.character(sys.call(-1))[1]

            # Print as normal
            if (prefix == "" || caller == "print_message"){
                cat(text, "\n")
            }
            # Print a new line before text, because the previous message didn't put a line break afterwards
            else{
                cat(prefix, text, "\n", sep = "")
            }
        }
        # Print without line break afterwards to be able to put time stamps behind the messages
        else{
            cat(text)
        }
        utils::flush.console()
    }

    # Remove line breaks and doubled blanks to receive the full message in one piece
    text_orig  <- gsub(seperator, " ", text_orig, fixed = TRUE)

    # Remove individual styling
    ansi_regex <- "(?>\\x1B\\[|\\033\\[)[0-9;:]*[a-zA-Z]"
    text_orig  <- gsub(ansi_regex, "", text_orig, perl = TRUE)

    # Set global last text, which is used to print time stamps
    .qol_messages[["last_message"]] <- text

    invisible(text_orig)
}


###############################################################################
# Headline and closing
###############################################################################

#' @param line_char The character that that forms the line.
#' @param max_width The maximum number of characters drawn, which determines the
#' line length of the headline.
#'
#' @examples
#' # Different headlines
#' print_headline("This is a headline")
#'
#' print_headline("[#00FFFF This is a different headline] with some color",
#'                line_char = "-")
#'
#' print_headline("[b]This is a very small[/b] and bold headline",
#'                line_char = ".",
#'                max_width = 60)
#'
#' @rdname messages
#'
#' @export
print_headline <- function(text,
                           ...,
                           line_char    = "=",
                           max_width    = getOption("width"),
                           always_print = FALSE,
                           utf8         = .qol_messages[["format"]][["utf8"]],
                           no_color     = .qol_messages[["no_color"]]){
    suppressed <- FALSE

    # Identify which function called the message
    caller <- sys.call(-1)

    if (is.call(caller) && is.name(caller[[1]])){
        # Get calling function and code line the message was called from
        caller <- ifelse(is.null(caller), "global", as.character(caller[[1]]))
    }
    else{
        caller <- "do.call"
    }

    # Don't print message, if messages are suppressed
    depth     <- sys.nframe()
    sys_calls <- get_condensed_sys_calls()

    if (!always_print && is_message_suppressed(sys_calls)){
        suppressed <- TRUE
        text_orig  <- text
    }
    else{
        # Put time stamp at the end of the last message if present
        print_time_stamp(FALSE, utf8, no_color)

        # Only get first element, if a vector is given
        if (length(text) > 1){
            text <- text[[1]]
        }

        # Only one character allowed
        if (nchar(line_char) > 1){
            line_char <- substr(line_char, 1, 1)
        }

        # Insert formatting
        text <- convert_square_brackets(text, ...)

        # Remove individual styling to be able to measure the right headline width
        ansi_regex <- "(?>\\x1B\\[|\\033\\[)[0-9;:]*[a-zA-Z]"
        text_orig  <- gsub(ansi_regex, "", text, perl = TRUE)

        # Pre format headline start
        text_measure <- paste0(strrep(line_char, 3), " ", text_orig, " ")
        text         <- paste0(strrep(line_char, 3), " ", text, " ")

        # Get console width and calculate remaining characters
        fill_width <- max(0, max_width - nchar(text_measure))

        if (!is_no_print_active()){
            # If no color should be printed, remove individual styling
            if (no_color){
                ansi_regex <- "(?>\\x1B\\[|\\033\\[)[0-9;:]*[a-zA-Z]"
                text       <- gsub(ansi_regex, "", text, perl = TRUE)
            }

            cat("\n", text, strrep(line_char, fill_width), "\n", sep = "")
            utils::flush.console()
        }
    }

    # Add message to global stack
    entry_nr <- length(.qol_messages[["stack"]]) + 1
    .qol_messages[["stack"]][[entry_nr]] <- list(type       = "HEADLINE",
                                                 text       = text_orig,
                                                 suppressed = suppressed,
                                                 new_line   = TRUE,
                                                 print_time = FALSE,
                                                 in_place   = FALSE,
                                                 depth      = depth,
                                                 caller     = caller,
                                                 call_stack = sys_calls,
                                                 time       = Sys.time())

    invisible(text_orig)
}


#' @param current_time The current time to create a time stamp.
#' @param caller_color Hex color code for the displayed caller function.
#' @param suppress FALSE by default. If TRUE triggers all the message
#' procedures to create a message stack but doesn't print the message to the
#' console.
#'
#' @examples
#' # Messages with time stamps
#' test_func <- function(){
#'     print_start_message()
#'     print_step("GREY", "Probably not so important")
#'     print_step("MAJOR", "This is a major step...")
#'     print_step("MINOR", "Sub step1")
#'     print_step("MINOR", "Sub step2")
#'     print_step("MINOR", "Sub step3")
#'     print_step("MAJOR", "[b]Finishing... [/b][#00FFFF with some color again!]")
#'     print_closing()
#' }
#'
#' test_func()
#'
#' # See what is going on in the message stack
#' message_stack <- get_message_stack()
#'
#' # Print messages on the same line instead of below each other
#' in_place_steps <- function(){
#'     print_start_message()
#'
#'     print_step("MAJOR", "Let's get started...")
#'
#'     for (i in seq_len(10)){
#'         print_step("Minor", "This is in place step [i] of 10", i = i, in_place = TRUE)
#'         Sys.sleep(0.25)
#'     }
#'
#'     print_step("MAJOR", "Loop has ended")
#'
#'     print_closing()
#' }
#'
#' in_place_steps()
#'
#' @rdname messages
#'
#' @export
print_start_message <- function(current_time = Sys.time(),
                                caller_color = "#63C2C9",
                                always_print = FALSE,
                                suppress     = FALSE,
                                utf8         = .qol_messages[["format"]][["utf8"]],
                                no_color     = .qol_messages[["no_color"]]){
    depth      <- sys.nframe()
    sys_calls  <- get_condensed_sys_calls()
    suppressed <- FALSE
    no_print   <- FALSE

    # Identify which function called the message
    caller <- sys.call(-1)

    if (is.call(caller) && is.name(caller[[1]])){
        caller <- as.character(sys.call(-1)[[1]])
        caller <- ifelse(length(caller) == 0, "global", caller)
    }
    else{
        caller <- "do.call"
    }

    # Whenever a code block is executed, a unique token is generated. Based on this
    # token it can be decided, whether we are starting a new message stack (in case we
    # have a new token) or we keep the current stack (in case of multiple functions
    # running as a block)
    current_session_token   <- format(sys.frame(1))
    current_execution_token <- get_execution_token(depth, sys_calls)

    # Don't print starting message, if messages are suppressed
    if (!always_print && is_message_suppressed(sys_calls)){
        # In case it is the first starting message in the current run or if the stack
        # has a not suppressed end not. Meaning a function or pipe operation has finished.
        if (!identical(current_session_token, .qol_messages[["last_session"]]) ||
            !identical(current_execution_token, .qol_messages[["last_execution"]]) ||
            message_stack_needs_reset()){

            reset_messages()
        }

        suppressed <- TRUE
        entry_nr   <- length(.qol_messages[["stack"]]) + 1
        text_orig  <- paste0(caller, " started")
    }
    else{
        # In case it is the first starting message in the current run or if the stack
        # has a not suppressed end not. Meaning a function or pipe operation has finished.
        if (!identical(current_session_token,   .qol_messages[["last_session"]]) ||
            !identical(current_execution_token, .qol_messages[["last_execution"]]) ||
            message_stack_needs_reset()){

            reset_messages()
            .qol_messages[["last_session"]]   <- current_session_token
            .qol_messages[["last_execution"]] <- current_execution_token

            entry_nr <- 1

            .qol_messages[["start_time"]] <- current_time
            start_timer()

            # Get current time in hh:mm:ss format
            current_time <- format(current_time, "%H:%M:%S")

            if (!is_no_print_active()){
                if (!no_print && !suppress){
                    # Print actual headline
                    text      <- paste0("[b][", caller_color, " ", caller, "][/b] started at ", current_time)
                    text_orig <- print_headline(text = text, no_color = no_color, always_print = TRUE)
                    cat("\n")
                    utils::flush.console()
                }
                else{
                    text_orig <- paste0("[b][", caller_color, " ", caller, "][/b] started at ", current_time)
                }
            }
            else{
                text_orig <- paste0("[b][", caller_color, " ", caller, "][/b] started at ", current_time)
            }
        }
        # In case it is another starting message within the same session
        else{
            entry_nr <- length(.qol_messages[["stack"]]) + 1

            suppressed <- TRUE
            no_print   <- TRUE

            # Get current time in hh:mm:ss format
            current_time <- format(current_time, "%H:%M:%S")

            text_orig <- paste0("[b][", caller_color, " ", caller, "][/b] started at ", current_time)
        }
    }

    # Special case, because otherwise HEADLINE is the first entry
    .qol_messages[["stack"]][[entry_nr]] <- NULL
    .qol_messages[["stack"]][[entry_nr]] <- list(type       = "START",
                                                 text       = text_orig,
                                                 suppressed = suppressed,
                                                 new_line   = TRUE,
                                                 print_time = FALSE,
                                                 in_place   = FALSE,
                                                 depth      = depth,
                                                 caller     = caller,
                                                 call_stack = sys_calls,
                                                 time       = Sys.time())

    invisible(text_orig)
}


#' @param time_threshold The total time spent is displayed in different colors from
#' green over yellow to red, depending on the threshold specified (in seconds).
#' @param start_time The time at which the function call started to calculate the
#' time difference and output the total time spent.
#'
#' @rdname messages
#'
#' @export
print_closing <- function(time_threshold = 2,
                          start_time     = .qol_messages[["start_time"]],
                          caller_color   = "#63C2C9",
                          always_print   = FALSE,
                          suppress       = FALSE,
                          utf8           = .qol_messages[["format"]][["utf8"]],
                          no_color       = .qol_messages[["no_color"]]){
    depth      <- sys.nframe()
    sys_calls  <- get_condensed_sys_calls()
    suppressed <- FALSE

    # Identify which function called the message
    caller <- sys.call(-1)

    if (is.call(caller) && is.name(caller[[1]])){
        caller <- as.character(sys.call(-1)[[1]])
        caller <- ifelse(length(caller) == 0, "global", caller)
    }
    else{
        caller <- "do.call"
    }

    # Don't print starting message, if messages are suppressed
    if (!always_print && is_message_suppressed(sys_calls)){
        suppressed <- TRUE
        entry_nr   <- length(.qol_messages[["stack"]]) + 1
        text_orig  <- paste0(caller, " ended")
    }
    else{
        # Whenever a code block is executed, a unique token is generated. Based on this
        # token it can be decided, whether we are starting a new message stack (in case we
        # have a new token) or we keep the current stack (in case of multiple functions
        # running as a block)
        current_session_token   <- format(sys.frame(1))
        current_execution_token <- get_execution_token(depth, sys_calls)

        if (is_stack() && (!identical(current_session_token,   .qol_messages[["last_session"]]) ||
                           !identical(current_execution_token, .qol_messages[["last_execution"]]))){
            if (!is_start_fake()){
                suppressed <- TRUE
            }
            else{
                reset_messages()
            }
        }

        # As safety net set a starting point, in case no starting message was used,
        # to imitate a full message stack.
        if (!is_stack()){
            if (length(.qol_messages[["stack"]]) == 0){
                .qol_messages[["stack"]] <- list(list(type = "START", suppressed = TRUE, text = "FAKE"))
            }
            else{
                .qol_messages[["stack"]] <- c(list(list(type = "START", suppressed = TRUE, text = "FAKE")), .qol_messages[["stack"]])
            }

            .qol_messages[["last_session"]]   <- current_session_token
            .qol_messages[["last_execution"]] <- current_execution_token
        }

        # Put time stamp at the end of the last message if present
        print_time_stamp(FALSE, utf8, no_color)

        stop_timer()

        if (is.null(start_time)){
            start_time <- Sys.time()
            print_message("NOTE", c("No starting time was set. If you didn't use print_starting_message(),",
                                    "you have to capture the starting time in your function manually and pass",
                                    "it to print_closing()."))
        }

        entry_nr <- length(.qol_messages[["stack"]]) + 1

        # Calculate time difference between provided starting time and current time
        formatted_time <- transform_time(start_time)

        # Generating a seamless color transition from green over yellow to red
        time_in_seconds <- as.numeric(Sys.time() - start_time, units = "secs")
        ratio           <- min(time_in_seconds / time_threshold, 1)
        color_ramp      <- grDevices::colorRamp(c("#32CD32", "#FFC90E", "#C93F3F"))

        # Get the RGB value for the current ratio
        rgb_value <- color_ramp(min(ratio, 1))

        # Convert RGB matrix to a Hex string
        hex_color <- grDevices::rgb(rgb_value[1], rgb_value[2], rgb_value[3], maxColorValue = 255)

        # Print actual headline
        if (!is_no_print_active() && !suppress){
            text      <- paste0("[b][", caller_color, " ", caller, "][/b] execution time: [", hex_color, " ", formatted_time, "]")
            text_orig <- print_headline(text = text, no_color = no_color, always_print = TRUE)
            cat("\n")
            utils::flush.console()
        }
        else{
            text_orig <- paste0("[b][", caller_color, " ", caller, "][/b] execution time: [", hex_color, " ", formatted_time, "]")
        }
    }

    # Special case, because otherwise HEADLINE would be last before the END
    .qol_messages[["stack"]][[entry_nr]] <- NULL
    .qol_messages[["stack"]][[entry_nr]] <- list(type       = "END",
                                                 text       = text_orig,
                                                 suppressed = suppressed,
                                                 new_line   = TRUE,
                                                 print_time = FALSE,
                                                 in_place   = FALSE,
                                                 depth      = depth,
                                                 caller     = caller,
                                                 call_stack = sys_calls,
                                                 time       = Sys.time())

    # Give list entries a name if the stack is closed
    name_stack_messages()

    invisible(text_orig)
}


#' @noRd
transform_time <- function(start_time){
    # Calculate time difference between provided starting time and current time
    time_in_seconds <- as.numeric(Sys.time() - start_time, units = "secs")

    # Convert time format into different units depending on time difference
    if (time_in_seconds < 1){
        formatted_time <- paste0(round(time_in_seconds * 1000), " ms")
    }
    else if (time_in_seconds < 60){
        formatted_time <- paste0(format(time_in_seconds, digits = 3, decimal.mark = ","), " sec")
    }
    else if (time_in_seconds < 3600){
        formatted_time <- paste0(format(time_in_seconds / 60, digits = 3, decimal.mark = ","), " min")
    }
    else{
        formatted_time <- paste0(format(time_in_seconds / 3600, digits = 3, decimal.mark = ","), " hours")
    }

    formatted_time
}


###############################################################################
# Major and minor steps
###############################################################################

#'
#' @param type The message type, so that the function knows which symbol and coloring
#' to use. Allowed are "note", "warning", "error", "neutral", "major", "minor" and "grey".
#' @param in_place Prints the step message on the same line as before, instead of in
#' the next line. This can e.g. be used inside loops.
#'
#' @rdname messages
#'
#' @export
print_step <- function(type,
                       text,
                       ...,
                       in_place     = FALSE,
                       always_print = FALSE,
                       utf8         = .qol_messages[["format"]][["utf8"]],
                       no_color     = .qol_messages[["no_color"]]){
    suppressed <- FALSE

    # Identify which function called the message
    caller <- sys.call(-1)

    if (is.call(caller) && is.name(caller[[1]])){
        # Get calling function and code line the message was called from
        caller <- ifelse(is.null(caller), "global", as.character(caller[[1]]))
    }
    else{
        caller <- "do.call"
    }

    if (!is.list(type) && !tolower(type) %in% c("major", "minor", "grey")){
        # Check if message type is a custom one and retrieve from global list
        if (!toupper(type) %in% names(.qol_messages[["custom_types"]])){
            type <- "major"
        }
        else{
            type            <- .qol_messages[["custom_types"]][[toupper(type)]]
            type[["timed"]] <- TRUE
        }
    }

    # Check how deep nested this function is called.
    depth     <- sys.nframe()
    sys_calls <- get_condensed_sys_calls()

    # Whenever a code block is executed, a unique token is generated. Based on this
    # token it can be decided, whether we are starting a new message stack (in case we
    # have a new token) or we keep the current stack (in case of multiple functions
    # running as a block)
    current_session_token   <- format(sys.frame(1))
    current_execution_token <- get_execution_token(depth, sys_calls)

    # In case it is the first starting message in the current run or if the stack
    # has a not suppressed end not. Meaning a function or pipe operation has finished.
    if (!identical(current_session_token,   .qol_messages[["last_session"]]) ||
        !identical(current_execution_token, .qol_messages[["last_execution"]])){

        if (awaits_time_print() && !is_no_print_active() && !is_message_suppressed(sys_calls)){
            cat("\n")
        }

        reset_messages()
    }
    else if (is.null(.qol_messages[["timer"]])){
        .qol_messages[["timer"]] <- Sys.time()
    }

    # Add message to global stack
    if (is_stack() && !is_full_stack()){
        entry_nr <- length(.qol_messages[["stack"]]) + 1
    }
    # If no starting point is set, the message is called inside a function where no
    # starting message was called. In this case only capture the last message and reset
    # the message stack. Otherwise the stack would just grow and never reset.
    else{
        if (length(.qol_messages[["stack"]]) > 1){
            .qol_messages[["stack"]]        <- list()
            .qol_messages[["last_message"]] <- NULL
            .qol_messages[["last_type"]]    <- NULL
        }

        entry_nr <- 1
    }

    if (!always_print && is_message_suppressed(sys_calls)){
        suppressed <- TRUE
    }
    else{
        # Put time stamp at the end of the last message if present
        print_time_stamp(in_place, utf8, no_color)

        # If messages are not suppressed
        text <- paste(text)
        text <- print_to_console(type = type,
                                 text = text,
                                 ...,
                                 utf8     = utf8,
                                 no_color = no_color,
                                 new_line = FALSE)
    }

    # Catch type for message stack on custom messages
    if (is.list(type)){
        type <- type[["type"]]
    }

    # Add message to global stack
    .qol_messages[["stack"]][[entry_nr]] <- list(type       = toupper(type),
                                                 text       = text,
                                                 suppressed = suppressed,
                                                 new_line   = FALSE,
                                                 print_time = TRUE,
                                                 in_place   = in_place,
                                                 depth      = depth,
                                                 caller     = caller,
                                                 call_stack = sys_calls,
                                                 time       = Sys.time())

    invisible(text)
}

###############################################################################
# Custom message
###############################################################################

#' @param type If displayed as a normal note, then this is the text displayed
#' in front of the message. This also appears as type in the message stack.
#' @param color The color of the message type.
#' @param ansi_icon The icon used when message is displayed in utf8 mode.
#' @param text_icon The icon used when message is displayed in text only mode.
#' @param ansi_wait_icon The icon used when a timed message is waiting for the end of
#' execution displayed in utf8 mode.
#' @param text_wait_icon The icon used when a timed message is waiting for the end of
#' execution displayed in text only mode.
#' @param indent How many spaces to indent the message.
#' @param text_bold FALSE by default. If TRUE prints the message text in bold letters.
#' @param text_italic FALSE by default. If TRUE prints the message text in italic letters.
#' @param text_underline FALSE by default. If TRUE prints the message text underlined.
#' @param text_color The color of the actual message text.
#' @param time_color The color used for the time stamps.
#'
#' @returns
#' [set_up_custom_message()]: Returns the global list of custom messages.
#'
#' @examples
#' # Set up a custom message
#' set_up_custom_message(type           = "HOTDOG",
#'                       color          = "#B27A01",
#'                       ansi_icon      = "\U0001f32d",
#'                       text_icon      = "IOI",
#'                       ansi_wait_icon = "\U00023f1",
#'                       text_wait_icon = "/",
#'                       indent         = 1)
#'
#' hotdog_print <- function(){
#'     print_start_message()
#'     print_message("HOTDOG", c("This is the first hotdog message! Hurray!",
#'                               "And it is also multiline in this version."))
#'     print_step("HOTDOG", "Or use as single line message with time stamps.")
#'     Sys.sleep(0.5)
#'     print_step("HOTDOG", "Or use as single line message with time stamps.")
#'     Sys.sleep(0.5)
#'     print_step("HOTDOG", "Or use as single line message with time stamps.")
#'     Sys.sleep(0.5)
#'     print_closing()
#' }
#'
#' hotdog_print()
#'
#' # See new message in the message stack
#' hotdog_stack <- get_message_stack()
#'
#' @rdname messages
#'
#' @export
set_up_custom_message <- function(type           = "UNICORN",
                                  color          = "#FF00FF",
                                  ansi_icon      = NULL,
                                  text_icon      = "^",
                                  ansi_wait_icon = NULL,
                                  text_wait_icon = "?",
                                  indent         = 1,
                                  text_bold      = FALSE,
                                  text_italic    = FALSE,
                                  text_underline = FALSE,
                                  text_color     = NULL,
                                  time_color     = "#6B6B6B"){
    # Set default icon here, because if set as standard argument, roxygen would
    # convert it into the emoji and then the LaTex pdf conversion would throw errors
    if (is.null(ansi_icon)){
        ansi_icon <- "\U0001f984"
    }

    if (is.null(ansi_wait_icon)){
        ansi_wait_icon <- "\u23f3\ufe0f"
    }

    type <- toupper(type)

    # Update global list of custom messages
    custom_message <- as.list(environment())
    .qol_messages[["custom_types"]][[type]] <- custom_message

    invisible(.qol_messages[["custom_types"]])
}

###############################################################################
# Helper functions
###############################################################################

#' Message Helper Functions
#'
#' @name message_helpers
#'
#' @description
#' [get_message_stack()]: Retrieves the current message stack as list or data frame.
#'
#' @param as_data_frame FALSE by default. If TRUE returns message stack information as
#' data frame.
#'
#' @seealso
#' Main printing functions: [print_message()], [print_headline()], [print_start_message()],
#' [print_closing()], [print_step()], [set_up_custom_message()]
#'
#' @returns
#' [get_message_stack()]: Returns a list of messages or a data frame.
#'
#' @rdname message_helpers
#'
#' @export
get_message_stack <- function(as_data_frame = FALSE){
    # Get stack as is
    if (!as_data_frame){
        .qol_messages[["stack"]]
    }
    # Convert to data frame without the whole call stack
    else{
        stack_df <- do.call(rbind, lapply(.qol_messages[["stack"]], function(message){
            message[["call_stack"]] <- NULL
            data.frame(message, stringsAsFactors = FALSE)
        }))

        rownames(stack_df) <- NULL

        stack_df
    }
}


#' @description
#' [set_no_print()]: FALSE by default. If set to TRUE the messages will be formatted
#' and returned but not printed to the console. Can e.g. be used in unit test
#' situations.
#'
#' @param value Can be TRUE or FALSE.
#'
#' @returns
#' [set_no_print()]: Returns the global no_print option.
#'
#' @rdname message_helpers
#'
#' @export
set_no_print <- function(value = FALSE){
    .qol_messages[["no_print"]] <- value

    invisible(.qol_messages[["no_print"]])
}


#' @description
#' [set_no_color()]: TRUE by default. Suppresses the color codes so that messages
#' can be printed clean.
#'
#' @param value Can be TRUE or FALSE.
#'
#' @returns
#' [set_no_color()]: Returns the global no_color option.
#'
#' @rdname message_helpers
#'
#' @export
set_no_color <- function(value = TRUE){
    .qol_messages[["no_color"]] <- value

    invisible(.qol_messages[["no_color"]])
}


#' @description
#' [print_stack_as_messages()]: Prints the message stack as actual messages (only
#' not suppressed messages). Can be used to trigger expect_message, expect_warning
#' or expect_error in unit tests.
#'
#' @param type The message type to filter.
#'
#' @returns
#' [print_stack_as_messages()]: Returns NULL.
#'
#' @rdname message_helpers
#'
#' @export
print_stack_as_messages <- function(type = NULL){
    for (message in .qol_messages[["stack"]]){
        # Skip suppressed messages
        if (message[["suppressed"]]){
            next
        }

        # Skip start and end point
        if (message[["type"]] %in% c("START", "END")){
            next
        }

        # Skip if messages are filtered by type
        if (!is.null(type) && toupper(type) != message[["type"]]){
            next
        }

        # Print message according to its type
        if (message[["type"]] == "ERROR"){
            stop(message[["text"]])
        }
        else if (message[["type"]] == "WARNING"){
            warning(message[["text"]])
        }
        else{
            message(message[["text"]])
        }
    }

    invisible(NULL)
}


#' @description
#' [convert_square_brackets()]: Convert different curly bracket patterns into ansi
#' formatting or passes the context of vectors into the text.
#'
#' @param text The text in which to replace the curly brackets.
#' @param ... The actual replacement vectors.
#'
#' @details
#' The message types in which you can enter custom texts, are capable of using
#' different styling operators. These are:
#'
#' - Insert list of elements: &#91;vector_name&#93;
#' - Adding conditional words, if list of elements has more than one element: &#91;?word&#93;
#' - Adding conditional singular/plural, depending on list of element length: &#91;?singular/plural&#93;
#' - Bold, italic and underline: &#91;b&#93;some text&#91;/b&#93;, &#91;i&#93;some text&#91;/i&#93;, &#91;u&#93;some text&#91;/u&#93;
#' - Coloring parts of the message: &#91;#FF00FF some text&#93;
#'
#' @returns
#' [convert_square_brackets()]: Returns formatted text.
#'
#' @rdname message_helpers
#'
#' @export
convert_square_brackets <- function(text, ...){
    arguments <- list(...)
    env       <- parent.frame()

    number_of_elements <- 1

    # Capture all extra arguments and put them into the current environment
    # so this function can see them for the {} actions
    if (length(arguments) > 0){
        number_of_elements <- length(arguments[[1]])

        env <- list2env(arguments, envir = new.env(parent = env))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Convert square brackets to curly brackets because the regex seems to work
    # better on them.
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    text <- gsub("\\[(/?[biu])\\]", "{\\1}",                text, perl = TRUE)
    text <- gsub("\\[(#[A-Fa-f0-9]{6}\\s+.*?)\\]", "{\\1}", text, perl = TRUE)
    text <- gsub("\\[(\\?.*?)\\]", "{\\1}",                 text, perl = TRUE)
    text <- gsub("\\[([a-zA-Z0-9_.]+)\\]", "{\\1}",         text, perl = TRUE)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Handle vector/variable insertion: [variable_name]
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    var_matches <- regmatches(text, gregexpr("(?<!\\?)\\{[a-zA-Z0-9_.]+\\}", text, perl = TRUE))[[1]]

    for (single_match in var_matches){
        # Remove curly brackets
        var_name <- gsub("[{}]", "", single_match)

        # Look for variable name in th current environment. If it exists replace it with
        # its actual contents.
        if (exists(var_name, envir = env)){
            # Get actual vector
            values <- get(var_name, envir = env)

            # Get replacement vector based on length
            replacement <- ifelse(length(values) > 1,
                                  paste(values, collapse = ", "),
                                  as.character(values))

            # Insert replacement in the provided text
            text <- gsub(single_match, replacement, text, fixed = TRUE)
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Insert conditional word, if length of first vector is greater than 1: [?s]
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    word_matches <- regmatches(text, gregexpr("\\{\\?[^/\\}]+\\}", text, perl = TRUE))[[1]]

    for (single_match in word_matches){
        # Get actual word
        word <- gsub("[{}?]", "", single_match)

        # Get replacement based on length
        replacement <- ifelse(number_of_elements != 1, word, "")

        # Insert replacement in the provided text
        text <- gsub(single_match, replacement, text, fixed = TRUE)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Conditionally insert singular or plural words, if length of first vector is
    # greater than 1: [?is/are]
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    slash_matches <- regmatches(text, gregexpr("\\{\\?[^\\}]*/[^\\}]+\\}", text))[[1]]

    for (single_match in slash_matches){
        # Get actual vector
        parts <- unlist(strsplit(gsub("[{}?]", "", single_match), "/"))

        # Get replacement based on length
        replacement <- ifelse(number_of_elements != 1, parts[2], parts[1])

        # Insert replacement in the provided text
        text <- gsub(single_match, replacement, text, fixed = TRUE)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Handle styling of bold [b], italic [i], underlined [u]
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    style_map <- list("\\{b\\}" = "\033[1m", "\\{/b\\}" = "\033[22m",
                      "\\{i\\}" = "\033[3m", "\\{/i\\}" = "\033[23m",
                      "\\{u\\}" = "\033[4m", "\\{/u\\}" = "\033[24m")

    # Just a quick find and replace loop
    for (tag in names(style_map)){
        text <- gsub(tag, style_map[[tag]], text)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Handle dynamic coloring with hex codes: [#FF00FF Some colored text]
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    color_matches <- regmatches(text, gregexpr("\\{#([A-Fa-f0-9]{6})\\s+([^\\}]+)\\}", text, perl = TRUE))[[1]]

    for (single_match in color_matches){
        # Remove curly brackets and hashtag
        expression <- sub("^\\{#([A-Fa-f0-9]{6})\\s+", "\\1|", single_match)
        expression <- sub("\\}$", "", expression)

        # Separate hex code and the actual text
        parts <- unlist(strsplit(expression, "\\|"))

        hex_color     <- paste0("#", parts[1])
        text_to_color <- parts[2]

        # Convert color to ansi codes
        replacement <- hex_to_ansi(text_to_color, hex_color = hex_color)

        # Add a closing code after the text to ensure the text coming after the colored
        # one is printed in normal color again.
        replacement <- paste0(replacement, "\033[0m")

        text <- gsub(single_match, replacement, text, fixed = TRUE)
    }

    invisible(text)
}


###############################################################################
# Iternal functions
###############################################################################

#' @description
#' Resets all global message options.
#'
#' @returns
#' Returns global message options.
#'
#' @noRd
reset_messages <- function(){
    .qol_messages[["stack"]]        <- list()
    .qol_messages[["start_time"]]   <- NULL
    .qol_messages[["timer"]]        <- Sys.time()
    .qol_messages[["last_message"]] <- NULL
    .qol_messages[["last_type"]]    <- NULL

    invisible(.qol_messages)
}


#' @description
#' Checks whether there is a START node in the message stack.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_stack <- function(){
    length(.qol_messages[["stack"]]) > 0 && .qol_messages[["stack"]][[1]][["type"]] == "START"
}


#' @description
#' Checks whether there is a START and END node in the message stack.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_full_stack <- function(){
    is_stack() &&
        .qol_messages[["stack"]][[length(.qol_messages[["stack"]])]][["type"]] == "END" &&
        !.qol_messages[["stack"]][[length(.qol_messages[["stack"]])]][["suppressed"]] &&
        .qol_messages[["stack"]][[1]][["caller"]] == .qol_messages[["stack"]][[length(.qol_messages[["stack"]])]][["caller"]] &&
        .qol_messages[["stack"]][[1]][["depth"]] == .qol_messages[["stack"]][[length(.qol_messages[["stack"]])]][["depth"]]
}


#' @description
#' Checks whether the stack has only one message which
#' is not a START message.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_single_non_starting_message <- function(){
    length(.qol_messages[["stack"]]) == 1 && .qol_messages[["stack"]][[1]][["type"]] != "START"
}


#' @description
#' Checks whether the first message in the stack
#' is suppressed.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_starting_message_suppressed <- function(){
    length(.qol_messages[["stack"]]) >= 1 && .qol_messages[["stack"]][[1]][["suppressed"]]
}


#' @description
#' Checks whether the last message in the stack is an error.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_last_message_error <- function(){
    length(.qol_messages[["stack"]]) >= 1 && .qol_messages[["stack"]][[length(.qol_messages[["stack"]])]][["type"]] == "ERROR"
}


#' @description
#' Checks whether the last message in the stack is an in place step.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_last_message_in_place <- function(){
    length(.qol_messages[["stack"]]) >= 1 && .qol_messages[["stack"]][[length(.qol_messages[["stack"]])]][["in_place"]]
}


#' @description
#' Checks whether the starting message is a fake. A fake start
#' message is inserted, if a closing message was called without starting message.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_start_fake <- function(){
    length(.qol_messages[["stack"]]) >= 1 &&
        .qol_messages[["stack"]][[1]][["type"]] == "START" &&
        .qol_messages[["stack"]][[1]][["text"]] == "FAKE"
}


#' @description
#' Condition on when the message stack needs to be reset.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
message_stack_needs_reset <- function(){
    is_full_stack() || is_single_non_starting_message() || is_starting_message_suppressed() || is_last_message_error()
}


#' @description
#' Checks whether the parent function is surrounded by suppressMessages.
#'
#' @param sys_calls The condensed sys.calls.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_message_suppressed <- function(sys_calls){
    targets <- c("suppressMessages", "suppressPackageStartupMessages", "suppressWarnings")

    # Vectorized substring check across all calls and all targets
    any(outer(sys_calls, targets, Vectorize(function(call, pattern){
        grepl(pattern, call, fixed = TRUE)
        })))
}


#' @description
#' Checks whether the last message set up a new line or if it
#' is necessary to manually print a new line before printing the actual message.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_new_line_necessary <- function(){
    !is.null(.qol_messages[["last_message"]]) && !.qol_messages[["stack"]][[length(.qol_messages[["stack"]])]][["new_line"]]
}


#' @description
#' Checks whether it is possible to print a time stamp
#' at the end of the last message.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_time_stamp_possible <- function(){
    !is.null(.qol_messages[["timer"]]) && !is.null(.qol_messages[["last_message"]])
}


#' @description
#' Checks whether the last message allows to put a time stamp at its end.
#'
#' @param message Message from a message in the stack.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_time_stamp_allowed <- function(message){
    if (is.null(message)){
        return(invisible(FALSE))
    }

    if (message[["print_time"]]){
        return(invisible(TRUE))
    }

    invisible(FALSE)
}


#' @description
#' Checks whether the global timer is running.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_timer_running <- function(){
    !is.null(.qol_messages[["timer"]])
}


#' @description
#' Checks whether the global option to suppress message
#' printing is active.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
is_no_print_active <- function(){
    .qol_messages[["no_print"]]
}


#' @description
#' Checks whether the last message in the stack awaits a time stamp.
#'
#' @returns
#' Returns TRUE or FALSE.
#'
#' @noRd
awaits_time_print <- function(){
    length(.qol_messages[["stack"]]) > 0 && .qol_messages[["stack"]][[length(.qol_messages[["stack"]])]][["print_time"]]
}


#' @description
#' Get a unique token for the current code block execution.
#'
#' @param depth The current sysframe depth this function is called from.
#' @param sys_calls The condensed sys.calls.
#'
#' @returns
#' Returns upper function call as string.
#'
#' @noRd
get_execution_token <- function(depth, sys_calls){
    # Only get the execution token, if the nesting is not too deep. Otherwise this
    # will become very resource heavy.
    if (depth < 20){
        # Get the condensed call stack
        sys_calls <- get_condensed_sys_calls()

        # If there is nothing left, the function was called from the global environment
        if (length(sys_calls) == 0){
            return("global")
        }

        # Return upper function call
        sys_calls[1]
    }
    else{
        .qol_messages[["last_execution"]]
    }
}


#' @description
#' Gets all information from the last not suppressed message.
#'
#' @returns
#' Returns a list.
#'
#' @noRd
get_last_active_message <- function(){
    last_entry <- length(.qol_messages[["stack"]])

    if (last_entry == 0){
        return(invisible(NULL))
    }

    for (i in last_entry:1){
        entry <- .qol_messages[["stack"]][[i]]

        if (!entry[["suppressed"]]){
            return(invisible(entry))
        }
    }

    invisible(NULL)
}


#' @description
#' Stops the global timer.
#'
#' @returns
#' Returns NULL.
#'
#' @noRd
stop_timer <- function(){
    .qol_messages[["timer"]] <- NULL

    invisible(.qol_messages[["timer"]])
}


#' @description
#' Starts the global timer.
#'
#' @returns
#' Returns the current time.
#'
#' @noRd
start_timer <- function(){
    .qol_messages[["timer"]] <- Sys.time()

    invisible(.qol_messages[["timer"]])
}


#' @description
#' Names the single messages in the global message stack.
#'
#' @returns
#' Returns the global message stack.
#'
#' @noRd
name_stack_messages <- function(){
    names(.qol_messages[["stack"]]) <- vapply(.qol_messages[["stack"]], function(message){

        caller <- message[["caller"]]
        type   <- message[["type"]]
        text   <- substr(message[["text"]], 1, 20)

        paste(caller, type, text, sep = " | ")

    }, character(1))

    invisible(.qol_messages[["stack"]])
}


#' @description
#' Converts the hourglass symbol.
#'
#' @param text Text of the last message.
#' @param last_message The last message from the stack.
#'
#' @returns
#' Returns a converted text.
#'
#' @noRd
convert_hourglass <- function(type = NULL,
                              text,
                              last_message,
                              utf8           = .qol_messages[["format"]][["utf8"]]){
    text <- .qol_messages[["last_message"]]
    type <- .qol_messages[["last_type"]]

    # Check if message type is a custom one and retrieve from global list
    if (!is.list(type) && toupper(type) %in% names(.qol_messages[["custom_types"]])){
        type <- .qol_messages[["custom_types"]][[toupper(type)]]
    }

    if (utf8){
        if (last_message[["type"]] == "MAJOR"){
            text <- gsub("\u23f3\ufe0f", "\u2794", text)
        }
        else if (last_message[["type"]] == "MINOR"){
            text <- gsub("\u23f3\ufe0f", "\u271a", text)
        }
        else if (last_message[["type"]] == "GREY"){
            # Do nothing
        }
        else{
            text <- gsub(type[["ansi_wait_icon"]], type[["ansi_icon"]], text)
        }
    }
    else{
        if (last_message[["type"]] == "MAJOR"){
            text <- gsub("\\?", ">", text)
        }
        else if (last_message[["type"]] == "MINOR"){
            text <- gsub("\\?", "+", text)
        }
        else if (last_message[["type"]] == "GREY"){
            # Do nothing
        }
        else{
            text <- gsub(paste0("\\", type[["text_wait_icon"]]), type[["text_icon"]], text)
        }
    }

    text
}


#' @description
#' Put time stamp at the end of the last message if present.
#'
#' @returns
#' Returns a converted text.
#'
#' @noRd
print_time_stamp <- function(in_place,
                             utf8     = .qol_messages[["format"]][["utf8"]],
                             no_color = .qol_messages[["no_color"]]){
    type <- .qol_messages[["last_type"]]

    if (is_time_stamp_possible()){
        last_message <- get_last_active_message()

        if (is_time_stamp_allowed(last_message)){
            # Calculate time difference between provided starting time and current time
            formatted_time <- transform_time(.qol_messages[["timer"]])
            start_timer()

            # Format message
            last_text <- convert_hourglass(type,
                                           .qol_messages[["last_message"]],
                                           last_message,
                                           utf8)

            time_color <- .qol_messages[["format"]][["time_color"]]

            # Check if message type is a custom one and retrieve from global list
            if (is.list(type)){
                time_color <- type[["time_color"]]
            }
            else if (!is.list(type) && toupper(type) %in% names(.qol_messages[["custom_types"]])){
                time_color <- .qol_messages[["custom_types"]][[toupper(type)]][["time_color"]]
            }

            time_stamp <- paste0(last_text, " [", time_color, " (", formatted_time, ")]")
            .qol_messages[["last_message"]] <- convert_square_brackets(time_stamp)

            # Print message
            if (!is_no_print_active()){
                # If no color should be printed, remove individual styling
                if (no_color){
                    ansi_regex <- "(?>\\x1B\\[|\\033\\[)[0-9;:]*[a-zA-Z]"
                    .qol_messages[["last_message"]] <- gsub(ansi_regex,
                                                            "",
                                                            .qol_messages[["last_message"]],
                                                            perl = TRUE)
                }

                # If an in place step is triggered then don't print a time stamp here,
                # but delete the current line for a fresh start.
                if (in_place){
                    # If there was a timed, non in place message before, print the
                    # time stamp.
                    if (!is_last_message_in_place()){
                        cat("\r", .qol_messages[["last_message"]], "\n", sep = "")
                    }

                    # Delete all text in current line
                    cat("\r", strrep(" ", getOption("width")), "\r", sep = "")
                }
                # For non in place messages print the time stamp
                else{
                    cat("\r", .qol_messages[["last_message"]], "\n", sep = "")
                }
                utils::flush.console()
            }
        }
        # Otherwise do nothing and just let the timer run without reset
    }
    # Start new timer
    else{
        # But only if the message before is something else than start
        if (!is_timer_running()){
            start_timer()
        }
    }

    invisible(NULL)
}


#' @description
#' Strips down the sys calls to reduce memory load.
#'
#' @returns
#' Returns a condensed and deparsed sys.call list.
#'
#' @noRd
# Put time stamp at the end of the last message if present
get_condensed_sys_calls <- function(){
    sys_calls <- sys.calls()

    call_select <- vapply(sys_calls, function(call){
        if (!is.call(call)){
            return(TRUE)
        }

        # Get the front main element of the call
        main_element <- call[[1]]

        # Filter out anonymous functions: "(function...)"
        if (typeof(main_element) %in% c("closure", "builtin", "special")){
            return(FALSE)
        }

        # Extract function name from the call
        if (is.call(main_element)){
            if (identical(main_element[[1]], quote(`(`))){
                return(FALSE)
            }

            function_name <- as.character(main_element[[1]])
        }
        else{
           function_name <- as.character(main_element)
        }

        if (length(function_name) == 0 || is.na(function_name)){
            return(TRUE)
        }

        # Remove functions that potentially cause heavy memory load or are redundant
        if (function_name %in% c("withCallingHandlers", "get_execution_token", "(",
                                 "tryCatch", "doTryCatch", "eval", "source", "local",
                                 "test_package", "test_single_file", "tinytest", "testthat",
                                 "rlang", "run_test_dir", "run_test_dir",
                                 "run_test_file", "lapply", "FUN", "is_message_suppressed",
                                 "get_condensed_sys_calls")){
            return(FALSE)
        }

        prefixes <- c("print_", "::", "run_pkg_hook", "[[", "expect_")

        if (any(outer(function_name, prefixes, startsWith))){
            return(FALSE)
        }

        TRUE
    }, logical(1))

    # Only keep selected calls
    sys_calls <- sys_calls[call_select]

    # Deparse calls
    vapply(sys_calls, function(call) {
        paste(deparse(call, nlines = 1), collapse = "")
    }, character(1))
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
# Printing the output is always set to FALSE, so the overall code can be tested
# but without drawing the whole outputs on screen.
###############################################################################

dummy_df      <- dummy_data(1000)
dummy_wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = c("sex", "education"),
                   values   = income,
                   na.rm    = TRUE,
                   formats  = list(sex = discrete_format(
                       "Total"  = 1:2,
                       "Male"   = 1,
                       "Female" = 2)))


# Simple long to wide transposition
dummy_df <- dummy_df |> sort_plus(by = sex)

wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = "sex",
                   values   = income,
                   weight   = weight)

expect_equal(names(wide_df), c("year", "1", "2", NA), info = "Simple long to wide transposition")


# Side by side long to wide transposition
dummy_df <- dummy_df |> sort_plus(by = c(sex, education))

wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = c("sex", "education"),
                   values   = income,
                   na.rm    = TRUE)

expect_equal(names(wide_df), c("year", "1", "2", "high", "low", "middle"), info = "Side by side long to wide transposition")


# Nested long to wide transposition
dummy_df <- dummy_df |> sort_plus(by = c(sex, education))

wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = "sex + education",
                   values   = income,
                   na.rm    = TRUE)

expect_equal(names(wide_df), c("year", "1_high", "1_low", "1_middle",
                                       "2_high", "2_low", "2_middle"), info = "Nested long to wide transposition")


# Transpose multiple value variables
dummy_df <- dummy_df |> sort_plus(by = sex)
wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = "sex",
                   values   = c(income, weight),
                   na.rm    = TRUE)

expect_equal(names(wide_df), c("year", "income_1", "income_2", "weight_1", "weight_2"), info = "Transpose multiple value variables")


# Using formats in long to wide transposition
wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = "sex",
                   values   = income,
                   na.rm    = TRUE,
                   formats  = list(sex =
                       discrete_format("Total"  = 1:2,
                                       "Male"   = 1,
                                       "Female" = 2)))

expect_equal(names(wide_df), c("year", "Total", "Male", "Female"), info = "Using formats in long to wide transposition")


# Simple wide to long transposition
wide_to_long <- dummy_wide_df |>
        transpose_plus(preserve = year,
                       pivot    = list(sex = c("Total", "Male", "Female")))

expect_equal(names(wide_to_long), c("year", "sex", "VALUE"), info = "Simple wide to long transposition")
expect_equal(collapse::funique(wide_to_long[["sex"]]), c("Total", "Male", "Female"), info = "Simple wide to long transposition")


# Transpose multiple variables from wide to long
wide_to_long <- dummy_wide_df |>
         transpose_plus(preserve = year,
                        pivot    = list(sex       = c("Male", "Female"),
                                        education = c("low", "middle", "high")),
                        formats  = list(sex =
                            discrete_format("Total"  = c("Male", "Female"),
                                            "Male"   = "Male",
                                            "Female" = "Female")))

expect_equal(names(wide_to_long), c("year", "BY", "VARIABLE", "VALUE"), info = "Transpose multiple variables from wide to long")
expect_equal(as.character(collapse::funique(wide_to_long[["VARIABLE"]])), c("Total", "Male", "Female", "high", "low", "middle"),
             info = "Simple wide to long transposition")


# Transpose multiple variables from wide to long (side by side)
wide_to_long <- dummy_wide_df |>
    transpose_plus(preserve = year,
                   pivot    = list(sex = c("Total", "Male", "Female"),
                                   sex = c("low", "middle", "high")))

expect_equal(names(wide_to_long), c("year", "sex", "value1", "value2"),
             info = "Transpose multiple variables from wide to long (side by side)")

wide_to_long <- dummy_wide_df |>
    transpose_plus(preserve = year,
                   pivot    = list(sex = c("Total", "Male", "Female"),
                                   sex = c("low", "middle", "high")),
                   values   = c(hello, world))

expect_equal(names(wide_to_long), c("year", "sex", "hello", "world"),
             info = "Transpose multiple variables from wide to long (side by side)")



###############################################################################
# Warning checks
###############################################################################

# Wide to long transposition doesn't support value parameter transposition
wide_df <- dummy_wide_df |>
    transpose_plus(preserve = year,
                   pivot    = list(sex = c("Male", "Female")),
                   values   = "Total")

expect_message(print_stack_as_messages("NOTE"), "<Values> parameter has no effect in wide to long transposition, when results are stacked.",
			   info = "Wide to long transposition doesn't support value parameter transposition")


# Wide to long transposition doesn't support weight parameter transposition
wide_df <- dummy_wide_df |>
   transpose_plus(preserve = year,
                  pivot    = list(sex = c("Male", "Female")),
                  weight   = "Total")

expect_message(print_stack_as_messages("NOTE"), "<Weight> parameter has no effect in wide to long transposition.",
               info = "Wide to long transposition doesn't support weight parameter transposition")


# Preserve variable in transposition is not part of the data frame
wide_df <- dummy_df |>
   transpose_plus(preserve = "test",
                  pivot    = "sex",
                  values   = income)

expect_warning(print_stack_as_messages("WARNING"), "The provided <preserve> variable",
               info = "Preserve variable in transposition is not part of the data frame")


# Value variable in transposition is also part of preserve
wide_df <- dummy_df |>
   transpose_plus(preserve = sex,
                  pivot    = "age",
                  values   = sex)

expect_warning(print_stack_as_messages("WARNING"), "The provided <values> variable",
               info = "Value variable in transposition is also part of preserve")

###############################################################################
# Abort checks
###############################################################################

# Wide to long transposition doesn't support nesting variables
wide_df <- dummy_wide_df |>
			transpose_plus(preserve = year,
						   pivot    = list(sex = "Male + Female"))

expect_error(print_stack_as_messages("ERROR"), "Nesting <pivot> variables in a wide to long transposition is not possible.",
             info = "Wide to long transposition doesn't support nesting variables")


# Abort transposition if pivot variable is part of preserve
wide_df <- dummy_wide_df |>
			transpose_plus(preserve = year,
						   pivot    = "year",
						   values   = income)

expect_error(print_stack_as_messages("ERROR"), "The provided <pivot> variable",
             info = "Abort transposition if pivot variable is part of preserve")


# Abort transposition if no valid value variable is provided
wide_df <- dummy_df |>
			transpose_plus(preserve = year,
						   pivot    = "sex")

expect_error(print_stack_as_messages("ERROR"), "No <values> provided. Transposition will be aborted.",
             info = "Abort transposition if no valid value variable is provided")


# Abort transposition if value variable is not part of the data frame
wide_df <- dummy_df |>
			transpose_plus(pivot  = "sex",
						   values = test)

expect_error(print_stack_as_messages("ERROR"), "No valid <values> to transpose provided. Transposition will be aborted.",
             info = "Abort transposition if value variable is not part of the data frame")


# Abort if value variable in transposition is also part of pivot
wide_df <- dummy_df |>
			transpose_plus(pivot  = "sex",
						   values = sex)

expect_error(print_stack_as_messages("ERROR"), "The provided <values> variable",
             info = "Abort if value variable in transposition is also part of pivot")


# Abort on duplicate variable names after transposition
wide_df <- dummy_df |>
			transpose_plus(preserve = year,
						   pivot    = c("sex", "education"),
						   values   = income)

expect_error(print_stack_as_messages("ERROR"), "Duplicate column names found:",
             info = "Abort on duplicate variable names after transposition")


# Abort if no valid pivot variable is provided in transposition
wide_df <- dummy_df |>
			transpose_plus(pivot  = "test",
						   values = income)

expect_error(print_stack_as_messages("ERROR"), "The provided <pivot> variable",
             info = "Abort if no valid pivot variable is provided in transposition")


# Abort side by side transposition, if list entries are of unequal lengths
wide_df <- dummy_wide_df |>
    transpose_plus(preserve = year,
                   pivot    = list(sex = c("Total", "Male", "Female", "low"),
                                   sex = c("low", "middle", "high")))

expect_error(print_stack_as_messages("ERROR"), "Every <pivot> list entry has to have the same number of variables for a",
             info = "Abort side by side transposition, if list entries are of unequal lengths")


# Abort side by side transposition, if list entries contain a unique variable name but also others
wide_df <- dummy_wide_df |>
    transpose_plus(preserve = year,
                   pivot    = list(sex  = c("Total", "Male", "Female", "low"),
                                   sex  = c("low", "middle", "high"),
                                   test = c("low", "middle", "high")))

expect_error(print_stack_as_messages("ERROR"), "The new result columns can only be set side by side in a wide",
             info = "Abort side by side transposition, if list entries contain a unique variable name but also others")


set_no_print()

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

###############################################################################
# Setup test data frames
###############################################################################

# One key
df1 <- data.frame(key = c(1, 1, 1, 2, 2, 2),
                  a   = c("a", "a", "a", "a", "a", "a"))

df2 <- data.frame(key = c(2, 3),
                  b   = c("b", "b"))

df3 <- data.frame(key = c(1, 2),
                  c   = c("c", "c"))

df4 <- data.frame(key  = c(1, 2),
                  key1 = c(1, 2),
                  a    = c("d", "d"))

# Multiple same keys
df1b <- data.frame(key1 = c(1, 1, 1, 2, 2, 2),
                   key2 = c("a", "a", "a", "a", "a", "a"),
                   a    = c("a", "a", "a", "a", "a", "a"))

df2b <- data.frame(key1 = c(2, 3),
                   key2 = c("a", "a"),
                   b    = c("b", "b"))

# Multiple different keys
df1c <- data.frame(key1 = c(1, 1, 1, 2, 2, 2),
                   key2 = c("a", "a", "a", "a", "a", "a"),
                   a    = c("a", "a", "a", "a", "a", "a"))

df2c <- data.frame(var1 = c(2, 3),
                   var2 = c("a", "a"),
                   b    = c("b", "b"))

df3c <- data.frame(any  = c(1, 2),
                   name = c("a", "a"),
                   c    = c("c", "c"))

###############################################################################
# Successful joins
###############################################################################

# Test all basic joins with one key
join_methods <- c("left", "right", "inner", "full", "outer", "left_inner", "right_inner")
joined_data  <- list()

for (method in seq_along(join_methods)){
    joined_data[[method]] <-
        multi_join(list(df1, df2),
                   on  = "key",
                   how = join_methods[[method]],
                   keep_indicators = TRUE)
}

expect_equal(length(joined_data), 7, info = "Test all basic joins with one key")


# Join on multiple keys
left_joined <- multi_join(list(df1b, df2b), on = c("key1", "key2"))

expect_true("b" %in% names(left_joined), info = "Join on multiple keys")

left_joined <- multi_join(list(df1b, df2b), on = list(key1, key2))

expect_true("b" %in% names(left_joined), info = "Join on multiple keys")


# Join more than two data frames
multiple_joined <- multi_join(list(df1, df2, df3), on = "key")

expect_true(all(c("b", "c") %in% names(multiple_joined)), info = "Join more than two data frames")


# Join multiple data frames on different variable names
multiple_joined <-
    multi_join(list(df1c, df2c, df3c),
               on = list(df1c = c("key1", "key2"),
                         df2c = c("var1", "var2"),
                         df3c = c("any", "name")))

expect_true(all(c("b", "c") %in% names(multiple_joined)), info = "Join multiple data frames on different variable names")

multiple_joined2 <-
    multi_join(list(df1c, df2c, df3c),
               on = list(df1c = c(key1, key2),
                         df2c = c(var1, var2),
                         df3c = c(any, name)))

expect_equal(multiple_joined, multiple_joined2, info = "Join multiple data frames on different variable names")


# Warning on invalid join method
left_joined <- multi_join(list(df1, df2), on = "key", how = "test")

expect_warning(print_stack_as_messages("WARNING"), "No valid join method provided, 'left' will be used.", info = "Warning on invalid join method")
expect_true("b" %in% names(left_joined), info = "Warning on invalid join method")


# Note on too many join methods provided
left_joined <- multi_join(list(df1, df2),
                                         on  = "key",
                                         how = c("left", "right"))

expect_message(print_stack_as_messages("NOTE"), "Too many join methods given in <how>. Excess methods will remain unused.",
               info = "Note on too many join methods provided")
expect_true("b" %in% names(left_joined), info = "Note on too many join methods provided")


# Warning duplicate variables on join
left_joined <- multi_join(list(df1, df4), on = "key")

expect_warning(print_stack_as_messages("WARNING"), "Duplicate variable names found", info = "Warning duplicate variables on join")
expect_true(collapse::fncol(left_joined) == 3, info = "Warning duplicate variables on join")

left_joined <- multi_join(list(df1, df4), on = list(df1 = "key",
                                                    df4 = "key1"))

expect_warning(print_stack_as_messages("WARNING"), "Duplicate variable names found", info = "Warning duplicate variables on join")
expect_true(collapse::fncol(left_joined) == 3, info = "Warning duplicate variables on join")

###############################################################################
# Abort checks
###############################################################################

# Abort join, if data frames aren't provided as list
left_joined <- multi_join(c(df1, df2), on = "key")

expect_error(print_stack_as_messages("ERROR"), "Data frames must be provided as a list. Join will be aborted.",
             info = "Abort join, if data frames aren't provided as list")


# Abort join, if only one data frames provided
left_joined <- multi_join(list(df1), on = "key")

expect_error(print_stack_as_messages("ERROR"), "At least two data frames are required. Join will be aborted.",
             info = "Abort join, if only one data frames provided")


# Abort join, if <on> variables are provided as incomplete unnamed list
left_joined <- multi_join(list(df1b, df2b), on = list(a = "key1", "key2"))

expect_error(print_stack_as_messages("ERROR"), "If all data frames have the same variable names for the <on> variables",
             info = "Abort join, if <on> variables are provided as incomplete unnamed list")


# Abort join, if <on> variables are not provided for every data frame
multiple_joined <- multi_join(list(df1c, df2c, df3c),
                              on = list(df1c = c("key1", "key2"),
                                        df2c = c("var1", "var2")))

expect_error(print_stack_as_messages("ERROR"), "Length of <on> doesn't match the number of provided data frames. Join will be aborted.",
             info = "Abort join, if <on> variables are not provided for every data frame")


# Abort join, if <on> variables are missing in data frame (equal names)
left_joined <- multi_join(list(df1, df2), on = "var")

expect_error(print_stack_as_messages("ERROR"), "Not all <on> variables", info = "Abort join, if <on> variables are missing in data frame (equal names)")


# Abort join, if <on> variables are missing in data frame (unequal names)
multiple_joined <- multi_join(list(df1c, df2c, df3c),
                              on = list(df1c = c("var1", "key2"),
                                        df2c = c("var1", "var2"),
                                        df3c = c("any", "name")))

expect_error(print_stack_as_messages("ERROR"), "Not all <on> variables", info = "Abort join, if <on> variables are missing in data frame (unequal names)")


# Abort join, if second of following data frame doesn't consist of only unique values (equal names)
left_joined <- multi_join(list(df1, df1), on = "key")

expect_error(print_stack_as_messages("ERROR"), "The second and all following data frames need to have unique combinations",
             info = "Abort join, if second of following data frame doesn't consist of only unique values (equal names)")


# Abort join, if second of following data frame doesn't consist of only unique values  (unequal names)
multiple_joined <- multi_join(list(df1c, df1c, df3c),
                              on = list(df1c = c("key1", "key2"),
                                        df1c = c("key1", "key2"),
                                        df3c = c("any", "name")))

expect_error(print_stack_as_messages("ERROR"), "The second and all following data frames need to have unique combinations",
             info = "Abort join, if second of following data frame doesn't consist of only unique values  (unequal names)")


# Abort join, if <on> variables have unequal length
multiple_joined <- multi_join(list(df1c, df2c, df3c),
                              on = list(df1c = c("key1", "key2"),
                                        df2c = c("var1"),
                                        df3c = c("any", "name")))

expect_error(print_stack_as_messages("ERROR"), "Unequal number of <on> variables provided", info = "Abort join, if <on> variables have unequal length")


set_no_print()

set_no_print(TRUE)

# Test master building
temp_file <- tempfile(fileext = ".Rmd")

build_master(dir = dirname(temp_file),
             master_name = basename(tools::file_path_sans_ext(temp_file)))

expect_true(file.exists(temp_file), info = "Test master building")


# Abort master building on invalid directory
build_master(dir = "Test")
expect_error(print_stack_as_messages("ERROR"), "Directory '", info = "Abort master building on invalid directory")


set_no_print()

set_no_print(TRUE)

# Resolving single macro variable
year <- 2026
text <- macro("The current year is &year")

expect_equal(text, "The current year is 2026", info = "Resolving single macro variable")


# Resolving combination of macro variables
year          <- 2026
some_variable <- "current"
current2026   <- "The current year is"

text <- macro("&&some_variable&year &year")

expect_equal(text, "The current year is 2026", info = "Resolving combination of macro variables")


# Resolving character vector containing macro variables
year          <- 2026
some_variable <- "current"
current2026   <- "The current year is"

char_vector <- c("The current year is &year", "The &some_variable year is &year", "&&some_variable&year &year")

text_vector <- apply_macro(char_vector)

expect_equal(text_vector, c("The current year is 2026",
                            "The current year is 2026",
                            "The current year is 2026"), info = "Resolving character vector containing macro variables")

###############################################################################
# Warning checks
###############################################################################

# Resolving macro variable throws a warning if text is a vector
year <- 2026
text <- macro(c("The current year is &year", "The current year is &year"))

expect_warning(print_stack_as_messages("WARNING"), "<Text> may only be of length one. The first Element will be used.",
               info = "Resolving macro variable throws a warning if text is a vector")

expect_equal(text, "The current year is 2026", info = "Resolving macro variable throws a warning if text is a vector")


# Resolving macro variable throws a warning if macro variable is a vector
year <- c(2026, 2025)
text <- macro("The current year is &year")

expect_warning(print_stack_as_messages("WARNING"), "Macro variable \'&year\' may only be of length one. The first Element will be used.",
               info = "Resolving macro variable throws a warning if macro variable is a vector")

expect_equal(text, "The current year is 2026", info = "Resolving macro variable throws a warning if macro variable is a vector")


# Resolving macro variable throws a warning if macro variable is neither character nor numeric
year <- TRUE
text <- macro("The current year is &year")

expect_warning(print_stack_as_messages("WARNING"), "Macro variable \'&year\' is a complex object. Macro variables may only be character or numeric.",
               info = "Resolving macro variable throws a warning if macro variable is neither character nor numeric")

expect_equal(text, "The current year is year", info = "Resolving macro variable throws a warning if macro variable is neither character nor numeric")

###############################################################################
# Abort checks
###############################################################################

# Resolving macro aborts if something other than a text is provided
text <- macro(1)

expect_error(print_stack_as_messages("ERROR"), "<Text> must be a character. Macro will be aborted.",
             info = "Resolving macro aborts if something other than a text is provided")


set_no_print()

set_no_print(TRUE)

test_env <- function(){
    envir <- new.env(parent = emptyenv())

    envir[["numeric1"]]   <- 1
    envir[["numeric2"]]   <- 2
    envir[["character"]]  <- "text"
    envir[["data_frame"]] <- dummy_data(10)
    envir[["another_frame"]] <- dummy_data(10)

    envir
}


# Free memory by names
env <- test_env()

removed <- free_memory(c("numeric1", "character"), envir = env)

expect_false(exists("numeric1",  envir = env), info = "Free memory by names")
expect_true(exists("numeric2",   envir = env), info = "Free memory by names")
expect_false(exists("character", envir = env), info = "Free memory by names")
expect_true(exists("data_frame", envir = env), info = "Free memory by names")

expect_equal(removed, c("numeric1", "character"), info = "Free memory by names")


# Free memory by object types
env <- test_env()

removed <- free_memory(types = "numeric", envir = env)

expect_false(exists("numeric1",  envir = env), info = "Free memory by object types")
expect_false(exists("numeric2",  envir = env), info = "Free memory by object types")
expect_true(exists("character",  envir = env), info = "Free memory by object types")
expect_true(exists("data_frame", envir = env), info = "Free memory by object types")

expect_equal(sort(removed), sort(c("numeric1", "numeric2")), info = "Free memory by object types")


# Free memory by names and object types
env <- test_env()

removed <- free_memory("numeric1", "numeric", envir = env)

expect_false(exists("numeric1",  envir = env), info = "Free memory by names and object types")
expect_true(exists("numeric2",   envir = env), info = "Free memory by names and object types")
expect_true(exists("character",  envir = env), info = "Free memory by names and object types")
expect_true(exists("data_frame", envir = env), info = "Free memory by names and object types")

expect_equal(removed, "numeric1", info = "Free memory by names and object types")


# Free memory by name ranges
env1 <- test_env()
env2 <- test_env()
env3 <- test_env()

removed1 <- free_memory("num:",   envir = env1)
removed2 <- free_memory(":frame", envir = env2)
removed3 <- free_memory(":a:",    envir = env3)

expect_equal(removed1, c("numeric1", "numeric2"), info = "Free memory by name ranges")
expect_equal(removed2, c("another_frame", "data_frame"), info = "Free memory by name ranges")
expect_equal(removed3, c("another_frame", "character", "data_frame"), info = "Free memory by name ranges")

###############################################################################
# Abort checks
###############################################################################

# Free memory aborts, if names are not characters
free_memory(names = 123, envir = test_env())
expect_error(print_stack_as_messages("ERROR"), "<Names> have to be provided as character. Objects will remain untouched.", info = "Free memory aborts, if names are not characters")


# Free memory aborts, if types are not characters
free_memory(types = 123, envir = test_env())
expect_error(print_stack_as_messages("ERROR"), "<Types> have to be provided as character. Objects will remain untouched.", info = "Free memory aborts, if types are not characters")


# Free memory aborts, if name has to many columns
free_memory(names = ":a:b:", envir = test_env())
expect_error(print_stack_as_messages("ERROR"), "<Name> has more than two colons which is not allowed. Objects will remain untouched.", info = "Free memory aborts, if name has to many columns")


# Free memory aborts, if no object is found
free_memory(names = "test", envir = test_env())
expect_error(print_stack_as_messages("ERROR"), "No object found. Objects will remain untouched.", info = "Free memory aborts, if no object is found")


set_no_print()

set_no_print(TRUE)

# Test theme building
temp_file <- tempfile(fileext = ".rstheme")
on.exit(unlink(temp_file), add = TRUE)

build_rstheme(file_path  = dirname(temp_file),
              theme_name = basename(tools::file_path_sans_ext(temp_file)))

expect_true(file.exists(temp_file), info = "Test theme building")


# Abort theme building on invalid directory
build_rstheme(file_path = "Test")

expect_error(print_stack_as_messages("ERROR"), "Directory '", info = "Abort theme building on invalid directory")


set_no_print()

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

test_df <- data.frame(var1 = c("a", "a", "a", "b", "b"),
                      var2 = c(1, 2, 3, 4, 5),
                      var3 = c(10, 20, 30, 40, 50))


# Compute evaluates simple expressions
result_df <- test_df |> compute.(sum = var2 + var3)

expected <- test_df[["var2"]] + test_df[["var3"]]

expect_equal(result_df[["sum"]], expected, info = "Compute evaluates simple expressions")


# Compute supports evaluation of functions
result_df <- test_df |> compute.(col_sum = collapse::fsum(var2))

expected <- collapse::fsum(test_df[["var2"]])

expect_equal(collapse::funique(result_df[["col_sum"]]), expected, info = "Compute supports evaluation of functions")


# Compute supports qol functions
result_df <- test_df |> compute.(row_sum = row_calculation("sum", var2, var3))

expected <- test_df[["var2"]] + test_df[["var3"]]

expect_equal(result_df[["row_sum"]], expected, info = "Compute supports qol functions")


# Compute supports value assignment
result_df <- test_df |> compute.(var4 = 1, var5 = "Hello")

expect_equal(collapse::funique(result_df[["var4"]]), 1, info = "Compute supports value assignment")
expect_equal(collapse::funique(result_df[["var5"]]), "Hello", info = "Compute supports value assignment")


# Compute respects do_if filtering
result_df <- test_df |>
    do_if(var1 == "a") |>
        compute.(sum = var2 + var3) |>
    end_do()

expected <- test_df[["var2"]] + test_df[["var3"]]
expected[test_df[["var1"]] != "a"] <- NA

expect_equal(result_df[["sum"]], expected, info = "Compute respects do_if filtering")


# Compute overwrites existing variable
result_df <- test_df |> compute.(var2 = var3)

expect_true(!identical(result_df[["var2"]], test_df[["var2"]]), info = "Compute overwrites existing variable")
expect_equal(result_df[["var2"]], result_df[["var3"]], info = "Compute overwrites existing variable")


# Compute as do over loop
variables  <- c("NEW_VAR1", "NEW_VAR2")
values     <- c(1, 2)

do_over_df <- test_df |> compute.(variables = values)

expect_true(all(c("NEW_VAR1", "NEW_VAR2") %in% names(do_over_df)), info = "Compute as do over loop")
expect_true(collapse::funique(do_over_df[["NEW_VAR1"]]) == 1, info = "Compute as do over loop")
expect_true(collapse::funique(do_over_df[["NEW_VAR2"]]) == 2, info = "Compute as do over loop")


# Compute handles multiple assignments
variables  <- c("NEW_VAR1", "NEW_VAR2")
values     <- c(1, 2)

result_df <- test_df |> compute.(sum       = var2 + var3,
                                 col_sum   = collapse::fsum(var2),
                                 row_sum   = row_calculation("sum", var2, var3),
                                 var4      = 1,
                                 var5      = "Hello",
                                 variables = values)

expect_true(all(c("sum", "col_sum", "row_sum", "var4", "var5", "NEW_VAR1", "NEW_VAR2") %in% names(result_df)),
            info = "Compute handles multiple assignments")


# Compute evaluates calls sequentially
result_df <- test_df |> compute.(sum1 = var2 + var3,
                                 sum2 = sum1 + var2)

expected <- test_df[["var2"]] * 2 + test_df[["var3"]]

expect_equal(result_df[["sum2"]], expected, info = "Compute evaluates calls sequentially")


# Compute evaluates list passed as argument
symbol <- list(sum = "var2 + var3")

result_df <- test_df |> compute.(symbol)

expected <- test_df[["var2"]] + test_df[["var3"]]

expect_equal(result_df[["sum"]], expected, info = "Compute evaluates list passed as argument")

###############################################################################
# Warning checks
###############################################################################

# Type mismatch in compute
result_df <- test_df |> compute.(var1 = var2)

expect_warning(print_stack_as_messages("WARNING"), "Type mismatch", info = "Type mismatch in compute")


# Adding multiple variables of the same name in compute throws a warning
result_df <- test_df |> compute.(var4 = 1, var4 = 2)

expect_warning(print_stack_as_messages("WARNING"), "Duplicate variable name",
               info = "Adding multiple variables of the same name in compute throws a warning")
expect_true("var4" %in% names(result_df))
expect_equal(collapse::funique(result_df[["var4"]]), 1)

###############################################################################
# Abort checks
###############################################################################

# Compute aborts with no assignment
result_df <- test_df |> compute.()

expect_error(print_stack_as_messages("ERROR"), "No assignments. Evaluation will be aborted.", info = "Compute aborts with no assignment")


# Compute aborts with no assignment
variables  <- c("NEW_VAR1", "NEW_VAR2", "NEW_VAR3")
values     <- c(1, 2)

do_over_df <- test_df |> compute.(variables = values)

expect_error(print_stack_as_messages("ERROR"), "Passed vectors are of unequal lengths.", info = "Compute aborts with no assignment")


set_no_print()

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
# Printing the output is always set to FALSE, so the overall code can be tested
# but without drawing the whole outputs on screen.
###############################################################################

small_df <- data.frame(var1 = c(3, 2, 1, 2),
                       var2 = c("c", "a", "a", "b"))


# Simple numeric sort
sort_df <- small_df |> sort_plus(by = var1)

expect_equal(sort_df[["var1"]], c(1, 2, 2, 3), info = "Simple numeric sort")


# Simple character sort
sort_df <- small_df |> sort_plus(by = var2)

expect_equal(sort_df[["var2"]], c("a", "a", "b", "c"), info = "Simple character sort")


# Simple numeric sort, descending
sort_df <- small_df |> sort_plus(by = var1, order = "descending")

expect_equal(sort_df[["var1"]], c(3, 2, 2, 1), info = "Simple numeric sort, descending")


# Sort with preserve
sort_df <- small_df |> sort_plus(by = var1, preserve = var2)

expect_equal(sort_df[["var1"]], c(3, 1, 2, 2), info = "Sort with preserve")
expect_equal(sort_df[["var2"]], c("c", "a", "a", "b"), info = "Sort with preserve")


# Sort with format
sort_df <- small_df |>
    sort_plus(by = c(var1, var2),
              formats = list(var2 =
                  discrete_format("1" = "b",
                                  "2" = "c",
                                  "3" = "a")))

expect_equal(sort_df[["var1"]], c(1, 2, 2, 3), info = "Sort with format")
expect_equal(sort_df[["var2"]], c("a", "b", "a", "c"), info = "Sort with format")


# Sort normal and with format
sort_df <- small_df |>
    sort_plus(by = var2,
              formats = list(var2 =
                                 discrete_format("1" = "a",
                                                 "2" = "c",
                                                 "3" = "b")))

expect_equal(sort_df[["var2"]], c("a", "a", "c", "b"), info = "Sort normal and with format")

###############################################################################
# Warning checks
###############################################################################

# Sort warning with invalid by variable
sort_df <- small_df |> sort_plus(by = var3)

expect_warning(print_stack_as_messages("WARNING"), "The provided <by> variable", info = "Sort warning with invalid by variable")


# Sort warning when by variable is also part of preserve, sort normal
sort_df <- small_df |> sort_plus(by = var1, preserve = var1)

expect_warning(print_stack_as_messages("WARNING"), "The provided <by> variable", info = "Sort warning when by variable is also part of preserve, sort normal")


# Sort warning when wrong order is specified, sorting in ascending order instead
sort_df <- small_df |> sort_plus(by = var1, order = "test")

expect_warning(print_stack_as_messages("WARNING"), "<Order> other than 'ascending'/'a' or 'descending'/'d' specified, which is", info = "Sort warning when wrong order is specified, sorting in ascending order instead")

expect_equal(sort_df[["var1"]], c(1, 2, 2, 3), info = "Sort warning when wrong order is specified, sorting in ascending order instead")


# Sort warning when trying to apply a multilabel, sort normal
sort_df <- small_df |>
	   sort_plus(by = var1,
				 formats = list(age =
								discrete_format("Total"          = 1:3,
												"Number 1 and 2" = 1:2,
												"Number 3"       = 3)))

expect_warning(print_stack_as_messages("WARNING"), "Applying a multilabel to variable", info = "Sort warning when trying to apply a multilabel, sort normal")

expect_equal(sort_df[["var1"]], c(1, 2, 2, 3), info = "Sort warning when trying to apply a multilabel, sort normal")


# Sort warning when trying to apply a format to a variable which is not in the data frame, sort normal
sort_df <- small_df |>
	   sort_plus(by = var1,
				 formats = list(var3 =
					 discrete_format("under 18"     = 0:17,
									 "18 and older" = 18:100)))

expect_warning(print_stack_as_messages("WARNING"), "The variable 'var3' is not part of the data frame.", info = "Sort warning when trying to apply a format to a variable which is not in the data frame, sort normal")

expect_equal(sort_df[["var1"]], c(1, 2, 2, 3), info = "Sort warning when trying to apply a format to a variable which is not in the data frame, sort normal")


# Sort warning when trying to apply a format to a variable which is not part of the by variables, sort normal
sort_df <- small_df |>
	   sort_plus(by = var1,
				 formats = list(var2 =
					 discrete_format("under 18"     = 0:17,
									 "18 and older" = 18:100)))

expect_message(print_stack_as_messages("NOTE"), "Format for variable", info = "Sort warning when trying to apply a format to a variable which is not part of the by variables, sort normal")

expect_equal(sort_df[["var1"]], c(1, 2, 2, 3), info = "Sort warning when trying to apply a format to a variable which is not part of the by variables, sort normal")


###############################################################################
# Abort checks
###############################################################################

# Sort aborts without by variables
sort_df <- small_df |> sort_plus()

expect_error(print_stack_as_messages("ERROR"), "No <by> variables provided. Sorting will be aborted.", info = "Sort aborts without by variables")


set_no_print()

set_no_color()

# Print normal messages utf8
test_func <- function(){
    print_start_message()
    print_message("NOTE",    "This is a note.",    utf8 = TRUE)
    print_message("WARNING", "This is a warning.", utf8 = TRUE)
    print_message("ERROR",   "This is an error.",  utf8 = TRUE)
    print_message("NEUTRAL", "This is neutral.",   utf8 = TRUE)
    print_closing()
}

test_func()

expect_equal(length(get_message_stack()), 6, info = "Print normal messages utf8")
expect_message(print_stack_as_messages("NOTE"),    "This is a note.",    info = "Print normal messages utf8")
expect_warning(print_stack_as_messages("WARNING"), "This is a warning.", info = "Print normal messages utf8")
expect_error(print_stack_as_messages("ERROR"),     "This is an error.",  info = "Print normal messages utf8")
expect_message(print_stack_as_messages("NEUTRAL"),  "This is neutral.",  info = "Print normal messages utf8")


# Print normal messages non-utf8
test_func <- function(){
    print_start_message()
    print_message("NOTE",    "This is a note.",    utf8 = FALSE)
    print_message("WARNING", "This is a warning.", utf8 = FALSE)
    print_message("ERROR",   "This is an error.",  utf8 = FALSE)
    print_message("NEUTRAL", "This is neutral.",   utf8 = FALSE)
    print_closing()
}

test_func()

expect_equal(length(get_message_stack()), 6, info = "Print normal messages non-utf8")
expect_message(print_stack_as_messages("NOTE"),    "This is a note.",    info = "Print normal messages non-utf8")
expect_warning(print_stack_as_messages("WARNING"), "This is a warning.", info = "Print normal messages non-utf8")
expect_error(print_stack_as_messages("ERROR"),     "This is an error.",  info = "Print normal messages non-utf8")
expect_message(print_stack_as_messages("NEUTRAL"), "This is neutral.",   info = "Print normal messages non-utf8")


# Print step messages utf8
test_func <- function(){
    print_start_message()
    print_headline("This is a headline.", utf8 = TRUE)
    print_step("GREY",  "Grey step.",     utf8 = TRUE)
    print_step("MAJOR", "Major step.",    utf8 = TRUE)
    print_step("MINOR", "Minor step.",    utf8 = TRUE)
    print_closing()
}

test_func()

expect_equal(length(get_message_stack()), 6, info = "Print step messages utf8")
expect_message(print_stack_as_messages("GREY"),  "Grey step.",  info = "Print step messages utf8")
expect_message(print_stack_as_messages("MAJOR"), "Major step.", info = "Print step messages utf8")
expect_message(print_stack_as_messages("MINOR"), "Minor step.", info = "Print step messages utf8")


# Print step messages non-utf8
test_func <- function(){
    print_start_message()
    print_headline("This is a headline.", utf8 = FALSE)
    print_step("GREY",  "Grey step.",     utf8 = FALSE)
    print_step("MAJOR", "Major step.",    utf8 = FALSE)
    print_step("MINOR", "Minor step.",    utf8 = FALSE)
    print_closing()
}

test_func()

expect_equal(length(get_message_stack()), 6, info = "Print step messages non-utf8")
expect_message(print_stack_as_messages("GREY"),  "Grey step.",  info = "Print step messages non-utf8")
expect_message(print_stack_as_messages("MAJOR"), "Major step.", info = "Print step messages non-utf8")
expect_message(print_stack_as_messages("MINOR"), "Minor step.", info = "Print step messages non-utf8")


# Print custom messages utf8
set_up_custom_message(text_bold      = TRUE,
                      text_italic    = TRUE,
                      text_underline = TRUE,
                      text_color     = "#FF00FF")

test_func <- function(){
    print_start_message()
    print_message("unicorn", "This is a unicorn message.", utf8 = TRUE)
    print_step("unicorn",    "This is a unicorn step.",    utf8 = TRUE)
    print_closing()
}

test_func()

expect_equal(length(get_message_stack()), 4, info = "Print custom messages utf8")
expect_message(print_stack_as_messages("UNICORN"), "This is a unicorn message.", info = "Print custom messages utf8")
expect_message(print_stack_as_messages("UNICORN"), "This is a unicorn step.",    info = "Print custom messages utf8")


# Print custom messages non-utf8
set_up_custom_message()

test_func <- function(){
    print_start_message()
    print_message("unicorn", "This is a unicorn message.", utf8 = FALSE)
    print_step("unicorn",    "This is a unicorn step.",    utf8 = FALSE)
    print_closing()
}

test_func()

expect_equal(length(get_message_stack()), 4, info = "Print custom messages non-utf8")
expect_message(print_stack_as_messages("UNICORN"), "This is a unicorn message.", info = "Print custom messages non-utf8")
expect_message(print_stack_as_messages("UNICORN"), "This is a unicorn step.",    info = "Print custom messages non-utf8")


# Print in place messages
test_func <- function(){
    print_start_message()

    for (i in seq_len(10)){
        print_step("Minor", "This is in place step [i] of 10", i = i, in_place = TRUE)
    }

    print_closing()
}

test_func()

expect_equal(length(get_message_stack()), 12, info = "Print in place messages")
expect_message(print_stack_as_messages("MINOR"), "This is in place step 1 of 10",  info = "Print custom messages non-utf8")
expect_message(print_stack_as_messages("MINOR"), "This is in place step 10 of 10", info = "Print custom messages non-utf8")


# Print messages without start and end only return a message stack of length 1
test_func <- function(){
    print_message("NOTE",    "This is a note.",    utf8 = TRUE)
    print_message("WARNING", "This is a warning.", utf8 = TRUE)
    print_message("ERROR",   "This is an error.",  utf8 = TRUE)
}

test_func()

expect_equal(length(get_message_stack()), 1, info = "Print messages without start and end only reutnr a message stack of length 1")


# Print closing without start creates a fake start
test_func <- function(){
    print_message("NEUTRAL", "This is neutral.")
    print_closing()
}

test_func()

expect_equal(length(get_message_stack()), 3, info = "Print closing without start creates a fake start")


# Message stack resets after new function call
test_func <- function(){
    print_start_message()
    print_message("NEUTRAL", "This is neutral.")
    print_closing()
}

test_func()
test_func()

expect_equal(length(get_message_stack()), 3, info = "Message stack resets after new function call")


# Styling messages
print_message("NEUTRAL", "[b][i][u][#FF00FF This is styled.][/u][/i][/b].")

expect_message(print_stack_as_messages("NEUTRAL"), "This is styled.", info = "Styling messages")


# Printing message in singular
vars <- "var1"

print_message("NEUTRAL", "[?This/These] [?is a/are] variable[?s]: [vars].", vars = vars)

expect_message(print_stack_as_messages("NEUTRAL"), "This is a variable: var1.", info = "Printing message in singular")


# Printing message in plural
vars <- c("var1", "var2")

print_message("NEUTRAL", "[?This/These] [?is a/are] variable[?s]: [vars].", vars = vars)

expect_message(print_stack_as_messages("NEUTRAL"), "These are variables: var1, var2.", info = "Printing message in singular")


# Get message stack as data frame
expect_inherits(get_message_stack(as_data_frame = TRUE), "data.frame", info = "Get message stack as data frame")

set_no_color(FALSE)

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

test_df <- dummy_data(10)

###############################################################################
# Keep
###############################################################################

# Different way of passing variables in keep
keep_df1 <- test_df |> keep(year)
keep_df2 <- test_df |> keep("year")

expect_identical(keep_df1, keep_df2, info = "Different way of passing variables in keep")


# Keep only one variable
keep_df <- test_df |> keep(year)

expect_equal(ncol(keep_df), 1, info = "Keep only one variable")

expect_true("year" %in% names(keep_df), info = "Keep only one variable")

expect_identical(keep_df[["year"]], test_df[["year"]], info = "Keep only one variable")


# Keep more than one variable
keep_df <- test_df |> keep(year, age, sex)

expect_equal(ncol(keep_df), 3, info = "Keep more than one variable")

expect_true(all(c("year", "sex", "age") %in% names(keep_df)), info = "Keep more than one variable")

expect_identical(keep_df[["year"]], test_df[["year"]], info = "Keep more than one variable")
expect_identical(keep_df[["age"]], test_df[["age"]], info = "Keep more than one variable")
expect_identical(keep_df[["sex"]], test_df[["sex"]], info = "Keep more than one variable")


# Keep range of variables
keep_df <- test_df |> keep(age:education)

expect_equal(ncol(keep_df), 3, info = "Keep range of variables")

expect_true(all(c("age", "sex", "education") %in% names(keep_df)), info = "Keep range of variables")

expect_identical(keep_df[["education"]], test_df[["education"]], info = "Keep range of variables")
expect_identical(keep_df[["age"]], test_df[["age"]], info = "Keep range of variables")


# Keep variables starting with letter
keep_df <- test_df |> keep("s:")

expect_equal(ncol(keep_df), 2, info = "Keep variables starting with letter")

expect_true(all(c("state", "sex") %in% names(keep_df)), info = "Keep variables starting with letter")


# Keep variables ending with letter
keep_df <- test_df |> keep(":id")

expect_equal(ncol(keep_df), 2, info = "Keep variables ending with letter")

expect_true(all(c("household_id", "person_id") %in% names(keep_df)), info = "Keep variables ending with letter")


# Keep variables containing letter
keep_df <- test_df |> keep(":on:")

expect_equal(ncol(keep_df), 4, info = "Keep variables containing letter")

expect_true(all(c("person_id", "first_person", "education") %in% names(keep_df)), info = "Keep variables containing letter")


# Variables to keep contain a variable name that isn't part of the data frame
keep_df1 <- test_df |> keep(year, age, sex, cats, dogs)
expect_warning(print_stack_as_messages("WARNING"), "The provided variable to keep", info = "Variables to keep contain a variable name that isn't part of the data frame")

keep_df2 <- test_df |> keep(year, age, sex)

expect_equal(ncol(keep_df1), 3, info = "Variables to keep contain a variable name that isn't part of the data frame")
expect_equal(ncol(keep_df2), 3, info = "Variables to keep contain a variable name that isn't part of the data frame")

expect_identical(keep_df1, keep_df2, info = "Variables to keep contain a variable name that isn't part of the data frame")

expect_true(all(c("year", "sex", "age") %in% names(keep_df1)), info = "Variables to keep contain a variable name that isn't part of the data frame")
expect_true(!all(c("cats", "dogs") %in% names(keep_df1)), info = "Variables to keep contain a variable name that isn't part of the data frame")


# Keep only variables that are not part of the data frame
keep_df <- test_df |> keep(cats, dogs)

expect_warning(print_stack_as_messages("WARNING"), "The provided variable to keep", info = "Keep only variables that are not part of the data frame")

expect_true(nrow(keep_df) == 0, info = "Keep only variables that are not part of the data frame")
expect_true(ncol(keep_df) == 0, info = "Keep only variables that are not part of the data frame")

expect_true(!all(c("cats", "dogs") %in% names(keep_df)), info = "Keep only variables that are not part of the data frame")


# Keep without any variables provided
keep_df <- test_df |> keep()

expect_identical(keep_df, test_df, info = "Keep without any variables provided")


# Keep with sorted variables
unsorted <- test_df |> keep(weight, age)
sorted   <- test_df |> keep(weight, age, order_vars = TRUE)

expect_equal(names(unsorted)[1], "age", info = "Keep with sorted variables")
expect_equal(names(sorted)[1], "weight", info = "Keep with sorted variables")

###############################################################################
# Drop
###############################################################################

# Different way of passing variables in drop
drop_df1 <- test_df |> dropp(year)
drop_df2 <- test_df |> dropp("year")

expect_identical(drop_df1, drop_df2, info = "Different way of passing variables in drop")


# Drop only one variable
drop_df <- test_df |> dropp(year)

expect_equal(ncol(drop_df), ncol(test_df) - 1, info = "Drop only one variable")

expect_true(!"year" %in% names(drop_df), info = "Drop only one variable")


# Drop more than one variable
drop_df <- test_df |> dropp(year, age, sex)

expect_equal(ncol(drop_df), ncol(test_df) - 3, info = "Drop more than one variable")

expect_true(!all(c("year", "sex", "age") %in% names(drop_df)), info = "Drop more than one variable")


# Drop range of variables
drop_df <- test_df |> dropp(age:education)

expect_equal(ncol(drop_df), ncol(test_df) - 3, info = "Drop range of variables")

expect_true(!all(c("education", "sex", "age") %in% names(drop_df)), info = "Drop range of variables")


# Drop variables starting with letter
drop_df <- test_df |> dropp("s:")

expect_true(!all(c("state", "sex") %in% names(drop_df)), info = "Drop variables starting with letter")


# Drop variables ending with letter
drop_df <- test_df |> dropp(":id")

expect_true(!all(c("household_id", "person_id") %in% names(drop_df)), info = "Drop variables ending with letter")


# Drop variables containing letter
drop_df <- test_df |> dropp(":on:")

expect_true(!all(c("person_id", "first_person", "education") %in% names(drop_df)), info = "Drop variables containing letter")


# Variables to drop contain a variable name that isn't part of the data frame
drop_df1 <- test_df |> dropp(year, age, sex, cats, dogs)

expect_warning(print_stack_as_messages("WARNING"), "The provided variable to drop", info = "Variables to drop contain a variable name that isn't part of the data frame")
drop_df2 <- test_df |> dropp(year, age, sex)

expect_equal(ncol(drop_df1), ncol(test_df) - 3, info = "Variables to drop contain a variable name that isn't part of the data frame")
expect_equal(ncol(drop_df2), ncol(test_df) - 3, info = "Variables to drop contain a variable name that isn't part of the data frame")

expect_identical(drop_df1, drop_df2, info = "Variables to drop contain a variable name that isn't part of the data frame")

expect_true(!all(c("year", "sex", "age") %in% names(drop_df1)), info = "Variables to drop contain a variable name that isn't part of the data frame")
expect_true(!all(c("cats", "dogs") %in% names(test_df)), info = "Variables to drop contain a variable name that isn't part of the data frame")


# Drop only variables that are not part of the data frame
drop_df <- test_df |> dropp(cats, dogs)

expect_warning(print_stack_as_messages("WARNING"), "The provided variable to drop", info = "Drop only variables that are not part of the data frame")

expect_identical(drop_df, test_df, info = "Drop only variables that are not part of the data frame")


# Drop without any variables provided
drop_df <- test_df |> dropp()

expect_identical(drop_df, test_df, info = "Drop without any variables provided")


set_no_print()

set_no_print(TRUE)

test_df <- data.frame(var1 = 1:5,
                      var2 = c(6, 7, 8, NA, 10),
                      var3 = 11:15)

# Row wise calculation
result <- test_df |> row_calculation("sum",  var1, var2, var3)

expect_equal(result, c(18, 21, 24, 18, 30), info = "Row wise calculation")


set_no_print()

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
# Printing the output is always set to FALSE, so the overall code can be tested
# but without drawing the whole outputs on screen.
###############################################################################

dummy_df      <- dummy_data(1000)
dummy_wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = c("sex", "education"),
                   values   = income,
                   na.rm    = TRUE,
                   formats  = list(sex = discrete_format(
                       "Total"  = 1:2,
                       "Male"   = 1,
                       "Female" = 2)))


# Simple long to wide transposition
dummy_df <- dummy_df |> sort_plus(by = sex)

wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = "sex",
                   values   = income,
                   weight   = weight)

expect_equal(names(wide_df), c("year", "1", "2", NA), info = "Simple long to wide transposition")


# Side by side long to wide transposition
dummy_df <- dummy_df |> sort_plus(by = c(sex, education))

wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = c("sex", "education"),
                   values   = income,
                   na.rm    = TRUE)

expect_equal(names(wide_df), c("year", "1", "2", "high", "low", "middle"), info = "Side by side long to wide transposition")


# Nested long to wide transposition
dummy_df <- dummy_df |> sort_plus(by = c(sex, education))

wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = "sex + education",
                   values   = income,
                   na.rm    = TRUE)

expect_equal(names(wide_df), c("year", "1_high", "1_low", "1_middle",
                                       "2_high", "2_low", "2_middle"), info = "Nested long to wide transposition")


# Transpose multiple value variables
dummy_df <- dummy_df |> sort_plus(by = sex)
wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = "sex",
                   values   = c(income, weight),
                   na.rm    = TRUE)

expect_equal(names(wide_df), c("year", "income_1", "income_2", "weight_1", "weight_2"), info = "Transpose multiple value variables")


# Using formats in long to wide transposition
wide_df <- dummy_df |>
    transpose_plus(preserve = year,
                   pivot    = "sex",
                   values   = income,
                   na.rm    = TRUE,
                   formats  = list(sex =
                       discrete_format("Total"  = 1:2,
                                       "Male"   = 1,
                                       "Female" = 2)))

expect_equal(names(wide_df), c("year", "Total", "Male", "Female"), info = "Using formats in long to wide transposition")


# Simple wide to long transposition
wide_to_long <- dummy_wide_df |>
        transpose_plus(preserve = year,
                       pivot    = list(sex = c("Total", "Male", "Female")))

expect_equal(names(wide_to_long), c("year", "sex", "VALUE"), info = "Simple wide to long transposition")
expect_equal(collapse::funique(wide_to_long[["sex"]]), c("Total", "Male", "Female"), info = "Simple wide to long transposition")


# Transpose multiple variables from wide to long
wide_to_long <- dummy_wide_df |>
         transpose_plus(preserve = year,
                        pivot    = list(sex       = c("Male", "Female"),
                                        education = c("low", "middle", "high")),
                        formats  = list(sex =
                            discrete_format("Total"  = c("Male", "Female"),
                                            "Male"   = "Male",
                                            "Female" = "Female")))

expect_equal(names(wide_to_long), c("year", "BY", "VARIABLE", "VALUE"), info = "Transpose multiple variables from wide to long")
expect_equal(as.character(collapse::funique(wide_to_long[["VARIABLE"]])), c("Total", "Male", "Female", "high", "low", "middle"),
             info = "Simple wide to long transposition")


# Transpose multiple variables from wide to long (side by side)
wide_to_long <- dummy_wide_df |>
    transpose_plus(preserve = year,
                   pivot    = list(sex = c("Total", "Male", "Female"),
                                   sex = c("low", "middle", "high")))

expect_equal(names(wide_to_long), c("year", "sex", "value1", "value2"),
             info = "Transpose multiple variables from wide to long (side by side)")

wide_to_long <- dummy_wide_df |>
    transpose_plus(preserve = year,
                   pivot    = list(sex = c("Total", "Male", "Female"),
                                   sex = c("low", "middle", "high")),
                   values   = c(hello, world))

expect_equal(names(wide_to_long), c("year", "sex", "hello", "world"),
             info = "Transpose multiple variables from wide to long (side by side)")



###############################################################################
# Warning checks
###############################################################################

# Wide to long transposition doesn't support value parameter transposition
wide_df <- dummy_wide_df |>
    transpose_plus(preserve = year,
                   pivot    = list(sex = c("Male", "Female")),
                   values   = "Total")

expect_message(print_stack_as_messages("NOTE"), "<Values> parameter has no effect in wide to long transposition, when results are stacked.",
			   info = "Wide to long transposition doesn't support value parameter transposition")


# Wide to long transposition doesn't support weight parameter transposition
wide_df <- dummy_wide_df |>
   transpose_plus(preserve = year,
                  pivot    = list(sex = c("Male", "Female")),
                  weight   = "Total")

expect_message(print_stack_as_messages("NOTE"), "<Weight> parameter has no effect in wide to long transposition.",
               info = "Wide to long transposition doesn't support weight parameter transposition")


# Preserve variable in transposition is not part of the data frame
wide_df <- dummy_df |>
   transpose_plus(preserve = "test",
                  pivot    = "sex",
                  value    = income)

expect_warning(print_stack_as_messages("WARNING"), "The provided <preserve> variable",
               info = "Preserve variable in transposition is not part of the data frame")


# Value variable in transposition is also part of preserve
wide_df <- dummy_df |>
   transpose_plus(preserve = sex,
                  pivot    = "age",
                  value    = sex)

expect_warning(print_stack_as_messages("WARNING"), "The provided <values> variable",
               info = "Value variable in transposition is also part of preserve")

###############################################################################
# Abort checks
###############################################################################

# Wide to long transposition doesn't support nesting variables
wide_df <- dummy_wide_df |>
			transpose_plus(preserve = year,
						   pivot    = list(sex = "Male + Female"))

expect_error(print_stack_as_messages("ERROR"), "Nesting <pivot> variables in a wide to long transposition is not possible.",
             info = "Wide to long transposition doesn't support nesting variables")


# Abort transposition if pivot variable is part of preserve
wide_df <- dummy_wide_df |>
			transpose_plus(preserve = year,
						   pivot    = "year",
						   value    = income)

expect_error(print_stack_as_messages("ERROR"), "The provided <pivot> variable",
             info = "Abort transposition if pivot variable is part of preserve")


# Abort transposition if no valid value variable is provided
wide_df <- dummy_df |>
			transpose_plus(preserve = year,
						   pivot    = "sex")

expect_error(print_stack_as_messages("ERROR"), "No <values> provided. Transposition will be aborted.",
             info = "Abort transposition if no valid value variable is provided")


# Abort transposition if value variable is not part of the data frame
wide_df <- dummy_df |>
			transpose_plus(pivot = "sex",
						   value = test)

expect_error(print_stack_as_messages("ERROR"), "No valid <values> to transpose provided. Transposition will be aborted.",
             info = "Abort transposition if value variable is not part of the data frame")


# Abort if value variable in transposition is also part of pivot
wide_df <- dummy_df |>
			transpose_plus(pivot = "sex",
						   value = sex)

expect_error(print_stack_as_messages("ERROR"), "The provided <values> variable",
             info = "Abort if value variable in transposition is also part of pivot")


# Abort on duplicate variable names after transposition
wide_df <- dummy_df |>
			transpose_plus(preserve = year,
						   pivot    = c("sex", "education"),
						   values   = income)

expect_error(print_stack_as_messages("ERROR"), "Duplicate column names found:",
             info = "Abort on duplicate variable names after transposition")


# Abort if no valid pivot variable is provided in transposition
wide_df <- dummy_df |>
			transpose_plus(pivot = "test",
						   value = income)

expect_error(print_stack_as_messages("ERROR"), "The provided <pivot> variable",
             info = "Abort if no valid pivot variable is provided in transposition")


# Abort side by side transposition, if list entries are of unequal lengths
wide_df <- dummy_wide_df |>
    transpose_plus(preserve = year,
                   pivot    = list(sex = c("Total", "Male", "Female", "low"),
                                   sex = c("low", "middle", "high")))

expect_error(print_stack_as_messages("ERROR"), "Every <pivot> list entry has to have the same number of variables for a",
             info = "Abort side by side transposition, if list entries are of unequal lengths")


# Abort side by side transposition, if list entries contain a unique variable name but also others
wide_df <- dummy_wide_df |>
    transpose_plus(preserve = year,
                   pivot    = list(sex  = c("Total", "Male", "Female", "low"),
                                   sex  = c("low", "middle", "high"),
                                   test = c("low", "middle", "high")))

expect_error(print_stack_as_messages("ERROR"), "The new result columns can only be set side by side in a wide",
             info = "Abort side by side transposition, if list entries contain a unique variable name but also others")


set_no_print()

#' qol - Quality of Life
#'
#' @author Tim Siebenmorgen
#'
#' @description
#' This package brings some quality of life concepts and functions inspired
#' by 'SAS' to 'R'. The main goal is to make descriptive evaluations easier, so
#' one can create bigger and more complex outputs in less time with less code.
#' Introducing format containers with multilabels, a more powerful summarise, which
#' is capable to output every possible combination of the provided grouping variables
#' in one go, tabulation functions which can create any table in different styles
#' and other more readable functions.
#'
#' In addition it offers an error handling which often catches errors and just
#' let's your code flow, even if there are small errors. You always get an
#' understandable message which helps you to get rid of the problem.
#'
#' The package builds on the incredibly fast data.table and collapse packages for
#' maximum speed and on the wonderful openxlsx2 package for maximum style.
#'
#' @section Imports:
#' data.table, collapse, openxlsx2, fst
#'
#' @section Minimal R Version:
#' 4.1.0 or higher
#'
#' @section Functions:
#' Creating formats: [discrete_format()], [interval_format()], [expand_formats()]
#'
#' Summarisation and tabulation: [summarise_plus()], [frequencies()], [crosstabs()], [any_table()], [export_with_style()], [combine_into_workbook()]
#'
#' Creating a custom table style: [excel_output_style()], [modify_output_style()],
#' [number_format_style()], [modify_number_formats()]
#'
#' Recoding: [recode.()], [recode_multi()]
#'
#' Selecting: [keep()], [dropp()], [inverse()], [vars_between()]
#'
#' Joining: [multi_join()]
#'
#' Transposing: [transpose_plus()]
#'
#' Sorting: [sort_plus()]
#'
#' If-statement: [if.()], [else_if.()], [else.()], [ifelse_multi()], [where.()], [do_if()], [else_do()], [end_do()], [end_all_do()]
#'
#' Monitoring: [monitor_start()], [monitor_end()], [monitor_next()], [monitor_plot()]
#'
#' Renaming: [rename_pattern()], [add_extension()], [remove_stat_extension()], [replace_except()]
#'
#' Retaining: [running_number()], [mark_case()], [retain_value()], [retain_stat()], [retain_variables()]
#'
#' Character Manipulation: [concat()], [sub_string()], [remove_blanks()]
#'
#' Generate dummy data: [dummy_data()]
#'
#' Conversion: [args_to_char()], [dots_to_char()], [get_origin_as_char()] [convert_numeric()], [convert_factor()],
#' [hex_to_256()], [hex_to_ansi()]
#'
#' Macros: [macro()], [apply_macro()]
#'
#' Saving And Loading: [libname()], [stack_data()], [save_file()], [save_file_multi()], [load_file()], [load_file_multi()]
#'
#' Reporting: [content_report()], [code_statistics()], [report_test_results()], [test_package()], [test_single_file()]
#'
#' Messages: [print_message()], [print_headline()], [print_start_message()], [print_closing()], [print_step()],
#' [set_up_custom_message()], [get_message_stack()], [set_no_print()], [print_stack_as_messages()], [convert_square_brackets()]
#'
#' Calculation: [compute.()], [row_calculation()]
#'
#' Small helpers: [setcolorder_by_pattern()], [drop_type_vars()], [fuse_variables()],
#' [get_excel_range()], [get_integer_length()], [round_values()]
#'
#' Split data frame: [split_by]
#'
#' Error handling: [resolve_intersection()], [part_of_df()], [remove_doubled_values()], [check_weight()]
#'
#' Global options: [set_style_options()], [get_style_options()], [reset_style_options()], [close_file()],
#' [set_labels()], [get_labels()], [set_print()], [get_print()], [set_monitor()],
#' [get_monitor()], [set_na.rm()], [get_na.rm()], [set_output()], [get_output()], [set_titles()],
#' [get_titles()], [set_footnotes()], [get_footnotes()], [set_print_miss()], [get_print_miss()],
#' [set_threads()], [get_threads()]
#'
#' Other: [build_master()], [build_rstheme()], [free_memory()], [qol_news()], [qol_chat()]
#'
#' @section Snippets:
#' \preformatted{
#' snippet splus
#'     summarise_plus(class      = c(var1, var2, ...),
#'                    values     = c(var1, var2, ...),
#'                    statistics = c("pct_group", "sum", "sum_wgt", "freq"),
#'                    weight     = weight_var,
#'                    formats    = list(var = format., ...),
#'                    nesting    = "deepest")
#'
#' snippet if.
#'          if.(condition, var = value) |>
#'     else_if.(condition, var = value) |>
#'     else.   (           var = value)
#'
#' snippet freq
#'     frequencies(variables = c(var1, var2, ...),
#'                 weight    = weight_var,
#'                 formats   = list(var = "format.", ...),
#'                 titles    = c(),
#'                 footnotes = c())
#'
#' snippet cross
#'     crosstabs(rows       = row_var,
#'               columns    = col_var,
#'               statistics = c("sum", "pct_row", "pct_column", "pct_total", "freq"),
#'               weight    = weight_var,
#'               formats   = list(var = format., ...),
#'               titles    = c(),
#'               footnotes = c())
#'
#' snippet any
#'     any_table(rows        = c("var1 + var2 + ...", "var1"),
#'               columns     = c("var3", "var3 + var4 + ..."),
#'               values      = c("value_var1", "value_var2"),
#'               statistics  = c("sum", "pct_group", "pct_value", "freq"),
#'               pct_group   = c("var1", "var2"),
#'               pct_value   = list(new_var = "numerator / denominator"),
#'               weight    = weight_var,
#'               formats     = list(var = format., ...),
#'               titles      = c(),
#'               footnotes   = c(),
#'               var_labels  = list("var1" = "My label", ...),
#'               stat_labels = list("pct" = "%"),
#'               box         = "")
#' }
#'
#' @keywords internal
"_PACKAGE"

#' Compute Any Possible Table
#'
#' @description
#' [any_table()] produces any possible descriptive table in 'Excel' format. Any number
#' of variables can be nested and crossed. The output is an individually styled
#' 'Excel' table, which also receives named ranges, making it easier to read the data back in.
#'
#' @param data_frame A data frame in which are the variables to tabulate.
#' @param rows A vector that provides single variables or variable combinations that
#' should appear in the table rows. To nest variables use the form:
#' "var1 + var2 + var3 + ...".
#' @param columns A vector that provides single variables or variable combinations that
#' should appear in the table rows. To nest variables use the form:
#' "var1 + var2 + var3 + ...".
#' @param values A vector containing all variables that should be summarised.
#' @param statistics Available functions:
#' - "sum"       -> Weighted and unweighted sum
#' - "sum_wgt"   -> Sum of all weights
#' - "freq"      -> Unweighted frequency
#' - "freq_g0"   -> Unweighted frequency of all values greater than zero
#' - "pct_group" -> Weighted and unweighted percentages within the respective group
#' - "pct_value" -> Weighted and unweighted percentages between value variables or of a variable expression
#' - "pct_total" -> Weighted and unweighted percentages compared to the grand total
#' - "mean"      -> Weighted and unweighted mean
#' - "median"    -> Weighted and unweighted median
#' - "mode"      -> Weighted and unweighted mode
#' - "min"       -> Minimum
#' - "max"       -> Maximum
#' - "sd"        -> Weighted and unweighted standard deviation
#' - "variance"  -> Weighted and unweighted standard variance
#' - "first"     -> First value
#' - "last"      -> Last value
#' - "pn"        -> Weighted and unweighted percentiles (any p1, p2, p3, ... possible)
#' Uses the SAS default way of calculating percentiles, which is Type 2 described in [quantile()].
#' - "missing"   -> Missings generated by the value variables
#' @param pct_group This option is used to determine which variable of the row and column
#' variables should add up to 100 %. Multiple variables can be specified in a vector to
#' generate multiple group percentages. You can also use the keywords "row_pct" or "col_pct"
#' to calculate total percentages for rows and columns regardless of the respective other dimension.
#' @param pct_value You can pass a list here, which contains the information for a new
#' variable name and between which of the value variables percentages should be computed.
#' @param pct_block Can be "rows" or "columns". Calculates percentages for the last
#' group of variables inside the individual row or column combinations.
#' @param compute A list in which individual calculations are handled. For the writing style
#' see the [compute.()] function.
#' @param formats A list in which is specified which formats should be applied to which variables.
#' @param by Compute tables stratified by the expressions of the provided variables.
#' @param weight Put in a weight variable to compute weighted results.
#' @param order_by Determine how the columns will be ordered. "values" orders the results by the
#' order you provide the variables in values. "stats" orders them by the order under statistics.
#' "values_stats" is a combination of both. "columns" keeps the order as given in columns
#' and "interleaved" alternates the stats.
#' @param titles Specify one or more table titles.
#' @param footnotes Specify one or more table footnotes.
#' @param var_labels A list in which is specified which label should be printed for
#' which variable instead of the variable name.
#' @param stat_labels A list in which is specified which label should be printed for
#' which statistic instead of the statistic name.
#' @param box Provide a text for the upper left box of the table.
#' @param workbook Insert a previously created workbook to expand the sheets instead of
#' creating a new file.
#' @param style A list of options can be passed to control the appearance of 'Excel' outputs.
#' Styles can be created with [excel_output_style()].
#' @param output The following output formats are available: excel and excel_nostyle.
#' @param na.rm FALSE by default. If TRUE removes all NA values from the variables.
#' @param print_miss FALSE by default. If TRUE outputs all possible categories of the
#' grouping variables based on the provided formats, even if there are no observations
#' for a combination.
#' @param print TRUE by default. If TRUE prints the output, if FALSE doesn't print anything. Can be used
#' if one only wants to catch the output data frame and workbook with meta information.
#' @param monitor FALSE by default. If TRUE, outputs two charts to visualize the functions time consumption.
#'
#' @details
#' [any_table()] is based on the 'SAS' procedure Proc Tabulate, which provides
#' efficient and readable ways to perform complex tabulations.
#'
#' With this function you can combine any number of variables in any possible way, all
#' at once. You just define which variables or variable combinations should end up in
#' the table rows and columns with a simple syntax. Listing variables in a vector like
#' c("var1", "var2", "var3",...) means to put variables below (in case of the
#' row variables) or besides (in case of the column variables) each other. Nesting variables
#' is as easy as putting a plus sign between them, e.g. c("var1 + var2", "var2" + "var3" + "var4", etc.).
#' And of course you can combine both versions.
#'
#' The real highlight is, that this function not only creates all the desired variable
#' combinations and exports them to an 'Excel' file, it prints a fully custom styled
#' table to a workbook. Setting up a custom, reusable style is as easy as setting up
#' options like: provide a color for the table header, set the font size for the row header,
#' should borders be drawn for the table cells yes/no, and so on. Merging doubled header texts,
#' happens automatically.
#'
#' With this function you basically can fully concentrate on designing a table, instead of
#' thinking hard about how to calculate where to put a border or to even manually prepare
#' a designed workbook.
#'
#' @return
#' Returns a list with the data table containing the results for the table, the formatted
#' 'Excel' workbook and the meta information needed for styling the final table.
#'
#' @seealso
#' Creating a custom table style: [excel_output_style()], [modify_output_style()],
#' [number_format_style()], [modify_number_formats()].
#'
#' Global style options: [set_style_options()], [set_labels()].
#'
#' Other global options: [set_titles()], [set_footnotes()], [set_print()], [set_monitor()],
#' [set_na.rm()], [set_print()], [set_print_miss()], [set_output()].
#'
#' Combine Excel workbooks: [combine_into_workbook()].
#'
#' Creating formats: [discrete_format()] and [interval_format()].
#'
#' Functions that can handle formats and styles: [frequencies()], [crosstabs()].
#'
#' Additional functions that can handle styles: [export_with_style()]
#'
#' Additional functions that can handle formats: [summarise_plus()], [recode.()],
#' [recode_multi()], [transpose_plus()], [sort_plus()]
#'
#' @examples
#' # NOTE: Some of the examples use the parameter 'output = "excel_nostyle"' to
#' #       make them run faster so that more examples can be provided. Take out
#' #       the parameter to get the fully styled excel table.
#' #       The global option for printing the output is also set to FALSE for
#' #       the same reason. When running the examples manually, don't run this
#' #       option here.
#' set_print(FALSE)
#' set_style_options(background_color = "")
#'
#' # Example data frame
#' my_data <- dummy_data(1000)
#' my_data[["person"]] <- 1
#'
#' # Formats
#' age. <- discrete_format(
#'     "Total"          = 0:100,
#'     "under 18"       = 0:17,
#'     "18 to under 25" = 18:24,
#'     "25 to under 55" = 25:54,
#'     "55 to under 65" = 55:64,
#'     "65 and older"   = 65:100)
#'
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' # NOTE: "!" in front of an expression makes the expression available for
#' #       calculations but prevents it from beeing printed out.
#' education. <- discrete_format(
#'     "!Total"           = c("low", "middle", "high"),
#'     "low education"    = "low",
#'     "middle education" = "middle",
#'     "high education"   = "high")
#'
#' state. <- discrete_format(
#'     "Germany"                       = 1:16,
#'     "Schleswig-Holstein"            = 1,
#'     "Hamburg"                       = 2,
#'     "Lower Saxony"                  = 3,
#'     "Bremen"                        = 4,
#'     "North Rhine-Westphalia"        = 5,
#'     "Hesse"                         = 6,
#'     "Rhineland-Palatinate"          = 7,
#'     "Baden-Württemberg"             = 8,
#'     "Bavaria"                       = 9,
#'     "Saarland"                      = 10,
#'     "West"                          = 1:10,
#'     "Berlin"                        = 11,
#'     "Brandenburg"                   = 12,
#'     "Mecklenburg-Western Pomerania" = 13,
#'     "Saxony"                        = 14,
#'     "Saxony-Anhalt"                 = 15,
#'     "Thuringia"                     = 16,
#'     "East"                          = 11:16)
#'
#' # Define style
#' set_style_options(column_widths = c(2, 15, 15, 15, 9))
#'
#' # Define titles and footnotes. If you want to add hyperlinks you can do so by
#' # adding "link:" followed by the hyperlink to the main text. Linking to another
#' # cell works with "cell:". To link to a file use "file:" an pass the full file
#' # path afterwards.
#' set_titles("This is title number 1",
#'            "This is title number 2 link: https://cran.r-project.org/",
#'            "This is title number 3 cell: W22",
#'            "This is title number 4 file: C:/MyFolder/MyFile.txt")
#'
#' set_footnotes("This is footnote number 1",
#'               "This is footnote number 2 file: C:/MyFolder/MyFile.txt",
#'               "This is footnote number 3 cell: W22",
#'               "This is footnote number 4 link: https://cran.r-project.org/")
#'
#' # Output complex tables with different percentages
#' my_data |> any_table(rows      = c("sex + age", "sex", "age"),
#'                      columns   = c("year", "education + year"),
#'                      values    = weight,
#'                      pct_group = c("row_pct", "col_pct"),
#'                      formats   = list(sex = sex., age = age.,
#'                                       education = education.),
#'                      na.rm     = TRUE)
#'
#' # If you want to get a clearer vision of what the result table looks like, in terms
#' # of the row and column categories, you can write the code like this, to make out
#' # the variable crossings and see the order.
#' # my_data |> any_table(columns   = c(            "year", "education + year"),
#' #                      rows      = c("sex + age",
#' #                                    "sex",
#' #                                    "age"),
#' #                      values    = weight,
#' #                      pct_group = c("sex", "age"),
#' #                      formats   = list(sex = sex., age = age.,
#' #                                       education = education.),
#' #                      output    = "excel_nostyle",
#' #                      na.rm     = TRUE)
#'
#' # Percentages based on value variables instead of categories
#' my_data |> any_table(rows       = c("age + year"),
#'                      columns    = "sex",
#'                      values     = c(probability, person),
#'                      statistics = c("pct_value", "sum", "freq"),
#'                      pct_value  = list(rate = "probability / person"),
#'                      weight     = weight,
#'                      formats    = list(sex = sex., age = age.),
#'                      output    = "excel_nostyle",
#'                      na.rm      = TRUE)
#'
#' # Percentages based on a formatted variable expression
#' my_data |> any_table(rows       = c("age + year"),
#'                      columns    = "sex",
#'                      values     = c(probability, person),
#'                      pct_value  = list(sex = "Total", age = "Total"),
#'                      weight     = weight,
#'                      formats    = list(sex = sex., age = age.),
#'                      output    = "excel_nostyle",
#'                      na.rm      = TRUE)
#'
#' # Percentages based on variable combination blocks
#' # NOTE: age + (year, education) becomes "age + year" and "age + education"
#' #       and will be sorted together. Also works in columns. You can also
#' #       do something like this: age + (year, sex + education, education).
#' my_data |> any_table(rows       = c("sex + (age, education)"),
#'                      columns    = "year",
#'                      values     = weight,
#'                      pct_block  = c("rows", "columns"),
#'                      formats    = list(sex = sex., age = age.,
#'                                        education = education.),
#'                      output     = "excel_nostyle",
#'                      na.rm      = TRUE)
#'
#' # Individual calculations
#' my_data |> any_table(rows       = c("age + year"),
#'                      columns    = "sex",
#'                      values     = "probability",
#'                      statistics = c("sum", "sum_wgt"),
#'                      compute    = list(percent    = probability_sum * 100 / sum_wgt,
#'                                        pct_square = percent ^ 2),
#'                      weight     = weight,
#'                      formats    = list(sex = sex., age = age.),
#'                      output    = "excel_nostyle",
#'                      na.rm      = TRUE)
#'
#' # Customize the visual appearance by adding variable and statistic labels. Both
#' # can also be set as a global option, if labels should be reused over multiple
#' # tables.
#' # Note: You don't have to describe every element. Sometimes a table can be more
#' # readable with less text. To completely remove a variable label just put in an
#' # empty text "" as label.#'
#' set_titles("This is title number 1",
#'            "This is title number 2",
#'            "This is title number 3")
#'
#' set_footnotes("This is footnote number 1",
#'               "This is footnote number 2",
#'               "This is footnote number 3")
#'
#' my_data |> any_table(rows        = c("age + year"),
#'                      columns     = "sex",
#'                      values      = weight,
#'                      statistics  = c("sum", "pct_group"),
#'                      order_by    = "interleaved",
#'                      formats     = list(sex = sex., age = age.),
#'                      var_labels  = list(age = "Age categories",
#'                                         sex = "", weight = ""),
#'                      stat_labels = list(pct = "%"),
#'                      output      = "excel_nostyle",
#'                      na.rm       = TRUE)
#'
#' # Individual styling can also be passed directly
#' my_style <- excel_output_style(header_back_color = "0077B6",
#'                                font              = "Times New Roman")
#'
#' # my_data |> any_table(rows       = c("age + year"),
#' #                      columns    = "sex",
#' #                      values     = c(probability, person),
#' #                      statistics = c("pct_value", "sum", "freq"),
#' #                      pct_value  = list(rate = "probability / person"),
#' #                      weight     = weight,
#' #                      formats    = list(sex = sex., age = age.),
#' #                      style      = my_style,
#' #                      na.rm      = TRUE)
#'
#' # Select specific statistics for specific variables
#' my_data |> any_table(rows       = "sex",
#'                      columns    = "year",
#'                      statistics = list("sum"       = c(weight, income),
#'                                        "pct_group" = balance),
#'                      formats    = list(sex = sex.),
#'                      output     = "excel_nostyle",
#'                      na.rm      = TRUE)
#'
#' # Pass on workbook to create more sheets in the same file
#' my_style <- my_style |> modify_output_style(sheet_name = "age_sex")
#'
#' result_list <- my_data |>
#'            any_table(rows       = "age",
#'                      columns    = "sex",
#'                      values     = weight,
#'                      statistics = "sum",
#'                      formats    = list(sex = sex., age = age.),
#'                      style      = my_style,
#'                      output     = "excel_nostyle",
#'                      na.rm      = TRUE,
#'                      print      = FALSE)
#'
#' my_style <- my_style |> modify_output_style(sheet_name = "edu_year")
#'
#' my_data |> any_table(workbook   = result_list[["workbook"]],
#'                      rows       = "education",
#'                      columns    = "year",
#'                      values     = weight,
#'                      statistics = "pct_group",
#'                      formats    = list(education = education.),
#'                      style      = my_style,
#'                      output     = "excel_nostyle",
#'                      na.rm      = TRUE)
#'
#' # In case you have a good amount of tables, you want to combine in a single workbook,
#' # you can also catch the outputs and combine them afterwards in one go.
#' # NOTE: This section is commented out to just show the principle. Otherwise the
#' #       example code would run for too long.
#' # set_style_options(sheet_name = "sheet1")
#' # tab1 <- my_data |> any_table(..., print = FALSE, output = "excel_nostyle")
#' #
#' # set_style_options(sheet_name = "sheet2")
#' # tab2 <- my_data |> any_table(..., print = FALSE, output = "excel_nostyle")
#' #
#' # set_style_options(sheet_name = "sheet3")
#' # tab3 <- my_data |> any_table(..., print = FALSE, output = "excel_nostyle")
#' #
#' # ...
#' #
#' # Every of the above tabs is a list, which contains the data table, an unstyled
#' # workbook and the meta information needed for the individual styling. These tabs
#' # can be input into the following function, which reads the meta information, styles
#' # each table individually and combines them as separate sheets into a single workbook.
#' # combine_into_workbook(tab1, tab2, tab3 file = "C:/My_folder/My_workbook.xlsx")
#'
#' # The result list from above also carries the transformed data frame if
#' # needed for further usage
#' any_table_df <- result_list[["table"]]
#'
#' # Output multiple complex tables by expressions of another variable.
#' # If you specify the sheet name as "by" in the output style, the sheet
#' # names are named by the variable expressions of the by-variable. Otherwise
#' # the given sheet named gets a running number.
#' my_style <- my_style |> modify_output_style(sheet_name = "by")
#'
#' my_data |> any_table(rows       = c("sex", "age"),
#'                      columns    = "education",
#'                      values     = weight,
#'                      by         = state,
#'                      statistics = c("sum", "pct_group"),
#'                      pct_group  = "education",
#'                      formats    = list(sex = sex., age = age., state = state.,
#'                                        education = education.),
#'                      output     = "excel_nostyle",
#'                      na.rm      = TRUE)
#'
#' # To save a table as xlsx file you have to set the path and filename in the
#' # style element
#' # Example files paths
#' table_file <- tempfile(fileext = ".xlsx")
#'
#' # Note: Normally you would directly input the path ("C:/MyPath/") and name ("MyFile.xlsx").
#' set_style_options(save_path  = dirname(table_file),
#'                   file       = basename(table_file),
#'                   sheet_name = "MyTable")
#'
#' my_data |> any_table(rows    = "sex",
#'                      columns = "year",
#'                      values  = weight,
#'                      formats = list(sex = sex.))
#'
#' # Manual cleanup for example
#' unlink(table_file)
#'
#' # Global options are permanently active until the current R session is closed.
#' # There are also functions to reset the values manually.
#' reset_style_options()
#' reset_qol_options()
#' close_file()
#'
#' set_print(TRUE)
#' set_style_options(background_color = "FFFFFF")
#'
#' @export
any_table <- function(data_frame,
                      rows,
                      columns        = "",
                      values,
                      statistics     = NULL,
                      pct_group      = c(),
                      pct_value      = list(),
                      pct_block      = c(),
                      compute        = list(),
                      formats        = list(),
                      by             = c(),
                      weight         = NULL,
                      order_by       = "stats",
                      titles         = .qol_options[["titles"]],
                      footnotes      = .qol_options[["footnotes"]],
                      var_labels     = .qol_options[["var_labels"]],
                      stat_labels    = .qol_options[["stat_labels"]],
                      box            = "",
                      workbook       = NULL,
                      style          = .qol_options[["excel_style"]],
                      output         = .qol_options[["output"]],
                      na.rm          = .qol_options[["na.rm"]],
                      print_miss     = .qol_options[["print_miss"]],
                      print          = .qol_options[["print"]],
                      monitor        = .qol_options[["monitor"]]){

    # Measure the time
    print_start_message()
	print_step("GREY", "Error handling")

    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Error handling", "Preparation")
    #-------------------------------------------------------------------------#

    ###########################################################################
    # Early evaluations
    ###########################################################################

    # First convert data frame to data table
    if (!data.table::is.data.table(data_frame)){
        data_frame <- data.table::as.data.table(data_frame)
    }

    # Evaluate formats early
    if (!is_list_of_dfs(formats)){
        formats_list <- as.list(substitute(formats))[-1]
        formats      <- evaluate_formats(formats_list)
    }

    # Remove empty formats and throw a warning. This can happen if there is e.g.
    # a typo in the format.
    for (variable in names(formats)){
        if (is.null(formats[[variable]])){
            formats[[variable]] <- NULL
            print_message("WARNING", "Format for variable '[variable]' does not exist and can't be applied.", variable = variable)
        }
    }

    # Apply macros to save path and file
    if (!is.null(style[["save_path"]])){
        style[["save_path"]] <- macro(style[["save_path"]])
    }

    if (!is.null(style[["file"]])){
        style[["file"]] <- macro(style[["file"]])
    }

    ###########################################################################
    # Error handling
    ###########################################################################

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Workbook
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Check whether the whole result list of one of the tabulation functions was passed
    if (!is.null(workbook) && is.list(workbook)){
        # If the workbook object is part of the list, extract it
        if ("workbook" %in% names(workbook)){
            workbook <- workbook[["workbook"]]
        }
        # If the list is some other list without workbook object, abort
        else{
            print_message("ERROR", c("Workbook object is invalid. You have to provide a workbook object",
                                     "created with one of the tabulation functions. Tabulation will be aborted."))
            return(invisible(NULL))
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Row variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Enable the use of macro variables
    rows <- apply_macro(rows)

    # Expand brackets first to be able to identify the unique variables used afterwards
    row_expand <- expand_brackets(rows)

    if (row_expand[[2]]){
        rows <- names(row_expand[[1]])
    }
    else{
        row_expand <- NULL
    }

    # Get row variables from provided combinations
    row_vars <- unlist_variables(rows)

    if (is.null(row_vars)){
        print_message("ERROR", "<Rows> variables must be provided in quotation marks. Tabulation will be aborted.")
        return(invisible(NULL))
    }

    row_vars <- data_frame |> part_of_df(row_vars, check_only = TRUE)

    if (is.list(row_vars)){
        print_message("ERROR", c("The provided <rows> variable[?s] '[vars]' [?is/are] not part of",
								 "the data frame. Tabulation will be aborted."), vars = row_vars[[1]])
        return(invisible(NULL))
    }

    if (length(rows) <= 1){
        if (length(rows) == 0 || rows == ""){
            print_message("ERROR", "No valid <rows> variables provided. Tabulation will be aborted.")
            return(invisible(NULL))
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Column variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Enable the use of macro variables
    columns <- apply_macro(columns)

    # Expand brackets first to be able to identify the unique variables used afterwards
    column_expand <- expand_brackets(columns)

    if (column_expand[[2]]){
        columns <- names(column_expand[[1]])
    }
    else{
        column_expand <- NULL
    }

    # Get row variables from provided combinations
    col_vars <- unlist_variables(columns)

    if (is.null(col_vars)){
        print_message("ERROR", "<Columns> variables must be provided in quotation marks. Tabulation will be aborted.")
        return(invisible(NULL))
    }

    col_vars <- data_frame |> part_of_df(col_vars, check_only = TRUE)

    if (is.list(col_vars)){
        print_message("ERROR", c("The provided <columns> variable[?s] '[vars]' [?is/are] not part of",
								 "the data frame. Tabulation will be aborted."), vars = col_vars[[1]])
        return(invisible(NULL))
    }

    # Make sure there is no column variable that is also a row variable.
    col_vars <- resolve_intersection(col_vars, row_vars, check_only = TRUE)

    if (is.list(col_vars)){
        print_message("ERROR", c("The provided <columns> variable[?s] '[vars]' [?is/are] also part of",
								 "the <rows> variables. Tabulation will be aborted."), vars = col_vars[[1]])
        return(invisible(NULL))
    }

    if (length(columns) <= 1){
        if (length(columns) == 0 || columns == ""){
            # Create empty pseudo variable to let the rest of the program run as normal
            data_frame[[".temp.var"]] <- 1
            columns    <- ".temp.var"
            col_vars   <- ".temp.var"
            var_labels <- c(var_labels, ".temp.var" = "")
            formats[[".temp.var"]] <- suppressMessages(discrete_format(" " = 1))
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # By variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Convert to character vectors
    by <- get_origin_as_char(by, substitute(by))

    # Make sure that the variables provided are part of the data frame.
    by <- data_frame |> part_of_df(by)

    # Make sure there is no row/column variable that is also a by variable.
    variables  <- c(row_vars, col_vars)
    by <- resolve_intersection(by, variables, check_only = TRUE)

    if (is.list(by)){
        print_message("WARNING", c("The provided <by> variable[?s] '[vars]' [?is/are] also part of",
								   "the <rows> or <columns> variables, which is not allowed. <by> will be removed."), vars = by[[1]])
        by <- c()
    }

    if (length(by) == 1){
        if (by == ""){
            print_message("WARNING", "No valid <by> variables provided. <by> will be removed.")
            by <- c()
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Weight
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    weight     <- get_origin_as_char(weight, substitute(weight))
    data_frame <- data_frame |> check_weight(weight)
    weight_var <- ".temp_weight"

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Statistics
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Translate statistics if possible
    temp_statistics <- tryCatch({
        # Force evaluation to see if it exists
        get_origin_as_char(statistics, substitute(statistics))
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    # In case statistics couldn't be evaluated, this means a list was passed,
    # which provides specific variables for specific statistics. This list needs
    # to be evaluated differently.
    vars_per_stat_list <- NULL

    if (is.null(temp_statistics)){
        vars_per_stat_list <- lapply(as.list(substitute(statistics)), function(element){
            element <- tolower(as.character(element))

            element[!element %in% "c"]
        })[-1]

        # Set to NULL again on empty list
        if (length(vars_per_stat_list) == 0){
            vars_per_stat_list <- NULL
            statistics         <- NULL
        }
        # If a real list was passed
        else{
            statistics <- names(vars_per_stat_list)
        }
    }
    # Prepare statistics vector as normal
    else{
        statistics <- tolower(temp_statistics)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Values
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Convert to character vectors
    values <- get_origin_as_char(values, substitute(values))

    # If no value variables are provided generate them automatically
    if (length(values) <= 1){
        if (length(values) == 0 || values == ""){
            # If a statistics list with variable names is given, extract them an set
            # them as value variables.
            if (!is.null(vars_per_stat_list)){
                values        <- collapse::funique(unlist(vars_per_stat_list))
                names(values) <- NULL
            }
            # If no value variables are provided generate a temporary variable which
            # outputs unweighted results.
            else{
                values               <- ".temp_values"
                var_labels           <- c(var_labels, .temp_values = "")
                data_frame[[values]] <- 1
            }
        }
    }

    # Make sure there is no row/column variable that is also a value variable.
    values <- resolve_intersection(values, variables, check_only = TRUE)

    if (is.list(values)){
        print_message("ERROR", c("The provided <rows>/<columns> variable[?s] '[vars]' [?is/are] also part of",
								 "the <values> variables. Tabulation will be aborted."), vars = values[[1]])
        return(invisible(NULL))
    }

    # Make sure that the variables provided are part of the data frame.
    values <- data_frame |> part_of_df(values)

    if (length(values) == 0){
        print_message("ERROR", "No valid <values> variables provided. Tabulation will be aborted.")
        return(invisible(NULL))
    }

    values <- remove_doubled_values(values)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Output
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Silent conversion of global options, which are invalid for any_table
    if (tolower(output) %in% c("console", "text")){
        output <- "excel"
    }

    # Check for invalid output option
    if (!tolower(output) %in% c("excel", "excel_nostyle")){
        print_message("WARNING", "<Output> format '[output]' not available. Using 'excel' instead.", output = output)

        output <- "excel"
    }
    else{
        output <- tolower(output)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Pre summarised data
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    pre_summed <- data_frame |> is_pre_summed(c(by, variables))

    # Extensiosn to preserve
    extensions <- c("_sum", "_pct_group_", "_pct_total", "_pct_value", "_pct_block", "_pct",
                    "_freq_g0", "_freq", "_mean", "_median", "_mode", "_min", "_max", "_first",
                    "_last", "_sum_wgt", "_p[0-9]+$", "sum_wgt", "_sd", "_variance", "_missing",
                    "by_vars", "TYPE_NR")

    # In case of using a pre summarised data frame, underscores are only allowed, if they carry
    # the statistics extension afterwards.
    if (pre_summed){
        # This basically serves as the default parameter
        if (length(statistics) == 0){
            statistics <- "sum"
        }

        # Only keep provided variables, otherwise it can happen later on that e.g. too many
        # variables are pivoted
        data_frame <- data_frame |> keep(by, variables, values, ".temp_weight")

        # If the TYPE variable isn't present in the data frame it will be generated here from
        # the provided variables.
        if (!"TYPE" %in% names(data_frame)){
            # For each observation check the non NA variables and combine their names in the TYPE
            # variable.
            data_frame[["TYPE"]] <- apply(data_frame[c(by, variables)], 1,
                function(variable){
                    var_name <- names(variable)[!is.na(variable)]

                    if (length(var_name) == 0){
                        NA_character_
                    }
                    else{
                        paste(var_name, collapse = "+")
                    }
                })

            data_frame |> data.table::setcolorder(c(by, variables, "TYPE"))
        }

        # Check if value variables have statistics extension
        pattern <- paste0("(", paste(extensions, collapse = "|"), ")$")

        # If one of the value variables hasn't got any of the above extension, add extensions automatically as provided
        if (!all(grepl(pattern, values))){
            print_message("WARNING", c("All <values> variables need to have the <statistic> extension in their variable name.",
									   "You can use the function 'add_extension' to achieve this. Provided <statistic> will be used.",
									   "This can lead to a wrong table structure, better add extensions by yourself."))

            data_frame <- data_frame |> add_extension(length(c(by, variables, "TYPE")) + 1,
                                                      rep(statistics, length(values))[1:length(values)])
        }

        # Set up options to make sure nothing errors below. Statistics is set to "mean"
        # because then summarise_plus takes a route where factor variables are kept in order.
        # With "sum" the order would be messed up.
        statistics <- "mean"
        pct_group  <- ""
        pct_value  <- ""
        weight     <- NULL
        formats    <- list()

        rm(pattern)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Order
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    order_by <- get_origin_as_char(order_by, substitute(order_by))

    # Correct ordering option if not set right
    order_by_vars <- NULL
    df_names      <- names(data_frame)

    # If only one argument is passed, it is assumed that it is one of the presets
    if (length(order_by) == 1){
        if (!tolower(order_by) %in% c("values", "stats", "values_stats", "columns", "interleaved", "blocks")){
            # If no preset is passed, check if a single variable name was passed
            order_by_vars    <- order_by[order_by %in% df_names]
            invalid_order_by <- order_by[!order_by %in% df_names]

            # No preset and no variable name was passed
            if (length(invalid_order_by) > 0){
                print_message("WARNING", c("<Order by> option '[order_by]' doesn't exist. Options 'values', 'stats', 'values_stats', 'columns'",
                                           "and 'interleaved' are available. <Order by> will be set to 'stats'."), order_by = order_by)
            }

            order_by <- "stats"
        }
    }
    # In case variable names were passed, format them with the underscore preserving
    # so that they match the pivoted variable names later on.
    else{
        order_by_vars <- replace_except(order_by, "_", "!!!", extensions)
        order_by      <- "stats"
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Statistics
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    pct_group <- get_origin_as_char(pct_group, substitute(pct_group))

    # If pct_group parameter is passed but is not selected as statistic, add it
    if (length(pct_group) > 0 && !"pct_group" %in% tolower(statistics)){
        if (!(length(pct_group) == 1 && pct_group == "")){
            statistics <- c(statistics, "pct_group")
        }
    }

    # If pct_group parameter is passed but is not selected as statistic, add it
    if (length(pct_value) > 0 && !"pct_value" %in% tolower(statistics)){
        if (!(length(pct_value) == 1 && pct_value == "")){
            statistics <- c(statistics, "pct_value")
        }
    }
    else if(length(pct_value) == 0 && "pct_value" %in% tolower(statistics)){
        print_message("NOTE", c("'pct_value' is specified in the statistics but the <pct_value> parameter",
                                " isn't set. 'pct_value' will be removed from statistics."))

        statistics <- statistics[!statistics %in% "pct_value"]
    }

    # If pct_block parameter is passed but is not selected as statistic, add it
    if (length(pct_block) > 0 && !"pct_block" %in% tolower(statistics)){
        if (!(length(pct_block) == 1 && pct_block == "")){
            statistics <- c(statistics, "pct_block")
        }
    }
    else if(length(pct_block) == 0 && "pct_block" %in% tolower(statistics)){
        print_message("NOTE", c("'pct_block' is specified in the statistics but the <pct_block> parameter",
                                " isn't set. 'pct_block' will be removed from statistics."))

        statistics <- statistics[!statistics %in% "pct_block"]
    }

    if (!pre_summed){
        list_of_statistics <- get_complete_statistics_list(statistics)

        # Check for invalid statistics
        valid_stats   <- statistics[statistics %in% c(names(list_of_statistics), "sum_wgt", "pct_value",
                                                      "pct_group", "pct_total", "pct_block", paste0("p", 1:100))]
        invalid_stats <- statistics[!statistics %in% c(names(list_of_statistics), "sum_wgt", "pct_value",
                                                       "pct_group", "pct_total", "pct_block", paste0("p", 1:100))]

        if (length(invalid_stats) > 0){
            print_message("WARNING", "<Statistic> '[invalid]' [?is/are] invalid and will be omitted.", invalid = invalid_stats)

            if (length(valid_stats) == 0){
                print_message("WARNING", "No valid <statistic> selected. 'sum' will be used.")

                statistics <- "sum"
            }
        }

        # This basically serves as the default parameter
        if (length(statistics) == 0){
            statistics <- "sum"
        }

        rm(list_of_statistics, invalid_stats)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Percentages
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    row_pct <- FALSE
    col_pct <- FALSE
    reenter_pct_group <- FALSE

    # Remove missing variables from pct_group
    if ("pct_group" %in% tolower(statistics)){
        # Keep groups for later sorting in case row or column percents where chosen
        order_pct <- sub("_pct", "", pct_group)

        invalid_pct <- pct_group[!pct_group %in% c(row_vars, col_vars)]

        # Check whether keywords for row or column percents where passed
        if (any(c("row_pct", "col_pct") %in% invalid_pct)){
            if ("row_pct" %in% invalid_pct){
                row_pct <- TRUE
            }
            if ("col_pct" %in% invalid_pct){
                col_pct <- TRUE
            }
        }

        if (length(invalid_pct) > 0){
            if (row_pct || col_pct){
                invalid_pct <- invalid_pct[!invalid_pct %in% c("row_pct", "col_pct")]
                reenter_pct_group <- TRUE
            }

            if (length(invalid_pct) > 0){
                print_message("WARNING", c("The variable[?s] '[invalid]' provided as <pct_group> [?is/are] not part",
										   "of the <rows> and <columns> variables. The variable[?s] will be omitted."), invalid = invalid_pct)

            }

            pct_group <- pct_group[pct_group %in% c(row_vars, col_vars)]

            # Save original statistics so that later on the order can be restored if
            # only row or column percentages were used.
            if (length(pct_group) == 0){
                statistics_orig <- statistics
                statistics      <- statistics[!statistics %in% "pct_group"]
            }
        }

        rm(invalid_pct)
    }

    # If pct_value is selected, make sure sum is also part of statistics
    flag_remove_sum <- FALSE

    if (!"sum" %in% statistics){
        if ("pct_value" %in% tolower(statistics) || row_pct || col_pct){
            statistics      <- c(statistics, "sum")
            flag_remove_sum <- TRUE
        }
    }

    # If pct_block is selected, make sure sum is also part of statistics
    flag_remove_block_sum <- FALSE

    pct_block <- get_origin_as_char(pct_block, substitute(pct_block))

    if (!all(pct_block %in% c("rows", "columns"))){
        print_message("WARNING", c("Invalid option for <pct_block>. <pct_block> can be 'rows' or 'columns', 'rows' will be used."))
        pct_block  <- "rows"
    }

    if (!"sum" %in% statistics){
        if ("pct_block" %in% tolower(statistics)){
            statistics            <- c(statistics, "sum")
            flag_remove_block_sum <- TRUE
        }
    }
    else{
        if ("pct_block" %in% tolower(statistics) && flag_remove_sum){
            flag_remove_block_sum <- TRUE
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Resolve macros
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    titles      <- apply_macro(titles)
    footnotes   <- apply_macro(footnotes)
    var_labels  <- apply_macro(var_labels)
    stat_labels <- apply_macro(stat_labels)

    ###########################################################################
    # Any tabulation starts
    ###########################################################################

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Summary", "Summary")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Computing stats")
    print_step("Minor", "Summarise")

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Some preparations beforehand
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Put together vector of grouping variables
    group_vars <- c(by, variables)

    # In case of group percentages order group variable to the last position
    if ("pct_group" %in% tolower(statistics) && length(pct_group) > 0){
        group_vars <- c(setdiff(group_vars, pct_group[1]), pct_group[1])
    }
    # If pct_group is specified in statistics but no group is provided set
    # last variable of group_vars as standard.
    else if ("pct_group" %in% tolower(statistics) && length(pct_group) == 0){
        pct_group <- group_vars[length(group_vars)]
    }

    # Put combinations in a single vector
    combinations <- as.vector(outer(rows, columns, paste, sep = " + "))

    # In case by variables are specified, add by to group variables and build
    # additional combinations
    if (length(by) > 0){
        combinations <- as.vector(outer(by, combinations, paste, sep = "+"))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Summarise data according to provided variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Compute statistics
    if (!pre_summed){
        any_tab <- suppressMessages(data_frame |>
              summarise_plus(class      = group_vars,
                             values     = values,
                             statistics = statistics,
                             formats    = formats,
                             weight     = weight_var,
                             nesting    = "all",
                             types      = combinations,
                             notes      = FALSE,
                             na.rm      = na.rm,
                             print_miss = print_miss)) |>
            rename_pattern("pct_group", paste0("pct_group_", pct_group[1])) |>
            collapse::fsubset(TYPE != "total")
    }
    else{
        # With pre summarised data just take the input data frame
        any_tab <- data_frame
    }

    if (is.null(any_tab)){
        print_message("ERROR", "Table could not be computed. Execution will be aborted.")
        return(invisible(NULL))
    }

    # Underscores are not allowed in column variables because when constructing the
    # table header later, the underscore is the sign by which the column names are split.
    # Having additional underscores would mess up this part and lead to errors.
    if (any(grepl("_", unlist(any_tab[col_vars])))){
        print_message("ERROR", "No underscores allowed in <columns> variable values. Execution will be aborted.")
        return(invisible(NULL))
    }

    # Convert grouping variables to factor to retain value order after sorting later on
    any_tab <- any_tab |> convert_factor(group_vars)

    # In case of pre summarised data frame remove the temporary weight variable
    if (pre_summed){
        any_tab <- any_tab |> collapse::fselect(-.temp_weight)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Handle by variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # In case of by variables fuse them into one
    if (length(by) > 0){
        any_tab <- any_tab |> fuse_variables("by_vars", by)

        # Normally extract the first variable from the TYPE combination into a
        # separate BY variable.
        if (!pre_summed){
            any_tab[["BY"]]   <- sub("\\+.*", "", any_tab[["TYPE"]])
            any_tab[["TYPE"]] <- sub("^[^+]*\\+\\s*", "", any_tab[["TYPE"]])
        }
        # In case of pre summarised data frame it is not clear, at which position
        # the by variable is in the TYPE.
        else{
            # Get TYPE variable as list of variable combinations
            type_split <- strsplit(any_tab[["TYPE"]], "\\+")

            # If by variable was found in a variable combination, put it into
            # a separate variable.
            any_tab[["BY"]] <- vapply(type_split, function(obs){
                by_found <- obs[obs %in% by]

                if (length(by_found) > 0){
                    by_found[1]
                }
                else{
                    ""
                }
            }, character(1L))

            # The TYPE variable itself is put back together as all variables which
            # are no by-Variables.
            any_tab[["TYPE"]] <- vapply(type_split, function(obs){
                paste(obs[!obs %in% by], collapse = "+")
            }, character(1L))
        }

        # Convert to factor to keep them in order when sorting later
        any_tab <- any_tab |> convert_factor(c("BY", "by_vars"))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Handle row percentages
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    if (any(grepl("pct", statistics))){ print_step("Minor", "Percentages") }

    # In case keyword row_pct was entered as group percentage, compute percentages
    # based on the row totals.
    if (row_pct){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Row percentages", "Summary")
        #---------------------------------------------------------------------#

        current_values <- paste0(values, "_sum")

        # To make the column variable independent row percentages work, the data frame
        # has to be summarised as before, but the column variables don't receive their
        # formats. This is to prevent multilabels from doubling up the sums.
        # This summarise is important because it makes sure, that the row variables have
        # the same NA values as the main any_tab.
        row_formats <- formats[names(formats) %in% c(by, row_vars)]

        group_df <- suppressMessages(data_frame |>
              summarise_plus(class      = group_vars,
                             values     = values,
                             statistics = "sum",
                             formats    = row_formats,
                             weight     = weight_var,
                             nesting    = "all",
                             types      = combinations,
                             notes      = FALSE,
                             na.rm      = na.rm,
                             print_miss = print_miss)) |>
            rename_pattern("_sum", "")


        # The TYPE is put into the grouping, otherwise the row variable expressions
        # would double up.
        row_group <- c(row_vars, "TYPE")

        # Clean up by variables
        if (length(by) > 0){
            group_df <- group_df |> fuse_variables("by_vars", by)

            group_df[["TYPE"]] <- sub("^[^+]*\\+\\s*", "", group_df[["TYPE"]])

            row_group <- c("by_vars", row_vars, "TYPE")
        }

        # Generate a pseudo weight for a simple summarise
        group_df[[".weight"]] <- 1

        result_list <- group_df |>
            matrix_summarise(values,
                             row_group,
                             group_df[[".weight"]],
                             "sum",
                             get_complete_statistics_list("sum"),
                             monitor_df,
                             FALSE)

        group_df <- result_list[[1]] |>
            rename_pattern("_sum", "_row_sum")

        # Join total sums to the main data frame
        any_tab <- any_tab |>
            collapse::join(group_df,
                           on      = row_group,
                           how     = "left",
                           verbose = FALSE,
                           overid  = 2)

        # Calculate row percentages per value variable
        total_sums <- paste0(values, "_row_sum")

        for (i in seq_along(values)){
            new_variable <- paste0(values[i], "_pct_group_row")

            any_tab[[new_variable]] <-
                data.table::fifelse(any_tab[[total_sums[i]]] > 0,
                                    any_tab[[current_values[i]]] * 100 / any_tab[[total_sums[i]]],
                                    NA)
        }

        # Drop total sums
        any_tab <- any_tab |> dropp(":_row_sum")

        rm(row_group, result_list, group_df, current_values, total_sums, new_variable)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Handle column percentages
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # In case keyword col_pct was entered as group percentage, compute percentages
    # based on the row totals.
    if (col_pct){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Column percentages", "Summary")
        #---------------------------------------------------------------------#

        current_values <- paste0(values, "_sum")

        # To make the row variable independent column percentages work, the data frame
        # has to be summarised as before, but the row variables don't receive their
        # formats. This is to prevent multilabels from doubling up the sums.
        # This summarise is important because it makes sure, that the row variables have
        # the same NA values as the main any_tab.
        col_formats <- formats[names(formats) %in% c(by, col_vars)]

        group_df <- suppressMessages(data_frame |>
             summarise_plus(class      = group_vars,
                            values     = values,
                            statistics = "sum",
                            formats    = col_formats,
                            weight     = weight_var,
                            nesting    = "all",
                            types      = combinations,
                            notes      = FALSE,
                            na.rm      = na.rm,
                            print_miss = print_miss)) |>
            rename_pattern("_sum", "")


        # The TYPE is put into the grouping, otherwise the row variable expressions
        # would double up.
        col_group <- c(col_vars, "TYPE")

        # Clean up by variables
        if (length(by) > 0){
            group_df <- group_df |> fuse_variables("by_vars", by)

            group_df[["TYPE"]] <- sub("^[^+]*\\+\\s*", "", group_df[["TYPE"]])

            col_group <- c("by_vars", col_vars, "TYPE")
        }

        # Generate a pseudo weight for a simple summarise
        group_df[[".weight"]] <- 1

        result_list <- group_df |>
            matrix_summarise(values,
                             col_group,
                             group_df[[".weight"]],
                             "sum",
                             get_complete_statistics_list("sum"),
                             monitor_df,
                             FALSE)

        group_df <- result_list[[1]] |>
            rename_pattern("_sum", "_col_sum")

        # Join total sums to the main data frame
        any_tab <- any_tab |>
            collapse::join(group_df,
                           on      = col_group,
                           how     = "left",
                           verbose = FALSE,
                           overid  = 2)

        # Calculate row percentages per value variable
        total_sums <- paste0(values, "_col_sum")

        for (i in seq_along(values)){
            new_variable <- paste0(values[i], "_pct_group_col")

            any_tab[[new_variable]] <-
                data.table::fifelse(any_tab[[total_sums[i]]] > 0,
                                    any_tab[[current_values[i]]] * 100 / any_tab[[total_sums[i]]],
                                    NA)
        }

        # Drop total sums
        any_tab <- any_tab |> dropp(":_col_sum")

        rm(col_group, result_list, group_df, current_values, total_sums, new_variable)
    }

    if (reenter_pct_group && !"pct_group" %in% statistics){
        statistics <- statistics_orig
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Handle group percentages
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # In case multiple group percentages should be computed, evaluate them in a loop
    # and join them to the main data frame.
    if (length(pct_group) > 1){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Additional group pct", "Summary")
        #---------------------------------------------------------------------#

        # Pre summarise data frame in order to compute additional group percentages faster.
        # This basically is the same approach as in the shortcut route of summarise_plus.
        current_values <- paste0(values, "_sum")

        group_df <- data_frame |>
            matrix_summarise(values,
                             group_vars,
                             data_frame[[weight_var]],
                             "sum",
                             get_complete_statistics_list("sum"),
                             monitor_df,
                             FALSE)

        group_df <- group_df[[1]] |> collapse::frename(stats::setNames(current_values, values))

        group_df[[".temp_weight"]] <- 1

        # Compute additional group percentages one by one
        for (group in seq_along(pct_group)){
            # First group was computed before so omit it here
            if (group == 1){
                next
            }

            # In case of group percentages order group variable to the last position
            group_vars <- c(setdiff(group_vars, pct_group[group]), pct_group[group])

            # Compute group percentages
            group_tab <- suppressMessages(group_df |>
                summarise_plus(class      = group_vars,
                               values     = values,
                               statistics = "pct_group",
                               formats    = formats,
                               weight     = weight_var,
                               nesting    = "all",
                               types      = combinations,
                               notes      = FALSE,
                               na.rm      = na.rm,
                               print_miss = print_miss)) |>
                rename_pattern("pct_group", paste0("pct_group_", pct_group[group])) |>
                drop_type_vars()

            merge_vars <- variables

            if (length(by) > 0){
                group_tab  <- group_tab |> fuse_variables("by_vars", by)
                merge_vars <- c("by_vars", variables)
            }

            # Join percentages to the main data frame
            any_tab <- any_tab |>
                collapse::join(group_tab,
                               on      = merge_vars,
                               how     = "left",
                               verbose = FALSE,
                               overid  = 2)
        }
        rm(group, group_tab, group_df, current_values)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Handle percentages based on value variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # In case percentages based on value variables should be computed
    if (length(pct_value) > 0){
        if (!(length(pct_value) == 1 && pct_value == "")){
            #---------------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Value percentages", "Summary")
            #---------------------------------------------------------------------#

            for (i in seq_along(pct_value)){
                value <- pct_value[[i]]
                name  <- names(pct_value)[i]

                # Skip iteration because given operation is not a division of value
                # variables. Instead it will be a categorical percentage computation,
                # which will be handled in another step down below.
                if (grepl("/", value, fixed = TRUE)){
                    # Separate provided variables first
                    eval_vars <- trimws(strsplit(value, split = "/")[[1]])

                    # Compute percentages
                    if (paste0(eval_vars[1], "_sum") %in% names(any_tab) &&
                        paste0(eval_vars[2], "_sum") %in% names(any_tab)){
                        any_tab[[paste0(name, "_pct_value")]] <-
                            any_tab[[paste0(eval_vars[1], "_sum")]] * 100 /
                            any_tab[[paste0(eval_vars[2], "_sum")]]
                    }
                    # Without sum percentages can't be computed
                    else{
                        flag_remove_sum <- FALSE

                        # Additional warnings for missing variables
                        if (!eval_vars[1] %in% names(data_frame)){
                            print_message("WARNING", "Variable '[eval]' not found in the data frame.", eval = eval_vars[1])
                        }
                        if (!eval_vars[2] %in% names(data_frame)){
                            print_message("WARNING", "Variable '[eval]' not found in the data frame.", eval = eval_vars[2])
                        }
                    }
                }
                else{
                    if (!name %in% names(any_tab)){
                        print_message("WARNING", c("Variable '[name]' not found in the data frame.",
    											   "Percentages can't be computed."), name = name)
                        next
                    }

                    # Get the super group values by subsetting the data frame by the given
                    # variable value.
                    if (length(by) == 0){
                        super_group <- group_vars[!group_vars %in% name]
                    }
                    else{
                        super_group <- c("by_vars", group_vars[!group_vars %in% c(name, by)])
                    }

                    super_tab <- any_tab |>
                        collapse::fsubset(any_tab[[name]] == value) |>
                        keep(super_group, ":_sum") |>
                        rename_pattern("_sum", "_sum_super")

                    if (collapse::fnrow(super_tab) == 0){
                        print_message("WARNING", c("Subsetting variable '[name]' by '[value]' results in an empty data frame.",
    											   "Percentages can't be computed."), name = name, value = value)
                        next
                    }

                    # Join the super group values back to the full data frame
                    any_tab <- collapse::join(any_tab, super_tab,
                                              on       = super_group,
                                              how      = "left",
                                              multiple = TRUE,
                                              verbose  = FALSE,
    										  overid   = 2)

                    # Compute percentages
                    eval_vars <- grep("_sum$", names(any_tab), value = TRUE)

                    for (sum_var in eval_vars){
                        any_tab[[paste0(name, "_pct_value")]] <-
                            any_tab[[sum_var]] * 100 /
                            any_tab[[paste0(sum_var, "_super")]]
                    }

                    any_tab <- any_tab |> dropp(":_super")
                }
            }
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Handle percentages based on value variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # This capture enables the compute parameter to be written without quotation
    # marks.
    compute <- as.list(substitute(compute))[-1]

    if (length(compute) > 0){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Individual calculations", "Summary")
        #---------------------------------------------------------------------#
        print_step("Minor", "Individual calculations")

        any_tab <- suppressMessages(any_tab |> compute.(compute))
    }

    # If no sums were selected in the statistics, drop them again now
    if (flag_remove_sum && !flag_remove_block_sum){
        any_tab <- any_tab |> dropp(":_sum")
    }

    # Abort if no values could be computed. This can especially happen during the various
    # percentage calculations if e.g. a wrong variable or expression was provided.
    if (length(by) == 0 && length(c(row_vars, col_vars, "TYPE", "TYPE_NR", "DEPTH")) == collapse::fncol(any_tab)){
        print_message("ERROR", c("After calculating the results, there are no valid values.",
								 "Tabulation will be aborted."))
        return(invisible(NULL))
    }

    if (length(by) > 0 && length(c(row_vars, col_vars, "by_vars", "TYPE", "TYPE_NR")) == collapse::fncol(any_tab)){
        print_message("ERROR", c("After calculating the results, there are no valid values.",
								 "Tabulation will be aborted."))
        return(invisible(NULL))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Round values
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Reorder variables according to statistics. This is necessary because some
    # percentages can only be computed after summarise_plus and therefor aren't ordered.
    if ("pct_group" %in% statistics){
        any_tab <- any_tab |> setcolorder_by_pattern(order_pct)
    }

    any_tab <- any_tab |> setcolorder_by_pattern(statistics)

    # Get value variable names
    if (length(by) == 0){
        var_vector <- c(variables, "TYPE", "TYPE_NR", "DEPTH")
        value_vars <- any_tab |> inverse(var_vector)
    }
    else{
        var_vector <- c(variables, "TYPE", "TYPE_NR", "DEPTH", "by_vars", "BY")
        value_vars <- any_tab |> inverse(var_vector)
    }

    # Round values according to number formats
    for (var_name in value_vars){
        # Get stat from variable name
        stat <- strsplit(var_name, split = "_")[[1]]
        stat <- stat[length(stat)]

        # Round values to the decimals places specified in the style
        if (tolower(stat) %in% c("sum", "freq", "freq", "mean", "median", "mode",
                                 "min", "max")){
            any_tab[[var_name]] <- round_values(any_tab[[var_name]],
                                          style[["number_formats"]][[paste0(stat, "_decimals")]])
        }
        else if(stat == "g0"){
            any_tab[[var_name]] <- round_values(any_tab[[var_name]],
                                          style[["number_formats"]][["freq_decimals"]])
        }
        else if(stat == "wgt"){
            any_tab[[var_name]] <- round_values(any_tab[[var_name]],
                                          style[["number_formats"]][["sum_decimals"]])
        }
        else if(length(stat) <= 3 && substr(stat, 1, 1) == "p" && grepl("^[0-9]$", substr(stat, 2, 2))){
            any_tab[[var_name]] <- round_values(any_tab[[var_name]],
                                          style[["number_formats"]][["p_decimals"]])
        }
        else{
            any_tab[[var_name]] <- round_values(any_tab[[var_name]],
                                          style[["number_formats"]][["pct_decimals"]])
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Tear apart the the summarised data frame by row and column combinations and
    # transpose columns to generate user defined combination. At the end put all
    # the pieces back together to form a fully printable result data frame.
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Transform table", "Transform")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Rearranging data for output table")

    part_combi_list       <- list()
    header_combi_list     <- list()
    col_header_df         <- list()
    row_header_dimensions <- list()
    last_number_of_rows   <- 0
    by_division           <- 1

    if (length(by) > 0){
        by_division <- length(collapse::funique(any_tab[["by_vars"]]))
    }

    # Underscores are not allowed in column and values variables, because when constructing the
    # table header later, the underscore is the symbol by which the variable names are split.
    # Having additional underscores would mess up this part and lead to errors. Therefor the
    # additional underscores will be temporarily replaced.
    rows       <- replace_except(rows,       "_", "!!!", extensions)
    columns    <- replace_except(columns,    "_", "!!!", extensions)
    value_vars <- replace_except(value_vars, "_", "!!!", extensions)
    value_sort <- replace_except(values,     "_", "!!!", extensions)
    names(any_tab)    <- replace_except(names(any_tab),    "_", "!!!", extensions)
    any_tab[["TYPE"]] <- replace_except(any_tab[["TYPE"]], "_", "!!!", extensions)

    # Sort type alphabetically to make finding the right combination easier
    any_tab[["TYPE"]] <- reorder_combination(any_tab[["TYPE"]])

    # Get number of column header variables by getting the maximum number of + signs in the
    # column variables.
    max_plus <- max(sapply(gregexpr("\\+", columns), function(var_to_test){
        if (var_to_test[1] == -1){
            1
        }
        else{
            length(var_to_test) + 1
        }}))

    index <- 1
    any_header <- NULL
    combined_id_vars <- NULL

    for (row_combi in rows){
        combined_col_df <- NULL

        # Get current single variables from row combination
        row_combi_vars <- collapse::funique(trimws(unlist(strsplit(row_combi, "\\+"))))

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # For every combination of row variables loop through all column variable
        # combinations
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        for (col_combi in columns){
            # Get current single variables from column combination
            col_combi_vars <- collapse::funique(trimws(unlist(strsplit(col_combi, "\\+"))))

            current_combi <- c(row_combi_vars, col_combi_vars)

            # Sort combination alphabetically
            sorted_combi <- paste(sort(current_combi), collapse = "+")

            # Keep only necessary variables
            if (length(by) > 0){
                current_combi <- c("BY", "by_vars", current_combi)
            }

            combi_df <- any_tab |> keep(current_combi, "TYPE", value_vars)

            # Combine variables to match TYPE variable and subset data frame by
            # current combination
            subset_type <- paste(sorted_combi, collapse = "+")

            # Sort to have the correct order after pivoting
            combi_df <- combi_df |>
                collapse::fsubset(TYPE == subset_type) |>
                data.table::setcolorder(current_combi, before = 1) |>
                data.table::setorderv(current_combi, na.last = TRUE)

            # When pre summarised data is used, it is possible to select types which
            # aren't present in the data frame. This leads to the data frame having no
            # observations at this point, which leads to an error down below. If this
            # happens, abort.
            if (collapse::fnrow(combi_df) == 0){
                print_message("ERROR", c("The variable combination of '[subset_type]' is not part of the provided data frame.",
										 "Tabulation will be aborted."), subset_type = subset_type)
                return(invisible(NULL))
            }

            # Rename the row variables to something neutral. In the next steps the
            # data frame is puzzled back together, but the thing is, that the row
            # variables can be individually in any order in the result table. If
            # this step would be omitted the variables would all be in a fixed order
            # in the result table.
            new_row_names <- paste0("var", seq_along(row_combi_vars))

            if (length(by) == 0){
                names(combi_df)[seq_along(row_combi_vars)] <- new_row_names

                id_vars <- new_row_names
            }
            else{
                end_col <- length(row_combi_vars) + 2
                names(combi_df)[3:end_col] <- new_row_names

                id_vars <- new_row_names
                id_vars <- c("BY", "by_vars", new_row_names)
            }

            #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            # Calculate block percentages
            #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

            # Make a copy of the value variables, because the vector is altered
            # during the block percentage calculation.
            pivot_values <- value_vars

            if (length(pct_block) > 0){
                #-------------------------------------------------------------------------#
                monitor_df <- monitor_df |> monitor_next("Block percentages", "Block percentages")
                #-------------------------------------------------------------------------#

                # Loop over rows and columns. The calculation is basically the same, but the
                # group variable selection is different.
                for (block in pct_block){
                    # On rows: select all  row variables but the last and all column variables
                    if (block == "rows"){
                        block_group  <- c(id_vars[-length(id_vars)], col_combi_vars)
                    }
                    # On columns: select all row variables and all column variables but the last
                    else{
                        block_group  <- c(id_vars, col_combi_vars[-length(col_combi_vars)])
                    }

                    # Get the sum variables and merge back the group maximums to the data frame
                    block_values <- pivot_values[endsWith(pivot_values, "_sum")]

                    combi_df <- suppressMessages(combi_df |>
                         summarise_plus(class      = block_group,
                                        values     = block_values,
                                        statistics = "max",
                                        notes      = FALSE,
                                        merge_back = TRUE))

                    # Run a do over loop to calculate all the block percentages for
                    # the different value variables.
                    new_pct  <- paste0(values, "_pct_block_", block)
                    max_vars <- paste0(block_values, "_max")

                    combi_df <- suppressMessages(combi_df |> if.(max_vars > 0, new_pct = block_values * 100 / max_vars))

                    # Add new percentages to values to pivot and remove the maximum variables
                    pivot_values <- c(pivot_values, new_pct)
                    combi_df     <- combi_df |> dropp(max_vars)
                }

                # Remove sum variables if they weren't specified in the statistics.
                # Also remove them from pivot vector.
                if (flag_remove_block_sum){
                    combi_df <- combi_df |> dropp(block_values)

                    pivot_values <- pivot_values[!pivot_values %in% block_values]
                }

                #-------------------------------------------------------------------------#
                monitor_df <- monitor_df |> monitor_next("Transform table", "Transform")
                #-------------------------------------------------------------------------#
            }

            #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            # Main transposition
            #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

            # If there is only one value provided and only one statistic selected
            # the pivoted variable names below only receive the column expressions
            # as names. The needed format is "value_stat_expression". In the mentioned
            # case this format is pre computed.
            if (length(values) == 1 && length(statistics) == 1){
                # If only group percentages are selected, but for multiple variables, then
                # the resulting variables are more than the given values, which isn't detected
                # with the condition above. This case has to be captured here, because it would
                # lead to wrong variable names and therefor a wrong header.
                if ((is.null(pct_group) || length(pct_group) <= 1) && !row_pct && !col_pct){
                    value_stat <- names(combi_df[collapse::fncol(combi_df)])

                    combi_df[[col_combi_vars[1]]] <-
                        paste0(value_stat, "_", combi_df[[col_combi_vars[1]]])
                }
            }
            # Fallback check if the user happens to input only wrong statistics but one.
            # In this case statistics is longer than 1 even though only one statistic was
            # computed. This leads to the condition above being omitted.
            else if (collapse::fncol(combi_df) - length(c(id_vars, col_combi_vars, "TYPE")) == 1){
                # If only group percentages are selected, but for multiple variables, then
                # the resulting variables are more than the given values, which isn't detected
                # with the condition above. This case has to be captured here, because it would
                # lead to wrong variable names and therefor a wrong header.
                if ((is.null(pct_group) || length(pct_group) <= 1) && !row_pct && !col_pct){
                    value_stat <- names(combi_df[collapse::fncol(combi_df)])

                    combi_df[[col_combi_vars[1]]] <-
                        paste0(value_stat, "_", combi_df[[col_combi_vars[1]]])
                }
            }

            # Pivot to wider format, which basically is the final format to print the data
            combi_df <- combi_df |>
                collapse::pivot(id     = id_vars,
                                names  = col_combi_vars,
                                values = pivot_values,
                                how    = "wider")

			# Remove excess underscore. This is created if no column variables are specified
            if (length(col_combi_vars) == 1 && col_combi_vars == ".temp.var"){
                names(combi_df) <- gsub("_ $", "", names(combi_df))
            }

            combi_df[id_vars] <- lapply(combi_df[id_vars], as.character)

            # Replace NA values with text so that they can be differentiated from empty
            # row header columns later on.
            row_var_cols <- seq_along(id_vars)
            combi_df[, row_var_cols][is.na(combi_df[, row_var_cols])] <- style[["na_symbol"]]

            # Sort interleaved
            if (tolower(order_by) == "interleaved"){
                combi_df <- combi_df |> order_interleaved(statistics)
            }

            #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            # Join different column results together
            #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

            if (is.null(combined_col_df)){
                row_labels       <- c()

                # Loop through all provided labels
                for (variable in row_combi_vars){
                    # Revert back underscores
                    variable <- gsub("!!!", "_", variable)

                    # If there are no labels specified add the variable names as labels
                    if (!variable %in% names(var_labels)){
                        row_labels <- c(row_labels, variable)
                        next
                    }

                    # Replace stat texts with provided labels
                    row_labels <- c(row_labels, var_labels[[variable]])
                }

				# Remove empty row labels so that no unnecessary slashes are created
                row_labels <- row_labels[row_labels != ""]

                # Insert labels as new variable and sort it to the front
                if (!all(row_labels == "")){
                    combi_df[["row.label"]] <- paste(row_labels, collapse = " / ")
                }
                # In case only empty labels were provided, just add empty character
                # so that the column can be identified as empty later on.
                else{
                    combi_df[["row.label"]] <- " "
                }

                # Order row label to the front
                if (length(by) == 0){
                    combi_df <- combi_df |>
                        data.table::setcolorder("row.label", before = 1)
                }
                else{
                    combi_df <- combi_df |>
                        data.table::setcolorder(c("BY", "by_vars", "row.label"), before = 1)
                }

                new_row_names <- c("row.label", new_row_names)

                # First iteration
                combined_col_df <- combi_df
            }
            # Following iterations
            else{
                # Check for duplicate variable names. If any duplicate is found abort.
                duplicates <- intersect(names(combined_col_df),
                                        setdiff(names(combi_df), c(new_row_names, "row.label", "BY", "by_vars")))

                if (length(duplicates) > 0){
                    print_message("ERROR", c("Duplicate <columns> names found: [dup].",
											 "If you are working with original values, consider making them unique by using formats."),
								  dup = paste(duplicates, collapse = ", "))
                    return(invisible(NULL))
                }

                # Join current data frame to the iterations before
                combined_col_df <- collapse::join(combined_col_df, combi_df,
                                                  on       = id_vars,
                                                  how      = "left",
                                                  verbose  = FALSE,
												  overid   = 2)

                # Remove variables which define the row labels so that they don't appear
                # when combining the column header below.
                combi_df <- suppressMessages(combi_df |>
                    dropp(new_row_names, "row.label", "BY", "by_vars"))
            }

            #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            # Build variable name table header for later use during formatting.
            # Column header only needs to be built once because other iterations
            # would just produce the same header.
            #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

            if (index == 1){
                combined_id_vars <- id_vars

                # Get data frame variable names
                col_header_df <- combi_df[0, ]

                # Fill header rows with column variable names
                for (var_index in seq_along(col_combi_vars)){
                    col_header_df[var_index, ] <- gsub("!!!", "_", col_combi_vars[[var_index]])
                }

                # If the header has fewer rows than the maximum header rows, fill up the header
                # with additional empty rows.
                header_diff <- max_plus - collapse::fnrow(col_header_df)
                if (header_diff > 0){
                    col_header_df <- rbind(col_header_df,
                                           stats::setNames(data.table::as.data.table(
                                               matrix("", header_diff, collapse::fncol(col_header_df))),
                                               names(col_header_df)))
                }

                # First iteration
                if (is.null(any_header)){
                    # Remove row header columns from column header data frame
                    row_header_var_count <- length(id_vars) + 1
                    any_header <- col_header_df[, -(1:row_header_var_count), drop = FALSE]
                }
                # Following iterations
                else{
                    # cbind current header data frame to the iterations before
                    any_header <- cbind(any_header, col_header_df)
                }
            }
            else{
                # Gather row header id variables
                combined_id_vars <- union(combined_id_vars, id_vars)
            }
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Combine data and store information
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Store combined column data frame in a list for later rbind
        part_combi_list[[row_combi]] <- combined_col_df

        # Get table dimensions for later use during formatting.
        name <- paste0("header_end_", paste(row_combi_vars, collapse = "%%%"))
        last_number_of_rows           <- (collapse::fnrow(combined_col_df) / by_division) + last_number_of_rows
        row_header_dimensions[[name]] <- last_number_of_rows

        name <- paste0("header_size_", paste(row_combi_vars, collapse = "%%%"))
        row_header_dimensions[[name]] <- length(row_combi_vars)

        index <- index + 1
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Put results together
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Put all computed data frames below each other to form a final result data frame.
    if (length(row_expand) == 0){
        any_tab <- data.table::rbindlist(part_combi_list, fill = TRUE, use.names = TRUE)
    }
    # If there are expanded brackets in the row combinations, also insert TYPE variable
    # back in to be able to sort the results.
    else{
        any_tab <- data.table::rbindlist(
            lapply(names(part_combi_list), function(combination){
                # Insert TYPE variable back in
                data_part           <- part_combi_list[[combination]]
                data_part[["TYPE"]] <- combination

                data_part
            }), fill = TRUE, use.names = TRUE)

        any_tab[["TYPE_ORIG"]] <- as.character(row_expand[any_tab[["TYPE"]]])
    }

    # If only specific variables should be kept per statistic, clean up the additional
    # not needed variables.
    if (!is.null(vars_per_stat_list)){
        variable_names <- names(any_header)

        # Generate search patterns like: variable name starts with list element
        # and somewhere in the variable name should appear the list element name.
        patterns <- sapply(names(vars_per_stat_list), function(name){
            elements <- vars_per_stat_list[[name]]
            paste0("^", elements, ".*", name, collapse = "|")
        })
        combined_pattern <- paste(patterns, collapse = "|")

        # Retrieve all variables to drop, which is the inverse vector of the patterns
        # that match. Drop these variables from the actual table and the header.
        vars_to_drop <- grep(combined_pattern, variable_names, value = TRUE, invert = TRUE)

        any_tab    <- suppressMessages(any_tab    |> dropp(vars_to_drop))
        any_header <- suppressMessages(any_header |> dropp(vars_to_drop))
    }

    # Reorder variables according to statistics
    if (tolower(order_by) == "stats" || tolower(order_by) == "values_stats" || tolower(order_by) == "blocks"){
        any_tab    <- any_tab    |> setcolorder_by_pattern(statistics)
        any_header <- any_header |> setcolorder_by_pattern(statistics)
    }

    # Reorder variables by provided values
    if (tolower(order_by) == "values" || tolower(order_by) == "values_stats"){
        any_tab    <- any_tab    |> setcolorder_by_pattern(value_sort)
        any_header <- any_header |> setcolorder_by_pattern(value_sort)
    }

    # Reorder variables by blocks depending on the root variable names
    if (tolower(order_by) == "blocks"){
        root_names <- collapse::funique(sub("_[^_]*$", "", names(any_header)))
        any_tab    <- any_tab    |> setcolorder_by_pattern(root_names)
        any_header <- any_header |> setcolorder_by_pattern(root_names)
    }

    if (length(order_by_vars) > 0){
        any_names <- names(any_tab)

        # Select all variables in blocks of the requested pattern
        ordered_cols <- unlist(lapply(order_by_vars, function(single_pattern){
            # Returns all variable names that match the provided pattern
            grep(single_pattern, any_names, value = TRUE, ignore.case = TRUE)
        }))

        # Put the ordered columns at the front of the data frame
        any_tab    <- any_tab    |> data.table::setcolorder(c(combined_id_vars, ordered_cols))
        any_header <- any_header |> data.table::setcolorder(ordered_cols)
    }
    else if (length(compute) > 0){
        any_names  <- replace_except(names(compute), "_", "!!!", extensions)

        any_tab    <- any_tab    |> setcolorder_by_pattern(any_names)
        any_header <- any_header |> setcolorder_by_pattern(any_names)
    }

    # After binding together the data frames it can happen, that some of the new var
    # variables end up at the end of the data frame instead of the front. Therefor
    # these columns need to be ordered to the front for safety.
    ordered_cols <- grep("^var[0-9]+$", names(any_tab), value = TRUE)
    any_tab <- any_tab |> data.table::setcolorder(c("row.label", ordered_cols), before = 1)

    # Sort row header columns if there are expanded combinations. What this does is
    # putting the combinations together in a way, so that what belongs together is
    # printed together. Meaning same combination roots are put below each other and
    # their cells get merged in the Excel table later.
    if (length(row_expand) > 0){
        # Sorting order: original not expanded group, all row header variables but the last
        #                expanded groups + last row header variable
        # This order keeps the root combinations together but the last variables
        # apart from another.
        sort_vars <- c("TYPE_ORIG", ordered_cols[-length(ordered_cols)],
                       "TYPE",      ordered_cols[length(ordered_cols)])

        # Get the unique values per variable in order of appearance. This is crucial
        # to keep the base order as is.
        levels_list <- lapply(any_tab[sort_vars], function(variable){
            collapse::funique(unlist(variable))
        })

        # Convert row header variables to factors
        for (variable in sort_vars){
            var_levels <- levels_list[[variable]]

            any_tab[[variable]] <- factor(any_tab[[variable]],
                                          levels  = var_levels,
                                          ordered = TRUE,
                                          exclude = NULL)
        }

        # Actual sorting
        data.table::setorderv(any_tab, sort_vars)

        # Convert back to character and drop TYPE variables
        any_tab[ordered_cols] <- lapply(any_tab[ordered_cols], as.character)
        any_tab <- any_tab |> collapse::fselect(-TYPE, -TYPE_ORIG)
    }

    # Get number of row header variables by getting the maximum number of + signs in the
    # row variables.
    max_plus <- max(sapply(gregexpr("\\+", rows), function(var_to_test){
        if (var_to_test[1] == -1){
            1
        }
        else{
            length(var_to_test) + 1
        }}))

    length_row_header <- max_plus + 1
    if (length(by) != 0){
        length_row_header <- length_row_header + 2
    }

    # Mark empty row header cells
    row_var_cols <- 1:length_row_header
    any_tab[, row_var_cols][is.na(any_tab[, row_var_cols])] <- ""

    # If all row labels are empty, delete the row header column
    if (all(any_tab[["row.label"]] == " ")){
        any_tab <- any_tab |> dropp("row.label")
    }
    # If only some labels are empty the row label column is printed. Problem: the cell
    # merging omits empty cells. Therefor convert empty cells into cells with a space
    # to merge the empty parts correctly.
    else{
        any_tab["row.label" == ""] <- " "
    }

    # Remove suppressed format expressions per do over for all row header variables
    # and drop column variables accordingly.
    any_tab <- suppressMessages(any_tab |>
                        if.(ordered_cols != "!:") |>
                        dropp(":_!:"))
    any_header <- suppressMessages(any_header |> dropp(":_!:"))

    # In between clean up to get a better overview
    rm(combi_df, combined_col_df, part_combi_list, col_combi, col_combi_vars,
       combinations, current_combi, index, last_number_of_rows, name, new_row_names,
       row_combi, row_combi_vars, sorted_combi, subset_type, group_vars, flag_remove_sum,
       length_row_header, col_header_df, header_diff, row_header_var_count, var_vector,
       var_name, var_index, value_sort, stat, row_pct, col_pct, pre_summed, ordered_cols,
       max_plus, id_vars, by_division, data_frame, pivot_values, extensions, combined_id_vars,
       order_by_vars)

    if (length(pct_block) > 0){
        rm(block_group, block_values, new_pct, max_vars)
    }

    # Grab all information, which is necessary to format the workbook. This list will be
    # returned at the end and can be grabbed by the workbook combine function.
    meta <- mget(c("rows", "columns", "statistics",
                   "by", "titles", "footnotes", "var_labels", "stat_labels",
                   "box", "any_header", "row_header_dimensions",
                   "style", "output", "na.rm", "print_miss"))

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Prepare table format for output
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Excel prepare", "Format")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Formatting table")

    # Setup styling in new workbook if no other is provided
    if (is.null(workbook)){
        workbook <- openxlsx2::wb_workbook() |>
            prepare_styles(list("title" = titles, "footnote" = footnotes), style)
    }
    # Update style options in provided workbook
    else{
        workbook <- workbook |>
            prepare_styles(list("title" = titles, "footnote" = footnotes), style)
    }

    monitor_df <- monitor_df |> monitor_end()

    # In case no by variables are provided
    if (length(by) == 0){
        wb_list <- format_any_excel(workbook, any_tab, rows, columns, statistics,
                                    by, titles, footnotes, var_labels, stat_labels,
                                    box, any_header,
                                    style, output, monitor_df = monitor_df)

        wb         <- wb_list[[1]]
        monitor_df <- wb_list[[2]]
    }
    # In case there are  by variables are provided
    else{
        if (!style[["by_as_subheaders"]]){
            wb_list <- format_any_by_excel(workbook, any_tab, rows, columns, statistics,
                                           by, titles, footnotes, var_labels, stat_labels,
                                           box, any_header,
                                           style, output, na.rm, print_miss, monitor_df)
        }
        else{
            wb_list <- format_by_single_table(workbook, any_tab, rows, columns, statistics,
                                              by, titles, footnotes, var_labels, stat_labels,
                                              box, any_header,
                                              style, output, na.rm, print_miss, monitor_df)
        }

        wb         <- wb_list[[1]]
        monitor_df <- wb_list[[2]]
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Output formatted table into different formats
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (print){
        print_step("MAJOR", "Output table")

        monitor_df <- monitor_df |> monitor_next("Output tables", "Output tables")

        # Check if only save path or file name is specified. If only one is specified
        # print a note. Otherwise the file would not be saved and opened without a
        # hint to why the file wasn't saved.
        if (is.null(style[["save_path"]]) + is.null(style[["file"]]) == 1){
            if (is.null(style[["save_path"]])){
                print_message("NOTE", c("No save path specified. Both save path and file name with extension",
                                        "need to be specified in the global options or style parameter for",
                                        "the file to be saved. File won't be saved."))
            }
            else{
                print_message("NOTE", c("No file name specified. Both save path and file name with extension",
                                        "need to be specified in the global options or style parameter for",
                                        "the file to be saved. File won't be saved."))
            }
        }

        # If no save path or file provided just open workbook
        if (is.null(style[["save_path"]]) || is.null(style[["file"]])){
            if (interactive()){
                wb$open()
            }
        }
        else{
            # If save path doesn't exist, just open workbook
            if (!file.exists(style[["save_path"]])){
                print_message("WARNING", "Path does not exist: [style]", style = style[["save_path"]])

                if (interactive()){
                    wb$open()
                }
            }
            # Save file
            else{
                extension <- tolower(tools::file_ext(style[["file"]]))

                # CSV export is based on the actual data frame
                if (extension == "csv"){
                    any_tab |> data.table::fwrite(file = paste0(style[["save_path"]], "/", style[["file"]]),
                                                  sep  = ";",
                                                  dec  = ",")
                }
                # While XLSX export is based on the workbook
                else if (extension == "xlsx"){
                    wb$save(file = paste0(style[["save_path"]], "/", style[["file"]]), overwrite = TRUE)
                }
                # If there is no extension or a wrong one, auto correct it
                else{
                    wb$save(file = paste0(style[["save_path"]], "/", basename(style[["file"]]), ".xlsx"), overwrite = TRUE)
                }
            }
        }
    }

    print_closing(10)

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_end()
    monitor_df |> monitor_plot(draw_plot = monitor)
    #-------------------------------------------------------------------------#

    invisible(structure(list("table"    = any_tab,
                             "workbook" = wb,
                             "meta"     = meta), class = "qol_table"))
}


###############################################################################
# Format any table for excel output
###############################################################################
#' Format Any Table Output (Excel Based)
#'
#' @description
#' Format any table with the provided row and column variables. Statistics
#' can be anything available.
#'
#' @param wb An already created workbook to add more sheets to.
#' @param any_tab The data frame which contains the information for this table.
#' @param rows The variable that appears in the table rows.
#' @param columns The variable that appears in the table columns.
#' @param statistics The user requested statistics.
#' @param by Separate the cross table output by the expressions of the provided variables.
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#' @param var_labels List which contains column variable names and their respective labels.
#' @param stat_labels List which contains statistic names and their respective labels.
#' @param box The text that should appear in the upper left box of the table.
#' @param any_header The column header carrying the variable names.
#' @param style A list containing the styling elements.
#' @param output Determines whether to style the output or to just quickly paste
#' the data.
#' @param by_info Text which contains the information which variable with which
#' expression is computed at the moment.Used for computation with by variables.
#' @param index Index of the current variable expression. Used for computation with
#' by variables.
#' @param monitor_df Data frame which stores the monitoring values.
#'
#' @return
#' Returns a list containing a formatted Excel workbook as well as the monitoring
#' data frame.
#'
#' @noRd
format_any_excel <- function(wb,
                             any_tab,
                             rows,
                             columns,
                             statistics,
                             by,
                             titles,
                             footnotes,
                             var_labels,
                             stat_labels,
                             box,
                             any_header,
                             style,
                             output,
                             by_info = NULL,
                             index   = NULL,
                             monitor_df){
    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_start("Excel prepare", "Format")
    #-------------------------------------------------------------------------#

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Setup header
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    if (is.null(by_info)){ print_step("MINOR", "Build header") }

    # Cut down percentage names to just "pct"
    names(any_tab) <- gsub("pct_group_", "pct group ",     names(any_tab))
    names(any_tab) <- gsub("pct_total",  "pct total",      names(any_tab))
    names(any_tab) <- gsub("pct_value",  "pct value",      names(any_tab))
    names(any_tab) <- gsub("pct_block_", "pct block ",     names(any_tab))
    names(any_tab) <- gsub("freq_g0",    "freq g0",        names(any_tab))
    names(any_tab) <- gsub("sum_wgt",    "weight_sum wgt", names(any_tab))

    # Build header from variable names
    multi_header <- build_multi_header(names(any_tab), any_header, var_labels, style)

    # Remove empty statistics rows, but keep multi_header with statistics because the information
    # is needed below for applying the correct number formats.
    column_header <- multi_header |> set_statistic_labels(stat_labels)
    column_header <- column_header[rowSums(column_header == "") != collapse::fncol(column_header), , drop = FALSE]

    stats_row <- multi_header[collapse::fnrow(multi_header), , drop = FALSE]

    # Get table ranges
    any_ranges <- get_any_tab_ranges(any_tab, column_header, stats_row,
                                     titles, footnotes, style)

    # Add empty columns to the header for the top left box at the beginning
    blank_columns <- matrix("", nrow = collapse::fnrow(column_header), ncol = any_ranges[["cat_col.width"]])
    column_header <- cbind(blank_columns, column_header)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Display additional table information
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Add box text
    if (box != ""){
        column_header[1, 1] <- box
    }
    # If no box text provided put in variable names of row headers
    else{
        column_header[1, 1] <- gsub("!!!", "_", paste(rows, collapse = "\n"))
    }

    # If function is called with by variables the sheet names have to be differentiated
    # and by info has to be written above the table.
    if (!is.null(by_info)){
        if (style[["sheet_name"]] == "by"){
            wb$add_worksheet(by,
                             grid_lines = style[["grid_lines"]])
        }
        else{
            wb$add_worksheet(paste0(style[["sheet_name"]], index),
                             grid_lines = style[["grid_lines"]])
        }
    }
    else{
        wb$add_worksheet(style[["sheet_name"]], grid_lines = style[["grid_lines"]])
    }

    # Rename the following stats back to match number formats in style element
    names(any_tab) <- gsub("sum.wgt", "sum_wgt", names(any_tab))

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Add table data and format according to style options
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    if (is.null(by_info)){ print_step("MINOR", "Insert data") }

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Excel data", "Format")
    #-------------------------------------------------------------------------#

    wb$add_data(x          = any_tab,
                start_col  = style[["start_column"]],
                start_row  = any_ranges[["table.row"]],
                col_names  = FALSE,
                na.strings = style[["na_symbol"]])


    # Add column header above table.
    wb$add_data(x          = column_header,
                start_col  = style[["start_column"]],
                start_row  = any_ranges[["header.row"]],
                col_names  = FALSE)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Table styling
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    if (is.null(by_info)){ print_step("MINOR", "Format titles and footnotes") }

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Excel titles/footnotes", "Format")
    #-------------------------------------------------------------------------#
    # Format titles and footnotes if there are any
    wb <- wb |>
        format_titles_foot_excel(titles, footnotes, any_ranges, style, output)

    # Only do the formatting when user specified it. With the excel_nostyle
    # option this whole part gets omitted to get a very quick unformatted
    # excel output.
    if (output == "excel"){
        if (is.null(by_info)){ print_step("MINOR", "Merge headers") }

        # Merge top left box
        wb$merge_cells(dims = any_ranges[["box_range"]])

        # Merge column and row headers
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel format col headers", "Format")
        #---------------------------------------------------------------------#
        wb <- wb |>
            handle_col_header_merge(column_header[, -c(1:any_ranges[["cat_col.width"]]), drop = FALSE], any_ranges, style)

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel format row headers", "Format")
        #---------------------------------------------------------------------#
        wb <- wb |>
            handle_row_header_merge(any_tab[, 1:any_ranges[["cat_col.width"]]], any_ranges)

        # Fill all cells with background color
        if (style[["background_color"]] != ""){
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Excel background", "Format")
            #-----------------------------------------------------------------#
            if (is.null(by_info)){ print_step("MINOR", "Fill background") }

            wb$add_fill(color = openxlsx2::wb_color(style[["background_color"]]))
            wb$set_cell_style_across(cols = "A:XFD", style = wb$get_cell_style(dims = "A1"))

            # Titles and footnotes need to get an extra fill
            wb$add_fill(dims = any_ranges[["title_range"]],    color = openxlsx2::wb_color(style[["background_color"]]))
            wb$add_fill(dims = any_ranges[["footnote_range"]], color = openxlsx2::wb_color(style[["background_color"]]))
        }

        # Style table
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel cell styles", "Format")
        #---------------------------------------------------------------------#
        if (is.null(by_info)){ print_step("MINOR", "Add cell styles") }

        wb <- wb |> handle_cell_styles(any_ranges, style)

        # Format subheaders, if present
        if (style[["by_as_subheaders"]]){
            if (is.null(by_info)){ print_step("MINOR", "Add subheaders") }

            # Get subheader ranges and the ones of the empty column header
            subheader_range     <- any_ranges[["subheader_range"]]
            pre_subheader_range <- any_ranges[["pre_subheader_range"]]

            # Merge subheaders and apply style
            for (sub_range in seq_along(subheader_range)){
                wb$add_data(x          = style[["subheaders"]][[sub_range]],
                            start_col  = any_ranges[["header.column"]] + any_ranges[["cat_col.width"]],
                            start_row  = any_ranges[["table.row"]] + (style[["subheader_rows"]][[sub_range]] - 1),
                            col_names  = FALSE,
                            na.strings = style[["na_symbol"]])
            }

            wb$merge_cells(dims = pre_subheader_range)
            wb$merge_cells(dims = subheader_range)
        }

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel number formats", "Format")
        #---------------------------------------------------------------------#
        if (is.null(by_info)){ print_step("MINOR", "Format numbers") }

        # Set up inner table number formats
        col_index <- 1
        chunks    <- sub("p[0-9]+$", "p", rle(stats_row)$values)

        for (type in chunks){
            wb$add_cell_style(dims                = any_ranges[[paste0("any_col_ranges", col_index)]],
                              apply_number_format = TRUE,
                              num_fmt_id          = wb$styles_mgr$get_numfmt_id(paste0(type, "_numfmt")))

            col_index <- col_index + 1
        }

        # Draw inner table cells as heat map with conditional formatting
        if (style[["as_heatmap"]]){
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Excel format heatmap", "Format")
            #-----------------------------------------------------------------#
            if (is.null(by_info)){ print_step("MINOR", "Add heatmap") }

            wb$add_conditional_formatting(dims  = any_ranges[["table_range"]],
                                          style = c(style[["heatmap_low_color"]],
                                                    style[["heatmap_middle_color"]],
                                                    style[["heatmap_high_color"]]),
                                          type  = "colorScale")
        }

        # Freeze headers. If both options are true they have to be set together, otherwise one
        # option would overwrite the other.
        if (style[["freeze_col_header"]] && style[["freeze_row_header"]]){
            wb$freeze_pane(first_active_col = any_ranges[["header.column"]] + any_ranges[["cat_col.width"]],
                           first_active_row = any_ranges[["table.row"]])
        }
        else if (style[["freeze_col_header"]]){
            wb$freeze_pane(first_active_col = any_ranges[["header.column"]] + any_ranges[["cat_col.width"]])
        }
        else if (style[["freeze_row_header"]]){
            wb$freeze_pane(first_active_row = any_ranges[["table.row"]])
        }

        # Adjust table dimensions
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel widths/heights", "Format")
        #---------------------------------------------------------------------#
        if (is.null(by_info)){ print_step("MINOR", "Adjust row heights and column widths") }

        wb <- wb |> handle_col_row_dimensions(any_ranges,
                                              collapse::fncol(any_tab),
                                              collapse::fnrow(any_tab) + collapse::fnrow(multi_header),
                                              style) |>
            handle_any_auto_dimensions(any_ranges, style) |>
            handle_header_table_dim(any_ranges, style)
    }

    wb$add_ignore_error(dims = any_ranges[["header_range"]],  number_stored_as_text = TRUE)
    wb$add_ignore_error(dims = any_ranges[["cat_col_range"]], number_stored_as_text = TRUE)
    wb$add_ignore_error(dims = any_ranges[["table_range"]],   number_stored_as_text = TRUE)

    wb$add_named_region(dims = any_ranges[["whole_tab_range"]], name = "table", local_sheet = TRUE)
    wb$add_named_region(dims = any_ranges[["table_range"]],     name = "data",  local_sheet = TRUE)

    monitor_df <- monitor_df |> monitor_end()

    # Return workbook
    list(wb, monitor_df)
}


#' Insert Statistic Labels
#'
#' @description
#' Give the statistics in the column header a custom label.
#'
#' @param column_header The complete column multi header.
#' @param stat_labels A list in which is specified which label should be printed for
#' which statistic instead of the statistic name.
#'
#' @return
#' Returns a multi layered column header with replaced statistic texts.
#'
#' @noRd
set_statistic_labels <- function(column_header, stat_labels){
    if (length(stat_labels) == 0){
        return(column_header)
    }

    # Loop through all provided labels
    for (i in seq_along(stat_labels)){
        name  <- names(stat_labels)[i]
        label <- stat_labels[[i]]

        # Omit label with missing variable name
        if (is.null(name) || name == ""){
            next
        }

        # Replace stat texts with provided labels
        column_header[collapse::fnrow(column_header), ] <- gsub(name, label, column_header[collapse::fnrow(column_header), ])
    }

    column_header
}


#' Insert Variable Labels
#'
#' @description
#' Give the variables in the column header a custom label.
#'
#' @param column_header The complete column multi header.
#' @param var_labels A list in which is specified which label should be printed for
#' which variable instead of the variable name.
#'
#' @return
#' Returns a multi layered column header with replaced variable texts.
#'
#' @noRd
set_col_variable_labels <- function(column_header, var_labels){
    if (length(var_labels) == 0){
        return(column_header)
    }

    # Order list by the length of the variable names (descending) to prevent that, if
    # the shortest variable name is at the top, and other variables follow with the
    # same base name (like: age, age1, age2), it would overwrite all other labels for
    # the following variables with the same base name.
    var_labels <- var_labels[order(nchar(names(var_labels)), decreasing = TRUE)]

    # Loop through all provided labels
    for (i in seq_along(var_labels)){
        name  <- names(var_labels)[i]
        label <- var_labels[[i]]

        # Omit label with missing variable name
        if (is.null(name) || name == ""){
            next
        }

        # Replace variable texts with provided labels
        column_header[,] <- gsub(paste0("^", name, "$"), label, as.matrix(column_header))
    }

    # If header only consists of one row it gets converted to a vector when using gsub above.
    # In case this happens, convert the vector back into a one row data frame.
    if (!is.matrix(column_header)){
        column_header <- data.table::as.data.table(as.list(column_header), stringsAsFactors = FALSE)
    }

    column_header
}


#' Build a Multi Header from Variable Names
#'
#' @description
#' Build a multi layered header from variable names of any_tab data frame. Analysis
#' variable name, stat and variable expressions are layerd into rows.
#'
#' @param var_names Variable names from any_tab.
#' @param any_header The column header carrying the variable names.
#' @param var_labels A list in which is specified which label should be printed for
#' which variable instead of the variable name.
#' @param style A list containing the styling elements.
#'
#' @return
#' Returns a data table with a multi layered header.
#'
#' @noRd
build_multi_header <- function(var_names,
                               any_header,
                               var_labels,
                               style){
    # Make sure no statistic is part of the variable labels
    extensions <- c("sum", "pctgroup", "pcttotal", "pctvalue", "pctblock", "pct",
                    "freqg0", "freq", "mean", "median", "mode", "min", "max", "first",
                    "last", "sumwgt", "p[0-9]+$", "sd", "variance", "missing")

    labels_to_drop <- grepl(paste(extensions, collapse = "|"), names(var_labels))

    var_labels <- var_labels[!labels_to_drop]

    # Replace variable texts with custom labels
    any_header      <- any_header |> set_col_variable_labels(var_labels)
    col_var_headers <- data.table::as.data.table(any_header, stringsAsFactors = FALSE)

    # Split up variable name into different parts
    header_parts <- strsplit(var_names, "_")
    max_parts    <- max(lengths(header_parts))

    # Remove row header variable names
    for (part in seq_along(header_parts)){
        if (length(header_parts[[part]]) > 1){
            break
        }

        header_parts[[part]] <- ""
    }

    # Loop through header parts
    header_matrix <- sapply(header_parts, function(parts){
        # Prepare empty header with max size
        multi_header <- rep("", max_parts)

        if (parts[1] != ""){
            # Identify different parts
            var_name    <- parts[1]
            stat        <- strsplit(parts[2], " ")[[1]][1]
            expressions <- parts[-c(1, 2)]

            # In case no column variables are provided, the expression will end
            # up as empty character vector. This is captured as empty character
            # of length 1, so that the header can be set up as usual.
            if (length(expressions) == 0){
                expressions <- ""
            }

            # Put multi header together: variable name, expressions, stat
            multi_header[1] <- var_name
            multi_header[seq(2, length(expressions) + 1)] <- expressions
            multi_header[max_parts] <- stat
        }

        multi_header
    })

    # Revert back underscores
    header_matrix <- gsub("!!!", "_", header_matrix)

    # Replace variable texts with custom labels
    header_matrix <- header_matrix |> set_col_variable_labels(var_labels)

    # Make sure header_matrix is treated as a matrix, even though there can be only one row
    if (is.null(dim(header_matrix))){
        dim(header_matrix) <- c(1, length(header_matrix))
    }

    # Drop completely empty rows
    if (collapse::fnrow(header_matrix) > 1){
        header_matrix <- data.table::as.data.table(
            header_matrix[, colSums(header_matrix != "" & header_matrix != " ") > 0, drop = FALSE])
    }
    else{
        header_matrix <- data.table::as.data.table(
            header_matrix[, colSums(header_matrix != "" & header_matrix != " "), drop = FALSE])
    }

    # Replace NA values
    header_matrix[is.na(header_matrix) | header_matrix == "NA"] <- style[["na_symbol"]]

    # Inject column variable headers into the multi header
    merged_header <- merge_headers(header_matrix, col_var_headers)

    # Get the more detailed blocks
    if (length(var_labels) > 0 && any(startsWith(tolower(names(var_labels)), "block"))){
        start_var <- (length(var_names) - collapse::fncol(merged_header)) + 1

        # Get the root variable names to count the number of group elements
        root_names    <- sub("^([^_]+_[^_]+)_.*$", "\\1", var_names[start_var:length(var_names)])
        block_lengths <- rle(root_names)[["lengths"]]

        # Generate block headlines
        block_labels <- var_labels[startsWith(tolower(names(var_labels)), "block")]
        block_labels <- fill_or_trim(block_labels, length(block_lengths))
        block_labels <- rep(block_labels, times = block_lengths)
        block_labels <- data.table::as.data.table(as.list(block_labels), stringsAsFactors = FALSE)

        # overwrite names of first table
        block_labels <- stats::setNames(block_labels, names(merged_header))

        # now rbind
        merged_header <- data.table::rbindlist(list(block_labels, merged_header), use.names = TRUE)
    }

    merged_header
}


#' Inject Variable Header into Multi Header
#'
#' @description
#' The multi layered headers based on variable values and variable names are generated
#' separately and need to be put together afterwards.
#'
#' @param value_header Header based on variable values.
#' @param variable_header Header based on variable names which will be injected.
#'
#' @return
#' Returns a data table with a complete multi layered header.
#'
#' @noRd
merge_headers <- function(value_header, variable_header){
    row_list <- list()

    # Loop over the value header rows to inject the variable headers into them
    for (i in seq_len(collapse::fnrow(value_header))){
        # Take a row from the value header
        row_list[[length(row_list) + 1]] <- value_header[i, , drop = FALSE]

        # If there are still rows in the variable header check whether to inject them
        if (i <= collapse::fnrow(variable_header)){
            variable_row <- variable_header[i, , drop = FALSE]

            # Only inject variable row, if the row isn't completely empty
            if (!all(variable_row == "")){
                row_list[[length(row_list) + 1]] <- variable_row
            }
        }
    }

    # Output as data table
    data.table::rbindlist(row_list, use.names = FALSE)
}


###############################################################################
# Format grouped by tables for excel output
###############################################################################
#' Format Any Table Output with by Variables (Excel Based)
#'
#' @description
#' Format any table with the provided row and column variables. Data frame is split
#' by by-variables, each table gets printed on a different sheet.
#'
#' @param wb An already created workbook to add more sheets to.
#' @param any_tab The data frame which contains the information for this table.
#' @param rows The variable that appears in the table rows.
#' @param columns The variable that appears in the table columns.
#' @param statistics The user requested statistics.
#' @param by Separate the cross table output by the expressions of the provided variables.
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#' @param var_labels List which contains column variable names and their respective labels.
#' @param stat_labels List which contains statistic names and their respective labels.
#' @param box The text that should appear in the upper left box of the table.
#' @param any_header The column header carrying the variable names.
#' @param style A list containing the styling elements.
#' @param output Determines whether to style the output or to just quickly paste
#' the data.
#' @param na.rm If TRUE removes all NA values from the tabulation.
#' @param print_miss FALSE by default. If TRUE outputs all possible categories of the
#' grouping variables based on the provided formats, even if there are no observations
#' for a combination.
#' @param monitor_df Data frame which stores the monitoring values.
#'
#' @return
#' Returns a list containing a formatted Excel workbook as well as the monitoring
#' data frame.
#'
#' @noRd
format_any_by_excel <- function(wb,
                                any_tab,
                                rows,
                                columns,
                                statistics,
                                by,
                                titles,
                                footnotes,
                                var_labels,
                                stat_labels,
                                box,
                                any_header,
                                style,
                                output,
                                na.rm,
                                print_miss,
                                monitor_df){

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Loop through all by variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    index <- 1

    for (by_var in by){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_start(paste0("Excel prepare (", by_var, ")"), "Format by")
        #---------------------------------------------------------------------#

        # Select by variables one by one
        any_by <- any_tab |>
            collapse::fsubset(any_tab[["BY"]] == by_var)

        # Extract unique values
        if (anyNA(any_by[["by_vars"]])){
            values <- c(collapse::funique(collapse::na_omit(any_by[["by_vars"]])), NA)
        }
        else{
            values <- collapse::funique(any_by[["by_vars"]])
        }

        monitor_df <- monitor_df |> monitor_end()

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Loop through all unique values to generate tables per expression
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        for (value in values){
            # In case NAs are removed
            if (is.na(value) && na.rm){
                next
            }

            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_start(paste0("Excel (", by_var, "_", value, ")"), "Format by")
            #-----------------------------------------------------------------#
            print_step("MINOR", "[by] = [value]", by = by_var, value = value)

            # Put additional by info together with the information which by variable
            # and which value is currently filtered.
            by_info <- paste0(by_var, " = ", value)

            # Filter table by current by variable and value
            if (!is.na(value)){
                any_temp <- any_by |>
                    collapse::fsubset(any_by[["by_vars"]] == value)
            }
            else{
                any_temp <- any_by |>
                    collapse::fsubset(is.na(any_by[["by_vars"]]))
            }

            any_temp <- any_temp |> dropp("BY", "by_vars")

            # Add by info below the titles
            if (length(titles) > 0){
                titles_temp <- c(titles, "", by_info)
            }
            # Or on top if there are no titles
            else{
                titles_temp <- by_info
            }

            # The print_miss option enables a shortcut in formatting the sheets after
            # the first one. Since it guarantees that all follow up sheets are printed
            # with the exact same table width and height, because all categories are
            # printed, only the first sheet must be formatted. All other sheets can
            # clone the entire style from the first sheet.
            if (index == 1 || !print_miss){
                # Generate tables as normal but base is filtered data frame
                wb_list <- format_any_excel(wb,
                                            any_temp,
                                            rows,
                                            columns,
                                            statistics,
                                            value,
                                            titles_temp,
                                            footnotes,
                                            var_labels,
                                            stat_labels,
                                            box,
                                            any_header,
                                            style,
                                            output,
                                            by_info,
                                            index,
                                            NULL)

            }
            # Clone style for follow up iterations
            else{
                # Generate tables with no style but base is filtered data frame
                wb_list <- format_any_excel(wb,
                                            any_temp,
                                            rows,
                                            columns,
                                            statistics,
                                            value,
                                            titles_temp,
                                            footnotes,
                                            var_labels,
                                            stat_labels,
                                            box,
                                            any_header,
                                            style,
                                            "excel_nostyle",
                                            by_info,
                                            index,
                                            NULL)

                # Clone entire style from first sheet
                wb_list[[1]]$clone_sheet_style(from = 1, to = index)
            }

            index <- index + 1

            wb <- wb_list[[1]]

            monitor_df <- monitor_df |> monitor_end()
        }

        monitor_df <- monitor_df |> monitor_end()
    }

    # Return workbook
    list(wb, monitor_df)
}


###############################################################################
# Format grouped by tables for excel output in one table
###############################################################################
#' Format Any Table Output with by Variables On One Sheet (Excel Based)
#'
#' @description
#' Format any table with the provided row and column variables. Data frame is printed
#' as one on one sheet. By-variables are used as subheaders.
#'
#' @param wb An already created workbook to add more sheets to.
#' @param any_tab The data frame which contains the information for this cross
#' table.
#' @param rows The variable that appears in the table rows.
#' @param columns The variable that appears in the table columns.
#' @param statistics The user requested statistics.
#' @param by Separate the cross table output by the expressions of the provided variables.
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#' @param var_labels List which contains column variable names and their respective labels.
#' @param stat_labels List which contains statistic names and their respective labels.
#' @param box The text that should appear in the upper left box of the table.
#' @param any_header The column header carrying the variable names.
#' @param style A list containing the styling elements.
#' @param output Determines whether to style the output or to just quickly paste
#' the data.
#' @param na.rm If TRUE removes all NA values from the tabulation.
#' @param print_miss FALSE by default. If TRUE outputs all possible categories of the
#' grouping variables based on the provided formats, even if there are no observations
#' for a combination.
#' @param monitor_df Data frame which stores the monitoring values.
#'
#' @return
#' Returns a list containing a formatted Excel workbook as well as the monitoring
#' data frame.
#'
#' @noRd
format_by_single_table <- function(wb,
                                   any_tab,
                                   rows,
                                   columns,
                                   statistics,
                                   by,
                                   titles,
                                   footnotes,
                                   var_labels,
                                   stat_labels,
                                   box,
                                   any_header,
                                   style,
                                   output,
                                   na.rm,
                                   print_miss,
                                   monitor_df){
    #---------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_start("Insert subheaders", "Format by")
    #---------------------------------------------------------------------#

    # Store the variable names for later
    var_names <- names(any_tab)

    # Get the unique by variable expressions as subheaders
    subheaders <- any_tab|>
        collapse::fselect("by_vars")|>
        collapse::funique("by_vars")

    style[["subheaders"]] <- subheaders[[1]]

    # Get number of row header variables by determining the number of character columns
    variable_types        <- vapply(any_tab, is.character, logical(1))
    number_of_row_headers <- which(!variable_types)[1] - 1
    number_of_values      <- collapse::fncol(any_tab) - number_of_row_headers

    # Split up data frame on by variable expressions
    grouping   <- collapse::GRP(any_tab, "by_vars", sort = FALSE)
    split_list <- any_tab |> collapse::gsplit(g = grouping, use.g.names = TRUE)

    # Insert by variable expressions as subheadings
    for (i in seq_along(split_list)){
        # Get subheader name and bind together as data frame row
        subheading <- split_list[[i]][1, "by_vars"]

        # Generate empty subheader line in separate parts to ensure that the value
        # columns will be set as numeric.
        sub_part1 <- as.list(rep(" ", number_of_row_headers))
        sub_part2 <- as.list(rep(NA_real_, number_of_values))

        # Combine both parts into data frame
        subheading_df <- data.table::as.data.table(c(sub_part1, sub_part2))

        # Put subheader above the initial data frame
        split_list[[i]] <- data.table::rbindlist(list(subheading_df, split_list[[i]]), use.names = FALSE)

        # Insert a helper column which indicates where in the data frame the headline sits.
        # This is later used to identify the subheader rows in the full data frame
        split_list[[i]][[".subheader"]]  <- 0
        split_list[[i]][1, ".subheader"] <- 1
    }

    # Bind full data frame back together
    any_tab <- data.table::rbindlist(split_list, use.names = FALSE)

    names(any_tab) <- c(var_names, ".subheader")
    any_tab <- any_tab |> collapse::fselect(-BY, -by_vars)

    # Set row numbers of subheaders in the helper column
    any_tab[[".subheader"]][any_tab[[".subheader"]] == 1] <- which(any_tab[[".subheader"]] == 1)

    subheader_rows <- any_tab |>
        collapse::fselect(".subheader") |>
        collapse::funique(".subheader") |>
        collapse::fsubset(.subheader != 0)

    style[["subheader_rows"]] <- subheader_rows[[1]]

    # Generate tables as normal but base is filtered data frame
    any_tab <- any_tab |> dropp(".subheader")

    monitor_df <- monitor_df |> monitor_end()

    wb_list <- format_any_excel(wb,
                                any_tab,
                                rows,
                                columns,
                                statistics,
                                NULL,
                                titles,
                                footnotes,
                                var_labels,
                                stat_labels,
                                box,
                                any_header,
                                style,
                                output,
                                monitor_df = monitor_df)

    # Return workbook
    list(wb_list[[1]], wb_list[[2]])
}


###############################################################################
# Combine multiple tables into one workbook
###############################################################################
#' Combine Multiple Tables Into One Workbook
#'
#' @description
#' Combines any number of tables created with [any_table()] into one workbook
#' and styles them according to their meta information.
#'
#' @param ... Provide any number of result lists output by [any_table()].
#' @param file If NULL, opens the output as temporary file. If a filename with path
#' is specified, saves the output to the specified path.
#' @param output The following output formats are available: excel and excel_nostyle.
#' @param print TRUE by default. If TRUE prints the output, if FALSE doesn't print anything. Can be used
#' if one only wants to catch the combined workbook.
#' @param monitor FALSE by default. If TRUE outputs two charts to visualize the functions time consumption.
#'
#' @return
#' A fully styled workbook containing the provided tables.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#' my_data[["person"]] <- 1
#'
#' # Formats
#' age. <- discrete_format(
#'     "Total"          = 0:100,
#'     "under 18"       = 0:17,
#'     "18 to under 25" = 18:24,
#'     "25 to under 55" = 25:54,
#'     "55 to under 65" = 55:64,
#'     "65 and older"   = 65:100)
#'
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' education. <- discrete_format(
#'     "Total"            = c("low", "middle", "high"),
#'     "low education"    = "low",
#'     "middle education" = "middle",
#'     "high education"   = "high")
#'
#' # Define style
#' set_style_options(column_widths = c(2, 15, 15, 15, 9))
#'
#' # Define titles and footnotes. If you want to add hyperlinks you can do so by
#' # adding "link:" followed by the hyperlink to the main text.
#' set_titles("This is title number 1 link: https://cran.r-project.org/",
#'            "This is title number 2 cell: W22",
#'            "This is title number 3 file: C:/MyFolder/MyFile.docx",
#'            "This is title number 4")
#' set_footnotes("This is footnote number 1 cell: W22",
#'               "This is footnote number 2 file: C:/MyFolder/MyFile.docx",
#'               "This is footnote number 3 link: https://cran.r-project.org/",
#'               "This is footnote number 4")
#'
#' # Catch the output and additionally use the options:
#' # print = FALSE and output = "excel_nostyle".
#' # This skips the styling and output part, so that the function runs faster.
#' # The styling is done later on.
#' set_print(FALSE)
#' set_output("excel_nostyle")
#' set_style_options(sheet_name = "big table")
#'
#' tab1 <- my_data |> any_table(rows       = c("sex + age", "sex", "age"),
#'                              columns    = c("year", "education + year"),
#'                              values     = weight,
#'                              statistics = c("sum", "pct_group"),
#'                              pct_group  = c("sex", "age", "education", "year"),
#'                              formats    = list(sex = sex., age = age.,
#'                                                education = education.),
#'                              na.rm      = TRUE)
#'
#' set_style_options(sheet_name = "age_sex")
#'
#' tab2 <- my_data |> any_table(rows       = "age",
#'                              columns    = "sex",
#'                              values     = weight,
#'                              statistics = "sum",
#'                              formats    = list(sex = sex., age = age.),
#'                              na.rm      = TRUE)
#'
#' set_style_options(sheet_name = "edu_year")
#'
#' tab3 <- my_data |> any_table(rows       = "education",
#'                              columns    = "year",
#'                              values     = weight,
#'                              statistics = "pct_group",
#'                              formats    = list(education = education.),
#'                              na.rm      = TRUE)
#'
#' # Every of the above tabs is a list, which contains the data table, an unstyled
#' # workbook and the meta information needed for the individual styling. These
#' # tabs can be input into the following function, which reads the meta information,
#' # styles each table individually and combines them as separate sheets into a single workbook.
#' combine_into_workbook(tab1, tab2, tab3)
#'
#' # Reset the global options afterwards
#' set_print(TRUE)
#' set_output("excel")
#'
#' @export
combine_into_workbook <- function(...,
                                  file    = NULL,
                                  output  = "excel",
                                  print   = TRUE,
                                  monitor = FALSE){
    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Prepare combine", "Prepare")
    #-------------------------------------------------------------------------#

    # Measure the time
    print_start_message()

    # Translate ... into a list if possible
    tables <- tryCatch({
        # Force evaluation to see if it exists
        list(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(tables)){
        print_message("ERROR", c("Unknown object found. Provide <any_table> or <export_with_style> results.",
								 "Combining tables to workbook will be aborted."))
        return(invisible(NULL))
    }

    for (table in tables){
        if (!is.list(table) || !all(c("table", "workbook", "meta") %in% names(table))){
            print_message("ERROR", c("Unknown object found. Provide <any_table> or <export_with_style> results.",
									 "Combining tables to workbook will be aborted."))
            return(invisible(NULL))
        }
    }

    tab_names <- as.character(substitute(list(...)))[-1]
    wb        <- openxlsx2::wb_workbook()

    i <- 1

    print_step("MAJOR", "Formatting tables")

    for (table in tables){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next(paste0("Format ", tab_names[i]), "Format tables")
        #---------------------------------------------------------------------#
        print_step("MINOR", "{tabs}", tabs = tab_names[i])

        meta <- table[["meta"]]

        wb <- wb |> prepare_styles(list("title" = meta[["titles"]], "footnote" = meta[["footnotes"]]), meta[["style"]])

        # Style data frame for export
        if (is.character(meta[[length(meta)]]) && meta[[length(meta)]] == "DATA"){
            wb_list <- suppressMessages(
                format_df_excel(wb, table[["table"]], meta[["titles"]], meta[["footnotes"]],
                                meta[["var_labels"]], meta[["style"]], meta[["column_align"]],
                                meta[["output"]], monitor_df))

            wb <- wb_list[[1]]
        }
        # Style any_table output for export
        else{
            # In case no by variables are provided
            if (length(meta[["by"]]) == 0){
                wb_list <- suppressMessages(
                    format_any_excel(wb, table[["table"]], meta[["rows"]], meta[["columns"]],
                                     meta[["statistics"]], meta[["by"]], meta[["titles"]],
                                     meta[["footnotes"]], meta[["var_labels"]], meta[["stat_labels"]],
                                     meta[["box"]], meta[["any_header"]],
                                     meta[["style"]], meta[["output"]], monitor_df = monitor_df))

                wb <- wb_list[[1]]
            }
            # In case there are by variables are provided
            else{
                wb_list <- suppressMessages(
                    format_any_by_excel(wb, table[["table"]], meta[["rows"]], meta[["columns"]],
                                        meta[["statistics"]], meta[["by"]], meta[["titles"]],
                                        meta[["footnotes"]], meta[["var_labels"]], meta[["stat_labels"]],
                                        meta[["box"]], meta[["any_header"]],
                                        meta[["style"]], meta[["output"]], meta[["na.rm"]], meta[["print_miss"]],
                                        monitor_df))

                wb <- wb_list[[1]]
            }
        }

        i <- i + 1
    }

    # Output formatted table into different formats
    if (print){
        print_step("MAJOR", "Exporting Excel Workbook")

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Output tables", "Output tables")
        #---------------------------------------------------------------------#

        if (is.null(file)){
            if (interactive()){
                wb$open()
            }
        }
        else if (!dir.exists(dirname(file)) || dirname(file) == "."){
            print_message("WARNING", "Directory '[file_name]' does not exist. File won't be saved.", file_name = dirname(file))

            if (interactive()){
                wb$open()
            }
        }
        else{
            wb$save(file = file, overwrite = TRUE)
        }
    }

    print_closing(10)

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_end()
    monitor_df |> monitor_plot(draw_plot = monitor)
    #-------------------------------------------------------------------------#

    invisible(wb)
}


###############################################################################
# Helper functions
###############################################################################
#' Make Space Between Plus Symbols Equal
#'
#' @description
#' Normalizes the space between plus symbols.
#'
#' @param combination Combination of variables connected with plus symbols.
#'
#' @return
#' Returns a normalized variable combination.
#'
#' @noRd
normalize_combination <- function(combination){
    # Remove extra spaces
    combination <- gsub("\\s*\\+\\s*", " + ", combination)
    combination <- gsub("\\s+", " ", combination)
    trimws(combination)
}


#' Expand Brackets To Generate Multiple Combinations
#'
#' @description
#' Expands variables passed in brackets to multiple full combinations.
#' "state + sex + (year age)" would become "state + sex + year" and
#' "state + sex + age".
#'
#' Can also expand combinations like "(state sex) + (year age)".
#'
#' @param combination Combination of variables connected with plus symbols.
#'
#' @return
#' Returns expanded variable combinations.
#'
#' @noRd
expand_combination <- function(combination){
    combination <- normalize_combination(combination)

    # Return combination as is, if there is no bracket to expand
    if (!grepl("\\(", combination)){
        return(stats::setNames(list(combination), combination))
    }

    # Find bracket groups
    matches <- gregexpr("\\([^\\)]+\\)", combination)
    groups  <- regmatches(combination, matches)[[1]]

    # Extract single variables
    variable_list <- lapply(groups, function(group){
        # Remove brackets
        content <- gsub("[()]", "", group)

        # Split only on commas
        terms <- strsplit(content, ",", fixed = TRUE)[[1]]

        # Trim whitespace around each alternative
        trimws(terms)
    })

    # Cartesian combinations
    combinations <- expand.grid(variable_list, stringsAsFactors = FALSE)

    # Create combination template to insert new combinations into
    template <- combination

    for (i in seq_along(groups)){
        template <- sub("\\([^\\)]+\\)", "%s", template)
    }

    # Put together the final expanded variable combinations
    results <- apply(combinations, 1, function(row){
        normalize_combination(do.call(sprintf, c(template, as.list(row))))
    })

    # mapping
    stats::setNames(as.list(rep(combination, length(results))), results)
}


#' Expand Brackets To Generate Multiple Combinations
#'
#' @description
#' Expands variables passed in brackets to multiple full combinations.
#' "state + sex + (year age)" would become "state + sex + year" and
#' "state + sex + age"
#'
#' @param combinations Combination vector of variables connected with plus symbols.
#'
#' @return
#' Returns expanded variable combination list.
#'
#' @noRd
expand_brackets <- function(combinations){
    # Expand all combinations
    result <- unlist(lapply(combinations, expand_combination), recursive = FALSE)

    # Remove duplicate combinations
    names <- result[collapse::funique(names(result))]
    list(names, any(grepl("\\(", unlist(names))))
}

#' Get Detailed Summary About A Data Frame
#'
#' @description
#' Prints a summary of a data frames contents, including details such as variable names,
#' types, unique values, missings and min/max values. It also tells you the number of
#' observations and variables present in the data frame, memory usage and the number
#' of duplicate observations.
#'
#' @param data_frame The data frame to get the content information from.
#' @param output The following output formats are available: console (default) or text.
#' @param monitor FALSE by default. If TRUE, outputs two charts to visualize the
#' functions time consumption.
#'
#' @details
#' [content_report()] is based on the 'SAS' procedure Proc Contents, which provides
#' a summary of global information one one hand like number of observations and variables among
#' many others and on the other hand shows per variable information like type and length.
#'
#' 'R' doesn't store the same information in a data frame like 'SAS', but there are many
#' useful information to get a quick overview of a data frame. With this function you don't
#' need to look at each variable individually. You can simply run it over a data frame and
#' get values for: number of unique values, missing values (absolute and relative), min and
#' max value as well as the top value.
#'
#' @return
#' Returns a list containing the global information as well as a data table containing
#' the per variable information.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(100)
#'
#' content_report(my_data)
#'
#' @export
content_report <- function(data_frame,
                           output  = "console",
                           monitor = FALSE){
    # Measure the time
    print_start_message()

    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Global information", "Global information")
    #-------------------------------------------------------------------------#

    ###########################################################################
    # Get global information
    ###########################################################################

    print_step("MAJOR", "Collecting global information")

    # Get the variable types as a vector
    column_types <- collapse::vtypes(data_frame)

    # Collect global information
    n_obs  <- collapse::fnrow(data_frame)
    n_vars <- collapse::fncol(data_frame)

    global_info <- list(
        name           = deparse(substitute(data_frame)),
        type           = class(data_frame)[[1]],
        n_obs          = n_obs,
        n_vars         = n_vars,
        n_cells        = n_obs * n_vars,
        duplicate_vars = get_duplicate_var_count(data_frame),
        duplicate_rows = collapse::fsum(collapse::fduplicated(data_frame)),
        type_counts    = collapse::fnobs(column_types, g = column_types),
        memory_usage   = format(utils::object.size(data_frame), units = "auto"))

    ###########################################################################
    # Per variable information
    ###########################################################################

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Type", "Per variable information")
    #-------------------------------------------------------------------------#

    print_step("MAJOR", "Collecting per variable information")

    variable_report <- data.table::data.table(
        pos = 1:collapse::fncol(data_frame),
        name = names(data_frame),
        type = column_types
    )

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Unique values", "Per variable information")
    #-------------------------------------------------------------------------#

    print_step("MINOR", "Number of unique values")

    variable_report[["unique"]] <- sapply(data_frame, data.table::uniqueN)

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Missing values", "Per variable information")
    #-------------------------------------------------------------------------#

    print_step("MINOR", "Number of missing values")

    variable_report[["na_count"]] <- sapply(data_frame, function(variable) collapse::fsum(is.na(variable)))
    variable_report[["na_pct"]]   <- sapply(data_frame, function(variable) round_values(collapse::fsum(is.na(variable))/length(variable) * 100, 1))

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Longest value", "Per variable information")
    #-------------------------------------------------------------------------#

    print_step("MINOR", "Longest value")

    variable_report[["longest"]] <- sapply(data_frame, get_max_length)

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Min/Max value", "Per variable information")
    #-------------------------------------------------------------------------#

    print_step("MINOR", "Min and max value")

    variable_report[["min_val"]] <- sapply(data_frame, format_value, type_func = collapse::fmin, type = "min")
    variable_report[["max_val"]] <- sapply(data_frame, format_value, type_func = max, type = "max")

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Top value", "Per variable information")
    #-------------------------------------------------------------------------#

    print_step("MINOR", "Top value")

    variable_report[["top_freq"]] <- sapply(data_frame, get_top_value)

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Create report", "Create report")
    #-------------------------------------------------------------------------#

    if (!is_no_print_active()){
        # Concatenate global information into one character
        headline  <- "--- DATA FRAME CONTENT REPORT "
        sub_head1 <- "... Variable Type Counts ....."
        sub_head2 <- "... Variable Details ........."

        console_width <- getOption("width") - nchar(headline)

        global_text <- c("",
                         paste0(headline, strrep("-", console_width)),
                         "",
                         sprintf("%-15s: %s", "Data frame",     global_info[["name"]]),
                         sprintf("%-15s: %s", "Type",           global_info[["type"]]),
                         sprintf("%-15s: %d", "Observations",   global_info[["n_obs"]]),
                         sprintf("%-15s: %d", "Variables",      global_info[["n_vars"]]),
                         sprintf("%-15s: %d", "Cells",          global_info[["total_cells"]]),
                         sprintf("%-15s: %s", "Memory Usage",   global_info[["memory_usage"]]),
                         sprintf("%-15s: %d", "Duplicate Vars", global_info[["duplicate_vars"]]),
                         sprintf("%-15s: %d", "Duplicate Rows", global_info[["duplicate_rows"]]),

                         "",
                         paste0(sub_head1, strrep(".", console_width)),
                         "",

                         utils::capture.output(global_info[["type_counts"]]),

                         "",
                         paste0(sub_head2, strrep(".", console_width)),
                         "\n")

        old <- options(max.print = .Machine$integer.max)
        on.exit(options(old), add = TRUE)

        # Print global information and data frame with per variable information
        if (output %in% c("console")){
            print_closing(10)

            cat(paste(global_text, collapse = "\n"))
            print.data.frame(variable_report)
        }
        else if (output == "text"){
            # Capture texts to print in variables
            global_text   <- utils::capture.output(cat(paste(global_text, collapse = "\n")))
            variable_text <- utils::capture.output(print.data.frame(variable_report))

            # Write texts to file and open editor
            temp_file <- tempfile(fileext = ".txt")
            writeLines(c(global_text, variable_text), temp_file, sep = "\n")

            if (interactive()){
                file.show(temp_file)
            }

            print_closing(10)
        }
    }

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_end()
    monitor_df |> monitor_plot(draw_plot = monitor)
    #-------------------------------------------------------------------------#

    # Return report list
    invisible(list(global    = global_info,
                   variables = variable_report))
}


#' Get Maximum Length Of A Variable
#'
#' @description
#' Get the maximum length of all variable types.
#'
#' @param variable Variable from which to get the length of the longest element.
#'
#' @return
#' Returns a numeric value
#'
#' @noRd
get_max_length <- function(variable){
    if (!any(!is.na(variable))){
        return(0)
    }

    # Just analyze the top 1000 because otherwise this would take to long
    variable <- variable[1:collapse::fmin(1000, collapse::fnrow(variable))]

    # Process doubles first because it should be more common that even integers are
    # saved as doubles.
    if (is.double(variable)){
        # Get unique values to avoid redundant formatting
        unique_values <- round_values(variable, 5) |>
            collapse::funique(sort = FALSE) |>
            collapse::na_omit()

        # Convert to character to count digits and decimal places
        chars <- format(unique_values, digits = 20, nsmall = 0, scientific = FALSE, trim = TRUE)

        # Remove trailing zeros to only get decimals if there really are any
        chars <- sub("\\.0+$", "", chars)
        chars <- sub("(\\.[0-9]*[1-9])0+$", "\\1", chars)

        return(collapse::fmax(collapse::vlengths(chars)))
    }

    # Process integers
    if (is.integer(variable)){
        # The negative sign will be ignored
        variable <- abs(as.numeric(variable))

        # log10(0) below is not defined, therefore get rid of all zeros
        variable[variable == 0] <- 1

        return(collapse::fmax(floor(log10(variable)) + 1))
    }

    # Process characters
    if (is.character(variable)){
        return(collapse::fmax(collapse::vlengths(variable), na.rm = TRUE))
    }

    # Fallback
    collapse::fmax(collapse::vlengths(as.character(variable)), na.rm = TRUE)
}


#' Get Minimum Or Maximum Value Of A Variable
#'
#' @description
#' Get the minimum or maximum value of a variable. For character variables retrieves
#' the alphabetically first or last entry.
#'
#' @param variable Variable from which to get the minimum or maximum value.
#' @param type_func Enter minimum or maximum function.
#' @param type The function type as character.
#'
#' @return
#' Returns a character.
#'
#' @noRd
format_value <- function(variable, type_func, type){
    if (collapse::allNA(variable)){
        return(NA_character_)
    }

    # For character values sort unique values and get first or last element of
    # the resulting vector.
    if (is.character(variable)){
        variable <- variable |>
            collapse::na_omit() |>
            collapse::funique() |>
            sort()

        if (type == "min"){
            return(variable[1])
        }
        else{
            return(variable[length(variable)])
        }
    }

    # Otherwise for numeric value get min or max value and return as character
    value <- type_func(variable, na.rm = TRUE)

    if (is.numeric(value)){
        return(as.character(round_values(value, 5)))
    }

    as.character(value)
}


#' Get The Top Value In Terms of Frequency
#'
#' @description
#' Get the value with the most number of observations.
#'
#' @param variable Variable from which to get the value with the top frequency.
#'
#' @return
#' Returns a character.
#'
#' @noRd
get_top_value <- function(variable){
    if (collapse::allNA(variable)){
        return("All NA")
    }

    # Round doubles, otherwise for doubles with decimals places this would result
    # in 0.0 % afterwards and would take forever
    if (is.double(variable)){
        variable <- round_values(variable, 3)
    }

    # Calculate frequencies
    counts <- collapse::fnobs(variable, g = variable)

    if (length(counts) == 0){
        return("NA")
    }

    # Check if all values are unique
    if (all(counts == 1) && length(variable) > 1){
        return("All unique")
    }

    # Evaluate top value percentage
    top_value <- names(counts)[which.max(counts)]
    top_freq  <- collapse::fmax(counts)
    pct       <- round_values((top_freq / length(variable)) * 100, 1)

    # Format percentage so that it can be displayed in a column with even width
    # despite uneven values.
    pct_formatted <- format(pct, width = 5, nsmall = 1, digits = 1)

    paste0(top_value, " (", pct_formatted, "%)")
}

#' Analyze R Scripts And Print Out Statistics
#'
#' @description
#' [code_statistics()] reads in a folder or entire folder structure, grabs all
#' 'R' script files and scans the contents for different patterns. It then outputs
#' a small report which shows of which parts the code consists.
#'
#' @param paths A full folder path in which there are 'R' scripts to be analyzed. Can
#' be a single path or a vector of paths.
#' @param recursive FALSE by default. If TRUE scans the provided paths recursive
#' for sub folders and the 'R' scripts within them.
#' @param output_per "overall" by default. Determines how detailed the results are
#' printed. Valid options are "overall", "folder" or "file".
#'
#' @return
#' Returns the results as a data frame.
#'
#' @examples
#' # Run this function on a directory, actually containing R script files
#' # to see some results.
#' code_statistics(tempdir())
#'
#' @export
code_statistics <- function(paths,
                            recursive  = FALSE,
                            output_per = "overall"){
    output_per <- tolower(output_per)

    if (!output_per %in% c("overall", "folder", "file")){
        print_message("WARNING", c("'[output_per]' is not a valid option. Available are 'overall', 'folder'",
                                   "or 'file'. 'overall' will be used."), output_per = output_per)

        output_per <- "overall"
    }

    # Loop through all the paths and analyze the R scripts. Gather the results
    # in a data frame and row bind everything together to a master data frame.
    result_list <- list()

    for (path in paths){
        if (!dir.exists(path)){
            print_message("WARNING", c("Path does not exist: [path]",
                                       "and will be skipped."), path = path)
            next
        }

        # Get all the R scripts from the path, scan them for patterns and add
        # the results to the main result data frame
        r_scripts   <- libname(path,
                               get_files  = TRUE,
                               recursive  = recursive,
                               extensions = "R")
        result_list <- c(result_list, scan_files(r_scripts))
    }

    # Row bind together all results to one data frame
    result_df <- data.table::rbindlist(result_list, use.names = TRUE)

    if (collapse::fnrow(result_df) == 1 && result_df[["total"]] == 0){
        return(invisible(NULL))
    }

    # Transpose data frame to make tabulation easier
    transpose_df <- result_df |>
        collapse::fselect(-total) |>
        collapse::pivot(c("path", "file")) |>
        collapse::fsubset(value > 0)

    # Create format for table output
    categories. <- suppressMessages(discrete_format(
        "Actual code lines"            = "actual_code",
        "Comments"                     = "comments",
        "Roxygen documentation"        = "documentation",
        "Blank lines"                  = "blank_line",
        "Only bracket lines"           = "bracket_line",
        "String only lines"            = "strings",
        "<- function() lines"          = "declared_functions",
        "Exported functions (@export)" = "exported_functions",
        "Internal functions (@noRd)"   = "internal_functions",
        "Unit tests"                   = "unit_tests"))

    # Print results
    if (output_per == "overall"){
        print_headline("[b]Overall code statistics[/b]")
        print_message("NEUTRAL", c("\nPaths that were scanned:", paths))

        suppressMessages(transpose_df |>
                        frequencies(variables = "variable",
                                    weight    = "value",
                                    formats   = list("variable" = categories.),
                                    style     = excel_output_style(number_formats = number_format_style(sum_decimals = 0))))
    }
    else if (output_per == "folder"){
        subfolders <- collapse::funique(transpose_df[["path"]])

        print_headline("[b]Code statistics per folder[/b]")
        print_message("NEUTRAL", c("\nPaths that were scanned:", subfolders))

        suppressMessages(transpose_df |>
                        frequencies(variables = "variable",
                                    by        = "path",
                                    weight    = "value",
                                    formats   = list("variable" = categories.),
                                    style     = excel_output_style(number_formats = number_format_style(sum_decimals = 0))))
    }
    else if (output_per == "file"){
        subfolders <- collapse::funique(transpose_df[["path"]])
        files      <- collapse::funique(transpose_df[["file"]])

        print_headline("[b]Code statistics per file[/b]")
        print_message("NEUTRAL", c("\nPaths that were scanned:", subfolders))
        print_message("NEUTRAL", c("\nFiles that were scanned:", files))

        suppressMessages(transpose_df |>
                        frequencies(variables = "variable",
                                    by        = "file",
                                    weight    = "value",
                                    formats   = list("variable" = categories.),
                                    style     = excel_output_style(number_formats = number_format_style(sum_decimals = 0))))
    }

    result_df
}


#' @description
#' Analyzes the provided 'R' script files and scans the contents for different patterns.
#'
#' @param files Vector of full paths to 'R' script files.
#'
#' @noRd
scan_files <- function(files){
    # If no R files are present in the currently scanned folder, just return an
    # empty data frame.
    if (length(files) == 0){
        return(list(data.frame(total              = 0,
                               actual_code        = 0,
                               comments           = 0,
                               documentation      = 0,
                               blank_lines        = 0,
                               bracket_line       = 0,
                               strings            = 0,
                               declared_functions = 0,
                               exported_functions = 0,
                               internal_functions = 0,
                               unit_tests         = 0)))
    }

    # Capture the different patterns from the script file
    file_statistics <- lapply(files, function(file){
        # Read in complete file and trim all blanks
        trimmed_lines <- trimws(readLines(file, warn = FALSE))

        # Total number of code lines
        total_lines <- length(trimmed_lines)

        # Empty lines and empty roxygen lines
        blank_line_ids <- trimmed_lines == "" | grepl("^#'\\s*$", trimmed_lines)

        # Exported and internal functions based on the roxygen tags
        exported_ids <- grepl("^#'\\s*@export\\s*$", trimmed_lines)
        internal_ids <- grepl("^#'\\s*@noRd\\s*$",   trimmed_lines)

        # Actual roxygen documentation. Roxygen lines carrying text which is not
        # already included in one of the above patterns.
        documentation_ids <- grepl("^#'\\s*.+", trimmed_lines) &
                                   !exported_ids & !internal_ids & !blank_line_ids

        # Regular comments without roxygen comments
        comment_ids <- grepl("^#", trimmed_lines) & !grepl("^#'", trimmed_lines)

        # Bracket only lines and constructs such as "} else {". Keep this as a
        # mutually exclusive line class, so trailing comments are included here.
        bracket_ids <- grepl("^[][{}(),[:space:]]+([#].*)?$|^}[[:space:]]*else[[:space:]]*\\{?([#].*)?$",
                             trimmed_lines)

        # Pure character lines
        string_ids <- grepl('^".*"$', trimmed_lines) | grepl("^'.*'$", trimmed_lines)

        # The actual function declaration lines without the ones happening in the examples
        # on roxygen lines.
        function_ids <- grepl("<-[[:space:]]*function[[:space:]]*\\(", trimmed_lines) & !documentation_ids

        # Unit test expectations
        unit_test_ids <- grepl("\\b(expect|expect_)[A-Za-z0-9_]*\\s*\\(", trimmed_lines)

        # The actual code is everything not already covered by the above patterns
        code_ids <- !(blank_line_ids |
                      documentation_ids |
                      exported_ids |
                      internal_ids |
                      comment_ids |
                      bracket_ids |
                      string_ids |
                      function_ids |
                      unit_test_ids)

        data.table::data.table(path               = dirname(file),
                               file               = basename(file),
                               total              = total_lines,
                               actual_code        = collapse::fsum(code_ids),
                               comments           = collapse::fsum(comment_ids),
                               documentation      = collapse::fsum(documentation_ids),
                               blank_line         = collapse::fsum(blank_line_ids),
                               bracket_line       = collapse::fsum(bracket_ids),
                               strings            = collapse::fsum(string_ids),
                               declared_functions = collapse::fsum(function_ids),
                               exported_functions = collapse::fsum(exported_ids),
                               internal_functions = collapse::fsum(internal_ids),
                               unit_tests         = collapse::fsum(unit_test_ids))
    })

    file_statistics
}

# Print code statistics
temp_file <- tempfile(fileext = ".R")

writeLines("print('Hello World')", con = temp_file)

result_df <- code_statistics(tempdir())
expect_true(result_df[["total"]] == 1, info = "Print code statistics")

result_df <- code_statistics(tempdir(), output_per = "folder")
expect_true(result_df[["total"]] == 1, info = "Print code statistics")

result_df <- code_statistics(tempdir(), output_per = "file")
expect_true(result_df[["total"]] == 1, info = "Print code statistics")

unlink(temp_file)

#' Convert Function Arguments to Character Vector
#'
#' @name convert_arguments
#'
#' @description
#' [args_to_char()]: Converts any argument passed as a single character or symbol as well
#' as character vectors or vector of symbols back as character vector.
#'
#' @param argument Function argument to convert.
#'
#' @return
#' Returns a character vector.
#'
#' @examples
#' # Example function with function parameter
#' print_vnames <- function(parameter){
#'     var_names <- args_to_char(substitute(parameter))
#'     print(var_names)
#' }
#'
#' print_vnames(age)
#' print_vnames("age")
#' print_vnames(c(age, sex, income, weight))
#' print_vnames(c("age", "sex", "income", "weight"))
#'
#' # You can also pass in a character vector, if you have stored variable names elsewhere
#' var_names <- c("age", "sex", "income", "weight")
#' print_vnames(var_names)
#'
#' # If you plan to use the function within other functions, better use get_origin_as_char()
#' print_vnames <- function(parameter){
#'     var_names <- get_origin_as_char(parameter, substitute(parameter))
#'     print(var_names)
#' }
#'
#' another_function <- function(parameter){
#'     print_vnames(parameter)
#' }
#'
#' another_function("age")
#' another_function(c("age", "sex", "income", "weight"))
#'
#' # Example function with ellipsis
#' print_vnames <- function(...){
#'     var_names <- dots_to_char(...)
#'     print(var_names)
#' }
#'
#' print_vnames(age)
#' print_vnames("age")
#' print_vnames(age, sex, income, weight)
#' print_vnames("age", "sex", "income", "weight")
#'
#' # You can also pass in a character vector, if you have stored variable names elsewhere
#' var_names <- c("age", "sex", "income", "weight")
#' print_vnames(var_names)
#'
#' @rdname convert_arguments
#'
#' @export
args_to_char <- function(argument){
    if (is.null(argument)){
        return(NULL)
    }

    # Single Character: Just return as is
    if (is.character(argument)){
        return(argument)
    }

    # Single symbol: Convert to character and return
    if (is.symbol(argument)){
        return(as.character(argument))
    }

    # Vector/List: If argument is a character vector or a vector of symbols,
    # c("year", "age") or c(year, age), then convert all elements to character.
    if (is.call(argument) &&  (identical(argument[[1]], quote(c)) ||
                               identical(argument[[1]], quote(list)))){
        return(vapply(argument[-1], as.character, character(1)))
    }

    # If there is a ":" in the call, this returns the call as is so that it can be used
    # as in e.g. keep/dropp to make use of ranges.
    if (is.call(argument) && identical(argument[[1]], quote(`:`))){
        return(paste(as.character(argument[[2]]),
                     as.character(argument[[3]]),
                     sep = ":"))
    }

    stop(" X ERROR: Something went wrong with the argument conversion.\n",
         "          Only single character and symbols, as well as vectors and flat lists are allowed.\n",
         "          Function will be aborted.")
}


#' Convert Ellipsis to Character Vector
#'
#' @description
#' [dots_to_char()]: When you define a function and want the user to be able to pass variable
#' names without the need to have them stored in a vector c() or list() beforehand and without
#' putting the names into quotation marks, you can convert this variable list passed as ...
#' into a character vector.
#'
#' Note: If the user passes a list of characters it is returned as given.
#'
#' @param ... Used for variable names listed in ... without the need to put them in c() or list().
#'
#' @rdname convert_arguments
#'
#' @export
dots_to_char <- function(...){
    expressions <- as.list(substitute(list(...)))[-1]

    # Resolve the different expressions into a character vector
    unlist(lapply(expressions, function(expression){
        # Character: Just return as is
        if (is.character(expression)){
            return(expression)
        }

        # Symbol: Try evaluation
        if (is.symbol(expression)){
            # Find the origin of the symbol by searching through the parent environments
            value <- get_origin_symbol(expression)

            if (is.character(value)){
                return(value)
            }

            return(as.character(expression))
        }

        # Calls: c(), list(), :
        args_to_char(expression)}), use.names = FALSE)
}


#' Get The Original Contents Of An Argument As Character
#'
#' @description
#' [get_origin_as_char()] is a wrapper that allows to retrieve the original contents
#' of the provided variable, whether called directly or nested in multiple function calls,
#' as a character vector.
#'
#' @param original The data frame which contains the columns to be checked.
#' @param substituted The grouping variables which potentially form unique combinations.
#'
#' @rdname convert_arguments
#'
#' @export
get_origin_as_char <- function(original, substituted){
    tryCatch({
        # Force evaluation to see if it exists
        if (is.character(original)){
            original
        }
        else{
            args_to_char(substituted)
        }
    }, error = function(e){
        # If evaluation failed (object not found), convert the symbol name
        args_to_char(substituted)
    })
}


#' Get The Original Symbol From Parent Environments
#'
#' @description
#' When nesting functions and trying to catch symbols, it
#' often happens that 'R' looses contact to the original symbol. This function looks
#' up the parent environments until it finds the original symbol and returns its
#' contents as character.
#'
#' In case the originally passed argument is an unresolvable symbol (e.g. a variable
#' name), this function returns original argument passed down as characxter instead of
#' the contents.
#'
#' @param symbol The symbol to look up in the parent environments.
#'
#' @return
#' The original symbols contents or name as character.
#'
#' @noRd
get_origin_symbol <- function(symbol){
    symbol <- as.character(symbol)

    # Loop through the thee of parent environments
    for (i in rev(seq_len(sys.nframe()))){
        env <- sys.frame(i)

        # If the symbol is found in the current parent environment return it
        value <- tryCatch({
            # Force evaluation to see if it exists
            get0(substitute(symbol), envir = env, inherits = FALSE)
        }, error = function(e){
            # Nested substitute:
            # 1. Inner: substitute(substitute(sym, env), list(sym = symbol))
            #    This builds the expression "substitute(var_name, env)"
            # 2. Outer: eval(...)
            #    This gets the expression in the correct context
            as.character(eval(substitute(
                substitute(sym, env), list(sym = as.name(symbol)))))
        })

        if (!is.null(value) && is.character(value)){
            return(value)
        }
    }

    symbol
}


#' Convert Variables
#'
#' @name convert_variables
#'
#' @description
#' [convert_numeric()] converts all given variables to numeric if possible. If
#' a variable contains none numerical values (not including NAs), the variable
#' will not be converted.
#'
#' @return
#' [convert_numeric()] returns the same data frame with converted variables where possible.
#'
#' @param data_frame A data frame containing variables to convert.
#' @param variables Variables from the data frame which should be converted.
#'
#' @examples
#' # Convert variables in a data frame to numeric where possible
#' test_df <- data.frame(var_a = c(1, 2, 3, NA, 4, 5),
#'                       var_b = c(1, 2, "Hello", NA, 4, 5))
#'
#' convert_df <- test_df |> convert_numeric(c("var_a", "var_b"))
#'
#' @rdname convert_variables
#'
#' @export
convert_numeric <- function(data_frame, variables){
    data_frame[variables] <-lapply(data_frame[variables], function(variable){
        # Try to convert to numeric
        var_converted <- suppressWarnings(as.numeric(variable))

        # If NA values are equal, conversion was successful
        if (all(is.na(variable) == is.na(var_converted))){
            var_converted
        }
        else{
            variable
        }
    })

    data_frame
}


#' @description
#' [convert_factor()] converts all given variables to factor.
#'
#' @return
#' [convert_factor()] returns the same data frame with converted variables.
#'
#' @examples
#' # Convert variables in a data frame to factor
#' test_df <- data.frame(var_a = c(1, 2, 3, 4, 5),
#'                       var_b = c("e", "c", "a", "d", "b"))
#'
#' convert_df <- test_df |> convert_factor("var_b")
#'
#' @rdname convert_variables
#'
#' @export
convert_factor <- function(data_frame, variables){
    for (variable in variables){
        if (is.character(data_frame[[variable]])){
            # Extract the number of labels from variable
            label_levels <- data_frame[[variable]] |>
                unlist(use.names = FALSE) |>
                collapse::funique() |>
                collapse::na_omit()

            # Convert variable to factor
            data_frame[[variable]] <- factor(
                data_frame[[variable]],
                levels  = label_levels,
                ordered = TRUE)
        }
    }

    data_frame
}


#' Convert Color Codes
#'
#' @name hex_ansi
#'
#' @description
#' [hex_to_256()]: Generate a 256-color 6x6x6 color cube.
#'
#' @param hex_color The hex color code to convert.
#'
#' @return
#' [hex_to_256()]: Returns 6x6x6 color cube.
#'
#' @examples
#' color_cube <- hex_to_ansi("#32CD32")
#'
#' @rdname hex_ansi
#'
#' @export
hex_to_256 <- function(hex_color){
    rgb <- as.vector(grDevices::col2rgb(hex_color))

    # Generate a 256-color 6x6x6 color cube: 16 + 36*r + 6*g + b
    # where r,g,b are 0-5
    values <- floor(rgb / 51)

    16 + (36 * values[1]) + (6 * values[2]) + values[3]
}


#' @description
#' [hex_to_ansi()]: Applies hex color and font weight to a text as ansi codes.
#'
#' @return
#' [hex_to_ansi()]: Returns formatted text.
#'
#' @param text The text which contains hex color codes to be convert into ansi
#' style formatting.
#' @param hex_color The hex color code to convert.
#' @param bold FALSE by default. If TRUE inserts ansi code for bold printing
#'
#' @examples
#' formatted_text <- hex_to_ansi("This is a text to test the coloring",
#'                               hex_color = "#32CD32", bold = TRUE)
#'
#' @rdname hex_ansi
#'
#' @export
hex_to_ansi <- function(text,
                        hex_color = NULL,
                        bold      = FALSE){
    prefix <- ""

    # If bold, text has to be surrounded with bold code
    if (bold){
        prefix <- paste0(prefix, "\033[1m")
    }

    # If color should be applied
    if (!is.null(hex_color)){
        color_256 <- hex_to_256(hex_color)

        # Use the 8-bit extended sequence: 38;5;Nm
        prefix <- paste0(prefix, sprintf("\033[38;5;%dm", color_256))
    }

    # Return fully styled text. Add old end as suffix
    paste0(prefix, text, "\033[0m")
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

dummy_df <- dummy_data(1000)
dummy_df[["age"]] <- as.character(dummy_df[["age"]])


# Convert to numeric only if possible
test_df1 <- dummy_df |> convert_numeric("age")
test_df2 <- dummy_df |> convert_numeric("education")

expect_true(is.numeric(test_df1[["age"]]), info = "Convert to numeric only if possible")
expect_true(is.character(test_df2[["age"]]), info = "Convert to numeric only if possible")
expect_true(is.character(test_df2[["education"]]), info = "Convert to numeric only if possible")


# Convert arguments to character vector
return_param <- function(parameter){
    args_to_char(substitute(parameter))
}

expect_equal(return_param(age), "age", info = "Convert arguments to character vector")
expect_equal(return_param("age"), "age", info = "Convert arguments to character vector")
expect_equal(return_param(c(age, sex, income, weight)), c("age", "sex", "income", "weight"), info = "Convert arguments to character vector")
expect_equal(return_param(c("age", "sex", "income", "weight")), c("age", "sex", "income", "weight"), info = "Convert arguments to character vector")


# Convert arguments to character vector in nested function
nested_function <- function(parameter){
    get_origin_as_char(parameter, substitute(parameter))
}

return_param <- function(parameter){
    nested_function(parameter)
}

expect_equal(return_param("age"), "age", info = "Convert arguments to character vector in nested function")
expect_equal(return_param(c("age", "sex", "income", "weight")), c("age", "sex", "income", "weight"), info = "Convert arguments to character vector in nested function")


# Convert ellipsis to character vector
return_param <- function(...){
    dots_to_char(...)
}

expect_equal(return_param(age), "age", info = "Convert ellipsis to character vector")
expect_equal(return_param("age"), "age", info = "Convert ellipsis to character vector")
expect_equal(return_param(c(age, sex, income, weight)), c("age", "sex", "income", "weight"), info = "Convert ellipsis to character vector")
expect_equal(return_param(c("age", "sex", "income", "weight")), c("age", "sex", "income", "weight"), info = "Convert ellipsis to character vector")


set_no_print()

#' Resolve Macro Variables In A Text
#'
#' @name macro
#'
#' @description
#' Variables written with a leading "&" inside a text will be resolved into the
#' character or numeric expression the corresponding object is holding.
#'
#' @param text A text containing macro variables to resolve.
#'
#' @return
#' [macro()]: Returns a character.
#'
#' @details
#' Macro variables in 'SAS' can be set up with %Let and can act as global accessible
#' variables. This in itself is nothing special to 'R' because basically every object
#' created outside a function is a global variable.
#'
#' To use these global objects within a text one has to use e.g. paste0("The current year is: ", year).
#' With the macro function one can write it like this: macro("The current year is &year").
#'
#' So where is the benefit? If implemented within a function, a parameter like "title" or
#' "footnote" in the tabulation functions, can resolve these variables without the need
#' of another function. You can just pass the character expression "The current year is &year"
#' and the implementation inside the function resolves the macro variable directly in place.
#'
#' @seealso
#' Functions that can make use of macro variables: [any_table()], [frequencies()],
#' [crosstabs()], [export_with_style()], [summarise_plus()]
#'
#' Within the tabulation functions titles, footnotes, var_labels and stat_labels
#' can resolve macros. Additionally they can be used in the [any_table()] rows and
#' columns parameter and in the [summarise_plus()] types parameter.
#'
#' @examples
#' # Resolving macro variable in single character
#' year <- 2026
#'
#' text <- macro("The current year is &year")
#'
#' # You can also combine multiple macro variables
#' some_variable <- "current"
#' current2026   <- "The current year is"
#'
#' text_combi <- macro("&&some_variable&year &year")
#'
#' @rdname macro
#'
#' @export
macro <- function(text){
    if (!is.character(text)){
        print_message("ERROR", "<Text> must be a character. Macro will be aborted.")
        return(invisible(text))
    }

    if (length(text) > 1){
        print_message("WARNING", "<Text> may only be of length one. The first Element will be used.")

        text <- text[[1]]
    }

    # Macro variables start with "&" and consist of alphanumeric characters. There can be a "."
    # at the end of a macro variable to symbolize the end of the variable name.
    regex_pattern <- "&[a-zA-Z0-9_]+(\\.|\\b)"

    # Limit the number of iterations to avoid the case of getting stuck in an endless loop
    for (i in 1:5){
        # Get the starting positions of the macro variables by the pattern above
        macro_vars_position <- gregexpr(regex_pattern, text, perl = TRUE)[[1]]

        # If no matching pattern was found end the loop
        if (length(macro_vars_position) == 1){
            if (macro_vars_position == -1){
                break
            }
        }

        # Get the full macro variable names
        macro_var_names <- regmatches(text, list(macro_vars_position))[[1]]
        var_names       <- gsub("^&|\\.$", "", macro_var_names)

        # Create a map of unique matches of macro variables so that variables that
        # come up multiple times in the same text are only processed once.
        macro_mapping <- unique(data.frame(macro_var = macro_var_names,
                                           var_name  = var_names,
                                           stringsAsFactors = FALSE))

        for (var_number in seq_len(collapse::fnrow(macro_mapping))){
            macro_var <- macro_mapping[var_number, "macro_var"]
            var_name  <- macro_mapping[var_number, "var_name"]

            # Try to get the original value of the macro variable up the environmental ladder
            original_symbol <- tryCatch({
                # Force evaluation to see if it exists
                #get(var_name, envir = parent.frame(), inherits = TRUE)
                dynGet(var_name, ifnotfound = NULL, inherits = TRUE)
            }, error = function(e){
                # Evaluation failed
                NULL
            })

            # If the object was found
            if (!is.null(original_symbol)){
                # Individual error handling
                if (length(original_symbol) > 1){
                    print_message("WARNING", "Macro variable '[macro_var]' may only be of length one. The first Element will be used.", macro_var = macro_var)

                    original_symbol <- original_symbol[[1]]
                }

                if (!is.character(original_symbol)){
                    # Silently convert numerics
                    if (is.numeric(original_symbol)){
                        original_symbol <- as.character(original_symbol)
                    }
                    # All other types will be skipped
                    else{
                        print_message("WARNING", c("Macro variable '[macro_var]' is a complex object. Macro variables may only be character or numeric.",
												   "Resolving will be skipped."), macro_var = macro_var)

                        text <- gsub(macro_var, var_name, text, fixed = TRUE)
                        next
                    }
                }

                # Replace the macro variable with the original content of the found object
                text <- gsub(macro_var, original_symbol, text, fixed = TRUE)
            }
            # If the object wasn't found
            else{
                print_message("WARNING", "Macro variable '[macro_var]' could not be resolved. Object doesn't exist.", macro_var = macro_var)

                text <- gsub(macro_var, var_name, text, fixed = TRUE)
            }
        }
    }

    text
}

#' @name macro
#'
#' @param character_vector A vector containing multiple character expressions in which
#' to resolve macro variables.
#'
#' @return
#' [apply_macro()]: Returns a character vector.
#'
#' @examples
#' # Resolving macro variable in character vector
#' char_vector <- c("The current year is &year",
#'                  "The &some_variable year is &year",
#'                  "&&some_variable&year &year")
#'
#' text_vector <- apply_macro(char_vector)
#'
#' @rdname macro
#'
#' @export
apply_macro <- function(character_vector){
    sapply(character_vector, macro, USE.NAMES = FALSE)
}

################################################################################
# Export functions
################################################################################
#' Perform Row Wise Calculations
#'
#' @description
#' Perform row wise calculations on numeric variables.
#'
#' @param data_frame A data frame in which are the values to be calculated.
#' @param statistics Available functions: "sum", "freq", "mean", "median", "mode",
#' "min", "max".
#' @param ... Variable names of the value variables.
#' @param round_digits The number of decimal places the values should be rounded to.
#'
#' @return
#' Returns a numeric vector.
#'
#' @examples
#' # Example data frame
#' my_data <- data.frame(var1 = 1:5,
#'                       var2 = c(6, 7, 8, NA, 10),
#'                       var3 = 11:15)
#'
#' # Calculate new variables
#' my_data[["sum"]]  <- my_data |> row_calculation("sum",  var1, var2, var3)
#' my_data[["mean"]] <- my_data |> row_calculation("mean", var1:var3)
#' @export
row_calculation <- function(data_frame,
                            statistics,
                            ...,
                            round_digits = NULL){
    # Just keep the selected variables and transpose the entire matrix. Since
    # the rows then become the columns, it is possible to just use column wise
    # operations on the matrix.
    var_matrix <- data_frame |>
        keep(...) |>
        t()

    if (!is.numeric(var_matrix)){
        print_message("ERROR", "Only numeric values allowed. Calculation will be aborted.")

        return(invisible(NA))
    }

    # Check if provided statistics is valid
    statistics <- get_origin_as_char(statistics, substitute(statistics))

    if (length(statistics) > 1){
        print_message("WARNING", "Only one <statistics> allowed at a time. The first element will be used.")

        statistics <- statistics[1]
    }

    if (!tolower(statistics) %in% c("sum", "mean", "median", "min", "max", "mode", "freq")){
        print_message("WARNING", "<Statistics> '[statistics]' [?is/are] invalid. 'sum' will be used.", statistics = tolower(statistics))

        statistics <- "sum"
    }

    # Get the desired operation
    stat_function <- switch(statistics,
                            sum    = collapse::fsum,
                            mean   = collapse::fmean,
                            median = collapse::fmedian,
                            min    = collapse::fmin,
                            max    = collapse::fmax,
                            mode   = collapse::fmode,
                            freq   = collapse::fnobs,
                            NULL)

    # Compute row result and return as vector
    result <- stat_function(var_matrix)

    if (!is.null(round_digits)){
        result <- round_values(result, round_digits)
    }

    result
}


################################################################################
# Internal functions used in summarise_plus
################################################################################
#' Unweighted Frequencies of Values Greater Than Zero
#'
#' @description
#' This stat always computes unweighted sums, completely ignoring provided weights.
#' It also just counts the values which are greater than 0.
#'
#' @param values The values to sum up.
#' @param group Grouping variables.
#'
#' @return
#' Returns uweighted frequencies of values greater than zero.
#'
#' @noRd
freq_g0_qol <- function(values, group){
    # Create a vector of same size as input vector and make every value a 1
    # that is greater than 0.
    # Sum up this new vector to get unweighted frequencies.
    values[values > 0] <- 1

    collapse::fsum(values, g = group)
}


#' Calculate Any Grouped And Weighted Percentiles
#'
#' @description
#' Calculates any percentile by grouping variables. Uses the SAS default way of
#' calculating percentiles, which is Type 2 described in [quantile()].
#'
#' @param values The values for which to compute percentiles.
#' @param weight A weighting variable.
#' @param group Grouping variables.
#' @param probs The percentiles that should be computed.
#'
#' @return
#' Returns da data.table with weighted percentiles.
#'
#' @noRd
percentiles_qol <- function(values, weight, group, probs){
    if (anyNA(values)){
        print_message("WARNING", "To calculate percentiles there may be no NA in the value variables.",
                      always_print = TRUE)
        return(NULL)
    }

    if (anyNA(weight)){
        print_message("WARNING", "To calculate percentiles there may be no NA in the weight variable.",
                      always_print = TRUE)
        return(NULL)
    }

    # If weight only consists of one unique value, then the results are unweighted.
    # In this case use the collapse method with Type 7.
    if (length(collapse::funique(weight)) == 1){
        collapse::BY(values, g = group, collapse::fquantile, w = weight, probs = probs)
    }
    # Otherwise the custom function for Type 2, which is not in collapse package.
    else{
        collapse::BY(values, g = group, calculate_single_percentiles, w = weight, probs = probs)
    }
}


#' @return
#' Returns a weighted percentile vector.
#'
#' @noRd
calculate_single_percentiles <- function(x, probs, w){
    # Sort values in ascending order and apply the same ordering to the weights.
    sort_order <- order(x)

    x <- x[sort_order]
    w <- w[sort_order]

    # Calculate total weight.
    sum_of_weights <- collapse::fsum(w)

    # If all weights sum to zero, weighted percentiles are undefined.
    if (sum_of_weights == 0){
        return(rep(NA_real_, length(probs)))
    }

    # Compute the weighted cumulative distribution function (CDF). Values range
    # from 0 to 1.
    weight_cdf <- collapse::fcumsum(w) / sum_of_weights

    # Find the largest index whose cumulative weight is still
    # less than or equal to the requested percentile.
    lower <- findInterval(probs, weight_cdf)

    # If the percentile lies below the first weighted observation, return the
    # minimum value.
    if (lower == 0){
        return(x[1])
    }

    # If the percentile lies beyond the last weighted observation,
    # return the maximum value.
    upper <- lower + 1

    if (upper > length(x)){
        return(x[lower])
    }

    # If the lower observation is still below the requested percentile,
    # use the next observation.
    if (weight_cdf[lower] < probs){
        x[upper]
    }
    # Otherwise return a weighted average of the two surrounding
    # observations.
    else{
        (w[lower] * x[lower] + w[upper] * x[upper]) / (w[lower] + w[upper])
    }
}


#' Get Missings of Combined Grouping Variables
#'
#' @description
#' Calculate the number of missings, which is generated by the combination of all
#' grouping variables.
#'
#' @param group_vars Grouping variables.
#' @param notes Flag if missings should be computed and message should be printed.
#' @param na.rm Flag if NA values are removed or not.
#'
#' @return
#' Returns a message stating the number of missing values based on grouping variables.
#'
#' @noRd
get_group_missings <- function(group_vars, notes, na.rm){
    if (na.rm){
        return(FALSE)
    }

    if (!notes){
        return(FALSE)
    }

    # Count combined missings of group variables. To count as missings only
    # one of the variables has to be NA in a position.
    missings <- collapse::fsum(collapse::missing_cases(group_vars))

    # If there are missings calculate percentage and output a message
    if (missings > 0){
        nobs      <- length(group_vars[[1]])
        percent   <- round(missings * 100 / nobs, 1)
        none_miss <- nobs - missings

        print_message("NOTE", "[missings] missings generated from grouping variables ([percent] %). Number of observations: [none_miss]/[nobs]",
					  missings  = format(missings, big.mark = ".", decimal.mark = ","),
				      percent   = percent,
				      none_miss = format(none_miss, big.mark = ".", decimal.mark = ","),
				      nobs      = format(nobs, big.mark = ".", decimal.mark = ","),
					  always_print = TRUE)
    }
}


#' Compute Percentages by Group
#'
#' @description
#' Calculate the percentages inside a specific group of variables by generating the
#' super group first and joining the super group totals back to the data frame.
#'
#' @param original_df The original unchanged data frame provided by the user.
#' @param summary_df Already summed up result data frame to build up upon.
#' @param statistics User provided statistics.
#' @param group_vars Grouping variables.
#' @param formats Provided list of variables with formats to apply.
#' @param values he values for which to compute percentages.
#' @param weight Weighting variable.
#' @param list_of_statistics A list of all statistics that can be computed.
#' @param monitor_df Data frame which stores the monitoring values.
#' @param fast_pct Flag that states whether the function can take a shortcut for faster
#' calculation.
#'
#' @return
#' Returns a data frame with added percentages, if the statistic was specified.
#'
#' @noRd
compute_group_percentages <- function(original_df,
                                      summary_df,
                                      statistics,
                                      group_vars,
                                      formats,
                                      values,
                                      weight,
                                      list_of_statistics,
                                      monitor_df,
                                      fast_pct){
    if (!"pct_group" %in% statistics){
        return(list(summary_df, monitor_df))
    }

    monitor_df <- monitor_df |> monitor_start("pct_group", paste0("Calc(", paste(group_vars, collapse = " + "), ")"))

    # Create pseudo group variable for totals
    original_df[[".temp_key"]] <- 1
    summary_df[[".temp_key"]]  <- 1

    # Get rid of the last grouping variable to get the grouping above the
    # desired grouping
    if (length(group_vars) > 1){
        super_group <- group_vars[-length(group_vars)]
    }
    else{
        super_group <- ".temp_key"
    }

    cut_off_var <- group_vars[length(group_vars)]

    # If no fast percentage (means original_df is the full data frame)
    if (!fast_pct){
        # It is faster to summarise first without formats
        # and then apply the formats to a much smaller data frame.
        super_df <- original_df |>
            matrix_summarise(values,
                             super_group,
                             original_df[[weight]],
                             c("sum", "sum_wgt"),
                             list_of_statistics,
                             monitor_df)

        super_df <- super_df[[1]]

        super_df <- super_df |>
            apply_format(formats, super_group) |>
            collapse::fselect(-sum_wgt)

        # Catch new variable names
        values <- super_df |> inverse(super_group)
    }
    # If fast percentage (means original_df is already summarised)
    else{
        values <- values[grepl("_sum$", values)]

        # Summarise first
        super_df <- original_df |>
            collapse::fgroup_by(super_group) |>
            collapse::fsummarise(across(values, collapse::fsum))

        super_df <- super_df |>
            apply_format(formats, super_group)
    }

    # Generate new variable names for super_group totals
    new_names <- paste0(values, "_qol")
    old_names <- values

    # Final summarise with formatted data frame
    super_df <- super_df |>
        collapse::fgroup_by(super_group) |>
        collapse::fsummarise(across(values, collapse::fsum)) |>
        collapse::frename(stats::setNames(old_names, new_names)) |>
        collapse::fselect(super_group, new_names)

    # Join super_df to summarized data frame
    if (is.null(super_group)){
        # In case of only totals
        joined_df <- collapse::join(summary_df, super_df,
                                    on       = ".temp_key",
                                    how      = "left",
                                    multiple = TRUE,
                                    verbose  = FALSE,
								    overid   = 2)
    }
    else{
        # In case there is at least one grouping variable
        joined_df <- collapse::join(summary_df, super_df,
                                    on      = super_group,
                                    how     = "left",
                                    verbose = FALSE,
								    overid  = 2)
    }

    # Return finished data frame
    joined_df <- joined_df |>
        calculate_percentages(values, "pct_group", cut_off_var) |>
        collapse::fselect(-.temp_key)

    monitor_df <- monitor_df |> monitor_end()

    list(joined_df, monitor_df)
}


#' Compute Percentages by Group
#'
#' @description
#' Calculate the percentages inside a specific group of variables, which builds upon
#' a data frame which already carries the summarised super groups.
#'
#' @param summary_df Already summed up result data frame to build up upon.
#' @param super_df Already summed up super group results.
#' @param statistics User provided statistics.
#' @param group_vars Grouping variables.
#' @param values he values for which to compute percentages.
#' @param monitor_df Data frame which stores the monitoring values.
#' @param fast_pct Flag that states whether the function can take a shortcut for faster
#' calculation.
#'
#' @return
#' Returns a data frame with added percentages, if the statistic was specified.
#'
#' @noRd
compute_group_percentages_short <- function(summary_df,
                                            super_df,
                                            statistics,
                                            group_vars,
                                            values,
                                            monitor_df,
                                            fast_pct){
    if (!"pct_group" %in% statistics){
        return(list(summary_df, monitor_df))
    }

    monitor_df <- monitor_df |> monitor_start("pct_group", paste0("Calc(", paste(group_vars, collapse = " + "), ")"))

    # Get rid of the last grouping variable to get the grouping above the
    # desired grouping
    super_group <- group_vars[-length(group_vars)]

    # Generate new variable names for super_group totals
    old_names <- values
    new_names <- paste0(values, "_qol")

    super_df <- super_df |>
        collapse::fselect(super_group, old_names) |>
        collapse::frename(stats::setNames(old_names, new_names))

    # Join data frames by super group
    joined_df <- collapse::join(summary_df, super_df,
                                on      = super_group,
                                how     = "left",
                                verbose = FALSE,
								overid  = 2)

    # Return finished data frame
    joined_df <- joined_df |>
        calculate_percentages(values, "pct_group", group_vars[length(group_vars)])

    monitor_df <- monitor_df |> monitor_end()

    list(joined_df, monitor_df)
}


#' Compute Total Percentages
#'
#' @description
#' Calculate the percentages based on the overall totals.
#'
#' @param original_df The original unchanged data frame provided by the user.
#' @param summary_df Already summed up result data frame to build up upon.
#' @param statistics User provided statistics.
#' @param group_vars Grouping variables.
#' @param values he values for which to compute percentages.
#' @param weight Weighting variable.
#' @param list_of_statistics A list of all statistics that can be computed.
#' @param monitor_df Data frame which stores the monitoring values.
#' @param fast_pct Flag that states whether the function can take a shortcut for faster
#' calculation.
#'
#' @return
#' Returns a data frame with added percentages, if the statistic was specified.
#'
#' @noRd
compute_total_percentages <- function(original_df,
                                      summary_df,
                                      statistics,
                                      group_vars,
                                      values,
                                      weight,
                                      list_of_statistics,
                                      monitor_df,
                                      fast_pct){
    if (!"pct_total" %in% statistics){
        return(list(summary_df, monitor_df))
    }

    monitor_df <- monitor_df |> monitor_start("pct_total", paste0("Calc(", paste(group_vars, collapse = " + "), ")"))

    # Generate new variable names for super_group totals
    new_names <- paste0(values, "_qol")

    # Compute the data frame with summaries of the super groups
    super_df <- original_df |> collapse::fungroup()

    # If no fast percentage (means original_df is the full data frame)
    if (!fast_pct){
        super_df <- super_df |>
            matrix_summarise(values,
                             NULL,
                             super_df[[weight]],
                             c("sum", "sum_wgt"),
                             list_of_statistics,
                             monitor_df)

        # Split results and monitor
        super_df <- super_df[[1]]

        # Catch new variable names
        values <- super_df |>
            collapse::fselect(-sum_wgt) |>
            inverse(group_vars)
    }
    # If fast percentage (means original_df is already summarised)
    else{
        values <- values[grepl("_sum$", values)]

        # Simple Summarise
        super_df <- original_df |>
            collapse::fsummarise(across(values, collapse::fsum))
    }

    # Generate new variable names for super_group totals
    new_names <- paste0(values, "_qol")
    old_names <- values

    super_df <- super_df |>
        collapse::frename(stats::setNames(old_names, new_names)) |>
        collapse::fselect(new_names)

    # Join super_df to summarized data frame
    summary_df[[".temp_key"]] <- 1
    super_df[[".temp_key"]]   <- 1

    joined_df <- collapse::join(summary_df, super_df,
                                on       = ".temp_key",
                                how      = "left",
                                multiple = TRUE,
                                verbose  = FALSE,
								overid   = 2)

    # Return finished data frame
    joined_df <- joined_df |>
        calculate_percentages(values, "pct_total", ".temp_key") |>
        collapse::fselect(-.temp_key)

    monitor_df <- monitor_df |> monitor_end()

    list(joined_df, monitor_df)

}


#' Compute Total Percentages
#'
#' @description
#' Calculate the percentages based on the overall totals, which builds upon
#' a data frame which already carries the summarised overall totals.
#'
#' @param summary_df Already summed up result data frame to build up upon.
#' @param total_df Already summed up result data frame to build up upon.
#' @param statistics User provided statistics.
#' @param values he values for which to compute percentages.
#' @param last_group_var The last variable of grouping variables.
#' @param pct_name Name of percentages.
#' @param monitor_df Data frame which stores the monitoring values.
#'
#' @return
#' Returns a data frame with added percentages, if the statistic was specified.
#'
#' @noRd
compute_total_percentages_short <- function(summary_df,
                                            total_df,
                                            statistics,
                                            values,
                                            last_group_var,
                                            pct_name = "pct_total",
                                            monitor_df){
    if (!pct_name %in% statistics){
        return(list(summary_df, monitor_df))
    }

    monitor_df <- monitor_df |> monitor_start(pct_name, paste0("Calc(", paste(last_group_var, collapse = " + "), ")"))

    # Join super_df to summarized data frame
    summary_df[[".temp_key"]] <- 1
    total_df[[".temp_key"]]   <- 1

    joined_df <- collapse::join(summary_df, total_df,
                                on       = ".temp_key",
                                how      = "left",
                                multiple = TRUE,
                                verbose  = FALSE,
								overid   = 2)

    # Return finished data frame
    if (pct_name == "pct_total"){
        joined_df <- joined_df |>
            calculate_percentages(values, pct_name, ".temp_key") |>
            collapse::fselect(-.temp_key)
    }
    else{
        # For group percentages with depth 1
        joined_df <- joined_df |>
            calculate_percentages(values, pct_name, last_group_var) |>
            collapse::fselect(-.temp_key)
    }

    monitor_df <- monitor_df |> monitor_end()

    list(joined_df, monitor_df)

}


#' Compute Percentages
#'
#' @description
#' The main process of calculating percentages.
#'
#' @param joined_df Data frame with joined super group or overall totals.
#' @param values he values for which to compute percentages.
#' @param last_group_var The last variable of grouping variables.
#' @param pct_name Name of percentages.
#'
#' @return
#' Returns a data frame with added percentages.
#'
#' @noRd
calculate_percentages <- function(joined_df, values, pct_name, last_group_var){
    # Specify numerators and denominators
    numerator <- values

    if (!all(values %in% names(joined_df))){
        # In case variable names have the statistic extension at the end
        numerator <- paste0(values, "_sum")
    }

    denominator <- paste0(values, "_qol")
    new_values  <- paste0(gsub("_sum$", "", values), "_", pct_name)

    # Compute percentages for every variable
    for (i in seq_along(numerator)){
        current_num     <- numerator[i]
        current_den     <- denominator[i]
        current_new_var <- new_values[i]

        # Evaluate percentage and set division by 0 to NA
        result <- joined_df[[current_num]] * 100 / joined_df[[current_den]]
        result[is.nan(result)] <- NA

        joined_df[[current_new_var]] <- result
    }

    joined_df |> dropp(denominator)
}

#' Concatenate Multiple Variables With Padding
#'
#' @description
#' Concatenate multiple variables inside a data frame into a new variable. An
#' automatic or individual padding can be applied. The padding character can be
#' chosen freely.
#'
#' The function can also be used to give a single variable a padding.
#'
#' @param data_frame A data frame which contains the the variables to concatenate.
#' @param ... The names of the variables to concatenate.
#' @param separator Characters to put in between variables when binding them together.
#' @param padding_char A single character which will be used to fill up the empty places.
#' @param padding_length A numeric vector containing the individual padding length per variable.
#' @param padding_right FALSE by default. If TRUE insert padding characters on the right side
#' instead of the left side.
#'
#' @return
#' Returns a character vector.
#'
#' @seealso
#' Other character manipulating functions: [sub_string()], [remove_blanks()]
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(100)
#'
#' # Concatenate variables as provided
#' my_data[["id1"]] <- my_data |> concat(household_id, state, age)
#'
#' # Concatenate variables with leading zeros. Each variable will
#' # receive an individual padding length according to their
#' # longest value.
#' my_data[["id2"]] <- my_data |> concat(household_id, state, age,
#'                                       padding_char = "0")
#'
#' # Concatenate variables with individual character and lengths.
#' my_data[["id2"]] <- my_data |> concat(household_id, state, age,
#'                                       padding_char   = "_",
#'                                       padding_length = c(5, 3, 4))
#'
#' # Padding a single variable in place
#' my_data[["state"]] <- my_data |> concat(state, padding_char = "0")
#'
#' @export
concat <- function(data_frame,
                   ...,
                   separator      = NULL,
                   padding_char   = NULL,
                   padding_length = NULL,
                   padding_right  = FALSE){
    variables <- dots_to_char(...)

    # If no padding or separator is defined just concatenate provided variables as they are
    if (is.null(padding_char) && is.null(padding_length) && is.null(separator)){
        return(do.call(paste0, data_frame[variables]))
    }

    # Padding character may only be of length one
    if (!is.null(padding_char) && collapse::vlengths(padding_char) != 1){
        print_message("WARNING", "<Padding chararacter> must be a single character. Concat will be done without <padding character>.")
        return(do.call(paste0, data_frame[variables]))
    }

    # If a separator is provided, concatenate variables with separator in between
    if (!is.null(separator)){
        separator <- as.character(separator)

        return(do.call(paste, c(data_frame[variables], sep = separator)))
    }

    # If no padding character is given use a blank
    if (is.null(padding_char)){
        padding_char <- " "
    }

    # Determine padding lengths. If NULL all given variables will be checked individually
    # for their maximum length.
    if (is.null(padding_length)){
        padding_length <- vapply(data_frame[variables],
                                 function(variable){
                                     collapse::fmax(nchar(variable), na.rm = TRUE)
                                 },
                                 integer(1))
    }
    # With given padding length it has to be checked whether the provided vector
    # is to long or to short in comparison to the number of provided variables.
    # Either way it has to be matched to the number of provided variables. Only if
    # it is of equal length nothing happens here.
    else{
        number_of_columns  <- length(variables)
        number_of_paddings <- length(padding_length)

        # If the padding vector is shorter than the number of provided variables,
        # the remaining variables will receive the individual maximum length as
        # padding length like above.
        if (number_of_paddings < number_of_columns){
            print_message("WARNING", c("<Padding length> is shorter than the number of variables.",
                                       "Missing lengths will be filled up using maximum individual variable length."))

            missing_length <- vapply(data_frame[variables[(number_of_paddings + 1):number_of_columns]],
                                     function(variable){
                                         collapse::fmax(nchar(variable), na.rm = TRUE)
                                     },
                                     integer(1))

            padding_length <- c(padding_length, missing_length)
        }
        # If the padding vector is longer than the number of provided variables,
        # juts trim it down.
        else if (number_of_paddings > number_of_columns){
            print_message("WARNING", c("<Padding length> is longer than the number of variables.",
                                       "Extra lengths will be ignored."))

            padding_length <- padding_length[seq_len(number_of_columns)]
        }
    }

    # Apply padding to all variables individually
    padded_variables <- Map(function(variable, padding_length){

        # NA values are just filled up with padding character
        variable[is.na(variable)] <- strrep(padding_char, padding_length)

        # Get the number of places to fill up with padding character per observation
        pad_per_observation <- pmax(padding_length - collapse::vlengths(variable), 0)

        # Concatenate padding and variable values
        if (!padding_right){
            paste0(strrep(padding_char, pad_per_observation), variable)
        }
        else{
            paste0(variable, strrep(padding_char, pad_per_observation))
        }
    },
    data_frame[variables],
    padding_length
    )

    # Concatenate all padded variables together to one single variable
    do.call(paste0, padded_variables)
}


#' Retrieve A Substring From A Character
#'
#' @description
#' [sub_string()] can extract parts of a character from the left side, right side
#' or from the middle. It is also able to start or end at specific letter sequences
#' instead of positions.
#'
#' @param data_frame A data frame which contains the variables to concatenate.
#' @param variable A character variable to extract parts from.
#' @param from The names of the variables to concatenate.
#' @param to A single character which will be used to fill up the empty places.
#' @param case_sensitive TRUE by default. When a character expression is passed as
#' from or to it makes a difference whether a letter is written in upper or lower case.
#' Pass FALSE to handle upper and lower case equally.
#'
#' @return
#' Returns parts of a character vector.
#'
#' @seealso
#' Other character manipulating functions: [concat()], [remove_blanks()]
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(100)
#'
#' # Extract text from the left
#' my_data[["left"]] <- my_data |> sub_string(education, to = 2)
#'
#' # Extract text from the right
#' my_data[["right"]] <- my_data |> sub_string(education, from = 2)
#'
#' # Extract text from the middle
#' my_data[["middle"]] <- my_data |> sub_string(education, from = 2, to = 3)
#'
#' # Find text and extract from the left
#' my_data[["left2"]] <- my_data |> sub_string(education, to = "l")
#'
#' # Find text and extract from the right
#' my_data[["right2"]] <- my_data |> sub_string(education, from = "l")
#'
#' # Find text and extract from the middle
#' my_data[["middle2"]] <- my_data |> sub_string(education, from = "i", to = "l")
#'
#' @export
sub_string <- function(data_frame,
                       variable,
                       from  = NULL,
                       to    = NULL,
                       case_sensitive = TRUE){
    variable <- get_origin_as_char(variable, substitute(variable))

    # If no value variables are provided abort
    if (length(variable) <= 1){
        if (length(variable) == 0 || variable == ""){
            print_message("ERROR", "No <variables> provided. Blank removal will be aborted.")
            return(invisible(NULL))
        }
    }

    # Adjust variable
    if (length(variable) > 1){
        print_message("WARNING", "<Variable> may only be of length one. The first Element will be used.")

        variable <- variable[[1]]
    }

    # Make sure that the variable provided is part of the data frame
    variable <- data_frame |> part_of_df(variable, check_only = TRUE)

    if (is.list(variable)){
        print_message("ERROR", c("The provided <variable> '[var]' is not part of",
								 "the data frame. Substring will be aborted."), var = variable[[1]])
        return(invisible(NULL))
    }

    if (!is.character(data_frame[[variable]])){
        print_message("ERROR", "<Variable> type must be character. Substring will be aborted.")
        return(invisible(NULL))
    }

    if (is.null(from) && is.null(to)){
        print_message("ERROR", "Neither <from> nor <to> is provided. Substring will be aborted.")
        return(invisible(NULL))
    }

    # Adjust from and length
    if (length(to) > 1){
        print_message("WARNING", "<To> may only be of length one. The first Element will be used.")

        to <- to[[1]]
    }

    if (length(from) > 1){
        print_message("WARNING", "<From> may only be of length one. The first Element will be used.")

        from <- from[[1]]
    }

    variable_vector <- data_frame[[variable]]

    # If to is a character, extract the text up until the first match of to
    if (is.character(to)){
        # If no match is found, to is adjusted to retrieve the whole text
        if (case_sensitive){
            to <- regexpr(to, variable_vector)
        }
        else{
            to <- regexpr(tolower(to), tolower(variable_vector))
        }
        to <- data.table::fifelse(to == -1, 9999, to)
    }

    # If from is a character, extract the text up until the first match of from
    if (is.character(from)){
        # If no match is found, from is adjusted to retrieve the whole text
        if (case_sensitive){
            from <- regexpr(from, variable_vector)
        }
        else{
            from <- regexpr(tolower(from), tolower(variable_vector))
        }
        from <- data.table::fifelse(from == -1, 1, from)
    }

    # In case no from position is given, substring starts at position 1 up to to-position
    if (is.null(from)){
        sub_variable <- substring(variable_vector, 1, to)
    }
    # In case no to position is given, substring starts at from and goes to the end
    else if (is.null(to)){
        sub_variable <- substring(variable_vector, from, nchar(variable_vector))
    }
    # In case both positions are given, extract text in between these two points
    else{
        sub_variable <- substring(variable_vector, from, to)
    }

    sub_variable
}


#' Remove Blanks
#'
#' @description
#' Removes leading and trailing blanks or both. Can also remove all blanks from a
#' character or normalize multiple blanks to single ones.
#'
#' @param data_frame A data frame which contains the character variables from which blanks
#' should be removed.
#' @param variable Variable name of the one from which to remove blanks.
#' @param which "all" by default. Can be "leading", "trailing", "trim", "normalize" or "all".
#' Determines which blanks should be removed
#'
#' @return
#' Returns a character vector with removed blanks.
#'
#' @seealso
#' Other character manipulating functions: [concat()], [sub_string()]
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(100)
#' my_data[["blanks"]] <- " This  is  a  test "
#'
#' # Remove blanks
#' my_data[["leading"]]   <- my_data |> remove_blanks(blanks, which = "leading")
#' my_data[["trailing"]]  <- my_data |> remove_blanks(blanks, which = "trailing")
#' my_data[["trim"]]      <- my_data |> remove_blanks(blanks, which = "trim")
#' my_data[["all"]]       <- my_data |> remove_blanks(blanks, which = "all")
#' my_data[["normalize"]] <- my_data |> remove_blanks(blanks, which = "normalize")
#'
#' @export
remove_blanks <- function(data_frame,
                          variable,
                          which = "all"){

    # Convert to character vectors
    variable <- get_origin_as_char(variable, substitute(variable))

    # If no value variable are provided abort
    if (length(variable) <= 1){
        if (length(variable) == 0 || variable == ""){
            print_message("ERROR", "No <variable> provided. Blank removal will be aborted.")
            return(invisible(NULL))
        }
    }

    # Adjust variable
    if (length(variable) > 1){
        print_message("WARNING", "<Variable> may only be of length one. The first Element will be used.")

        variable <- variable[[1]]
    }

    # Make sure that the provided variable is part of the data frame.
    variable <- data_frame |> part_of_df(variable)

    if (length(variable) == 0){
        print_message("ERROR", "No valid <variable> provided. Blank removal will be aborted.")
        return(invisible(NULL))
    }

    # Abort if non-character variable are selected
    variable_vector <- unlist(data_frame[variable])

    if (!is.character(variable_vector[[1]])){
        print_message("ERROR", "Blank removal only works with a character <variable>. Blank removal will be aborted.")
        return(invisible(NULL))
    }

    # Check if which is a valid option
    which <- tolower(which)

    if (!which %in% c("trim", "leading", "trailing", "all", "normalize")){
        print_message("WARNING", c("Invalid option for <which> provided. Allowed are 'trim', 'leading', 'trailing', 'all' and",
								   "'normalize'. 'all' will be used."))

        which <- "all"
    }

    # Remove blanks according to selected option
    variable_vector <- switch(which,
                              trim      = trimws(variable_vector, which = "both"),
                              leading   = trimws(variable_vector, which = "left"),
                              trailing  = trimws(variable_vector, which = "right"),
                              all       = gsub(" +", "", variable_vector),
                              normalize = {
                                  variable_vector <- trimws(variable_vector)
                                  gsub(" +", " ", variable_vector)
                              })

    variable_vector
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

test_df <- data.frame(
    var1 = c(1, 10, 100),
    var2 = c(1, NA, 3),
    var3 = c("abc", "ab", "a"))

text_df <- data.frame(
    var1 = c("This is a text", "Hello World", "this is a Text"))

text_df2 <- data.frame(
    var1 = c(" This  is  a  text ", " Hello  World ", " this  is  a  Text "),
    var2 = c(1, 2, 3))


# Concatenating multiple variables as provided
test_df[["concat_vec"]] <- test_df |> concat(var1, var2, var3)

expect_equal(test_df[["concat_vec"]], c("11abc", "10NAab", "1003a"), info = "Concatenating multiple variables as provided")


# Concatenating multiple variables with automatic padding
test_df[["concat_vec"]] <- test_df |> concat(var1, var2, var3,
                                padding_char = "0")

expect_equal(test_df[["concat_vec"]], c("0011abc", "01000ab", "100300a"), info = "Concatenating multiple variables with automatic padding")


# Concatenating multiple variables with individual padding length
test_df[["concat_vec"]] <- test_df |> concat(var1, var2, var3,
                                padding_char   = "0",
                                padding_length = c(5, 6, 7))

expect_equal(test_df[["concat_vec"]], c("000010000010000abc", "0001000000000000ab", "00100000003000000a"), info = "Concatenating multiple variables with individual padding length")


# Concatenating multiple variables with individual padding length and no padding character provided
test_df[["concat_vec"]] <- test_df |> concat(var1, var2, var3,
                                padding_length = c(5, 6, 7))

expect_equal(test_df[["concat_vec"]], c("    1     1    abc", "   10           ab", "  100     3      a"), info = "Concatenating multiple variables with individual padding length and no padding character provided")


# Concatenating multiple variables with individual padding length and too small values removes NA values
test_df[["concat_vec"]] <- test_df |> concat(var1, var2, var3,
                                padding_length = c(1, 1, 1))

expect_equal(test_df[["concat_vec"]], c("11abc", "10 ab", "1003a"), info = "Concatenating multiple variables with individual padding length and too small values removes NA values")


# Pad single variable
test_df[["concat_vec"]] <- test_df |> concat(var1, padding_char = "0")

expect_equal(test_df[["concat_vec"]], c("001", "010", "100"), info = "Pad single variable")


# Pad single variable with individual length
test_df[["concat_vec"]] <- test_df |> concat(var1, padding_char   = "_",
                                      padding_length = 5)

expect_equal(test_df[["concat_vec"]], c("____1", "___10", "__100"), info = "Pad single variable with individual length")


# Pad single variable on the right side
test_df[["concat_vec"]] <- test_df |> concat(var1, padding_char = "0", padding_right = TRUE)

expect_equal(test_df[["concat_vec"]], c("100", "100", "100"), info = "Pad single variable on the right side")


# Concatenating multiple variables with automatic padding on the right side
test_df[["concat_vec"]] <- test_df |> concat(var1, var2, var3,
                                padding_char = "0", padding_right = TRUE)

expect_equal(test_df[["concat_vec"]], c("1001abc", "1000ab0", "1003a00"), info = "Concatenating multiple variables with automatic padding on the right side")


# Substring from the left with numeric value
test_df[["sub_vec"]] <- text_df |> sub_string(var1, to = 3)

expect_equal(test_df[["sub_vec"]], c("Thi", "Hel", "thi"), info = "Substring from the left with numeric value")


# Substring from the right with numeric value
test_df[["sub_vec"]] <- text_df |> sub_string(var1, from = 5)

expect_equal(test_df[["sub_vec"]], c(" is a text", "o World", " is a Text"), info = "Substring from the right with numeric value")


# Substring in the middle with numeric values
test_df[["sub_vec"]] <- text_df |> sub_string(var1, from = 5, to = 10)

expect_equal(test_df[["sub_vec"]], c(" is a ", "o Worl", " is a "), info = "Substring in the middle with numeric values")


# Substring from the left with character value
test_df[["sub_vec"]] <- text_df |> sub_string(var1, to = "s")

expect_equal(test_df[["sub_vec"]], c("This", "Hello World", "this"), info = "Substring from the left with character value")


# Substring from the right with character value
test_df[["sub_vec"]] <- text_df |> sub_string(var1, from = "s")

expect_equal(test_df[["sub_vec"]], c("s is a text", "Hello World", "s is a Text"), info = "Substring from the right with character value")


# Substring in the middle with character values
test_df[["sub_vec"]] <- text_df |> sub_string(var1, from = "t", to = "s")

expect_equal(test_df[["sub_vec"]], c("", "Hello World", "this"), info = "Substring in the middle with character values")


# Substring in the middle with character values and no case sensitivity
test_df[["sub_vec"]] <- text_df |> sub_string(var1, from = "t", to = "s", case_sensitive = FALSE)

expect_equal(test_df[["sub_vec"]], c("This", "Hello World", "this"), info = "Substring in the middle with character values and no case sensitivity")


# Substring in the middle with character values and no case sensitivity
test_df[["sub_vec"]] <- text_df |> sub_string(var1, from = "t", to = "s", case_sensitive = FALSE)

expect_equal(test_df[["sub_vec"]], c("This", "Hello World", "this"), info = "Substring in the middle with character values and no case sensitivity")


# Remove leading blanks
text_df2[["gone_blanks"]] <- text_df2 |> remove_blanks(var1, which = "leading")

expect_equal(text_df2[["gone_blanks"]], c("This  is  a  text ", "Hello  World ", "this  is  a  Text "), info = "Remove leading blanks")


# Remove trailing blanks
text_df2[["gone_blanks"]] <- text_df2 |> remove_blanks(var1, which = "trailing")

expect_equal(text_df2[["gone_blanks"]], c(" This  is  a  text", " Hello  World", " this  is  a  Text"), info = "Remove trailing blanks")


# Remove leading and trailing blanks
text_df2[["gone_blanks"]] <- text_df2 |> remove_blanks(var1, which = "trim")

expect_equal(text_df2[["gone_blanks"]], c("This  is  a  text", "Hello  World", "this  is  a  Text"), info = "Remove leading and trailing blanks")


# Remove all blanks
text_df2[["gone_blanks"]] <- text_df2 |> remove_blanks(var1, which = "all")

expect_equal(text_df2[["gone_blanks"]], c("Thisisatext", "HelloWorld", "thisisaText"), info = "Remove all blanks")


# Normalize blanks
text_df2[["gone_blanks"]] <- text_df2 |> remove_blanks(var1, which = "normalize")

expect_equal(text_df2[["gone_blanks"]], c("This is a text", "Hello World", "this is a Text"), info = "Normalize blanks")

###############################################################################
# Warning checks
###############################################################################

# Concatenating only possible with single length padding_char
test_df[["concat_vec"]] <- test_df |> concat(var1, var2, var3,
                                padding_char = "00")

expect_warning(print_stack_as_messages("WARNING"), "<Padding chararacter> must be a single character. Concat will be done without <padding character>.",
               info = "Concatenating only possible with single length padding_char")

expect_equal(test_df[["concat_vec"]], c("11abc", "10NAab", "1003a"), info = "Concatenating only possible with single length padding_char")

test_df[["concat_vec"]] <- test_df |> concat(var1, var2, var3,
                                               padding_char = "")

expect_warning(print_stack_as_messages("WARNING"), "<Padding chararacter> must be a single character. Concat will be done without <padding character>.",
               info = "Concatenating only possible with single length padding_char")

expect_equal(test_df[["concat_vec"]], c("11abc", "10NAab", "1003a"), info = "Concatenating only possible with single length padding_char")


# Concatenating with too short individual padding length
test_df[["concat_vec"]] <- test_df |> concat(var1, var2, var3,
                                               padding_char   = "0",
                                               padding_length = c(5, 6))

expect_warning(print_stack_as_messages("WARNING"), "<Padding length> is shorter than the number of variables.",
               info = "Concatenating with too short individual padding length")

expect_equal(test_df[["concat_vec"]], c("00001000001abc", "000100000000ab", "0010000000300a"), info = "Concatenating with too short individual padding length")


# Concatenating with too long individual padding length
test_df[["concat_vec"]] <- test_df |> concat(var1, var2, var3,
                                padding_char   = "0",
                                padding_length = c(5, 6, 7, 8))

expect_warning(print_stack_as_messages("WARNING"), "<Padding length> is longer than the number of variables.",
               info = "Concatenating with too long individual padding length")

expect_equal(test_df[["concat_vec"]], c("000010000010000abc", "0001000000000000ab", "00100000003000000a"), info = "Concatenating with too long individual padding length")


# Substring throws a warning when variable is provided as a vector
test_df[["sub_vec"]] <- text_df |> sub_string(c(var1, var1), to = 3)

expect_warning(print_stack_as_messages("WARNING"), "<Variable> may only be of length one. The first Element will be used.",
               info = "Substring throws a warning when variable is provided as a vector")


# Substring throws a warning when to is provided as a vector
test_df[["sub_vec"]] <- text_df |> sub_string(var1, to = c(3, 5))

expect_warning(print_stack_as_messages("WARNING"), "<To> may only be of length one. The first Element will be used.",
               info = "Substring throws a warning when to is provided as a vector")


# Substring throws a warning when from is provided as a vector
test_df[["sub_vec"]] <- text_df |> sub_string(var1, from = c(3, 5))

expect_warning(print_stack_as_messages("WARNING"), "<From> may only be of length one. The first Element will be used.",
               info = "Substring throws a warning when from is provided as a vector")


# Remove all blanks with invalid <which> option
text_df2[["gone_blanks"]] <- text_df2 |> remove_blanks(var1, which = "lala")

expect_warning(print_stack_as_messages("WARNING"), "Invalid option for <which> provided. Allowed are 'trim', 'leading', 'trailing', 'all' and",
               info = "Remove all blanks with invalid <which> option")

expect_equal(text_df2[["gone_blanks"]], c("Thisisatext", "HelloWorld", "thisisaText"), info = "Remove all blanks with invalid <which> option")


# Remove blanks only of the first provided variable
text_df2[["gone_blanks"]] <- text_df2 |> remove_blanks(c(var1, var2), which = "all")

expect_warning(print_stack_as_messages("WARNING"), "<Variable> may only be of length one. The first Element will be used.",
               info = "Remove blanks only of the first provided variable")

expect_equal(text_df2[["gone_blanks"]], c("Thisisatext", "HelloWorld", "thisisaText"), info = "Remove blanks only of the first provided variable")

###############################################################################
# Abort checks
###############################################################################

# Substring aborts if variable is not part of the data frame
test_df[["sub_vec"]] <- text_df |> sub_string(var2, to = 3)

expect_error(print_stack_as_messages("ERROR"), "The provided <variable> '", info = "Substring aborts if variable is not part of the data frame")


# Substring aborts if variable is not character
test_df[["sub_vec"]] <- test_df |> sub_string(var1, to = 3)

expect_error(print_stack_as_messages("ERROR"), "<Variable> type must be character. Substring will be aborted.",
               info = "Substring aborts if variable is not character")


# Substring aborts if neither from nor to are provided
test_df[["sub_vec"]] <- text_df |> sub_string(var1)

expect_error(print_stack_as_messages("ERROR"), "Neither <from> nor <to> is provided. Substring will be aborted.",
               info = "Substring aborts if neither from nor to are provided")


# Abort blank removal with no variable provided
text_df2[["gone_blanks"]] <- text_df2 |> remove_blanks()

expect_error(print_stack_as_messages("ERROR"), "No <variable> provided. Blank removal will be aborted.",
               info = "Abort blank removal with no variable provided")


# Abort blank removal if variable is not part of the data frame
text_df2[["gone_blanks"]] <- text_df2 |> remove_blanks(var3)

expect_error(print_stack_as_messages("ERROR"), "No valid <variable> provided. Blank removal will be aborted.",
               info = "Abort blank removal if variable is not part of the data frame")


# Abort blank removal if variable is not character
text_df2[["gone_blanks"]] <- text_df2 |> remove_blanks(var2)

expect_error(print_stack_as_messages("ERROR"), "Blank removal only works with a character <variable>. Blank removal will be aborted.",
               info = "Abort blank removal if variable is not character")


set_no_print()

#' Remove Objects From Memory By Name And Type
#'
#' @description
#' Remove objects by name or type to free up memory. Uses the base [rm()] function
#' but provides more flexible ways to remove objects.
#'
#' @param names Object names to be removed.
#' @param types Object types to be removed.
#' @param envir The environment to remove the objects from.
#'
#' @details
#' [free_memory()] is based on the 'SAS' function Proc Datasets. Among other things
#' this procedure is able to remove datasets from memory. But not only by writing
#' out the full file name, it is capable of looking for datasets starting with, ending
#' with or containing a certain text. Additionally certain object types can be removed.
#'
#' @return Returns removed objects.
#'
#' @examples
#' # Example data frames
#' my_data1 <- dummy_data(10)
#' my_data2 <- dummy_data(10)
#' data     <- dummy_data(10)
#' file     <- dummy_data(10)
#'
#' # Free memory by name
#' free_memory("my_:")
#' free_memory(c("data", "file"))
#'
#' # Free by type
#' my_data1 <- dummy_data(10)
#' my_data2 <- dummy_data(10)
#' data     <- dummy_data(10)
#' file     <- dummy_data(10)
#'
#' free_memory(type = "data.frame")
#'
#' @export
free_memory <- function(names = NULL,
                        types = NULL,
                        envir = .GlobalEnv){

    # If neither names nor types are given, ALL objects will be removed
    objects <- ls(envir = envir)

    # Get objects based on provided names
    if (!is.null(names)){
        if (!is.character(names)){
            print_message("ERROR", "<Names> have to be provided as character. Objects will remain untouched.")
            return(invisible(NULL))
        }

        # If more than one object name is provided, treat them as separate names
        object_vector <- character(0)

        for (name in names){
            # Count the colons, there may not be more than two, otherwise the function
            # should abort
            colon_count <- nchar(gsub("[^:]", "", name))

            # In case of a single name just get the object with the provided name
            if (colon_count == 0){
                object_vector <- c(object_vector, intersect(objects, name))
            }
            # In case of a range
            else if (colon_count == 1){
                # When at the end ("text:"), select all objects which start with the characters
                # coming before the colon. ":" acts as a placeholder for those who come after.
                if (endsWith(name, ":")){
                    search_term   <- substring(name, 1, nchar(name) - 1)
                    object_vector <- c(object_vector, objects[startsWith(objects, search_term)])
                }
                # When at the start (":name"), select all objects which end with the characters
                # following the colon. ":" acts as a placeholder for everything that comes before
                else if (startsWith(name, ":")){
                    search_term   <- substring(name, 2)
                    object_vector <- c(object_vector, objects[endsWith(objects, gsub(":", "", name))])
                }
            }
            # In case there are two colons, one at the start and one at the end (":name:"),
            # look for the text inbetween.
            else if (colon_count == 2 && startsWith(name, ":") && endsWith(name, ":")){
                search_term   <- gsub(":", "", name)
                object_vector <- c(object_vector, objects[grepl(search_term, objects, fixed = TRUE)])

            }
            # If there are more than two colons abort
            else{
                print_message("ERROR", "<Name> has more than two colons which is not allowed. Objects will remain untouched.")
                return(invisible(NULL))
            }
        }

        objects <- object_vector |> collapse::funique()
    }

    # Get objects based on provided types
    if (!is.null(types)){
        if (!is.character(types)){
            print_message("ERROR", "<Types> have to be provided as character. Objects will remain untouched.")
            return(invisible(NULL))
        }

        # Get a logical vector with the length of the number of objects
        objects_to_keep <- logical(length(objects))

        # Look up original symbols and get their type. If the object type is within the
        # provided types it will be kept otherwise dropped.
        for (i in seq_along(objects_to_keep)){
            object <- get(objects[i], envir = envir)
            objects_to_keep[i] <- inherits(object, types)
        }

        objects <- objects[objects_to_keep]
    }

    # Abort if no object found
    if (length(objects) == 0){
        print_message("ERROR", "No object found. Objects will remain untouched.")
        return(invisible(NULL))
    }

    # Compute memory usage of the objects to be removed
    object_sizes <- vapply(objects,
                           function(object){
                               utils::object.size(get(object, envir = envir))
                           },
                           numeric(1))

    total_bytes <- collapse::fsum(object_sizes)

    print_step("MAJOR", "'free_memory' removed [objects] object{?s} and freed approximately [memory].",
                     objects = length(objects),
                     memory  = format(structure(total_bytes, class = "object_size"), units = "auto"))

    # Remove objects from memory
    rm(list = objects, envir = envir)
    gc(verbose = FALSE)

    invisible(objects)
}

#' Display Frequency Tables of Single Variables
#'
#' @description
#' [frequencies()] produces two kinds of tables for a quick overview of single variables.
#' The first table is for a broader overview and contains mean, sd, min, max, freq and missings.
#' The second table is the actual frequency table which shows the weighted sums, percentages
#' and unweighted frequencies per expression.
#'
#' @param data_frame A data frame in which are the variables to tabulate.
#' @param variables A vector of single variables to create frequency tables for.
#' @param formats A list in which is specified which formats should be applied to which variables.
#' @param by Compute tables stratified by the expressions of the provided variables.
#' @param weight Put in a weight variable to compute weighted results.
#' @param means FALSE by default. If TRUE prints a small summarising table which contains mean, sd,
#' min, max, total freq and missing values.
#' @param titles Specify one or more table titles.
#' @param footnotes Specify one or more table footnotes.
#' @param style A list of options can be passed to control the appearance of 'Excel' outputs.
#' Styles can be created with [excel_output_style()].
#' @param output The following output formats are available: console (default), text,
#' excel and excel_nostyle.
#' @param na.rm FALSE by default. If TRUE removes all NA values from the variables.
#' @param print_miss FALSE by default. If TRUE outputs all possible categories of the
#' grouping variables based on the provided formats, even if there are no observations
#' for a combination.
#' @param print TRUE by default. If TRUE prints the output, if FALSE doesn't print anything. Can be used
#' if one only wants to catch the output data frame.
#' @param monitor FALSE by default. If TRUE, outputs two charts to visualize the functions time consumption.
#'
#' @details
#' [frequencies()] is based on the 'SAS' procedure Proc Freq, which provides
#' efficient and readable ways to output frequency tables.
#'
#' To create a frequency table you only need to provide a single variable.
#' Nothing special about this. The real power comes into play, when you output your
#' tables as a fully styled 'Excel' workbook. Setting up a custom, reusable style is
#' as easy as setting up options like: provide a color for the table header, set the
#' font size for the row header, should borders be drawn for the table cells yes/no,
#' and so on.
#'
#' You also can provide multiple single variables to generate multiple, evenly designed
#' tables, all at once. For just a quick overview, rather than fully designed
#' tables, you can also just output the tables in ASCII style format.
#'
#' @return
#' Returns a list of two data tables containing the results for the frequency tables.
#'
#' @seealso
#' Creating a custom table style: [excel_output_style()], [modify_output_style()],
#' [number_format_style()], [modify_number_formats()].
#'
#' Global style options: [set_style_options()], [set_labels()].
#'
#' Other global options: [set_titles()], [set_footnotes()], [set_print()], [set_monitor()],
#' [set_na.rm()], [set_print()], [set_print_miss()], [set_output()].
#'
#' Creating formats: [discrete_format()] and [interval_format()].
#'
#' Functions that can handle formats and styles: [crosstabs()], [any_table()].
#'
#' Additional functions that can handle styles: [export_with_style()]
#'
#' Additional functions that can handle formats: [summarise_plus()], [recode.()],
#' [recode_multi()], [transpose_plus()], [sort_plus()]
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Define titles and footnotes.
#' set_titles("This is title number 1",
#'            "This is title number 2",
#'            "This is title number 3")
#'
#' set_footnotes("This is footnote number 1",
#'               "This is footnote number 2",
#'               "This is footnote number 3")
#'
#' # Output frequencies tables
#' my_data |> frequencies(sex)
#' my_data |> frequencies(c(age, education),
#'                        weight = weight)
#'
#' # Also works with characters
#' my_data |> frequencies("sex")
#' my_data |> frequencies(c("age", "education"),
#'                        weight = "weight")
#'
#' # Applying
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' my_data |> frequencies(sex, formats = list(sex = sex.))
#'
#' # Split frequencies by expressions of another variable
#' my_data |> frequencies(sex, by = education)
#'
#' # Get a list with two data tables for further usage
#' result_list <- my_data |> frequencies(sex, formats = list(sex = sex.))
#'
#' # Output in text file
#' my_data |> frequencies(sex, output = "text")
#'
#' # Output to Excel
#' my_data |> frequencies(sex, output = "excel")
#'
#' # If you want to add hyperlinksto titles and footnotes you can do so by
#' # adding "link:" followed by the hyperlink to the main text. Linking to another
#' # cell works with "cell:". To link to a file use "file:" an pass the full file
#' # path afterwards.
#' set_titles("This is title number 1",
#'            "This is title number 2 link: https://cran.r-project.org/",
#'            "This is title number 3 cell: W22",
#'            "This is title number 4 file: C:/MyFolder/MyFile.txt")
#'
#' set_footnotes("This is footnote number 1",
#'               "This is footnote number 2 file: C:/MyFolder/MyFile.txt",
#'               "This is footnote number 3 cell: W22",
#'               "This is footnote number 4 link: https://cran.r-project.org/")
#'
#' # Individual styling can also be passed directly
#' my_style <- excel_output_style(header_back_color = "0077B6",
#'                                font              = "Times New Roman")
#'
#' my_data |> frequencies(sex, output = "excel", style = my_style)
#'
#' # To save a table as xlsx file you have to set the path and filename in the
#' # style element
#' # Example files paths
#' table_file <- tempfile(fileext = ".xlsx")
#'
#' # Note: Normally you would directly input the path ("C:/MyPath/") and name ("MyFile.xlsx").
#' #       With the set_style_options you can also set a table style globally.
#' set_style_options(save_path  = dirname(table_file),
#'                   file       = basename(table_file),
#'                   sheet_name = "MyTable")
#'
#' my_data |> frequencies(sex, output = "excel")
#'
#' # Manual cleanup for example
#' unlink(table_file)
#'
#' # Global options are permanently active until the current R session is closed.
#' # There are also functions to reset the values manually.
#' reset_style_options()
#' reset_qol_options()
#' close_file()
#'
#' @export
frequencies <- function(data_frame,
                        variables,
                        formats    = c(),
                        by         = c(),
                        weight     = NULL,
                        means      = FALSE,
                        titles     = .qol_options[["titles"]],
                        footnotes  = .qol_options[["footnotes"]],
                        style      = .qol_options[["excel_style"]],
                        output     = .qol_options[["output"]],
                        na.rm      = .qol_options[["na.rm"]],
                        print_miss = .qol_options[["print_miss"]],
                        print      = .qol_options[["print"]],
                        monitor    = .qol_options[["monitor"]]){

    # Measure the time
    print_start_message()
	print_step("GREY", "Error handling")

    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Error handling", "Preparation")
    #-------------------------------------------------------------------------#

    ###########################################################################
    # Early evaluations
    ###########################################################################

    # First convert data frame to data table
    if (!data.table::is.data.table(data_frame)){
        data_frame <- data.table::as.data.table(data_frame)
    }

    # Evaluate formats early
    if (!is_list_of_dfs(formats)){
        formats_list <- as.list(substitute(formats))[-1]
        formats      <- evaluate_formats(formats_list)
    }

    # Remove empty formats and throw a warning. This can happen if there is e.g.
    # a typo in the format.
    for (variable in names(formats)){
        if (is.null(formats[[variable]])){
            formats[[variable]] <- NULL
            print_message("WARNING", "Format for variable '[variable]' does not exist and can't be applied.", variable = variable)
        }
    }

    # Apply macros to save path and file
    if (!is.null(style[["save_path"]])){
        style[["save_path"]] <- macro(style[["save_path"]])
    }

    if (!is.null(style[["file"]])){
        style[["file"]] <- macro(style[["file"]])
    }

    ###########################################################################
    # Error handling
    ###########################################################################

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    variables <- get_origin_as_char(variables, substitute(variables))

    # Make sure that the variables provided are part of the data frame.
    variables <- data_frame |> part_of_df(variables)

    if (length(variables) <= 1){
        if (length(variables) == 0 || variables == ""){
            print_message("ERROR", "No valid <variables> provided. Frequencies will be aborted.")
            return(invisible(NULL))
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # By variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    by <- get_origin_as_char(by, substitute(by))

    # Make sure that the variables provided are part of the data frame.
    by <- data_frame |> part_of_df(by)

    by <- resolve_intersection(by, variables)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Weight
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    weight     <- get_origin_as_char(weight, substitute(weight))
    data_frame <- data_frame |> check_weight(weight)
    weight_var <- ".temp_weight"

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Output
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Check for invalid output option
    if (!tolower(output) %in% c("console", "text", "excel", "excel_nostyle")){
        print_message("WARNING", "<Output> format '[output]' not available. Using 'console' instead.", output = output)

        output <- "console"
    }
    else{
        output <- tolower(output)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Resolve macros
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    titles    <- apply_macro(titles)
    footnotes <- apply_macro(footnotes)

    ###########################################################################
    # Frequency starts
    ###########################################################################

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Mean summary", "Summary")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Computing stats.")

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Summarise mean table according to provided variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (print_miss && means){
        print_message("NOTE", "Wenn <print_miss> is TRUE, there will be no mean tables. <Means> is set to FALSE.")

        means <- FALSE
    }

    if (means){
        # For mean tabulation only numeric variables are allowed. So here all character
        # variables are excluded temporarily.
        vars_mean <- variables

        for (variable in vars_mean){
            if (is.character(data_frame[[variable]])){
                vars_mean <- vars_mean[vars_mean != variable]
            }
        }

        # In case no by variables are specified
        if (length(by) == 0){
            # Compute stats for mean summary
            mean_tab <- suppressMessages(data_frame |>
                 summarise_plus(values     = vars_mean,
                                statistics = c("mean", "sd", "min", "max", "freq", "missing"),
                                weight     = weight_var,
                                notes      = FALSE,
                                na.rm      = na.rm) |>
                     drop_type_vars())
        }
        # In case by variables are specified
        else{
            mean_tab <- suppressMessages(data_frame |>
                 summarise_plus(class      = by,
                                values     = vars_mean,
                                statistics = c("mean", "sd", "min", "max", "freq", "missing"),
                                formats    = formats,
                                weight     = weight_var,
                                nesting    = "all",
                                types      = by,
                                notes      = FALSE,
                                na.rm      = na.rm))

            if (!is.null(mean_tab)){
                mean_tab <- mean_tab |>
                    fuse_variables("by_vars", by) |>
                    collapse::fsubset(DEPTH > 0)
            }
        }

        # Put results for each variable below each other instead of beside
        mean_columns <- c("mean", "sd", "min", "max", "freq", "miss")
        all_rows     <- list()

        for (variable in vars_mean){
            # Extract all stats for a single variable and give them uniform names
            var_columns <- grep(variable, names(mean_tab), value = TRUE)
            var_row <- suppressMessages(mean_tab |> keep("TYPE", "by_vars", var_columns))

            if (length(by) == 0){
                names(var_row) <- mean_columns
            }
            else{
                names(var_row) <- c("TYPE", "by_vars", mean_columns)
            }

            # Put variable name in front and put formatted row below the other ones
            # if there are any.
            var_row              <- cbind(variable, var_row)
            all_rows[[variable]] <- var_row
        }

        mean_tab <- data.table::rbindlist(all_rows, use.names = TRUE, fill = TRUE)
    }
    else{
        mean_tab <- c()
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Summarise results for a frequency output per variable
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Freq summary", "Summary")
    #-------------------------------------------------------------------------#

    data_frame[["var"]] <- 1

    # In case no by variables are specified
    if (length(by) == 0){
        freq_tab <- suppressMessages(data_frame |>
             summarise_plus(class      = variables,
                            values     = "var",
                            statistics = c("sum", "pct_group", "freq"),
                            formats    = formats,
                            weight     = weight_var,
                            nesting    = "single",
                            notes      = FALSE,
                            na.rm      = na.rm,
                            print_miss = print_miss))
    }
    # In case by variables are specified
    else{
        group_vars   <- c(by, variables)
        combinations <- as.vector(outer(by, variables, paste, sep = "+"))
        combinations <- c("total", combinations)

        freq_tab <- suppressMessages(data_frame |>
             summarise_plus(class      = group_vars,
                            values     = "var",
                            statistics = c("sum", "pct_group", "freq"),
                            formats    = formats,
                            weight     = weight_var,
                            nesting    = "all",
                            types      = combinations,
                            notes      = FALSE,
                            na.rm      = na.rm,
                            print_miss = print_miss)) |>
             fuse_variables("fused_vars", variables) |>
             fuse_variables("by_vars", by)

        freq_tab[1, "by_vars"] <- "total"
        freq_tab[["BY"]]       <- sub("\\+.*", "", freq_tab[["TYPE"]])
        freq_tab[["TYPE"]]     <- sub(".*\\+", "", freq_tab[["TYPE"]])
    }

    if (is.null(freq_tab)){
        print_message("ERROR", "Frequencies could not be computed.")
        return(invisible(NULL))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Prepare table format for output
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    print_step("MAJOR", "Formatting tables.")

    if (output %in% c("console", "text")){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Format tables", "Format tables")
        #---------------------------------------------------------------------#

        # In case no by variables are provided
        if (length(by) == 0){
            complete_mean  <- format_mean_text(mean_tab, vars_mean, mean_columns, means)
            complete_freq  <- format_freq_text(freq_tab, variables, formats,
                                               by, titles, footnotes, style)

            complete_table <- c(complete_mean, complete_freq)
        }
        # In case there are  by variables are provided
        else{
            complete_table <- format_by_text(mean_tab, freq_tab, variables, mean_columns,
                                             formats, by, titles, footnotes, na.rm, means,
                                             style)
        }
    }
    else if (output == "excel" || output == "excel_nostyle"){
        wb <- openxlsx2::wb_workbook() |>
            prepare_styles(list("title" = titles, "footnote" = footnotes), style)

        monitor_df <- monitor_df |> monitor_end()

        # In case no by variables are provided
        if (length(by) == 0){
            wb_list <- format_mean_excel(mean_tab, mean_columns, style, output, means, wb = wb,
                                         monitor_df = monitor_df)
            wb_list <- format_freq_excel(wb_list[[1]], freq_tab, variables, formats,
                                         by, titles, footnotes, style, output, monitor_df = wb_list[[2]])

            wb         <- wb_list[[1]]
            monitor_df <- wb_list[[2]]
        }
        # In case there are  by variables are provided
        else{
            wb_list <- format_by_excel(mean_tab, freq_tab, variables, mean_columns,
                                       formats, by, titles, footnotes, style, output,
                                       na.rm, print_miss, wb, monitor_df, means)

            wb         <- wb_list[[1]]
            monitor_df <- wb_list[[2]]
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Output formatted table into different formats
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (print){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Output tables", "Output tables")
        #---------------------------------------------------------------------#

        if (output %in% c("console")){
            print_closing()
            cat(paste(complete_table, collapse = "\n"), "\n\n")
        }
        # Open in text editor
        else if (output == "text"){
            temp_file <- tempfile(fileext = ".txt")
            writeLines(complete_table, temp_file)

            if (interactive()){
                file.show(temp_file)
            }

            print_closing()
        }
        else if (output == "excel" || output == "excel_nostyle"){
            # Check if only save path or file name is specified. If only one is specified
            # print a note. Otherwise the file would not be saved and opened without a
            # hint to why the file wasn't saved.
            if (is.null(style[["save_path"]]) + is.null(style[["file"]]) == 1){
                if (is.null(style[["save_path"]])){
                    print_message("NOTE", c("No save path specified. Both save path and file name with extension",
                                            "need to be specified in the global options or style parameter for",
                                            "the file to be saved. File won't be saved."))
                }
                else{
                    print_message("NOTE", c("No file name specified. Both save path and file name with extension",
                                            "need to be specified in the global options or style parameter for",
                                            "the file to be saved. File won't be saved."))
                }
            }

            # If no save path or file provided just open workbook
            if (is.null(style[["save_path"]]) || is.null(style[["file"]])){
                if (interactive()){
                    wb$open()
                }
            }
            else{
                # If save path doesn't exist, just open workbook
                if (!file.exists(style[["save_path"]])){
                    print_message("WARNING", "Path does not exist: ", style[["save_path"]])

                    if (interactive()){
                        wb$open()
                    }
                }
                # Save file
                else{
                    extension <- tolower(tools::file_ext(style[["file"]]))

                    # CSV export is based on the actual data frame
                    if (extension == "csv"){
                        freq_tab |> data.table::fwrite(file = paste0(style[["save_path"]], "/", style[["file"]]),
                                                       sep  = ";",
                                                       dec  = ",")
                    }
                    # While XLSX export is based on the workbook
                    else if (extension == "xlsx"){
                        wb$save(file = paste0(style[["save_path"]], "/", style[["file"]]), overwrite = TRUE)
                    }
                    # If there is no extension or a wrong one, auto correct it
                    else{
                        wb$save(file = paste0(style[["save_path"]], "/", basename(style[["file"]]), ".xlsx"), overwrite = TRUE)
                    }
                }
            }

            print_closing()
        }
    }
    else{
        print_closing()
    }

    monitor_df <- monitor_df |> monitor_end()
    monitor_df |> monitor_plot(draw_plot = monitor)

    invisible(structure(list("mean" = mean_tab,
                             "freq" = freq_tab), class = "qol_freq"))
}

###############################################################################
# Format numbers with decimal points as thousands
###############################################################################
#' Enhance Readability for Large Numbers
#'
#' @description
#' Format numbers with big marks to make them more readable in the text output.
#' Also set up number of decimals.
#'
#' @param number The number to format.
#' @param width Width provided by the largest number in the table column to format
#' all numbers i a column at an even width.
#' @param stat Stat that determines the numbers formatting.
#' @param decimals Number of decimals.
#'
#' @return
#' Returns a formatted number.
#'
#' @noRd
format_number <- function(number,
                          width,
                          stat,
                          decimals = 1){
    number <- unlist(number)

    # Format numbers to make them more readable
    if (stat %in% c("freq", "miss", "var_freq", "var_cum_freq")){
        output_format <- format(number,
                                format       = "d",
                                decimal.mark = ",",
                                big.mark     = ".",
                                scientific   = FALSE,
                                nsmall       = 0)
    }
    else if (stat %in% c("min", "max")){
        output_format <- format(round_values(number, decimals),
                                format       = "d",
                                decimal.mark = ",",
                                big.mark     = ".",
                                scientific   = FALSE)
    }
    # Any other stat with decimal numbers
    else{
        output_format <- format(round_values(number, decimals),
                                format       = "f",
                                decimal.mark = ",",
                                big.mark     = ".",
                                scientific   = FALSE,
                                nsmall       = decimals)
    }

    # Remove NA
    output_format <- sub("NA", "", output_format)

    # Output new format
    format(output_format,
           width   = width,
           justify = "right")
}

###############################################################################
# Get maximum column width for equal formatting
###############################################################################
#' Get Maximum Column Width
#'
#' @description
#' Get maximum column width based on the largest text or number written in the column.
#'
#' @param data_frame The data frame which contains the columns to measure.
#' @param columns_to_format The actual columns from the data frame to be measured for
#' their widest texts or numbers.
#' @param decimals Number of decimals for numbers.
#'
#' @return
#' Returns a vector of column widths.
#'
#' @noRd
get_column_width <- function(data_frame,
                             columns_to_format,
                             decimals = 1){
    all_widths <- c()

    # Get the maximum column width by looking up which is the largest number
    # individually in each column.
    for (column in columns_to_format){
        max_width <- collapse::vlengths(column)

        if (column %in% c("freq", "miss", "var_freq", "var_cum_freq")){
            max_width <- max(max_width,
                             collapse::vlengths(format(data_frame[[column]],
                                                      format     = "d",
                                                      big.mark   = ",",
                                                      scientific = FALSE,
                                                      nsmall     = 0)))
        }
        # Any other stat with decimal numbers
        else{
            max_width <- max(max_width,
                             collapse::vlengths(format(round_values(data_frame[[column]], decimals),
                                                      format     = "f",
                                                      big.mark   = ",",
                                                      scientific = FALSE,
                                                      nsmall     = decimals)))
        }

        all_widths <- c(all_widths, max_width)
    }

    all_widths
}

###############################################################################
# Format mean table for console output
###############################################################################
#' Format Mean Table Output (Text Based)
#'
#' @description
#' Format a short summarising table for all numeric variables.
#'
#' @param mean_tab The data frame which contains the information for this summarising
#' table.
#' @param variables The variables that should appear in the table.
#' @param mean_columns The column names from the data frames that should appear
#' in the table.
#' @param means FALSE by default. If TRUE prints a small summarising table which contains mean, sd,
#' min, max, total freq and missing values.
#'
#' @return
#' Returns a character vector with all formatted rows for the table.
#'
#' @noRd
format_mean_text <- function(mean_tab,
                             variables,
                             mean_columns,
                             means){
    if (is.null(mean_tab) || !means || collapse::fnrow(mean_tab) == 0){
        return(c())
    }

    if ("TYPE" %in% names(mean_tab)){
        mean_tab <- mean_tab |>
            dropp("TYPE", "by_vars")
    }

    # Get the maximum width of the provided variable names to determine the width
    # of the first column.
    first_column_width <- collapse::vlengths("variable")

    for (variable in variables){
        first_column_width <- max(first_column_width,
                                  collapse::vlengths(mean_tab[["variable"]]))
    }

    # Set header row formatting. Loop through all header columns and give
    # each column the optimal width.
    column_widths  <- get_column_width(mean_tab, mean_columns)
    header_columns <- c()

    for (i in seq_len(length(mean_columns))){
        header_columns <- c(header_columns,
                            format(mean_columns[i],
                                   width   = column_widths[i],
                                   justify = "right"))
    }

    # Put together the first two rows of the table as complete table header
    header_row <- paste0(format("variable",
                                width   = first_column_width,
                                justify = "left"), "  |   ",
                         paste(header_columns, collapse = "   "))

    complete_header <- c(header_row,
                         strrep("-", collapse::vlengths(header_row)))

    # Format table column by column. Basically concatenate first column text and
    # corresponding values together while keeping the individual maximum column
    # width in mind.
    first_column <- paste0(format(mean_tab[["variable"]],
                                  width   = first_column_width,
                                  justify = "left"), "  |   ")

    # Loop through stat columns and format each number in a readable way
    formatted_cols <- vector("list", length(mean_columns))

    for (column in seq_len(length(mean_columns))){
        formatted_cols[[column]] <- format_number(mean_tab[column + 1],
                                                  width = column_widths[column],
                                                  stat  = names(mean_tab[column + 1]))
    }

    # Convert list to matrix
    formatted_matrix <- do.call(cbind, formatted_cols)

    # Convert matrix to single formatted rows
    all_rows <- paste0(first_column,
                       apply(formatted_matrix, 1, paste, collapse = "   "))

    # Output formatted result
    c("", complete_header, all_rows)
}

###############################################################################
# Format mean table for excel output
###############################################################################
#' Format Mean Table Output (Excel Based)
#'
#' @description
#' Format a short summarising table for all numeric variables.
#'
#' @param mean_tab The data frame which contains the information for this summarising
#' table.
#' @param mean_columns The column names from the data frames that should appear
#' in the table.
#' @param style A list containing the styling elements.
#' @param output Determines whether to style the output or to just quickly paste
#' the data.
#' @param means FALSE by default. If TRUE prints a small summarising table which contains mean, sd,
#' min, max, total freq and missing values.
#' @param wb An already created workbook to add more sheets to.
#' @param index Index of the current variable expression. Used for computation with
#' by variables.
#' @param monitor_df Data frame which stores the monitoring values.
#'
#' @return
#' Returns a list containing a formatted Excel workbook as well as the monitoring
#' data frame.
#'
#' @noRd
format_mean_excel <- function(mean_tab,
                              mean_columns,
                              style = excel_output_style(),
                              output,
                              means,
                              wb    = NULL,
                              index = NULL,
                              monitor_df){
    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_start("Excel prepare (mean)", "Format mean")
    #-------------------------------------------------------------------------#

    if (is.null(mean_tab) || !means || collapse::fnrow(mean_tab) == 0){
        monitor_df <- monitor_df |> monitor_end()

        return(list(wb, monitor_df))
    }

    # Drop TYPE if present
    if ("TYPE" %in% names(mean_tab)){
        mean_tab <- mean_tab |> dropp("TYPE", "by_vars")
    }

    # Round values according to style options
    mean_tab <- mean_tab |>
        collapse::fmutate(mean = round_values(mean, style[["number_formats"]][["mean_decimals"]]),
                          sd   = round_values(sd,   style[["number_formats"]][["sd_decimals"]]),
                          min  = round_values(min,  style[["number_formats"]][["min_decimals"]]),
                          max  = round_values(max,  style[["number_formats"]][["max_decimals"]]))

    # Build header row
    header     <- c("variable", mean_columns)
    table_data <- rbind(header, as.matrix(mean_tab[, c("variable", mean_columns)]))

    # Get table ranges
    mean_ranges <- get_mean_tab_ranges(mean_tab, style)

    # If function is called with by variables the sheet names have to be differentiated
    if (!is.null(index)){
        wb$add_worksheet(paste0("means", index),
                         grid_lines = style[["grid_lines"]])
    }
    # If function is called without by variables
    else{
        wb$add_worksheet("means", grid_lines = style[["grid_lines"]])
    }

    # Add table data and format according to style options
    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Excel data (mean)", "Format mean")
    #-------------------------------------------------------------------------#

    wb$add_data(x          = mean_tab,
                start_col  = style[["start_column"]],
                start_row  = style[["start_row"]],
                na.strings = style[["na_symbol"]])

    # Only do the formatting when user specified it. With the excel_nostyle
    # option this whole part gets omitted to get a very quick unformatted
    # excel output.
    if (output == "excel"){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel cell styles (mean)", "Format mean")
        #---------------------------------------------------------------------#
        wb <- wb |> handle_cell_styles(mean_ranges, style)

        # Set up inner table number formats
        col_index <- 1

        for (type in c("mean", "sd", "min", "max", "freq", "missing")){
            wb$add_cell_style(dims                = mean_ranges[[paste0("mean_col_ranges", col_index)]],
                              apply_number_format = TRUE,
                              num_fmt_id          = wb$styles_mgr$get_numfmt_id(paste0(type, "_numfmt")))

            col_index <- col_index + 1
        }

        # If there are width or heights defined
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel widths/heights (mean)", "Format mean")
        #---------------------------------------------------------------------#

        column_width <- style[["column_widths"]]
        row_heights  <- style[["row_heights"]]

        if (length(column_width) == 1 && length(row_heights) == 1){
            if (column_width != "auto" || row_heights != "auto"){
                wb <- wb |> handle_col_row_dimensions(mean_ranges,
                                                      collapse::fncol(mean_tab) + (style[["start_column"]] - 1),
                                                      collapse::fnrow(mean_tab) + (style[["start_row"]] - 1),
                                                      style)
            }
        }
        # If both widths and heights should be determined automatically.
        else{
            wb <- wb |> handle_col_row_dimensions(mean_ranges,
                                                  style[["start_column"]],
                                                  collapse::fnrow(mean_tab) + (style[["start_row"]] - 1),
                                                  style)
        }

        # If autofit columns is selected, set a manual size after the first table
        # column which is sufficient in most cases. The integrated autofit often
        # gives to narrow columns.
        if (length(column_width) == 1){
            if (column_width == "auto"){
                start_col <- style[["start_column"]] + 1

                wb$set_col_widths(cols   = start_col:mean_ranges[["table.end"]],
                                  widths = 10)
            }
        }

        wb$add_named_region(dims = mean_ranges[["whole_tab_range"]], name = "table", local_sheet = TRUE)
        wb$add_named_region(dims = mean_ranges[["table_range"]],     name = "data",  local_sheet = TRUE)
    }

    # Return workbook
    monitor_df <- monitor_df |> monitor_end()

    list(wb, monitor_df)
}

###############################################################################
# Format freq table for console output
###############################################################################
#' Format Frequency Table Output (Text Based)
#'
#' @description
#' Format a frequency table for each variable which shows the counts, sums and
#' percentages per variable expression.
#'
#' @param freq_tab The data frame which contains the information for this frequency
#' table.
#' @param variables The variables that should appear in the table.
#' @param formats A list in which is specified which formats should be applied to
#' which variable.
#' @param by Separate the frequency output by the expressions of the provided variables.
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#'
#' @return
#' Returns a character vector with all formatted rows for the tables.
#'
#' @noRd
format_freq_text <- function(freq_tab,
                             variables,
                             formats,
                             by,
                             titles,
                             footnotes,
                             style){
    complete_tabs <- c()

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Generate a separate frequency table for each variable
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    for (variable in variables){
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Prepare header
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Compute additional stats
        var_tab <- compute_cumulative(freq_tab, variable, formats)

        # If no weight variable was applied
        if (identical(var_tab[["var_sum"]], var_tab[["var_freq"]])){
            freq_columns <- c("var_freq", "pct_group", "var_cum_freq",
                              "var_cum_pct")

            freq_top_columns    <- c("    ", "       ", "cumulative", "cumulative")
            freq_bottom_columns <- c("freq", "percent",       "freq",    "percent")
        }
        # If weight variable was applied
        else{
            freq_columns <- c("var_sum", "pct_group", "var_cum_sum",
                              "var_cum_pct", "var_freq", "var_cum_freq")

            freq_top_columns    <- c("   ", "       ", "cumulative", "cumulative", "    ", "cumulative")
            freq_bottom_columns <- c("sum", "percent",        "sum",    "percent", "freq",       "freq")
        }

        var_tab <- var_tab |>
            data.table::setcolorder(c("fused_vars", freq_columns))

        # Get the maximum width of the provided variable names to determine the width
        # of the first column.
        first_column_width <- max(collapse::vlengths(variable), collapse::vlengths("total"))

        first_column_width <- max(first_column_width,
                                  collapse::vlengths(collapse::na_omit(var_tab[["fused_vars"]])))

        # Set header row formatting. Loop through all header columns and give
        # each column the optimal width.
        column_widths         <- get_column_width(var_tab, freq_columns, decimals = 3)
        header_top_columns    <- c()
        header_bottom_columns <- c()

        var_tab <- var_tab |> collapse::frename("pct_group" = "var_pct_group")

        for (i in seq_len(length(freq_bottom_columns))){
            header_top_columns <- c(header_top_columns,
                                    format(freq_top_columns[i],
                                           width   = column_widths[i],
                                           justify = "right"))

            header_bottom_columns <- c(header_bottom_columns,
                                       format(freq_bottom_columns[i],
                                              width   = column_widths[i],
                                              justify = "right"))
        }

        # Put together the first two rows of the table as complete table header
        header_top_row <- paste0(format("",
                                        width   = first_column_width,
                                        justify = "left"), "  |  ",
                                 paste(header_top_columns,    collapse = "  "), "\n")

        header_row <- paste0(header_top_row,
                             format(variable,
                                    width   = first_column_width,
                                    justify = "left"), "  |  ",
                             paste(header_bottom_columns, collapse = "  "))

        complete_header <- c(header_row,
                             strrep("-", collapse::vlengths(header_top_row)))

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Format table column by column. Basically concatenate first column text and
        # corresponding values together while keeping the individual maximum column
        # width in mind.
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        first_column <- paste0(format(var_tab[["fused_vars"]],
                                      width   = first_column_width,
                                      justify = "left"), "  |  ")

        # Loop through stat columns and format each number in a readable way
        formatted_cols <- vector("list", length(freq_columns))

        for (column in seq_len(length(freq_columns))){
            # Weighted sums
            if (column %in% c(1, 3)){
                formatted_cols[[column]] <- format_number(var_tab[column + 1],
                                                          width    = column_widths[column],
                                                          stat     = names(var_tab[column + 1]),
                                                          decimals = style[["number_formats"]][["sum_decimals"]])
            }
            # Percentages and Unweighted freq
            else{
                formatted_cols[[column]] <- format_number(var_tab[column + 1],
                                                          width = column_widths[column],
                                                          stat  = names(var_tab[column + 1]))
            }
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Put everything together
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Convert list to matrix
        formatted_matrix <- do.call(cbind, formatted_cols)

        # Convert matrix to single formatted rows
        all_rows <- paste0(first_column,
                           apply(formatted_matrix, 1, paste, collapse = "  "))

        # If no multilabel formats are applied separate total row at the bottom from the rest
        if (!is_multilabel(formats, variable)){
            rows <- length(all_rows)

            all_rows <- c(all_rows[1:(rows - 1)],
                          strrep("-", collapse::vlengths(header_top_row)),
                          all_rows[rows])
        }
        # If multilabel formats are applied remove total row
        else{
            all_rows <- all_rows[-length(all_rows)]

            # Only do this step with no by variables here otherwise the separate
            # by function down below will handle this message, so that it appears only
            # once and not for each loop.
            if (length(by) == 0){
                print_message("NOTE", c("The format for variable '[variable]' is a multilabel.",
										"In this case cumulative results aren't computed properly."),
										variable = variable, always_print = TRUE)
            }
        }

        # Insert titles and footnotes
        complete_table <- c(complete_header, all_rows)

        if (length(titles) > 0){
            complete_table <- c(titles, "", complete_table)
        }
        if (length(footnotes) > 0){
            complete_table <- c(complete_table, "____________", footnotes)
        }

        # Output formatted result
        complete_tabs <- c(complete_tabs, "\n", complete_table, "\n")

    }

    # Output formatted result
    complete_tabs
}


#' Compute Cumulative Variables
#'
#' @description
#' Compute separate cumulative variables for frequencies, weighted sums and
#' percentages.
#'
#' @param freq_tab The data frame which contains the information for this frequency
#' table.
#' @param variabls The currently computed variable.
#' @param formats A list in which is specified which formats should be applied to
#' which variable.
#'
#' @return
#' Returns a data frame with added cumulative variables.
#'
#' @noRd
compute_cumulative <- function(freq_tab,
                               variable,
                               formats){
    # Sort total to bottom if no by variables are specified
    if (!"BY" %in% names(freq_tab)){
        # Reduce data frame to the necessary rows
        var_tab <- freq_tab |> collapse::fsubset(TYPE %in% c("total", variable))

        var_tab <- var_tab |> data.table::setorder(-"DEPTH")
    }
    # Compute new total if by variables are specified because otherwise the grand total would be
    # used for every single expression.
    else{
        total_temp <- freq_tab |>
            collapse::fsubset(TYPE %in% variable) |>
            collapse::fsummarise(across(c("var_sum", "var_pct_group", "var_freq"), collapse::fsum))

        var_tab <- data.table::rbindlist(list(freq_tab, total_temp), use.names = TRUE, fill = TRUE)
        var_tab <- var_tab |>
            collapse::fsubset(TYPE %in% c(variable, NA))
    }

    var_tab[["fused_vars"]][collapse::fnrow(var_tab)] <- "total"

    var_tab[["var_cum_sum"]]  <- collapse::fcumsum(var_tab[["var_sum"]])
    var_tab[["var_cum_pct"]]  <- collapse::fcumsum(var_tab[["var_pct_group"]])
    var_tab[["var_cum_freq"]] <- collapse::fcumsum(var_tab[["var_freq"]])

    # Erase total cumulative values
    var_tab[["var_cum_sum"]][collapse::fnrow(var_tab)]  <- NA
    var_tab[["var_cum_pct"]][collapse::fnrow(var_tab)]  <- NA
    var_tab[["var_cum_freq"]][collapse::fnrow(var_tab)] <- NA

    # Order columns in order of freq table
    var_tab |> collapse::frename("var_pct_group" = "pct_group")
}

###############################################################################
# Format freq table for excel output
###############################################################################
#' Format Frequency Table Output (Excel Based)
#'
#' @description
#' Format a frequency table for each variable which shows the counts, sums and
#' percentages per variable expression.
#'
#' @param wb An already created workbook to add more sheets to.
#' @param freq_tab The data frame which contains the information for this frequency
#' table.
#' @param variables The variables that should appear in the table.
#' @param formats A list in which is specified which formats should be applied to
#' which variable.
#' @param by Separate the frequency output by the expressions of the provided variables.
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#' @param style A list containing the styling elements.
#' @param output Determines whether to style the output or to just quickly paste
#' the data.
#' @param by_info Text which contains the information which variable with which
#' expression is computed at the moment.Used for computation with by variables.
#' @param index Index of the current variable expression. Used for computation with
#' by variables.
#' @param monitor_df Data frame which stores the monitoring values.
#'
#' @return
#' Returns a list containing a formatted Excel workbook as well as the monitoring
#' data frame.
#'
#' @noRd
format_freq_excel <- function(wb,
                              freq_tab,
                              variables,
                              formats,
                              by,
                              titles,
                              footnotes,
                              style,
                              output,
                              by_info = NULL,
                              index   = NULL,
                              monitor_df){
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Generate a separate frequency table for each variable
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    for (variable in variables){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_start("Excel prepare (freq)", "Format freq")
        #---------------------------------------------------------------------#

        # Compute additional stats
        var_tab <- compute_cumulative(freq_tab, variable, formats)

        # Round values according to style options
        var_tab <- var_tab |>
            collapse::fmutate(var_sum     = round_values(var_sum,     style[["number_formats"]][["sum_decimals"]]),
                              var_cum_sum = round_values(var_cum_sum, style[["number_formats"]][["sum_decimals"]]),
                              pct_group   = round_values(pct_group,   style[["number_formats"]][["pct_decimals"]]),
                              var_cum_pct = round_values(var_cum_pct, style[["number_formats"]][["pct_decimals"]]))

        # If no weight variable was applied
        if (identical(var_tab[["var_sum"]], var_tab[["var_freq"]])){
            freq_columns <- c("var_freq", "pct_group", "var_cum_freq",
                              "var_cum_pct")

            freq_col_labels <- c("freq", "percent", "cumulative freq", "cumulative percent")

            by_info_col <- style[["start_column"]] + 6
        }
        # If weight variable was applied
        else{
            freq_columns <- c("var_sum", "pct_group", "var_cum_sum",
                              "var_cum_pct", "var_freq", "var_cum_freq")

            freq_col_labels <- c("sum", "percent", "cumulative sum",
                                 "cumulative percent", "freq", "cumulative freq")

            by_info_col <- style[["start_column"]] + 8
        }

        # If multilabel, remove automatically computed total observation.
        # Only do this step with no by variables here otherwise the separate
        # by function down below will handle this message, so that it appears only
        # once and not for each loop.
        var_is_multilabel <- FALSE

        if (is_multilabel(formats, variable)){
            var_is_multilabel <- TRUE

            if (length(by) == 0){
                print_message("NOTE", c("The format for variable '[variable]' is a multilabel.",
										"In this case cumulative results aren't computed properly."),
										variable = variable, always_print = TRUE)
            }

            var_tab <- var_tab |> collapse::fsubset(TYPE != "total")
        }

        var_tab <- var_tab |>
            keep("fused_vars", freq_columns, order_vars = TRUE)

        names(var_tab) <- c(variable, freq_col_labels)

        # Get table ranges
        freq_ranges <- get_freq_tab_ranges(var_tab, titles, footnotes, style)

        # If function is called with by variables the sheet names have to be differentiated
        # and by info has to be written besides the table.
        if (!is.null(by_info)){
            wb$add_worksheet(paste0(variable, index),
                                    grid_lines = style[["grid_lines"]])
            wb$add_data(x         = by_info,
                        start_col = by_info_col,
                        start_row = style[["start_row"]])
        }
        else{
            wb$add_worksheet(variable, grid_lines = style[["grid_lines"]])
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Apply style
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel titles/footnotes (freq)", "Format freq")
        #---------------------------------------------------------------------#
        # Format titles and footnotes if there are any
        wb <- wb |>
            format_titles_foot_excel(titles, footnotes, freq_ranges, style, output)

        # Add table data and format according to style options
        monitor_df <- monitor_df |> monitor_next("Excel data (freq)", "Format freq")

        wb$add_data(x          = var_tab,
                    start_col  = style[["start_column"]],
                    start_row  = freq_ranges[["header.row"]],
                    na.strings = style[["na_symbol"]])

        # Only do the formatting when user specified it. With the excel_nostyle
        # option this whole part gets omitted to get a very quick unformatted
        # excel output.
        if (output == "excel"){
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Excel cell styles (freq)", "Format freq")
            #-----------------------------------------------------------------#
            wb <- wb |> handle_cell_styles(freq_ranges, style)

            # Set up inner table number formats
            col_index <- 1
            col_types <- c("sum", "pct", "sum", "pct", "freq", "freq")

            if (!"sum" %in% names(var_tab)){
                col_types <- c("freq", "pct", "freq", "pct")
            }

            for (type in col_types){
                wb$add_cell_style(dims                = freq_ranges[[paste0("freq_col_ranges", col_index)]],
                                  apply_number_format = TRUE,
                                  num_fmt_id          = wb$styles_mgr$get_numfmt_id(paste0(type, "_numfmt")))

                col_index <- col_index + 1
            }

            # Draw inner table cells as heat map with conditional formatting
            if (style[["as_heatmap"]]){
                #-----------------------------------------------------------------#
                monitor_df <- monitor_df |> monitor_next("Excel format heatmap", "Format")
                #-----------------------------------------------------------------#

                # First exclude the total value, if present, otherwise heat map won't
                # work, because total will always be the highest value.
                if (!var_is_multilabel){
                    old_number <- max(2, as.numeric(sub(".*(\\d+)$", "\\1", freq_ranges[["freq_col_ranges1"]])))
                    new_range  <- sub(paste0(old_number, "$"), old_number - 1, freq_ranges[["freq_col_ranges1"]])
                }
                else{
                    new_range <- freq_ranges[["freq_col_ranges1"]]
                }

                wb$add_conditional_formatting(dims  = new_range,
                                              style = c(style[["heatmap_low_color"]],
                                                        style[["heatmap_middle_color"]],
                                                        style[["heatmap_high_color"]]),
                                              type  = "colorScale")
            }

            # Adjust table dimensions
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Excel widths/heights (freq)", "Format freq")
            #-----------------------------------------------------------------#
            wb <- wb |> handle_col_row_dimensions(freq_ranges,
                                                  collapse::fncol(var_tab) + (style[["start_column"]] - 1),
                                                  collapse::fnrow(var_tab) + (style[["start_row"]] - 1),
                                                  style)

            wb <- wb |> handle_auto_dimensions(freq_ranges,
                                               style)

            wb <- wb |> handle_header_table_dim(freq_ranges,
                                                style)

            wb$add_ignore_error(dims = freq_ranges[["cat_col_range"]], number_stored_as_text = TRUE)

            wb$add_named_region(dims = freq_ranges[["whole_tab_range"]], name = "table", local_sheet = TRUE)
            wb$add_named_region(dims = freq_ranges[["table_range"]],     name = "data",  local_sheet = TRUE)
        }

        monitor_df <- monitor_df |> monitor_end()
    }

    # Return workbook
    list(wb, monitor_df)
}

###############################################################################
# Format grouped by tables for console output
###############################################################################
#' Format Frequency Table Output with by Variables (Text Based)
#'
#' @description
#' Format a frequency table for each variable which shows the counts, sums and
#' percentages per variable expression.
#'
#' @param mean_tab The data frame which contains the information for this summarising
#' table.
#' @param freq_tab The data frame which contains the information for this frequency
#' table.
#' @param variables The variables that should appear in the table.
#' @param mean_columns The column names from the data frames that should appear
#' in the table.
#' @param formats A list in which is specified which formats should be applied to
#' which variable.
#' @param by Separate the output by the expressions of the provided variables.
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#' @param na.rm If TRUE removes all NA values from the tabulation.
#' @param means FALSE by default. If TRUE prints a small summarising table which contains mean, sd,
#' min, max, total freq and missing values.
#'
#' @return
#' Returns a character vector with all formatted rows for the tables.
#'
#' @noRd
format_by_text <- function(mean_tab,
                           freq_tab,
                           variables,
                           mean_columns,
                           formats,
                           by,
                           titles,
                           footnotes,
                           na.rm,
                           means,
                           style){
    # Print message if multilabels are applied
    for (variable in variables){
        if (is_multilabel(formats, variable)){
            print_message("NOTE", c("The format for variable '[variable]' is a multilabel.",
									"In this case cumulative results aren't computed properly."),
									variable = variable, always_print = TRUE)
        }
    }

    complete_tabs <- c()

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Loop through all by variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    for (by_var in by){
        # Select by variables one by one
        if (means){
            mean_by <- mean_tab |>
                collapse::fsubset(mean_tab[["by_vars"]] == by_var)
        }
        else{
            mean_temp <- mean_tab
        }

        freq_by <- freq_tab |>
            collapse::fsubset(freq_tab[["BY"]] == by_var)

        # Extract unique values
        if (anyNA(freq_by[["by_vars"]])){
            values <- c(collapse::funique(collapse::na_omit(freq_by[["by_vars"]])), NA)
        }
        else{
            values <- collapse::funique(freq_by[["by_vars"]])
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Loop through all unique values to generate tables per expression
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        for (value in values){
            # In case NAs are removed
            if (is.na(value) && na.rm){
                next
            }

            # Put additional header together with the information which by variable
            # and which value is currently filtered.
            header <- paste0("| ", by_var, " = ", value, " |")

            complete_header <- c("\n", strrep("-", collapse::vlengths(header)),
                                 header,
                                 strrep("-", collapse::vlengths(header)))

            # Filter table by current by variable and value
            if (!is.na(value)){
                if (means){
                    mean_temp <- mean_by |>
                        collapse::fsubset(mean_by[["TYPE"]] == value)
                }

                freq_temp <- freq_by |>
                    collapse::fsubset(freq_by[["by_vars"]] == "total" | freq_by[["by_vars"]] == value)
            }
            else{
                if (means){
                    mean_temp <- mean_by |>
                        collapse::fsubset(is.na(mean_by[["TYPE"]]))
                }

                freq_temp <- freq_by |>
                    collapse::fsubset(freq_by[["by_vars"]] == "total" | is.na(freq_by[["by_vars"]]))
            }

            # Generate mean table as normal but base is filtered data frame
            current_mean <- format_mean_text(mean_temp,
                                             variables,
                                             mean_columns,
                                             means)

            # Generate frequency tables as normal but base is filtered data frame
            current_freq <- format_freq_text(freq_temp,
                                             variables,
                                             formats,
                                             by,
                                             titles,
                                             footnotes,
                                             style)

            # Output formatted result
            complete_tabs <- c(complete_tabs, complete_header, current_mean, current_freq)
        }
    }

    # Output formatted result
    complete_tabs
}

###############################################################################
# Format grouped by tables for excel output
###############################################################################
#' Format Frequency Table Output with by Variables (Excel Based)
#'
#' @description
#' Format a frequency table for each variable which shows the counts, sums and
#' percentages per variable expression.
#'
#' @param mean_tab The data frame which contains the information for this summarising
#' table.
#' @param freq_tab The data frame which contains the information for this frequency
#' table.
#' @param variables The variables that should appear in the table.
#' @param mean_columns The column names from the data frames that should appear
#' in the table.
#' @param formats A list in which is specified which formats should be applied to
#' which variable.
#' @param by Separate the output by the expressions of the provided variables.
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#' @param style A list containing the styling elements.
#' @param output Determines whether to style the output or to just quickly paste
#' the data.
#' @param na.rm If TRUE removes all NA values from the tabulation.
#' @param print_miss FALSE by default. If TRUE outputs all possible categories of the
#' grouping variables based on the provided formats, even if there are no observations
#' for a combination.
#' @param wb An already created workbook to add more sheets to.
#' @param monitor_df Data frame which stores the monitoring values.
#' @param means FALSE by default. If TRUE prints a small summarising table which contains mean, sd,
#' min, max, total freq and missing values.
#'
#' @return
#' Returns a list containing a formatted Excel workbook as well as the monitoring
#' data frame.
#'
#' @noRd
format_by_excel <- function(mean_tab,
                            freq_tab,
                            variables,
                            mean_columns,
                            formats,
                            by,
                            titles,
                            footnotes,
                            style,
                            output,
                            na.rm,
                            print_miss,
                            wb,
                            monitor_df,
                            means){
    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_start("Excel prepare (by)", "Format by")
    #-------------------------------------------------------------------------#

    # Print message if multilabels are applied
    for (variable in variables){
        if (is_multilabel(formats, variable)){
            print_message("NOTE", c("The format for variable '[variable]' is a multilabel.",
									"In this case cumulative results aren't computed properly."),
									variable = variable, always_print = TRUE)
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Loop through all by variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    index <- 1

    for (by_var in by){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next(paste0("Excel prepare (", by_var, ")"), "Format by")
        #---------------------------------------------------------------------#

        # Select by variables one by one
        if (means){
            mean_by <- mean_tab |>
                collapse::fsubset(mean_tab[["by_vars"]] == by_var)
        }
        else{
            mean_temp <- mean_tab
        }

        freq_by <- freq_tab |>
            collapse::fsubset(freq_tab[["BY"]] == by_var)

        # Extract unique values
        if (anyNA(freq_by[["by_vars"]])){
            values <- c(collapse::funique(collapse::na_omit(freq_by[["by_vars"]])), NA)
        }
        else{
            values <- collapse::funique(freq_by[["by_vars"]])
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Loop through all unique values to generate tables per expression
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        for (value in values){
            # In case NAs are removed
            if (is.na(value) && na.rm){
                next
            }

            print_step("MINOR", "[by] = [value]", by = by, value = value)

            # Put additional by info together with the information which by variable
            # and which value is currently filtered.
            by_info <- paste0(by_var, " = ", value)

            # Filter table by current by variable and value
            if (!is.na(value)){
                if (means){
                    #---------------------------------------------------------#
                    monitor_df <- monitor_df |> monitor_next(paste0("Excel mean (", by_var, "_", value, ")"), "Format by")
                    #---------------------------------------------------------#

                    mean_temp <- mean_by |>
                        collapse::fsubset(mean_by[["TYPE"]] == value)
                }

                freq_temp <- freq_by |>
                    collapse::fsubset(freq_by[["by_vars"]] == "total" | freq_by[["by_vars"]] == value)
            }
            else{
                if (means){
                    mean_temp <- mean_by |>
                        collapse::fsubset(is.na(mean_by[["TYPE"]]))
                }

                freq_temp <- freq_by |>
                    collapse::fsubset(freq_by[["by_vars"]] == "total" | is.na(freq_by[["by_vars"]]))
            }

            # Generate mean table as normal but base is filtered data frame
            if (!is.null(mean_temp) && means && collapse::fnrow(mean_temp) > 0){
                wb_list <- format_mean_excel(mean_temp,
                                             mean_columns,
                                             style,
                                             output,
                                             means,
                                             wb,
                                             index,
                                             NULL)

                wb_list[[1]]$add_data(x         = by_info,
                                      start_col = style[["start_column"]] + 8,
                                      start_row = style[["start_row"]])
            }
            else{
                wb_list      <- list()
                wb_list[[1]] <- wb
            }

            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next(paste0("Excel freq (", by_var, "_", value, ")"), "Format by")
            #-----------------------------------------------------------------#

            # The print_miss option enables a shortcut in formatting the sheets after
            # the first one. Since it guarantees that all follow up sheets are printed
            # with the exact same table width and height, because all categories are
            # printed, only the first sheet must be formatted. All other sheets can
            # clone the entire style from the first sheet.
            if (index == 1 || !print_miss){
                # Generate frequency tables as normal but base is filtered data frame
                wb_list <- format_freq_excel(wb_list[[1]],
                                             freq_temp,
                                             variables,
                                             formats,
                                             by,
                                             titles,
                                             footnotes,
                                             style,
                                             output,
                                             by_info,
                                             index,
                                             NULL)
            }
            # Clone style for follow up iterations
            else{
                # Generate tables with no style but base is filtered data frame
                wb_list <- format_freq_excel(wb_list[[1]],
                                             freq_temp,
                                             variables,
                                             formats,
                                             by,
                                             titles,
                                             footnotes,
                                             style,
                                             "excel_nostyle",
                                             by_info,
                                             index,
                                             NULL)

                # Clone entire style from the corresponding sheet. If multiple variables
                # are given to produce separate frequency tables, the sheets are produced
                # one after the other. The following loop manages to copy the sheet style
                # from the corresponding base to the respective current sheet.
                for (sheet_index in seq_along(variables)){
                    wb_list[[1]]$clone_sheet_style(from = sheet_index,
                                                   to   = (index * length(variables)) - (length(variables) - sheet_index))
                }
            }

            index <- index + 1

            wb <- wb_list[[1]]
        }
    }

    #-----------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Order sheets", "Format by")
    #-----------------------------------------------------------------#

    # Generate a data frame consisting of the sheet names and their indexes
    sort_sheets_df <- data.table::data.table(names = wb$get_sheet_names(),
                                             order = openxlsx2::wb_get_order(wb))

    # Give every variable its index in provided order
    for (variable_number in seq_along(variables)){
        matches <- startsWith(sort_sheets_df[["names"]], variables[variable_number])

        sort_sheets_df[["var_order"]][matches] <- variable_number
    }

    # Sort by provided index to have the sheets per variable together instead of alternating
    sort_sheets_df |> data.table::setorderv("var_order")
    wb$set_order(sort_sheets_df[["order"]])

    monitor_df <- monitor_df |> monitor_end()

    # Return workbook
    list(wb, monitor_df)
}

#' Display Cross Table of Two Variables
#'
#' @description
#' [crosstabs()] produces a cross table of two variables. Statistics can be
#' weighted sums, unweighted frequencies or different percentages.
#'
#' @param data_frame A data frame in which are the variables to tabulate.
#' @param rows The variable that appears in the table rows.
#' @param columns The variable that appears in the table columns.
#' @param show_total TRUE by default. Whether to print row and column totals or not.
#' @param statistics The user requested statistics.Available functions:
#' - "sum"        -> Weighted and unweighted sum
#' - "freq"       -> Unweighted frequency
#' - "pct_row"    -> Weighted and unweighted row percentages
#' - "pct_column" -> Weighted and unweighted column percentages
#' - "pct_total"  -> Weighted and unweighted percentages compared to the grand total
#' @param formats A list in which is specified which formats should be applied to which variables.
#' @param by Compute tables stratified by the expressions of the provided variables.
#' @param weight Put in a weight variable to compute weighted results.
#' @param titles Specify one or more table titles.
#' @param footnotes Specify one or more table footnotes.
#' @param style A list of options can be passed to control the appearance of 'Excel' outputs.
#' Styles can be created with [excel_output_style()].
#' @param output The following output formats are available: console (default), text,
#' excel and excel_nostyle.
#' @param na.rm FALSE by default. If TRUE removes all NA values from the variables.
#' @param print_miss FALSE by default. If TRUE outputs all possible categories of the
#' grouping variables based on the provided formats, even if there are no observations
#' for a combination.
#' @param print TRUE by default. If TRUE prints the output, if FALSE doesn't print anything. Can be used
#' if one only wants to catch the output data frame.
#' @param monitor FALSE by default. If TRUE, outputs two charts to visualize the functions time consumption.
#'
#' @details
#' [crosstabs()] is based on the 'SAS' procedure Proc Freq, which provides
#' efficient and readable ways to perform cross tabulations.
#'
#' To create a cross table you only need to provide a variable for the rows and columns.
#' Nothing special about this. The real power comes into play, when you output your
#' tables as a fully styled 'Excel' workbook. Setting up a custom, reusable style is
#' as easy as setting up options like: provide a color for the table header, set the
#' font size for the row header, should borders be drawn for the table cells yes/no,
#' and so on.
#'
#' You can not only output sums and frequencies, but also different percentages, all
#' set up in separate, evenly designed tables. For just a quick overview, rather than
#' fully designed tables, you can also just output the tables in ASCII style format.
#'
#' @return
#' Returns a data tables containing the results for the cross table.
#'
#' @seealso
#' Creating a custom table style: [excel_output_style()], [modify_output_style()],
#' [number_format_style()], [modify_number_formats()].
#'
#' Global style options: [set_style_options()], [set_labels()].
#'
#' Other global options: [set_titles()], [set_footnotes()], [set_print()], [set_monitor()],
#' [set_na.rm()], [set_print()], [set_print_miss()], [set_output()].
#'
#' Creating formats: [discrete_format()] and [interval_format()].
#'
#' Functions that can handle formats and styles: [frequencies()], [any_table()].
#'
#' Additional functions that can handle styles: [export_with_style()]
#'
#' Additional functions that can handle formats: [summarise_plus()], [recode.()],
#' [recode_multi()], [transpose_plus()], [sort_plus()]
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Define titles and footnotes.
#' set_titles("This is title number 1",
#'            "This is title number 2",
#'            "This is title number 3")
#'
#' set_footnotes("This is footnote number 1",
#'               "This is footnote number 2",
#'               "This is footnote number 3")
#'
#' # Output cross tables
#' my_data |> crosstabs(state, sex)
#' my_data |> crosstabs(state, sex,
#'                      weight = "weight")
#'
#' # Also works with characters
#' my_data |> crosstabs("state", "sex")
#' my_data |> crosstabs("state", "sex",
#'                      weight = "weight")
#'
#' # Applying formats
#' age. <- discrete_format(
#'     "Total"          = 0:100,
#'     "under 18"       = 0:17,
#'     "18 to under 25" = 18:24,
#'     "25 to under 55" = 25:54,
#'     "55 to under 65" = 55:64,
#'     "65 and older"   = 65:100)
#'
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' my_data |> crosstabs(age, sex,
#'                      formats = list(age = age., sex = sex.))
#'
#' # Split cross table by expressions of another variable
#' my_data |> crosstabs(state, sex, by = education)
#'
#' # Compute different stats
#' my_data |> crosstabs(state, sex,
#'                      statistics = c("sum", "freq", "pct_row", "pct_column", "pct_total"))
#'
#' # Get a list with two data tables for further usage
#' result_list <- my_data |> crosstabs(age, sex,
#'                                     formats = list(age = age., sex = sex.))
#'
#' # Output in text file
#' my_data |> crosstabs(state, sex, output = "text")
#'
#' # Output to Excel
#' my_data |> crosstabs(state, sex, output = "excel")
#'
#' # If you want to add hyperlinksto titles and footnotes you can do so by
#' # adding "link:" followed by the hyperlink to the main text. Linking to another
#' # cell works with "cell:". To link to a file use "file:" an pass the full file
#' # path afterwards.
#' set_titles("This is title number 1",
#'            "This is title number 2 link: https://cran.r-project.org/",
#'            "This is title number 3 cell: W22",
#'            "This is title number 4 file: C:/MyFolder/MyFile.txt")
#'
#' set_footnotes("This is footnote number 1",
#'               "This is footnote number 2 file: C:/MyFolder/MyFile.txt",
#'               "This is footnote number 3 cell: W22",
#'               "This is footnote number 4 link: https://cran.r-project.org/")
#'
#' # Individual styling can also be passed directly
#' my_style <- excel_output_style(header_back_color = "0077B6",
#'                                font              = "Times New Roman")
#'
#' my_data |> crosstabs(age, sex, output = "excel", style = my_style,
#'                      formats = list(age = age., sex = sex.))
#'
#' # To save a table as xlsx file you have to set the path and filename in the
#' # style element
#' # Example files paths
#' table_file <- tempfile(fileext = ".xlsx")
#'
#' # Note: Normally you would directly input the path ("C:/MyPath/") and name ("MyFile.xlsx").
#' #       With the set_style_options you can also set a table style globally.
#' set_style_options(save_path  = dirname(table_file),
#'                   file       = basename(table_file),
#'                   sheet_name = "MyTable")
#'
#' my_data |> crosstabs(state, sex, output = "excel")
#'
#' # Manual cleanup for example
#' unlink(table_file)
#'
#' # Global options are permanently active until the current R session is closed.
#' # There are also functions to reset the values manually.
#' reset_style_options()
#' reset_qol_options()
#' close_file()
#'
#' @export
crosstabs <- function(data_frame,
                      rows,
                      columns,
                      show_total = TRUE,
                      statistics = "sum",
                      formats    = c(),
                      by         = c(),
                      weight     = NULL,
                      titles     = .qol_options[["titles"]],
                      footnotes  = .qol_options[["footnotes"]],
                      style      = .qol_options[["excel_style"]],
                      output     = .qol_options[["output"]],
                      na.rm      = .qol_options[["na.rm"]],
                      print_miss = .qol_options[["print_miss"]],
                      print      = .qol_options[["print"]],
                      monitor    = .qol_options[["monitor"]]){

    # Measure the time
    print_start_message()
	print_step("GREY", "Error handling")

    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Error handling", "Preparation")
    #-------------------------------------------------------------------------#

    ###########################################################################
    # Early evaluations
    ###########################################################################

    # First convert data frame to data table
    if (!data.table::is.data.table(data_frame)){
        data_frame <- data.table::as.data.table(data_frame)
    }

    # Evaluate formats early
    if (!is_list_of_dfs(formats)){
        formats_list <- as.list(substitute(formats))[-1]
        formats      <- evaluate_formats(formats_list)
    }

    # Remove empty formats and throw a warning. This can happen if there is e.g.
    # a typo in the format.
    for (variable in names(formats)){
        if (is.null(formats[[variable]])){
            formats[[variable]] <- NULL
            print_message("WARNING", "Format for variable '[variable]' does not exist and can't be applied.", variable = variable)
        }
    }

    # Apply macros to save path and file
    if (!is.null(style[["save_path"]])){
        style[["save_path"]] <- macro(style[["save_path"]])
    }

    if (!is.null(style[["file"]])){
        style[["file"]] <- macro(style[["file"]])
    }

    ###########################################################################
    # Error handling
    ###########################################################################

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Row variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    rows <- get_origin_as_char(rows, substitute(rows))

    # Make sure that the variables provided are part of the data frame.
    rows <- data_frame |> part_of_df(rows)

    if (length(rows) <= 1){
        if (length(rows) == 0 || rows == ""){
            print_message("ERROR", "No valid <rows> variable provided. Crosstabs will be aborted.")
            return(invisible(NULL))
        }
    }

    if (length(rows) > 1){
        rows <- rows[[1]]
        print_message("WARNING", "Only one variable for <rows> allowed. First variable will be used.")
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Column variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    columns <- get_origin_as_char(columns, substitute(columns))

    # Make sure that the variables provided are part of the data frame.
    columns <- data_frame |> part_of_df(columns)

    # Make sure there is no column variable that is also a row variable.
    columns <- resolve_intersection(columns, rows, check_only = TRUE)

    if (is.list(columns)){
        print_message("ERROR", c("The provided <columns> variable '[columns]' is also part of",
								 "the <rows> variables. Crosstabs will be aborted."), columns = columns[[1]])
        return(invisible(NULL))
    }

    if (length(columns) <= 1){
        if (length(columns) == 0 || columns == ""){
            print_message("ERROR", "No valid <columns> variable provided. Crosstabs will be aborted.")
            return(invisible(NULL))
        }
    }

    if (length(columns) > 1){
        columns <- columns[[2]]
        print_message("WARNING", "Only one variable for <columns> allowed. First variable will be used.")
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # By variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    by <- get_origin_as_char(by, substitute(by))

    # Make sure that the variables provided are part of the data frame.
    by <- data_frame |> part_of_df(by)

    # Make sure there is no class variable that is also a value variable.
    variables <- c(rows, columns)
    by        <- resolve_intersection(by, variables)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Weight
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    weight     <- get_origin_as_char(weight, substitute(weight))
    data_frame <- data_frame |> check_weight(weight)
    weight_var <- ".temp_weight"

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Output
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Check for invalid output option
    if (!tolower(output) %in% c("console", "text", "excel", "excel_nostyle")){
        print_message("WARNING", "<Output> format '[output]' not available. Using 'console' instead.", output = output)

        output <- "console"
    }
    else{
        output <- tolower(output)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Statistics
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    statistics <- get_origin_as_char(statistics, substitute(statistics))
    statistics <- tolower(statistics)

    # Get the intersection of the requested statistics to make sure
    # only valid actions are passed down
    statistics <- statistics[collapse::funique(statistics) %in%
                             c("sum", "freq", "pct_row", "pct_column", "pct_total")]

    # If no valid statistics selected, default to sum
    if (length(statistics) == 0){
        statistics <- "sum"
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Resolve macros
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    titles    <- apply_macro(titles)
    footnotes <- apply_macro(footnotes)

    ###########################################################################
    # Cross tabulation starts
    ###########################################################################

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Summary", "Summary")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Computing stats.")

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Summarise data according to provided variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Summarise results for a crosstabs output
    data_frame[["var"]] <- 1

    # Put together vector of grouping variables
    group_vars <- c(by, rows, columns)

    # In case no by variables are specified
    if (length(by) == 0){
        cross_tab <- suppressMessages(data_frame |>
             summarise_plus(class      = group_vars,
                            values     = "var",
                            statistics = c("sum", "pct_group", "pct_total", "freq"),
                            formats    = formats,
                            weight     = weight_var,
                            nesting    = "deepest",
                            notes      = FALSE,
                            na.rm      = na.rm,
                            print_miss = print_miss)) |>
            drop_type_vars() |>
            collapse::frename(var_pct_group = var_pct_row)

        # Get the expressions of the column variable to use them later as column names
        column_names <- collapse::funique(as.character(cross_tab[[columns]]))

        # Add unique ids to column variables so that the pivoted variable names receive
        # running numbers instead of the values or labels.
        run_nr_df        <- data.table::data.table(column_names)
        names(run_nr_df) <- columns
        run_nr_df[["run_nr"]] <- suppressMessages(run_nr_df |> running_number())

        cross_tab <- cross_tab |>
            collapse::join(run_nr_df,
                           on      = columns,
                           how     = "left",
                           verbose = FALSE,
						   overid  = 2)

        # Pivot to wider format, which basically is the final format to print the data
        cross_tab <- cross_tab |>
            collapse::pivot(id     = rows,
                            names  = "run_nr",
                            values = c("var_sum", "var_pct_row", "var_pct_total", "var_freq"),
                            how    = "wider")
    }
    # In case by variables are specified
    else{
        combinations <- as.vector(outer(by, paste0(rows, "+", columns), paste, sep = "+"))

        cross_tab <- suppressMessages(data_frame |>
            summarise_plus(class      = group_vars,
                           values     = "var",
                           statistics = c("sum", "pct_group", "pct_total", "freq"),
                           formats    = formats,
                           weight     = weight_var,
                           nesting    = "all",
                           types      = combinations,
                           notes      = FALSE,
                           na.rm      = na.rm,
                           print_miss = print_miss)) |>
            collapse::frename(var_pct_group = var_pct_row) |>
            fuse_variables("by_vars", by)

        cross_tab[["BY"]] <- sub("\\+.*", "", cross_tab[["TYPE"]])

        # Get the expressions of the column variable to use them later as column names
        column_names <- unique(as.character(cross_tab[[columns]]))

        # Add unique ids to column variables so that the pivoted variable names receive
        # running numbers instead of the values or labels.
        run_nr_df        <- data.table::data.table(column_names)
        names(run_nr_df) <- columns
        run_nr_df[["run_nr"]] <- suppressMessages(run_nr_df |> running_number())

        cross_tab <- cross_tab |>
            collapse::join(run_nr_df,
                           on      = columns,
                           how     = "left",
                           verbose = FALSE,
						   overid  = 2)

        # Pivot to wider format, which basically is the final format to print the data
        cross_tab <- cross_tab |>
            collapse::pivot(id     = c("BY", "by_vars", rows),
                            names  = "run_nr",
                            values = c("var_sum", "var_pct_row", "var_pct_total", "var_freq"),
                            how    = "wider")
    }

    if (is.null(cross_tab)){
        print_message("ERROR", "Crosstab could not be computed.")
        return(invisible(NULL))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Prepare table format for output
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    print_step("MAJOR", "Formatting tables.")

    if (output %in% c("console", "text")){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Format tables", "Format tables")
        #---------------------------------------------------------------------#

        # In case no by variables are provided
        if (length(by) == 0){
            complete_table  <- format_cross_text(cross_tab, rows, columns, column_names,
                                                 statistics, formats, by, titles, footnotes, show_total)
        }
        # In case there are  by variables are provided
        else{
            complete_table <- format_cross_by_text(cross_tab, rows, columns, column_names,
                                                   statistics, formats, by, titles, footnotes, na.rm, show_total)
        }
    }
    else if (output == "excel" || output == "excel_nostyle"){
        wb <- openxlsx2::wb_workbook() |>
            prepare_styles(list("title" = titles, "footnote" = footnotes), style)

        monitor_df <- monitor_df |> monitor_end()

        # In case no by variables are provided
        if (length(by) == 0){
            wb_list <- format_cross_excel(wb, cross_tab, rows, columns, column_names,
                                          statistics, formats, by, titles, footnotes, style, output, show_total,
                                          monitor_df = monitor_df)

            wb         <- wb_list[[1]]
            monitor_df <- wb_list[[2]]
        }
        # In case there are  by variables are provided
        else{
            wb_list <- format_cross_by_excel(cross_tab, rows, columns, column_names,
                                             statistics, formats, by, titles, footnotes,
                                             style, output, show_total, na.rm, print_miss, wb, monitor_df)

            wb         <- wb_list[[1]]
            monitor_df <- wb_list[[2]]
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Output formatted table into different formats
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (print){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Output tables", "Output tables")
        #---------------------------------------------------------------------#

        if (output %in% c("console")){
            print_closing()

            cat(paste(complete_table, collapse = "\n"), "\n\n")
        }
        # Open in text editor
        else if (output == "text"){
            temp_file <- tempfile(fileext = ".txt")
            writeLines(complete_table, temp_file)

            if (interactive()){
                file.show(temp_file)
            }

            print_closing()
        }
        else if (output == "excel" || output == "excel_nostyle"){
            # Check if only save path or file name is specified. If only one is specified
            # print a note. Otherwise the file would not be saved and opened without a
            # hint to why the file wasn't saved.
            if (is.null(style[["save_path"]]) + is.null(style[["file"]]) == 1){
                if (is.null(style[["save_path"]])){
                    print_message("NOTE", c("No save path specified. Both save path and file name with extension",
                                            "need to be specified in the global options or style parameter for",
                                            "the file to be saved. File won't be saved."))
                }
                else{
                    print_message("NOTE", c("No file name specified. Both save path and file name with extension",
                                            "need to be specified in the global options or style parameter for",
                                            "the file to be saved. File won't be saved."))
                }
            }

            # If no save path or file provided just open workbook
            if (is.null(style[["save_path"]]) || is.null(style[["file"]])){
                if (interactive()){
                    wb$open()
                }
            }
            else{
                # If save path doesn't exist, just open workbook
                if (!file.exists(style[["save_path"]])){
                    print_message("WARNING", "Path does not exist: [style]", style = style[["save_path"]])

                    if (interactive()){
                        wb$open()
                    }
                }
                # Save file
                else{
                    extension <- tolower(tools::file_ext(style[["file"]]))

                    # CSV export is based on the actual data frame
                    if (extension == "csv"){
                        cross_tab |> data.table::fwrite(file = paste0(style[["save_path"]], "/", style[["file"]]),
                                                        sep  = ";",
                                                        dec  = ",")
                    }
                    # While XLSX export is based on the workbook
                    else if (extension == "xlsx"){
                        wb$save(file = paste0(style[["save_path"]], "/", style[["file"]]), overwrite = TRUE)
                    }
                    # If there is no extension or a wrong one, auto correct it
                    else{
                        wb$save(file = paste0(style[["save_path"]], "/", basename(style[["file"]]), ".xlsx"), overwrite = TRUE)
                    }
                }
            }

            print_closing()
        }
    }
    else{
        print_closing()
    }

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_end()
    monitor_df |> monitor_plot(draw_plot = monitor)
    #-------------------------------------------------------------------------#

    invisible(cross_tab)
}

###############################################################################
# Format cross table for console output
###############################################################################
#' Format Cross Table Output (Text Based)
#'
#' @description
#' Format a cross table with the provided row and column variables. Statistics
#' can be weighted sums, unweighted frequencies or different percentages.
#'
#' @param cross_tab The data frame which contains the information for this cross
#' table.
#' @param rows The variable that appears in the table rows.
#' @param columns The variable that appears in the table columns.
#' @param column_names The real column names that appear in the table.
#' @param statistics The user requested statistics.
#' @param formats A list in which is specified which formats should be applied to
#' which variable.
#' @param by Separate the frequency output by the expressions of the provided variables.
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#'
#' @return
#' Returns a character vector with all formatted rows for the tables.
#'
#' @noRd
format_cross_text <- function(cross_tab,
                              rows,
                              columns,
                              column_names,
                              statistics,
                              formats,
                              by,
                              titles,
                              footnotes,
                              show_total){
    complete_tabs <- c()

    # Set equal number of maximum column width
    sum_tab          <- setup_print_table(cross_tab, rows, "sum")
    longest_value    <- collapse::vlengths(format(sum_tab[collapse::fnrow(sum_tab), "total"], scientific = FALSE))
    max_column_width <- max(8, longest_value + 7)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Compute individual cross tables for each operation
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    for (stat in statistics){
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Prepare header
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Setup cross table for formatting
        var_tab <- setup_print_table(cross_tab, rows, stat)

        # Get the maximum width of the provided variable names to determine the width
        # of the first column.
        first_column_width <- collapse::vlengths(paste0(rows, " / ", columns))

        first_column_width <- max(first_column_width,
                                  collapse::vlengths(collapse::na_omit(as.character(var_tab[[rows]]))))

        # In case of multilabels don't add total column
        if(is_multilabel(formats, columns) || !show_total){
            multi_header   <- wrap_header(column_names, max_column_width)

            var_tab        <- var_tab |> dropp("total")
            formatted_cols <- vector("list", length(column_names))
        }
        # Use the max column width to evenly spread the columns.
        # Wrap the column headers according to this width.
        else{
            multi_header   <- wrap_header(c(column_names, "total"), max_column_width)

            formatted_cols <- vector("list", length(c(column_names, "total")))
        }

        # Format the first column and connect it with the rest of the header
        first_column <- c(paste0(format(paste0(rows, " / ", columns),
                                        width   = first_column_width,
                                        justify = "left"), " |"),
                          rep(paste0(format("",
                                            width   = first_column_width,
                                            justify = "left"), " |"),
                              collapse::fnrow(multi_header) - 1))
        multi_header <- cbind(first_column, multi_header)

        # Visually separate total column if there is one
        total_col <- FALSE

        if (trimws(multi_header[1, collapse::fncol(multi_header)]) == "total"){
            total_col <- TRUE

            multi_header[1, collapse::fncol(multi_header)] <-
                paste0(" | ", multi_header[1, collapse::fncol(multi_header)])
        }

        # Convert header matrix into character vector with single lines
        complete_header <- apply(multi_header, 1, paste, collapse = " ")
        complete_header <- c(complete_header, strrep("-", collapse::vlengths(complete_header[1])))

        # If row multilabel, remove automatically computed total observation
        if (is_multilabel(formats, rows) || !show_total){
            var_tab <- var_tab |>
                collapse::fsubset(var_tab[[rows]] != "total")

            if (stat == "pct_row"){
                if (length(by) == 0){
                    print_message("NOTE", c("The format for variable '[rows]' is a multilabel.",
											"In this case row percentages aren't computed properly."), rows = rows,
											always_print = TRUE)
                }
            }
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Format table column by column. Basically concatenate first column text and
        # corresponding values together while keeping the individual maximum column
        # width in mind.
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        first_column <- paste0(format(var_tab[[rows]],
                                      width   = first_column_width,
                                      justify = "left"), " | ")

        # Loop through columns and format each number in a readable way
        for (column in seq_len(length(formatted_cols))){
            # Percentages and Unweighted freq
            if (stat %in% c("pct_row", "pct_column", "pct_total")){
                formatted_cols[[column]] <- format_number(var_tab[column + 1],
                                                          width = max_column_width,
                                                          stat  = names(var_tab[column + 1]))
            }
            # Unweighted frequencies
            else if (stat %in% c("freq") ||
                     identical(as.numeric(cross_tab[["var_sum_1"]]),
                               as.numeric(cross_tab[["var_freq_1"]]))){
                formatted_cols[[column]] <- format_number(var_tab[column + 1],
                                                          width    = max_column_width,
                                                          stat     = names(var_tab[column + 1]),
                                                          decimals = 0)
            }
            # Weighted sums
            else if (stat %in% c("sum")){
                formatted_cols[[column]] <- format_number(var_tab[column + 1],
                                                          width    = max_column_width,
                                                          stat     = names(var_tab[column + 1]),
                                                          decimals = 3)
            }

            # Visually separate total column if there is one
            if (column == length(formatted_cols) && total_col){
                formatted_cols[[column]] <- paste0(" | ", formatted_cols[[column]])
            }
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Put everything together
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Convert list to matrix
        formatted_matrix <- do.call(cbind, formatted_cols)

        # Convert matrix to single formatted rows
        all_rows <- paste0(first_column,
                           apply(formatted_matrix, 1, paste, collapse = " "))

        # If no multilabel formats are applied separate total row at the bottom from the rest
        if (!is_multilabel(formats, rows) && show_total){
            rows_temp <- length(all_rows)

            all_rows <- c(all_rows[1:(rows_temp - 1)],
                          strrep("-", collapse::vlengths(complete_header[1])),
                          all_rows[rows_temp])
        }

        # Insert titles and footnotes
        complete_table <- c(complete_header, all_rows)

        if (length(titles) > 0){
            complete_table <- c(titles, "", complete_table)
        }
        if (length(footnotes) > 0){
            complete_table <- c(complete_table, "____________", footnotes)
        }

        # Output formatted result
        complete_tabs <- c(complete_tabs, "\n", complete_table, "\n")
    }

    # Output formatted result
    complete_tabs
}


#' Wrap Text Based Table Header
#'
#' @description
#' The width for each column gets a maximum value. After that maximum value the
#' header wraps to the next line to ensure the table doesn't gets too wide.
#'
#' @param cross_tab The data frame which contains the information for this cross
#' table.
#' @param rows The variables that appear in the table rows.
#' @param stat The currently computed stat.
#'
#' @return
#' Returns a data table trimmed down to the necessary variables and observations.
#'
#' @noRd
wrap_header <- function(column_names, width = 10){
    # Split each column name into single letters
    letter_list <- strsplit(column_names, "")

    wrapped <- lapply(letter_list, function(word){
        # How many chunks are needed for this label
        n_chunks <- ceiling(length(word) / (width - 1))

        # Extract substrings of max 'width' characters
        parts <- character(n_chunks)

        for (i in seq_len(n_chunks)){
            start    <- (i - 1) * (width - 1) + 1
            end      <- min(i * (width - 1), length(word))
            parts[i] <- paste(word[start:end], collapse = "")
        }

        format(parts, width = width, justify = "right")
    })

    # Find the maximum number of lines across all labels
    max_lines <- max(sapply(wrapped, length))

    # Pad shorter labels with empty strings so they all align
    wrapped_padded <- lapply(wrapped, function(parts){
        c(parts, rep(format("", width = width, justify = "right"), max_lines - length(parts)))
    })

    # Return as a matrix: rows = lines, cols = labels
    do.call(cbind, wrapped_padded)
}


#' Trim Cross Table to the Needed Parts
#'
#' @description
#' Cross tables are computed by stats. Here the stats get reduced the the current
#' stat in the loop. Column percentages get newly computed because they aren't
#' computed beforehand.
#'
#' @param cross_tab The data frame which contains the information for this cross
#' table.
#' @param rows The variables that appear in the table rows.
#' @param stat The currently computed stat.
#'
#' @return
#' Returns a data table trimmed down to the necessary variables and observations.
#'
#' @noRd
setup_print_table <- function(cross_tab, rows, stat){
    # Compute column percentages
    if (stat == "pct_column"){
        var_tab <- cross_tab |>
            collapse::get_vars(c(rows, "sum"), regex = TRUE)

        # Compute column percentages and rename variables
        for (column in names(var_tab[-1])){
            new_var <- sub("sum", "pct_column", column)
            var_tab[[new_var]] <- var_tab[[column]] * 100 / collapse::fsum(var_tab[[column]])
        }

        # Get only the needed variables for this iteration
        var_tab <- var_tab |>
            collapse::get_vars(c(rows, stat), regex = TRUE)
    }
    else{
        # Get only the needed variables for this iteration
        var_tab <- cross_tab |>
            collapse::get_vars(c(rows, stat), regex = TRUE)
    }

    # Compute row and column sums
    var_tab[["total"]]    <- rowSums(var_tab[, -1], na.rm = TRUE)

    column_totals         <- var_tab[1, ]
    column_totals[[rows]] <- "total"
    column_totals[-1]     <- colSums(var_tab[, -1], na.rm = TRUE)
    var_tab               <- rbind(var_tab, column_totals)

    # Set unneeded totals to NA
    if (stat == "pct_column"){
        var_tab[["total"]] <- NA
    }
    else if (stat == "pct_row"){
        var_tab[collapse::fnrow(var_tab), -1] <- NA
    }

    var_tab
}

###############################################################################
# Format cross table for excel output
###############################################################################
#' Format Cross Table Output (Excel Based)
#'
#' @description
#' Format a cross table with the provided row and column variables. Statistics
#' can be weighted sums, unweighted frequencies or different percentages.
#'
#' @param wb An already created workbook to add more sheets to.
#' @param cross_tab The data frame which contains the information for this cross
#' table.
#' @param rows The variable that appears in the table rows.
#' @param columns The variable that appears in the table columns.
#' @param column_names The real column names that appear in the table.
#' @param statistics The user requested statistics.
#' @param formats A list in which is specified which formats should be applied to
#' which variable.
#' @param by Separate the cross table output by the expressions of the provided variables.
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#' @param style A list containing the styling elements.
#' @param output Determines whether to style the output or to just quickly paste
#' the data.
#' @param by_info Text which contains the information which variable with which
#' expression is computed at the moment.Used for computation with by variables.
#' @param index Index of the current variable expression. Used for computation with
#' by variables.
#' @param monitor_df Data frame which stores the monitoring values.
#'
#' @return
#' Returns a list containing a formatted Excel workbook as well as the monitoring
#' data frame.
#'
#' @noRd
format_cross_excel <- function(wb,
                               cross_tab,
                               rows,
                               columns,
                               column_names,
                               statistics,
                               formats,
                               by,
                               titles,
                               footnotes,
                               style,
                               output,
                               show_total,
                               by_info = NULL,
                               index   = NULL,
                               monitor_df){
    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_start("Excel prepare", "Format")
    #-------------------------------------------------------------------------#

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Compute individual cross tables for each operation
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    for (stat in statistics){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel prepare", "Format")
        #---------------------------------------------------------------------#

        # Setup cross table for formatting
        var_tab <- setup_print_table(cross_tab, rows, stat)

        # Round values according to style options
        if (stat %in% c("sum", "freq")){
            var_tab <- var_tab |>
                collapse::fmutate(across(is.numeric, round,
                                         digits = style[["number_formats"]][["sum_decimals"]]))
        }
        else{
            var_tab <- var_tab |>
                collapse::fmutate(across(is.numeric, round,
                                         digits = style[["number_formats"]][["pct_decimals"]]))
        }

        # If row multilabel, remove automatically computed total observation
        if (is_multilabel(formats, rows) || !show_total){
            var_tab <- var_tab |>
                collapse::fsubset(var_tab[[rows]] != "total")

            if (stat == "pct_row"){
                if (length(by) == 0){
                    print_message("NOTE", c("The format for variable '[rows]' is a multilabel.",
											"In this case row percentages aren't computed properly."), rows = rows,
											always_print = TRUE)
                }
            }
        }

        column_names[is.na(column_names)] <- style[["na_symbol"]]

        # If column multilabel, remove automatically computed total column.
        # Set the real column names according to number of columns.
        if (is_multilabel(formats, columns) || !show_total){
            var_tab <- var_tab |> dropp("total")

            names(var_tab) <- c(paste0(rows, " / ", columns), column_names)
        }
        else{
            names(var_tab) <- c(paste0(rows, " / ", columns), column_names, "total")
        }

        # Get table ranges
        cross_ranges <- get_table_ranges(var_tab, titles, footnotes, style)

        # If function is called with by variables the sheet names have to be differentiated
        # and by info has to be written besides the table.
        if (!is.null(by_info)){
            by_info_col <- style[["start_column"]] + length(var_tab) + 1

            wb$add_worksheet(paste0(stat, index),
                             grid_lines = style[["grid_lines"]])
            wb$add_data(x         = by_info,
                        start_col = by_info_col,
                        start_row = style[["start_row"]])
        }
        else{
            wb$add_worksheet(stat, grid_lines = style[["grid_lines"]])
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Apply style
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel titles/footnotes", "Format")
        #---------------------------------------------------------------------#
        # Format titles and footnotes if there are any
        wb <- wb |>
            format_titles_foot_excel(titles, footnotes, cross_ranges, style, output)

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel data", "Format")
        #---------------------------------------------------------------------#

        # Add table data and format according to style options
        wb$add_data(x          = var_tab,
                    start_col  = style[["start_column"]],
                    start_row  = cross_ranges[["header.row"]],
                    na.strings = style[["na_symbol"]])

        # Only do the formatting when user specified it. With the excel_nostyle
        # option this whole part gets omitted to get a very quick unformatted
        # excel output.
        if (output == "excel"){
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Excel cell styles", "Format")
            #-----------------------------------------------------------------#
            wb <- wb |> handle_cell_styles(cross_ranges, style)

            if (stat == "sum" || stat== "freq"){
                wb$add_cell_style(dims                = cross_ranges[["table_range"]],
                                  apply_number_format = TRUE,
                                  num_fmt_id          = wb$styles_mgr$get_numfmt_id(paste0(stat, "_numfmt")))
            }
            else{
                wb$add_cell_style(dims                = cross_ranges[["table_range"]],
                                  apply_number_format = TRUE,
                                  num_fmt_id          = wb$styles_mgr$get_numfmt_id("pct_numfmt"))
            }

            # Draw inner table cells as heat map with conditional formatting
            if (style[["as_heatmap"]]){
                #-----------------------------------------------------------------#
                monitor_df <- monitor_df |> monitor_next("Excel format heatmap", "Format")
                #-----------------------------------------------------------------#

                wb$add_conditional_formatting(dims  = cross_ranges[["table_range"]],
                                              style = c(style[["heatmap_low_color"]],
                                                        style[["heatmap_middle_color"]],
                                                        style[["heatmap_high_color"]]),
                                              type  = "colorScale")
            }

            # Adjust table dimensions
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Excel widths/heights", "Format")
            #-----------------------------------------------------------------#
            wb <- wb |> handle_col_row_dimensions(cross_ranges,
                                                  collapse::fncol(var_tab) + (style[["start_column"]] - 1),
                                                  collapse::fnrow(var_tab) + (style[["start_row"]] - 1),
                                                  style)

            wb <- wb |> handle_auto_dimensions(cross_ranges,
                                               style)

            wb <- wb |> handle_header_table_dim(cross_ranges,
                                                style)

            wb$add_ignore_error(dims = cross_ranges[["header_range"]],  number_stored_as_text = TRUE)
            wb$add_ignore_error(dims = cross_ranges[["cat_col_range"]], number_stored_as_text = TRUE)

            wb$add_named_region(dims = cross_ranges[["whole_tab_range"]], name = "table", local_sheet = TRUE)
            wb$add_named_region(dims = cross_ranges[["table_range"]],     name = "data",  local_sheet = TRUE)
        }
    }

    monitor_df <- monitor_df |> monitor_end()

    # Return workbook
    list(wb, monitor_df)
}

###############################################################################
# Format grouped by tables for console output
###############################################################################
#' Format Cross Table Output with by Variables (Text Based)
#'
#' @description
#' Format a cross table with the provided row and column variables. Statistics
#' can be weighted sums, unweighted frequencies or different percentages.
#'
#' @param cross_tab The data frame which contains the information for this cross
#' table.
#' @param rows The variable that appears in the table rows.
#' @param columns The variable that appears in the table columns.
#' @param column_names The real column names that appear in the table.
#' @param statistics The user requested statistics.
#' @param formats A list in which is specified which formats should be applied to
#' which variable.
#' @param by Separate the output by the expressions of the provided variables.
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#' @param na.rm If TRUE removes all NA values from the tabulation.
#'
#' @return
#' Returns a character vector with all formatted rows for the tables.
#'
#' @noRd
format_cross_by_text <- function(cross_tab,
                                 rows,
                                 columns,
                                 column_names,
                                 statistics,
                                 formats,
                                 by,
                                 titles,
                                 footnotes,
                                 na.rm,
                                  show_total){
    # Print message if multilabels are applied
    if (is_multilabel(formats, rows)){
        if ("pct_row" %in% statistics){
            print_message("NOTE", c("The format for variable '[rows]' is a multilabel.",
									"In this case row percentages aren't computed properly."), rows = rows,
									always_print = TRUE)
        }
    }

    complete_tabs <- c()

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Loop through all by variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    for (by_var in by){
        # Select by variables one by one
        cross_by <- cross_tab |>
            collapse::fsubset(cross_tab[["BY"]] == by_var)

        # Extract unique values
        if (anyNA(cross_by[["by_vars"]])){
            values <- c(collapse::funique(collapse::na_omit(cross_by[["by_vars"]])), NA)
        }
        else{
            values <- collapse::funique(cross_by[["by_vars"]])
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Loop through all unique values to generate tables per expression
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        for (value in values){
            # In case NAs are removed
            if (is.na(value) && na.rm){
                next
            }

            # Put additional header together with the information which by variable
            # and which value is currently filtered.
            header <- paste0("| ", by_var, " = ", value, " |")

            complete_header <- c("\n", strrep("-", collapse::vlengths(header)),
                                 header,
                                 strrep("-", collapse::vlengths(header)))

            # Filter table by current by variable and value
            if (!is.na(value)){
                cross_temp <- cross_by |>
                    collapse::fsubset(cross_by[["by_vars"]] == value)
            }
            else{
                cross_temp <- cross_by |>
                    collapse::fsubset(is.na(cross_by[["by_vars"]]))
            }

            # Generate frequency tables as normal but base is filtered data frame
            current_cross <- format_cross_text(cross_temp,
                                               rows,
                                               columns,
                                               column_names,
                                               statistics,
                                               formats,
                                               by,
                                               titles,
                                               footnotes,
                                               show_total)

            # Output formatted result
            complete_tabs <- c(complete_tabs, complete_header, current_cross)
        }
    }

    # Output formatted result
    complete_tabs
}

###############################################################################
# Format grouped by tables for excel output
###############################################################################
#' Format Cross Table Output with by Variables (Excel Based)
#'
#' @description
#' Format a cross table with the provided row and column variables. Statistics
#' can be weighted sums, unweighted frequencies or different percentages.
#'
#' @param cross_tab The data frame which contains the information for this cross
#' table.
#' @param rows The variable that appears in the table rows.
#' @param columns The variable that appears in the table columns.
#' @param column_names The real column names that appear in the table.
#' @param statistics The user requested statistics.
#' @param formats A list in which is specified which formats should be applied to
#' which variable.
#' @param by Separate the output by the expressions of the provided variables.
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#' @param style A list containing the styling elements.
#' @param output Determines whether to style the output or to just quickly paste
#' the data.
#' @param na.rm If TRUE removes all NA values from the tabulation.
#' @param print_miss FALSE by default. If TRUE outputs all possible categories of the
#' grouping variables based on the provided formats, even if there are no observations
#' for a combination.
#' @param wb An already created workbook to add more sheets to.
#' @param monitor_df Data frame which stores the monitoring values.
#'
#' @return
#' Returns a list containing a formatted Excel workbook as well as the monitoring
#' data frame.
#'
#' @noRd
format_cross_by_excel <- function(cross_tab,
                                  rows,
                                  columns,
                                  column_names,
                                  statistics,
                                  formats,
                                  by,
                                  titles,
                                  footnotes,
                                  style,
                                  output,
                                  show_total,
                                  na.rm,
                                  print_miss,
                                  wb,
                                  monitor_df){
    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_start("Excel prepare (by)", "Format by")
    #-------------------------------------------------------------------------#

    # Print message if multilabels are applied
    if (is_multilabel(formats, rows)){
        if ("pct_row" %in% statistics){
            print_message("NOTE", c("The format for variable '[rows]' is a multilabel.",
									"In this case row percentages aren't computed properly."), rows = rows,
									always_print = TRUE)
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Loop through all by variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    index <- 1

    monitor_df <- monitor_df |> monitor_end()

    for (by_var in by){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_start(paste0("Excel prepare (", by_var, ")"), "Format by")
        #---------------------------------------------------------------------#

        # Select by variables one by one
        cross_by <- cross_tab |>
            collapse::fsubset(cross_tab[["BY"]] == by_var)

        # Extract unique values
        if (anyNA(cross_by[["by_vars"]])){
            values <- c(collapse::funique(collapse::na_omit(cross_by[["by_vars"]])), NA)
        }
        else{
            values <- collapse::funique(cross_by[["by_vars"]])
        }

        monitor_df <- monitor_df |> monitor_end()

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Loop through all unique values to generate tables per expression
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        for (value in values){
            # In case NAs are removed
            if (is.na(value) && na.rm){
                next
            }

            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_start(paste0("Excel (", by_var, "_", value, ")"), "Format by")
            #-----------------------------------------------------------------#
            print_step("MINOR", "[by] = [value]", by = by, value = value)

            # Put additional by info together with the information which by variable
            # and which value is currently filtered.
            by_info <- paste0(by_var, " = ", value)

            # Filter table by current by variable and value
            if (!is.na(value)){
                cross_temp <- cross_by |>
                    collapse::fsubset(cross_by[["by_vars"]] == value)
            }
            else{
                cross_temp <- cross_by |>
                    collapse::fsubset(is.na(cross_by[["by_vars"]]))
            }

            # The print_miss option enables a shortcut in formatting the sheets after
            # the first one. Since it guarantees that all follow up sheets are printed
            # with the exact same table width and height, because all categories are
            # printed, only the first sheet must be formatted. All other sheets can
            # clone the entire style from the first sheet.
            if (index == 1 || !print_miss){
                # Generate frequency tables as normal but base is filtered data frame
                wb_list <- format_cross_excel(wb,
                                              cross_temp,
                                              rows,
                                              columns,
                                              column_names,
                                              statistics,
                                              formats,
                                              by,
                                              titles,
                                              footnotes,
                                              style,
                                              output,
                                              show_total,
                                              by_info,
                                              index,
                                              NULL)
            }
            # Clone style for follow up iterations
            else{
                # Generate tables with no style but base is filtered data frame
                wb_list <- format_cross_excel(wb,
                                              cross_temp,
                                              rows,
                                              columns,
                                              column_names,
                                              statistics,
                                              formats,
                                              by,
                                              titles,
                                              footnotes,
                                              style,
                                              "excel_nostyle",
                                              show_total,
                                              by_info,
                                              index,
                                              NULL)

                # Clone entire style from first sheet
                wb_list[[1]]$clone_sheet_style(from = 1, to = index)
            }

            index <- index + 1

            wb <- wb_list[[1]]

            monitor_df <- monitor_df |> monitor_end()
        }

        monitor_df <- monitor_df |> monitor_end()
    }

    # Return workbook
    list(wb, monitor_df)
}

#' Reorder Variable Combination Alphabetically
#'
#' @description
#' [summarise_plus()] can create multiple variable combinations from the grouping
#' variables. The order in which the TYPE variable is generated is always the order
#' which the user provided. In [any_table()] different column variable combinations
#' have to be joined back to the same row variable combinations. The problem is, that
#' the order of row variables inside the loop can differ from the order in the TYPE
#' variable. Therefor this function can order this TYPE alphabetically, to make a
#' join possible.
#'
#' @param combination A combination of variables in the form "var1 + var2 + var3 + ...".
#'
#' @return
#' Returns an alphabetically sorted combination of variables.
#'
#' @noRd
reorder_combination <- function(combination){
    # Remove spaces if there are any
    combination <- gsub(" ", "", combination, fixed = TRUE)

    # Extract unique combinations so that only a short vector has to be reordered
    unique_combi  <- collapse::funique(combination)
    unique_vector <- strsplit(unique_combi, "+", fixed = TRUE)

    # Reorder variables of combinations alphabetically and combine them again into single combinations
    map <- stats::setNames(vapply(unique_vector,
                function(variable){
                    paste(sort(variable), collapse = "+")
                }, character(1)),
            unique_combi)

    # For each element of combination, retrieve value of map whose name equals that element.
    # Basically replacing the unordered with ordered combinations in one go.
    unname(map[combination])
}


#' Check if Format is a Multilabel
#'
#' @description
#' Check for a multilabel, meaning if a format can produce multiple new categories
#' from a single original value.
#'
#' @param format_list The list of variables with corresponding formats.
#' @param variable The variable whose format should be checked.
#'
#' @return
#' Returns TRUE or FALSE depending on the result.
#'
#' @noRd
is_multilabel <- function(format_list, variable){
    if (length(format_list) == 0){
        return(FALSE)
    }

    if (!variable %in% names(format_list)){
        return(FALSE)
    }

    # Select the current format
    format <- format_list[[variable]]

    # In case of discrete format a simple check for duplicated values is sufficient
    if (names(format)[1] == "value" && any(duplicated(format[["value"]]))){
        return(TRUE)
    }

    # In case of an interval format the format has to be joined with itself to
    # see whether the result produces more observations than the format had before.
    if ("from" %in% names(format)){
        format_temp <- format |> collapse::frename("label" = "label_check")

        format_dt      <- data.table::copy(format)
        format_temp_dt <- data.table::copy(format_temp)

        data.table::setkey(format_dt,      from, to)
        data.table::setkey(format_temp_dt, from, to)

        # Merge data frame with format by range
        format_check <- data.table::foverlaps(format_dt, format_temp_dt,
                                              by.x = c("from", "to"),
                                              by.y = c("from", "to"))

        if (collapse::fnrow(format_check) > collapse::fnrow(format)){
            return(TRUE)
        }
    }

    # No multilabel detected
    FALSE
}


#' Order Statistics Alternating
#'
#' @description
#' In [any_table()] the table columns can be ordered in different ways. Interleaved
#' means that columns of different statistics appear alternated in the table.
#'
#' @param data_frame The data frame which contains the columns to be ordered.
#' @param patterns The statistics in the order they should be interleaved.
#'
#' @return
#' Returns a data frame with interleaved ordered statistics columns.
#'
#' @noRd
order_interleaved <- function(data_frame, patterns){
    # Find columns for each pattern
    columns <- lapply(patterns, function(pattern){
        grep(pattern, names(data_frame), value = TRUE)
    })

    # Interleave columns. Pad shorter groups with NA and bind alternating.
    max_length  <- max(lengths(columns))

    interleaved <- as.vector(
        do.call(rbind, lapply(columns, function(column){
            c(column, rep(NA, max_length - length(column)))
        }))
    )
    interleaved <- interleaved[!is.na(interleaved)]

    # Order columns
    data_frame |> data.table::setcolorder(interleaved, after = ncol(data_frame))
}


#' Check If Data Frame Is Already Summarised
#'
#' @description
#' [any_table()] can be used to tabulate individual data but also pre summarised
#' data. This check is used to determine which route to use.
#'
#' @param data_frame The data frame which contains the columns to be checked.
#' @param group_vars The grouping variables which potentially form unique combinations.
#'
#' @return
#' Returns TRUE or FALSE depending on the result.
#'
#' @noRd
is_pre_summed <- function(data_frame, group_vars){
    !collapse::any_duplicated(data_frame[group_vars])
}


#' Unlist Variables
#'
#' @description
#' Unlists variables concatenated by + signs. This is used to get the single variables
#' in e.g. [any_table()], where rows and columns are passed like c("sex", "sex + age").
#'
#' @param var_names A character vector of variable names.
#'
#' @return
#' Character vector.
#'
#' @noRd
unlist_variables <- function(var_names){
    # This errors if a value without quotation marks is passed. In this case return NULL
    # so that the function using this check can abort.
    tryCatch({
        collapse::funique(trimws(unlist(strsplit(var_names, "\\+"))))
    }, error = function(e){
        NULL
    })
}


#' Check For Character Vectors
#'
#' @description
#' Checks for valid character vectors.
#'
#' @param vector_to_check A vector to check.
#'
#' @return
#' Returns TRUE or FALSE depending on the result.
#'
#' @noRd
is_valid_vector <- function(vector_to_check){
    is.vector(vector_to_check) && (length(vector_to_check) > 1 || !is.null(vector_to_check))
}


#' Catch Specific Message
#'
#' @description
#' Catches a specific message and excludes it from the output while printing all
#' other messages as normal.
#'
#' @param call_function A function which outputs messages.
#' @param pattern The message pattern for which to look out for.
#'
#' @return
#' Returns the call_functions return value.
#'
#' @noRd
suppress_specific_message <- function(call_function, pattern){
    withCallingHandlers(call_function,
                        message = function(single_message){
                            if (grepl(pattern, conditionMessage(single_message))){
                                invokeRestart("muffleMessage")
                            }
                        })
}


#' Check If List Is A Named List
#'
#' @description
#' Checks if an object is a list and all elements carry a name.
#'
#' @param list_to_check A list to be checked.
#'
#' @return
#' Returns TRUE or FALSE.
#'
#' @noRd
is_named_list <- function(list_to_check){
    is.list(list_to_check) && !is.null(names(list_to_check)) && all(names(list_to_check) != "")
}

#' Compute New Variables
#'
#' @description
#' Compute new variables without having to write the name of the data frame multiple times.
#' It can handle all kinds of operations like simple value assignment or calculations, but also
#' can make use of other functions.
#'
#' In addition it can be used in a conditional [do_if()] block.
#'
#' @param data_frame A data frame in which to compute new variables.
#' @param ... The calculations that should be executed.
#' @param monitor FALSE by default. If TRUE, outputs two charts to visualize the
#' functions time consumption.
#'
#' @details
#' The loop you can use within [compute.()] is based on the 'SAS' do-over-loop. This
#' type of loop iterates over every vector that appears in the loop in parallel.
#' Means that in the first iteration all the first vector elements are used, in the
#' second iteration all second elements of every vector, and so on. With this loop
#' you don't have the need to construct an outer loop, but can directly pass in
#' different vectors and let the function handle the loop inside.
#'
#' @return Returns a data frame with newly computed variables.
#'
#' @seealso
#' The following functions can make use of the [do_if()] filter variables:
#'
#' Conditions: [if.()], [else_if.()], [else.()]
#'
#' Filter Data Frame: [where.()]
#'
#' Create new Variables: [compute.()]
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Simple assignment
#' assign_df <- my_data |> compute.(new_var1 = 1,
#'                                 new_var2 = "Hello")
#'
#' # Simple calculation
#' sum_df <- my_data |> compute.(new_sum = age + sex)
#'
#' # Using functions
#' mean_df <- my_data |> compute.(new_mean = collapse::fmean(age))
#'
#' # Using qol functions
#' qol_df <- my_data |> compute.(row_sum = row_calculation("sum", state, age, sex))
#'
#' # Use compute.() as a do-over-loop. In this kind of loop all vectors will be
#' # advanced one iteration at a time in parallel.
#' new_vars <- c("var1", "var2", "var3")
#' money    <- c("income", "expenses", "balance")
#' multi    <- c(1, 2, 3)
#'
#' do_over_df <- my_data |> compute.(new_vars = money * multi)
#'
#' # You can also do all at once
#' all_df <- my_data |> compute.(new_var1 = 1,
#'                               new_var2 = "Hello",
#'                               new_sum  = age + sex,
#'                               new_mean = mean(age),
#'                               row_sum  = row_calculation("sum", state, age, sex),
#'                               new_vars = multi * money)
#'
#' # compute.() can be used in a do_if() situation and is aware of overarching
#' # conditions.
#' age_west. <- discrete_format("under 18"     = 0:17,
#'                              "18 and older" = 18:100)
#'
#' age_east. <- discrete_format("under 65"     = 0:64,
#'                              "65 and older" = 65:100)
#'
#' do_if_df <- my_data |>
#'     do_if(state < 11) |>
#'         compute.(region    = "West",
#'                  age_group = recode.(age = age_west.)) |>
#'     else_do() |>
#'         compute.(region    = "East",
#'                  age_group = recode.(age = age_east.)) |>
#'     end_do()
#'
#' @export
compute. <- function(data_frame,
                     ...,
                     monitor = .qol_options[["monitor"]]){
    # Measure the time
    print_start_message()

    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Get do_if() condition", "do_if()")
    #-------------------------------------------------------------------------#
    parent_env  <- parent.frame()
    assignments <- as.list(substitute(list(...)))[-1]

    # In case compute.() is used inside a function and only the passed object
    # is returned as unevaluated symbol, try to get the contents.
    if (length(assignments) == 1 && is.symbol(assignments[[1]])){
        # Evaluate the symbol in the parent environment to get the actual contents
        resolved_list <- tryCatch({
            # Force evaluation to see if it exists
            as.list(eval(assignments[[1]], envir = parent_env))
        }, error = function(e){
            # Evaluation failed
            NULL
        })

        if (!is.null(resolved_list) && is_named_list(resolved_list)){
            # Convert the list elements into standard language expressions
            assignments <- lapply(resolved_list, function(calculation){
                if (is.character(calculation)){
                    str2lang(calculation)
                }
                else{
                    calculation
                }
            })
        }
    }

    if (length(assignments) == 0){
        print_message("ERROR", "No assignments. Evaluation will be aborted.")
        return(invisible(data_frame))
    }

    # Check whether compute is used within an if.() statement. If that is the
    # case, then a hidden parameter is passed, carrying the condition. Additionally
    # check, if there are any do_if() conditions active and apply them
    if_condition_list <- NULL
    if_suppressed     <- FALSE

    if (".if_condition" %in% names(assignments)){
        condition                      <- TRUE
        if_condition_list              <- eval(assignments[[".if_condition"]], envir = parent_env)
        assignments[[".if_condition"]] <- NULL
    }
    else{
        condition <- data_frame |> combined_condition(TRUE)
    }

    if (".if_parent_frame" %in% names(assignments)){
        parent_env                        <- eval(assignments[[".if_parent_frame"]], envir = parent_env)
        assignments[[".if_parent_frame"]] <- NULL
    }

    if (".if_suppressed" %in% names(assignments)){
        assignments[[".if_suppressed"]] <- NULL
        if_suppressed <- TRUE
    }

    # The condition and the variable assignments are torn apart here, so that
    # only the unique variable and vector names are captured as characters.
    used_variables <- unique(c(names(assignments), # Get all variables to which a value should be assigned
                               unlist(lapply(assignments, all.names)))) # Get all assigned variables

    # Get the original contents of vectors if there are any and put them in a list.
    # List names are the symbols names from above and the elements hold the actual contents.
    content_list <- mget(used_variables, envir = parent_env, ifnotfound = list(NULL))

    # Remove invalid empty entries. This is the case if a variable is an original
    # variable name and not a vector of variable names.
    content_list <- Filter(is_valid_vector, content_list)

    list_entry_lengths <- lengths(content_list)

    if (length(collapse::funique(list_entry_lengths)) > 1){
        print_message("ERROR", c("Passed vectors are of unequal lengths. All vectors must have an",
								 "equal number of elements. Evaluation will be aborted."))
        return(invisible(data_frame))
    }

    # Check for duplicate variable names
    duplicate_vars <- duplicated(names(assignments))

    if (any(duplicate_vars)){
        duplicate_vars <- names(assignments)[duplicate_vars]

        print_message("WARNING", "Duplicate variable name[?s] '{var}' in compute. The last entry will overwrite the others before.",
                      var = duplicate_vars)
    }

    print_step("MAJOR", "Computing stats")

    variable_names <- names(data_frame)
    number_of_rows <- collapse::fnrow(data_frame)

    # Go trough each assignment and collect the translated calls
    for (entry in seq_along(assignments)){
        variable    <- names(assignments)[[entry]]
        calculation <- assignments[[variable]]
        calc_text   <- deparse(calculation)

        # Get used variables especially for this condition in combination mit
        # assigned variables.
        used_variables <- unique(c(names(assignments)[[entry]], # Get all variables to which a value should be assigned
                                   unlist(lapply(assignments[[variable]], all.names)))) # Get all assigned variables

        # If no vector was passed then evaluate as normal.
        if (!any(used_variables %in% names(content_list))){
            #-------------------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next(paste0(variable, " = ", calc_text), "Non vector")
            #-------------------------------------------------------------------------#
            print_step("MINOR", "[var] = [calc]", var = variable, calc = calc_text)

            # This step is important to make this function work in a nested situation.
            # Normally variable would be the name of what was last passed as a parameter.
            # If "if.()" is used nested inside a function this can basically be any placeholder.
            # So here we go up the ladder to get the original name of the variable.
            original_var <- get_origin_symbol(variable)

            # Evaluate complete assignment first without condition
            expression <- get_custom_functions(calculation, parent_env)
            value      <- suppressMessages(eval(expression, envir = data_frame))

            # If there already is a variable with the given name pick the existing value as fallback
            flag_var_in_data <- FALSE

            if (original_var %in% variable_names){
                flag_var_in_data <- TRUE

                # Check if existing variable type is of same type as assigned value.
                # Put out a warning on type mismatch.
                is_type_missmatch <- check_types(data_frame, original_var, value)

                # If result is NULL, then there is a variable with all NA values
                # which has to be converted into the right type.
                if (is.null(is_type_missmatch)){
                    if (is.character(value)){
                        data_frame[[variable]] <- NA_character_
                    }
                    else if (is.numeric(value)){
                        data_frame[[variable]] <- NA_real_
                    }
                }
                # Convert to character on type miss match
                else if (is_type_missmatch){
                    data_frame[[original_var]] <- as.character(data_frame[[original_var]])
                    value <- as.character(value)
                }
            }

            # Values are now conditionally evaluated. Which route is used depends on whether
            # this function is called on its own or via an if statement.
            if (!is.null(if_condition_list)){
                for (i in seq_len(length(if_condition_list))){
                    # Merge the given condition as well as the active filter variables together
                    # into one logical vector
                    full_condition <- condition & unlist(if_condition_list[[i]])

                    # When function is called via if statement and variable is in the
                    # data frame, then variable will be created, but fallback value is
                    # not NA, but the values which are already there. This is in case of
                    # else_if.() or else.().
                    if (flag_var_in_data){
                        conditional_value <- data.table::fifelse(full_condition, value, data_frame[[original_var]])
                    }
                    # When function is called via if statement and variable is not already in the
                    # data frame, then variable will be created.
                    else{
                        if (length(full_condition) == 1){
                            conditional_value <- value
                        }
                        else{
                            conditional_value <- data.table::fifelse(full_condition, value, NA)
                        }
                    }

                    # Only add call, if there are evaluated values. With else_if.() or else.()
                    # it can happen that no values are generated, in case there is no original
                    # variable in the data frame.
                    if (length(conditional_value) > 0){
                        new_call <- stats::setNames(list(conditional_value), variable)

                        # Add to call list
                        data_frame <- collapse::ftransform(data_frame, new_call)
                    }
                }
            }
            # When function is called on its own, variable will be created, ignoring whether
            # it is already in the data frame or not.
            else{
                full_condition <- condition

                if (length(full_condition) == 1){
                    conditional_value <- value
                }
                else{
                    # When function is called with active do_if statement and variable
                    # is in the data frame, then variable will be created, but fallback
                    # value is not NA, but the values which are already there.
                    if (flag_var_in_data){
                        conditional_value <- data.table::fifelse(full_condition, value, data_frame[[original_var]])
                    }
                    else{
                        conditional_value <- data.table::fifelse(full_condition, value, NA)
                    }
                }

                # Only add call, if there are evaluated values. With else_if.() or else.()
                # it can happen that no values are generated, in case there is no original
                # variable in the data frame.
                if (length(conditional_value) > 0){
                    new_call <- stats::setNames(list(conditional_value), variable)

                    # Add to call list
                    data_frame <- collapse::ftransform(data_frame, new_call)
                }
            }
        }
        # If a vector was passed, then evaluate as a do over loop. Meaning for each
        # vector the same elements are used simultaneously one after another.
        else{
            #-------------------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next(paste0(variable, " = ", calc_text), "Vector based")
            #-------------------------------------------------------------------------#

            call_list <- list()

            # Do over loop per element
            for (element in seq_len(list_entry_lengths[1])){
                # Get the respective first elements as symbols in a new list
                replace_list <- lapply(content_list, function(vector){
                    if (is.numeric(vector)){
                        expression <- vector[element]
                    }
                    else{
                        expression <- as.name(vector[element])
                    }
                })

                # Get the target variable from the assignment list and look whether
                # it is a vector or not. If it is a vector, extract the current element
                # and use its name.
                target_variable <- variable

                if (target_variable %in% names(content_list)){
                    target_variable <- content_list[[target_variable]][element]
                }

                # Evaluate complete assignment first without condition
                value_var  <- do.call(substitute, list(calculation, replace_list))
                expression <- get_custom_functions(value_var, parent_env)
                value      <- suppressMessages(eval(expression, envir = data_frame))

                print_step("MINOR", "{target} = [value]", target = target_variable, value = deparse(value_var))

                # Look up, if single value was passed or vector of values. Only if a
                # vector of values is passed, which has fewer observations than the
                # data frame, it will be processed element wise.
                number_of_values <- length(value)

                if (number_of_values > 1 && number_of_values < number_of_rows){
                    value <- value[element]
                }

                # If there already is a variable with the given name pick the existing value as fallback
                flag_var_in_data <- FALSE

                if (target_variable %in% variable_names){
                    flag_var_in_data <- TRUE

                    # Check if existing variable type is of same type as assigned value.
                    # Put out a warning on type mismatch.
                    is_type_missmatch <- check_types(data_frame, target_variable, value)

                    # If result is NULL, then there is a variable with all NA values
                    # which has to be converted into the right type.
                    if (is.null(is_type_missmatch)){
                        if (is.character(value)){
                            data_frame[[variable]] <- NA_character_
                        }
                        else if (is.numeric(value)){
                            data_frame[[variable]] <- NA_real_
                        }
                    }
                    # Convert to character on type miss match
                    else if (is_type_missmatch){
                        data_frame[[target_variable]] <- as.character(data_frame[[target_variable]])
                        value <- as.character(value)
                    }
                }

                # Values are now conditionally evaluated. Which route is used depends on whether
                # this function is called on its own or via an if statement.
                if (!is.null(if_condition_list)){
                    # In case there is a condition without do over behavior always
                    # take the first element, because there is no other.
                    if (length(if_condition_list) == 1){
                        full_condition <- condition & if_condition_list[[1]]
                    }
                    # Otherwise move along the list
                    else{
                        full_condition <- condition & if_condition_list[[element]]
                    }

                    # When function is called via if statement and variable is in the
                    # data frame, then variable will be created, but fallback value is
                    # not NA, but the values which are already there. This is in case of
                    # else_if.() or else.().
                    if (flag_var_in_data){
                        conditional_value <- data.table::fifelse(full_condition, value, data_frame[[target_variable]])
                    }
                    # When function is called via if statement and variable is not already in the
                    # data frame, then variable will be created.
                    else{
                        if (length(full_condition) == 1){
                            conditional_value <- value
                        }
                        else{
                            conditional_value <- data.table::fifelse(full_condition, value, NA)
                        }
                    }
                }
                # When function is called on its own, variable will be created, ignoring whether
                # it is already in the data frame or not.
                else{
                    full_condition <- condition

                    if (length(full_condition) == 1){
                        conditional_value <- value
                    }
                    else{
                        # When function is called with active do_if statement and variable
                        # is in the data frame, then variable will be created, but fallback
                        # value is not NA, but the values which are already there.
                        if (flag_var_in_data){
                            conditional_value <- data.table::fifelse(full_condition, value, data_frame[[target_variable]])
                        }
                        else{
                            conditional_value <- data.table::fifelse(full_condition, value, NA)
                        }
                    }
                }

                new_call <- stats::setNames(list(conditional_value), target_variable)

                # Add to call list
                call_list <- c(call_list, new_call)
            }

            if (length(call_list) > 0){
                # If there are any duplicate names in list, the call_list has to be cleaned
                # up first.
                if (length(call_list) > 1 && any(duplicated(names(call_list)))){
                    # Reverse order of the list, so that the last applyed value
                    # beats the others.
                    call_list <- rev(call_list)

                    # Group up duplicate list names in their own list
                    call_list <- split(call_list, names(call_list))

                    # Lists with duplicate variable names will be merged together, because
                    # it is actually one variable with different expressions. Probably
                    # generated within a do-over-loop. fcoalesce takes the first value
                    # per line which is not NA, so first value beats other values, if there
                    # are any. Since the list was reversed before it is actually the last
                    # assigned value that beats the others.
                    call_list <- lapply(call_list, function(element){
                        if (length(element) > 1) {
                            data.table::fcoalesce(element)
                        }
                        else{
                            element[[1]]
                        }
                    })
                }

                data_frame <- collapse::ftransform(data_frame, call_list)
            }
        }
    }

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Add variables", "Vector based")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Add variables to data frame")

    if (if_suppressed){
        print_closing(suppress = TRUE)
    }
    else{
        print_closing()
    }

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_end()
    monitor_df |> monitor_plot(draw_plot = monitor)
    #-------------------------------------------------------------------------#

    data_frame
}


#' Translate Expressions
#'
#' @description
#' Translates expressions like calculations, functions based on vectors and
#' custom functions from this package into the final expression to evaluate.
#' All expressions basically stay the same, just the custom functions from this
#' package get the "data_frame" argument injected.
#'
#' @param expression The expression to be translated.
#' @param env The parent environment in which the expression has to be evaluated.
#'
#' @return
#' Returns a translated expression
#'
#' @noRd
get_custom_functions <- function(expression, env){
    # If the expression is not just a number or text, check what type of call it is
    if (is.call(expression)){
        # Decide whether expression is a function or a simple calculation
        custom_function <- tryCatch({
            # Force evaluation to see if it exists
            eval(expression[[1]], env)
        }, error = function(e){
            # Evaluation failed
            NULL
        })

        # If expression is a function check whether it is a function from this package
        if (is.function(custom_function)){
            function_args <- names(formals(custom_function))

            # If the first function argument is "data_frame" then the function is
            # from this package. In this case "data_frame" need to be added as first
            # argument so that the function can be evaluated correctly.
            if (length(function_args) > 0 && function_args[1] == "data_frame"){
                expression <- as.call(c(expression[[1]], quote(data_frame), as.list(expression[-1])))
            }
        }
    }

    expression
}


#' Check for Identical Types
#'
#' @description
#' Check if a certain value is of the same type as a variable.
#'
#' @param data_frame The data frame which contains the variables to check.
#' @param variable The main variable whose data type should be compared to value.
#' @param current The current value whose data type should be compared to variable.
#'
#' @return
#' Returns a further formatted workbook.
#'
#' @noRd
check_types <- function(data_frame, variable, current){
    # Abort if all values are NA
    if (all(is.na(data_frame[[variable]]))){
        return(NULL)
    }

    type_c <- typeof(current)
    type_d <- typeof(data_frame[[variable]])

    # Abort, if types are identical
    if (identical(type_c, type_d)){
        return(FALSE)
    }
    # Also abort, if both types are of a numerical type
    else if(type_c %in% c("integer", "double") && type_d %in% c("integer", "double")){
        return(FALSE)
    }

    print_message("WARNING", c("Type mismatch: Current value [current] is of type [type_c] but should be of",
							   "type [type_d]. [variable] will be converted to character."),
                  current = current[1], type_c = type_c, type_d = type_d, variable = variable,
				  always_print = TRUE)

    TRUE
}

#' Drop automatically generated Variables
#'
#' @description
#' If [summarise_plus()] is used with the nested options "all" or "single", three
#' hierarchical metadata variables are automatically generated: TYPE, TYPE_NR and
#' DEPTH. With this functions these variables are dropped.
#'
#' - **`TYPE`** (character): The active grouping combination for the row (e.g., `"year+sex"`), or `"total"` for the grand total.
#' - **`TYPE_NR`** (integer): A unique sequential identifier assigned to each distinct grouping combination (`TYPE`).
#' - **`DEPTH`** (integer): The number of active grouping variables in the row's combination (e.g., `0` for `"total"`, `1` for single variables, etc.).
#'
#' @param data_frame The data frame with automatically generated variables.
#'
#' @return
#' Returns a data frame without the variables TYPE, TYPE_NR and DEPTH.
#'
#' @examples
#' # Example format
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Call function
#' all_possible <- my_data |>
#'     summarise_plus(class      = c(year, sex),
#'                    values     = c(income, probability),
#'                    statistics = c("sum", "mean", "freq"),
#'                    formats    = list(sex = "sex."),
#'                    weight     = weight,
#'                    nesting    = "all",
#'                    na.rm      = TRUE) |>
#'     drop_type_vars()
#'
#' @export
drop_type_vars <- function(data_frame){
    data_frame |> dropp("TYPE", "TYPE_NR", "DEPTH")
}


#' Fuse Multiple Variables
#'
#' @description
#' When you have a situation where you have multiple variables with different NA
#' values that happen to be in different places (where one variable has a value the
#' other is NA and vice versa) you can fuse these together to a single variable.
#'
#' @param data_frame A data frame with variables to fuse.
#' @param new_variable_name The name of the new fused variable.
#' @param variables_to_fuse A vector with the variables that should be fused together.
#' @param drop_original_vars Whether to drop or keep the original values. TRUE by default.
#'
#' @return
#' Returns a data frame without the variables TYPE, TYPE_NR and DEPTH.
#'
#' @examples
#' # Example format
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Call function
#' all_possible <- my_data |>
#'     summarise_plus(class      = c(year, sex),
#'                    values     = c(income, probability),
#'                    statistics = c("sum", "mean", "freq"),
#'                    formats    = list(sex = "sex."),
#'                    weight     = weight,
#'                    nesting    = "all",
#'                    na.rm      = TRUE)
#'
#' all_possible <- all_possible[DEPTH <= 1] |>
#'     fuse_variables("fusion", c("year", "sex"))
#'
#' # NOTE: You can generally use this function to fuse variables. What is done in
#' #       multiple steps above can be achieved by just using nested = "single" in
#' #       summarise_plus.
#' single <- my_data |>
#'     summarise_plus(class      = c(year, sex),
#'                    values     = c(income, probability),
#'                    statistics = c("sum", "mean", "freq"),
#'                    formats    = list(sex = "sex."),
#'                    weight     = weight,
#'                    nesting    = "single",
#'                    na.rm      = TRUE)
#'
#' @export
fuse_variables <- function(data_frame,
                           new_variable_name,
                           variables_to_fuse,
                           drop_original_vars = TRUE){
    # Convert to character
    new_variable_name <- gsub("\"", "", deparse(substitute(new_variable_name)))

    if (length(new_variable_name) > 1){
        print_message("ERROR", "No vector allowed. Only one new variable can be generated.")
        return(invisible(data_frame))
    }

    # Convert to character vectors
    variables_to_fuse <- get_origin_as_char(variables_to_fuse, substitute(variables_to_fuse))

    # Make sure that the variables provided are part of the data frame.
    variables_to_fuse <- data_frame |> part_of_df(variables_to_fuse)

    # Convert the group columns to character so coalesce works properly
    selected_data   <- data_frame[variables_to_fuse]
    selected_data[] <- lapply(selected_data, as.character)

    # Create a new variable with the first non-NA value from the group columns
    data_frame[[new_variable_name]] <- do.call(data.table::fcoalesce, selected_data)

    if (drop_original_vars){
        data_frame <- data_frame |> dropp(variables_to_fuse)
    }

    data_frame |> data.table::setcolorder(new_variable_name)
}


#' Order Columns by Variable Name Patterns
#'
#' @description
#' Order variables in a data frame based on a pattern rather than whole variable
#' names. E.g. grab every variable that contains "sum" in it's name and order
#' them together so that they appear next to each other.
#'
#' @param data_frame The data frame to be ordered.
#' @param pattern The pattern which is used for ordering the data frame columns.
#'
#' @return
#' Returns a reordered data frame with the ordered variables at the end.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Summarise data
#' all_nested <- my_data |>
#'     summarise_plus(class      = c(year, sex),
#'                    values     = c(weight, income),
#'                    statistics = c("sum", "pct_group", "pct_total", "sum_wgt", "freq"),
#'                    weight     = weight,
#'                    nesting    = "deepest",
#'                    na.rm      = TRUE)
#'
#' # Set a different column order
#' new_order <- all_nested |> setcolorder_by_pattern(c("pct", "freq", "sum"))
#'
#' @export
setcolorder_by_pattern <- function(data_frame, pattern){
    # Copy data frame or else the original input data frame will be altered too
    data_frame <- data.table::copy(data_frame)

    # Select all variables in blocks of the requested pattern
    ordered_cols <- unlist(lapply(pattern, function(single_pattern){
        # Returns all variable names that match the provided pattern
        grep(single_pattern, names(data_frame),
             value = TRUE, ignore.case = TRUE)
    }))

    ordered_cols <- collapse::funique(ordered_cols)

    # Put the ordered columns at the end of the data frame
    data_frame |> data.table::setcolorder(ordered_cols, after = collapse::fncol(data_frame))
}


#' Add Empty Variables In A Given Range
#'
#' @description
#' Add empty variables to a data frame in the provided range. Basically does in a
#' data frame, what paste0("age", 1:10) does for a vector.
#'
#' @param data_frame A data frame to add variables to.
#' @param var_range A range of variables to add, provided in the form: var_name1:var_name10.
#'
#' @return
#' Returns a data frame with added variables.
#'
#' @examples
#' # Example data frames
#' my_data <- dummy_data(100)
#'
#' # Add variable range
#' my_data <- my_data |> add_variable_range(status1:status12)
#'
#' @export
add_variable_range <- function(data_frame, var_range){
    var_range <- get_origin_as_char(var_range, substitute(var_range))

    # Using regex to capture prefix, start number, and end number
    # Meaning: (prefix)(start):(prefix)(end)
    pattern <- "^([A-Za-z_.]+)([0-9]+):\\1([0-9]+)$"

    if (!grepl(pattern, var_range)){
        print_message("ERROR", c("Variable range has to be provided in the form 'var_name1:var_name10'.",
								 "Variable names must match. No variables will be added."))
        return(invisible(data_frame))
    }

    # Put together variable names
    prefix <- gsub(pattern, "\\1", var_range)
    start  <- as.numeric(gsub(pattern, "\\2", var_range))
    end    <- as.numeric(gsub(pattern, "\\3", var_range))

    var_names <- paste0(prefix, start:end)

    # If variable names are already part of the data frame, abort
    invalid_variables <- var_names[var_names %in% names(data_frame)]

    if (length(invalid_variables) > 0){
        print_message("ERROR", c("Some variables are already part of the data frame: [invalid].",
								 "No variables will be added."), invalid = invalid_variables)
        return(invisible(data_frame))
    }

    # Create new variables in their own data frame first
    new_columns <- collapse::qDF(matrix(NA_integer_,
                                        nrow = collapse::fnrow(data_frame),
                                        ncol = end))

    names(new_columns) <- paste0(prefix, start:end)

    # Add new empty variables to data frame
    collapse::add_vars(data_frame, new_columns)
}


#' Get Integer Length
#'
#' @description
#' Get the number of digits of an integer variable.
#'
#' @param variable The integer variable from which to get the length.
#'
#' @return
#' Returns a vector with the number of digits places.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(100)
#'
#' my_data[["age_length"]] <- get_integer_length(my_data[["age"]])
#'
#' @export
get_integer_length <- function(variable){
    if (!is.numeric(variable)){
        print_message("ERROR", "Only numeric values allowed.")

        return(invisible(NA_integer_))
    }

    if (!is.integer(variable)){
        print_message("WARNING", "Variable is not an integer and will be floored. Decimal places won't count.")

        variable <- floor(variable)
    }

    nchar(abs(variable))
}


#' Check For Duplicate Variable Names
#'
#' @name duplicates
#'
#' @description
#' Checks for duplicate variable names in a data frame, e.g. AGE, age and Age.
#'
#' @param data_frame The data frame which variable names to check for duplicates.
#'
#' @return
#' [get_duplicate_var_names()]: Returns a vector of duplicate variable names.
#'
#' @examples
#' # Example data frame
#' my_data <- data.frame(age    = 1,
#'                       AGE    = 2,
#'                       Age    = 3,
#'                       sex    = 1)
#'
#' dup_var_names <- my_data |> get_duplicate_var_names()
#'
#' @rdname duplicates
#'
#' @export
get_duplicate_var_names <- function(data_frame){
    var_names <- names(data_frame)

    duplicate_names <- duplicated(tolower(var_names)) |
        duplicated(tolower(var_names), fromLast = TRUE)

    var_names[duplicate_names]
}


#' @description
#' Counts the number of duplicated variables in a data frame. If a variable
#' appears three times, e.g. AGE, age and Age, the variable count will be one.
#'
#' @return
#' [get_duplicate_var_count()]: Returns the count of duplicated variables as a
#' numeric value.
#'
#' @examples
#' dup_var_count <- my_data |> get_duplicate_var_count()
#'
#' @rdname duplicates
#'
#' @export
get_duplicate_var_count <- function(data_frame){
    duplicate_names <- data_frame |> get_duplicate_var_names()

    unique_names <- collapse::funique(tolower(duplicate_names))

    length(unique_names)
}


#' Round Values With Half Rounded Up And Multiples Of X
#'
#' @name round_values
#'
#' @description
#' This function rounds values according to DIN 1333 (round half up) or as an
#' alternative in multiples of a given value.
#'
#' @param values Numeric values to round.
#' @param digits The number of decimal places the values should be rounded to.
#' @param multiple The multiple to round the values to.
#'
#' @return
#' [round_values()]: Returns rounded values.
#'
#' @examples
#' # With vectors
#' round_numbers1 <- round_values(c(-0.5, -0.4, 0.1, 0.49, 0.5, 1.5, 2.5, 3.2))
#' round_numbers2 <- round_values(c(-0.5, -0.49, 0.17, 0.499, 0.51, 1.549, 2.51, 3.25),
#'                                digits = 1)
#' round_numbers3 <- round_values(c(-0.3, -0.24, 1.17, 2.749, 0.25, 1.549, 2.75, 3.25),
#'                                multiple = 0.5)
#'
#' # With a data frame
#' my_data <- dummy_data(100)
#'
#' my_data[["income_round1"]] <- my_data[["income"]] |> round_values()
#' my_data[["income_round2"]] <- my_data[["income"]] |> round_values(multiple = 100)
#'
#'
#' @rdname round_values
#'
#' @export
round_values <- function(values,
                         digits   = 0,
                         multiple = NULL){
    if (!is.numeric(values)){
        # Silently return if all values are NA
        if (is.logical(values)){
            return(invisible(values))
        }

        print_message("ERROR", "Only numeric values allowed. Rounding will be aborted.")
        return(invisible(values))
    }

    eps <- .Machine[["double.eps"]]

    if (!is.null(multiple)){
        # Round in multiples of x
        values <- trunc(abs(values) / multiple + (0.5 + sqrt(eps))) * multiple * sign(values)
    }
    else{
        # Pre calculate the multiplier
        p <- 10 ^ digits

        # Using vectorized math to reduce the number of temporary objects created otherwise
        values <- trunc(abs(values) * p + (0.5 + sqrt(eps))) / p * sign(values)
    }

    values
}


#' @description
#' [round_multi()]: Rounds multiple variables at once inside a data frame.
#'
#' @param data_frame The data frame in which the variables to be rounded are found.
#' @param variables The variable names of which to round the values.
#' @param new_names The new names of the rounded variables. If provided adds variables,
#' if not overwrites the existing variables with rounded values.
#'
#' @return
#' [round_multi()]: Returns a data frame with rounded values.
#'
#' @examples
#' # Round multiple variables in a data frame
#' my_data <- my_data |> round_multi(variables = c(income,  expenses,  balance),
#'                                   new_names = c(incomeR, expensesR, balanceR),
#'                                   digits = 1)
#'
#' @rdname round_values
#'
#' @export
round_multi <- function(data_frame,
                        variables,
                        new_names = NULL,
                        digits    = 0,
                        multiple  = NULL){
    variables <- get_origin_as_char(variables, substitute(variables))
    new_names <- get_origin_as_char(new_names, substitute(new_names))

    # Make sure that the variables provided are part of the data frame.
    variables <- data_frame |> part_of_df(variables)

    # If no variables are provided return data frame without rounding
    if (length(variables) == 0){
        return(invisible(data_frame))
    }

    # If no variables are provided return data frame without rounding
    if (!is.null(new_names) && length(variables) != length(new_names)){
        print_message("ERROR", "<Variables> and <new_names> are of unequal length. Rounding will be aborted.")
        return(invisible(data_frame))
    }

    # Expand digits and multiples to variable count, if fewer values than variables
    # are given.
    var_count    <- length(variables)
    digits_count <- length(digits)

    digits <- digits[seq_len(min(digits_count, var_count))]

    if (!is.null(digits) && digits_count < var_count){
        digits <- c(digits, rep(digits[digits_count], var_count - digits_count))
    }

    multiple_count <- length(multiple)

    multiple <- multiple[seq_len(min(multiple_count, var_count))]

    if (!is.null(multiple) && multiple_count < var_count){
        multiple <- c(multiple, rep(multiple[multiple_count], var_count - multiple_count))
    }

    # Loop through all provided variables and check whether they are numeric.
    # Non numeric variables will be excluded.
    invalid_vars <- c()

    for (i in seq_along(variables)){
        variable <- variables[[i]]
        new_name <- new_names[[i]]
        digit    <- digits[[i]]
        multi    <- multiple[[i]]

        # Exclude non numeric variables
        if (!is.numeric(data_frame[[variable]])){
            invalid_vars <- c(invalid_vars, variable)
            variables    <- variables[!variable %in% variables]
        }
        # Actual rounding
        else{
            # Overwrite existing variables if no new names are provided
            if (is.null(new_name)){
                new_name <- variable
            }

            data_frame[[new_name]] <- data_frame[[variable]] |>
                round_values(digits = digit, multiple = multi)
        }
    }

    if (length(invalid_vars)){
        print_message("WARNING", c("Only numeric values allowed. [invalid] [?is/are] not numeric, and will",
                                   "be excluded from rounding."), invalid = invalid_vars)
    }

    data_frame
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

dummy_df <- dummy_data(100)
dup_df   <- data.frame(age = 1, AGE = 2, Age = 3,
                       sex = 1, SeX = 2,
                       Height_male = 1, Height_female = 1,
                       weight = 2)

test_df <- data.frame(var1 = c(-2.5, -2.4, 2.4, 2.5),
                      var2 = c(2.4, NA, NA, 2.5),
                      var3 = c("a", "b", "c", "d"))


# Add new NA column range with helper function
added_df <- dummy_df |>
      add_variable_range(status1:status3)

expect_equal(utils::tail(names(added_df), 3), c("status1", "status2", "status3"), info = "Add new NA column range with helper function")


# Get duplicate variable names
expect_equal(get_duplicate_var_names(dup_df),
             c("age", "AGE", "Age", "sex", "SeX"), info = "Get duplicate variable names")


# Get duplicate variable count
expect_equal(get_duplicate_var_count(dup_df), 2, info = "Get duplicate variable count")


# Round values half up
values_to_round <- c(-2.5, -2.4, -0.5, -0.4, 0.4, 0.5, 2.4, 2.5)

expect_equal(round_values(values_to_round), c(-3, -2, -1, 0, 0, 1, 2, 3), info = "Round values half up")


# Round values with NULL digits returns values unchanged
values_to_round <- c(-2.5789546547, -5.4563546547, 0.1234546547, 1.45465475465)

expect_equal(round_values(values_to_round, digits = NULL), values_to_round,
             info = "Round values with NULL digits returns values unchanged")


# Round to multiples
values_to_round <- c(-7.5, -7.4, -2.5, -2.4, -0.5, -0.4, 0.4, 0.5, 2.4, 2.5, 7.4, 7.5)

expect_equal(round_values(values_to_round, multiple = 5), c(-10, -5, -5, 0, 0, 0, 0, 0, 0, 5, 5, 10),
             info = "Round to multiples")


# Round multiple variables in a data frame half up
new_df <- test_df |> round_multi(c(var1, var2))

expect_equal(new_df[["var1"]], c(-3, -2, 2, 3), info = "Round multiple variables in a data frame half up")
expect_equal(new_df[["var2"]], c(2, NA, NA, 3), info = "Round multiple variables in a data frame half up")


# Round multiple variables in a data frame to multiples
new_df <- test_df |> round_multi(c(var1, var2), multiple = 5)

expect_equal(new_df[["var1"]], c(-5, 0, 0, 5), info = "Round multiple variables in a data frame to multiples")
expect_equal(new_df[["var2"]], c(0, NA, NA, 5), info = "Round multiple variables in a data frame to multiples")


# Round multiple variables in a data frame and add them as new variables
new_df <- test_df |> round_multi(variables = c(var1, var2), new_names = c(round1, round2))

expect_equal(new_df[["round1"]], c(-3, -2, 2, 3),   info = "Round multiple variables in a data frame and add them as new variables")
expect_equal(new_df[["round2"]], c(2, NA, NA, 3),   info = "Round multiple variables in a data frame and add them as new variables")
expect_true(collapse::fsum(new_df[["var1"]] == new_df[["round1"]]) == 0,
            info = "Round multiple variables in a data frame and add them as new variables")
expect_true(collapse::fsum(new_df[["var2"]] == new_df[["round2"]]) == 0,
            info = "Round multiple variables in a data frame and add them as new variables")

###############################################################################
# Abort checks
###############################################################################

# Adding NA variables to data frame aborts on wrong pattern
added_df <- dummy_df |>
     add_variable_range(status1:age3)

expect_error(print_stack_as_messages("ERROR"), "Variable range has to be provided in the form 'var_name1:var_name10'.", info = "Adding NA variables to data frame aborts on wrong pattern")


# Adding NA variables aborts, if variables are already part of data frame
added_df <- dummy_df |>
     add_variable_range(status1:status3)

added_df2 <- added_df |>
     add_variable_range(status1:status3)

expect_error(print_stack_as_messages("ERROR"), "Some variables are already part of the data frame: ", info = "Adding NA variables aborts, if variables are already part of data frame")


# Abort round values if values are not numeric
round_values("Test")

expect_error(print_stack_as_messages("ERROR"), "Only numeric values allowed.", info = "Abort round values if values are not numeric")


# Round multiple variables in a data frame aborts on unequal variable name lengths
new_df <- test_df |> round_multi(variables = c(var1, var2), new_names = c(round1, round2, round3))

expect_error(print_stack_as_messages("ERROR"), "<Variables> and <new_names> are of unequal length. Rounding will be aborted.",
             info = "Round multiple variables in a data frame aborts on unequal variable name lengths")


# Round multiple variables in a data frame throws a warning on non numeric variables
new_df <- test_df |> round_multi(variables = c(var1, var2, var3), new_names = c(round1, round2, round3))

expect_warning(print_stack_as_messages("WARNING"), "Only numeric values allowed. var3 is not numeric, and will",
             info = "Round multiple variables in a data frame throws a warning on non numeric variables")


set_no_print()

#' High Level Import From And Export To CSV And XLSX
#'
#' @name import_export
#'
#' @description
#' [import_data()]: A wrapper for [data.table::fread()] and [openxlsx2::wb_to_df()],
#' providing basic import functionality with minimal code. Uses "Latin-1" encoding.
#'
#' @param infile Full file path with extension to a csv or xlsx file to be imported.
#' @param sheet Only used in xlsx import. Which sheet of the workbook to import.
#' @param region Only used in xlsx import. Can either be an 'Excel' range like 'A1:BY27'
#' or the name of a named region.
#' @param data_frame A data frame to export.
#' @param outfile Full file path with extension. Allowed extensions are ".csv", ".txt" and ".xlsx".
#' @param separator Only used in CSV/TXT-export. Defines the single character value separator.
#' @param decimal Only used in CSV/TXT-export. Defines the single character decimal character.
#' @param var_names TRUE by default. Whether to export variable names or not.
#'
#' @details
#' [import_data()] and [export_data()] are based on the 'SAS' procedures Proc Import and Proc Export,
#' which provide a very straight forward syntax. While 'SAS' can import many different formats with
#' these procedures, these 'R' versions concentrate on importing CSV, TXT and XLSX files.
#'
#' The main goal here is to just provide as few as possible parameters to tackle most of the imports
#' and exports. These error handling also tries to let an import and export happen, even though
#' a parameter wasn't provided in the correct way.
#'
#' @return
#' Returns a data frame.
#'
#' @seealso
#' Functions that can export with style: [frequencies()], [crosstabs()], [any_table()],
#' [export_with_style()].
#'
#' Creating a custom table style: [excel_output_style()], [modify_output_style()],
#' [number_format_style()], [modify_number_formats()].
#'
#' Global style options: [set_style_options()], [set_labels()].
#'
#' @examples
#' # Example files
#' csv_file  <- system.file("extdata", "qol_example_data_csv.csv",   package = "qol")
#' txt_file  <- system.file("extdata", "qol_example_data_txt.txt",   package = "qol")
#' xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")
#'
#' # Import: Provide full file path
#' my_csv  <- import_data(csv_file)
#' my_csv  <- import_data(txt_file)
#' my_xlsx <- import_data(xlsx_file)
#'
#' # Import specific regions
#' range_import <- import_data(xlsx_file, region = "B4:H32")
#' name_import  <- import_data(xlsx_file, region = "test_region")
#'
#' # Import from another sheet
#' sheet_import <- import_data(xlsx_file, sheet = "Sheet 2")
#'
#' @rdname import_export
#'
#' @export
import_data <- function(infile,
                        sheet     = 1,
                        region    = NULL,
                        separator = "auto",
                        decimal   = "auto",
                        var_names = TRUE){
    print_start_message(suppress = TRUE)

    ###########################################################################
    # Error handling
    ###########################################################################

    if (inherits(infile, "wbWorkbook")){
        extension <- "xlsx"
    }
    else{
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Path
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Abort on vector provided as path
        if (!is.character(infile) || length(infile) != 1){
            print_message("ERROR", "<Infile> must be a single character. Import will be aborted.")

            return(invisible(NULL))
        }

        # Apply macros to infile
        infile <- macro(infile)

        # Abort on invalid path
        if (!dir.exists(dirname(infile)) || dirname(infile) == "."){
            print_message("ERROR", c("Path does not exist: [infile]",
									 "Import will be aborted."), infile = infile)

            return(invisible(NULL))
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # File extension
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        extension <- tolower(tools::file_ext(infile))

        if (extension == ""){
            print_message("ERROR", c("No file extension provided in <infile>. 'csv' and 'xlsx' are allowed.",
									 "Import will be aborted."))

            return(invisible(NULL))
        }

        if (!extension %in% c("csv", "txt", "xlsx")){
            print_message("ERROR", c("Only 'csv', 'txt' or 'xlsx' are allowed as file extensions in the <infile>.",
									 "Import will be aborted."))

            return(invisible(NULL))
        }
    }

    ###########################################################################
    # Import
    ###########################################################################

    if (extension %in% c("csv", "txt")){
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Separator
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Separator may only consist of one character
        if (!is.character(separator)){
            print_message("WARNING", "<Separator> must be provided as character. Automatic detection will be used.")

            separator <- "auto"
        }

        if (length(separator) != 1L || (separator != "auto" && nchar(separator) != 1L)){
            print_message("WARNING", "<Separator> may only be one character. Automatic detection will be used.")

            separator <- "auto"
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Decimal
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Decimal may only consist of one character
        if (!is.character(decimal)){
            print_message("WARNING", "<Decimal> must be provided as character. Automatic detection will be used.")

            decimal <- "auto"
        }

        if (length(decimal) != 1L || (decimal != "auto" && nchar(decimal) != 1L)){
            print_message("WARNING", "<Decimal> may only be one character. Automatic detection will be used.")

            decimal <- "auto"
        }

        # Decimal may not be equal to separator
        if (separator != "auto" && decimal != "auto" && separator == decimal){
            print_message("WARNING", "<Decimal> may not be the same character as the <separator>. Automatic detection will be used.")

            decimal <- "auto"
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Actual import
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        data_frame <- data.table::fread(file     = infile,
                                        sep      = separator,
                                        dec      = decimal,
                                        header   = var_names,
                                        encoding = "Latin-1")
    }
    # xlsx
    else if (extension == "xlsx"){
        if (!is.null(region)){
            # Region may only consist of one character
            if (!is.character(region)){
                print_message("WARNING", c("Region must be provided as character. Allowed are specific ranges like 'A1:BY27' or",
										   "the names of named regions. The whole file will be read."))

                region <- NULL
            }

            if (length(region) != 1L){
                print_message("WARNING", "Only one character element allowed for region. The whole file will be read.")

                region <- NULL
            }
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Actual import
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # If valid region is provided
        if (!is.null(region)){
            # If it's a specific region
            if (grepl(":", region, fixed = TRUE)){
                data_frame <- openxlsx2::wb_to_df(file            = infile,
                                                  sheet           = sheet,
                                                  dims            = region,
                                                  col_names       = var_names,
                                                  skip_empty_rows = TRUE,
                                                  skip_empty_cols = TRUE,
                                                  na.strings      = "")
            }
            # If it's a named region
            else{
                # Try to grab named region if it exists
                data_frame <- tryCatch({
                    openxlsx2::wb_to_df(file            = infile,
                                        sheet           = sheet,
                                        named_region    = region,
                                        col_names       = var_names,
                                        skip_empty_rows = TRUE,
                                        skip_empty_cols = TRUE,
                                        na.strings      = "")
                }, error = function(e){
                    # Read whole file on error. In this case the named region didn't exist in the file
                    print_message("WARNING", c("Region '[region]' doesn't exist in sheet '[sheet]'.",
											   "The whole file will be read."), region = region, sheet = sheet)

                    openxlsx2::wb_to_df(file            = infile,
                                        sheet           = sheet,
                                        col_names       = var_names,
                                        skip_empty_rows = TRUE,
                                        skip_empty_cols = TRUE,
                                        na.strings      = "")
                })
            }
        }
        # Otherwise read whole file
        else{
            data_frame <- openxlsx2::wb_to_df(file            = infile,
                                              sheet           = sheet,
                                              col_names       = var_names,
                                              skip_empty_rows = TRUE,
                                              skip_empty_cols = TRUE,
                                              na.strings      = "")
        }
    }

    # Convert xlsx files to data.table
    if (!data.table::is.data.table(data_frame)){
        data_frame <- data.table::as.data.table(data_frame)
    }

    if (extension == "xlsx"){
        print_closing(5)
    }
    else{
        print_closing()
    }

    invisible(data_frame)
}


#' @description
#' [import_multi()]: Runs multiple imports on all provided files. Is able to import
#' all sheets from xlsx files.
#'
#' @param file_list [import_multi()]: A character vector containing full file paths.
#'
#' [export_multi()]: A list of data frames.
#'
#' @return
#' Multi functions: Returns a list of data frames.
#'
#' @examples
#' # Import multiple files at once
#' all_files <- import_multi(c(csv_file, txt_file, xlsx_file))
#'
#' @rdname import_export
#'
#' @export
import_multi <- function(file_list,
                         sheet     = "all",
                         region    = NULL,
                         separator = "auto",
                         decimal   = "auto",
                         var_names = TRUE){
    print_start_message()

    # Loop through all files and import them one after another
    result_list <- list()

    for (i in seq_along(file_list)){
        infile <- file_list[[i]]

        filename  <- tools::file_path_sans_ext(basename(infile))
        extension <- tolower(tools::file_ext(infile))

        # For CSV files just do a simple import
        if (extension %in% c("csv", "txt")){
            print_step("MAJOR", "Importing: [infile]", infile = infile)

            result_list[[filename]] <- suppressMessages(
                import_data(infile    = infile,
                            separator = separator,
                            decimal   = decimal,
                            var_names = var_names))
        }
        # For XLSX files there are two possible path to go
        else if (extension == "xlsx"){
            # If all sheets should be imported from a single file
            if (sheet == "all"){
                print_step("MAJOR", "Loading file: [infile]", infile = infile)

                # Load file as a workbook first to be able to extract the sheet names
                wb <- openxlsx2::wb_load(infile)
                sheet_names <- openxlsx2::wb_get_sheet_names(wb)

                print_step("MAJOR", "Importing:")

                # Import all sheets one after another by name
                for (sheet_name in sheet_names){
                    print_step("MINOR", "Sheet: [sheet]", sheet = sheet_name)

                    result_list[[paste0(filename, "_", sheet_name)]] <-
                        suppressMessages(
                            import_data(infile    = wb,
                                        sheet     = sheet_name,
                                        region    = region,
                                        var_names = var_names))
                }
            }
            # If only a single sheet from each file should be imported, just do a
            # simple import
            else{
                result_list[[filename]] <- suppressMessages(
                    import_data(infile    = infile,
                                sheet     = sheet,
                                region    = region,
                                var_names = var_names))
            }
        }
    }

    print_closing(10)

    invisible(result_list)
}


#' @description
#' [export_data()]: A wrapper for [data.table::fwrite()] and [openxlsx2::wb_save()],
#' providing basic export functionality with minimal code.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(100)
#'
#' # Example export file paths
#' export_csv  <- tempfile(fileext = ".csv")
#' export_txt  <- tempfile(fileext = ".txt")
#' export_xlsx <- tempfile(fileext = ".xlsx")
#'
#' # Export: Provide full file path
#' my_data |> export_data(export_csv)
#' my_data |> export_data(export_txt)
#' my_data |> export_data(export_xlsx)
#'
#' @rdname import_export
#'
#' @export
export_data <- function(data_frame,
                        outfile,
                        separator = ";",
                        decimal   = ",",
                        var_names = TRUE){
    print_start_message(suppress = TRUE)

    ###########################################################################
    # Error handling
    ###########################################################################

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Path
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Abort on vector provided as path
    if (!is.character(outfile) || length(outfile) != 1){
        print_message("ERROR", "<Outfile> must be a single character. Export will be aborted.")

        return(invisible(data_frame))
    }

    # Apply macros to outfile
    outfile <- macro(outfile)

    # Abort on invalid path
    if (!dir.exists(dirname(outfile)) || dirname(outfile) == "."){
        print_message("ERROR", c("Path does not exist: [outfile]",
								 "Export will be aborted."), outfile = outfile)

        return(invisible(data_frame))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # File extension
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    extension <- tolower(tools::file_ext(outfile))

    if (extension == ""){
        print_message("WARNING", "No file extension provided in <outfile>. 'csv' will be used.")

        outfile   <- paste0(outfile, ".csv")
        extension <- "csv"
    }

    if (!extension %in% c("csv", "txt", "xlsx")){
        print_message("WARNING", "Only 'csv', 'txt' or 'xlsx' are allowed as file extensions in the <outfile>. 'csv' will be used.")

        outfile <- sub(extension, "csv", outfile, ignore.case = TRUE)
        extension <- "csv"
    }

    if (extension %in% c("csv", "txt")){
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Separator
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Separator may only consist of one character
        if (!is.character(separator)){
            print_message("WARNING", "<Separator> must be provided as character. ';' will be used.")

            separator <- ";"
        }

        if (length(separator) != 1L || nchar(separator) != 1L){
            print_message("WARNING", "<Separator> may only be one character. ';' will be used.")

            separator <- ";"
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # Decimal
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        # Decimal may only consist of one character
        if (!is.character(decimal)){
            print_message("WARNING", "<Decimal> must be provided as character. ',' will be used.")

            decimal <- ","
        }

        if (length(decimal) != 1L || nchar(decimal) != 1L){
            print_message("WARNING", "<Decimal> may only be one character. ',' will be used.")

            decimal <- ","
        }

        # Decimal may not be equal to separator
        if (separator == decimal){
            print_message("WARNING", "<Decimal> may not be the same character as the <separator>. ',' will be used.")

            decimal <- ","
        }

    ###########################################################################
    # Export
    ###########################################################################

        data_frame |> data.table::fwrite(file = outfile,
                                         sep  = separator,
                                         dec  = decimal,
                                         col.names = var_names,
                                         bom  = TRUE)
    }
    # xlsx
    else if (extension == "xlsx"){
        wb <- openxlsx2::wb_workbook()
        wb$add_worksheet()
        wb$add_data(x          = data_frame,
                    col_names  = var_names,
                    na.strings = "")

        # Add named region so there is no need to specify the exact region when importing back
        wb$add_named_region(dims = openxlsx2::wb_dims(rows = seq.int(1, collapse::fnrow(data_frame) + var_names),
                                                      cols = seq.int(1, collapse::fncol(data_frame))),
                            name = "data", local_sheet = TRUE)

        wb$save(file = outfile, overwrite = TRUE)
    }

    if (extension == "xlsx"){
        print_closing(5)
    }
    else{
        print_closing()
    }

    invisible(data_frame)
}


#' @description
#' [export_multi()]: Runs multiple exports on a list of data frames. Is able to export
#' to a single xlsx file with multiple sheets.
#'
#' @param out_path The file path where to save the exported files.
#' @param into_sheets TRUE by default. Whether to export all data frames into multiple
#' sheets of a single xlsx file or into separate xlsx files.
#'
#' @examples
#' # Example data frame list
#' my_list <- list(first      = dummy_data(10),
#'                 second.txt = dummy_data(10))
#'
#' # Export multiple data frames into one xlsx file
#' # with multiple sheets
#' export_multi(my_list, tempdir())
#'
#' # Export multiple data frames into multiple xlsx files
#' export_multi(my_list, tempdir(), into_sheets = FALSE)
#'
#' # Export multiple data frames into multiple csv/txt files
#' export_multi(my_list, tempdir(), separator = ";")
#'
#' # Manual cleanup for example
#' file1 <- file.path(tempdir(), "my_list.xlsx")
#' file2 <- file.path(tempdir(), "first.xlsx")
#' file3 <- file.path(tempdir(), "second.xlsx")
#' file4 <- file.path(tempdir(), "first.csv")
#' file5 <- file.path(tempdir(), "second.txt")
#'
#' unlink(c(export_csv, export_xlsx,
#'          file1, file2, file3, file4, file5))
#'
#' @rdname import_export
#'
#' @export
export_multi <- function(file_list,
                         out_path,
                         into_sheets = TRUE,
                         separator   = NULL,
                         decimal     = ",",
                         var_names   = TRUE){
    print_start_message()

    if (into_sheets && is.null(separator)){
        wb <- openxlsx2::wb_workbook()
    }

    # Loop through all files and export them one after another
    for (i in seq_along(file_list)){
        data_frame <- file_list[[i]]
        filename   <- names(file_list)[[i]]

        print_step("MAJOR", "Exporting: [filename]", filename = filename)

        # For CSV files just do a simple export
        if (!is.null(separator)){
            extension <- tolower(tools::file_ext(filename))

            if (!extension %in% c("csv", "txt")){
                if (!extension == ""){
                    print_message("WARNING", "Only 'csv', 'txt' or 'xlsx' are allowed as file extensions. 'csv' will be used.")
                }

                filename <- paste0(filename, ".csv")
            }

            suppressMessages(export_data(data_frame,
                                         outfile   = paste0(out_path, "/", filename),
                                         separator = separator,
                                         decimal   = decimal,
                                         var_names = var_names))
        }
        # For XLSX files there are two possible path to go
        else{
            # If all data frames should be exported into the same file with multiple sheets
            if (into_sheets){
                # Fill all sheets one after another with data
                wb$add_worksheet(filename)
                wb$add_data(x          = data_frame,
                            col_names  = var_names,
                            na.strings = "")

                # Add named region so there is no need to specify the exact region when importing back
                wb$add_named_region(dims = openxlsx2::wb_dims(rows = seq.int(1, collapse::fnrow(data_frame) + var_names),
                                                              cols = seq.int(1, collapse::fncol(data_frame))),
                                    name = "data", local_sheet = TRUE)
            }
            # If only a single sheet from each file should be imported, just do a
            # simple import
            else{
                suppressMessages(export_data(data_frame,
                                             outfile   = paste0(out_path, "/", filename, ".xlsx"),
                                             separator = separator,
                                             decimal   = decimal,
                                             var_names = var_names))
            }
        }
    }

    if (into_sheets && is.null(separator)){
        filename <- get_origin_as_char(file_list, substitute(file_list))
        wb$save(file = paste0(out_path, "/", filename, ".xlsx"), overwrite = TRUE)
    }

    print_closing(10)

    invisible(file_list)
}

set_no_print(TRUE)

# Example csv, txt and xlsx files load
csv_file  <- system.file("extdata", "qol_example_data_csv.csv",   package = "qol")
txt_file  <- system.file("extdata", "qol_example_data_txt.txt",   package = "qol")
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")

expect_true(file.exists(csv_file), info = "Example csv, txt and xlsx files load")
expect_true(file.exists(txt_file), info = "Example csv, txt and xlsx files load")
expect_true(file.exists(xlsx_file), info = "Example csv, txt and xlsx files load")

###############################################################################
# Export
###############################################################################

export_df <- dummy_data(10)
export_list     <- list(first  = export_df,
                        second = export_df)
export_list_txt <- list(first.txt  = export_df,
                        second.txt = export_df)


# CSV export works correctly
temp_file <- tempfile(fileext = ".csv")
on.exit(unlink(temp_file), add = TRUE)

export_df |> export_data(temp_file)

expect_true(file.exists(temp_file), info = "CSV export works correctly")


# TXT export works correctly
temp_file <- tempfile(fileext = ".txt")
on.exit(unlink(temp_file), add = TRUE)

export_df |> export_data(temp_file)

expect_true(file.exists(temp_file), info = "TXT export works correctly")


# XLSX export works correctly
temp_file <- tempfile(fileext = ".xlsx")
on.exit(unlink(temp_file), add = TRUE)

export_df |> export_data(temp_file)

expect_true(file.exists(temp_file), info = "XLSX export works correctly")


# Abort export with invalid path
export_df |> export_data(1)

expect_error(print_stack_as_messages("ERROR"), "<Outfile> must be a single character. Export will be aborted.", info = "Abort export with invalid path")


# Abort export with non existing path
export_df |> export_data("Test")

expect_error(print_stack_as_messages("ERROR"), "Path does not exist:", info = "Abort export with non existing path")


# Warning in export on missing file extension
temp_file <- tempfile(fileext = "")
on.exit(unlink(temp_file), add = TRUE)

export_df |> export_data(temp_file)

expect_warning(print_stack_as_messages("WARNING"), "No file extension provided in <outfile>. 'csv' will be used.", info = "Warning in export on missing file extension")

expect_true(file.exists(paste0(temp_file, ".csv")), info = "Warning in export on missing file extension")


# Warning in export on invalid file extension
temp_file <- tempfile(fileext = ".test")
on.exit(unlink(temp_file), add = TRUE)

export_df |> export_data(temp_file)

expect_warning(print_stack_as_messages("WARNING"), "Only 'csv', 'txt' or 'xlsx' are allowed as file extensions in the <outfile>. 'csv' will be used.", info = "Warning in export on invalid file extension")

temp_file <- sub(".test", ".csv", temp_file, ignore.case = TRUE)
expect_true(file.exists(temp_file), info = "Warning in export on invalid file extension")


# Warning in export if separator is not provided as character
temp_file <- tempfile(fileext = ".csv")
on.exit(unlink(temp_file), add = TRUE)

export_df |> export_data(temp_file, separator = 1)

expect_warning(print_stack_as_messages("WARNING"), "<Separator> must be provided as character. ';' will be used.", info = "Warning in export if separator is not provided as character")

expect_true(file.exists(temp_file), info = "Warning in export if separator is not provided as character")


# Warning in export if separator is longer than one character
temp_file <- tempfile(fileext = ".csv")
on.exit(unlink(temp_file), add = TRUE)

export_df |> export_data(temp_file, separator = ";;")

expect_warning(print_stack_as_messages("WARNING"), "<Separator> may only be one character. ';' will be used.", info = "Warning in export if separator is longer than one character")

expect_true(file.exists(temp_file), info = "Warning in export if separator is longer than one character")


# Warning in export if decimal is not provided as character
temp_file <- tempfile(fileext = ".csv")
on.exit(unlink(temp_file), add = TRUE)

export_df |> export_data(temp_file, decimal = 1)

expect_warning(print_stack_as_messages("WARNING"), "<Decimal> must be provided as character. ',' will be used.", info = "Warning in export if decimal is not provided as character")

expect_true(file.exists(temp_file), info = "Warning in export if decimal is not provided as character")


# Warning in export if decimal is longer than one character
temp_file <- tempfile(fileext = ".csv")
on.exit(unlink(temp_file), add = TRUE)

export_df |> export_data(temp_file, decimal = ",,")

expect_warning(print_stack_as_messages("WARNING"), "<Decimal> may only be one character. ',' will be used.", info = "Warning in export if decimal is longer than one character")

expect_true(file.exists(temp_file), info = "Warning in export if decimal is longer than one character")


# Warning in export if separator and decimal are equal
temp_file <- tempfile(fileext = ".csv")
on.exit(unlink(temp_file), add = TRUE)

export_df |> export_data(temp_file, decimal = ";")

expect_warning(print_stack_as_messages("WARNING"), "<Decimal> may not be the same character as the <separator>. ',' will be used.", info = "Warning in export if separator and decimal are equal")

expect_true(file.exists(temp_file), info = "Warning in export if separator and decimal are equal")


# Multi CSV export works correctly
file1 <- file.path(tempdir(), "first.csv")
file2 <- file.path(tempdir(), "second.csv")
on.exit(unlink(c(file1, file2)), add = TRUE)

export_multi(export_list, tempdir(), separator = ";")

expect_true(file.exists(file1), info = "Multi CSV export works correctly")
expect_true(file.exists(file2), info = "Multi CSV export works correctly")


# Multi TXT export works correctly
file1 <- file.path(tempdir(), "first.txt")
file2 <- file.path(tempdir(), "second.txt")
on.exit(unlink(c(file1, file2)), add = TRUE)

export_multi(export_list_txt, tempdir(), separator = ";")

expect_true(file.exists(file1), info = "Multi TXT export works correctly")
expect_true(file.exists(file2), info = "Multi TXT export works correctly")


# Multi XLSX export works correctly
file1 <- file.path(tempdir(), "first.xlsx")
file2 <- file.path(tempdir(), "second.xlsx")
on.exit(unlink(c(file1, file2)), add = TRUE)

export_multi(export_list, tempdir(), into_sheets = FALSE)

expect_true(file.exists(file1), info = "Multi XLSX export works correctly")
expect_true(file.exists(file2), info = "Multi XLSX export works correctly")


# Multi sheet export into one XLSX file works correctly
file <- file.path(tempdir(), "export_list.xlsx")
on.exit(unlink(file), add = TRUE)

export_multi(export_list, tempdir())

expect_true(file.exists(file), info = "Multi sheet export into one XLSX file works correctly")

###############################################################################
# Import
###############################################################################


# CSV import works correctly
csv_file <- system.file("extdata", "qol_example_data_csv.csv", package = "qol")

infile <- import_data(csv_file)

expect_equal(nrow(infile), 100, info = "CSV import works correctly")
expect_equal(ncol(infile), 11, info = "CSV import works correctly")


# TXT import works correctly
txt_file <- system.file("extdata", "qol_example_data_txt.txt", package = "qol")

infile <- import_data(txt_file)

expect_equal(nrow(infile), 100, info = "TXT import works correctly")
expect_equal(ncol(infile), 11, info = "TXT import works correctly")


# CSV import without first row as column names
csv_file <- system.file("extdata", "qol_example_data_csv.csv", package = "qol")

infile <- import_data(csv_file, var_names = FALSE)

expect_equal(nrow(infile), 101, info = "CSV import without first row as column names")


# XLSX import works correctly
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")

infile <- import_data(xlsx_file)

expect_equal(nrow(infile), 100, info = "XLSX import works correctly")
expect_equal(ncol(infile), 11, info = "XLSX import works correctly")


# XLSX import with specific region
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")

infile <- import_data(xlsx_file, sheet = 2, region = "A1:C11")

expect_equal(nrow(infile), 10, info = "XLSX import with specific region")
expect_equal(ncol(infile), 3, info = "XLSX import with specific region")


# XLSX import with named region
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")

infile <- import_data(xlsx_file, region = "test_region")

expect_equal(nrow(infile), 100, info = "XLSX import with named region")
expect_equal(ncol(infile), 11, info = "XLSX import with named region")


# XLSX import without first row as column names
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")

infile <- import_data(xlsx_file, var_names = FALSE)

expect_equal(nrow(infile), 101, info = "XLSX import without first row as column names")


# Abort import with invalid path
import_data(1)

expect_error(print_stack_as_messages("ERROR"), "<Infile> must be a single character. Import will be aborted.", info = "Abort import with invalid path")


# Abort import with non existing path
import_data("Test")

expect_error(print_stack_as_messages("ERROR"), "Path does not exist:", info = "Abort import with non existing path")


# Abort import on missing file extension
csv_file <- gsub(".csv", "", system.file("extdata", "qol_example_data_csv.csv", package = "qol"))

infile <- import_data(csv_file)

expect_error(print_stack_as_messages("ERROR"), "No file extension provided in <infile>. 'csv' and 'xlsx' are allowed.", info = "Abort import on missing file extension")


# Abort import on invalid file extension
csv_file <- gsub(".csv", ".test", system.file("extdata", "qol_example_data_csv.csv", package = "qol"))

infile <- import_data(csv_file)

expect_error(print_stack_as_messages("ERROR"), "Only 'csv', 'txt' or 'xlsx' are allowed as file extensions in the <infile>.", info = "Abort import on invalid file extension")


# Warning in import if separator is not provided as character
csv_file <- system.file("extdata", "qol_example_data_csv.csv", package = "qol")

infile <- import_data(csv_file, separator = 1)

expect_warning(print_stack_as_messages("WARNING"), "<Separator> must be provided as character. Automatic detection will be used.", info = "Warning in import if separator is not provided as character")

expect_equal(nrow(infile), 100, info = "Warning in import if separator is not provided as character")
expect_equal(ncol(infile), 11, info = "Warning in import if separator is not provided as character")


# Warning in import if separator is longer than one character
csv_file <- system.file("extdata", "qol_example_data_csv.csv", package = "qol")

infile <- import_data(csv_file, separator = ";;")

expect_warning(print_stack_as_messages("WARNING"), "<Separator> may only be one character. Automatic detection will be used.", info = "Warning in import if separator is longer than one character")

expect_equal(nrow(infile), 100, info = "Warning in import if separator is longer than one character")
expect_equal(ncol(infile), 11, info = "Warning in import if separator is longer than one character")


# Warning in import if decimal is not provided as character
csv_file <- system.file("extdata", "qol_example_data_csv.csv", package = "qol")

infile <- import_data(csv_file, decimal = 1)

expect_warning(print_stack_as_messages("WARNING"), "<Decimal> must be provided as character. Automatic detection will be used.", info = "Warning in import if decimal is not provided as character")

expect_equal(nrow(infile), 100, info = "Warning in import if decimal is not provided as character")
expect_equal(ncol(infile), 11, info = "Warning in import if decimal is not provided as character")


# Warning in import if decimal is longer than one character
csv_file <- system.file("extdata", "qol_example_data_csv.csv", package = "qol")

infile <- import_data(csv_file, decimal = ",,")

expect_warning(print_stack_as_messages("WARNING"), "<Decimal> may only be one character. Automatic detection will be used.", info = "Warning in import if decimal is longer than one character")

expect_equal(nrow(infile), 100, info = "Warning in import if decimal is longer than one character")
expect_equal(ncol(infile), 11, info = "Warning in import if decimal is longer than one character")


# Warning in import if separator and decimal are equal
csv_file <- system.file("extdata", "qol_example_data_csv.csv", package = "qol")

infile <- import_data(csv_file, separator = ";", decimal = ";")

expect_warning(print_stack_as_messages("WARNING"), "<Decimal> may not be the same character as the <separator>. Automatic detection will be used.", info = "Warning in import if separator and decimal are equal")

expect_equal(nrow(infile), 100, info = "Warning in import if separator and decimal are equal")
expect_equal(ncol(infile), 11, info = "Warning in import if separator and decimal are equal")


# Warning in import if region is not provided as character
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")

infile <- import_data(xlsx_file, region = 1)

expect_warning(print_stack_as_messages("WARNING"), "Region must be provided as character. Allowed are specific ranges like 'A1:BY27' or", info = "Warning in import if region is not provided as character")

expect_equal(nrow(infile), 100, info = "Warning in import if region is not provided as character")
expect_equal(ncol(infile), 11, info = "Warning in import if region is not provided as character")


# Warning in import if region is a vector
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")

infile <- import_data(xlsx_file, region = c("A", "B"))

expect_warning(print_stack_as_messages("WARNING"), "Only one character element allowed for region. The whole file will be read.", info = "Warning in import if region is a vector")

expect_equal(nrow(infile), 100, info = "Warning in import if region is a vector")
expect_equal(ncol(infile), 11, info = "Warning in import if region is a vector")


# Warning in import if region not found
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")

infile <- import_data(xlsx_file, region = "Test")

expect_warning(print_stack_as_messages("WARNING"), "Region 'Test' doesn't exist in sheet", info = "Warning in import if region not found")

expect_equal(nrow(infile), 100, info = "Warning in import if region not found")
expect_equal(ncol(infile), 11, info = "Warning in import if region not found")


# Multi file import works correctly (single sheet)
csv_file  <- system.file("extdata", "qol_example_data_csv.csv",   package = "qol")
txt_file  <- system.file("extdata", "qol_example_data_txt.txt",   package = "qol")
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")
files <- c(csv_file, txt_file, xlsx_file)

infile <- import_multi(files, 1)

expect_inherits(infile, "list", info = "Multi file import works correctly (single sheet)")
expect_equal(length(infile), 3, info = "Multi file import works correctly (single sheet)")


# Multi file import works correctly (all sheets)
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")

infile <- import_multi(xlsx_file)

expect_inherits(infile, "list", info = "Multi file import works correctly (all sheets)")
expect_equal(length(infile), 2, info = "Multi file import works correctly (all sheets)")


# Multi file import works correctly (selected sheets)
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")

infile <- import_multi(xlsx_file, sheet = c(1, 2))

expect_inherits(infile, "list", info = "Multi file import works correctly (selected sheets)")
expect_equal(length(infile), 2, info = "Multi file import works correctly (selected sheets)")


# Multi file import can stack data
csv_file  <- system.file("extdata", "qol_example_data_csv.csv",   package = "qol")
txt_file  <- system.file("extdata", "qol_example_data_txt.txt",   package = "qol")
files     <- c(csv_file, txt_file)

csv_import <- import_data(csv_file)
txt_import <- import_data(txt_file)
all_import <- import_multi(files, stack_files = TRUE)

expect_equal(collapse::fnrow(csv_import) + collapse::fnrow(txt_import),
             collapse::fnrow(all_import), info = "Multi file import can stack data")


set_no_print()

###############################################################################
# Globals
###############################################################################

# Variables entered here won't appear in Notes during devtools::check().
# The from - to variables can't be handled otherwise, because they
# are used in a data.table context which doesn't support the rlang .data
# keyword.
utils::globalVariables(c("qol_ID", "qol_from", "qol_to", "from", "to", "delta",
                         "across", ":=", "TYPE", "TYPE_NR", "DEPTH", ".subheader",
                         "var_sum", "var_cum_sum", "pct_group", "var_cum_pct",
                         "mean", "sd", "min", "max", "sum_wgt", ".temp_key", ".temp_weight",
                         "var_pct_row", ".pseudo_preserve", "BY", "by_vars", "VALUE",
                         "first_person", "age", "age_factor", "income", "income_factor",
                         "expenses", "expenses_factor", "file", "result", "TYPE_ORIG",
                         "total", "value", "pseudo_group"))

###############################################################################
# Start up
###############################################################################

# Start up message
.onAttach <- function(libname, pkgname){
    current_version <- utils::packageVersion("qol")

    print_headline("[#63C2C9 [b]Quality Of Life][/b]")
    print_message("NEUTRAL", c("\n The [#32CD32 [b]qol-package[/b]] brings powerful concepts from [#63C2C9 [b]SAS[/b]] to [#C93F3F [b]R[/b]] to make life easier",
                               "and produce bigger and more complex outputs in less time with less code.",
                               "",
                               "> Use [#32CD32 ?qol] to get an overview.",
                               "> To view the changelog type: [#32CD32 qol_news()]",
                               "> Chat with the repository:   [#32CD32 qol_chat()]"))
    print_headline("[#63C2C9 [b]Current version:[/b]] [b][version][/b]", version = current_version)
}

###############################################################################
# General global options
###############################################################################

# Internal environment to store global options
.qol_options <- new.env(parent = emptyenv())
.qol_options[["excel_style"]] <- excel_output_style()
.qol_options[["var_labels"]]  <- list()
.qol_options[["stat_labels"]] <- list()
.qol_options[["print"]]       <- TRUE
.qol_options[["monitor"]]     <- FALSE
.qol_options[["na.rm"]]       <- FALSE
.qol_options[["print_miss"]]  <- FALSE
.qol_options[["output"]]      <- "console"
.qol_options[["threads"]]     <- suppressMessages(fst::threads_fst())

###############################################################################
# Message system
###############################################################################

# Internal environment to store message styling options
.qol_messages <- new.env(parent = emptyenv())
.qol_messages[["format"]] <- list(utf8                = l10n_info()[["UTF-8"]] && .Platform$GUI != "Rgui",
                                  time_color          = "#6B6B6B",
                                  note_ansi           = hex_to_ansi(" \u2139\ufe0f NOTE: ", hex_color = "#63C2C9", bold = TRUE),
                                  note_pt             = hex_to_ansi(" ~ NOTE: ",            hex_color = "#63C2C9", bold = TRUE),
                                  note_indent_ansi    = hex_to_ansi(" \u2139\ufe0f\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0", hex_color = "#63C2C9", bold = TRUE),
                                  note_indent_pt      = hex_to_ansi(" ~\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0",            hex_color = "#63C2C9", bold = TRUE),
                                  warning_ansi        = hex_to_ansi(" \u26a0\ufe0f WARNING: ", hex_color = "#FFC90E", bold = TRUE),
                                  warning_pt          = hex_to_ansi(" ! WARNING: ",            hex_color = "#FFC90E", bold = TRUE),
                                  warning_indent_ansi = hex_to_ansi(" \u26a0\ufe0f\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0", hex_color = "#FFC90E", bold = TRUE),
                                  warning_indent_pt   = hex_to_ansi(" !\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0",            hex_color = "#FFC90E", bold = TRUE),
                                  error_ansi          = hex_to_ansi(" \u274c ERROR: ", hex_color = "#C93F3F", bold = TRUE),
                                  error_pt            = hex_to_ansi(" X ERROR: ",      hex_color = "#C93F3F", bold = TRUE),
                                  error_indent_ansi   = hex_to_ansi(" \u274c\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0", hex_color = "#C93F3F", bold = TRUE),
                                  error_indent_pt     = hex_to_ansi(" X\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0\u00a0",      hex_color = "#C93F3F", bold = TRUE),
                                  neutral_ansi        = hex_to_ansi("\u00a0", hex_color = "#63C2C9", bold = TRUE),
                                  neutral_pt          = hex_to_ansi("\u00a0", hex_color = "#63C2C9", bold = TRUE),
                                  neutral_indent_ansi = hex_to_ansi("\u00a0", hex_color = "#63C2C9", bold = TRUE),
                                  neutral_indent_pt   = hex_to_ansi("\u00a0", hex_color = "#63C2C9", bold = TRUE),
                                  grey_ansi           = hex_to_ansi(" \u2601 ",                 hex_color = "#6B6B6B", bold = TRUE),
                                  grey_pt             = hex_to_ansi(" * ",                      hex_color = "#6B6B6B", bold = TRUE),
                                  grey_indent_ansi    = hex_to_ansi("\u00a0\u00a0\u00a0\u00a0", hex_color = "#6B6B6B", bold = TRUE),
                                  grey_indent_pt      = hex_to_ansi("\u00a0\u00a0\u00a0",       hex_color = "#6B6B6B", bold = TRUE),
                                  major_ansi          = hex_to_ansi(" \u23f3\ufe0f ",           hex_color = "#32CD32", bold = TRUE),
                                  major_pt            = hex_to_ansi(" > ",                      hex_color = "#32CD32", bold = TRUE),
                                  major_indent_ansi   = hex_to_ansi("\u00a0\u00a0\u00a0\u00a0", hex_color = "#32CD32", bold = TRUE),
                                  major_indent_pt     = hex_to_ansi("\u00a0\u00a0\u00a0",       hex_color = "#32CD32", bold = TRUE),
                                  minor_ansi          = hex_to_ansi(" \u00a0\u00a0\u00a0\u23f3\ufe0f ", hex_color = "#32CD32", bold = TRUE),
                                  minor_pt            = hex_to_ansi(" \u00a0\u00a0+ ",                  hex_color = "#32CD32", bold = TRUE),
                                  minor_indent_ansi   = hex_to_ansi("\u00a0\u00a0\u00a0\u00a0\u00a0",   hex_color = "#32CD32", bold = TRUE),
                                  minor_indent_pt     = hex_to_ansi("\u00a0\u00a0\u00a0",               hex_color = "#32CD32", bold = TRUE))
.qol_messages[["custom_types"]]   <- list()
.qol_messages[["stack"]]          <- list()
.qol_messages[["start_time"]]     <- NULL
.qol_messages[["timer"]]          <- NULL
.qol_messages[["last_message"]]   <- NULL
.qol_messages[["last_session"]]   <- NULL
.qol_messages[["last_execution"]] <- NULL
.qol_messages[["no_print"]]       <- FALSE
.qol_messages[["no_color"]]       <- nzchar(Sys.getenv("NO_COLOR"))

#' Build a Theme From Scratch
#'
#' @description
#' Build your own theme by just setting up the colors for the different parts of
#' RStudio. A theme file will be exported which can be added by going to:
#'
#' Tools -> Global Options -> Appearance -> Add
#'
#' @param file_path The path to which the theme file should be saved.
#' @param theme_name The themes name.
#' @param dark_theme Handles some elements not covered with the other parameters.
#' @param editor_background Base background color in the editor.
#' @param editor_headline Mostly used for the headlines of the environment panel.
#' @param editor_font Base font color of the editor.
#' @param toolbar Base toolbar and frame color.
#' @param tab Color of inactive tabs.
#' @param selected_tab Color of active tabs.
#' @param line_number The color of the line numbers on the left.
#' @param print_margin Color of the vertical line showing the print margin.
#' @param cursor Cursor color.
#' @param selection The background color of the current selection.
#' @param smart_highlight Background color of smart highlighted words.
#' @param bracket_highlight Background color of highlighted bracket pairs.
#' @param active_line Color for the active line the cursor is in.
#' @param whitespace Color for whitespace characters.
#' @param debug_line Color of the current debug line.
#' @param scrollbar Color of the scrollbars.
#' @param scrollbar_hover Highlight color when hovering over a scrollbar.
#' @param scrollbar_active Highlight color when clicking on a scrollbar.
#' @param class_name Code color for class names (like package names).
#' @param keyword Code color for fixed keywords (like function, if, else).
#' @param language_constant Code color for language constants (like the @ keywords).
#' @param function_name Code color for base and package functions.
#' @param numeric Code color for numeric values.
#' @param string Code color for string values.
#' @param regex Code color for regex expressions.
#' @param variable Code color for variables, parameters and arguments.
#' @param comment Code color for comments.
#' @param symbol Code Color of symbols (like <-, brackets).
#' @param console_code Color of executed Code in the Console.
#' @param markdown_code Background color of code passages in a markdown file.
#'
#' @details
#' In the 'SAS Enterprise Guide' the user is able to not only choose a given theme, but
#' to also pick the colors for the different parts of the editor by themselves. Everyone
#' has a different taste of what colors look pleasing to the eyes, so you should be able
#' to choose them by yourself.
#'
#' @return
#' Saves a complete theme file.
#'
#' @examples
#' # Example export file paths
#' # NOTE: These tempfiles are only for the examples. In reality you just call the
#' # main function and put in your desired path and name directly.
#' temp_file <- tempfile(fileext = ".rstheme")
#' file_name <- basename(tools::file_path_sans_ext(temp_file))
#'
#' # Example theme
#' build_rstheme(file_path         = dirname(temp_file),
#'               theme_name        = file_name,
#'               editor_background = "#417291",
#'               editor_headline   = "#602BCA",
#'               editor_font       = "#C75C48")
#'
#' # Manual cleanup for example
#' unlink(temp_file)
#'
#' @export
build_rstheme <- function(file_path,
                          theme_name        = "qol_green",
                          dark_theme        = TRUE,
                          editor_background = "#062625",
                          editor_headline   = "#3B3B3B",
                          editor_font       = "#C3B79D",
                          toolbar           = "#2E2E2E",
                          tab               = "#3B3B3B",
                          selected_tab      = "#062625",
                          line_number       = "#C3B79D",
                          print_margin      = "#3B3B3B",
                          cursor            = "#CCCCCC",
                          selection         = "#1B436E",
                          smart_highlight   = "#3686dc",
                          bracket_highlight = "#595959",
                          active_line       = "#202324",
                          whitespace        = "#CCCCCC",
                          debug_line        = "#F18889",
                          scrollbar         = "#3B3B3B",
                          scrollbar_hover   = "#595959",
                          scrollbar_active  = "#BFBFBF",
                          class_name        = "#BEDD1A",
                          keyword           = "#FFC90E",
                          language_constant = "#FFC90E",
                          function_name     = "#C3B79D",
                          numeric           = "#C93F3F",
                          string            = "#63C2C9",
                          regex             = "#E8E6E3",
                          variable          = "#E8E6E3",
                          comment           = "#32CD32",
                          symbol            = "#C3B79D",
                          console_code      = "#C3B79D",
                          markdown_code     = "#083332"){
    # Measure the time
    print_start_message()

    # Apply macros to file path
    file_path <- macro(file_path)

    # Check if folder exists; ... is for testing
    if (!dir.exists(file_path) || dirname(file_path) == "."){
        print_message("ERROR", "Directory '[file_path]' does not exist.", file_path = file_path)
        return(invisible(NULL))
    }

    # Change main body to light theme if specified
    if (dark_theme == FALSE){
        rstheme_body <- gsub(".rstudio-themes-dark", ".rstudio-themes-flat", rstheme_body)
        rstheme_body <- gsub(".rstudio-themes-flat > .rstudio-themes-flat",
                             ".rstudio-themes-flat > .rstudio-themes-default", rstheme_body)
    }

    # Put together the complete theme file
    rstheme <- c(paste0("/* rs-theme-name: ", theme_name, " */"),
                 paste0("/* rs-theme-is-dark: ", dark_theme, " */"),
                 ":root {",
                 "  /* editor */",
                 paste0("  --editor-background: ", editor_background, ";"),
                 paste0("  --editor-headline: ", editor_headline, ";"),
                 paste0("  --editor-font: ", editor_font, ";"),
                 "",
                 paste0("  --toolbar: ", toolbar, ";"),
                 paste0("  --tab: ", tab, ";"),
                 paste0("  --selected-tab: ", selected_tab, ";"),
                 "",
                 paste0("  --line-number: ", line_number, ";"),
                 paste0("  --print_margin: ", print_margin, ";"),
                 "",
                 paste0("  --cursor: ", cursor, ";"),
                 paste0("  --selection: ", selection, ";"),
                 paste0("  --smart-highlight: ", smart_highlight, ";"),
                 paste0("  --bracket-highlight: ", bracket_highlight, ";"),
                 paste0("  --active-line: ", active_line, ";"),
                 paste0("  --whitespace: ", whitespace, ";"),
                 paste0("  --debug-line: ", debug_line, ";"),
                 "",
                 paste0("  --scrollbar: ", scrollbar, ";"),
                 paste0("  --scrollbar-hover: ", scrollbar_hover, ";"),
                 paste0("  --scrollbar-active: ", scrollbar_active, ";"),
                 "",
                 "  /* syntax */",
                 paste0("  --class-name: ", class_name, ";"),
                 paste0("  --keyword: ", keyword, ";"),
                 paste0("  --language-constant: ", language_constant, ";"),
                 paste0("  --function-declaration: ", function_name, ";"),
                 paste0("  --numeric-constant: ", numeric, ";"),
                 paste0("  --string: ", string, ";"),
                 paste0("  --regex: ", regex, ";"),
                 paste0("  --variable: ", variable, ";"),
                 paste0("  --comment: ", comment, ";"),
                 paste0("  --symbols: ", symbol, ";"),
                 "",
                 "  /* console */",
                 paste0("  --console-code: ", console_code, ";"),
                 "",
                 "  /* markdown */",
                 paste0("  --code-line: ", markdown_code, ";"),
                 "",
                 "  /* ??? */",
                 paste0("  --invalid-font: #ced2cf;"),
                 paste0("  --invalid-background: #b798bf;"),
                 "}",
                 rstheme_body,
                 generate_xterm_color_block())

    # Save file
    path <- ifelse(grepl("/$", file_path), file_path, paste0(file_path, "/"))

    writeLines(rstheme, paste0(path, theme_name, ".rstheme"))

    print_closing()

    invisible(rstheme)
}


#' Auto Generate Last Block of Theme File
#'
#' @description
#' Build the rather large last block of color code values automatically.
#'
#' @return
#' Returns the fully put together last block of a theme file as character vector.
#'
#' @noRd
generate_xterm_color_block <- function(){
    # Generate full 256-color xterm CSS file
    # Define the 16 base ANSI colors
    base_colors <- c("#000000", "#800000", "#008000", "#808000",
                     "#000080", "#800080", "#008080", "#c0c0c0",
                     "#808080", "#ff0000", "#00ff00", "#ffff00",
                     "#0000ff", "#ff00ff", "#00ffff", "#ffffff")

    # Build the 6x6x6 color cube (indices 16-231)
    cube_levels <- c(0, 95, 135, 175, 215, 255)
    cube_colors <- c()

    for (r in cube_levels){
        for (g in cube_levels){
            for (b in cube_levels){
                cube_colors <- c(cube_colors, sprintf("#%02x%02x%02x", r, g, b))
            }
        }
    }

    # Build grayscale ramp (indices 232-255)
    gray_levels <- seq(8, 238, by = 10)
    gray_colors <- sprintf("#%02x%02x%02x", gray_levels, gray_levels, gray_levels)

    # Combine all 256 colors
    xterm_palette <- c(base_colors, cube_colors, gray_colors)

    # Generate CSS lines for each color
    unlist(lapply(seq_along(xterm_palette) - 1, function(i){
        color <- xterm_palette[i + 1]

        c(sprintf(".xtermColor%d { color: %s !important; }", i, color),
          sprintf(".xtermBgColor%d { background-color: %s; }", i, color))
    }))
}


# The main body which remains unchanged because the colors are saved in global variables.
rstheme_body <- '.ace_gutter {
  color: var(--line-number);
}

.ace_print-margin {
  width: 2px;
  box-shadow: 1px 0 0 0 var(--print_margin) inset;
}

.ace_editor,
.rstudio-themes-flat.ace_editor_theme .profvis-flamegraph,
.rstudio-themes-flat.ace_editor_theme,
.rstudio-themes-flat .ace_editor_theme {
  background-color: var(--editor-background);
  color: var(--editor-font);
  line-height: 1.3 !important;
}

.ace_cursor {
  border-color: var(--cursor);
  width: 2px;
}

.ace_marker-layer .ace_selection {
  background-color: var(--selection);
}

.ace_selection.ace_start {
  box-shadow: 0 0 3px 0px var(--selection);
  border-radius: 2px;
}

.ace_marker-layer .ace_step {
  background: var(--selection);
}

.ace_marker-layer .ace_bracket {
  outline: 1px solid var(--bracket-highlight);
  background-color: var(--bracket-highlight);
}

.ace_marker-layer .ace_active-line {
  outline: 2px solid var(--active-line);
  background-color: var(--active-line);
}

.ace_gutter-active-line {
  background-color: var(--active-line);
}

.ace_marker-layer .ace_selected-word {
  background-color: var(--smart-highlight);
}

.ace_invisible {
  color: var(--whitespace);
}

.ace_operator{
  color: var(--symbols);
}

.ace_keyword:not(.ace_operator),
.ace_meta,
.ace_storage,
.ace_storage.ace_type,
.ace_support.ace_type {
  color: var(--keyword);
}

.ace_constant.ace_character,
.ace_keyword.ace_other.ace_unit,
.ace_support.ace_constant,
.ace_variable.ace_parameter {
  color: var(--string);
}

.ace_constant.ace_language,
.ace_keyword.ace_virtual-comment {
  color: var(--language-constant);
}

.ace_constant.ace_numeric,
.ace_constant.ace_other {
  color: var(--numeric-constant);
}

.ace_invalid {
  color: var(--invalid-font);
  background-color: var(invalid-background);
}

.ace_invalid.ace_deprecated {
  color: var(--invalid-font);
  background-color: var(invalid-background);
}

.ace_gutter-cell:not([style*="top"]) {
  position: relative;
}

.ace_line .ace_fold {
  background: none;
  height: inherit;
  margin: 0;
  border: none;
  position: relative;
}

.ace_line .ace_fold::before {
  color: grey;
  content: "...";
  position: absolute;
  left: 50%;
  transform: translateX(-50%);
}

.ace_gutter .ace_fold-widget {
  cursor: pointer;
  background: no-repeat border-box transparent;
  border: none;
  border-radius: 0;
  box-shadow: none;
  width: 17px;
  position: absolute;
  top: 50%;
  right: 0;
  transform: translate(-50%, -50%);
}

.ace_gutter .ace_fold-widget.ace_closed {
  background-image: url("data:image/svg+xml;charset=utf-8,%3Csvg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 15 15"%3E%3Cpath opacity=".1" fill="%23fff" d="M3 3h9v9H3z"/%3E%3Cpath d="M11 4v7H4V4h7m1-1H3v9h9V3z" fill="%235a5a5a"/%3E%3Cpath fill="none" stroke="%23c5c5c5" stroke-miterlimit="10" d="M10 7.5H5M7.5 5v5"/%3E%3C/svg%3E");
}

.ace_gutter .ace_fold-widget.ace_open {
  background-image: url("data:image/svg+xml;charset=utf-8,%3Csvg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 15 15"%3E%3Cpath d="M11 4v7H4V4h7m1-1H3v9h9V3z" fill="%235a5a5a"/%3E%3Cpath fill="none" stroke="%23c5c5c5" stroke-miterlimit="10" d="M10 7.5H5"/%3E%3C/svg%3E");
  opacity: 0;
  transition: opacity 0.5s;
}

.ace_gutter:hover .ace_fold-widget.ace_open {
  opacity: 1;
}

.ace_entity.ace_name.ace_function,
.ace_identifier.ace_support.ace_function:not(.ace_virtual-comment) {
  color: var(--function-declaration);
}

.ace_identifier,
.ace_variable {
  color: var(--variable);
}

.ace_support.ace_class,
.ace_support.ace_type {
  color: var(--class-name);
}

.ace_heading,
.ace_markup.ace_heading {
  color: var(--language-constant);
}

.ace_markup.ace_heading .ace_entity.ace_name.ace_tag,
.ace_entity.ace_other.ace_attribute-name,
.ace_meta.ace_tag,
.ace_string {
  color: var(--string);
}

.ace_string.ace_regexp {
  color: var(--regex);
}

.ace_comment {
  color: var(--comment);
}

/* rmarkdown code */
[class="ace_support ace_function"] {
  color: var(--string);
}

/* rmarkdown emphasis */
.ace_constant.ace_language.ace_boolean.ace_text {
  font-style: italic;
}

/* rmarkdown strong emphasis */
.ace_constant.ace_numeric.ace_text {
  font-weight: bold;
}

.ace_indent-guide {
  box-shadow: 1px 0 0 0 var(--whitespace);
  display: inline-block;
}

.nocolor.ace_editor .ace_line span {
  color: var(--string);
}

.ace_marker-layer .ace_foreign_line {
  position: absolute;
  z-index: -1;
  background-color: var(--code-line);
}

.ace_marker-layer .ace_active_debug_line {
  position: absolute;
  width: 100% !important;
  z-index: -1;
  background-color: var(--debug-line);
}

.ace_marker-layer .ace_find_line {
  position: absolute;
  z-index: -1;
  background-color: var(--smart-highlight);
}lsl

.ace_console_error {
  color: var(--invalid-font);
  background-color: var(--invalid-background);
  font-weight: bold;
}

#rstudio_console_output .ace_keyword {
  color: var(--console-code);
  font-weight: normal;
}

#rstudio_console_output .ace_constant.ace_language {
  color: var(--language-constant);
}

.rstudio-themes-flat.editor_dark.ace_editor_theme a {
  color: var(--editor-font) !important;
}

.ace_layer {
  z-index: 3;
}

.ace_layer.ace_print-margin-layer {
  z-index: 2;
}

.ace_layer.ace_marker-layer {
  z-index: 1;
}

.rstudio-themes-flat.rstudio-themes-dark-menus .ace_editor.ace_autocomplete {
  background: var(--tab);
  border: solid 1px var(--toolbar) !important;
  color: var(--editor-font);
}

.rstudio-themes-flat.rstudio-themes-dark-menus
  .ace_editor.ace_autocomplete
  .ace_marker-layer
  .ace_active-line,
.rstudio-themes-flat.rstudio-themes-dark-menus
  .ace_editor.ace_autocomplete
  .ace_marker-layer
  .ace_line-hover {
  background: var(--editor-background);
  border: none;
}

/* positioning (exclude rstudio server) */
.rstudio-themes-flat > .rstudio-themes-dark div:last-child[style*="top: 29px"] {
  top: 24px !important;
  right: 0 !important;
  bottom: 0 !important;
  left: 0 !important;
}

.rstudio-themes-flat > .rstudio-themes-dark div:last-child[style*="top: 5px"] {
  top: 0 !important;
  right: 0 !important;
  bottom: 0 !important;
  left: 0 !important;
}

.rstudio-themes-dark
  :-webkit-any(div:last-child[style*="top: 5px"], div:last-child[style*="top: 29px"])
  .gwt-SplitLayoutPanel-Workbench {
  top: 0 !important;
  right: 0 !important;
  bottom: 0 !important;
  left: 0 !important;
}

/* background */
.rstudio-themes-flat > .rstudio-themes-dark,
.rstudio-themes-flat
  > .rstudio-themes-dark
  :-webkit-any(.gwt-TabLayoutPanel, .gwt-TabLayoutPanelContent) {
  background: var(--toolbar) !important;
}

/* splitters */
.rstudio-themes-flat > .rstudio-themes-dark .gwt-SplitLayoutPanel-HDragger {
  background: var(--toolbar) !important;
  cursor: ew-resize;
  border-color: var(--toolbar);
}

.rstudio-themes-flat > .rstudio-themes-dark .gwt-SplitLayoutPanel-VDragger {
  background: var(--toolbar) !important;
  cursor: ns-resize;
  border-color: var(--toolbar);
  /* box-shadow: 0 -1px 0 var(--editor-background) inset; */
}

/* window containers */
.rstudio-themes-flat > .rstudio-themes-dark
  :-webkit-any(.windowframe, .rstheme_minimizedWindowObject)
  > div:last-child {
  border-radius: 0;
  border: none !important;
}

/* main toolbar */
.rstudio-themes-dark > div:last-child > div > div > .rstheme_toolbarWrapper,
.rstudio-themes-dark [role="application"] > div > .rstheme_toolbarWrapper {
  background-color: var(--toolbar) !important;
  border-color: var(--toolbar) !important;
}

/* toolbars */
.rstudio-themes-dark .rstheme_toolbarWrapper,
.rstudio-themes-dark .rstheme_secondaryToolbar {
  background-color: var(--editor-background) !important;
  border-color: var(--editor-background) !important;
}

/* other toolbars */
.rstudio-themes-dark .rstudio-themes-background {
  background-color: var(--editor-headline) !important;
  border-top-color: var(--editor-headline) !important;
  border-bottom-color: var(--editor-headline) !important;
  border-radius: 0;
}

/* git status toolbar */
.rstudio-themes-flat .rstudio-themes-dark .GD15MCFCB2C,
.rstudio-themes-flat .rstudio-themes-dark .GGBOEFPDC4C {
  background-color: var(--editor-headline) !important;
}

/* files breadcrumb gradient */
.rstudio-themes-dark .breadcrumb > div:last-child {
  background: var(--editor-background) !important;
  left: inherit !important;
}

/* tabs container */
.rstudio-themes-dark .gwt-TabLayoutPanelTabs,
/* windows */
.rstudio-themes-dark .windowframe,
/* remove gradients */
.rstudio-themes-dark :-webkit-any(.rstheme_multiPodUtilityTabArea, .GD15MCFCCS, .GJQ3LUQCCS, .GGBOEFPDDS) {
  background: var(--toolbar) !important;
  border-color: var(--toolbar) !important;
}

/* tabs */
.rstudio-themes-dark
  .gwt-TabLayoutPanelTab:not(.gwt-TabLayoutPanelTab-selected)
  .rstheme_tabLayoutCenter,
.rstudio-themes-dark .rstheme_minimizedWindowObject .rstheme_center {
  background-color: var(--tab) !important;
  color: var(--editor-font) !important;
  border-color: var(--tab) !important;
  border-right: 1px solid var(--toolbar) !important;
  border-radius: 0;
  cursor: pointer;
}

.rstudio-themes-dark .gwt-TabLayoutPanelTab .gwt-Label {
  cursor: pointer;
}

/* active tab */
.rstudio-themes-dark
  .gwt-TabLayoutPanelTab.gwt-TabLayoutPanelTab-selected
  .rstheme_tabLayoutCenter {
  background-color: var(--selected-tab) !important;
  color: white !important;
}

/* editor */
#rstudio_shell_widget.ace_editor.ace_scroller,
.ace_editor[id^="rstudio_source_text_editor"]:not(#rstudio_source_text_editor_0)
  :-webkit-any(.ace_gutter, .ace_scroller, .ace_scrollbar-v) {
  box-shadow: inset 0 6px 6px -6px black;
}

/* datagrid */
.rstudio-themes-dark .dataGridHeader,
.rstudio-themes-dark tr[__gwt_header_row] > :-webkit-any(td, th),
.rstudio-themes-dark .dataTables_info {
  background-color: var(--editor-headline) !important;
}

/* data table */
.rstudio-themes-dark .dataTable :-webkit-any(th, .first-child) {
  background-color: var(--editor-headline) !important;
}

.rstudio-themes-dark .dataTable :-webkit-any(th, td, tr) {
  border-color: silver !important;
}

.rstudio-themes-dark .dataTable tbody > tr:hover > td {
  background-color: var(--editor-headline) !important;
}

/* search background */
.rstudio-themes-dark
  .gwt-TabLayoutPanelContent
  > div:nth-last-child(2):nth-child(4)
  .rstudio-themes-background {
  background-color: var(--toolbar) !important;
}

/* search input */
.rstudio-themes-dark
  :-webkit-any(.rstheme_toolbarWrapper, #rstudio_find_replace_bar)
  .search,
.rstudio-themes-flat .themedPopupPanel .search {
  background-color: var(--editor-headline) !important;
  color: var(--editor-font) !important;
  border-radius: 0 !important;
  border: 1px solid transparent;
}

.rstudio-themes-dark
  .gwt-TabLayoutPanelContent
  > div:nth-last-child(2):nth-child(4)
  .rstudio-themes-background
  .search {
  height: 18px;
}

.rstudio-themes-dark .search:-webkit-any(:focus, :focus-within) {
  outline: 1px solid var(--editor-font);
  outline-offset: -1px;
}

/* scrollbars */
::-webkit-scrollbar {
  height: calc(17px * 10 / 14) !important;
}
::-webkit-scrollbar-thumb {
  border: none !important;
  border-radius: 0 !important;
  background-color: var(--scrollbar) !important;
}

::-webkit-scrollbar-thumb:hover {
  background-color: var(--scrollbar-hover) !important;
}

::-webkit-scrollbar-thumb:active {
  background-color: var(--scrollbar-active) !important;
}

::-webkit-scrollbar,
::-webkit-scrollbar-corner,
::-webkit-scrollbar-track {
  background-color: transparent !important;
}

::-webkit-scrollbar-track:vertical {
  box-shadow: inset 1px 0 0 0 var(--editor-background);
}

/* menus */
.rstudio-themes-flat .themedPopupPanel,
.rstudio-themes-flat .popupMiddleCenter,
.rstudio-themes-flat .menuPopupMiddleCenter {
  color: var(--editor-font) !important;
  background-color: var(--tab) !important;
  border-color: var(--editor-headline) !important;
}

.rstudio-themes-flat .themedPopupPanel ::-webkit-scrollbar:vertical {
  width: 10px !important;
}

.rstudio-themes-flat .themedPopupPanel ::-webkit-scrollbar-track:vertical {
  box-shadow: none;
}

.rstudio-themes-flat .gwt-MenuItem-selected {
  background-color: var(--editor-headline) !important;
}

.rstudio-themes-flat .gwt-MenuItemSeparator > .menuSeparatorInner {
  border-color: var(--editor-headline) !important;
}

/* terminal */
.terminal {
  background-color: var(--editor-background);
  color: var(--editor-font);
  font-feature-settings: "liga" 12;
  position: relative;
  user-select: none;
  -ms-user-select: none;
  -webkit-user-select: none;
}
.terminal.xterm-cursor-style-block.focus:not(.xterm-cursor-blink-on)
  .terminal-cursor {
  background-color: var(--editor-headline);
  color: var(--cursor);
}
.terminal.focus.xterm-cursor-style-bar:not(.xterm-cursor-blink-on)
  .terminal-cursor::before,
.terminal.focus.xterm-cursor-style-underline:not(.xterm-cursor-blink-on)
  .terminal-cursor::before {
  content: "";
  position: absolute;
  background-color: var(--editor-headline);
}
.terminal:not(.focus) .terminal-cursor {
  outline: 1px solid var(--editor-headline);
  outline-offset: -1px;
}
.terminal .xterm-selection div {
  position: absolute;
  background-color: var(--editor-headline);
}
.terminal .xterm-viewport {
  background-color: var(--string);
  overflow-y: scroll;
}
.xtermInvertColor {
  color: var(--editor-background);
}
.xtermInvertBgColor {
  background-color: var(--editor-background);
}
.xtermBold {
  font-weight: bold;
}
.xtermUnderline {
  text-decoration: underline;
}
.xtermBlink {
  text-decoration: blink;
}
.xtermHidden {
  visibility: hidden;
}
.xtermItalic {
  font-style: italic;
}
.xtermStrike {
  text-decoration: line-through;
}
'

#' Error Handling
#'
#' @name error_handling
#'
#' @description
#' [resolve_intersection()]: Compares if two vectors have intersecting values. If TRUE,
#' removes the intersection values from the base vector
#'
#' @param base The base vector from which to remove any intersecting values.
#' @param vector_to_check The vector for which intersections should be checked.
#' @param check_only Returns a list of invalid entries instead of a vector. Additionally
#' it doesn't throw a warning.
#'
#' @return
#' Returns a vector or list.
#'
#' @examples
#' # Resolve intersection between two vectors
#' vec1 <- c("a", "b", "c", "d")
#' vec2 <- c("e", "f", "a", "g")
#'
#' vec1 <- resolve_intersection(vec1, vec2)
#'
#' @rdname error_handling
#'
#' @export
resolve_intersection <- function(base, vector_to_check, check_only = FALSE){
    # Get argument names used in the error message below
    base_name            <- deparse(substitute(base))
    vector_to_check_name <- deparse(substitute(vector_to_check))

    # Check for intersection between the two vectors
    intersection <- base %in% vector_to_check

    invalid_base <- base[intersection]
    base         <- base[!intersection]

    # Check only, if another function will be aborted because of this check.
    if (check_only){
        if (length(invalid_base) > 0){
            return(list(invalid_base, FALSE))
        }
    }
    # Otherwise throw a warning
    else{
        if (length(invalid_base) > 0){
            print_message("WARNING", c("The provided <{base_name}> variable[?s] '[invalid]' [?is/are] also part of",
									   "the <[to_check]> variables. [?This/These] variable[?s] will be omitted as <[base_name]> variable[?s] during computation."),
                          base_name = base_name, invalid = invalid_base, to_check = vector_to_check_name,
                          always_print = TRUE)
        }
    }

    base
}


#' Check If Variable Is Part Of A Data Frame
#'
#' @description
#' [part_of_df()]: Check if variable names are part of a data frame. If not, remove them
#' from the given vector.
#'
#' @param data_frame A data frame in which to look up variable names.
#' @param var_names A character vector of variable names.
#'
#' @examples
#' # Check if variables are part of a data frame
#' my_data   <- dummy_data(100)
#' var_names <- c("year", "state", "age", "test")
#'
#' var_names <- my_data |> part_of_df(var_names)
#'
#' @rdname error_handling
#'
#' @export
part_of_df <- function(data_frame, var_names, check_only = FALSE){
    # Get argument name used in the error message below
    vector_name <- deparse(substitute(var_names))

    # Check if variable names are part of the data frame
    intersection <- var_names %in% names(data_frame)

    invalid_vars <- var_names[!intersection]
    var_names    <- var_names[intersection]

    # Check only, if another function will be aborted because of this check.
    if (check_only){
        if (length(invalid_vars) > 0){
            return(list(invalid_vars, FALSE))
        }
    }
    # Otherwise throw a warning
    else{
        if (length(invalid_vars) > 0){
            print_message("WARNING", c("The provided <{vector}> variable[?s] '[invalid]' [?is/are] not part of",
                                       "the data frame. This variable will be omitted during computation."),
                          invalid = invalid_vars, vector = vector_name,
                          always_print = TRUE)
        }
    }

    var_names
}


#' Remove Doubled Values From A Vector
#'
#' @description
#' [remove_doubled_values()]: Remove values from a vector that appear more than once.
#'
#' @examples
#' # Remove doubled values
#' var_names <- c("year", "state", "state", "age")
#'
#' var_names <- remove_doubled_values(var_names)
#'
#' @rdname error_handling
#'
#' @export
remove_doubled_values <- function(var_names){
    # Get argument name used in the error message below
    vector_name <- deparse(substitute(var_names))

    # Only keep unique values
    var_names_unique <- collapse::funique(var_names)

    if (length(var_names) > length(var_names_unique)){
        print_message("WARNING", "Some <{vector}> variables are provided more than once. The doubled entries will be omitted.",
                      vector = vector_name,
                      always_print = TRUE)
    }

    var_names_unique
}


#' Check For Suitable Weight Variable
#'
#' @description
#' [check_weight()]: Check if a weight variable was provided. If TRUE, check whether it
#' can be used else add a temporary weight variable.
#'
#' @examples
#' # Check the provided weight variable
#' var_names <- my_data |> check_weight("weight")
#'
#' @rdname error_handling
#'
#' @export
check_weight <- function(data_frame, var_names){
    # Create temporary weight column if none is provided.
    # Also get the name of the weight variable as string.
    if (is.null(var_names) || length(var_names) > 1){
        data_frame[[".temp_weight"]] <- 1

        if (length(var_names) > 1){
            print_message("WARNING", "Only one variable for weight allowed. Evaluations will be unweighted.",
                          always_print = TRUE)
        }
    }
    else if (!var_names %in% names(data_frame)){
        data_frame[[".temp_weight"]] <- 1

        print_message("WARNING", "Provided weight variable is not part of the data frame. Unweighted results will be computed.",
                      always_print = TRUE)
    }
    else if (!is.numeric(data_frame[[var_names]])){
        data_frame[[".temp_weight"]] <- 1

        print_message("WARNING", "Provided weight variable is not numeric. Unweighted results will be computed.",
                      always_print = TRUE)
    }
    else{
        # NA values in weight lead to errors therefor convert them to 0
        if (anyNA(data_frame[[var_names]])){
            print_message("NOTE", "Missing values in weight variable '[var]' will be converted to 0.", var = var_names,
                          always_print = TRUE)
        }
        data_frame[[var_names]][is.na(data_frame[[var_names]])] <- 0

        # @Hack: so I don't have to check if .temp_weight exists later on
        data_frame[[".temp_weight"]] <- data_frame[[var_names]]
    }

    data_frame
}

#' Sort Data Frame Rows With Some Additions
#'
#' @description
#' Sort data frame rows by the provided variables. [sort_plus()] is also able to
#' preserve the current order of certain variables and only sort other variables
#' within this order. As another option one can sort a variable with the help of
#' formats, which can be used to e.g. sort a character variable in another than
#' alphabetical order without creating a temporary variable just for sorting.
#'
#' @param data_frame A data frame to summarise.
#' @param by A variable vector which contains the variables to sort by.
#' @param preserve A vector containing all variables which current order should be
#' preserved.
#' @param order A vector containing the sorting order for each variable. 'ascending'/'a'
#' or 'descending'/'d' can be used. If there are less orders given than by variables
#' provided, the last given sorting order will be used for the additional by variables.
#' @param formats A list in which is specified which formats should be used to sort
#' certain variables.
#' @param na.last TRUE by default. Specifies whether NA values should come last or first.
#'
#' @details
#' [sort_plus()] is just very loosely based on the 'SAS' procedure Proc Sort. It tries
#' to keep the simplicity, but with some added features.
#'
#' @return
#' Returns a sorted data table.
#'
#' @seealso
#' Creating formats: [discrete_format()] and [interval_format()].
#'
#' Functions that also make use of formats: [frequencies()], [crosstabs()],
#' [any_table()], [recode.()], [recode_multi()], [transpose_plus()].
#'
#' @examples
#' # Example formats
#' education. <- discrete_format(
#'     "1" = "low",
#'     "2" = "middle",
#'     "3" = "high")
#'
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Simple sorting
#' sort_df1 <- my_data |> sort_plus(by = c(state, sex, age))
#' sort_df2 <- my_data |> sort_plus(by    = c(state, sex, age),
#'                                  order = c("ascending", "descending"))
#'
#' # Character variables will normally be sorted alphabetically. With the help
#' # of a format this variable can be sorted in a completely different way.
#' sort_df3 <- my_data |> sort_plus(by      = education,
#'                                  formats = list(education = education.))
#'
#' # Preserve the order of the character variable, otherwise it couldn't stay in
#' # it's current order.
#' sort_df4 <- sort_df3 |> sort_plus(by       = age,
#'                                   preserve = education)
#'
#' # Or do all at once
#' sort_df5 <- my_data |> sort_plus(by       = c(sex, education),
#'                                  preserve = state,
#'                                  formats = list(education = education.))
#'
#' @export
sort_plus <- function(data_frame,
                      by,
                      preserve = NULL,
                      order    = "ascending",
                      formats  = c(),
                      na.last  = TRUE){
    # Measure the time
    print_start_message()
    print_step("GREY", "Error handling")

    ###########################################################################
    # Early evaluations
    ###########################################################################

    # First convert data frame to data table
    if (!data.table::is.data.table(data_frame)){
        data_frame <- data.table::as.data.table(data_frame)
    }

    # Evaluate formats early
    if (!is_list_of_dfs(formats)){
        formats_list <- as.list(substitute(formats))[-1]
        formats      <- evaluate_formats(formats_list)
    }

    ###########################################################################
    # Error handling
    ###########################################################################

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Preserve variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    preserve <- get_origin_as_char(preserve, substitute(preserve))

    # Make sure that the variables provided are part of the data frame.
    preserve <- data_frame |> part_of_df(preserve)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # By variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    by <- get_origin_as_char(by, substitute(by))

    # If no value variables are provided abort
    if (length(by) <= 1){
        if (length(by) == 0 || by == ""){
            print_message("ERROR", "No <by> variables provided. Sorting will be aborted.")
            return(invisible(NULL))
        }
    }

    # Make sure there is no column variable that is also a row variable.
    by <- resolve_intersection(by, preserve)

    # Make sure that the variables provided are part of the data frame.
    by <- data_frame |> part_of_df(by)

    if (length(by) == 0){
        print_message("ERROR", "No valid <by> variable provided. Sorting will be aborted.")
        return(invisible(NULL))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Double entries
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    preserve <- remove_doubled_values(preserve)
    by       <- remove_doubled_values(by)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Order
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # If there are invalid orders specified transform them to ascending order
    invalid_order <- any(!tolower(order) %in% c("ascending", "descending", "a", "d"))

    if (invalid_order){
        print_message("WARNING", c("<Order> other than 'ascending'/'a' or 'descending'/'d' specified, which is\n",
								   "not allowed. 'ascending' will be used instead."))
    }

    # Convert character expressions into numbers, which is needed later on in the
    # data.table sort.
    order[order %in% c("descending", "d")] <- -1
    order[order != -1] <- 1

    # Fill up order vector with the last element or trim it down, if too many
    # elements are provided.
    order <- fill_or_trim(order, length(by))

    ###########################################################################
    # Sorting starts
    ###########################################################################

    # If formats are provided, they will be used to temporarily recode new variables,
    # which are used to order the data frame in format order first.
    if (!is.null(formats)){
        print_step("MAJOR", "Preparing formats")

        for (variable in names(formats)){
            if (is_multilabel(formats, variable)){
                print_message("WARNING", c("Applying a multilabel to variable '[variable]' is not allowed.\n",
										   "Format won't be applied. Variable is skipped."), variable = variable)
                next
            }

            if (!variable %in% names(data_frame)){
                print_message("WARNING", c("The variable '[variable]' is not part of the data frame.\n",
										   "Format cannot be applied. Variable is skipped."), variable = variable)
                next
            }

            if (!variable %in% by){
                print_message("NOTE", c("Format for variable '[variable]' is provided, but the variable itself.",
										"is not part of the <by> variables. Format won't be applied. Variable is skipped."), variable = variable)
                next
            }

            # To make recode work "variable" can't be passed directly, because recode
            # would take it as "variable" instead of e.g. "age". Therefore a named list
            # with all the needed elements has to be constructed beforehand.
            new_column      <- paste0(".sort_", variable)
            argument_list   <- list(data_frame)
            argument_list[[variable]] <- formats[[variable]]

            # Now use do.call to run e.g.: recode.(data_frame, ".sort_age", age = formats[["age"]])
            data_frame[new_column] <- suppressMessages(do.call(recode., argument_list))

            # Format new sorting variable as factor to sort in provided format order
            # Extract the number of labels from variable
            label_levels <- formats[[variable]] |>
                unlist(use.names = FALSE) |>
                collapse::funique() |>
                collapse::na_omit()

            # Convert variable to factor
            data_frame[[new_column]] <- factor(
                data_frame[[new_column]],
                levels  = label_levels,
                ordered = TRUE)

            by[by == variable] <- new_column
        }
    }

    if (!is.null(preserve)){
        print_step("MAJOR", "Preserving: [var]", var = preserve)

        # If there are variables to preserve, they will be converted into factor
        # variables before sorting. To restore them afterwards, to there original
        # format, grab it here.
        original_class <- lapply(data_frame[preserve], class)

        # Convert preserve variables to factor so that they keep their order during sort
        data_frame <- data_frame |> convert_factor(preserve)
        new_by     <- c(preserve, by)

        # Fill up order vector with the last element or trim it down, if too many
        # elements are provided.
        order <- fill_or_trim(order, length(new_by))

        # Actual sort
        sort_df <- data.table::copy(data_frame) |>
            data.table::setorderv(cols    = new_by,
                                  order   = order,
                                  na.last = na.last)

        # Convert preserve variables back
        for (variable in names(original_class)){
            # Only check first element. If original type is factor it is stored
            # in two parts 'ordered' and 'factor'.
            if (original_class[[variable]][1] == "numeric"){
                sort_df[[variable]] <- as.numeric(sort_df[[variable]])
            }
            else if (original_class[[variable]][1] == "logical"){
                sort_df[[variable]] <- as.logical(sort_df[[variable]])
            }
            else if (original_class[[variable]][1] == "character"){
                sort_df[[variable]] <- as.character(sort_df[[variable]])
            }

        }
    }
    else{
        sort_df <- data.table::copy(data_frame) |>
            data.table::setorderv(cols    = by,
                                  order   = order,
                                  na.last = na.last)
    }

    # Clean up: Remove all temporary sort variables, if formats where used
    sort_vars <- grep("^\\.sort", names(data_frame), value = TRUE)
    sort_df   <- suppressMessages(sort_df |> dropp(sort_vars))

    print_closing(5)

    sort_df
}

#' Fast And Powerful Yet Simple To Use Transpose
#'
#' @description
#' [transpose_plus()] is able to reshape a data frame from long to wide and from
#' wide to long. In the long to wide transposition variables can be nested or
#' placed side by side. With the wide to long transposition it is also possible
#' to transpose multiple variables at once.
#'
#' Additionally [transpose_plus()] is able to weight results before transposing
#' them from long to wide.
#'
#' The function also makes use of formats, which means you don't need to create
#' variables storing the new variable names before transposition. You can just
#' use formats to name the new variables and with multilabels you can even generate
#' new variable expressions at the same time.
#'
#' @param data_frame A data frame to transpose
#' @param preserve Variables to keep and preserve in their current form.
#' @param pivot A vector that provides the expressions of single variables or
#' variable combinations that should be transposed. To nest variables use the form:
#' "var1 + var2 + var3 + ...".
#' @param values A vector containing all value variables that should be transposed.
#' @param formats A list in which is specified which formats should be applied to which
#' variables.
#' @param weight Put in a weight variable to compute weighted results.
#' @param na.rm FALSE by default. If TRUE removes all NA values from the preserve and
#' pivot variables.
#' @param monitor FALSE by default. If TRUE, outputs two charts to visualize the
#' functions time consumption.
#'
#' @details
#' [transpose_plus()] is just very loosely based on the 'SAS' procedure Proc Transpose,
#' and the possibilities of a Data-Step transposition using loops.
#'
#' The transposition methods 'SAS' has to offer are actually fairly weak. Which is weird
#' because all tools are there to have another powerful function. So [transpose_plus()]
#' tries to create the function 'SAS' should have.
#'
#' The function is able to interpret which transposition direction the user wants by just
#' looking at what the user provided with the function parameters. For a long to wide
#' transposition it is natural to just provide variables to transpose. While it is also
#' just natural to provide new variable names when transposing from wide to long. That alone
#' reduces the number of parameters the user has to enter to perform a simple transposition.
#'
#' The real magic happens when formats come into play. With their help you can not only name
#' new variables or their expressions, but you can also generate completely new expressions
#' with no effort, just with the help of multilabels.
#'
#' @return
#' Returns a transposed data table.
#'
#' @seealso
#' Creating formats: [discrete_format()] and [interval_format()].
#'
#' Functions that also make use of formats: [frequencies()], [crosstabs()],
#' [any_table()], [recode.()], [recode_multi()], [sort_plus()].
#'
#' @examples
#' # Example formats
#' age. <- discrete_format(
#'     "Total"          = 0:100,
#'     "under 18"       = 0:17,
#'     "18 to under 25" = 18:24,
#'     "25 to under 55" = 25:54,
#'     "55 to under 65" = 55:64,
#'     "65 and older"   = 65:100)
#'
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' sex2. <- discrete_format(
#'     "Total"  = c("Male", "Female"),
#'     "Male"   = "Male",
#'     "Female" = "Female")
#'
#' income. <- interval_format(
#'     "Total"              =    0:100000,
#'     "below 500"          =    0:500,
#'     "500 to under 1000"  =  500:1000,
#'     "1000 to under 2000" = 1000:2000,
#'     "2000 and more"      = 2000:100000)
#'
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Transpose from long to wide and use a multilabel to generate additional categories
#' long_to_wide <- my_data |>
#'     transpose_plus(preserve = c(year, age),
#'                    pivot    = c("sex", "education"),
#'                    values   = income,
#'                    formats  = list(sex = sex., age = age.),
#'                    weight   = weight,
#'                    na.rm    = TRUE)
#'
#' # Transpose back from wide to long and put results below each other. To trigger
#' # this behavior every list entry in pivot has to have a different name.
#' wide_to_long <- long_to_wide |>
#'     transpose_plus(preserve = c(year, age),
#'                    pivot    = list(sex       = c("Total", "Male", "Female"),
#'                                    education = c("low", "middle", "high")))
#'
#' # Transpose from long to wide and use a multilabel to generate additional categories
#' long_to_wide <- my_data |>
#'     transpose_plus(preserve = c(year, age),
#'                    pivot    = "sex",
#'                    values   = c(income, weight),
#'                    formats  = list(sex = sex., age = age.),
#'                    weight   = weight,
#'                    na.rm    = TRUE) |>
#'     rename_multi("income_Total"  = "Total",
#'                  "income_Male"   = "Male",
#'                  "income_Female" = "Female")
#'
#' # Transpose back from wide to long but this time put results side by side.
#' # To do that every list entry has to have the same name. The values parameter
#' # is then used to give the new value variables a name. For the expressions of
#' # the new categorical variable the variable names from the first pivot list
#' # entry are used.
#' wide_to_long <- long_to_wide |>
#'     transpose_plus(preserve = c(year, age),
#'                    values   = c(income, weight),
#'                    pivot    = list(sex = c("Total", "Male", "Female"),
#'                                    sex = c("weight_Total", "weight_Male", "weight_Female")))
#'
#' # Nesting variables in long to wide transposition
#' nested <- my_data |>
#'     transpose_plus(preserve = c(year, age),
#'                    pivot    = "sex + education",
#'                    values   = income,
#'                    formats  = list(sex = sex., age = age.),
#'                    weight   = weight,
#'                    na.rm    = TRUE)
#'
#' # Or both, nested and un-nested, at the same time
#' both <- my_data |>
#'     transpose_plus(preserve = c(year, age),
#'                    pivot    = c("sex + education", "sex", "education"),
#'                    values   = income,
#'                    formats  = list(sex = sex., age = age.),
#'                    weight   = weight,
#'                    na.rm    = TRUE)
#'
#' @export
transpose_plus <- function(data_frame,
                           preserve   = NULL,
                           pivot      = NULL,
                           values     = NULL,
                           formats    = c(),
                           weight     = NULL,
                           na.rm      = .qol_options[["na.rm"]],
                           monitor    = .qol_options[["monitor"]]){
    # Measure the time
    print_start_message()
    print_step("GREY", "Error handling")

    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Error handling", "Preparation")
    #-------------------------------------------------------------------------#

    ###########################################################################
    # Early evaluations
    ###########################################################################

    # First convert data frame to data table
    if (!data.table::is.data.table(data_frame)){
        data_frame <- data.table::as.data.table(data_frame)
    }

    # Evaluate formats early
    if (!is_list_of_dfs(formats)){
        formats_list <- as.list(substitute(formats))[-1]
        formats      <- evaluate_formats(formats_list)
    }

    # If all pivot list/vector entries have a name, transposition will be wide to long
    long_to_wide <- TRUE

    if (is.null(pivot) || (!is.null(names(pivot)) && all(nzchar(names(pivot))))){
        long_to_wide <- FALSE
        side_by_side <- FALSE

        # In case there is more than one variable to pivot, it will be checked, whether
        # all variables have the same name (put results side by side) or whether they
        # are different (results will be put below each other).
        if (length(pivot) > 1){
            unique_var_names <- collapse::funique(names(pivot))

            # If the number of unique variable names is equal to 1 in a multi variable
            # name list, then all variable names are the same and results should be
            # put together side by side.
            if (length(unique_var_names) == 1){
                if (length(collapse::funique(lengths(pivot))) > 1){
                    print_message("ERROR", c("Every <pivot> list entry has to have the same number of variables for a",
                                             "side by side transposition to work. Transposition will be aborted."))
                    return(invisible(NULL))
                }

                side_by_side <- TRUE

                # In case of no provided values (which are the variable names in this case)
                # there will be dummy names.
                values <- get_origin_as_char(values, substitute(values))

                if (is.null(values)){
                    values <- paste0("value", seq_len(length(pivot)))
                }
            }
            # In case of a mix of equal and unequal names transposition will be aborted.
            # It is only possible to do one or the other (side by side vs. below each other)
            else if (length(unique_var_names) < length(pivot)){
                print_message("ERROR", c("The new result columns can only be set side by side in a wide",
                                         "to long transposition, if all variable names are the same.",
                                         "Transposition will be aborted."))
                return(invisible(NULL))
            }
        }
    }

    ###########################################################################
    # Error handling
    ###########################################################################

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Wide to long
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (!long_to_wide){
        # Crossing variables not possible when transposing into long format.
        # Loop through all variable vectors in named list. If there is any crossing
        # listed, abort.
        for (variables in pivot){
            if (any(grepl("+", variables, fixed = TRUE))){
                print_message("ERROR", c("Nesting <pivot> variables in a wide to long transposition is not possible.",
										 "Transposition will be aborted."))
                return(invisible(NULL))
            }
        }

        # Values and weight has no effect in wide to long transposition
        if (!is.null(values) && !side_by_side){
            print_message("NOTE", "<Values> parameter has no effect in wide to long transposition, when results are stacked.")
        }

        if (!is.null(weight)){
            print_message("NOTE", "<Weight> parameter has no effect in wide to long transposition.")
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Weight
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (long_to_wide){
        weight     <- get_origin_as_char(weight, substitute(weight))
        data_frame <- data_frame |> check_weight(weight)
        weight_var <- ".temp_weight"
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Preserve variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    preserve <- get_origin_as_char(preserve, substitute(preserve))

    # Make sure that the variables provided are part of the data frame.
    preserve <- data_frame |> part_of_df(preserve)

    # If no preserve variables are provided, create a pseudo preserve variable
    if (length(preserve) == 0){
        preserve <- ".pseudo_preserve"
        data_frame[[".pseudo_preserve"]] <- 1
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Pivot variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (is.null(pivot)){
        variable_names <- names(data_frame)
        pivot          <- list(".transposed" = variable_names[!variable_names %in% preserve])
    }

    # Get pivot variables from provided combinations
    if (long_to_wide){
        pivot_vars <- unlist_variables(pivot)
    }
    else{
        pivot_vars <- collapse::funique(unlist(pivot, use.names = FALSE))
    }

    if (is.null(pivot_vars)){
        print_message("ERROR", "<Pivot> variables must be provided in quotation marks. Transposition will be aborted.")
        return(invisible(NULL))
    }

    pivot_vars <- data_frame |> part_of_df(pivot_vars, check_only = TRUE)

    if (is.list(pivot_vars)){
        print_message("ERROR", c("The provided <pivot> variable[?s] '[vars]' [?is/are] not part of",
								 "the data frame. Transposition will be aborted."), vars = pivot_vars[[1]])
        return(invisible(NULL))
    }

    if (length(pivot_vars) <= 1){
        if (length(pivot_vars) == 0 || pivot_vars == ""){
            print_message("ERROR", "No valid <pivot> variables provided. Transposition will be aborted.")
            return(invisible(NULL))
        }
    }

    # Make sure there is no pivot variable that is also a preserve variable.
    pivot_vars <- resolve_intersection(pivot_vars, preserve, check_only = TRUE)

    if (is.list(pivot_vars)){
        print_message("ERROR", c("The provided <pivot> variable[?s] '[vars]' [?is/are] also part of",
								 "the <preserve> variables, which is not allowed. Transposition will be aborted."), vars = pivot_vars[[1]])
        return(invisible(NULL))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Value variables
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (long_to_wide){
        values <- get_origin_as_char(values, substitute(values))

        # If no value variables are provided abort
        if (length(values) <= 1){
            if (length(values) == 0 || values == ""){
                print_message("ERROR", "No <values> provided. Transposition will be aborted.")
                return(invisible(NULL))
            }
        }

        # Make sure there is no values variable that is also a preserve variable.
        values <- resolve_intersection(values, preserve)

        # Make sure there is no values variable that is also a pivot variable.
        values <- resolve_intersection(values, pivot_vars, check_only = TRUE)

        if (is.list(values)){
            print_message("ERROR", c("The provided <values> variable[?s] '[vars]' [?is/are] also part of",
									 "the <pivot> variables, which is not allowed. Transposition will be aborted."), vars = values[[1]])
            return(invisible(NULL))
        }

        values <- data_frame |> part_of_df(values, check_only = TRUE)

        if (is.list(values)){
            print_message("ERROR", "No valid <values> to transpose provided. Transposition will be aborted.")
            return(invisible(NULL))
        }

        if (length(values) == 0){
            print_message("ERROR", "No <values> variables provided. Summarise will be aborted.")
            return(invisible(NULL))
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Double entries
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    preserve <- remove_doubled_values(preserve)

    if (long_to_wide){
        pivot_vars <- remove_doubled_values(pivot_vars)
        values     <- remove_doubled_values(values)
    }

    ###########################################################################
    # Transposition starts
    ###########################################################################

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Preparation", "Preparation")
    #-------------------------------------------------------------------------#

    # Check for pre summarised data, to be able to take a shortcut
    pre_summed <- data_frame |> is_pre_summed(c(preserve, pivot_vars))

    # Determine transposition method - only necessary for long to wide
    # transposition. For wide to long a named list is already given.
    # If variable names are combined with a + sign, then they will be crossed
    # during transposition.
    if (long_to_wide){
        # Store variable combinations in a list
        transpose_methods <- stats::setNames(
            lapply(pivot, function(variables){
                trimws(strsplit(variables, "\\+", fixed = FALSE)[[1]])
            }), pivot)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Long -> Wide
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (long_to_wide){
        # Summarise data first in order to apply formats, if specified
        if (!pre_summed && !is.null(formats)){
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Summarise", "Long to wide")
            #-----------------------------------------------------------------#
            print_step("MAJOR", "Summarising data.")

            group_vars <- c(preserve, pivot_vars)

            data_frame <- suppressMessages(data_frame |>
                summarise_plus(class      = group_vars,
                               values     = values,
                               statistics = "sum",
                               formats    = formats,
                               weight     = weight_var,
                               nesting    = "deepest",
                               notes      = FALSE,
                               na.rm      = na.rm)) |>
                remove_stat_extension("sum")
        }

        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Transpose", "Long to wide")
        #---------------------------------------------------------------------#

        combined_df <- NULL

        # The intended behavior of the collapse pivot is to cross all variables provided as
        # 'names'. If a variable combination is provided, the variables will be crossed but
        # if they are provided separately, they will be put beside each other. Therefore
        # each pivot has to be done sequentially.
        print_step("MAJOR", "Transposing long to wide.")

        for (method in transpose_methods){
            print_step("MINOR", paste(method, collapse = " + "))

            transpose_df <- data_frame |>
                collapse::pivot(id     = preserve,
                                names  = method,
                                values = values,
                                how    = "wider")

            # Remove NA columns afterwards.
            if (na.rm){
                none_NA_columns <- which(!is.na(names(transpose_df)) &
                                         !grepl("NA_|_NA", names(transpose_df)))
                transpose_df    <- transpose_df |> collapse::fselect(none_NA_columns)
            }

            # Combine transposed data frames
            if (is.null(combined_df)){
                combined_df <- transpose_df
            }
            # Following iterations
            else{
                # Only keep new value columns
                transpose_df <- transpose_df |> dropp(preserve)

                # Check for duplicate variable names. If any duplicate is found abort.
                duplicates <- intersect(names(combined_df), names(transpose_df))

                if (length(duplicates) > 0){
                    print_message("ERROR", c("Duplicate column names found: [duplicate].",
											 "If you are working with original values, consider making them unique by using formats."),
                                  duplicate = paste(duplicates, collapse = ", "))
                    return(invisible(NULL))
                }

                # cbind current data frame to the iterations before
                combined_df <- cbind(combined_df, transpose_df)
            }
        }
    }
    # Wide to long
    else{
        print_step("MAJOR", "Transposing wide to long")

        combined_df <- NULL

        # Each given list entry will be transposed sequentially
        for (i in seq_along(pivot)){
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Transpose", "Wide to long")
            #-----------------------------------------------------------------#
            variable <- names(pivot)[i]

            # Only keep the necessary variables because otherwise all variables will be transposed.
            # Since it should be possible to transpose multiple variables into multiple categories,
            # this step is essential.
            vars_to_keep <- c(preserve, pivot[[i]])

            print_step("MINOR", "[var] = [keep]", var = variable, keep = vars_to_keep[!vars_to_keep %in% preserve])

            # Determine new variable name. If only one new variable provided take the name from
            # the named list, otherwise use a general name.
            var_name <- "VARIABLE"

            if (length(names(pivot)) == 1){
                var_name <- variable
            }

            # After transposition into long format a BY variable is computed containing the list
            # elements name. This way one can identify which category the expressions belong to,
            # if multiple transpositions are put back together afterwards.
            transpose_df <- data_frame |>
                keep(vars_to_keep) |>
                collapse::pivot(id    = preserve,
                                names = list(variable = var_name,
                                              value   = "VALUE"),
                                how   = "longer",
                                na.rm = na.rm) |>
                collapse::fmutate(BY = variable) |>
                data.table::setcolorder(c("BY", var_name, "VALUE"), after = length(preserve))

            transpose_df[[var_name]] <- as.character(transpose_df[[var_name]])

            # Recode variable, if format is given
            if (!is.null(formats) && variable %in% names(formats)){
                #-------------------------------------------------------------#
                monitor_df <- monitor_df |> monitor_next("Summarise", "Wide to long")
                #-------------------------------------------------------------#

                # Set up new group vars for summarise. This also could be done with a
                # recode, but in case of a multilabel there has to be an additional summarise.
                # Therefore just use the summarise in both versions.
                group_vars <- c(preserve, "BY", var_name)

                # Keeping it simple and just rename the variable name to the generic one
                # in a copy of formats vector. Always copy because otherwise on the next
                # iteration there would be two entries called "VARIABLE".
                formats_copy <- formats
                names(formats_copy)[names(formats_copy) == variable] <- var_name

                transpose_df <- suppressMessages(transpose_df |>
                   summarise_plus(class      = group_vars,
                                  values     = VALUE,
                                  statistics = "sum",
                                  formats    = formats_copy,
                                  nesting    = "deepest",
                                  notes      = FALSE,
                                  na.rm      = na.rm)) |>
                    remove_stat_extension("sum") |>
                    drop_type_vars()

                transpose_df <- suppressMessages(transpose_df |> sort_plus(by = var_name))
            }

            # Combine transposed data frames
            if (is.null(combined_df)){
                # If only one new variable provided drop BY-variable
                if (length(names(pivot)) == 1){
                    transpose_df <- transpose_df |> collapse::fselect(-BY)
                }
                else if (side_by_side){
                    transpose_df <- suppressMessages(transpose_df |>
                        collapse::fselect(-BY) |>
                        rename_multi("VARIABLE" = variable,
                                     "VALUE"    = values[i]))
                }

                combined_df <- transpose_df
            }
            # Following iterations
            else{
                if (!side_by_side){
                    # rbind current data frame to the iterations before
                    combined_df <- rbind(combined_df, transpose_df)
                }
                else{
                    # On a side by side transposition all variable expressions are
                    # renamed, so that a new primary key variable exists on which
                    # the iterations can be joined. Whether the result makes sense
                    # or not, is entirely in the users hands to decide.
                    current_expressions <- collapse::funique(transpose_df[["VARIABLE"]])
                    rename_map          <- stats::setNames(pivot[[1]], current_expressions)

                    transpose_df[["VARIABLE"]] <- rename_map[transpose_df[["VARIABLE"]]]

                    # Prepare data frame for the join
                    transpose_df <- suppressMessages(transpose_df |>
                        collapse::fselect(-BY) |>
                        rename_multi("VARIABLE" = variable,
                                     "VALUE"    = values[i]))

                    # Join to main data frame
                    combined_df <- combined_df |>
                        collapse::join(transpose_df,
                                       on      = c(preserve, variable),
                                       how     = "left",
                                       verbose = FALSE,
                                       overid  = 2)
                }
            }
        }
    }

    # Drop pseudo preserve variable if it is there
    if (names(combined_df)[[1]] == ".pseudo_preserve"){
        combined_df <- combined_df |> collapse::fselect(-.pseudo_preserve)
    }

    print_closing()

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_end()
    monitor_df |> monitor_plot(draw_plot = monitor)
    #-------------------------------------------------------------------------#

    combined_df
}

#' Check If Path Exists And Retrieve Files
#'
#' @description
#' [libname()] checks if a given path exists and writes a message in the console accordingly.
#' Optional all files from the given path can be retrieved as a named character vector.
#'
#' @param path A folder path.
#' @param get_files FALSE by default. If TRUE returns a named character vector containing
#' file paths.
#' @param extensions Specify file extensions to be kept in the list when retrieving files.
#' @param recursive FALSE by default. If TRUE scans the provided folder for additional files
#' inside sub folders. Only work in combination with get_files = TRUE.
#'
#' @return
#' Returns the given file path or a named character vector containing file paths.
#'
#' @examples
#' my_path   <- libname("C:/My_Path/")
#' file_list <- libname("C:/My_Path/", get_files = TRUE)
#'
#' @export
libname <- function(path,
                    get_files  = FALSE,
                    extensions = NULL,
                    recursive  = FALSE){
    path <- macro(path)

    if (!file.exists(path) || dirname(path) == "."){
        print_message("ERROR", "Path does not exist: [path]", path = path)
        return(invisible(NULL))
    }

    print_start_message(suppress = TRUE)

    if (get_files){
		print_step("GREY", "Get files")

        # Retrieve all file paths from provided path
        files <- list.files(path, full.names = TRUE, recursive = recursive)

        # Strip paths and only keep file names with extension
        files <- files[!dir.exists(files)]

        if (length(files) == 0){
            print_message("ERROR", "No files found in directory: [path]", path = path)
            return(invisible(NULL))
        }

        # Return named character vector
        print_step("MAJOR", "Filepaths successfully retrieved: [path]", path = path)

        # Return all files as list if no extensions where specified
        if (is.null(extensions)){
            # Print file names and return
            if (length(files) > 0){
                for (file in files){
                    print_step("MINOR", basename(file))
                }
            }
            else{
                print_message("NOTE", "No files found.")
            }

            print_closing(suppress = TRUE)
            return(invisible(stats::setNames(files, basename(files))))
        }
        # Filter extensions
        else{
            file_list <- stats::setNames(files, basename(files))

            # Create pattern to grab all extensions at once
            pattern <- paste0("(", paste(gsub("\\.", "\\\\.", extensions), collapse = "|"), ")$")
            files   <- file_list[grepl(pattern, file_list, ignore.case = TRUE)]

            print_step("MAJOR", "Only keeping files with extension[?s]: [extensions]", extensions = extensions)

            # Print file names and return
            if (length(files) > 0){
                for (file in files){
                    print_step("MINOR", basename(file))
                }
            }
            else{
                print_message("NOTE", "No files found.")
            }

            print_closing(suppress = TRUE)

            return(invisible(files))
        }
    }

    print_step("MAJOR", "Path successfully assigned: [path]", path = path)
    print_closing(suppress = TRUE)

    path
}


#' Stack Multiple Data Frames
#'
#' @description
#' Stacks multiple data frames and matches column names.
#'
#' @param ... Put in multiple data frames to stack them in the provided order.
#' @param id Adds an ID column to indicate the different data frames.
#' @param compress No compression by default. If compression receives any value, [stack_data()]
#' will convert character variables to numeric or integer where possible. If set to "factor"
#' all non numeric character variables will be converted to factors.
#' @param guessing_rows 100 by default. [stack_data()] takes a sample of rows to determine of
#' which type variables are.
#'
#' @return
#' Returns a stacked data frame.
#'
#' @examples
#' # Example data frames
#' my_data1 <- dummy_data(100)
#' my_data2 <- dummy_data(100)
#' my_data3 <- dummy_data(100)
#' my_data4 <- dummy_data(100)
#' my_data5 <- dummy_data(100)
#'
#' # Stack data frames
#' stacked_df <- stack_data(my_data1,
#'                          my_data2,
#'                          my_data3,
#'                          my_data4,
#'                          my_data5)
#'
#' @export
stack_data <- function(...,
                       id            = FALSE,
                       compress      = NULL,
                       guessing_rows = 100){
    # Measure the time
    print_start_message(suppress = TRUE)

    # Translate ... into a list if possible
    df_list <- tryCatch({
        # Force evaluation to see if it exists
        list(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    if (is.null(df_list)){
        print_message("ERROR", "Unknown object found. Retaining will be aborted.")
        return(invisible(NULL))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Row binding
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Make sure df_list is a flat list of data frames. Otherwise rbindlist tries
    # to stuff the list within the list into a single data frame column (with
    # success and no error by the way).
    if (any(sapply(df_list, function(element){ !is.data.frame(element) }))){
        element_names <- sapply(substitute(list(...))[-1], deparse)

        df_list <- do.call(c, lapply(seq_along(df_list), function(i){
            element <- df_list[[i]]

            # Set the objects name as list entry name
            if (is.data.frame(element)){
                stats::setNames(list(element), element_names[i])
            }
            # The lists already have their names
            else{
                element
            }
        }))
    }

    # Generate an ID variable and order it to the last position which indicates the single data frame
    if (id){
        data_frame <- data.table::rbindlist(df_list, use.names = TRUE, fill = TRUE, idcol = "ID")
        data_frame <- data_frame |> data.table::setcolorder("ID", after = collapse::fncol(data_frame))
    }
    # Stack without ID variable
    else{
        data_frame <- data.table::rbindlist(df_list, use.names = TRUE, fill = TRUE)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Compression
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Compress numeric values to integers if possible and conditionally convert characters to factors
    if (!is.null(compress)){
        columns <- names(data_frame)

        for (column in columns){
            single_column <- data_frame[[column]]

            # Check all character columns
            if (is.character(single_column)){
                # Take a sample of the first x non-NA values to guess type
                sample_column <- single_column |>
                    collapse::na_omit() |>
                    collapse::fsubset(1:min(guessing_rows))

                if (length(sample_column) == 0){
                    next
                }

                # Try to convert to check whether it is numeric
                is_numeric <- suppressWarnings(!any(is.na(as.numeric(sample_column))))

                if (is_numeric){
                    # Check for decimals to decide between integer or double
                    has_decimals <- any(grepl("\\.", sample_column))

                    # Convert
                    if (!has_decimals){
                        data.table::set(data_frame, j = column, value = as.integer(single_column))
                    }
                    else{
                        data.table::set(data_frame, j = column, value = as.numeric(single_column))
                    }
                }
                # Convert character values to factors
                else if (compress == "factor"){
                    data.table::set(data_frame, j = column, value = as.factor(single_column))
                }
            }
            # Check if numeric columns can be converted to integer
            else if (is.numeric(single_column) && !is.integer(single_column) && !is.factor(single_column)){
                # Only convert to integer if it doesn't result in data loss
                if (all(single_column == floor(single_column), na.rm = TRUE)){
                    data.table::set(data_frame, j = column, value = as.integer(single_column))
                }
            }
        }
    }

    print_closing()

    data_frame
}


#' Save And Load FST And RDS Files
#'
#' @name save_load
#'
#' @description
#' [save_file()]: Saves fst and rds files. Offers variable selection and observation
#' subsetting. By default the function has a write protection, which has to be
#' explicitly turned off to be able to overwrite files.
#'
#' @param data_frame The data fame to be saved.
#' @param path The file path to save to/load from.
#' @param file The file name including the file extension.
#' @param keep The variables to keep in the file.
#' @param where A condition on which to subset the observations.
#' @param compress The amount of compression to use from 0 to 100. The lower the value,
#' the larger the file size and faster the saving speed. Uses maximum compression by
#' default.
#' @param protect TRUE by default. Throws an error if a file already exists. If FALSE,
#' overwrites existing files.
#' @param ... Used internally to suppress messages, when using the multi save/load version.
#'
#' @return
#' Returns a data frame. [load_file_multi()] can also return a list of data frame.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(100)
#'
#' # Save files
#' # NOTE: Normally you would pass in the path and file as character. For the
#' #       examples this is handled differently to provide runnable examples.
#' my_data |> save_file(path = tempdir(),
#'                      file = "testfile.fst")
#' my_data |> save_file(path = tempdir(),
#'                      file = "testfile.rds")
#'
#' # Save file and only keep specific variables
#' # NOTE: Since the temporary file already exists now if you run the above code,
#' #       all the following save operations would throw errors because by default
#' #       the function write protects existing files. So for the following examples
#' #       the write protection is turned off.
#' my_data |> save_file(path    = tempdir(),
#'                      file    = "testfile.fst",
#'                      keep    = c(sex, age, state),
#'                      protect = FALSE)
#'
#' # Save file and subset observations
#' my_data |> save_file(path    = tempdir(),
#'                      file    = "testfile.fst",
#'                      where   = sex == 1 & age > 65,
#'                      protect = FALSE)
#'
#' @rdname save_load
#'
#' @export
save_file <- function(data_frame,
                      path,
                      file,
                      keep     = NULL,
                      where    = NULL,
                      compress = 100,
                      protect  = TRUE,
                      ...){
    print_start_message(suppress = TRUE)
    suppress <- ifelse(is.null(unlist(list(...))), FALSE, TRUE)

    ###########################################################################
    # Error handling
    ###########################################################################

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # File extension
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    extension <- tolower(tools::file_ext(file))

    if (extension == ""){
        print_message("WARNING", "No file extension provided in <file>. 'fst' will be used.")

        file      <- paste0(file, ".fst")
        extension <- "fst"
    }

    if (!extension %in% c("fst", "rds")){
        print_message("WARNING", c("Only 'fst' or 'rds' are allowed as file extensions in <file>.",
								   "'fst' will be used."))

        file      <- paste0(tools::file_path_sans_ext(file), ".fst")
        extension <- "fst"
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # File path
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    path <- macro(path)
    file <- macro(file)

    full_path <- file.path(path, file)

    if (file.exists(full_path) && protect){
        print_message("ERROR", c("File already exists: [full_path]",
								 "Set <protect> to FALSE to overwrite files."), full_path = full_path)

        return(invisible(data_frame))
    }

    if (!file.exists(path)){
        print_message("ERROR", "Path does not exist: [full_path]", full_path = full_path)

        return(invisible(data_frame))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Keep
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Convert to character vectors
    keep_eval <- substitute(keep)

    if (is.call(keep_eval)){
        keep <- get_origin_as_char(keep, keep_eval)
    }

    # Check if variables exist in the data frame. If not, abort.
    keep <- data_frame |> part_of_df(keep, check_only = TRUE)

    if (is.list(keep)){
        print_message("ERROR", c("The provided variables to <keep> '[keep_vars]' are not part of",
								 "the data frame. Saving will be aborted."), keep_vars = keep[[1]])
        return(invisible(data_frame))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Compress
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    compress <- min(100, max(0, compress))

    ###########################################################################
    # Saving
    ###########################################################################

    # Select and subset data frame
    if (!is.null(keep)){
        data_frame <- suppressMessages(data_frame |>
                                           collapse::fselect(keep) |>
                                           data.table::setcolorder(keep) |>
                                           if.(where))
    }
    else{
        data_frame <- suppressMessages(data_frame |> if.(where))
    }

    if (extension == "fst"){
        suppressMessages(fst::threads_fst(.qol_options[["threads"]]))

        # Save data frame
        suppressWarnings(suppressMessages(
            data_frame |> fst::write_fst(full_path, compress = compress)))
    }
    if (extension == "rds"){
        data_frame |> saveRDS(full_path, compress = compress == 100)
    }

    # Revert used threads
    suppressMessages(fst::threads_fst(NULL))

    if (extension == "fst"){
        print_closing(5)
    }
    else{
        print_closing(15)
    }

    invisible(data_frame)
}


#' @description
#' [save_file_multi()]: Saves multiple files.
#'
#' @param data_frame_list A list of data frames.
#' @param file_list A character vector containing full file paths.
#' @param keep_list Can be a single variable name, a vector or a list of vectors containing
#' the variables to keep per file. If there are fewer list entries than files to load, the
#' last list element will be used repeatedly.
#'
#' @examples
#' # Example lists
#' my_df_list <- list(dummy_data(10),
#'                    dummy_data(10))
#'
#' file1 <- file.path(tempdir(), "first.fst")
#' file2 <- file.path(tempdir(), "second.rds")
#' my_file_list <- list(file1, file2)
#'
#' # Save multiple files at once
#' save_file_multi(data_frame_list = my_df_list,
#'                 file_list       = my_file_list,
#'                 protect         = FALSE)
#'
#' # Save multiple files and only keep specific variables
#' save_file_multi(data_frame_list = my_df_list,
#'                 file_list       = my_file_list,
#'                 keep_list       = c(sex, age, state),
#'                 protect         = FALSE)
#'
#' # Save multiple files and keep different variables per data frame
#' save_file_multi(data_frame_list = my_df_list,
#'                 file_list       = my_file_list,
#'                 keep_list       = list(c(person_id, first_person),
#'                                        c(NUTS3, income, weight)),
#'                 protect         = FALSE)
#'
#' unlink(c(file1, file2,
#'          file.path(tempdir(), "testfile.fst"),
#'          file.path(tempdir(), "testfile.rds")))
#'
#' @rdname save_load
#'
#' @export
save_file_multi <- function(data_frame_list,
                            file_list,
                            keep_list = NULL,
                            compress  = 100,
                            protect   = TRUE){
    print_start_message()

    if (length(data_frame_list) != length(file_list)){
        print_message("ERROR", "Data frame and file list are of unequal lengths. Saving will be aborted.")

        return(invisible(file_list))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Keep
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # This whole block enables the user to pass different things into keep_list
    keep_eval <- substitute(keep_list)

    # If a single variable name is passed, transform into a list with one character element
    if (is.name(keep_eval)){
        keep_list <- list(get_origin_as_char(keep_list, substitute(keep_list)))
    }
    # If a vector is passed, transform into a list with one character vector
    else if (is.call(keep_eval) && identical(keep_eval[[1]], as.name("c"))){
        keep_list <- list(get_origin_as_char(keep_list, substitute(keep_list)))
    }
    # If a list is passed, make sure it is a list of character vectors
    else if (is.call(keep_eval)){
        # Loop through each list entry and convert vectors to characters
        keep_list <- lapply(as.list(keep_eval)[-1], function(element){
            variables <- as.list(element)[-1]

            # In case of a list of symbols. Variable names are passed without quotation marks.
            if (all(sapply(variables, is.symbol))){
                all.vars(element)
            }
            # In case of a list of character vectors. Variable names are passed with quotation marks.
            else{
                unlist(variables)
            }
        })
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Loading
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Loop through all files and import them one after another
    length_keep <- length(keep_list)

    print_step("MAJOR", "Saving files")

    # Loop through all files and load them in one after another
    for (i in seq_along(file_list)){
        infile  <- data_frame_list[[i]]
        outfile <- file_list[[i]]

        print_step("MINOR", "{outfile}", outfile = outfile)

        # If there are variables to keep
        if (!is.null(keep_eval)){
            # Iterate through the keep list along with the files
            if (i < length_keep){
                vars_to_keep <- keep_list[[i]]
            }
            # If the iterator is bigger than the number of list elements, just
            # use the last list element for every other iteration.
            else{
                vars_to_keep <- keep_list[[length_keep]]
            }
        }
        else{
            vars_to_keep <- NULL
        }

        # Load file and add to list
        save_file(data_frame = infile,
                  path       = dirname(outfile),
                  file       = basename(outfile),
                  keep       = vars_to_keep,
                  compress   = compress,
                  protect    = protect,
                  suppress   = TRUE)
    }

    # Revert used threads
    suppressMessages(fst::threads_fst(NULL))

    print_closing(15)

    invisible(file_list)
}


#' @description
#' [load_file()]: Loads fst and rds files. Provided variables to keep are read in case
#' insensitive and are returned in provided order. Additionally a subset can be defined
#' directly.
#'
#' @examples
#' # Example files
#' fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
#' rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")
#'
#' # Load file
#' my_fst <- load_file(path = dirname(fst_file),
#'                     file = basename(rds_file))
#' my_rds <- load_file(path = dirname(rds_file),
#'                     file = basename(rds_file))
#'
#' # Load file and only keep specific variables
#' # NOTE: Variable names can be written case insensitive. Meaning if a variable
#' #       is stored as "age" and you write "AGE" in keep, the function will find
#' #       the variable and rename it to "AGE".
#' my_fst_keep<- load_file(path = dirname(fst_file),
#'                         file = basename(rds_file),
#'                         keep = c(AGE, INCOME_class, State, weight))
#'
#' # Load file and subset observations
#' my_fst_where <- load_file(path  = dirname(fst_file),
#'                           file  = basename(rds_file),
#'                           where = sex == 1 & age > 65)
#'
#' @rdname save_load
#'
#' @export
load_file <- function(path,
                      file,
                      keep  = NULL,
                      where = NULL,
                      ...){
    print_start_message(suppress = TRUE)
    suppress <- ifelse(is.null(unlist(list(...))), FALSE, TRUE)

    ###########################################################################
    # Error handling
    ###########################################################################

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # File path
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    path <- macro(path)
    file <- macro(file)

    full_path <- file.path(path, file)

    if (!file.exists(full_path)){
        print_message("ERROR", "File does not exist: [full_path]", full_path = full_path)

        return(invisible(NULL))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # File extension
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    extension <- tolower(tools::file_ext(file))

    if (extension == ""){
        print_message("ERROR", c("No file extension provided in <file>. 'fst' and 'rds' are allowed.",
								 "Loading file will be aborted."))

        return(invisible(NULL))
    }

    if (!extension %in% c("fst", "rds")){
        print_message("ERROR", c("Only 'fst' or 'rds' are allowed as file extensions in <file>.",
								 "Loading file will be aborted."))

        return(invisible(NULL))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Keep
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Convert to character vectors
    keep_eval <- substitute(keep)

    if (is.call(keep_eval)){
        keep <- get_origin_as_char(keep, keep_eval)
    }

    ###########################################################################
    # Loading
    ###########################################################################

    if (extension == "fst"){
        suppressMessages(fst::threads_fst(.qol_options[["threads"]]))

        if (!is.null(keep)){
            # Get metadata first to be able to read in the variable names case insensitive
            meta_data      <- suppressMessages(fst::metadata_fst(full_path))
            variable_names <- meta_data[["columnNames"]]

            # Check if variables exist in the meta data
            invalid_keep <- keep[!tolower(keep) %in% tolower(variable_names)]

            if (length(invalid_keep) > 0){
                print_message("WARNING", "Variables not found: [invalid]", invalid = invalid_keep)
            }

            # Use fast match to identify existing columns
            matching_ids <- collapse::fmatch(tolower(keep), tolower(variable_names))
            valid_ids    <- !is.na(matching_ids)

            # Select existing variables
            vars_to_keep  <- variable_names[matching_ids[valid_ids]]
            keep_names    <- keep[valid_ids]

            # Read in existing variables
            data_frame <- suppressMessages(fst::read_fst(full_path,
                                                         columns       = vars_to_keep,
                                                         as.data.table = TRUE))

            # Rename variables to the names provided in keep
            data_frame |> data.table::setnames(vars_to_keep, keep_names)
        }
        # Without keep read in the whole data set
        else{
            data_frame <- suppressWarnings(suppressMessages(
                fst::read_fst(full_path, as.data.table = TRUE)))
        }
    }
    if (extension == "rds"){
        # RDS format is not able to read in only specific variables. Meaning the
        # whole file has to be read in first.
        data_frame <- data.table::as.data.table(readRDS(full_path))

        if (!is.null(keep)){
            variable_names <- names(data_frame)

            # Check if variables exist in the meta data
            invalid_keep <- variable_names[!tolower(keep) %in% tolower(variable_names)]

            if (length(invalid_keep) > 0){
                print_message("WARNING", "Variables not found: [invalid]", invalid = invalid_keep)
            }

            # Use fast match to identify existing columns
            matching_ids <- collapse::fmatch(tolower(keep), tolower(variable_names))
            valid_ids    <- !is.na(matching_ids)

            # Read in existing variables
            vars_to_keep  <- variable_names[matching_ids[valid_ids]]
            keep_names    <- keep[valid_ids]
            invalid_names <- keep[is.na(vars_to_keep)]

            if (length(invalid_names) > 0){
                print_message("WARNING", "Variables not found: [invalid]", invalid = missing)
            }

            # Only keep the desired variables and rename variables to the names provided in keep
            data_frame <- data_frame |>
                collapse::fselect(vars_to_keep) |>
                data.table::setcolorder(vars_to_keep)

            data_frame |> data.table::setnames(vars_to_keep, keep_names)
        }
    }

    data_frame <- suppressMessages(data_frame |> if.(where))

    # Revert used threads
    suppressMessages(fst::threads_fst(NULL))

    if (extension == "fst"){
        print_closing(5, suppress = suppress)
    }
    else{
        print_closing(15, suppress = suppress)
    }

    invisible(data_frame)
}


#' @description
#' [load_file_multi()]: Loads multiple files and stores them into a list or directly
#' stacks the files into one data frame.
#'
#' @param stack_files TRUE by default. Stacks data frames after loading them. If FALSE,
#' returns all data frames in a list.
#'
#' @examples
#' # Load multiple files and stack them
#' stack_files <- load_file_multi(c(fst_file, rds_file))
#'
#' # Load multiple files and output them in a list
#' list_files <- load_file_multi(file_list   = c(fst_file, rds_file),
#'                               stack_files = FALSE)
#'
#' # Load multiple files and only keep specific variables
#' all_files_keep <- load_file_multi(file_list = c(fst_file, rds_file),
#'                                   keep_list = c(Sex, AGE, stAte))
#'
#' # Load multiple files and keep different variables per data frame
#' all_files_diff <- load_file_multi(file_list = c(fst_file, rds_file),
#'                                   keep_list = list(c(Person_ID, First_Person),
#'                                                    c(nuts3, Income, WEIGHT)))
#'
#' @rdname save_load
#'
#' @export
load_file_multi <- function(file_list,
                            keep_list   = NULL,
                            stack_files = TRUE){
    print_start_message()

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Keep
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # This whole block enables the user to pass different things into keep_list
    keep_eval <- substitute(keep_list)

    # If a single variable name is passed, transform into a list with one character element
    if (is.name(keep_eval)){
        keep_list <- list(get_origin_as_char(keep_list, substitute(keep_list)))
    }
    # If a vector is passed, transform into a list with one character vector
    else if (is.call(keep_eval) && identical(keep_eval[[1]], as.name("c"))){
        keep_list <- list(get_origin_as_char(keep_list, substitute(keep_list)))
    }
    # If a list is passed, make sure it is a list of character vectors
    else if (is.call(keep_eval)){
        # Loop through each list entry and convert vectors to characters
        keep_list <- lapply(as.list(keep_eval)[-1], function(element){
            variables <- as.list(element)[-1]

            # In case of a list of symbols. Variable names are passed without quotation marks.
            if (all(sapply(variables, is.symbol))){
                all.vars(element)
            }
            # In case of a list of character vectors. Variable names are passed with quotation marks.
            else{
                unlist(variables)
            }
        })
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Loading
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Loop through all files and import them one after another
    length_keep <- length(keep_list)
    result_list <- list()

    print_step("MAJOR", "Loading files")

    # Loop through all files and load them in one after another
    for (i in seq_along(file_list)){
        infile <- file_list[[i]]

        print_step("MINOR", "{infile}", infile = infile)

        # If there are variables to keep
        if (!is.null(keep_eval)){
            # Iterate through the keep list along with the files
            if (i < length_keep){
                vars_to_keep <- keep_list[[i]]
            }
            # If the iterator is bigger than the number of list elements, just
            # use the last list element for every other iteration.
            else{
                vars_to_keep <- keep_list[[length_keep]]
            }
        }
        else{
            vars_to_keep <- NULL
        }

        # Load file and add to list
        result_list[[basename(infile)]] <- load_file(path = dirname(infile),
                                                     file = basename(infile),
                                                     keep = vars_to_keep,
                                                     suppress = TRUE)
    }

    # Stack data frames
    if (stack_files){
        print_step("MAJOR", "Stacking files")

        result_list <- suppressMessages(stack_data(result_list))
    }

    # Revert used threads
    suppressMessages(fst::threads_fst(NULL))

    print_closing(15)

    invisible(result_list)
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

dummy_df1 <- dummy_data(100)
dummy_df2 <- dummy_data(100)
dummy_df3 <- dummy_data(100)
df_list   <- list("file1" = dummy_data(10),
                  "file2" = dummy_data(10),
                  "file3" = dummy_data(10))

external_path <- system.file("extdata",  package = "qol")


# Stack data frames
new_df1 <- stack_data(dummy_df1, dummy_df2, dummy_df3)
new_df2 <- stack_data(dummy_df1, dummy_df2, dummy_df3, compress = "factor")

expect_equal(collapse::fnrow(new_df1), 300, info = "Stack data fames")
expect_equal(collapse::fnrow(new_df2), 300, info = "Stack data fames")
expect_equal(class(new_df1[["education"]]), "character", info = "Stack data fames")
expect_equal(class(new_df2[["education"]]), "factor", info = "Stack data fames")


# Stack data frames with id column
new_df <- stack_data(dummy_df1, dummy_df2, dummy_df3, id = TRUE)

expect_equal(max(new_df[["ID"]]), 3, info = "Stack data fames with id column")


# Stack mix of data frames and list of data frames
new_df <- stack_data(dummy_df1, df_list, dummy_df2, id = TRUE)

expect_equal(collapse::fnrow(new_df), 230, info = "Stack mix of data frames and list of data frames")
expect_true(all(c("dummy_df1", "dummy_df2", "file1", "file2", "file3") %in% collapse::funique(new_df[["ID"]])),
                info = "Stack mix of data frames and list of data frames")


# Retrieve path with libname
my_path <- libname(external_path)

expect_message(print_stack_as_messages("MAJOR"), "Path successfully assigned: ", info = "Retrieve path with libname")
expect_equal(my_path, external_path, info = "Retrieve path with libname")


# Retrieve files from path with libname
my_path <- libname(external_path, get_files = TRUE)

expect_true(all(c("qol_example_data_csv.csv",  "qol_example_data_fst.fst", "qol_example_data_rds.rds",
                  "qol_example_data_txt.txt", "qol_example_data_xlsx.xlsx", "qol_nuts.csv",
                  "qol_tinytest_results.fst", "qol_table.min.css.txt", "qol_table.min.js.txt") %in% names(my_path)),
             info = "Retrieve files from path with libname")


# Retrieve files with specific extensions from path with libname
my_path <- libname(external_path, get_files = TRUE, extensions = c("txt", "xlsx"))

expect_true(all(c("qol_example_data_txt.txt", "qol_example_data_xlsx.xlsx", "qol_table.min.css.txt", "qol_table.min.js.txt")
            %in% names(my_path)), info = "Retrieve files with specific extensions from path with libname")

###############################################################################
# Loading
###############################################################################

# Example fst and rds files exist
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

expect_true(file.exists(fst_file), info = "Example fst and rds files exist")
expect_true(file.exists(rds_file), info = "Example fst and rds files exist")


# Loading files with keep doesn't reorder them
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

fst_keep <- load_file(dirname(fst_file), basename(fst_file), keep = c(Sex, aGe, STATE))
rds_keep <- load_file(dirname(rds_file), basename(rds_file), keep = c(Sex, aGe, STATE))

expect_equal(names(fst_keep), c("Sex", "aGe", "STATE"), info = "Loading files with keep renames variables and reorders them")
expect_equal(names(rds_keep), c("Sex", "aGe", "STATE"), info = "Loading files with keep renames variables and reorders them")


# Loading files with keep_var_order keeps variables in order
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

fst_keep <- load_file(dirname(fst_file), basename(fst_file), keep = c(Sex, aGe, STATE), keep_var_order = TRUE)
rds_keep <- load_file(dirname(rds_file), basename(rds_file), keep = c(Sex, aGe, STATE), keep_var_order = TRUE)

expect_equal(names(fst_keep), c("STATE", "aGe", "Sex"), info = "Loading files with keep_var_order keeps variables in order")
expect_equal(names(rds_keep), c("STATE", "aGe", "Sex"), info = "Loading files with keep_var_order keeps variables in order")


# Loading files with named list renames variables
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")

fst_keep <- load_file(dirname(fst_file), basename(fst_file), keep = c(Sex = var1, aGe = var2, STATE = var3))

expect_equal(names(fst_keep), c("var1", "var2", "var3"), info = "Loading files with named list renames variables")


# Loading files with subset
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

fst_keep <- load_file(dirname(fst_file), basename(fst_file), where = first_person == 1)
rds_keep <- load_file(dirname(rds_file), basename(rds_file), where = first_person == 1)

expect_equal(collapse::funique(fst_keep[["first_person"]]), 1, info = "Loading files with subset")
expect_equal(collapse::funique(rds_keep[["first_person"]]), 1, info = "Loading files with subset")


# Loading files with subset works on original and renamed variables
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

fst_keep <- load_file(dirname(fst_file), basename(fst_file), keep = c(Sex = var1, aGe = var2), where = Sex  == 1)
rds_keep <- load_file(dirname(rds_file), basename(rds_file), keep = c(Sex = var1, aGe = var2), where = var1 == 1)

expect_equal(collapse::funique(fst_keep[["var1"]]), 1, info = "Loading files with subset works on original and renamed variables")
expect_equal(collapse::funique(rds_keep[["var1"]]), 1, info = "Loading files with subset works on original and renamed variables")


# Loading files with subset can handle SAS style conditions
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")

fst_keep <- load_file(dirname(fst_file), basename(fst_file), where = "15 <= age < 65 and sex == 2")

expect_equal(collapse::funique(fst_keep[["sex"]]), 2, info = "Loading files with subset can handle SAS style conditions")
expect_false(any(00:14  %in% fst_keep[["age"]]), info = "Loading files with subset can handle SAS style conditions")
expect_false(any(65:100 %in% fst_keep[["age"]]), info = "Loading files with subset can handle SAS style conditions")


# Loading files by reference returns an fst_table
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")

fst_ref <- load_file(dirname(fst_file), basename(fst_file), by_reference = TRUE)

expect_true(inherits(fst_ref, "fst_table"), info = "Loading files by reference returns an fst_table")


# Loading multiple files and stack them
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

fst_df <- load_file(dirname(fst_file), basename(fst_file))
rds_df <- load_file(dirname(rds_file), basename(rds_file))
all_df <- load_file_multi(c(fst_file, rds_file))

expect_equal(names(fst_df), names(all_df), info = "Loading multiple files and stack them")
expect_equal(collapse::fnrow(fst_df) + collapse::fnrow(rds_df),
             collapse::fnrow(all_df), info = "Loading multiple files and stack them")


# Loading multiple files and return as list
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

fst_df <- load_file(dirname(fst_file), basename(fst_file))
rds_df <- load_file(dirname(rds_file), basename(rds_file))
all_df <- load_file_multi(c(fst_file, rds_file), stack_files = FALSE)

expect_equal(names(fst_df), names(all_df[[1]]), info = "Loading multiple files and return as list")
expect_equal(fst_df, all_df[[1]], info = "Loading multiple files and return as list")
expect_equal(rds_df, all_df[[2]], info = "Loading multiple files and return as list")


# Loading multiple files with keep and stack them
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

fst_df <- load_file(dirname(fst_file), basename(fst_file), keep = c("sex", "age"))
rds_df <- load_file(dirname(rds_file), basename(rds_file), keep = c("income", "state"))
all_df <- load_file_multi(c(fst_file, rds_file), keep_list = list(c("sex", "age"), c("income", "state")))

expect_equal(names(all_df), c("sex", "age", "income", "state"), info = "Loading multiple files with keep and stack them")
expect_equal(collapse::fnrow(fst_df) + collapse::fnrow(rds_df),
             collapse::fnrow(all_df), info = "Loading multiple files with keep and stack them")


# Loading multiple files with keep and return as list
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

fst_df <- load_file(dirname(fst_file), basename(fst_file), keep = c("sex", "age"))
rds_df <- load_file(dirname(rds_file), basename(rds_file), keep = c("income", "state"))
all_df <- load_file_multi(c(fst_file, rds_file),
                          keep_list   = list(c("sex", "age"), c("income", "state")),
                          stack_files = FALSE)

expect_equal(names(fst_df), names(all_df[[1]]), info = "Loading multiple files with keep and return as list")
expect_equal(fst_df, all_df[[1]], info = "Loading multiple files with keep and return as list")
expect_equal(rds_df, all_df[[2]], info = "Loading multiple files with keep and return as list")


# Loading multiple files and rename variables
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

all_df <- load_file_multi(c(fst_file, rds_file),
                          keep_list   = list(c("sex"    = var1, "age"   = var2),
                                             c("income" = var3, "state" = var4)),
                          stack_files = FALSE)

expect_equal(names(all_df[[1]]), c("var1", "var2"), info = "Loading multiple files and rename variables")
expect_equal(names(all_df[[2]]), c("var3", "var4"), info = "Loading multiple files and rename variables")

###############################################################################
# Saving
###############################################################################

# Saving file works correctly
fst_file <- tempfile(fileext = ".fst")
rds_file <- tempfile(fileext = ".rds")
on.exit(unlink(c(fst_file, rds_file)), add = TRUE)

dummy_df1 |> save_file(dirname(fst_file), basename(fst_file))
dummy_df1 |> save_file(dirname(rds_file), basename(rds_file))

expect_true(file.exists(fst_file), info = "Saving file works correctly")
expect_true(file.exists(rds_file), info = "Saving file works correctly")

fst_df <- load_file(dirname(fst_file), basename(fst_file))
rds_df <- load_file(dirname(rds_file), basename(rds_file))

expect_equal(dummy_df1, fst_df, info = "Saving file works correctly")
expect_equal(dummy_df1, rds_df, info = "Saving file works correctly")


# Saving file with keep works correctly
fst_file <- tempfile(fileext = ".fst")
rds_file <- tempfile(fileext = ".rds")
on.exit(unlink(c(fst_file, rds_file)), add = TRUE)

dummy_df1 |> save_file(dirname(fst_file), basename(fst_file), keep = c(sex, age, state))
dummy_df1 |> save_file(dirname(rds_file), basename(rds_file), keep = c(sex, age, state))

expect_true(file.exists(fst_file), info = "Saving file with keep works correctly")
expect_true(file.exists(rds_file), info = "Saving file with keep works correctly")

fst_df <- load_file(dirname(fst_file), basename(fst_file))
rds_df <- load_file(dirname(rds_file), basename(rds_file))

expect_equal(names(fst_df), c("sex", "age", "state"), info = "Saving file with keep works correctly")
expect_equal(names(rds_df), c("sex", "age", "state"), info = "Saving file with keep works correctly")


# Saving files with named list renames variables
fst_file <- tempfile(fileext = ".fst")
rds_file <- tempfile(fileext = ".rds")
on.exit(unlink(c(fst_file, rds_file)), add = TRUE)

dummy_df1 |> save_file(dirname(fst_file), basename(fst_file), keep = c(sex = var1, age = var2, state = var3))
dummy_df1 |> save_file(dirname(rds_file), basename(rds_file), keep = c(sex = var1, age = var2, state = var3))

fst_df <- load_file(dirname(fst_file), basename(fst_file))
rds_df <- load_file(dirname(rds_file), basename(rds_file))

expect_equal(names(fst_df), c("var1", "var2", "var3"), info = "Saving files with named list renames variables")
expect_equal(names(rds_df), c("var1", "var2", "var3"), info = "Saving files with named list renames variables")


# Saving file with subset works correctly
fst_file <- tempfile(fileext = ".fst")
rds_file <- tempfile(fileext = ".rds")
on.exit(unlink(c(fst_file, rds_file)), add = TRUE)

dummy_df1 |> save_file(dirname(fst_file), basename(fst_file), where = first_person == 1)
dummy_df1 |> save_file(dirname(rds_file), basename(rds_file), where = first_person == 1)

expect_true(file.exists(fst_file), info = "Saving file with subset works correctly")
expect_true(file.exists(rds_file), info = "Saving file with subset works correctly")

fst_df <- load_file(dirname(fst_file), basename(fst_file))
rds_df <- load_file(dirname(rds_file), basename(rds_file))

expect_equal(collapse::funique(fst_df[["first_person"]]), 1, info = "Saving file with subset works correctly")
expect_equal(collapse::funique(rds_df[["first_person"]]), 1, info = "Saving file with subset works correctly")


# Saving files with subset works on original and renamed variables
fst_file <- tempfile(fileext = ".fst")
rds_file <- tempfile(fileext = ".rds")
on.exit(unlink(c(fst_file, rds_file)), add = TRUE)

dummy_df1 |> save_file(dirname(fst_file), basename(fst_file), keep = c(sex = var1, age = var2), where = sex  == 1)
dummy_df1 |> save_file(dirname(rds_file), basename(rds_file), keep = c(sex = var1, age = var2), where = var1 == 1)

fst_df <- load_file(dirname(fst_file), basename(fst_file))
rds_df <- load_file(dirname(rds_file), basename(rds_file))

expect_equal(collapse::funique(fst_df[["var1"]]), 1, info = "Saving files with subset works on original and renamed variables")
expect_equal(collapse::funique(rds_df[["var1"]]), 1, info = "Saving files with subset works on original and renamed variables")


# Loading files with subset can handle SAS style conditions
fst_file <- tempfile(fileext = ".fst")
on.exit(unlink(fst_file), add = TRUE)

dummy_df1 |> save_file(dirname(fst_file), basename(fst_file), where = "15 <= age < 65 and sex == 2")

fst_df <- load_file(dirname(fst_file), basename(fst_file))

expect_equal(collapse::funique(fst_df[["sex"]]), 2, info = "Saving files with subset can handle SAS style conditions")
expect_false(any(00:14  %in% fst_df[["age"]]), info = "Saving files with subset can handle SAS style conditions")
expect_false(any(65:100 %in% fst_df[["age"]]), info = "Saving files with subset can handle SAS style conditions")


# Saving multiple files
fst_file <- tempfile(fileext = ".fst")
rds_file <- tempfile(fileext = ".rds")
on.exit(unlink(c(fst_file, rds_file)), add = TRUE)

save_file_multi(data_frame_list = list(dummy_df1, dummy_df2),
                file_list       = c(fst_file, rds_file))

expect_true(file.exists(fst_file), info = "Saving multiple files")
expect_true(file.exists(rds_file), info = "Saving multiple files")


# Saving multiple files with keep
fst_file <- tempfile(fileext = ".fst")
rds_file <- tempfile(fileext = ".rds")
on.exit(unlink(c(fst_file, rds_file)), add = TRUE)

save_file_multi(data_frame_list = list(dummy_df1, dummy_df2),
                file_list       = c(fst_file, rds_file),
                keep_list       = list(c("sex", "age"), c("income", "state")))

expect_true(file.exists(fst_file), info = "Saving multiple files with keep")
expect_true(file.exists(rds_file), info = "Saving multiple files with keep")

all_df <- load_file_multi(c(fst_file, rds_file), stack_files = FALSE)

expect_equal(names(all_df[[1]]), c("sex", "age"), info = "Saving multiple files with keep")
expect_equal(names(all_df[[2]]), c("income", "state"), info = "Saving multiple files with keep")

###############################################################################
# Warning checks
###############################################################################

# Loading throws warning when variables are not in loaded file
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

load_file(dirname(fst_file), basename(fst_file), keep = "test")
expect_warning(print_stack_as_messages("WARNING"), "Variables not found:", info = "Loading throws warning when variables are not in loaded file")

load_file(dirname(rds_file), basename(rds_file), keep = "test")
expect_warning(print_stack_as_messages("WARNING"), "Variables not found:", info = "Loading throws warning when variables are not in loaded file")


# Saving sets fst as file extension if it is otherwise invalid
test_dir <- tempdir()

file <- "test_file"

expected_path1 <- file.path(test_dir, paste0(file, "1.fst"))
expected_path2 <- file.path(test_dir, paste0(file, "2.fst"))

on.exit(unlink(c(expected_path1, expected_path2)), add = TRUE)

dummy_df1 |> save_file(test_dir, paste0(file, "1"))
expect_warning(print_stack_as_messages("WARNING"), "No file extension provided in <file>. 'fst' will be used.", info = "Saving sets fst as file extension if it is otherwise invalid")

dummy_df1 |> save_file(test_dir, paste0(file, "2.test"))
expect_warning(print_stack_as_messages("WARNING"), "Only 'fst' or 'rds' are allowed as file extensions in <file>.", info = "Saving sets fst as file extension if it is otherwise invalid")

expect_true(file.exists(expected_path1), info = "Saving sets fst as file extension if it is otherwise invalid")
expect_true(file.exists(expected_path2), info = "Saving sets fst as file extension if it is otherwise invalid")

###############################################################################
# Abort checks
###############################################################################

# Abort libname on invalid path
my_path <- libname("Test")

expect_error(print_stack_as_messages("ERROR"), "Path does not exist: ", info = "Abort libname on invalid path")


# Loading aborts with invalid file path
file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")

load_file(dirname(file), "test.fst")
expect_error(print_stack_as_messages("ERROR"), "File does not exist:", info = "Loading aborts with invalid file path")


# Loading aborts with invalid file extension
temp_file1 <- tempfile(fileext = "")
temp_file2 <- tempfile(fileext = ".test")
write("test", temp_file1)
write("test", temp_file2)

load_file(dirname(temp_file1), basename(temp_file1))
expect_error(print_stack_as_messages("ERROR"), "No file extension provided in <file>.", info = "Loading aborts with invalid file extension")

load_file(dirname(temp_file2), basename(temp_file2))
expect_error(print_stack_as_messages("ERROR"), "Only 'fst' or 'rds' are allowed as file extensions in <file>.", info = "Loading aborts with invalid file extension")
on.exit(unlink(c(temp_file1, temp_file2)), add = TRUE)


# Saving file has write protection
fst_file <- tempfile(fileext = ".fst")
rds_file <- tempfile(fileext = ".rds")
on.exit(unlink(c(fst_file, rds_file)), add = TRUE)

dummy_df1 |> save_file(dirname(fst_file), basename(fst_file))
dummy_df1 |> save_file(dirname(rds_file), basename(rds_file))

expect_true(file.exists(fst_file), info = "Saving file has write protection")
expect_true(file.exists(rds_file), info = "Saving file has write protection")

dummy_df1 |> save_file(dirname(fst_file), basename(fst_file))
expect_error(print_stack_as_messages("ERROR"), "File already exists:", info = "Saving file has write protection")

dummy_df1 |> save_file(dirname(rds_file), basename(rds_file))
expect_error(print_stack_as_messages("ERROR"), "File already exists:", info = "Saving file has write protection")


# Saving file aborts with invalid keep variables
fst_file <- tempfile(fileext = ".fst")
on.exit(unlink(fst_file), add = TRUE)

dummy_df1 |> save_file(dirname(fst_file), basename(fst_file), keep = "test")

expect_error(print_stack_as_messages("ERROR"), "The provided variables to <keep>", info = "Saving file aborts with invalid keep variables")

expect_false(file.exists(fst_file), info = "Saving file aborts with invalid keep variables")


# Saving multiple files aborts if data frame and file list are of unequal lengths
fst_file <- tempfile(fileext = ".fst")
rds_file <- tempfile(fileext = ".rds")
on.exit(unlink(c(fst_file, rds_file)), add = TRUE)

save_file_multi(data_frame_list = list(dummy_df1), file_list = c(fst_file, rds_file))

expect_error(print_stack_as_messages("ERROR"), "Data frame and file list are of unequal lengths. Saving will be aborted.", info = "Saving multiple files aborts if data frame and file list are of unequal lengths")

expect_false(file.exists(fst_file), info = "Saving multiple files aborts if data frame and file list are of unequal lengths")
expect_false(file.exists(rds_file), info = "Saving multiple files aborts if data frame and file list are of unequal lengths")


set_no_print()

#' Monitor Time Consumption
#'
#' @description
#' The monitor functions offer a simple way to keep track of timings and visualize them in charts.
#' If used throughout a longer syntax it is useful to identify bottlenecks or just get a better feeling
#' which passages take more time than others.
#'
#' @name monitor
#'
#' @param monitor_df A data table in which the delta times with their respective section names are stored.
#' @param section A named section for which to store delta times.
#' @param group Optionally pass a broader group name to be able to plot summarised delta times in addition to the detailed ones.
#' "Total" as default value.
#'
#' @return
#' Returns a small data table with section-, group-names and corresponding delta times.
#'
#' @examples
#' # Example data frame
#' monitor_df <- NULL |> monitor_start("Generate data frame", "Preparation")
#'
#' my_data <- dummy_data(1000)
#'
#' # Formats
#' monitor_df <- monitor_df |> monitor_next("Create formats", "Preparation")
#'
#' age. <- discrete_format(
#'     "Total"          = 0:100,
#'     "under 18"       = 0:17,
#'     "18 to under 25" = 18:24,
#'     "25 to under 55" = 25:54,
#'     "55 to under 65" = 55:65,
#'     "65 and older"   = 65:100)
#'
#' sex. <- discrete_format(
#'     "Total"  = 1:2,
#'     "Male"   = 1,
#'     "Female" = 2)
#'
#' # Evaluations
#' monitor_df <- monitor_df |> monitor_next("Nested summarise", "Summarise")
#'
#' all_nested <- my_data |>
#'     summarise_plus(class      = c(year, sex, age),
#'                    values     = income,
#'                    statistics = c("sum", "pct_group", "pct_total", "sum_wgt", "freq"),
#'                    formats    = list(sex = "sex.", age = "age."),
#'                    weight     = weight,
#'                    nesting    = "deepest",
#'                    na.rm      = TRUE)
#'
#' monitor_df <- monitor_df |> monitor_next("All summarise", "Summarise")
#'
#' all_possible <- my_data |>
#'     summarise_plus(class      = c(year, sex, age),
#'                    values     = c(probability),
#'                    statistics = c("sum", "p1", "p99", "min", "max", "freq", "freq_g0"),
#'                    formats    = list(sex    = "sex.",
#'                                      age    = "age."),
#'                    weight     = weight,
#'                    nesting    = "all",
#'                    na.rm      = TRUE)
#'
#' monitor_df <- monitor_df |> monitor_end()
#'
#' # For detailed plot
#' monitor_df |> monitor_plot()
#'
#' # For summarised plot
#' monitor_df |> monitor_plot(by = "group")
#'
#' # NOTE: The more complex functions in this package have a detailed monitoring
#' #       integrated which can be viewed by setting the argument 'monitor' to TRUE.
#'
#' @rdname monitor
#'
#' @keywords internal
NULL


#' @description
#' [monitor_start()] starts a new timing and adds this as an observation to the monitoring
#' data table. Pass NULL as monitor_df if you call the function for the first time to create a new
#' monitoring data table.
#'
#' @rdname monitor
#'
#' @export
monitor_start <- function(monitor_df, section, group = "Total"){
    if (is.null(monitor_df)){
        monitor_df <- data.table::data.table()
    }

    # Set up starting time in a new row of monitoring data frame
    monitor_df[collapse::fnrow(monitor_df) + 1, "group"]   <- as.character(group)
    monitor_df[collapse::fnrow(monitor_df),     "section"] <- as.character(section)
    monitor_df[collapse::fnrow(monitor_df),     "start"]   <- as.numeric(Sys.time())

    monitor_df
}


#' @description
#' [monitor_end()] ends the current timing and calculates corresponding delta.
#'
#' @rdname monitor
#'
#' @export
monitor_end <- function(monitor_df){
    # Set end time in last timed row
    monitor_df[collapse::fnrow(monitor_df), "end"]   <- as.numeric(Sys.time())
    monitor_df[collapse::fnrow(monitor_df), "delta"] <- monitor_df[collapse::fnrow(monitor_df), "end"] - monitor_df[collapse::fnrow(monitor_df), "start"]

    monitor_df
}


#' @description
#' [monitor_next()] ends the current timing and calculates corresponding delta. In addition directly starts
#' a new timing for a new section.
#'
#' @rdname monitor
#'
#' @export
monitor_next <- function(monitor_df, section, group = "Total"){
    # End current section and directly start new one
    monitor_df <- monitor_df |> monitor_end()
    monitor_df <- monitor_df |> monitor_start(section, group)

    monitor_df
}


#' @description
#' [monitor_plot()] outputs two charts to visualize the saved delta times.
#'
#' @param by Use "section" for a detailed plot and "group" for summarised categories.
#' @param draw_plot Conditionally draw plots. TRUE by default.
#'
#' @rdname monitor
#'
#' @export
monitor_plot <- function(monitor_df, by = "section", draw_plot = TRUE){
    if (!draw_plot){
        return(invisible(monitor_df))
    }

    old_par <-graphics:: par(no.readonly = TRUE)
    on.exit(graphics::par(old_par))

    monitor_df[monitor_df[["group"]] == "Calc(pseudo_group)", "group"] <- "Calc(total)"

    if (by == "group"){
        label_levels <- monitor_df[["group"]] |>
            unlist(use.names = FALSE) |>
            collapse::funique() |>
            collapse::na_omit()

        monitor_df[["group"]] <- factor(
            monitor_df[["group"]],
            levels = label_levels,
            ordered = TRUE)

        monitor_df <- monitor_df |>
            collapse::fgroup_by("group") |>
            collapse::fsummarise(delta = collapse::fsum(delta))
    }
    else{
        by <- "section"

        label_levels <- monitor_df[["section"]] |>
            unlist(use.names = FALSE) |>
            collapse::funique() |>
            collapse::na_omit()

        monitor_df[["section"]] <- factor(
            monitor_df[["section"]],
            levels = label_levels,
            ordered = TRUE)

        monitor_df <- monitor_df |>
            collapse::fgroup_by("section") |>
            collapse::fsummarise(delta = collapse::fsum(delta))
    }

    # Calculate total time consumption
    total_time <- round_values(collapse::fsum(monitor_df[["delta"]]), 3)

    # Setup plot dimensions
    graphics::par(las      = 2,             # Vertical labels
                  mfrow    = c(2, 1),       # Two rows, one column on page
                  cex.main = 0.8,           # Smaller font size for title
                  cex.lab  = 0.8,           # Smaller font size for labels
                  cex.axis = 0.6,           # Smaller font size for axis
                  mar      = c(6, 4, 2, 1)) # Margins: bottom, left, top, right

    # Line chart
    plot(monitor_df[["delta"]],                            # Values
         main = paste0("Time Consumption - ", total_time), # Main title
         type = "l",                                       # Line without markers
         xlab = "", ylab = "Seconds",                      # Axis texts
         ylim = c(0, max(monitor_df[["delta"]]) * 1.2),    # y-axis starts from zero up to the highest value
         axes = FALSE,                                     # Removes all lines around diagram
         col  = 4,                                         # Blue color for line
         lwd  = 2)                                         # Line thickness

    # Set up axis
    graphics::axis(side   = 1,                         # x-Axis
                   at     = seq_len(collapse::fnrow(monitor_df)), # How many ticks are shown
                   labels = monitor_df[[by]])          # Labels
    graphics::axis(side   = 2)                         # Show y-Axis as stated in plot

    # Stacked bar chart 100 %
    percentages <- monitor_df[["delta"]] / sum(monitor_df[["delta"]])

    # Setup plot dimensions
    color_palette <- grDevices::rainbow(collapse::fnrow(monitor_df))

    graphics::par(mar = c(5, 1, 2, 1))              # Margins: bottom, left, top, right

    graphics::barplot(as.matrix(percentages),       # Convert to matrix for stacked bars
                      xlab  = "", ylab = "",        # Axis texts
                      axes  = FALSE,                # Removes all lines around diagram
                      horiz = TRUE,                 # Bars
                      col   = color_palette,        # Rainbow colors
                      main = "Time Distribution in Percent") # Different colors

    # Set up legend
    graphics::legend("bottomleft",              # General legend position
                     legend = monitor_df[[by]], # Values
                     fill   = color_palette,    # Different colors
                     cex    = 0.7,              # Smaller font size
                     ncol   = 3,                # Number of horizontal categories
                     inset  = c(0, -0.4),       # Move legend slightly down
                     xpd    = TRUE)             # Legend can be drawn outside th chart area

    invisible(monitor_df)
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

dummy_df <- dummy_data(100)


# Starting monitoring creates new data table and variables
monitor_df <- NULL |> monitor_start("Unit test")

expect_equal(nrow(monitor_df), 1, info = "Starting monitoring creates new data table and variables")
expect_equal(ncol(monitor_df), 3, info = "Starting monitoring creates new data table and variables")
expect_true("Unit test" %in% monitor_df[["section"]], info = "Starting monitoring creates new data table and variables")


# Ending monitoring stores system time in current observation and calculates delta
monitor_df <- NULL |> monitor_start("Unit test")
monitor_df <- monitor_df |> monitor_end()

expect_equal(nrow(monitor_df), 1, info = "Ending monitoring stores system time in current observation and calculates delta")
expect_equal(ncol(monitor_df), 5, info = "Ending monitoring stores system time in current observation and calculates delta")
expect_true("delta" %in% names(monitor_df), info = "Ending monitoring stores system time in current observation and calculates delta")
expect_true("Unit test" %in% monitor_df[["section"]], info = "Ending monitoring stores system time in current observation and calculates delta")


# Next monitoring ends the current one and directly starts a new one
monitor_df <- NULL |> monitor_start("Unit test")
monitor_df <- monitor_df |> monitor_next("Next")
monitor_df <- monitor_df |> monitor_end()

expect_equal(nrow(monitor_df), 2, info = "Next monitoring ends the current one and directly starts a new one")
expect_equal(ncol(monitor_df), 5, info = "Next monitoring ends the current one and directly starts a new one")
expect_true("Unit test" %in% monitor_df[["section"]], info = "Next monitoring ends the current one and directly starts a new one")
expect_true("Next" %in% monitor_df[["section"]], info = "Next monitoring ends the current one and directly starts a new one")


# Plotting monitoring resets par to the values before
old_par <- graphics::par()

monitor_df <- NULL |> monitor_start("Unit test")
monitor_df <- monitor_df |> monitor_next("Next")
monitor_df <- monitor_df |> monitor_end()

monitor_df |> monitor_plot()
monitor_df |> monitor_plot(by = "group")

expect_identical(old_par, graphics::par(), info = "Plotting monitoring resets par to the values before")


set_no_print()

#' Split Data Frame By Variable Expressions Or Condition
#'
#' @description
#' Split up a data frame based on variable expressions or on conditions to receive
#' multiple smaller data frames. Both possibilities can be used at the same time.
#'
#' @param data_frame A data frame which should be split up into multiple data frames.
#' @param ... Pass in one or multiple variables and/or conditions on which the provided
#' data frame should be splitted.
#' @param formats A list in which is specified which formats should be applied to which
#' variables.
#' @param inverse Uses the inverse conditions to split up the data frame.
#' @param monitor FALSE by default. If TRUE, outputs two charts to visualize the
#' functions time consumption.
#'
#' @details
#' [split_by()] is based on the explicit Output from 'SAS'. With the Output function
#' one can - among other things - explicitly tell 'SAS' which observation to output into
#' which data set. Which enables the user to output one observation into one or multiple
#' data sets.
#'
#' Instead of subsetting the same data frame multiple times manually, you can subset it
#' multiple times at once with this function.
#'
#' @return Returns a list of data frames split by variable expressions and/or conditions.
#' The lists names are the variable expressions or conditions.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Split by variable expressions
#' split_var_df <- my_data |> split_by(sex)
#'
#' # Split by conditions
#' split_cond_df <- my_data |> split_by(sex == 1 & age <  18,
#'                                      sex == 2 & age >= 18)
#'
#' # Split by condition with inverse group
#' split_inv_df <- my_data |> split_by(sex == 1, inverse = TRUE)
#'
#' # Split by variables and conditions
#' split_combi_df <- my_data |> split_by(state, education,
#'                                       sex == 1, age < 18)
#'
#' # Split by variable expressions using formats
#' state. <- discrete_format(
#'     "Germany"                       = 1:16,
#'     "Schleswig-Holstein"            = 1,
#'     "Hamburg"                       = 2,
#'     "Lower Saxony"                  = 3,
#'     "Bremen"                        = 4,
#'     "North Rhine-Westphalia"        = 5,
#'     "Hesse"                         = 6,
#'     "Rhineland-Palatinate"          = 7,
#'     "Baden-Württemberg"             = 8,
#'     "Bavaria"                       = 9,
#'     "Saarland"                      = 10,
#'     "West"                          = 1:10,
#'     "Berlin"                        = 11,
#'     "Brandenburg"                   = 12,
#'     "Mecklenburg-Western Pomerania" = 13,
#'     "Saxony"                        = 14,
#'     "Saxony-Anhalt"                 = 15,
#'     "Thuringia"                     = 16,
#'     "East"                          = 11:16)
#'
#' split_format_df <- my_data |> split_by(state,
#'                                        formats = list(state = state.))
#'
#' @export
split_by <- function(data_frame,
                     ...,
                     formats = list(),
                     inverse = FALSE,
                     monitor = .qol_options[["monitor"]]){
    # Measure the time
    print_start_message(suppress = TRUE)

    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Preparation", "Preparation")
    #-------------------------------------------------------------------------#

    # Convert to character in case of single variables
    conditions <- tryCatch({
        # Force evaluation to see if it exists
        dots_to_char(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    # Get actual condition or a mix of variables and conditions
    if (is.null(conditions)){
        conditions <- as.list(substitute(list(...)))[-1]
    }

    # If only variables are provided, check if they are part of the data frame
    if (is.character(conditions)){
        conditions <- data_frame |> part_of_df(conditions)
    }

    # Split data frame by conditions and store the split data frames into list
    data_list <- list()

    # Apply formats if there are any first
    formats_list <- formats

    if (!is_list_of_dfs(formats)){
        formats_list <- as.list(substitute(formats))[-1]
        formats_list <- evaluate_formats(formats_list)
    }

    if (length(formats_list) > 0){
        #-------------------------------------------------------------------------#
        monitor_df <- NULL |> monitor_start("Applying formats", "Formats")
        #-------------------------------------------------------------------------#

        data_frame <- suppressMessages(do.call(recode_multi, c(list(data_frame), formats)))
    }

    for (condition in conditions){
        #-------------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next(deparse(condition), "Condition")
        #-------------------------------------------------------------------------#

        # Split by value if a single variable name is provided
        if (is.name(condition) || is.character(condition)){
            condition <- as.character(condition)

            # Generate new split list with NA entries
            new_entry <- data_frame |> collapse::rsplit(by = data_frame[[condition]])

            if (NA %in% names(new_entry)){
                names(new_entry)[length(new_entry)] <- paste0("NA_", condition)
            }

            # Check for duplicate list entry names
            duplicate_entry <- intersect(names(data_list), names(new_entry))

            if (length(duplicate_entry) > 0){
                print_message("ERROR", c("Variable '[condition]' caused duplicate list entry names. Try to make them unique",
										 "using formats. Splitting will be aborted."), condition = condition)
                return(data_frame)
            }

            data_list <- c(data_list, new_entry)
        }
        # If a condition is provided split by it
        else if (is.language(condition)){
            # Evaluate the condition first to make it usable below
            real_condition <- eval(substitute(condition), envir = data_frame, enclos = parent.frame())

            # Add the subset or the inverse one as list element. The list element name is the condition iteself.
            if (!inverse){
                data_list <- c(data_list,
                               stats::setNames(list(data_frame |> collapse::fsubset(real_condition)), gsub('"', "'", deparse(condition))))
            }
            else{
                data_list <- c(data_list,
                               stats::setNames(list(data_frame |> collapse::fsubset(!real_condition)),
                                               paste0("not (", gsub('"', "'", deparse(condition)), ")")))
            }
        }
        # If something else is given abort
        else{
            print_message("ERROR", "Only single variables or conditions allowed. Splitting will be aborted.")
            return(data_frame)
        }
    }

    print_closing()

    #---------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_end()
    monitor_df |> monitor_plot(draw_plot = monitor)
    #---------------------------------------------------------------------#

    data_list
}

#' Print Styled Tinytest Results
#'
#' @name reporter
#'
#' @description
#' Styles the results of tinytest to get a better visual overview.
#'
#' @param tiny_results The results produced by tinytest.
#' @param utf8 Whether to display complex characters or just plain text.
#'
#' @return
#' Returns the results.
#'
#' @examples
#' # Example results
#' result_file <- system.file("extdata", "qol_tinytest_results.fst", package = "qol")
#' results     <- load_file(dirname(result_file), basename(result_file))
#'
#' # Display results
#' results |> report_test_results()
#'
#' # Normally you would do this:
#' # tinytest::test_all("PATH TO PACKAGE") |> report_test_results()
#' # or
#' # tinytest::test_package("PACKAGE NAME", testdir = "inst/tinytest") |> report_test_results()
#'
#' # To test the whole package with the custom reporter use:
#' # test_package("qol")
#'
#' @rdname reporter
#'
#' @export
report_test_results <- function(tiny_results, utf8 = .qol_messages[["format"]][["utf8"]]){
    result_df <- as.data.frame(tiny_results)

    if (!all(c("result", "file", "call", "first", "short", "diff") %in% names(result_df))){
        print_message("ERROR", "Input file is not a tinytest result list. Reporting will be aborted.")

        return(invisible(tiny_results))
    }

    # Get base summary results
    total  <- collapse::fnrow(result_df)
    passed <- collapse::fsum(result_df[["result"]])
    failed <- total - passed

    # Set up a theme which is used to color the output
    if (utf8){
        symbol_total  <- "\ud83d\udcdd"
        symbol_pass   <- "\u2705"
        symbol_fail  <- "\u274c"
        symbol_script <- "\U{1f4c4} FILE:"
    }
    else{
        symbol_total  <- ""
        symbol_pass   <- ""
        symbol_fail   <- ""
        symbol_script <- "FILE:"
    }

    # Set up custom messages
    set_up_custom_message(ansi_icon = "\u274c",       text_icon = "X", type = ".FAIL",       color = "#C93F3F")
    set_up_custom_message(ansi_icon = "\ud83d\udd0d", text_icon = ">", type = ".PATH",       color = "#BEDD1A")
    set_up_custom_message(ansi_icon = "\u2699\ufe0f", text_icon = "^", type = ".CALL",       color = "#32CD32")
    set_up_custom_message(ansi_icon = "\U0001f4e6",   text_icon = "?", type = ".TYPE",       color = "#8F8DC7")
    set_up_custom_message(ansi_icon = "\u26a0\ufe0f", text_icon = "!", type = ".DIFFERENCE", color = "#FFC90E")

    # Set up main headline
    print_headline("[b][#63C2C9 Results][/b]: [tests] [b]TESTS:[/b] [total] | [pass] [b][#32CD32 PASSED:][/b] [passed] | [fail] [b][#C93F3F FAILED:][/b] [failed]",
                   tests = symbol_total, pass = symbol_pass, fail = symbol_fail,
                   total = total, passed = passed, failed = failed)

    # If there are no failed tests, then return
    if (failed == 0){
        print_headline("[pass] [b]All tests [#32CD32 PASSED][/b]", pass = symbol_pass)

        return(invisible(tiny_results))
    }

    # Loop through all files and print out failed tests per file
    test_files <- collapse::funique(result_df[["file"]])

    for (test_file in test_files){
        # Get number of failed tests in the current file. If there are none, skip the file.
        failed_df      <- result_df |> collapse::fsubset(file == test_file & result == FALSE)
        failed_in_file <- collapse::fnrow(failed_df)

        if (failed_in_file == 0){
            next
        }

        # Get full absolute path to test file
        full_path <- normalizePath(file.path("inst/tinytest", test_file), winslash = "/", mustWork = FALSE)

        # Set the headline for the file
        print_headline("[b][#63C2C9 {script}][/b] [b][test_file][/b]",
                       script = symbol_script, test_file = test_file,
                       line_char = "-")

        # Now loop through all failed tests and print the results per file
        for (i in seq_len(failed_in_file)){
            # Extract the info parameter from the expect call to use this as headline
            # and additionally print a clean call.
            expect_call <- failed_df[["call"]][i]
            info_text   <- regmatches(expect_call, regexec('info\\s*=\\s*["\']([^"\']+)["\']', expect_call))[[1]][2]
            expect_call <- gsub(',\\s*info\\s*=\\s*["\'].*?["\']', '', expect_call)

            # Get the calls line number
            line_number <- failed_df[["first"]][i]

            # Get fail type and the actual difference
            fail_type  <- failed_df[["short"]][i]
            difference <- unlist(strsplit(failed_df[["diff"]][i], "\n"))

            # Print out fail block
            print_message(".FAIL", "[info_text]", info_text = info_text)
            print_message(".PATH", 'Line: [line_number], file.edit("[full_path]")', line_number = line_number, full_path = full_path)
            print_message(".CALL", "[expect_call]", expect_call = expect_call)
            print_message(".TYPE", "[fail_type]", fail_type = fail_type)

            # Print out formatted difference text or as is depending on failure type
            if (startsWith(difference[1], "Found")){
                difference <- substr(difference, 1, getOption("width") * 3)
            }

            print_message(".DIFFERENCE", difference)
            cat("\n")
        }
    }

    print_headline("[b]END OF REPORT[/b]")

    invisible(tiny_results)
}


#' @description
#' [test_package()]: Runs all set up unit tests with tinytest and outputs the results with the
#' custom reporter.
#'
#' @param package_name Name of the package to test.
#' @param multithread FALSE by default. Whether to run tests multithreaded.
#' NOTE: To make this work you have to manually run [devtools::install()] first.
#'
#' @rdname reporter
#'
#' @export
test_package <- function(package_name,
                         multithread = FALSE){
    if (!requireNamespace("tinytest", quietly = TRUE)){
        print_message("NOTE", 'To use this function you have to install "tinytest": install.packages("tinytest")')

        return(invisible(NULL))
    }

    if (multithread){
        results <- tinytest::test_package("qol", testdir = "inst/tinytest", ncpu = .qol_options[["threads"]])
    }
    else{
        results <- tinytest::test_package("qol", testdir = "inst/tinytest", color = TRUE)
    }

    # Clean up the temporary folder after testing
    test_dir <- tempdir()
    unlink(list.files(test_dir, full.names = TRUE), recursive = TRUE)

    results |> report_test_results()
}


#' @description
#' [test_single_file()]: Runs a single unit test file with tinytest.
#'
#' @param file_name Name of the file to test.
#'
#' @rdname reporter
#'
#' @export
test_single_file <- function(file_name){
    if (!requireNamespace("tinytest", quietly = TRUE)){
        print_message("NOTE", 'To use this function you have to install "tinytest": install.packages("tinytest")')

        return(invisible(NULL))
    }

    file_name <- tools::file_path_sans_ext(basename(file_name))

    results <- tinytest::run_test_file(paste0("inst/tinytest/test-", file_name, ".R"))

    # Clean up the temporary folder after testing
    test_dir <- tempdir()
    unlink(list.files(test_dir, full.names = TRUE), recursive = TRUE)

    results |> report_test_results()
}

#' If - Else if - Else Statements
#'
#' @description
#' These functions make if statements more readable. Especially if an if block becomes bigger
#' it can be hard to read with multiple nested if_else statements. With these new functions
#' if blocks can be written like in other languages with a clear and simpler structure. In
#' addition not only for one variable can a new value be assigned, but for multiple.
#'
#' @name if_else
#'
#' @param data_frame A data frame on which to apply an if statement.
#' @param condition The condition on which a value should be passed to a variable.
#' @param ... The assignment of what should happen when condition becomes TRUE.
#'
#' @details
#' The loop you can use within these functions is based on the 'SAS' do-over-loop. This
#' type of loop iterates over every vector that appears in the loop in parallel.
#' Means that in the first iteration all the first vector elements are used, in the
#' second iteration all second elements of every vector, and so on. With this loop
#' you don't have the need to construct an outer loop, but can directly pass in
#' different vectors and let the function handle the loop inside.
#'
#' @return Returns a data frame with conditionally computed variables.
#'
#' @seealso
#' [ifelse()] with multiple conditions and 'SAS' writing style: [ifelse_multi()]
#'
#' The following functions can make use of the [do_if()] filter variables:
#'
#' Conditions: [if.()], [else_if.()], [else.()]
#'
#' Filter Data Frame: [where.()]
#'
#' Create new Variables: [compute.()]
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Call function
#' new_df <- my_data |>
#'          if.(age < 18,             age_group = "under 18") |>
#'     else_if.(age >= 18 & age < 65, age_group = "18 to under 65") |>
#'     else.   (                      age_group = "65 and older")
#'
#' # Or with multiple variables
#' new_df <- my_data |>
#'          if.(age < 18,             age_group = "under 18"      , age_num = 1L) |>
#'     else_if.(age >= 18 & age < 65, age_group = "18 to under 65", age_num = 2L) |>
#'     else.   (                      age_group = "65 and older",   age_num = 3L)
#'
#' # NOTE: As in other languages the following if blocks won't produce the same result.
#' #       if.() will overwrite existing values, while else_if.() will not.
#' state_df <- my_data |>
#'          if.(state == 1, state_a = "State 1") |>
#'     else_if.(state < 11, state_a = "West") |>
#'     else.   (            state_a = "East")
#'
#' state_df <- state_df |>
#'       if.(state == 1, state_b = "State 1") |>
#'       if.(state < 11, state_b = "West") |>
#'     else.(            state_b = "East")
#'
#' # Add multiple variables based on single conditions
#' multi_df <- my_data |>
#'          if.(age < 18 & education == "low",                  var1 = 1,
#'                                                              var2 = TRUE,
#'                                                              var3 = "Category 1") |>
#'     else_if.(age >= 18 & education %in% c("middle", "high"), var1 = 2,
#'                                                              var2 = FALSE,
#'                                                              var3 = "Category 2") |>
#'     else.   (                                                var1 = 3,
#'                                                              var2 = FALSE,
#'                                                              var3 = "Category 3")
#'
#' # Use if.() as a do-over-loop. In this kind of loop all vectors will be
#' # advanced one iteration at a time in parallel.
#' money    <- c("income", "expenses", "balance", "probability")
#' new_vars <- c("var1", "var2", "var3", "var4")
#' result   <- c(1, 2, 3, 4)
#'
#' do_over_df <- my_data |>
#'       if.(money > 0, new_vars = result) |>
#'     else.(           new_vars = 0)
#'
#' # It is also possible to select character expressions based on whether they
#' # start/end with or contain a certain text.
#' text_select_df <- my_data |>
#'     if.(income_class == "01.:",  start    = 1) |>
#'     if.(income_class == ":more", end      = 1) |>
#'     if.(education    == ":i:",   contains = 1)
#'
#' # Select observations by condition instead of generating new variable
#' subset_df <- my_data |> if.(sex == 1)
#'
#' # Explicitly delete observations
#' subset_df <- my_data |> if.(sex == 1, delete)
#'
#' # Select all non NA observations by variable
#' subset_df <- my_data |> if.(sex)
#'
#' # All these functions can be used in a do_if() situation and are aware of
#' # overarching conditions.
#' do_if_df <- my_data |>
#'     do_if(state < 11) |>
#'           if.(age < 18, new_var = 1) |>
#'         else.(          new_var = 2) |>
#'     else_do() |>
#'           if.(age < 18, new_var = 3) |>
#'         else.(          new_var = 4) |>
#'     end_do()
#'
#' @rdname if_else
#'
#' @keywords internal
NULL


#' @description
#' [if.()] always creates a new variable if the given variable name is not part of the given
#' data frame. If there already is a variable with the given name, the existing values will
#' be overwritten if the condition is TRUE.
#'
#' If no new variable is provided, [if.()] will select observations by the given condition
#' instead.
#'
#' @rdname if_else
#'
#' @export
if. <- function(data_frame, condition, ...){
    # Measure the time
    print_start_message(suppress = TRUE)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Early evaluations
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    parent_env <- parent.frame()

    # This step is important to make this function work in a nested situation.
    # Normally condition would be the name of what was last passed as a parameter.
    # If "if.()" is used nested inside a function this can basically be any placeholder.
    # So here we go up the ladder to get the original condition as a call.
    condition <- substitute(condition)

    while (is.name(condition) && exists(as.character(condition), parent_env)){
        condition <- eval(substitute(substitute(expression, parent_env), list(expression = condition)))
    }

    # When the condition is passed as character, then parse it to enable all the
    # SAS like writing styles.
    if (is.character(condition)){
       condition <- parse_conditions(condition, na.rm = FALSE)

       # If a name is returned then a single variable name was passed. In this case
       # revert to character to ensure it is processed right down the road.
       if (is.name(condition)){
           condition <- as.character(condition)
       }
       # Otherwise transform into a call
       else{
           condition <- as.call(condition)
       }
    }

    assignments <- as.list(substitute(list(...)))[-1]

    # Check for "delete" keyword
    flag_delete <- FALSE

    if (length(assignments) == 1 && assignments[[1]] == "delete"){
        flag_delete      <- TRUE
        assignments[[1]] <- NULL
    }

    # The condition and the variable assignments are torn apart here, so that
    # only the unique variable and vector names are captured as characters.
    used_variables <- unique(all.vars(condition)) # Get all assigned variables

    # Get the original contents of vectors if there are any and put them in a list.
    # List names are the symbols names from above and the elements hold the actual contents.
    content_list <- mget(used_variables, envir = parent_env, ifnotfound = list(NULL))

    # Remove invalid empty entries. This is the case if a variable is an original
    # variable name and not a vector of variable names.
    content_list <- Filter(is_valid_vector, content_list)

    list_entry_lengths <- lengths(content_list)

    # If no vector was passed in the condition or the assignments, then evaluate
    # as normal if statement.
    flag_filter    <- FALSE
    condition_list <- list()

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Variable assignment route
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (length(assignments) > 0){
        # If a vector was passed in the condition or the assignments, then evaluate
        # the if statement as a do over loop. Meaning for each vector the same elements
        # are used simultaneously one after another.
        list_entry_lengths <- lengths(content_list)

        if (length(list_entry_lengths) > 0 && length(collapse::funique(list_entry_lengths)) != 1){
            print_message("ERROR", c("Passed vectors are of unequal lengths. All vectors must have an",
									 "equal number of elements. Evaluation will be aborted."))
            return(invisible(data_frame))
        }

        #---------------------------------------------------------------------#
        # Normal evaluation
        #---------------------------------------------------------------------#

        if (length(list_entry_lengths) == 0){
            condition <- translate_condition(condition)
            condition <- eval(condition, envir = data_frame, enclos = parent_env)
            condition[is.na(condition)] <- FALSE

            # Check whether there are additional conditions active via do_if
            full_condition <- data_frame |> combined_condition(condition)

            # Add to call list
            condition_list <- list(full_condition)
        }
        #---------------------------------------------------------------------#
        # Do over evaluation
        #---------------------------------------------------------------------#
        else{
            # Do over loop per element
            for (element in seq_len(list_entry_lengths[1])){
                # Get the respective first elements as symbols in a new list
                replace_list <- lapply(content_list, function(vector){
                    if (is.numeric(vector) || is.logical(vector)){
                        expression <- vector[element]
                    }
                    else{
                        # Check if vector element is part of the data frame to determine
                        # whether it is a variable or a character expression. Convert
                        # it accordingly.
                        expression <- data_frame |> part_of_df(vector[element], check_only = TRUE)

                        if (is.list(expression)){
                            expression <- as.character(vector[element])
                        }
                        else{
                            expression <- as.name(vector[element])
                        }
                    }

                    # Check if there are colons as placeholders. If this is the
                    # case, convert these expressions to character so that they
                    # can be evaluated correctly later on.
                    number_of_colons <- nchar(gsub("[^:]", "", as.character(expression)))

                    if (number_of_colons > 0){
                        as.character(expression)
                    }
                    else{
                        expression
                    }
                })

                # Insert the current variables into the condition
                current_condition <- do.call(substitute, list(condition, replace_list))
                current_condition <- translate_condition(current_condition)
                condition_list[[element]] <- eval(substitute(current_condition), envir = data_frame, enclos = parent_env)

                # Check whether there are additional conditions active via do_if and
                # add to call list
                condition_list[[element]] <- data_frame |> combined_condition(condition_list[[element]])
                condition_list[[element]][is.na(condition_list[[element]])] <- FALSE
            }
        }
    }
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Filtering route
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Filter observations, if there are no assignments given
    else{
        flag_filter <- TRUE

        #---------------------------------------------------------------------#
        # Normal evaluation
        #---------------------------------------------------------------------#

        # If no vector was passed in the condition, then evaluate as normal
        if (length(content_list) == 0){
            condition <- translate_condition(condition)

            # If observations should be deleted, reverse condition before evaluation
            if (flag_delete){
                condition <- call("!", condition)
            }

            condition <- eval(condition, envir = data_frame, enclos = parent_env)

            # Remember rows to tell the user how many rows have been removed
            rows_before <- data_frame |> collapse::fnrow()

            # Evaluate normal condition
            if (is.logical(condition)){
                # Check whether there are additional conditions active via do_if
                full_condition <- data_frame |> combined_condition(condition)

                data_frame <- data_frame |> collapse::fsubset(full_condition)
            }
            # If a single variable name is given like 'age' this will be evaluated as:
            # !is.na(age). So it is basically a short form like 'If age;' in SAS.
            else{
                # If it is a single character variable it has to be checked first if it is part
                # of the data frame. If not, no subsetting will take place.
                if (is.character(condition) && length(condition) == 1){
                    condition <- data_frame |> part_of_df(condition)

                    if (length(condition) == 0){
                        print_message("ERROR", "No variable for subsetting provided. Data frame remains as is.")
                    }
                    else{
                        # Check whether there are additional conditions active via do_if
                        full_condition <- data_frame |> combined_condition(!is.na(data_frame[[condition]]))

                        data_frame <- data_frame |> collapse::fsubset(full_condition)
                    }
                }
                else if (length(condition) != collapse::fnrow(data_frame)){
                    print_message("ERROR", "Only single variables and conditions allowed. Data frame remains as is.")
                }
                # Evaluate single variable
                else{
                    # Check whether there are additional conditions active via do_if
                    full_condition <- data_frame |> combined_condition(!is.na(condition))

                    data_frame <- data_frame |> collapse::fsubset(full_condition)
                }
            }

            # Output info message
            rows_after <- data_frame |> collapse::fnrow()

            print_step("MAJOR", "Removed [removed] observations. Data frame now has [still_there] observations.",
                    removed = format(rows_before - rows_after,
                                     format = "d", decimal.mark = ",", big.mark = ".", scientific = FALSE),
                    still_there = format(rows_after,
                                         format = "d", decimal.mark = ",", big.mark = ".", scientific = FALSE))
        }
        #---------------------------------------------------------------------#
        # Do over evaluation
        #---------------------------------------------------------------------#
        # If a vector was passed in the condition, then evaluate the if statement
        # as a do over loop. Meaning for each vector the same elements are used
        # simultaneously one after another.
        else{
            original_condition <- condition

            for (element in seq_len(list_entry_lengths[1])){
                # Get the respective first elements as symbols in a new list
                replace_list <- lapply(content_list, function(vector){
                    if (is.numeric(vector) || is.logical(vector)){
                        expression <- vector[element]
                    }
                    else{
                        # Check if vector element is part of the data frame to determine
                        # whether it is a variable or a character expression. Convert
                        # it accordingly.
                        expression <- data_frame |> part_of_df(vector[element], check_only = TRUE)

                        if (is.list(expression)){
                            expression <- as.character(vector[element])
                        }
                        else{
                            expression <- as.name(vector[element])
                        }
                    }
                })

                # Evaluate complete assignment first without condition
                condition  <- do.call(substitute, list(original_condition, replace_list))

                condition <- translate_condition(condition)
                condition <- eval(condition, envir = data_frame, enclos = parent_env)
                condition[is.na(condition)] <- FALSE

                # Remember rows to tell the user how many rows have been removed
                rows_before <- data_frame |> collapse::fnrow()

                # Evaluate normal condition
                if (is.logical(condition)){
                    # Check whether there are additional conditions active via do_if
                    full_condition <- data_frame |> combined_condition(condition)

                    data_frame <- data_frame |> collapse::fsubset(full_condition)
                }
                # If a single variable name is given like 'age' this will be evaluated as:
                # !is.na(age). So it is basically a short form like 'If age;' in SAS.
                else{
                    # If it is a single character variable it has to be checked first if it is part
                    # of the data frame. If not, no subsetting will take place.
                    if (is.character(condition) && length(condition) == 1){
                        condition <- data_frame |> part_of_df(condition)

                        if (length(condition) == 0){
                            print_message("ERROR", "No variable for subsetting provided. Data frame remains as is.")
                        }
                        else{
                            # Check whether there are additional conditions active via do_if
                            full_condition <- data_frame |> combined_condition(!is.na(data_frame[[condition]]))

                            data_frame <- data_frame |> collapse::fsubset(full_condition)
                        }
                    }
                    else if (length(condition) != collapse::fnrow(data_frame)){
                        print_message("ERROR", "Only single variables and conditions allowed. Data frame remains as is.")
                    }
                    # Evaluate single variable
                    else{
                        # Check whether there are additional conditions active via do_if
                        full_condition <- data_frame |> combined_condition(!is.na(condition))

                        data_frame <- data_frame |> collapse::fsubset(full_condition)
                    }
                }

                # Output info message
                rows_after <- data_frame |> collapse::fnrow()

                print_step("MAJOR", "Removed [removed] observations. Data frame now has [still_there] observations.",
                           removed = format(rows_before - rows_after,
                                            format = "d", decimal.mark = ",", big.mark = ".", scientific = FALSE),
                           still_there = format(rows_after,
                                                format = "d", decimal.mark = ",", big.mark = ".", scientific = FALSE))
            }
        }
    }

    # Evaluate calculations conditionally. Making use of hidden parameters.
    if (!flag_filter){
        data_frame <- data_frame |> compute.(..., .if_condition    = condition_list,
                                                  .if_parent_frame = parent_env,
                                                  .if_suppressed   = TRUE)
    }

    print_closing()

    data_frame
}


#' @description
#' [else_if.()] only acts if there already is a variable with the given name. Only NA values
#' will get new values if condition is TRUE. The existing values will not be overwritten.
#'
#' @rdname if_else
#'
#' @export
else_if. <- function(data_frame, condition, ...){
    # Measure the time
    print_start_message(suppress = TRUE)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Early evaluations
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    parent_env  <- parent.frame()

    # This step is important to make this function work in a nested situation.
    # Normally condition would be the name of what was last passed as a parameter.
    # If "if.()" is used nested inside a function this can basically be any placeholder.
    # So here we go up the ladder to get the original condition as a call.
    condition <- substitute(condition)

    while (is.name(condition) && exists(as.character(condition), parent_env)){
        condition <- eval(substitute(substitute(expression, parent_env), list(expression = condition)))
    }

    # When the condition is passed as character, then parse it to enable all the
    # SAS like writing styles.
    if (is.character(condition)){
        condition <- parse_conditions(condition, na.rm = FALSE)

        # If a name is returned then a single variable name was passed. In this case
        # revert to character to ensure it is processed right down the road.
        if (is.name(condition)){
            condition <- as.character(condition)
        }
        # Otherwise transform into a call
        else{
            condition <- as.call(condition)
        }
    }

    assignments <- as.list(substitute(list(...)))[-1]

    # The condition and the variable assignments are torn apart here, so that
    # only the unique variable and vector names are captured as characters.
    used_variables <- unique(c(all.vars(condition), # Get all variables from the condition
                               names(assignments), # Get all variables to which a value should be assigned
                               unlist(lapply(assignments, all.names)))) # Get all assigned variables

    # Get the original contents of vectors if there are any and put them in a list.
    # List names are the symbols names from above and the elements hold the actual contents.
    content_list <- mget(used_variables, envir = parent_env, ifnotfound = list(NULL))

    # Remove invalid empty entries. This is the case if a variable is an original
    # variable name and not a vector of variable names.
    content_list <- Filter(is_valid_vector, content_list)

    # If no vector was passed in the condition or the assignments, then evaluate
    # as normal if statement.
    condition_list <- list()

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Variable assignment
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (length(assignments) > 0){
        for (entry in seq_along(assignments)){
            variable    <- names(assignments)[[entry]]
            calculation <- assignments[[variable]]
            calc_text   <- deparse(calculation)

            # Get used variables especially for this condition in combination mit
            # assigned variables.
            used_variables <- unique(c(all.vars(condition), # Get all variables from the condition
                                       variable, # Get all variables to which a value should be assigned
                                       unlist(lapply(calculation, all.names)))) # Get all assigned variables

            #---------------------------------------------------------------------#
            # Normal evaluation
            #---------------------------------------------------------------------#

            # If no vector was passed in the condition or the assignments, then evaluate
            # as normal.
            if (!any(used_variables %in% names(content_list))){
                condition <- translate_condition(condition)
                condition <- eval(condition, envir = data_frame, enclos = parent_env)
                condition[is.na(condition)] <- FALSE

                # This step is important to make this function work in a nested situation.
                # Normally variable would be the name of what was last passed as a parameter.
                # If "if.()" is used nested inside a function this can basically be any placeholder.
                # So here we go up the ladder to get the original name of the variable.
                original_var <- get_origin_symbol(variable)

                # Variable has to exist in data frame
                if (!original_var %in% names(data_frame)){
                    next
                }

                # Check whether there are additional conditions active via do_if
                full_condition <- data_frame |> combined_condition(is.na(data_frame[[original_var]]) & condition)

                # Add to call list
                condition_list <- c(condition_list, list(full_condition))
            }
            #---------------------------------------------------------------------#
            # Do over evaluation
            #---------------------------------------------------------------------#
            # If a vector was passed in the condition or the assignments, then evaluate
            # the if statement as a do over loop. Meaning for each vector the same elements
            # are used simultaneously one after another.
            else{
                list_entry_lengths <- lengths(content_list)

                if (length(collapse::funique(list_entry_lengths)) != 1){
                    print_message("ERROR", c("Passed vectors are of unequal lengths. All vectors must have an",
											 "equal number of elements. Evaluation will be aborted."))
                    return(invisible(data_frame))
                }

                if (length(assignments) == 0){
                    print_message("ERROR", c("When using vectors in conditions, there must be a variable assignment.",
											 "Evaluation will be aborted."))
                    return(invisible(data_frame))
                }

                # Do over loop per element
                for (element in seq_len(list_entry_lengths[1])){
                    # Get the respective first elements as symbols in a new list
                    replace_list <- lapply(content_list, function(vector){
                        if (is.numeric(vector) || is.logical(vector)){
                            expression <- vector[element]
                        }
                        else{
                            # Check if vector element is part of the data frame to determine
                            # whether it is a variable or a character expression. Convert
                            # it accordingly.
                            expression <- data_frame |> part_of_df(vector[element], check_only = TRUE)

                            if (is.list(expression)){
                                expression <- as.character(vector[element])
                            }
                            else{
                                expression <- as.name(vector[element])
                            }
                        }

                        # Check if there are colons as placeholders. If this is the
                        # case, convert these expressions to character so that they
                        # can be evaluated correctly later on.
                        number_of_colons <- nchar(gsub("[^:]", "", as.character(expression)))

                        if (number_of_colons > 0){
                            as.character(expression)
                        }
                        else{
                            expression
                        }
                    })

                    # Get the target variable from the assignment list and look whether
                    # it is a vector or not. If it is a vector, extract the current element
                    # and use its name.
                    target_variable <- names(assignments)[1]

                    if (target_variable %in% names(content_list)){
                        target_variable <- content_list[[target_variable]][element]
                    }

                    # Variable has to exist in data frame
                    if (!target_variable %in% names(data_frame)){
                        next
                    }

                    # Check whether there are additional conditions active via do_if
                    # Insert the current variables into the condition
                    current_condition <- do.call(substitute, list(condition, replace_list))
                    current_condition <- translate_condition(current_condition)
                    condition_list[[element]] <- eval(substitute(current_condition), envir = data_frame, enclos = parent_env)

                    # Check whether there are additional conditions active via do_if and
                    # add to call list
                    condition_list[[element]] <- data_frame |> combined_condition(is.na(data_frame[[target_variable]]) & condition_list[[element]])
                    condition_list[[element]][is.na(condition_list[[element]])] <- FALSE
                }
            }
        }

        if (length(condition_list) > 0){
            # Evaluate calculations conditionally. Making use of hidden parameters.
            data_frame <- data_frame |> compute.(..., .if_condition    = condition_list,
                                                      .if_parent_frame = parent_env,
                                                      .if_suppressed   = TRUE)
        }
        # condition_list can be empty in case the variables to assign values to
        # are not already in the data frame.
        else{
            print_message("WARNING", c("No valid target variables found in data frame. For else_if. to work the",
                                       "variables you assign values to must already be present in the data frame.",
                                       "To computation will be executed."))
        }
    }
    else{
        print_message("WARNING", "No assignments found. If you want to filter observations use if.() instead.")
    }

    print_closing()

    data_frame
}


#' @description
#' [else.()] only acts if there already is a variable with the given name. Sets every
#' remaining NA in given variable to the given value.
#'
#' @rdname if_else
#'
#' @export
else. <- function(data_frame, ...){
    # Measure the time
    print_start_message(suppress = TRUE)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Early evaluations
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    parent_env  <- parent.frame()
    assignments <- as.list(substitute(list(...)))[-1]

    # The condition and the variable assignments are torn apart here, so that
    # only the unique variable and vector names are captured as characters.
    used_variables <- unique(c(names(assignments), # Get all variables to which a value should be assigned
                               unlist(lapply(assignments, all.names)))) # Get all assigned variables

    # Get the original contents of vectors if there are any and put them in a list.
    # List names are the symbols names from above and the elements hold the actual contents.
    content_list <- mget(used_variables, envir = parent_env, ifnotfound = list(NULL))

    # Remove invalid empty entries. This is the case if a variable is an original
    # variable name and not a vector of variable names.
    content_list <- Filter(is_valid_vector, content_list)

    # If no vector was passed in the condition or the assignments, then evaluate
    # as normal if statement.
    condition_list <- list()

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Variable assignment
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (length(assignments) > 0){
        for (entry in seq_along(assignments)){
            variable    <- names(assignments)[[entry]]
            calculation <- assignments[[variable]]
            calc_text   <- deparse(calculation)

            # Get used variables especially for this condition in combination mit
            # assigned variables.
            used_variables <- unique(c(variable, # Get all variables to which a value should be assigned
                                       unlist(lapply(calculation, all.names)))) # Get all assigned variables

            #---------------------------------------------------------------------#
            # Normal evaluation
            #---------------------------------------------------------------------#

            # If no vector was passed in the condition or the assignments, then evaluate
            # as normal.
            if (!any(used_variables %in% names(content_list))){
                # This step is important to make this function work in a nested situation.
                # Normally variable would be the name of what was last passed as a parameter.
                # If "if.()" is used nested inside a function this can basically be any placeholder.
                # So here we go up the ladder to get the original name of the variable.
                original_var <- get_origin_symbol(variable)

                # Variable has to exist in data frame
                if (!original_var %in% names(data_frame)){
                    next
                }

                # Check whether there are additional conditions active via do_if
                full_condition <- data_frame |> combined_condition(is.na(data_frame[[original_var]]))

                # Add to call list
                condition_list <- c(condition_list, list(full_condition))
            }
            #---------------------------------------------------------------------#
            # Do over evaluation
            #---------------------------------------------------------------------#
            # If a vector was passed in the condition or the assignments, then evaluate
            # the if statement as a do over loop. Meaning for each vector the same elements
            # are used simultaneously one after another.
            else{
                list_entry_lengths <- lengths(content_list)

                if (length(collapse::funique(list_entry_lengths)) != 1){
                    print_message("ERROR", c("Passed vectors are of unequal lengths. All vectors must have an",
											 "equal number of elements. Evaluation will be aborted."))
                    return(invisible(data_frame))
                }

                if (length(assignments) == 0){
                    print_message("ERROR", c("When using vectors in conditions, there must be a variable assignment.",
											 "Evaluation will be aborted."))
                    return(invisible(data_frame))
                }

                # Do over loop per element
                for (element in seq_len(list_entry_lengths[1])){
                    # Get the respective first elements as symbols in a new list
                    replace_list <- lapply(content_list, function(vector){
                        if (is.numeric(vector) || is.logical(vector)){
                            expression <- vector[element]
                        }
                        else{
                            # Check if vector element is part of the data frame to determine
                            # whether it is a variable or a character expression. Convert
                            # it accordingly.
                            expression <- data_frame |> part_of_df(vector[element], check_only = TRUE)

                            if (is.list(expression)){
                                expression <- as.character(vector[element])
                            }
                            else{
                                expression <- as.name(vector[element])
                            }
                        }

                        # Check if there are colons as placeholders. If this is the
                        # case, convert these expressions to character so that they
                        # can be evaluated correctly later on.
                        number_of_colons <- nchar(gsub("[^:]", "", as.character(expression)))

                        if (number_of_colons > 0){
                            as.character(expression)
                        }
                        else{
                            expression
                        }
                    })

                    # Get the target variable from the assignment list and look whether
                    # it is a vector or not. If it is a vector, extract the current element
                    # and use its name.
                    target_variable <- names(assignments)[1]

                    if (target_variable %in% names(content_list)){
                        target_variable <- content_list[[target_variable]][element]
                    }

                    # Variable has to exist in data frame
                    if (!target_variable %in% names(data_frame)){
                        next
                    }

                    # Check whether there are additional conditions active via do_if and
                    # add to call list
                    condition_list[[element]] <- data_frame |> combined_condition(is.na(data_frame[[target_variable]]))
                    condition_list[[element]][is.na(condition_list[[element]])] <- FALSE
                }
            }
        }
    }

    if (length(condition_list) > 0){
        # Evaluate calculations conditionally. Making use of hidden parameters.
        data_frame <- data_frame |> compute.(..., .if_condition    = condition_list,
                                                  .if_parent_frame = parent_env,
                                                  .if_suppressed   = TRUE)
    }
    # condition_list can be empty in case the variables to assign values to
    # are not already in the data frame.
    else{
        print_message("WARNING", c("No valid target variables found in data frame. For else_if. to work the",
                                   "variables you assign values to must already be present in the data frame.",
                                   "To computation will be executed."))
    }

    print_closing()

    data_frame
}


#' Translate Colon In Condition
#'
#' @description
#' Translates colons used as placeholders in a condition. A colon can stand for
#' an expression starting/ending with or containing a letter combination.
#'
#' @param condition The condition on which the data frame should be subsetted.
#'
#' @return
#' Returns a translated condition.
#'
#' @noRd
translate_condition <- function(condition){
    if (is.call(condition)){
        operator <- condition[[1]]

        # Replace doubled logical operators
        if (identical(operator, as.name("&&"))){
            print_message("WARNING", "Replaced '&&' with '&' for vectorized evaluation.")

            condition[[1]] <- as.name("&")
            operator       <- condition[[1]]
        }

        if (identical(operator, as.name("||"))){
            print_message("WARNING", "Replaced '||' with '|' for vectorized evaluation.")

            condition[[1]] <- as.name("|")
            operator       <- condition[[1]]
        }

        # Only check the expression, if its an equality check
        if (identical(operator, as.name("==")) || identical(operator, as.name("!="))){
            variable   <- condition[[2]]
            expression <- condition[[3]]

            if (is.character(expression) && length(expression) == 1){
                flag_translated <- FALSE
                translated_call <- condition

                # Count the colons, there may not be more than two, otherwise the function
                # should skip the expression.
                colon_count <- nchar(gsub("[^:]", "", expression))

                if (colon_count == 1){
                    # When at the end ("text:"), select all variables which start with the characters
                    # coming before the colon. ":" acts as a placeholder for those who come after.
                    if (endsWith(expression, ":")){
                        search_term <- substring(expression, 1, nchar(expression) - 1)

                        translated_call <- call("startsWith", variable, search_term)

                        flag_translated <- TRUE
                    }
                    # When at the start (":text"), select all variables which end with the characters
                    # following the colon. ":" acts as a placeholder for everything that comes before.
                    else if (startsWith(expression, ":")){
                        search_term <- substring(expression, 2)

                        translated_call <- call("endsWith", variable, search_term)

                        flag_translated <- TRUE
                    }
                }
                # In case there are two colons, one at the start and one at the end (":text:"),
                # look for the text inbetween inside the variable expressions
                else if (colon_count == 2){
                    if (startsWith(expression, ":") && endsWith(expression, ":")){
                        search_term <- gsub(":", "", expression)

                        translated_call <- call("grepl", search_term, variable, fixed = TRUE)

                        flag_translated <- TRUE
                    }
                }

                # Return call as is or as NOT depending on the operator
                if (identical(operator, as.name("!=")) && flag_translated){
                    return(call("!", translated_call))
                }
                else{
                    return(translated_call)
                }
            }
        }

        # Tear up the condition into its single expressions and check each expression
        # for a colon. Then put the single translated expressions back into a fully
        # functional condition.
        as.call(lapply(condition, translate_condition))
    }
    # Return condition as is, if it is a logical instead of a call
    else{
        condition
    }
}


#' Translate Colon In Condition
#'
#' @description
#' Translates colons used as placeholders in a condition. A colon can stand for
#' an expression starting/ending with or containing a letter combination.
#'
#' @param condition The condition on which the data frame should be subsetted.
#'
#' @return
#' Returns a translated condition.
#'
#' @noRd
combined_condition <- function(data_frame, condition){
    # Check whether there are any filter variables
    df_names <- names(data_frame)

    filter_variables <- df_names[startsWith(df_names, ".do_if_select")]

    # If there are no filter variables return a pseudo always TRUE
    if (length(filter_variables) == 0){
        filter_variables <- TRUE
    }
    # Otherwise get the logical filter variables
    else{
        filter_variables <- data_frame |> collapse::fselect(filter_variables)
    }

    # Merge the given condition as well as the active filter variables together
    # into one logical vector
    Reduce(`&`, c(list(condition), filter_variables))
}


###############################################################################
# Quick filtering
###############################################################################
#' Filter Data Frame With Direct View
#'
#' @description
#' Filter observations and variables and directly view the result on screen.
#'
#' @param data_frame A data frame on which to apply filters.
#' @param condition The condition on which to filter observations.
#' @param keep The Variables to keep in the result data frame.
#'
#' @return Returns a filtered data frame.
#'
#' @seealso
#' [ifelse()] with multiple conditions and 'SAS' writing style: [ifelse_multi()]
#'
#' The following functions can make use of the [do_if()] filter variables:
#'
#' Conditions: [if.()], [else_if.()], [else.()]
#'
#' Filter Data Frame: [where.()]
#'
#' Create new Variables: [compute.()]
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Get a quick filtered view
#' my_data |> where.(sex == 1 & age < 25,
#'                   c(sex, age, household_id, education))
#'
#' @export
where. <- function(data_frame,
                   condition = NULL,
                   keep      = NULL){
    # Manage conditional subsetting
    condition <- substitute(condition)
    condition <- translate_condition(condition)
    condition <- eval(condition, envir = data_frame, enclos = parent.frame())

    full_condition <- data_frame |> combined_condition(condition)

    data_frame <- data_frame |> collapse::fsubset(full_condition)

    # Manage keeping variables
    keep <- get_origin_as_char(keep, substitute(keep))

    if (!is.null(keep)){
        data_frame <- data_frame |> keep(keep, order_vars = TRUE)
    }

    # Abort without observations
    if (collapse::fnrow(data_frame) == 0){
        print_message("WARNING", "No observations left in the data frame.")
        return(invisible(data_frame))
    }

    # Abort without variables
    if (collapse::fncol(data_frame) == 0){
        print_message("WARNING", "No variables left in the data frame.")
        return(invisible(data_frame))
    }

    # View data frame in new window
    if (interactive()){
        data_frame |> utils::View()
    }

    invisible(data_frame)
}


###############################################################################
# Do if
###############################################################################
#' Lock In A Condition
#'
#' @name do_if
#'
#' @description
#' Creates a filter variable based on the given condition. This variable can be
#' accessed by [if.()], [else_if.()], [else.()], [where.()] and [compute.()], enabling these
#' functions to work with an overarching condition. This function can also be
#' used to nest multiple overarching conditions.
#'
#' @param data_frame A data frame on which to apply filters.
#' @param condition The condition lock in.
#'
#' @return Returns a data frame with a new filter variable.
#'
#' @seealso
#' The following functions can make use of the [do_if()] filter variables:
#'
#' Conditions: [if.()], [else_if.()], [else.()]
#'
#' Filter Data Frame: [where.()]
#'
#' Create new Variables: [compute.()]
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Create a simple do-if-block
#' do_if_df <- my_data |>
#'     do_if(state < 11) |>
#'           if.(age < 18, new_var = 1) |>
#'         else.(          new_var = 2) |>
#'     else_do() |>
#'           if.(age < 18, new_var = 3) |>
#'         else.(          new_var = 4) |>
#'     end_do()
#'
#' # do_if() can also be nested
#' do_if_df <- my_data |>
#'     do_if(state < 11) |>
#'         do_if(sex == 1) |>
#'               if.(age < 18, new_var = 1) |>
#'             else.(          new_var = 2) |>
#'         else_do() |>
#'               if.(age < 18, new_var = 3) |>
#'             else.(          new_var = 4) |>
#'         end_do() |>
#'     else_do() |>
#'         do_if(sex == 1) |>
#'               if.(age < 18, new_var = 5) |>
#'             else.(          new_var = 6) |>
#'         else_do() |>
#'               if.(age < 18, new_var = 7) |>
#'             else.(          new_var = 8) |>
#'         end_do() |>
#'     end_do()
#'
#' # NOTE: Close the do-if-blocks with end_do() to remove the temporary logical
#' #       filter variables.
#'
#' # Probably a logical filter variable is exactly what you want. In this case
#' # just run do_if() without closing the block.
#' logic_filter_df <- my_data |> do_if(state < 11)
#'
#' @rdname do_if
#'
#' @export
do_if <- function(data_frame, condition){
    parent_env <- parent.frame()

    # Evaluate condition to get a logical vector
    condition <- translate_condition(substitute(condition))
    condition <- eval(condition, envir = data_frame, enclos = parent_env)

    # Check whether there are already filter variables and add a new one. Every
    # filter variable receives a running number.
    filter_variables <- grep("^\\.do_if_select", names(data_frame), value = TRUE)
    new_filter       <- paste0(".do_if_select", length(filter_variables) + 1)

    # Add logical condition vector as new filter variable and return
    data_frame[[new_filter]] <- condition

    data_frame
}


#' @description
#' [else_do()]: Checks for existing filter variables and reverses the condition
#' of the last filter variable.
#'
#' @rdname do_if
#'
#' @export
else_do <- function(data_frame){
    # Get existing filter variables
    filter_variables <- grep("^\\.do_if_select", names(data_frame), value = TRUE)

    if (length(filter_variables) == 0){
        print_message("ERROR", "No active filter variable found. else_do() will be ignored.")
        return(invisible(data_frame))
    }

    # Get last filter variable and get the inverse condition
    last_filter               <- filter_variables[[length(filter_variables)]]
    data_frame[[last_filter]] <- !data_frame[[last_filter]]

    data_frame
}


#' @description
#' [end_do()]: Drops the last filter variable.
#'
#' @rdname do_if
#'
#' @export
end_do <- function(data_frame){
    # Get existing filter variables
    filter_variables <- grep("^\\.do_if_select", names(data_frame), value = TRUE)

    if (length(filter_variables) == 0){
        print_message("ERROR", "No active filter variable found. end_do() will be ignored.")
        return(invisible(data_frame))
    }

    # Get last filter variable and drop it
    last_filter <- filter_variables[[length(filter_variables)]]

    data_frame |> dropp(last_filter)
}


#' @description
#' [end_all_do()]: Drops all filter variables.
#'
#' @rdname do_if
#'
#' @export
end_all_do <- function(data_frame){
    # Get existing filter variables
    filter_variables <- grep("^\\.do_if_select", names(data_frame), value = TRUE)

    if (length(filter_variables) == 0){
        print_message("ERROR", "No active filter variable found. end_all_do() will be ignored.")
        return(invisible(data_frame))
    }

    # Drop all filter variables
    data_frame |> dropp(filter_variables)
}


###############################################################################
# Condition
###############################################################################
#' Do Multiple ifelse At Once
#'
#' @description
#' [ifelse_multi()] handles multiple conditions and value assignments at once by
#' nesting multiple ifelse statements. It is special in the case that it takes the
#' conditions as unevaluated characters to be able to parse them before evaluation.
#' This enables a SAS like writing style of the conditions.
#'
#' @param data_frame A data frame on which to apply multiple ifelse statements.
#' @param ... The conditions as character and the assignments in the form:
#' "condition" = value, "condition" = value, ...
#' @param else. The default value which is applied when no condition results in
#' an assignment.
#' @param do_if Define an overarching condition that will be used on all other conditions.
#' @param na.rm TRUE by default. Sets the default value even though a condition results
#' in NA. If FALSE leaves NA as missing value.
#'
#' @details
#' The function takes in conditions as characters to be able to parse the conditions.
#' The parsing allows to write conditions in a SAS like way. For example the condition
#'
#' "age >= 15 & age < 65" can be written as "15 <= age < 65"
#'
#' Additionally and/or (case insensitive) are recognized keywords which will be translated into &/|.
#'
#' "age >= 15 and age < 65 or sex == 1" becomes "age >= 15 & age < 65 | sex == 1"
#'
#' And [macro()] variables are allowed in any spot, e.g.:
#'
#' YEAR <- 2025\
#' "year == &YEAR" translates to "year == 2025".
#'
#' Single = get translated to ==, the SAS not ^= is translated to !=.
#'
#' In SAS you can write "age in (1 4 7 15 21)" or "age not in (2, 4, 37, 82)" which
#' gets translated into "age %in% (1, 4, 7, 15, 21)" and "!(age %in% (2, 4, 37, 82))".
#'
#' When using == or != operators you can check if character expressions start with
#' (text:), end with (:text) or contain a certain text (:text:). These also get
#' translated.
#'
#' @return Returns a vector with a conditionally computed variable.
#'
#' @seealso
#' Conditions: [if.()], [else_if.()], [else.()]
#'
#' Filter Data Frame: [where.()]
#'
#' Create new Variables: [compute.()]
#'
#' Resolve macro variables: [macro()], [apply_macro()]
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(100)
#'
#' # Simple ifelse statement
#' my_data[["under18"]]    <- my_data |> ifelse_multi(" age < 18 " = 1,       else. = 0)
#' my_data[["middle_age"]] <- my_data |> ifelse_multi(" 15 <= age < 65 " = 1, else. = 0)
#'
#' my_data[["age_gr"]] <- my_data |>
#'     ifelse_multi(" age < 18 "       = "under 18",
#'                  " 18 <= age < 25 " = "18 to under 25",
#'                  " 25 <= age < 50 " = "25 to under 50",
#'                  " 50 <= age < 65 " = "50 to under 65",
#'                  " 65 <= age      " = "65 and more")
#'
#' # With overarching do_if condition
#' my_data[["age_gr_edu"]] <- my_data |>
#'     ifelse_multi(do_if = " education in ('middle' 'high') ",
#'                      " age < 18 "       = "under 18",
#'                      " 18 <= age < 25 " = "18 to under 25",
#'                      " 25 <= age < 50 " = "25 to under 50",
#'                      " 50 <= age < 65 " = "50 to under 65",
#'                      " 65 <= age      " = "65 and more")
#'
#' # And/or translation
#' my_data[["and"]] <- my_data |> ifelse_multi(" age > 65 and sex = 1 " = 1,
#'                                             " age > 65 and sex = 2 " = 2,
#'                                             else. = 0)
#'
#' my_data[["or"]] <- my_data |> ifelse_multi(" age > 65 or sex = 1 " = 1,
#'                                            " age > 65 or sex = 2 " = 2,
#'                                            else. = 0)
#'
#' # "in" translation
#' my_data[["in"]] <- my_data |> ifelse_multi(" age in (1 10 25 65 90) " = 1, else. = 0)
#'
#' # Colon translation: start/ends with and contains
#' my_data[["start"]]    <- my_data |> ifelse_multi(" education == 'lo:' "  = 1, else. = 0)
#' my_data[["end"]]      <- my_data |> ifelse_multi(" education == ':le' "  = 1, else. = 0)
#' my_data[["contains"]] <- my_data |> ifelse_multi(" education == ':ig:' " = 1, else. = 0)
#'
#' # Macro variables can be integrated in any place
#' variable     <- "age"
#' age_to_check <- 18
#' value_to_set <- "under 18"
#'
#' my_data[["macro"]] <- my_data |>
#'     ifelse_multi(" &variable < &age_to_check " = "&value_to_set",
#'                  else.                         = "other")
#'
#' # NA translation
#' my_data[["NA"]]    <- my_data |> ifelse_multi(" age == .       " = 1, else. = 0)
#' my_data[["notNA"]] <- my_data |> ifelse_multi(" education != . " = 1, else. = 0)
#'
#' # Pass in existing variable values
#' my_data[["income_mix"]] <- my_data |>
#'     ifelse_multi(" age < 50 " = income, else. = expenses)
#'
#' @export
ifelse_multi <- function(data_frame,
                         ...,
                         else. = NA,
                         do_if = NULL,
                         na.rm = TRUE){
    # Measure the time
    print_start_message(suppress = TRUE)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Convert conditions
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    conditions <- tryCatch({
        # Force evaluation to see if it exists
        list(...)
    }, error = function(e){
        # Evaluation failed
        NULL
    })

    # In case a variable name was passed as result value, meaning a name without
    # quotation marks, the condition is NULL here and has to be evaluated differently.
    if (is.null(conditions)){
        # Get unevaluated arguments and convert into a list of symbols
        conditions <- substitute(list(...))
        conditions <- sapply(conditions[-1], as.name)
    }

    if (is.null(conditions) || !is_named_list(conditions)){
        print_message("ERROR", c("You have to pass conditions and assignments in the form",
                                 "<condition> = <value>, <condition> = <value>, ...",
                                 "Evaluation will be aborted."))
        return(invisible(data_frame))
    }

    # Parse the condition and resolve and/or, SAS like "15 <= age < 65" and macros
    parsed_conditions <- lapply(names(conditions), parse_conditions, na.rm = na.rm)

    # Apply macro variables to result values
    result_values <- unlist(conditions)

    if (is.character(result_values)){
        result_values <- apply_macro(result_values)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Convert else.
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    else_temp <- tryCatch({
        # Force evaluation to see if it exists
        !is.na(else.)
    }, error = function(e){
        NULL
    })

    # In case a variable name was passed as result value, meaning a name without
    # quotation marks, the parameter is NULL here and has to be evaluated differently.
    if (is.null(else_temp)){
        else. <- as.name(substitute(else.))
    }

    # Look up of which type all the variables are
    variable_names <- names(data_frame)

    types <- vapply(c(result_values, list(else.)), function(element){
        if (!is.name(element) && !is.symbol(element) && is.na(element)){
            return(TRUE)
        }

        # Variable names need to be looked up in the data frame
        if (is.name(element)){
            var_name <- as.character(element)

            if (!var_name %in% variable_names){
                print_message("ERROR", c("Variable [var] is not part of the data frame. Evaluation will be aborted."), var = var_name)
                return(NA)
            }

            # Check if variable is numeric
            is.numeric(data_frame[[var_name]])
        }
        # Check if value is numeric
        else{
            is.numeric(element)
        }

    }, logical(1))

    # Set a flag whether the passed values are of the same type
    is_mixed_type <- any(types) && any(!types)

    # Check if a type conversion is needed
    if (is_mixed_type){
        print_message("NOTE", "Mixed value types found. Result will be converted to character.")

        # Apply macro on character expressions
        if (is.character(else.)){
            else. <- macro(else.)
        }
        # In case of variable names alter the type in the provided data frame
        else if (is.name(else.)){
            value_char               <- as.character(else.)
            data_frame[[value_char]] <- as.character(data_frame[[value_char]])
        }
        # Otherwise convert the value directly
        else{
            else. <- as.character(else.)
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Convert do_if
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # If there was an overarching condition specified, parse and inject it into
    # all the other conditions.
    if(!is.null(do_if)){
        parsed_do_if <- parse_conditions(do_if, na.rm = na.rm)

        parsed_conditions <- lapply(parsed_conditions, function(condition){
                call("&", parsed_do_if, condition)
            })
    }

    if (any(sapply(parsed_conditions, is.null))){
        print_message("ERROR", c("The <do_if> condition couldn't be parsed and is NULL."))
        return(invisible(NA))
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Put together the nested ifelse statement
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # The single conditions and assignments are now nested into multiple ifelse
    # functions.
    result <- else.

    for(i in rev(seq_along(conditions))){
        value <- result_values[[i]]

        # If there are mixed types
        if (is_mixed_type){
            # In case of variable names alter the type in the provided data frame
            if (is.name(value)){
                value_char               <- as.character(value)
                data_frame[[value_char]] <- as.character(data_frame[[value_char]])
            }
            # Otherwise convert the value directly
            else{
                value <- as.character(value)
            }
        }

        # Nest the ifelse statement together
        result <- as.call(list(data.table::fifelse, parsed_conditions[[i]], value, result))
    }

    # Evaluate the nested ifelse functions and return result
    result <- eval(result, envir = data_frame)

    print_closing()

    result
}


#' @description
#' Parses unevaluated character conditions to translate SAS like syntax into
#' R syntax.
#'
#' @param condition A single unevaluated condition as character.
#' @param na.rm Sets the default value even though a condition results in NA.
#' If FALSE leaves NA as missing value.
#'
#' @return Returns a translated condition as language object.
#'
#' @noRd
parse_conditions <- function(condition, na.rm = TRUE){
    # Replace 'and' and 'or' with the actual operators. Also translate the not
    # operator and turn single = into ==.
    condition <- gsub("\\bAND\\b", " & ", condition, ignore.case = TRUE)
    condition <- gsub("\\bOR\\b",  " | ", condition, ignore.case = TRUE)
    condition <- gsub("\\^=", " != ",     condition)
    condition <- gsub("(?<!<|>|=|!)={1}(?!=)", " == ", condition, perl = TRUE)

    # Replace == . and != . with is.na() and !is.na()
    condition <- gsub("([a-zA-Z0-9_]+)[[:space:]]*==[[:space:]]*\\.", "is.na(\\1)",  condition)
    condition <- gsub("([a-zA-Z0-9_]+)[[:space:]]*!=[[:space:]]*\\.", "!is.na(\\1)", condition)

    # Resolve macro variables
    condition <- macro(condition)

    # Parse the in keyword
    condition <- parse_in(condition)
    condition <- parse_colon(condition)

    # Look out for SAS like pattern: 15 <= age < 65, meaning:
    # [number] [operator] [variable_name] [operator] [number]
    # and transform to:
    # [number] [operator] [variable_name] & [variable_name] [operator] [number]
    pattern <- "([0-9.]+)\\s*(<=|<|>=|>)\\s*([a-zA-Z_][a-zA-Z0-9_.]*)\\s*(<=|<|>=|>)\\s*([0-9.]+)"

    while(grepl(pattern, condition)){
        condition <- gsub(pattern, "(\\1 \\2 \\3 & \\3 \\4 \\5)", condition, perl = TRUE)
    }

    if (!has_balanced_quotes(condition)){
        print_message("ERROR", "Condition ' [condition] ' has a missing quotation mark. Condition couldn't be parsed.",
                      condition = condition)
        return(NULL)
    }

    # If na.rm is TRUE, then NA values will be translated into the else. value
    if (na.rm){
        condition <- str2lang(condition)

        bquote(!is.na(.(condition)) & .(condition))
    }
    # Otherwise NA stays NA
    else{
        str2lang(condition)
    }
}


#' @description
#' Parses unevaluated character conditions to translate SAS like syntax into
#' R syntax.
#'
#' @noRd
parse_in <- function(condition){
    # This regex tackles patterns like:
    # [variable] in ([value], [value], ... / [value] [value] ...)
    # [variable] not in ([value], [value], ... / [value] [value] ...)
    in_pattern <- paste0("([a-zA-Z_][a-zA-Z0-9_.]*)\\s+",
                         "(not\\s+)?in\\s*\\(([^)]*)\\)")

    bracket_pattern <- '"[^"]*"|\'[^\']*\'|[^[:space:],]+'

    # Get the positions of matching patterns and the patterns themselves
    machting_positions <- gregexpr(in_pattern, condition, perl = TRUE, ignore.case = TRUE)
    actual_matches     <- regmatches(condition, machting_positions)[[1]]

    if(length(actual_matches) == 0){
        return(condition)
    }

    # Translate into R syntax. Loop through the condition as long as "in" statements
    # are found to be translated.
    while(grepl(in_pattern, condition, perl = TRUE, ignore.case = TRUE)){
        # Extract first "in" statement
        in_match <- regmatches(condition, regexec(in_pattern, condition, perl = TRUE, ignore.case = TRUE))[[1]]

        # Separate the different parts of the statement
        variable <- in_match[2]
        not_part <- in_match[3]
        values   <- trimws(in_match[4])

        # Convert character of values into a vector of values while preserving
        # blanks in character expressions.
        values <- regmatches(values, gregexpr(bracket_pattern, values, perl = TRUE))[[1]]

        # Actual translation into the R %in% statement
        replacement <- sprintf("%s %%in%% c(%s)", variable, paste(values, collapse = ", "))

        # Add negation for "not in"
        if(nzchar(trimws(not_part))){
            replacement <- sprintf("!(%s)", replacement)
        }

        # Replace first occurrence
        condition <- sub(in_pattern, replacement, condition,
                         perl = TRUE, ignore.case = TRUE)
    }

    # Return translated condition
    condition
}


#' @description
#' Translates colons used as placeholders in a condition. A colon can stand for
#' an expression starting/ending with or containing a letter combination.
#'
#' @noRd
parse_colon <- function(condition){
    # Look for variable == "text:" / variable != ":text:" patterns
    pattern <- paste0("([a-zA-Z_][a-zA-Z0-9_.]*)\\s*(==|!=)\\s*",
                      "(['\"])",
                      "([^'\"]*:[^'\"]*)\\3")

    # Replace all colon expressions one after another
    while(grepl(pattern, condition, perl = TRUE)){
        # Extract first colon match
        colon_match <- regmatches(condition, regexec(pattern, condition, perl = TRUE))[[1]]

        # Separate the different parts of the statement
        variable   <- colon_match[2]
        operator   <- colon_match[3]
        expression <- colon_match[5]

        # Count colons inside expression
        colon_count <- lengths(regmatches(expression, gregexpr(":", expression, fixed = TRUE)))
        replacement <- colon_match[1]

        if(colon_count == 1){
            # When at the end ("text:"), select all variables which start with the characters
            # coming before the colon. ":" acts as a placeholder for those who come after.
            if(endsWith(expression, ":")){
                search_term <- substring(expression, 1, nchar(expression) - 1)
                replacement <- sprintf('startsWith(%s, "%s")', variable, search_term)
            }
            # When at the start (":text"), select all variables which end with the characters
            # following the colon. ":" acts as a placeholder for everything that comes before.
            else if(startsWith(expression, ":")){
                search_term <- substring(expression, 2)
                replacement <- sprintf('endsWith(%s, "%s")', variable, search_term)
            }
        }
        # In case there are two colons, one at the start and one at the end (":text:"),
        # look for the text inbetween inside the variable expressions
        else if(colon_count == 2){
            if (startsWith(expression, ":") && endsWith(expression, ":")){
                search_term <- gsub(":", "", expression, fixed = TRUE)
                replacement <- sprintf('grepl("%s", %s, fixed = TRUE)', search_term, variable)
            }
        }

        # Negate translated expression if original operator was !=
        if(operator == "!=" && replacement != colon_match[1]){
            replacement <- sprintf("!(%s)", replacement)
        }

        # Replace first occurrence
        condition <- sub(pattern, replacement, condition, perl = TRUE)
    }

    # Return translated condition
    condition
}


#' @description
#' Checks for balanced quotation marks inside the condition string.
#'
#' @noRd
has_balanced_quotes <- function(condition){
    number_of_double <- lengths(regmatches(condition, gregexpr('"', condition, fixed = TRUE)))
    number_of_single <- lengths(regmatches(condition, gregexpr("'", condition, fixed = TRUE)))

    number_of_double %% 2 == 0 && number_of_single %% 2 == 0
}

set_no_print(TRUE)

###############################################################################
# Suppressing some functions messages because they only output the information
# on how much time they took.
###############################################################################

dummy_df <- dummy_data(1000)

ifelse_df <- data.table::data.table(age  = c(10, 20, 30, 70, NA),
                                    sex  = c(1, 2, 1, 2, NA),
                                    var1 = c(10, 20, 10, 20, 10),
                                    name = c("Hello World", "Hello Again", "Hello", "World", NA))


# if. can convert values conditionally
test_df <- dummy_df |>
         if.(age < 18,             age_group = "under 18") |>
    else_if.(age >= 18 & age < 65, age_group = "18 to under 65") |>
    else.   (                      age_group = "65 and older")

expect_true("age_group" %in% names(test_df), info = "if. can convert values conditionally")
expect_true(all(c("under 18", "18 to under 65", "65 and older") %in% test_df[["age_group"]]), info = "if. can convert values conditionally")


# if. can handle vectors
test_df <- dummy_df |>
         if.(age < 18,             income_new = income) |>
    else_if.(age >= 18 & age < 65, income_new = income) |>
    else.   (                      income_new = income)

expect_true("income_new" %in% names(test_df), info = "if. can handle vectors")
expect_equal(test_df[["income_new"]], test_df[["income"]], info = "if. can handle vectors")


# if. doesn't overwrite existing values with NA
test_df           <- dummy_df
test_df[["var1"]] <- ifelse(test_df[["first_person"]] == 1, NA, 1)
test_df[["var2"]] <- 1
test_df           <- test_df |> if.(var1 == 1, var2 = 2)

expect_true(all(collapse::funique(test_df[["var2"]]) %in% c(1, 2)), info = "if. doesn't overwrite existing values with NA")
expect_true(!anyNA(test_df[["var2"]]), info = "if. doesn't overwrite existing values with NA")


# else_if. doesn't work without if.
test_df <- dummy_df |>
    else_if.(age >= 18 & age < 65, age_group = "18 to under 65")

expect_true(!"age_group" %in% names(test_df), info = "else_if. doesn't work without if.")
expect_warning(print_stack_as_messages("WARNING"), "No valid target variables found in data frame. For else_if. to work the",
               info = "else_if. doesn't work without if.")


# else. doesn't work without if.
test_df <- dummy_df |>
    else.(age_group = "65 and older")

expect_true(!"age_group" %in% names(test_df), info = "else. doesn't work without if.")
expect_warning(print_stack_as_messages("WARNING"), "No valid target variables found in data frame. For else_if. to work the",
               info = "else. doesn't work without if.")


# else_if. only alters NA values conditionally
test_df1 <- dummy_df |>
         if.(age >= 18 & age < 65, age_group = "under 18") |>
    else_if.(age >= 18 & age < 65, age_group = "18 to under 65")

test_df2 <- dummy_df |>
         if.(age < 0,              age_group = "under 18") |>
    else_if.(age >= 18 & age < 65, age_group = "18 to under 65")

expect_true("age_group" %in% names(test_df1), info = "else_if. only alters NA values conditionally")
expect_true("age_group" %in% names(test_df2), info = "else_if. only alters NA values conditionally")
expect_true(!"18 to under 65" %in% test_df1[["age_group"]], info = "else_if. only alters NA values conditionally")
expect_true(!"under 18" %in% test_df2[["age_group"]], info = "else_if. only alters NA values conditionally")
expect_true(NA %in% test_df2[["age_group"]], info = "else_if. only alters NA values conditionally")


# else. only alters every other NA value
test_df1 <- dummy_df |>
    collapse::fsubset(!is.na(age)) |>
      if.(age < 101, age_group = "under 100") |>
    else.(           age_group = "100 and older")

test_df2 <- dummy_df |>
    collapse::fsubset(!is.na(age)) |>
      if.(age < 0, age_group = "under 18") |>
    else.(         age_group = "18 and older")

expect_true("age_group" %in% names(test_df1), info = "else. only alters every other NA value")
expect_true("age_group" %in% names(test_df2), info = "else. only alters every other NA value")
expect_true(!"100 and older" %in% test_df1[["age_group"]], info = "else. only alters every other NA value")
expect_true(!NA %in% test_df2[["age_group"]], info = "else. only alters every other NA value")


# if. and else_if. are not the same
state_df_a <- dummy_df |>
         if.(state <= 3, state_a = "State 1-3") |>
    else_if.(state < 11, state_a = "West") |>
    else.   (            state_a = "East")

state_df_b <- dummy_df |>
      if.(state <= 3, state_b = "State 1-3") |>
      if.(state < 11, state_b = "West") |>
    else.(            state_b = "East")

expect_false(identical(state_df_a[["state_a"]], state_df_b[["state_b"]]),
             info = "if. and else_if. are not the same")


# if. converts NA values into format of follow up value
result_df <- dummy_df |>
         if.(state <= 3, var1 = NA, var2 = NA) |>
    else_if.(state >  3, var1 = 1,  var2 = "Hello")

expect_equal(collapse::funique(result_df[["var1"]])[-1], 1,       info = "if. converts NA values into format of follow up value")
expect_equal(collapse::funique(result_df[["var2"]])[-1], "Hello", info = "if. converts NA values into format of follow up value")


# if. can assign a single NA value
test_df           <- dummy_df
test_df[["var1"]] <- 1
test_df           <- test_df |> if.(state <= 3, var1 = NA)

expect_true(all(is.na(test_df[["var1"]][test_df[["state"]] <= 3])), info = "if. can assign a single NA value")


# if. can check for variable expressions starting with letter
letter_df <- dummy_df |> if.(education == "m:", edu = 1)
test_df   <- letter_df |> if.(edu)

expect_true(all(collapse::funique(letter_df[["edu"]]) %in% c(1, NA)), info = "if. can check for variable expressions starting with letter")
expect_true(collapse::funique(test_df[["education"]]) == "middle", info = "if. can check for variable expressions starting with letter")


# if. can check for variable expressions ending with letter
letter_df <- dummy_df |> if.(education == ":w", edu = 1)
test_df   <- letter_df |> if.(edu)

expect_true(all(collapse::funique(letter_df[["edu"]]) %in% c(1, NA)), info = "if. can check for variable expressions ending with letter")
expect_true(collapse::funique(test_df[["education"]]) == "low", info = "if. can check for variable expressions ending with letter")


# if. can check for variable expressions containing a letter
letter_df <- dummy_df |> if.(education == ":g:", edu = 1)
test_df   <- letter_df |> if.(edu)

expect_true(all(collapse::funique(letter_df[["edu"]]) %in% c(1, NA)), info = "if. can check for variable expressions containing a letter")
expect_true(collapse::funique(test_df[["education"]]) == "high", info = "if. can check for variable expressions containing a letter")


# if. can check for variable expressions starting with letter (unequal)
letter_df <- dummy_df |> if.(education != "m:", edu = 1)
test_df   <- letter_df |> if.(edu)

expect_true(all(collapse::funique(letter_df[["edu"]]) %in% c(1, NA)), info = "if. can check for variable expressions starting with letter (unequal)")
expect_true(!"middle" %in% collapse::funique(test_df[["education"]]), info = "if. can check for variable expressions starting with letter (unequal)")


# if. can check for variable expressions ending with letter (unequal)
letter_df <- dummy_df |> if.(education != ":w", edu = 1)
test_df   <- letter_df |> if.(edu)

expect_true(all(collapse::funique(letter_df[["edu"]]) %in% c(1, NA)), info = "if. can check for variable expressions ending with letter (unequal)")
expect_true(!"low" %in% collapse::funique(test_df[["education"]]), info = "if. can check for variable expressions ending with letter (unequal)")


# if. can check for variable expressions containing a letter (unequal)
letter_df <- dummy_df |> if.(education != ":g:", edu = 1)
test_df   <- letter_df |> if.(edu)

expect_true(all(collapse::funique(letter_df[["edu"]]) %in% c(1, NA)), info = "if. can check for variable expressions containing a letter (unequal)")
expect_true(!"high" %in% collapse::funique(test_df[["education"]]), info = "if. can check for variable expressions containing a letter (unequal)")


# if. as do over loop
vars1  <- c("income", "balance")
vars2  <- c("VAR1", "VAR2")
values <- c(1, 2)

do_over_df <- dummy_df |> if.(vars1 > 0, vars2 = values)

expect_true(all(c("VAR1", "VAR2") %in% names(do_over_df)), info = "if. as do over loop")
expect_true(all(collapse::funique(do_over_df[["VAR1"]]) %in% c(1, NA)), info = "if. as do over loop")
expect_true(all(collapse::funique(do_over_df[["VAR2"]]) %in% c(2, NA)), info = "if. as do over loop")

do_over_df <- do_over_df |> else_if.(vars1 > 500, vars2 = 3)
expect_true(all(collapse::funique(do_over_df[["VAR1"]]) %in% c(1, 3, NA)), info = "if. as do over loop")
expect_true(all(collapse::funique(do_over_df[["VAR2"]]) %in% c(2, 3, NA)), info = "if. as do over loop")

do_over_df <- do_over_df |> else.(vars2 = 4)
expect_true(all(collapse::funique(do_over_df[["VAR1"]]) %in% c(1, 3, 4)), info = "if. as do over loop")
expect_true(all(collapse::funique(do_over_df[["VAR2"]]) %in% c(2, 3, 4)), info = "if. as do over loop")


# if. as do over loop doesn't overwrite existing values with NA
vars   <- c("var2", "var2")
values <- c(2, 3)

test_df           <- dummy_df
test_df[["var1"]] <- ifelse(test_df[["first_person"]] == 1, NA, 1)
test_df[["var2"]] <- 1
test_df           <- test_df |> if.(var1 == 1, vars = values)

expect_true(all(collapse::funique(test_df[["var2"]]) %in% c(1, 3)), info = "if. as do over loop doesn't overwrite existing values with NA")
expect_true(!anyNA(test_df[["var2"]]), info = "if. as do over loop doesn't overwrite existing values with NA")


# if. as do over loop can handle different values in the same variable
dummy_df <- dummy_data(1000)
class_df <- dummy_df |>
         if.(income <     1,                 income1 = 0) |>
    else_if.(income >=    1 & income < 4000, income1 = 1) |>
    else_if.(income >= 4000 & income < 5000, income1 = 2) |>
    else_if.(income >= 5000,                 income1 = 3)

lower_bound <- c(   1, 4000)
upper_bound <- c(4000, 5000)
values      <- c(   1,    2)

do_over_df <- class_df |>
         if.(income <  1,                                  income2 = 0) |>
    else_if.(income >= lower_bound & income < upper_bound, income2 = values) |>
    else_if.(income >= 5000,                               income2 = 3)

expect_equal(do_over_df[["income1"]], do_over_df[["income2"]],
             info = "if. as do over loop can handle different values in the same variable")


# do_if blocks
test_df <- dummy_df |>
    do_if(sex == 1) |>
             if.(age < 18,             age_group = 1) |>
        else_if.(age >= 18 & age < 65, age_group = 2) |>
        else.   (                      age_group = 3) |>
    else_do() |>
             if.(age < 18,             age_group = 4) |>
        else_if.(age >= 18 & age < 65, age_group = 5) |>
        else.   (                      age_group = 6) |>
    end_do()

  male_df <- test_df |> collapse::fsubset(sex == 1)
female_df <- test_df |> collapse::fsubset(sex == 2)

expect_true( all(collapse::funique(male_df[["age_group"]]) %in% c(1, 2, 3)), info = "do_if blocks")
expect_true(!all(collapse::funique(male_df[["age_group"]]) %in% c(4, 5, 6)), info = "do_if blocks")
expect_true( all(collapse::funique(female_df[["age_group"]]) %in% c(4, 5, 6)), info = "do_if blocks")
expect_true(!all(collapse::funique(female_df[["age_group"]]) %in% c(1, 2, 3)), info = "do_if blocks")


# do_if blocks with end_all_do
test_df <- dummy_df |>
    do_if(sex == 1) |>
    end_all_do()

expect_equal(test_df, dummy_df, info = "do_if blocks with end_all_do")


# if. can do all kinds of calculations
variables  <- c("NEW_VAR1", "NEW_VAR2")
values     <- c(1, 2)

result_df <- dummy_df |> if.(state < 11, sum       = state + age,
                                         col_sum   = collapse::fsum(age),
                                         row_sum   = row_calculation("sum", state, age),
                                         var1      = 1,
                                         var2      = "Hello",
                                         variables = values) |>
                   else_if.(state == 11, sum       = state + age,
                                         col_sum   = collapse::fsum(age),
                                         row_sum   = row_calculation("sum", state, age),
                                         var1      = 1,
                                         var2      = "Hello",
                                         variables = values) |>
                                   else.(sum       = state + age,
                                         col_sum   = collapse::fsum(age),
                                         row_sum   = row_calculation("sum", state, age),
                                         var1      = 1,
                                         var2      = "Hello",
                                         variables = values)

expect_true(all(c("sum", "col_sum", "row_sum", "var1", "var2", "NEW_VAR1", "NEW_VAR2") %in% names(result_df)),
            info = "if. can do all kinds of calculations")


# if. and else_if. work with parsed character conditions
vars1  <- c("income", "balance")
vars2  <- c("VAR1", "VAR2")
values <- c(1, 2)

test_df <- dummy_df |>
         if.("15 <= age < 65",        age_group = "15 to under 65") |>
    else_if.("age < 15 or age >= 65", age_group = "other") |>
         if.("vars1 > 0",             vars2     = values)

expect_true("age_group" %in% names(test_df), info = "if. and else_if. work with parsed character conditions")
expect_true(all(c("15 to under 65", "other", NA) %in% test_df[["age_group"]]), info = "if. and else_if. work with parsed character conditions")

expect_true(all(c("VAR1", "VAR2") %in% names(test_df)), info = "if. and else_if. work with parsed character conditions")
expect_true(all(collapse::funique(test_df[["VAR1"]]) %in% c(1, NA)), info = "if. and else_if. work with parsed character conditions")
expect_true(all(collapse::funique(test_df[["VAR2"]]) %in% c(2, NA)), info = "if. and else_if. work with parsed character conditions")


# else_if. throws a warning in do over loop if variables are missing in data frame
vars1  <- c("income", "balance")
vars2  <- c("VAR1", "VAR2")
values <- c(1, 2)

test_df <- dummy_df |>
    if.("15 <= age < 65", age_group = "15 to under 65") |>
    else_if.("vars1 > 0", vars2     = values)

expect_warning(print_stack_as_messages("WARNING"), "No valid target variables found in data frame. For else_if. to work the",
               info = "else_if. throws a warning in do over loop if variables are missing in data frame")

###############################################################################
# if. for subsetting
###############################################################################

# Subset data frame with if.
test_df <- dummy_df |> if.(sex == 1)

expect_true(collapse::fnrow(test_df) < collapse::fnrow(dummy_df), info = "Subset data frame with if.")
expect_true(!2 %in% test_df[["sex"]], info = "Subset data frame with if.")


# Subset data frame with if. (unequal)
test_df <- dummy_df |> if.(sex != 1)

expect_true(collapse::fnrow(test_df) < collapse::fnrow(dummy_df), info = "Subset data frame with if. (unequal)")
expect_true(!1 %in% test_df[["sex"]], info = "Subset data frame with if. (unequal)")


# Subset data frame with if., when only providing single variable
test_df <- dummy_df |> if.(sex)

expect_true(collapse::fnrow(test_df) < collapse::fnrow(dummy_df), info = "Subset data frame with if., when only providing single variable")
expect_true(!NA %in% test_df[["sex"]], info = "Subset data frame with if., when only providing single variable")


# Subset data frame with if., when only providing single variable as character
test_df <- dummy_df |> if.("sex")

expect_true(collapse::fnrow(test_df) < collapse::fnrow(dummy_df), info = "Subset data frame with if., when only providing single variable as character")
expect_true(!NA %in% test_df[["sex"]], info = "Subset data frame with if., when only providing single variable as character")


# Delete observations with 'delete' keyword
test_df <- dummy_df |> if.(sex == 1, delete)

expect_true(collapse::fnrow(test_df) < collapse::fnrow(dummy_df), info = "Delete observations with 'delete' keyword")
expect_true(!1 %in% test_df[["sex"]], info = "Delete observations with 'delete' keyword")


# if. as do over loop for subsetting
vars    <- c("income", "balance")
values1 <- c(1000, 0)
values2 <- c("low", "high")
values3 <- FALSE

test_df <- dummy_df
test_df[["logical"]] <- TRUE

do_over_df1 <- dummy_df |> if.(vars > values1)
do_over_df2 <- dummy_df |> if.(education != values2)
do_over_df3 <- test_df  |> if.(logical == values3)

expect_true(collapse::fmin(do_over_df1[["income"]])  > 1000, info = "if. as do over loop for subsetting")
expect_true(collapse::fmin(do_over_df1[["balance"]]) > 0,    info = "if. as do over loop for subsetting")
expect_true(collapse::funique(do_over_df2[["education"]]) == "middle", info = "if. as do over loop for subsetting")
expect_true(collapse::fnrow(do_over_df3) == 0, info = "if. as do over loop for subsetting")

# Warning on doubled logical operators
test_df <- dummy_df |> if.((sex == 1 && education == "high") || sex == 2)

expect_warning(print_stack_as_messages("WARNING"), "Replaced",  info = "Warning on doubled logical operators")

unique_male   <- test_df |> if.(sex == 1)
unique_male   <- collapse::funique(unique_male[["education"]])
unique_female <- test_df |> if.(sex == 2)
unique_female <- collapse::funique(unique_female[["education"]])

expect_true("high" %in% unique_male, info = "Warning on doubled logical operators")
expect_equal(length(unique_male), 1, info = "Warning on doubled logical operators")
expect_true(all(c("low", "middle", "high") %in% unique_female), info = "Warning on doubled logical operators")


# Abort subset with if., if variable is not part of the data frame
test_df <- dummy_df |> if.("test")

expect_error(print_stack_as_messages("ERROR"), "No variable for subsetting provided. Data frame remains as is.",
               info = "Abort subset with if., if variable is not part of the data frame")

expect_equal(test_df, dummy_df, info = "Abort subset with if., if variable is not part of the data frame")


# Abort subset with if., if multiple variables are provided
test_df <- dummy_df |> if.(c("age", "sex"))

expect_error(print_stack_as_messages("ERROR"), "Only single variables and conditions allowed. Data frame remains as is.",
               info = "Abort subset with if., if multiple variables are provided")

expect_equal(test_df, dummy_df, info = "Abort subset with if., if multiple variables are provided")

###############################################################################
# Where.
###############################################################################

# Filter data frame with where.
test_df <- dummy_df |> where.(sex == 1, c("sex", "age"))

expect_equal(names(test_df), c("sex", "age"), info = "Filter data frame with where.")
expect_true(!any(c(2, NA) %in% test_df[["sex"]]), info = "Filter data frame with where.")


# where. aborts with a warning, if no observations left
test_df <- dummy_df |> where.(sex == 0)

expect_warning(print_stack_as_messages("WARNING"), "No observations left in the data frame.",
               info = "where. aborts with a warning, if no observations left")


# where. works with parsed character conditions
test_df <- dummy_df |> where.(" 15 <= age < 65 and sex == 2 ")

expect_true(!any(c(1, NA) %in% test_df[["sex"]]), info = "where. works with parsed character conditions")
expect_true(!any(c(0:14, 65:100) %in% test_df[["age"]]), info = "where. works with parsed character conditions")

###############################################################################
# ifelse_multi
###############################################################################

# ifelse_multi basic recoding works
result <- ifelse_df |> ifelse_multi("age < 18" = 1,
                                    "age < 65" = 2,
                                    else.      = 3)

expect_equal(result, c(1, 2, 2, 3, 3), info = "ifelse_multi basic recoding works")


# ifelse_multi SAS intervals are translated
result <- ifelse_df |> ifelse_multi(" 0 <= age < 30" = 1,
                                    "30 <= age < 65" = 2,
                                    else.            = 3)

expect_equal(result, c(1, 1, 2, 3, 3), info = "ifelse_multi SAS intervals are translated")


# ifelse_multi IN works with commas
result <- ifelse_df |> ifelse_multi("age in (10, 20)" = 1, else. = 0)

expect_equal(result, c(1, 1, 0, 0, 0), info = "ifelse_multi IN works with commas")


# ifelse_multi IN works with spaces
result <- ifelse_df |> ifelse_multi("age in (10 20)" = 1, else. = 0)

expect_equal(result, c(1, 1, 0, 0, 0), info = "ifelse_multi IN works with spaces")


# ifelse_multi NOT IN works
result <- ifelse_df |> ifelse_multi("age not in (10 20)" = 1, else. = 0)

expect_equal(result, c(0, 0, 1, 1, 1), info = "ifelse_multi NOT IN works")


# ifelse_multi character with spaces works with IN
result <- ifelse_df |> ifelse_multi("name in ('Hello World' 'Hello Again')" = 1, else. = 0)

expect_equal(result, c(1, 1, 0, 0, 0), info = "ifelse_multi character with spaces works with IN")


# ifelse_multi character expressions starting with letter
result <- ifelse_df |> ifelse_multi("name == 'Hell:'" = 1, else. = 0)

expect_equal(result, c(1, 1, 1, 0, 0), info = "ifelse_multi character expressions starting with letter")


# ifelse_multi character expressions ending with letter
result <- ifelse_df |> ifelse_multi("name == ':orld'" = 1, else. = 0)

expect_equal(result, c(1, 0, 0, 1, 0), info = "ifelse_multi character expressions ending with letter")


# ifelse_multi character expressions contain letter
result <- ifelse_df |> ifelse_multi("name == ':ga:'" = 1, else. = 0)

expect_equal(result, c(0, 1, 0, 0, 0), info = "ifelse_multi character expressions contain letter")


# ifelse_multi negated character expressions contain letter
result <- ifelse_df |> ifelse_multi("name != ':ga:'" = 1, else. = 0)

expect_equal(result, c(1, 0, 1, 1, 1), info = "ifelse_multi negated character expressions contain letter")


# ifelse_multi AND is translated
result <- ifelse_df |> ifelse_multi("age < 18 and sex == 1" = 1, else. = 0)

expect_equal(result, c(1, 0, 0, 0, 0), info = "ifelse_multi AND is translated")


# ifelse_multi OR is translated
result <- ifelse_df |> ifelse_multi("age < 18 or sex == 2" = 1, else. = 0)

expect_equal(result, c(1, 1, 0, 1, 0), info = "ifelse_multi OR is translated")


# ifelse_multi na.rm = FALSE preserves NA
result <- ifelse_df |> ifelse_multi("age < 18" = 1, else. = 0, na.rm = FALSE)

expect_equal(result, c(1, 0, 0, 0, NA), info = "ifelse_multi na.rm = FALSE preserves NA")


# ifelse_multi can handle macro variables
variable <- "sex"
gender   <- 1
value    <- "male"

result <- ifelse_df |> ifelse_multi("&variable == &gender" = "&value")

expect_equal(result, c("male", NA, "male", NA, NA), info = "ifelse_multi can handle macro variables")


# ifelse_multi keeps the variable values when assigning NA without an else.
result <- ifelse_df |> ifelse_multi("var1 == 10" = NA)

expect_equal(result, c(NA, 20, NA, 20, NA), info = "ifelse_multi NA is translated")

result <- ifelse_df |> ifelse_multi("var1 == 10" = NA, else. = 30)

expect_equal(result, c(NA, 30, NA, 30, NA), info = "ifelse_multi NA is translated")


# ifelse_multi NA is translated
result1 <- ifelse_df |> ifelse_multi("age == ." = 1, else. = 0)
result2 <- ifelse_df |> ifelse_multi("age != ." = 1, else. = 0)

expect_equal(result1, c(0, 0, 0, 0, 1), info = "ifelse_multi NA is translated")
expect_equal(result2, c(1, 1, 1, 1, 0), info = "ifelse_multi NA is translated")


# ifelse_multi can assign the values of variables
result <- ifelse_df |> ifelse_multi("age < 30" = sex, else. = var1)

expect_equal(result, c(1, 2, 10, 20, 10), info = "ifelse_multi can assign the values of variables")


# ifelse_multi do_if condition is used on all other conditions
result <- ifelse_df |> ifelse_multi(do_if = "sex == 1",
                                    "age < 18" = 1,
                                    "age < 65" = 2,
                                    else.      = 3)

expect_equal(result, c(1, 3, 2, 3, 3), info = "ifelse_multi do_if condition is used on all other conditions")


# ifelse_multi throws a note on value conversion
result <- ifelse_df |> ifelse_multi("age < 18" = 1,
                                    else.      = "2")

expect_equal(result, c("1", "2", "2", "2", "2"), info = "ifelse_multi throws a note on value conversion")
expect_message(print_stack_as_messages("NOTE"), "Mixed value types found. Result will be converted to character.",
             info = "ifelse_multi throws a note on value conversion")

result <- ifelse_df |> ifelse_multi("age < 18" = "1",
                                    else.      = 2)

expect_equal(result, c("1", "2", "2", "2", "2"), info = "ifelse_multi throws a note on value conversion")
expect_message(print_stack_as_messages("NOTE"), "Mixed value types found. Result will be converted to character.",
               info = "ifelse_multi throws a note on value conversion")


# ifelse_multi aborts on missing quotation mark
result <- ifelse_df |> ifelse_multi("name in ('Hello World 'Hello Again')" = 1, else. = 0)

expect_error(print_stack_as_messages("ERROR"), "Condition couldn't be parsed.",
             info = "ifelse_multi aborts on missing quotation mark")


# ifelse_multi aborts on non named list condition
result <- ifelse_df |> ifelse_multi("name in ('Hello World 'Hello Again')")

expect_error(print_stack_as_messages("ERROR"), "You have to pass conditions and assignments in the form",
             info = "ifelse_multi aborts on non named list condition")

###############################################################################
# Abort checks
###############################################################################

# if. as do over loop aborts on vectors of unequal lengths
vars1  <- c("income", "balance")
vars2  <- c("VAR1", "VAR2", "VAR3")
values <- c(1, 2)

do_over_df <- dummy_df |> if.(vars1 > 0, vars2 = values)

expect_error(print_stack_as_messages("ERROR"), "Passed vectors are of unequal lengths.",
               info = "if. as do over loop aborts on vectors of unequal lengths")

do_over_df <- dummy_df |> else_if.(vars1 > 0, vars2 = values)

expect_error(print_stack_as_messages("ERROR"), "Passed vectors are of unequal lengths.",
               info = "if. as do over loop aborts on vectors of unequal lengths")

do_over_df <- dummy_df |> else.(vars2 = values, vars2 = values)

expect_error(print_stack_as_messages("ERROR"), "Passed vectors are of unequal lengths.",
               info = "if. as do over loop aborts on vectors of unequal lengths")


# if. as do over loop aborts without variable assignment
vars <- c("income", "balance")

do_over_df <- dummy_df |> else_if.(vars > 0)

expect_warning(print_stack_as_messages("WARNING"), "No assignments found. If you want to filter observations",
               info = "if. as do over loop aborts without variable assignment")


# else_do aborts when there is no active filter
test_df <- dummy_df |> else_do()

expect_error(print_stack_as_messages("ERROR"), "No active filter variable found.",
               info = "else_do aborts when there is no active filter")


# end_do aborts when there is no active filter
test_df <- dummy_df |> end_do()

expect_error(print_stack_as_messages("ERROR"), "No active filter variable found.",
               info = "end_do aborts when there is no active filter")


# end_all_do aborts when there is no active filter
test_df <- dummy_df |> end_all_do()

expect_error(print_stack_as_messages("ERROR"), "No active filter variable found.",
               info = "end_all_do aborts when there is no active filter")


set_no_print()

#' Keep and Drop Variables Inside a Data Frame
#'
#' @name keep_dropp
#'
#' @param data_frame A data frame which should be reduced to (keep) or by (drop) the specified variables.
#' @param ... The variable names to keep/drop.
#' @param order_vars [keep()]: At the end variables are ordered as specified in the command.
#'
#' @return Returns a reduced data table.
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(1000)
#'
#' # Call function
#' new_dt1 <- my_data |> keep(year, age, sex)
#' new_dt2 <- my_data |> keep(weight, income, education, sex, order_vars = TRUE)
#' new_dt3 <- my_data |> dropp(year, age, sex)
#'
#' # Also works with characters
#' new_dt4 <- my_data |> keep("year", "age", "sex")
#' new_dt5 <- my_data |> dropp("year", "age", "sex")
#'
#' # Or variable names stored as a character vector
#' var_names <- c("age", "sex", "income", "weight")
#'
#' new_dt6 <- my_data |> keep(var_names)
#' new_dt7 <- my_data |> dropp(var_names)
#'
#' # You can also keep or drop a range of variables
#' new_dt8 <- my_data |> keep(year, state:income)
#' new_dt9 <- my_data |> dropp(year, state:income)
#'
#' # You can also use the colon as a placeholder for any text
#' start1 <- my_data |> keep("s:") # Variable names start with "s"
#' start2 <- my_data |> dropp("s:")
#'
#' end1 <- my_data |> keep(":id") # Variable names end with "id"
#' end2 <- my_data |> dropp(":id")
#'
#' contain1 <- my_data |> keep(":on:") # Variable names which contain "on"
#' contain2 <- my_data |> dropp(":on:")
#'
#' @rdname keep_dropp
#'
#' @keywords internal
NULL


#' @description
#' [keep()] enables you to put in a vector of variable names which then are kept
#' inside the given data frame. All other variables are dropped.
#'
#' @rdname keep_dropp
#'
#' @export
keep <- function(data_frame, ..., order_vars = FALSE){
    # Convert to character vectors
    variables_temp <- dots_to_char(...)

    if (length(variables_temp) == 0){
        return(invisible(data_frame))
    }

    original_order <- names(data_frame)

    # Check if there are any colons in the selection and deparse variables accordingly
    variables <- data_frame |> deparse_colon(variables_temp)

    if (length(variables) == 0){
        print_message("WARNING", "No variables found with the given pattern. Data frame remains untouched.")
        return(invisible(data_frame))
    }

    # Check if all variables are part of the data frame
    provided_vars <- variables
    invalid_vars  <- variables[!variables %in% names(data_frame)]
    variables     <- variables[variables %in% names(data_frame)]

    if (length(invalid_vars) > 0){
        print_message("WARNING", c("The provided variable to keep '[invalid]' is not part of",
								   "the data frame. This variable will be omitted."), invalid = invalid_vars)
        rm(provided_vars, invalid_vars)
    }

    # Convert to data.table if necessary
    if (!data.table::is.data.table(data_frame)){
        data_frame <- data.table::as.data.table(data_frame)
    }

    # Get the negative vector of variables to keep and every variable name inside
    # given data frame to produce a vector with the variables to drop
    name_vector <- names(data_frame)
    drop_variables <- name_vector[!name_vector %in% variables]

    # Use data.table function set to drop the desired variables
    data_frame <- data_frame |> collapse::fselect(variables)

    # Order variables
    if (!order_vars){
        data.table::setcolorder(data_frame, original_order, skip_absent = TRUE)
    }

    data_frame
}


#' @description
#' [dropp()] enables you to put in a vector of variable names which then are dropped
#' from the given data frame. All other variables are kept.
#'
#' @rdname keep_dropp
#'
#' @export
dropp <- function(data_frame, ...){
    # Convert to character vectors
    variables_temp <- dots_to_char(...)

    if (length(variables_temp) == 0){
        return(invisible(data_frame))
    }

    # Check if there are any colons in the selection and deparse variables accordingly
    variables <- data_frame |> deparse_colon(variables_temp)

    if (length(variables) == 0){
        print_message("WARNING", "No variables found with the given pattern. Data frame remains untouched.")
        return(invisible(data_frame))
    }

    # Check if all variables are part of the data frame
    provided_vars <- variables
    invalid_vars  <- variables[!variables %in% names(data_frame)]
    variables     <- variables[variables %in% names(data_frame)]

    if (length(invalid_vars) > 0){
        print_message("WARNING", c("The provided variable to drop '[invalid]' is not part of",
								   "the data frame. This variable will be omitted."),
								   invalid = invalid_vars)

        if (length(variables) == 0){
            return(invisible(data_frame))
        }
    }

    # Convert to data.table if necessary
    if (!data.table::is.data.table(data_frame)){
        data_frame <- data.table::as.data.table(data_frame)
    }

    # Use data.table function set to drop the desired variables
    collapse::fselect(data_frame, variables) <- NULL

    data_frame
}

#' Helper For Selecting Variable Ranges
#'
#' @description
#' When using a colon in certain places, variable ranges should be selected, depending
#' on where the colon is placed.
#'
#' @param data_frame The data frame which contains the variable names to be selected.
#' @param variable_vector A vector containing the variable names to be checked.
#'
#' @return
#' A deparsed vector of variable names.
#'
#' @noRd
deparse_colon <- function(data_frame, variable_vector){
    variables <- character(0)

    # Loop through variable vector to decide whether it is a single variable or
    # a variable range with colon should be selected.
    for (variable in variable_vector){
        only_colons <- gsub(":", "", variable)

        # If it's just colons with no text, skip "variable" and put out a warning
        if (nchar(only_colons) == 0){
            print_message("WARNING", "A selection only contained colons. Selection will be ignored.",
                          always_print = TRUE)
            next
        }

        # Count the colons, there may not be more than two, otherwise the function
        # should skip the variable.
        colon_count <- nchar(gsub("[^:]", "", variable))

        # In case of a single variable just add it to the vector
        if (colon_count == 0){
            variables <- c(variables, variable)
        }
        # In case of a variable range
        else if (colon_count == 1){
            # When at the end ("text:"), select all variables which start with the characters
            # coming before the colon. ":" acts as a placeholder for those who come after.
            if (endsWith(variable, ":")){
                search_term <- substring(variable, 1, nchar(variable) - 1)
                matches     <- names(data_frame)[startsWith(names(data_frame), search_term)]

                variables   <- c(variables, matches)
            }
            # When at the start (":text"), select all variables which end with the characters
            # following the colon. ":" acts as a placeholder for everything that comes before.
            else if (startsWith(variable, ":")){
                search_term <- substring(variable, 2)
                matches     <- names(data_frame)[endsWith(names(data_frame), search_term)]

                variables   <- c(variables, matches)
            }
            # If colon is in the middle ("var1:var10"), select a range of varaibles between
            # these two.
            else{
                parts     <- strsplit(variable, ":", fixed = TRUE)[[1]]

                variables <- c(variables, data_frame |> vars_between(parts[1], parts[2]))
            }
        }
        # In case there are two colons, one at the start and one at the end (":text:"),
        # look for the text inbetween inside the variable names.
        else if (colon_count == 2 && startsWith(variable, ":") && endsWith(variable, ":")){
            search_term <- gsub(":", "", variable)
            matches     <- names(data_frame)[grepl(search_term, names(data_frame), fixed = TRUE)]

            variables   <- c(variables, matches)

        }
        # If there are more than two colons skip variable
        else{
            print_message("WARNING", c("Selection '[variable]' contains more than two colons which is\n",
									   "not allowed. Selection will be ignored."), variable = variable,
									   always_print = TRUE)
        }
    }

    variables
}

#' Export Data Frame With Style
#'
#' @description
#' [export_with_style()] prints a data frame as an individually styled 'Excel' table. Titles,
#' footnotes and labels for variable names can optionally be added.
#'
#' @param data_frame A data frame to print.
#' @param titles Specify one or more table titles.
#' @param footnotes Specify one or more table footnotes.
#' @param var_labels A list in which is specified which label should be printed for
#' which variable instead of the variable name.
#' @param workbook Insert a previously created workbook to expand the sheets instead of
#' creating a new file.
#' @param style A list of options can be passed to control the appearance of 'Excel' outputs.
#' Styles can be created with [excel_output_style()].
#' @param column_align Individually align the data frame columns in the final 'Excel' output.
#' @param output The following output formats are available: excel and excel_nostyle.
#' @param print TRUE by default. If TRUE prints the output, if FALSE doesn't print anything. Can be used
#' if one only wants to catch the output workbook.
#' @param monitor FALSE by default. If TRUE outputs two charts to visualize the functions time consumption.
#'
#' @details
#' [export_with_style()] is based on the 'SAS' procedure Proc Print, which outputs the data frame
#' as is into a styled table.
#'
#' @return
#' Returns a list with the data table containing the results for the table, the formatted
#' 'Excel' workbook and the meta information needed for styling the final table.
#'
#' @seealso
#' Creating a custom table style: [excel_output_style()], [modify_output_style()],
#' [number_format_style()], [modify_number_formats()].
#'
#' Global style options: [set_style_options()], [set_labels()].
#'
#' Other global options: [set_titles()], [set_footnotes()], [set_print()], [set_monitor()],
#' [set_na.rm()], [set_print()], [set_print_miss()], [set_output()].
#'
#' Combine Excel workbooks: [combine_into_workbook()].
#'
#' Functions that can handle styles: [frequencies()], [crosstabs()], [any_table()].
#'
#' @examples
#' # Example data frame
#' my_data <- dummy_data(100)
#' set_style_options(background_color = "")
#'
#' # Define style
#' set_style_options(column_widths = c(2, 15, 15, 15, 9))
#'
#' # Define titles and footnotes. If you want to add hyperlinks you can do so by
#' # adding "link:" followed by the hyperlink to the main text. Linking to another
#' # cell works with "cell:". To link to a file use "file:" an pass the full file
#' # path afterwards.
#' set_titles("This is title number 1",
#'            "This is title number 2 link: https://cran.r-project.org/",
#'            "This is title number 3 cell: W22",
#'            "This is title number 4 file: C:/MyFolder/MyFile.txt")
#'
#' set_footnotes("This is footnote number 1",
#'               "This is footnote number 2 file: C:/MyFolder/MyFile.txt",
#'               "This is footnote number 3 cell: W22",
#'               "This is footnote number 4 link: https://cran.r-project.org/")
#'
#' # Print styled data frame
#' my_data |> export_with_style()
#'
#' # Retrieve formatted workbook, original table and meta information for further usage
#' export_list <- my_data |> export_with_style()
#'
#' # To save a table as xlsx file you have to set the path and filename in the
#' # style element
#' # Example files paths
#' table_file <- tempfile(fileext = ".xlsx")
#'
#' # Note: Normally you would directly input the path ("C:/MyPath/") and
#' #       name ("MyFile.xlsx").
#' #       With the set_style_options you can also set a table style globally.
#' set_style_options(save_path  = dirname(table_file),
#'                   file       = basename(table_file),
#'                   sheet_name = "MyTable")
#'
#' my_data |> export_with_style()
#'
#' # In case you have a good amount of tables, you want to combine in a single
#' # workbook, you can also catch the outputs and combine them afterwards in one go.
#' set_style_options(sheet_name = "sheet1")
#' tab1 <- my_data |> export_with_style(print = FALSE)
#'
#' set_titles(NULL)
#' set_style_options(sheet_name = "sheet2")
#' tab2 <- my_data |> export_with_style(print = FALSE)
#'
#' set_footnotes(NULL)
#' set_style_options(sheet_name = "sheet3")
#' tab3 <- my_data |> export_with_style(print = FALSE, output = "excel_nostyle")
#'
#' # Every of the above tabs is a list, which contains the data table, an unstyled
#' # workbook and the meta information needed for the individual styling. These tabs
#' # can be input into the following function, which reads the meta information,
#' # styles each table individually and combines them as separate sheets into a
#' # single workbook.
#' combine_into_workbook(tab1, tab2, tab3)
#' combine_into_workbook(tab1, tab2, tab3, file = table_file)
#'
#' # Manual cleanup for example
#' unlink(table_file)
#'
#' # Global options are permanently active until the current R session is closed.
#' # There are also functions to reset the values manually.
#' reset_style_options()
#' reset_qol_options()
#' close_file()
#'
#' set_print(TRUE)
#' set_style_options(background_color = "FFFFFF")
#'
#' @export
export_with_style <- function(data_frame,
                              titles       = .qol_options[["titles"]],
                              footnotes    = .qol_options[["footnotes"]],
                              var_labels   = .qol_options[["var_labels"]],
                              workbook     = NULL,
                              style        = .qol_options[["excel_style"]],
                              column_align = NULL,
                              output       = .qol_options[["output"]],
                              print        = .qol_options[["print"]],
                              monitor      = .qol_options[["monitor"]]){
    print_start_message()
	print_step("GREY", "Error handling")

    #-------------------------------------------------------------------------#
    monitor_df <- NULL |> monitor_start("Error handling", "Error handling")
    #-------------------------------------------------------------------------#

    # Apply macros to save path and file
    if (!is.null(style[["save_path"]])){
        style[["save_path"]] <- macro(style[["save_path"]])
    }

    if (!is.null(style[["file"]])){
        style[["file"]] <- macro(style[["file"]])
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Workbook
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Check whether the whole result list of one of the tabulation functions was passed
    if (!is.null(workbook) && is.list(workbook)){
        # If the workbook object is part of the list, extract it
        if ("workbook" %in% names(workbook)){
            workbook <- workbook[["workbook"]]
        }
        # If the list is some other list without workbook object, abort
        else{
            print_message("ERROR", c("Workbook object is invalid. You have to provide a workbook object",
                                     "created with one of the tabulation functions. Tabulation will be aborted."))
            return(invisible(NULL))
        }
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Output
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Silent conversion of global options, which are invalid for any_table
    if (tolower(output) %in% c("console", "text")){
        output <- "excel"
    }

    # Check for invalid output option
    if (!tolower(output) %in% c("excel", "excel_nostyle")){
        print_message("WARNING", "<Output> format '[output]' not available. Using 'excel' instead.", output = output)

        output <- "excel"
    }
    else{
        output <- tolower(output)
    }

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Resolve macros
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    titles     <- apply_macro(titles)
    footnotes  <- apply_macro(footnotes)
    var_labels <- apply_macro(var_labels)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Prepare table format for output
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    # Grab all information, which is necessary to format the workbook. This list will be
    # returned at the end and can be grabbed by the workbook combine function.
    meta <- c(mget(c("titles", "footnotes", "var_labels", "style", "column_align", "output")), "DATA")

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Excel prepare", "Format")
    #-------------------------------------------------------------------------#

    # Setup styling in new workbook if no other is provided
    if (is.null(workbook)){
        workbook <- openxlsx2::wb_workbook() |>
            prepare_styles(list("title" = titles, "footnote" = footnotes), style)
    }
    # Update style options in provided workbook
    else{
        workbook <- workbook |>
            prepare_styles(list("title" = titles, "footnote" = footnotes), style)
    }

    monitor_df <- monitor_df |> monitor_end()

    # Style data frame for export
    wb_list <- format_df_excel(workbook, data_frame, titles, footnotes, var_labels,
                               style, column_align, output, monitor_df)

    wb         <- wb_list[[1]]
    monitor_df <- wb_list[[2]]

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Output formatted table into different formats
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    if (print){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Output tables", "Output tables")
        #---------------------------------------------------------------------#
        print_step("MAJOR", "Output")

        # Check if only save path or file name is specified. If only one is specified
        # print a note. Otherwise the file would not be saved and opened without a
        # hint to why the file wasn't saved.
        if (is.null(style[["save_path"]]) + is.null(style[["file"]]) == 1){
            if (is.null(style[["save_path"]])){
                print_message("NOTE", c("No save path specified. Both save path and file name with extension",
                                        "need to be specified in the global options or style parameter for",
                                        "the file to be saved. File won't be saved."))
            }
            else{
                print_message("NOTE", c("No file name specified. Both save path and file name with extension",
                                        "need to be specified in the global options or style parameter for",
                                        "the file to be saved. File won't be saved."))
            }
        }

        # If no save path or file provided just open workbook
        if (is.null(style[["save_path"]]) || is.null(style[["file"]])){
            if (interactive()){
                wb$open()
            }
        }
        else{
            # If no save path or file provided just open workbook
            if (!file.exists(style[["save_path"]])){
                print_message("WARNING", "Path does not exist: [style]", style = style[["save_path"]])

                if (interactive()){
                    wb$open()
                }
            }
            # Save file
            else{
                # NOTE: A CSV export like in any_table, crosstabs or frequencies makes no sense
                #       here. export_data should be used for that, since this function here is
                #       explicitly for styled data frame export.
                wb$save(file = paste0(style[["save_path"]], "/", style[["file"]]), overwrite = TRUE)
            }
        }
    }

    print_closing(5)

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_end()
    monitor_df |> monitor_plot(draw_plot = monitor)
    #-------------------------------------------------------------------------#

    invisible(structure(list("table"    = data_frame,
                             "workbook" = wb,
                             "meta"     = meta), class = "qol_styled_data"))
}


###############################################################################
# Format data frame for excel output
###############################################################################
#' Format Data Frame Output (Excel Based)
#'
#' @description
#' Format data frame according to the provided style options.
#'
#' @param wb An already created workbook to add more sheets to.
#' @param data_frame The data frame which contains the information for this cross
#' table..
#' @param titles Character vector of titles to display above the table.
#' @param footnotes Character vector of footnotes to display under the table.
#' @param var_labels List which contains column variable names and their respective labels.
#' @param style A list containing the styling elements.
#' @param column_align Individually align the data frame columns in the final 'Excel' output.
#' @param output Determines whether to style the output or to just quickly paste
#' the data.
#' @param monitor_df Data frame which stores the monitoring values.
#'
#' @return
#' Returns a list containing a formatted Excel workbook as well as the monitoring
#' data frame.
#'
#' @noRd
format_df_excel <- function(wb,
                            data_frame,
                            titles,
                            footnotes,
                            var_labels,
                            style,
                            column_align,
                            output,
                            monitor_df){

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Add data
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_start("Excel prepare", "Format")
    #-------------------------------------------------------------------------#

    # Get table ranges
    df_ranges <- get_df_ranges(data_frame, titles, footnotes, style)

    data_frame <- data_frame |> set_labels_as_names(var_labels)

    # Add sheet
    wb$add_worksheet(style[["sheet_name"]], grid_lines = style[["grid_lines"]])

    # Add table data and format according to style options
    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Excel data", "Format")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Writing data to workbook")

    wb$add_data(x           = data_frame,
                start_col   = style[["start_column"]],
                start_row   = df_ranges[["header.row"]],
                col_names   = TRUE,
                with_filter = style[["filters"]],
                na.strings  = style[["na_symbol"]])

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Apply style
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    #-------------------------------------------------------------------------#
    monitor_df <- monitor_df |> monitor_next("Excel titles/footnotes", "Format")
    #-------------------------------------------------------------------------#
    print_step("MAJOR", "Formatting data")
    print_step("MINOR", "Format titles and footnotes")

    # Format titles and footnotes if there are any
    wb <- wb |>
        format_titles_foot_excel(titles, footnotes, df_ranges, style, output)

    # Only do the formatting when user specified it. With the excel_nostyle
    # option this whole part gets omitted to get a very quick unformatted
    # excel output.
    if (output == "excel"){
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel cell styles", "Format")
        #---------------------------------------------------------------------#
        print_step("MINOR", "Add cell styles")

        wb <- wb |> handle_cell_styles(df_ranges, style)

        # Fill all cells with background color
        if (style[["background_color"]] != ""){
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Excel background", "Format")
            #-----------------------------------------------------------------#
            print_step("MINOR", "Fill background")

            wb$add_fill(color = openxlsx2::wb_color(style[["background_color"]]))
            wb$set_cell_style_across(cols = "A:XFD", style = wb$get_cell_style(dims = "A1"))

            # Titles and footnotes need to get an extra fill
            wb$add_fill(dims = df_ranges[["title_range"]],    color = openxlsx2::wb_color(style[["background_color"]]))
            wb$add_fill(dims = df_ranges[["footnote_range"]], color = openxlsx2::wb_color(style[["background_color"]]))
        }

        # Format individual column alignments
        if (!is.null(column_align)){
            print_step("MINOR", "Individual column alignment")

            single_columns <- split_up_ranges(df_ranges[["table_range"]])
            column_align   <- fill_or_trim(column_align, length(single_columns))

            for (i in seq_len(length(single_columns))){
                wb$add_cell_style(dims       = single_columns[i],
                                  horizontal = column_align[i])
            }
        }

        # Apply number formats
        if (df_ranges[["num_format.length"]] > 0){
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Excel number formats", "Format")
            #-----------------------------------------------------------------#
            print_step("MINOR", "Format numbers")

            # Set up inner table number formats
            for (i in 1:df_ranges[["num_format.length"]]){
                wb$add_cell_style(dims                = df_ranges[[paste0("df_col_ranges", i)]],
                                  apply_number_format = TRUE,
                                  num_fmt_id          = wb$styles_mgr$get_numfmt_id(paste0(df_ranges[[paste0("df_col_types", i)]], "_numfmt")))
                wb$add_ignore_error(dims = df_ranges[[paste0("df_col_ranges", i)]], number_stored_as_text = TRUE)
            }
        }

        # Draw inner table cells as heat map with conditional formatting
        if (style[["as_heatmap"]]){
            #-----------------------------------------------------------------#
            monitor_df <- monitor_df |> monitor_next("Excel format heatmap", "Format")
            #-----------------------------------------------------------------#
            print_step("MINOR", "Add heatmap")

            wb$add_conditional_formatting(dims  = df_ranges[["table_range"]],
                                          style = c(style[["heatmap_low_color"]],
                                                    style[["heatmap_middle_color"]],
                                                    style[["heatmap_high_color"]]),
                                          type  = "colorScale")
        }

        # Freeze headers. If both options are true they have to be set together, otherwise one
        # option would overwrite the other.
        if (style[["freeze_col_header"]] && style[["freeze_row_header"]]){
            wb$freeze_pane(first_active_col = df_ranges[["header.column"]] + df_ranges[["cat_col.width"]],
                           first_active_row = df_ranges[["table.row"]])
        }
        else if (style[["freeze_col_header"]]){
            wb$freeze_pane(first_active_col = df_ranges[["header.column"]] + df_ranges[["cat_col.width"]])
        }
        else if (style[["freeze_row_header"]]){
            wb$freeze_pane(first_active_row = df_ranges[["table.row"]])
        }

        # Adjust table dimensions
        #---------------------------------------------------------------------#
        monitor_df <- monitor_df |> monitor_next("Excel widths/heights", "Format")
        #---------------------------------------------------------------------#
        print_step("MINOR", "Adjust row heights and column widths")

        wb <- wb |> handle_col_row_dimensions(df_ranges,
                                              collapse::fncol(data_frame) + (style[["start_column"]] - 1),
                                              collapse::fnrow(data_frame) + (style[["start_row"]] - 1),
                                              style) |>
            handle_any_auto_dimensions(df_ranges, style) |>
            handle_header_table_dim(df_ranges, style)

        wb$add_ignore_error(dims = df_ranges[["whole_tab_range"]], number_stored_as_text = TRUE)

        wb$add_named_region(dims = df_ranges[["whole_tab_range"]], name = "table", local_sheet = TRUE)
        wb$add_named_region(dims = df_ranges[["table_range"]],     name = "data",  local_sheet = TRUE)
    }

    monitor_df <- monitor_df |> monitor_end()

    # Return workbook
    list(wb, monitor_df)
}


#' Insert Variable Labels
#'
#' @description
#' Give the variables in the column header a custom label.
#'
#' @param column_header The complete column multi header.
#' @param var_labels A list in which is specified which label should be printed for
#' which variable instead of the variable name.
#'
#' @return
#' Returns a multi layered column header with replaced variable texts.
#'
#' @noRd
set_labels_as_names <- function(data_frame, var_labels){
    if (length(var_labels) == 0){
        return(data_frame)
    }

    valid_labels <- data_frame |> part_of_df(names(var_labels))

    # If no valid variable names for renaming are provided, return
    if (length(valid_labels) == 0){
        return(data_frame)
    }

    # Only keep valid labels
    var_labels <- var_labels[names(var_labels) %in% valid_labels]

    # Rename variables according to specified labels
    collapse::frename(data_frame,
                      stats::setNames(names(var_labels),
                                      as.character(var_labels)),
                      .nse = FALSE)
}

#' Apply Formats to a Data Frame
#'
#' @description
#' Used to apply formats to a data frame. Handles single labels with a simple join
#' and also multilabels by computing the cartesian product.
#'
#' @param data_frame The data frame to which formats should be joined.
#' @param formats A list in which is specified which formats should be applied
#' to which variables.
#' @param group_vars A vector containing all grouping variables.
#'
#' @return
#' Returns a data table to which a format data frame was joined.
#'
#' @noRd
apply_format <- function(data_frame, formats, group_vars = NULL, print_miss = TRUE){
    if (length(formats) == 0){
        return(data_frame)
    }

    arguments     <- formats

    # See comment above about numeric conversion
    inverse_group <- group_vars[!group_vars %in% names(arguments)]
    temp_data     <- data_frame |> convert_numeric(inverse_group)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Loop through all given variables and join each format with the data frame
    # at a time
    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    for (current_var in names(arguments)){
        format_df <- arguments[[current_var]]

        if (is.null(format_df)){
            print_message("WARNING", "Format for '[current_var]' not found. Formatting for variable '[current_var]' will be omitted.",
                          current_var = current_var, always_print = TRUE)
            next
        }

        if (is.factor(temp_data[[current_var]])){
            print_message("NOTE", c("'[current_var]' is a factor variable. Formats only work if the visible character values",
									"are specified as input values and not the factor levels."), current_var = current_var,
									always_print = TRUE)
        }

        # Security checks to ensure types are right
        if (!is.null(group_vars)){
            if (!current_var %in% group_vars){
                # No Warning because this exit is used to omit missing group variables
                next
            }
        }

        if (!current_var %in% names(temp_data)){
            print_message("WARNING", "Variable '[current_var]' not found in the input data frame. No Format will be applied.",
                          current_var = current_var)
            next
        }

        # Look up variable names in format data frame to check whether it is an
        # interval or discrete format
        interval_variables <- c("from", "to")
        actual_variables <- names(format_df)[1:2]

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # In case of interval format
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        if (identical(interval_variables, actual_variables)){
            # Separate NAs from rest of the data frame because the used join
            # can't handle them
            na_positions <- is.na(temp_data[[current_var]])
            temp_na      <- temp_data |> collapse::fsubset(na_positions)
            temp_data    <- temp_data |> collapse::fsubset(!na_positions)

            # Generate pseudo variables for range merging
            temp_data[["qol_from"]] <- temp_data[[current_var]]
            temp_data[["qol_to"]]   <- temp_data[[current_var]]

            data_frame[["qol_ID"]] <- seq_len(collapse::fnrow(data_frame))

            # Make a copy of format data frame or otherwise the original will be
            # altered by the following key sorting
            format_dt <- data.table::copy(format_df)

            # Set key variables
            data.table::setkey(format_dt, from, to)

            # Merge data frame with format by range
            temp_data <- data.table::foverlaps(temp_data, format_dt,
                                               by.x = c("qol_from", "qol_to"),
                                               by.y = c("from", "to"))

            temp_data |> collapse::fselect(current_var, "qol_from", "qol_to", "from", "to") <- NULL
            temp_data <- temp_data |> collapse::frename("label" = current_var, .nse = FALSE)

            # Put NAs back into full data frame
            temp_data <- data.table::rbindlist(list(temp_data, temp_na), fill = TRUE)

            all_levels <- format_df[3] |>
                unlist(use.names = FALSE) |>
                collapse::funique() |>
                collapse::na_omit()
        }

        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # In case of discrete format
        #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        else{
            # Rename label column to be specific to the variable
            format_df <- format_df |> collapse::frename("value" = current_var, .nse = FALSE)

            # Perform a cartesian join which joins all possibilities together, meaning it doubles observations
            temp_data <- collapse::join(temp_data, format_df,
                                        on       = current_var,
                                        how      = "left",
                                        multiple = TRUE,
                                        verbose  = FALSE,
										overid   = 2)

            # If not all values are represented in the format container, check where there are gaps
            # and fill them at the affected positions. This implements the "other" keyword in formats.
            na_positions <- which(is.na(temp_data[["label"]]) & !is.na(temp_data[[current_var]]))
            if (length(na_positions) > 0){
                if (as.character(.Machine[["integer.max"]]) %in% tolower(format_df[[current_var]])){
                    temp_data[["label"]][na_positions] <- format_df[["label"]][tolower(format_df[[current_var]]) == as.character(.Machine[["integer.max"]])]
                }
                else if(.Machine[["integer.max"]] %in% format_df[[current_var]]){
                    temp_data[["label"]][na_positions] <- format_df[["label"]][format_df[[current_var]] == .Machine[["integer.max"]]]
                }
                else{
                    temp_data[["label"]][na_positions] <- temp_data[[current_var]][na_positions]
                }
            }

            # Drop current variable and rename newly joined label to current variable name
            temp_data |> collapse::fselect(current_var) <- NULL
            temp_data <- temp_data |> collapse::frename("label" = current_var, .nse = FALSE)

            # Extract the number of labels from format
            all_levels <- format_df[-1] |>
                unlist(use.names = FALSE) |>
                collapse::funique() |>
                collapse::na_omit()
        }

        # Make sure that the labels will appear in order of the format when
        # applied to a variable in a data frame
        temp_data[[current_var]] <- factor(
            temp_data[[current_var]],
            levels  = all_levels,
            ordered = TRUE)
    }

    temp_data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/free_memory.R
\name{free_memory}
\alias{free_memory}
\title{Remove Objects From Memory By Name And Type}
\usage{
free_memory(names = NULL, types = NULL, envir = .GlobalEnv)
}
\arguments{
\item{names}{Object names to be removed.}

\item{types}{Object types to be removed.}

\item{envir}{The environment to remove the objects from.}
}
\value{
Returns removed objects.
}
\description{
Remove objects by name or type to free up memory. Uses the base \code{\link[=rm]{rm()}} function
but provides more flexible ways to remove objects.
}
\details{
\code{\link[=free_memory]{free_memory()}} is based on the 'SAS' function Proc Datasets. Among other things
this procedure is able to remove datasets from memory. But not only by writing
out the full file name, it is capable of looking for datasets starting with, ending
with or containing a certain text. Additionally certain object types can be removed.
}
\examples{
# Example data frames
my_data1 <- dummy_data(10)
my_data2 <- dummy_data(10)
data     <- dummy_data(10)
file     <- dummy_data(10)

# Free memory by name
free_memory("my_:")
free_memory(c("data", "file"))

# Free by type
my_data1 <- dummy_data(10)
my_data2 <- dummy_data(10)
data     <- dummy_data(10)
file     <- dummy_data(10)

free_memory(type = "data.frame")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/excel_helpers.R
\name{number_format_style}
\alias{number_format_style}
\title{Number Formats Used by \code{\link[=any_table]{any_table()}}}
\usage{
number_format_style(
  pct_excel = "0.0",
  freq_excel = "#,###,##0",
  freq.g0_excel = "#,###,##0",
  sum_excel = "#,###,##0",
  sum.wgt_excel = "#,###,##0",
  mean_excel = "#,###,##0",
  median_excel = "#,###,##0",
  mode_excel = "#,###,##0",
  min_excel = "#,###,##0",
  max_excel = "#,###,##0",
  sd_excel = "#,###,##0.000",
  variance_excel = "#,###,##0.000",
  first_excel = "#,###,##0",
  last_excel = "#,###,##0",
  p_excel = "#,###,##0",
  missing_excel = "#,###,##0",
  pct_decimals = 1,
  freq_decimals = 0,
  freq.g0_decimals = 0,
  sum_decimals = 3,
  sum.wgt_decimals = 3,
  mean_decimals = 2,
  median_decimals = 2,
  mode_decimals = 2,
  min_decimals = 2,
  max_decimals = 2,
  sd_decimals = 3,
  variance_decimals = 3,
  first_decimals = 0,
  last_decimals = 0,
  p_decimals = 2,
  missing_decimals = 0
)
}
\arguments{
\item{pct_excel}{Number format for percentage applied in Excel workbook.}

\item{freq_excel}{Number format for frequency applied in Excel workbook.}

\item{freq.g0_excel}{Number format for frequency greater zero applied in Excel workbook.}

\item{sum_excel}{Number format for sum applied in Excel workbook.}

\item{sum.wgt_excel}{Number format for sum of weights applied in Excel workbook.}

\item{mean_excel}{Number format for mean applied in Excel workbook.}

\item{median_excel}{Number format for median applied in Excel workbook.}

\item{mode_excel}{Number format for mode applied in Excel workbook.}

\item{min_excel}{Number format for min applied in Excel workbook.}

\item{max_excel}{Number format for max applied in Excel workbook.}

\item{sd_excel}{Number format for sd applied in Excel workbook.}

\item{variance_excel}{Number format for variance applied in Excel workbook.}

\item{first_excel}{Number format for first applied in Excel workbook.}

\item{last_excel}{Number format for last applied in Excel workbook.}

\item{p_excel}{Number format for percentile applied in Excel workbook.}

\item{missing_excel}{Number format for missing applied in Excel workbook.}

\item{pct_decimals}{Number of decimals for percentage.}

\item{freq_decimals}{Number of decimals for frequency.}

\item{freq.g0_decimals}{Number of decimals for frequency greater zero.}

\item{sum_decimals}{Number of decimals for sum.}

\item{sum.wgt_decimals}{Number of decimals for sum of weights.}

\item{mean_decimals}{Number of decimals for mean.}

\item{median_decimals}{Number of decimals for median.}

\item{mode_decimals}{Number of decimals for mode.}

\item{min_decimals}{Number of decimals for min.}

\item{max_decimals}{Number of decimals for max.}

\item{sd_decimals}{Number of decimals for sd.}

\item{variance_decimals}{Number of decimals for variance.}

\item{first_decimals}{Number of decimals for first.}

\item{last_decimals}{Number of decimals for last.}

\item{p_decimals}{Number of decimals for percentile.}

\item{missing_decimals}{Number of decimals for missing.}
}
\value{
Returns a list of named number format options.
}
\description{
Set individual number formats for the different statistics in tables produced
with \code{\link[=any_table]{any_table()}}.
}
\details{
\code{\link[=number_format_style]{number_format_style()}} is based on 'SAS' number formats and the Output Delivery
System (ODS), which provides efficient and readable ways to set up different table
styles.

With the number format style you have full control over formatting numbers according
to the different statistics. There is no need to think about calculating the right
place to input the number formats and how to do this in a loop for multiple cells.
Just input the different number formats and decimals for the different statistics
and everything else is handled by the functions capable of using number styles.

The concept basically is: design over complex calculations.
}
\examples{
# For default values
format_list <- number_format_style()

# Set specific options, the rest will be set to default values
format_list <- number_format_style(pct_excel    = "0.00000000",
                                   pct_decimals = 8)

# IMPORTANT: Don't forget to add individual formats to an excel style, otherwise
# they won't come into affect.
excel_style <- excel_output_style(number_formats = format_list)

}
\seealso{
Creating a custom table style: \code{\link[=excel_output_style]{excel_output_style()}}, \code{\link[=modify_output_style]{modify_output_style()}},
\code{\link[=modify_number_formats]{modify_number_formats()}}.

Global style options: \code{\link[=set_style_options]{set_style_options()}}, \code{\link[=set_labels]{set_labels()}}.

Functions that can handle styles: \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}}, \code{\link[=any_table]{any_table()}},
\code{\link[=export_with_style]{export_with_style()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dummy_data.R
\name{dummy_data}
\alias{dummy_data}
\title{Dummy Data}
\usage{
dummy_data(
  no_obs = 25000,
  insert_na = TRUE,
  monitor = .qol_options[["monitor"]]
)
}
\arguments{
\item{no_obs}{Number of observations.}

\item{insert_na}{TRUE by default. Inserts random NA values into variables.}

\item{monitor}{FALSE by default. If TRUE outputs two charts to visualize the functions time consumption.}
}
\value{
Returns a dummy data table.
}
\description{
The dummy data frame contains a few randomly generated variables like year, sex,
age, income and weight to test out functionalities. It can be generated with the desired
number of observations.
}
\examples{
my_data <- dummy_data(1000)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renaming.R
\name{rename_pattern}
\alias{rename_pattern}
\title{Replace Patterns Inside Variable Names}
\usage{
rename_pattern(data_frame, old_pattern, new_pattern)
}
\arguments{
\item{data_frame}{The data frame in which there are variables to be renamed.}

\item{old_pattern}{The pattern which should be replaced in the variable names.}

\item{new_pattern}{The pattern which should be set in place for the old one.}
}
\value{
Returns a data frame with renamed variables.
}
\description{
Replace a certain pattern inside a variable name with a new one. This can be
used if there are multiple different variable names which have a pattern in
common (e.g. all end in "_sum" but start different), so that there don't have
to be multiple rename variable calls.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Summarise data
all_nested <- my_data |>
    summarise_plus(class      = c(year, sex),
                   values     = c(weight, income),
                   statistics = c("sum", "pct_group", "pct_total", "sum_wgt", "freq"),
                   weight     = weight,
                   nesting    = "deepest",
                   na.rm      = TRUE)

# Rename variables by repacing patterns
new_names <- all_nested |>
    rename_pattern("pct", "percent") |>
    rename_pattern("_sum", "")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_export.R
\name{import_export}
\alias{import_export}
\alias{import_data}
\alias{import_multi}
\alias{export_data}
\alias{export_multi}
\title{High Level Import From And Export To CSV And XLSX}
\usage{
import_data(
  infile,
  sheet = 1,
  region = NULL,
  separator = "auto",
  decimal = "auto",
  var_names = TRUE
)

import_multi(
  file_list,
  sheet = "all",
  region = NULL,
  separator = "auto",
  decimal = "auto",
  var_names = TRUE
)

export_data(
  data_frame,
  outfile,
  separator = ";",
  decimal = ",",
  var_names = TRUE
)

export_multi(
  file_list,
  out_path,
  into_sheets = TRUE,
  separator = NULL,
  decimal = ",",
  var_names = TRUE
)
}
\arguments{
\item{infile}{Full file path with extension to a csv or xlsx file to be imported.}

\item{sheet}{Only used in xlsx import. Which sheet of the workbook to import.}

\item{region}{Only used in xlsx import. Can either be an 'Excel' range like 'A1:BY27'
or the name of a named region.}

\item{separator}{Only used in CSV/TXT-export. Defines the single character value separator.}

\item{decimal}{Only used in CSV/TXT-export. Defines the single character decimal character.}

\item{var_names}{TRUE by default. Whether to export variable names or not.}

\item{file_list}{\code{\link[=import_multi]{import_multi()}}: A character vector containing full file paths.

\code{\link[=export_multi]{export_multi()}}: A list of data frames.}

\item{data_frame}{A data frame to export.}

\item{outfile}{Full file path with extension. Allowed extensions are ".csv", ".txt" and ".xlsx".}

\item{out_path}{The file path where to save the exported files.}

\item{into_sheets}{TRUE by default. Whether to export all data frames into multiple
sheets of a single xlsx file or into separate xlsx files.}
}
\value{
Returns a data frame.

Multi functions: Returns a list of data frames.
}
\description{
\code{\link[=import_data]{import_data()}}: A wrapper for \code{\link[data.table:fread]{data.table::fread()}} and \code{\link[openxlsx2:wb_to_df]{openxlsx2::wb_to_df()}},
providing basic import functionality with minimal code. Uses "Latin-1" encoding.

\code{\link[=import_multi]{import_multi()}}: Runs multiple imports on all provided files. Is able to import
all sheets from xlsx files.

\code{\link[=export_data]{export_data()}}: A wrapper for \code{\link[data.table:fwrite]{data.table::fwrite()}} and \code{\link[openxlsx2:wb_save]{openxlsx2::wb_save()}},
providing basic export functionality with minimal code.

\code{\link[=export_multi]{export_multi()}}: Runs multiple exports on a list of data frames. Is able to export
to a single xlsx file with multiple sheets.
}
\details{
\code{\link[=import_data]{import_data()}} and \code{\link[=export_data]{export_data()}} are based on the 'SAS' procedures Proc Import and Proc Export,
which provide a very straight forward syntax. While 'SAS' can import many different formats with
these procedures, these 'R' versions concentrate on importing CSV, TXT and XLSX files.

The main goal here is to just provide as few as possible parameters to tackle most of the imports
and exports. These error handling also tries to let an import and export happen, even though
a parameter wasn't provided in the correct way.
}
\examples{
# Example files
csv_file  <- system.file("extdata", "qol_example_data_csv.csv",   package = "qol")
txt_file  <- system.file("extdata", "qol_example_data_txt.txt",   package = "qol")
xlsx_file <- system.file("extdata", "qol_example_data_xlsx.xlsx", package = "qol")

# Import: Provide full file path
my_csv  <- import_data(csv_file)
my_csv  <- import_data(txt_file)
my_xlsx <- import_data(xlsx_file)

# Import specific regions
range_import <- import_data(xlsx_file, region = "B4:H32")
name_import  <- import_data(xlsx_file, region = "test_region")

# Import from another sheet
sheet_import <- import_data(xlsx_file, sheet = "Sheet 2")

# Import multiple files at once
all_files <- import_multi(c(csv_file, txt_file, xlsx_file))

# Example data frame
my_data <- dummy_data(100)

# Example export file paths
export_csv  <- tempfile(fileext = ".csv")
export_txt  <- tempfile(fileext = ".txt")
export_xlsx <- tempfile(fileext = ".xlsx")

# Export: Provide full file path
my_data |> export_data(export_csv)
my_data |> export_data(export_txt)
my_data |> export_data(export_xlsx)

# Example data frame list
my_list <- list(first      = dummy_data(10),
                second.txt = dummy_data(10))

# Export multiple data frames into one xlsx file
# with multiple sheets
export_multi(my_list, tempdir())

# Export multiple data frames into multiple xlsx files
export_multi(my_list, tempdir(), into_sheets = FALSE)

# Export multiple data frames into multiple csv/txt files
export_multi(my_list, tempdir(), separator = ";")

# Manual cleanup for example
file1 <- file.path(tempdir(), "my_list.xlsx")
file2 <- file.path(tempdir(), "first.xlsx")
file3 <- file.path(tempdir(), "second.xlsx")
file4 <- file.path(tempdir(), "first.csv")
file5 <- file.path(tempdir(), "second.txt")

unlink(c(export_csv, export_xlsx,
         file1, file2, file3, file4, file5))

}
\seealso{
Functions that can export with style: \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}}, \code{\link[=any_table]{any_table()}},
\code{\link[=export_with_style]{export_with_style()}}.

Creating a custom table style: \code{\link[=excel_output_style]{excel_output_style()}}, \code{\link[=modify_output_style]{modify_output_style()}},
\code{\link[=number_format_style]{number_format_style()}}, \code{\link[=modify_number_formats]{modify_number_formats()}}.

Global style options: \code{\link[=set_style_options]{set_style_options()}}, \code{\link[=set_labels]{set_labels()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/small_helpers.R
\name{drop_type_vars}
\alias{drop_type_vars}
\title{Drop automatically generated Variables}
\usage{
drop_type_vars(data_frame)
}
\arguments{
\item{data_frame}{The data frame with automatically generated variables.}
}
\value{
Returns a data frame without the variables TYPE, TYPE_NR and DEPTH.
}
\description{
If \code{\link[=summarise_plus]{summarise_plus()}} is used with the nested options "all" or "single", three
hierarchical metadata variables are automatically generated: TYPE, TYPE_NR and
DEPTH. With this functions these variables are dropped.
\itemize{
\item \strong{\code{TYPE}} (character): The active grouping combination for the row (e.g., \code{"year+sex"}), or \code{"total"} for the grand total.
\item \strong{\code{TYPE_NR}} (integer): A unique sequential identifier assigned to each distinct grouping combination (\code{TYPE}).
\item \strong{\code{DEPTH}} (integer): The number of active grouping variables in the row's combination (e.g., \code{0} for \code{"total"}, \code{1} for single variables, etc.).
}
}
\examples{
# Example format
sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

# Example data frame
my_data <- dummy_data(1000)

# Call function
all_possible <- my_data |>
    summarise_plus(class      = c(year, sex),
                   values     = c(income, probability),
                   statistics = c("sum", "mean", "freq"),
                   formats    = list(sex = "sex."),
                   weight     = weight,
                   nesting    = "all",
                   na.rm      = TRUE) |>
    drop_type_vars()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_format.R
\name{expand_formats}
\alias{expand_formats}
\title{Expand A List Of Formats To Generate All Possible Combinations}
\usage{
expand_formats(..., names = NULL)
}
\arguments{
\item{...}{A list containing format data frames.}

\item{names}{Custom variable names for the newly generated data frame.}
}
\value{
Returns a data frames with all format label combinations.
}
\description{
Generates a data frame which contains all nested combinations of the provided
format labels.
}
\examples{
# Example formats
sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

# Expand formats
expand_df <- expand_formats(sex., age.,
                            names = c("sex", "age"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute.R
\name{compute.}
\alias{compute.}
\title{Compute New Variables}
\usage{
compute.(data_frame, ..., monitor = .qol_options[["monitor"]])
}
\arguments{
\item{data_frame}{A data frame in which to compute new variables.}

\item{...}{The calculations that should be executed.}

\item{monitor}{FALSE by default. If TRUE, outputs two charts to visualize the
functions time consumption.}
}
\value{
Returns a data frame with newly computed variables.
}
\description{
Compute new variables without having to write the name of the data frame multiple times.
It can handle all kinds of operations like simple value assignment or calculations, but also
can make use of other functions.

In addition it can be used in a conditional \code{\link[=do_if]{do_if()}} block.
}
\details{
The loop you can use within \code{\link[=compute.]{compute.()}} is based on the 'SAS' do-over-loop. This
type of loop iterates over every vector that appears in the loop in parallel.
Means that in the first iteration all the first vector elements are used, in the
second iteration all second elements of every vector, and so on. With this loop
you don't have the need to construct an outer loop, but can directly pass in
different vectors and let the function handle the loop inside.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Simple assignment
assign_df <- my_data |> compute.(new_var1 = 1,
                                new_var2 = "Hello")

# Simple calculation
sum_df <- my_data |> compute.(new_sum = age + sex)

# Using functions
mean_df <- my_data |> compute.(new_mean = collapse::fmean(age))

# Using qol functions
qol_df <- my_data |> compute.(row_sum = row_calculation("sum", state, age, sex))

# Use compute.() as a do-over-loop. In this kind of loop all vectors will be
# advanced one iteration at a time in parallel.
new_vars <- c("var1", "var2", "var3")
money    <- c("income", "expenses", "balance")
multi    <- c(1, 2, 3)

do_over_df <- my_data |> compute.(new_vars = money * multi)

# You can also do all at once
all_df <- my_data |> compute.(new_var1 = 1,
                              new_var2 = "Hello",
                              new_sum  = age + sex,
                              new_mean = mean(age),
                              row_sum  = row_calculation("sum", state, age, sex),
                              new_vars = multi * money)

# compute.() can be used in a do_if() situation and is aware of overarching
# conditions.
age_west. <- discrete_format("under 18"     = 0:17,
                             "18 and older" = 18:100)

age_east. <- discrete_format("under 65"     = 0:64,
                             "65 and older" = 65:100)

do_if_df <- my_data |>
    do_if(state < 11) |>
        compute.(region    = "West",
                 age_group = recode.(age = age_west.)) |>
    else_do() |>
        compute.(region    = "East",
                 age_group = recode.(age = age_east.)) |>
    end_do()

}
\seealso{
The following functions can make use of the \code{\link[=do_if]{do_if()}} filter variables:

Conditions: \code{\link[=if.]{if.()}}, \code{\link[=else_if.]{else_if.()}}, \code{\link[=else.]{else.()}}

Filter Data Frame: \code{\link[=where.]{where.()}}

Create new Variables: \code{\link[=compute.]{compute.()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statistics.R
\name{row_calculation}
\alias{row_calculation}
\title{Perform Row Wise Calculations}
\usage{
row_calculation(data_frame, statistics, ..., round_digits = NULL)
}
\arguments{
\item{data_frame}{A data frame in which are the values to be calculated.}

\item{statistics}{Available functions: "sum", "freq", "mean", "median", "mode",
"min", "max".}

\item{...}{Variable names of the value variables.}

\item{round_digits}{The number of decimal places the values should be rounded to.}
}
\value{
Returns a numeric vector.
}
\description{
Perform row wise calculations on numeric variables.
}
\examples{
# Example data frame
my_data <- data.frame(var1 = 1:5,
                      var2 = c(6, 7, 8, NA, 10),
                      var3 = 11:15)

# Calculate new variables
my_data[["sum"]]  <- my_data |> row_calculation("sum",  var1, var2, var3)
my_data[["mean"]] <- my_data |> row_calculation("mean", var1:var3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/any_table.R
\name{any_table}
\alias{any_table}
\title{Compute Any Possible Table}
\usage{
any_table(
  data_frame,
  rows,
  columns = "",
  values,
  statistics = NULL,
  pct_group = c(),
  pct_value = list(),
  pct_block = c(),
  compute = list(),
  formats = list(),
  by = c(),
  weight = NULL,
  order_by = "stats",
  titles = .qol_options[["titles"]],
  footnotes = .qol_options[["footnotes"]],
  var_labels = .qol_options[["var_labels"]],
  stat_labels = .qol_options[["stat_labels"]],
  box = "",
  workbook = NULL,
  style = .qol_options[["excel_style"]],
  output = .qol_options[["output"]],
  na.rm = .qol_options[["na.rm"]],
  print_miss = .qol_options[["print_miss"]],
  print = .qol_options[["print"]],
  monitor = .qol_options[["monitor"]]
)
}
\arguments{
\item{data_frame}{A data frame in which are the variables to tabulate.}

\item{rows}{A vector that provides single variables or variable combinations that
should appear in the table rows. To nest variables use the form:
"var1 + var2 + var3 + ...".}

\item{columns}{A vector that provides single variables or variable combinations that
should appear in the table rows. To nest variables use the form:
"var1 + var2 + var3 + ...".}

\item{values}{A vector containing all variables that should be summarised.}

\item{statistics}{Available functions:
\itemize{
\item "sum"       -> Weighted and unweighted sum
\item "sum_wgt"   -> Sum of all weights
\item "freq"      -> Unweighted frequency
\item "freq_g0"   -> Unweighted frequency of all values greater than zero
\item "pct_group" -> Weighted and unweighted percentages within the respective group
\item "pct_value" -> Weighted and unweighted percentages between value variables or of a variable expression
\item "pct_total" -> Weighted and unweighted percentages compared to the grand total
\item "mean"      -> Weighted and unweighted mean
\item "median"    -> Weighted and unweighted median
\item "mode"      -> Weighted and unweighted mode
\item "min"       -> Minimum
\item "max"       -> Maximum
\item "sd"        -> Weighted and unweighted standard deviation
\item "variance"  -> Weighted and unweighted standard variance
\item "first"     -> First value
\item "last"      -> Last value
\item "pn"        -> Weighted and unweighted percentiles (any p1, p2, p3, ... possible)
Uses the SAS default way of calculating percentiles, which is Type 2 described in \code{\link[=quantile]{quantile()}}.
\item "missing"   -> Missings generated by the value variables
}}

\item{pct_group}{This option is used to determine which variable of the row and column
variables should add up to 100 \%. Multiple variables can be specified in a vector to
generate multiple group percentages. You can also use the keywords "row_pct" or "col_pct"
to calculate total percentages for rows and columns regardless of the respective other dimension.}

\item{pct_value}{You can pass a list here, which contains the information for a new
variable name and between which of the value variables percentages should be computed.}

\item{pct_block}{Can be "rows" or "columns". Calculates percentages for the last
group of variables inside the individual row or column combinations.}

\item{compute}{A list in which individual calculations are handled. For the writing style
see the \code{\link[=compute.]{compute.()}} function.}

\item{formats}{A list in which is specified which formats should be applied to which variables.}

\item{by}{Compute tables stratified by the expressions of the provided variables.}

\item{weight}{Put in a weight variable to compute weighted results.}

\item{order_by}{Determine how the columns will be ordered. "values" orders the results by the
order you provide the variables in values. "stats" orders them by the order under statistics.
"values_stats" is a combination of both. "columns" keeps the order as given in columns
and "interleaved" alternates the stats.}

\item{titles}{Specify one or more table titles.}

\item{footnotes}{Specify one or more table footnotes.}

\item{var_labels}{A list in which is specified which label should be printed for
which variable instead of the variable name.}

\item{stat_labels}{A list in which is specified which label should be printed for
which statistic instead of the statistic name.}

\item{box}{Provide a text for the upper left box of the table.}

\item{workbook}{Insert a previously created workbook to expand the sheets instead of
creating a new file.}

\item{style}{A list of options can be passed to control the appearance of 'Excel' outputs.
Styles can be created with \code{\link[=excel_output_style]{excel_output_style()}}.}

\item{output}{The following output formats are available: excel and excel_nostyle.}

\item{na.rm}{FALSE by default. If TRUE removes all NA values from the variables.}

\item{print_miss}{FALSE by default. If TRUE outputs all possible categories of the
grouping variables based on the provided formats, even if there are no observations
for a combination.}

\item{print}{TRUE by default. If TRUE prints the output, if FALSE doesn't print anything. Can be used
if one only wants to catch the output data frame and workbook with meta information.}

\item{monitor}{FALSE by default. If TRUE, outputs two charts to visualize the functions time consumption.}
}
\value{
Returns a list with the data table containing the results for the table, the formatted
'Excel' workbook and the meta information needed for styling the final table.
}
\description{
\code{\link[=any_table]{any_table()}} produces any possible descriptive table in 'Excel' format. Any number
of variables can be nested and crossed. The output is an individually styled
'Excel' table, which also receives named ranges, making it easier to read the data back in.
}
\details{
\code{\link[=any_table]{any_table()}} is based on the 'SAS' procedure Proc Tabulate, which provides
efficient and readable ways to perform complex tabulations.

With this function you can combine any number of variables in any possible way, all
at once. You just define which variables or variable combinations should end up in
the table rows and columns with a simple syntax. Listing variables in a vector like
c("var1", "var2", "var3",...) means to put variables below (in case of the
row variables) or besides (in case of the column variables) each other. Nesting variables
is as easy as putting a plus sign between them, e.g. c("var1 + var2", "var2" + "var3" + "var4", etc.).
And of course you can combine both versions.

The real highlight is, that this function not only creates all the desired variable
combinations and exports them to an 'Excel' file, it prints a fully custom styled
table to a workbook. Setting up a custom, reusable style is as easy as setting up
options like: provide a color for the table header, set the font size for the row header,
should borders be drawn for the table cells yes/no, and so on. Merging doubled header texts,
happens automatically.

With this function you basically can fully concentrate on designing a table, instead of
thinking hard about how to calculate where to put a border or to even manually prepare
a designed workbook.
}
\examples{
# NOTE: Some of the examples use the parameter 'output = "excel_nostyle"' to
#       make them run faster so that more examples can be provided. Take out
#       the parameter to get the fully styled excel table.
#       The global option for printing the output is also set to FALSE for
#       the same reason. When running the examples manually, don't run this
#       option here.
set_print(FALSE)
set_style_options(background_color = "")

# Example data frame
my_data <- dummy_data(1000)
my_data[["person"]] <- 1

# Formats
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

# NOTE: "!" in front of an expression makes the expression available for
#       calculations but prevents it from beeing printed out.
education. <- discrete_format(
    "!Total"           = c("low", "middle", "high"),
    "low education"    = "low",
    "middle education" = "middle",
    "high education"   = "high")

state. <- discrete_format(
    "Germany"                       = 1:16,
    "Schleswig-Holstein"            = 1,
    "Hamburg"                       = 2,
    "Lower Saxony"                  = 3,
    "Bremen"                        = 4,
    "North Rhine-Westphalia"        = 5,
    "Hesse"                         = 6,
    "Rhineland-Palatinate"          = 7,
    "Baden-Württemberg"             = 8,
    "Bavaria"                       = 9,
    "Saarland"                      = 10,
    "West"                          = 1:10,
    "Berlin"                        = 11,
    "Brandenburg"                   = 12,
    "Mecklenburg-Western Pomerania" = 13,
    "Saxony"                        = 14,
    "Saxony-Anhalt"                 = 15,
    "Thuringia"                     = 16,
    "East"                          = 11:16)

# Define style
set_style_options(column_widths = c(2, 15, 15, 15, 9))

# Define titles and footnotes. If you want to add hyperlinks you can do so by
# adding "link:" followed by the hyperlink to the main text. Linking to another
# cell works with "cell:". To link to a file use "file:" an pass the full file
# path afterwards.
set_titles("This is title number 1",
           "This is title number 2 link: https://cran.r-project.org/",
           "This is title number 3 cell: W22",
           "This is title number 4 file: C:/MyFolder/MyFile.txt")

set_footnotes("This is footnote number 1",
              "This is footnote number 2 file: C:/MyFolder/MyFile.txt",
              "This is footnote number 3 cell: W22",
              "This is footnote number 4 link: https://cran.r-project.org/")

# Output complex tables with different percentages
my_data |> any_table(rows      = c("sex + age", "sex", "age"),
                     columns   = c("year", "education + year"),
                     values    = weight,
                     pct_group = c("row_pct", "col_pct"),
                     formats   = list(sex = sex., age = age.,
                                      education = education.),
                     na.rm     = TRUE)

# If you want to get a clearer vision of what the result table looks like, in terms
# of the row and column categories, you can write the code like this, to make out
# the variable crossings and see the order.
# my_data |> any_table(columns   = c(            "year", "education + year"),
#                      rows      = c("sex + age",
#                                    "sex",
#                                    "age"),
#                      values    = weight,
#                      pct_group = c("sex", "age"),
#                      formats   = list(sex = sex., age = age.,
#                                       education = education.),
#                      output    = "excel_nostyle",
#                      na.rm     = TRUE)

# Percentages based on value variables instead of categories
my_data |> any_table(rows       = c("age + year"),
                     columns    = "sex",
                     values     = c(probability, person),
                     statistics = c("pct_value", "sum", "freq"),
                     pct_value  = list(rate = "probability / person"),
                     weight     = weight,
                     formats    = list(sex = sex., age = age.),
                     output    = "excel_nostyle",
                     na.rm      = TRUE)

# Percentages based on a formatted variable expression
my_data |> any_table(rows       = c("age + year"),
                     columns    = "sex",
                     values     = c(probability, person),
                     pct_value  = list(sex = "Total", age = "Total"),
                     weight     = weight,
                     formats    = list(sex = sex., age = age.),
                     output    = "excel_nostyle",
                     na.rm      = TRUE)

# Percentages based on variable combination blocks
# NOTE: age + (year, education) becomes "age + year" and "age + education"
#       and will be sorted together. Also works in columns. You can also
#       do something like this: age + (year, sex + education, education).
my_data |> any_table(rows       = c("sex + (age, education)"),
                     columns    = "year",
                     values     = weight,
                     pct_block  = c("rows", "columns"),
                     formats    = list(sex = sex., age = age.,
                                       education = education.),
                     output     = "excel_nostyle",
                     na.rm      = TRUE)

# Individual calculations
my_data |> any_table(rows       = c("age + year"),
                     columns    = "sex",
                     values     = "probability",
                     statistics = c("sum", "sum_wgt"),
                     compute    = list(percent    = probability_sum * 100 / sum_wgt,
                                       pct_square = percent ^ 2),
                     weight     = weight,
                     formats    = list(sex = sex., age = age.),
                     output    = "excel_nostyle",
                     na.rm      = TRUE)

# Customize the visual appearance by adding variable and statistic labels. Both
# can also be set as a global option, if labels should be reused over multiple
# tables.
# Note: You don't have to describe every element. Sometimes a table can be more
# readable with less text. To completely remove a variable label just put in an
# empty text "" as label.#'
set_titles("This is title number 1",
           "This is title number 2",
           "This is title number 3")

set_footnotes("This is footnote number 1",
              "This is footnote number 2",
              "This is footnote number 3")

my_data |> any_table(rows        = c("age + year"),
                     columns     = "sex",
                     values      = weight,
                     statistics  = c("sum", "pct_group"),
                     order_by    = "interleaved",
                     formats     = list(sex = sex., age = age.),
                     var_labels  = list(age = "Age categories",
                                        sex = "", weight = ""),
                     stat_labels = list(pct = "\%"),
                     output      = "excel_nostyle",
                     na.rm       = TRUE)

# Individual styling can also be passed directly
my_style <- excel_output_style(header_back_color = "0077B6",
                               font              = "Times New Roman")

# my_data |> any_table(rows       = c("age + year"),
#                      columns    = "sex",
#                      values     = c(probability, person),
#                      statistics = c("pct_value", "sum", "freq"),
#                      pct_value  = list(rate = "probability / person"),
#                      weight     = weight,
#                      formats    = list(sex = sex., age = age.),
#                      style      = my_style,
#                      na.rm      = TRUE)

# Select specific statistics for specific variables
my_data |> any_table(rows       = "sex",
                     columns    = "year",
                     statistics = list("sum"       = c(weight, income),
                                       "pct_group" = balance),
                     formats    = list(sex = sex.),
                     output     = "excel_nostyle",
                     na.rm      = TRUE)

# Pass on workbook to create more sheets in the same file
my_style <- my_style |> modify_output_style(sheet_name = "age_sex")

result_list <- my_data |>
           any_table(rows       = "age",
                     columns    = "sex",
                     values     = weight,
                     statistics = "sum",
                     formats    = list(sex = sex., age = age.),
                     style      = my_style,
                     output     = "excel_nostyle",
                     na.rm      = TRUE,
                     print      = FALSE)

my_style <- my_style |> modify_output_style(sheet_name = "edu_year")

my_data |> any_table(workbook   = result_list[["workbook"]],
                     rows       = "education",
                     columns    = "year",
                     values     = weight,
                     statistics = "pct_group",
                     formats    = list(education = education.),
                     style      = my_style,
                     output     = "excel_nostyle",
                     na.rm      = TRUE)

# In case you have a good amount of tables, you want to combine in a single workbook,
# you can also catch the outputs and combine them afterwards in one go.
# NOTE: This section is commented out to just show the principle. Otherwise the
#       example code would run for too long.
# set_style_options(sheet_name = "sheet1")
# tab1 <- my_data |> any_table(..., print = FALSE, output = "excel_nostyle")
#
# set_style_options(sheet_name = "sheet2")
# tab2 <- my_data |> any_table(..., print = FALSE, output = "excel_nostyle")
#
# set_style_options(sheet_name = "sheet3")
# tab3 <- my_data |> any_table(..., print = FALSE, output = "excel_nostyle")
#
# ...
#
# Every of the above tabs is a list, which contains the data table, an unstyled
# workbook and the meta information needed for the individual styling. These tabs
# can be input into the following function, which reads the meta information, styles
# each table individually and combines them as separate sheets into a single workbook.
# combine_into_workbook(tab1, tab2, tab3 file = "C:/My_folder/My_workbook.xlsx")

# The result list from above also carries the transformed data frame if
# needed for further usage
any_table_df <- result_list[["table"]]

# Output multiple complex tables by expressions of another variable.
# If you specify the sheet name as "by" in the output style, the sheet
# names are named by the variable expressions of the by-variable. Otherwise
# the given sheet named gets a running number.
my_style <- my_style |> modify_output_style(sheet_name = "by")

my_data |> any_table(rows       = c("sex", "age"),
                     columns    = "education",
                     values     = weight,
                     by         = state,
                     statistics = c("sum", "pct_group"),
                     pct_group  = "education",
                     formats    = list(sex = sex., age = age., state = state.,
                                       education = education.),
                     output     = "excel_nostyle",
                     na.rm      = TRUE)

# To save a table as xlsx file you have to set the path and filename in the
# style element
# Example files paths
table_file <- tempfile(fileext = ".xlsx")

# Note: Normally you would directly input the path ("C:/MyPath/") and name ("MyFile.xlsx").
set_style_options(save_path  = dirname(table_file),
                  file       = basename(table_file),
                  sheet_name = "MyTable")

my_data |> any_table(rows    = "sex",
                     columns = "year",
                     values  = weight,
                     formats = list(sex = sex.))

# Manual cleanup for example
unlink(table_file)

# Global options are permanently active until the current R session is closed.
# There are also functions to reset the values manually.
reset_style_options()
reset_qol_options()
close_file()

set_print(TRUE)
set_style_options(background_color = "FFFFFF")

}
\seealso{
Creating a custom table style: \code{\link[=excel_output_style]{excel_output_style()}}, \code{\link[=modify_output_style]{modify_output_style()}},
\code{\link[=number_format_style]{number_format_style()}}, \code{\link[=modify_number_formats]{modify_number_formats()}}.

Global style options: \code{\link[=set_style_options]{set_style_options()}}, \code{\link[=set_labels]{set_labels()}}.

Other global options: \code{\link[=set_titles]{set_titles()}}, \code{\link[=set_footnotes]{set_footnotes()}}, \code{\link[=set_print]{set_print()}}, \code{\link[=set_monitor]{set_monitor()}},
\code{\link[=set_na.rm]{set_na.rm()}}, \code{\link[=set_print]{set_print()}}, \code{\link[=set_print_miss]{set_print_miss()}}, \code{\link[=set_output]{set_output()}}.

Combine Excel workbooks: \code{\link[=combine_into_workbook]{combine_into_workbook()}}.

Creating formats: \code{\link[=discrete_format]{discrete_format()}} and \code{\link[=interval_format]{interval_format()}}.

Functions that can handle formats and styles: \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}}.

Additional functions that can handle styles: \code{\link[=export_with_style]{export_with_style()}}

Additional functions that can handle formats: \code{\link[=summarise_plus]{summarise_plus()}}, \code{\link[=recode.]{recode.()}},
\code{\link[=recode_multi]{recode_multi()}}, \code{\link[=transpose_plus]{transpose_plus()}}, \code{\link[=sort_plus]{sort_plus()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renaming.R
\name{first_row_as_names}
\alias{first_row_as_names}
\title{Set First Data Frame Row As Variable Names}
\usage{
first_row_as_names(data_frame)
}
\arguments{
\item{data_frame}{A data frame for which to set new variable names.}
}
\value{
Returns a data frame with renamed variables.
}
\description{
Sets the first row of a data frame as variable names and deletes it. In case
of NA, numeric values or empty characters in the first row, the old names are kept.
}
\examples{
# Example data frame
my_data <- data.frame(
              var1 = c("id", 1, 2, 3),
              var2 = c(NA, "a", "b", "c"),
              var3 = c("value", 1, 2, 3),
              var4 = c("", "a", "b", "c"),
              var5 = c(1, 2, 3, 4))

my_data <- my_data |> first_row_as_names()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renaming.R
\name{remove_stat_extension}
\alias{remove_stat_extension}
\title{Replace Statistic From Variable Names}
\usage{
remove_stat_extension(data_frame, statistics)
}
\arguments{
\item{data_frame}{The data frame in which there are variables to be renamed.}

\item{statistics}{Statistic extensions that should be removed from the variable names.}
}
\value{
Returns a data frame with renamed variables.
}
\description{
Remove the statistic name from variable names, so that they get back their old
names without extension.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Summarise data
all_nested <- my_data |>
    summarise_plus(class      = c(year, sex),
                   values     = c(weight, income),
                   statistics = c("sum", "pct_group", "pct_total", "sum_wgt", "freq"),
                   weight     = weight,
                   nesting    = "deepest",
                   na.rm      = TRUE)

# Remove statistic extension
new_names <- all_nested |> remove_stat_extension("sum")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{qol_options}
\alias{qol_options}
\alias{set_print}
\alias{get_print}
\alias{set_monitor}
\alias{get_monitor}
\alias{set_na.rm}
\alias{get_na.rm}
\alias{set_print_miss}
\alias{get_print_miss}
\alias{set_output}
\alias{get_output}
\alias{set_titles}
\alias{get_titles}
\alias{set_footnotes}
\alias{get_footnotes}
\alias{set_threads}
\alias{get_threads}
\title{Set Global Print Option}
\usage{
set_print(...)

get_print()

set_monitor(...)

get_monitor()

set_na.rm(...)

get_na.rm()

set_print_miss(...)

get_print_miss()

set_output(...)

get_output()

set_titles(...)

get_titles()

set_footnotes(...)

get_footnotes()

set_threads(...)

get_threads()
}
\arguments{
\item{...}{\code{\link[=set_threads]{set_threads()}}: Put in the number of threads to use or NULL to reset.}
}
\value{
\code{\link[=set_print]{set_print()}}: Changed global print option.

\code{\link[=get_print]{get_print()}}: TRUE or FALSE.

\code{\link[=set_monitor]{set_monitor()}}: Changed global monitor option.

\code{\link[=get_monitor]{get_monitor()}}: TRUE or FALSE.

\code{\link[=set_na.rm]{set_na.rm()}}: Changed global na.rm option.

\code{\link[=get_na.rm]{get_na.rm()}}: TRUE or FALSE.

\code{\link[=set_print_miss]{set_print_miss()}}: Changed global print_miss option.

\code{\link[=get_print_miss]{get_print_miss()}}: TRUE or FALSE.

\code{\link[=set_output]{set_output()}}: Changed global output option.

\code{\link[=get_output]{get_output()}}: Current output option as character.

\code{\link[=set_titles]{set_titles()}}: Changed global titles.

\code{\link[=get_titles]{get_titles()}}: Current titles as character.

\code{\link[=set_footnotes]{set_footnotes()}}: Changed global footnotes.

\code{\link[=get_footnotes]{get_footnotes()}}: Current footnotes as character.

\code{\link[=set_threads]{set_threads()}}: Changed global number of used threads.

\code{\link[=get_threads]{get_threads()}}: Current number of used threads.
}
\description{
\code{\link[=set_print]{set_print()}}: Set the print option globally for the tabulation and export to
Excel functions.

\code{\link[=get_print]{get_print()}}: Get the globally stored print option.

\code{\link[=set_monitor]{set_monitor()}}: Set the monitor option globally for the heavier functions which are
able to show how they work internally.

\code{\link[=get_monitor]{get_monitor()}}: Get the globally stored monitor option.

\code{\link[=set_na.rm]{set_na.rm()}}: Set the na.rm option globally for each function which can remove
NA values.

\code{\link[=get_na.rm]{get_na.rm()}}: Get the globally stored na.rm option.

\code{\link[=set_print_miss]{set_print_miss()}}: Set the print_miss option globally for each function which can display
missing categories.

\code{\link[=get_print_miss]{get_print_miss()}}: Get the globally stored print_miss option.

\code{\link[=set_output]{set_output()}}: Set the output option globally for each function that can output
results to "console", "text", "excel" or "excel_nostyle".

\code{\link[=get_output]{get_output()}}: Get the globally stored output option.

\code{\link[=set_titles]{set_titles()}}: Set the titles globally for each function that can print titles
above the output table.

\code{\link[=get_titles]{get_titles()}}: Get the globally stored titles.

\code{\link[=set_footnotes]{set_footnotes()}}: Set the footnotes globally for each function that can print footnotes
above the output table.

\code{\link[=get_footnotes]{get_footnotes()}}: Get the globally stored footnotes.

\code{\link[=set_threads]{set_threads()}}: Globally sets the number of used threads for the save and load
file functions.

\code{\link[=get_threads]{get_threads()}}: Get the globally stored number of used threads.
}
\examples{
# Set and get the global print option
set_print(FALSE)
set_print(TRUE)

get_print()

# Set and get the global monitor option
set_monitor(TRUE)
set_monitor(FALSE)

get_monitor()

# Set and get the global NA removal option
set_na.rm(TRUE)
set_na.rm(FALSE)

get_na.rm()

# Set and get the global print missing option
set_print_miss(TRUE)
set_print_miss(FALSE)

get_print_miss()

# Set and get the global output option
set_output("excel")

get_output()

# Set and get table titles globally
set_titles("This is title number 1 link: https://cran.r-project.org/",
           "This is title number 2 cell: W22",
           "This is title number 3 file: C:/MyFolder/MyFile.docx",
           "This is title number 4")

# To reset titles do
set_titles(NULL)

get_titles()

# Set and get table footnotes globally
set_footnotes("This is footnote number 1 link: https://cran.r-project.org/",
              "This is footnote number 2 cell: W22",
              "This is footnote number 3 file: C:/MyFolder/MyFile.docx",
              "This is footnote number 4")

# To reset footnotes do
set_footnotes(NULL)

get_footnotes()

# Set and get used threads globally
set_threads(8)

get_threads()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sort_plus.R
\name{sort_plus}
\alias{sort_plus}
\title{Sort Data Frame Rows With Some Additions}
\usage{
sort_plus(
  data_frame,
  by,
  preserve = NULL,
  order = "ascending",
  formats = c(),
  na.last = TRUE
)
}
\arguments{
\item{data_frame}{A data frame to summarise.}

\item{by}{A variable vector which contains the variables to sort by.}

\item{preserve}{A vector containing all variables which current order should be
preserved.}

\item{order}{A vector containing the sorting order for each variable. 'ascending'/'a'
or 'descending'/'d' can be used. If there are less orders given than by variables
provided, the last given sorting order will be used for the additional by variables.}

\item{formats}{A list in which is specified which formats should be used to sort
certain variables.}

\item{na.last}{TRUE by default. Specifies whether NA values should come last or first.}
}
\value{
Returns a sorted data table.
}
\description{
Sort data frame rows by the provided variables. \code{\link[=sort_plus]{sort_plus()}} is also able to
preserve the current order of certain variables and only sort other variables
within this order. As another option one can sort a variable with the help of
formats, which can be used to e.g. sort a character variable in another than
alphabetical order without creating a temporary variable just for sorting.
}
\details{
\code{\link[=sort_plus]{sort_plus()}} is just very loosely based on the 'SAS' procedure Proc Sort. It tries
to keep the simplicity, but with some added features.
}
\examples{
# Example formats
education. <- discrete_format(
    "1" = "low",
    "2" = "middle",
    "3" = "high")

# Example data frame
my_data <- dummy_data(1000)

# Simple sorting
sort_df1 <- my_data |> sort_plus(by = c(state, sex, age))
sort_df2 <- my_data |> sort_plus(by    = c(state, sex, age),
                                 order = c("ascending", "descending"))

# Character variables will normally be sorted alphabetically. With the help
# of a format this variable can be sorted in a completely different way.
sort_df3 <- my_data |> sort_plus(by      = education,
                                 formats = list(education = education.))

# Preserve the order of the character variable, otherwise it couldn't stay in
# it's current order.
sort_df4 <- sort_df3 |> sort_plus(by       = age,
                                  preserve = education)

# Or do all at once
sort_df5 <- my_data |> sort_plus(by       = c(sex, education),
                                 preserve = state,
                                 formats = list(education = education.))

}
\seealso{
Creating formats: \code{\link[=discrete_format]{discrete_format()}} and \code{\link[=interval_format]{interval_format()}}.

Functions that also make use of formats: \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}},
\code{\link[=any_table]{any_table()}}, \code{\link[=recode.]{recode.()}}, \code{\link[=recode_multi]{recode_multi()}}, \code{\link[=transpose_plus]{transpose_plus()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/small_helpers.R
\name{duplicates}
\alias{duplicates}
\alias{get_duplicate_var_names}
\alias{get_duplicate_var_count}
\title{Check For Duplicate Variable Names}
\usage{
get_duplicate_var_names(data_frame)

get_duplicate_var_count(data_frame)
}
\arguments{
\item{data_frame}{The data frame which variable names to check for duplicates.}
}
\value{
\code{\link[=get_duplicate_var_names]{get_duplicate_var_names()}}: Returns a vector of duplicate variable names.

\code{\link[=get_duplicate_var_count]{get_duplicate_var_count()}}: Returns the count of duplicated variables as a
numeric value.
}
\description{
Checks for duplicate variable names in a data frame, e.g. AGE, age and Age.

Counts the number of duplicated variables in a data frame. If a variable
appears three times, e.g. AGE, age and Age, the variable count will be one.
}
\examples{
# Example data frame
my_data <- data.frame(age    = 1,
                      AGE    = 2,
                      Age    = 3,
                      sex    = 1)

dup_var_names <- my_data |> get_duplicate_var_names()

dup_var_count <- my_data |> get_duplicate_var_count()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/messages.R
\name{message_helpers}
\alias{message_helpers}
\alias{get_message_stack}
\alias{set_no_print}
\alias{set_no_color}
\alias{print_stack_as_messages}
\alias{convert_square_brackets}
\title{Message Helper Functions}
\usage{
get_message_stack(as_data_frame = FALSE)

set_no_print(value = FALSE)

set_no_color(value = TRUE)

print_stack_as_messages(type = NULL)

convert_square_brackets(text, ...)
}
\arguments{
\item{as_data_frame}{FALSE by default. If TRUE returns message stack information as
data frame.}

\item{value}{Can be TRUE or FALSE.}

\item{type}{The message type to filter.}

\item{text}{The text in which to replace the curly brackets.}

\item{...}{The actual replacement vectors.}
}
\value{
\code{\link[=get_message_stack]{get_message_stack()}}: Returns a list of messages or a data frame.

\code{\link[=set_no_print]{set_no_print()}}: Returns the global no_print option.

\code{\link[=set_no_color]{set_no_color()}}: Returns the global no_color option.

\code{\link[=print_stack_as_messages]{print_stack_as_messages()}}: Returns NULL.

\code{\link[=convert_square_brackets]{convert_square_brackets()}}: Returns formatted text.
}
\description{
\code{\link[=get_message_stack]{get_message_stack()}}: Retrieves the current message stack as list or data frame.

\code{\link[=set_no_print]{set_no_print()}}: FALSE by default. If set to TRUE the messages will be formatted
and returned but not printed to the console. Can e.g. be used in unit test
situations.

\code{\link[=set_no_color]{set_no_color()}}: TRUE by default. Suppresses the color codes so that messages
can be printed clean.

\code{\link[=print_stack_as_messages]{print_stack_as_messages()}}: Prints the message stack as actual messages (only
not suppressed messages). Can be used to trigger expect_message, expect_warning
or expect_error in unit tests.

\code{\link[=convert_square_brackets]{convert_square_brackets()}}: Convert different curly bracket patterns into ansi
formatting or passes the context of vectors into the text.
}
\details{
The message types in which you can enter custom texts, are capable of using
different styling operators. These are:
\itemize{
\item Insert list of elements: [vector_name]
\item Adding conditional words, if list of elements has more than one element: [?word]
\item Adding conditional singular/plural, depending on list of element length: [?singular/plural]
\item Bold, italic and underline: [b]some text[/b], [i]some text[/i], [u]some text[/u]
\item Coloring parts of the message: [#FF00FF some text]
}
}
\seealso{
Main printing functions: \code{\link[=print_message]{print_message()}}, \code{\link[=print_headline]{print_headline()}}, \code{\link[=print_start_message]{print_start_message()}},
\code{\link[=print_closing]{print_closing()}}, \code{\link[=print_step]{print_step()}}, \code{\link[=set_up_custom_message]{set_up_custom_message()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/split_by.R
\name{split_by}
\alias{split_by}
\title{Split Data Frame By Variable Expressions Or Condition}
\usage{
split_by(
  data_frame,
  ...,
  formats = list(),
  inverse = FALSE,
  monitor = .qol_options[["monitor"]]
)
}
\arguments{
\item{data_frame}{A data frame which should be split up into multiple data frames.}

\item{...}{Pass in one or multiple variables and/or conditions on which the provided
data frame should be splitted.}

\item{formats}{A list in which is specified which formats should be applied to which
variables.}

\item{inverse}{Uses the inverse conditions to split up the data frame.}

\item{monitor}{FALSE by default. If TRUE, outputs two charts to visualize the
functions time consumption.}
}
\value{
Returns a list of data frames split by variable expressions and/or conditions.
The lists names are the variable expressions or conditions.
}
\description{
Split up a data frame based on variable expressions or on conditions to receive
multiple smaller data frames. Both possibilities can be used at the same time.
}
\details{
\code{\link[=split_by]{split_by()}} is based on the explicit Output from 'SAS'. With the Output function
one can - among other things - explicitly tell 'SAS' which observation to output into
which data set. Which enables the user to output one observation into one or multiple
data sets.

Instead of subsetting the same data frame multiple times manually, you can subset it
multiple times at once with this function.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Split by variable expressions
split_var_df <- my_data |> split_by(sex)

# Split by conditions
split_cond_df <- my_data |> split_by(sex == 1 & age <  18,
                                     sex == 2 & age >= 18)

# Split by condition with inverse group
split_inv_df <- my_data |> split_by(sex == 1, inverse = TRUE)

# Split by variables and conditions
split_combi_df <- my_data |> split_by(state, education,
                                      sex == 1, age < 18)

# Split by variable expressions using formats
state. <- discrete_format(
    "Germany"                       = 1:16,
    "Schleswig-Holstein"            = 1,
    "Hamburg"                       = 2,
    "Lower Saxony"                  = 3,
    "Bremen"                        = 4,
    "North Rhine-Westphalia"        = 5,
    "Hesse"                         = 6,
    "Rhineland-Palatinate"          = 7,
    "Baden-Württemberg"             = 8,
    "Bavaria"                       = 9,
    "Saarland"                      = 10,
    "West"                          = 1:10,
    "Berlin"                        = 11,
    "Brandenburg"                   = 12,
    "Mecklenburg-Western Pomerania" = 13,
    "Saxony"                        = 14,
    "Saxony-Anhalt"                 = 15,
    "Thuringia"                     = 16,
    "East"                          = 11:16)

split_format_df <- my_data |> split_by(state,
                                       formats = list(state = state.))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/excel_helpers.R
\name{get_excel_range}
\alias{get_excel_range}
\title{Converts Numbers into 'Excel' Ranges}
\usage{
get_excel_range(
  row = NULL,
  column = NULL,
  from_row = NULL,
  from_column = NULL,
  to_row = NULL,
  to_column = NULL
)
}
\arguments{
\item{row}{Single row number.}

\item{column}{Single column number.}

\item{from_row}{Range start row.}

\item{from_column}{Range start column.}

\item{to_row}{Range end row.}

\item{to_column}{Range end column.}
}
\value{
Returns a character with an 'Excel' range.
}
\description{
Converts a column number into the according letter to form a cell reference
like it is used in 'Excel' (e.g "A1"). Also can compute a range from cell to
cell (e.g. "A1:BY22").
}
\examples{
single_cell <- get_excel_range(row = 1, column = 6)
range       <- get_excel_range(from_row = 1, from_column = 6,
                                 to_row = 5,   to_column = 35)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/messages.R
\name{messages}
\alias{messages}
\alias{print_message}
\alias{print_headline}
\alias{print_start_message}
\alias{print_closing}
\alias{print_step}
\alias{set_up_custom_message}
\title{Print Styled Messages}
\usage{
print_message(
  type,
  text,
  ...,
  always_print = FALSE,
  utf8 = .qol_messages[["format"]][["utf8"]],
  no_color = .qol_messages[["no_color"]]
)

print_headline(
  text,
  ...,
  line_char = "=",
  max_width = getOption("width"),
  always_print = FALSE,
  utf8 = .qol_messages[["format"]][["utf8"]],
  no_color = .qol_messages[["no_color"]]
)

print_start_message(
  current_time = Sys.time(),
  caller_color = "#63C2C9",
  always_print = FALSE,
  suppress = FALSE,
  utf8 = .qol_messages[["format"]][["utf8"]],
  no_color = .qol_messages[["no_color"]]
)

print_closing(
  time_threshold = 2,
  start_time = .qol_messages[["start_time"]],
  caller_color = "#63C2C9",
  always_print = FALSE,
  suppress = FALSE,
  utf8 = .qol_messages[["format"]][["utf8"]],
  no_color = .qol_messages[["no_color"]]
)

print_step(
  type,
  text,
  ...,
  in_place = FALSE,
  always_print = FALSE,
  utf8 = .qol_messages[["format"]][["utf8"]],
  no_color = .qol_messages[["no_color"]]
)

set_up_custom_message(
  type = "UNICORN",
  color = "#FF00FF",
  ansi_icon = NULL,
  text_icon = "^",
  ansi_wait_icon = NULL,
  text_wait_icon = "?",
  indent = 1,
  text_bold = FALSE,
  text_italic = FALSE,
  text_underline = FALSE,
  text_color = NULL,
  time_color = "#6B6B6B"
)
}
\arguments{
\item{type}{If displayed as a normal note, then this is the text displayed
in front of the message. This also appears as type in the message stack.}

\item{text}{The message text to display.}

\item{...}{Additional information to display like variable names. To use
these write [NAME YOU PUT IN] in the text.}

\item{always_print}{FALSE by default. If TRUE, prints headlines even in deeper
nested situations.}

\item{utf8}{Whether to display complex characters or just plain text.}

\item{no_color}{Whether to display color codes or suppress them.}

\item{line_char}{The character that that forms the line.}

\item{max_width}{The maximum number of characters drawn, which determines the
line length of the headline.}

\item{current_time}{The current time to create a time stamp.}

\item{caller_color}{Hex color code for the displayed caller function.}

\item{suppress}{FALSE by default. If TRUE triggers all the message
procedures to create a message stack but doesn't print the message to the
console.}

\item{time_threshold}{The total time spent is displayed in different colors from
green over yellow to red, depending on the threshold specified (in seconds).}

\item{start_time}{The time at which the function call started to calculate the
time difference and output the total time spent.}

\item{in_place}{Prints the step message on the same line as before, instead of in
the next line. This can e.g. be used inside loops.}

\item{color}{The color of the message type.}

\item{ansi_icon}{The icon used when message is displayed in utf8 mode.}

\item{text_icon}{The icon used when message is displayed in text only mode.}

\item{ansi_wait_icon}{The icon used when a timed message is waiting for the end of
execution displayed in utf8 mode.}

\item{text_wait_icon}{The icon used when a timed message is waiting for the end of
execution displayed in text only mode.}

\item{indent}{How many spaces to indent the message.}

\item{text_bold}{FALSE by default. If TRUE prints the message text in bold letters.}

\item{text_italic}{FALSE by default. If TRUE prints the message text in italic letters.}

\item{text_underline}{FALSE by default. If TRUE prints the message text underlined.}

\item{text_color}{The color of the actual message text.}

\item{time_color}{The color used for the time stamps.}
}
\value{
Return text without styling or total running time.

\code{\link[=set_up_custom_message]{set_up_custom_message()}}: Returns the global list of custom messages.
}
\description{
Printing styled messages for different occasions. There are notifications,
warnings, errors, function call headlines, progress and function closing
messages. Or just a neutral one.
}
\details{
The message types in which you can enter custom texts, are capable of using
different styling operators. These are:
\itemize{
\item Insert list of elements: [vector_name]
\item Adding conditional words, if list of elements has more than one element: [?word]
\item Adding conditional singular/plural, depending on list of element length: [?singular/plural]
\item Bold, italic and underline: [b]some text[/b], [i]some text[/i], [u]some text[/u]
\item Coloring parts of the message: [#FF00FF some text]
}
}
\examples{
# Example messages
print_message("NOTE", c("Just a quick note that you can also insert e.g.[? a / ]variable",
                        "name[?s] like this: [listing].",
                        "Depending on the number of variables you can also alter the text."),
             listing = c("age", "state", "NUTS3"))

print_message("WARNING", "Just a quick [#FF00FF colored warning]!")

print_message("ERROR", "Or a [b]bold[/b], [i]italic[/i] and [u]underlined[/u] error.")

print_message("NEUTRAL", c("You can also just output [u]plain text[/u] if you like and use",
                           "[#FFFF00 [b]all the different[/b] [i]formatting options.[/i]]"))

# Different headlines
print_headline("This is a headline")

print_headline("[#00FFFF This is a different headline] with some color",
               line_char = "-")

print_headline("[b]This is a very small[/b] and bold headline",
               line_char = ".",
               max_width = 60)

# Messages with time stamps
test_func <- function(){
    print_start_message()
    print_step("GREY", "Probably not so important")
    print_step("MAJOR", "This is a major step...")
    print_step("MINOR", "Sub step1")
    print_step("MINOR", "Sub step2")
    print_step("MINOR", "Sub step3")
    print_step("MAJOR", "[b]Finishing... [/b][#00FFFF with some color again!]")
    print_closing()
}

test_func()

# See what is going on in the message stack
message_stack <- get_message_stack()

# Print messages on the same line instead of below each other
in_place_steps <- function(){
    print_start_message()

    print_step("MAJOR", "Let's get started...")

    for (i in seq_len(10)){
        print_step("Minor", "This is in place step [i] of 10", i = i, in_place = TRUE)
        Sys.sleep(0.25)
    }

    print_step("MAJOR", "Loop has ended")

    print_closing()
}

in_place_steps()

# Set up a custom message
set_up_custom_message(type           = "HOTDOG",
                      color          = "#B27A01",
                      ansi_icon      = "\U0001f32d",
                      text_icon      = "IOI",
                      ansi_wait_icon = "\U00023f1",
                      text_wait_icon = "/",
                      indent         = 1)

hotdog_print <- function(){
    print_start_message()
    print_message("HOTDOG", c("This is the first hotdog message! Hurray!",
                              "And it is also multiline in this version."))
    print_step("HOTDOG", "Or use as single line message with time stamps.")
    Sys.sleep(0.5)
    print_step("HOTDOG", "Or use as single line message with time stamps.")
    Sys.sleep(0.5)
    print_step("HOTDOG", "Or use as single line message with time stamps.")
    Sys.sleep(0.5)
    print_closing()
}

hotdog_print()

# See new message in the message stack
hotdog_stack <- get_message_stack()

}
\seealso{
Also have a look at the small helpers: \code{\link[=get_message_stack]{get_message_stack()}}, \code{\link[=set_no_print]{set_no_print()}},
\code{\link[=set_no_color]{set_no_color()}}, \code{\link[=print_stack_as_messages]{print_stack_as_messages()}}, \code{\link[=convert_square_brackets]{convert_square_brackets()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conversion.R
\name{convert_variables}
\alias{convert_variables}
\alias{convert_numeric}
\alias{convert_factor}
\title{Convert Variables}
\usage{
convert_numeric(data_frame, variables)

convert_factor(data_frame, variables)
}
\arguments{
\item{data_frame}{A data frame containing variables to convert.}

\item{variables}{Variables from the data frame which should be converted.}
}
\value{
\code{\link[=convert_numeric]{convert_numeric()}} returns the same data frame with converted variables where possible.

\code{\link[=convert_factor]{convert_factor()}} returns the same data frame with converted variables.
}
\description{
\code{\link[=convert_numeric]{convert_numeric()}} converts all given variables to numeric if possible. If
a variable contains none numerical values (not including NAs), the variable
will not be converted.

\code{\link[=convert_factor]{convert_factor()}} converts all given variables to factor.
}
\examples{
# Convert variables in a data frame to numeric where possible
test_df <- data.frame(var_a = c(1, 2, 3, NA, 4, 5),
                      var_b = c(1, 2, "Hello", NA, 4, 5))

convert_df <- test_df |> convert_numeric(c("var_a", "var_b"))

# Convert variables in a data frame to factor
test_df <- data.frame(var_a = c(1, 2, 3, 4, 5),
                      var_b = c("e", "c", "a", "d", "b"))

convert_df <- test_df |> convert_factor("var_b")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/strings.R
\name{concat}
\alias{concat}
\title{Concatenate Multiple Variables With Padding}
\usage{
concat(
  data_frame,
  ...,
  separator = NULL,
  padding_char = NULL,
  padding_length = NULL,
  padding_right = FALSE
)
}
\arguments{
\item{data_frame}{A data frame which contains the the variables to concatenate.}

\item{...}{The names of the variables to concatenate.}

\item{separator}{Characters to put in between variables when binding them together.}

\item{padding_char}{A single character which will be used to fill up the empty places.}

\item{padding_length}{A numeric vector containing the individual padding length per variable.}

\item{padding_right}{FALSE by default. If TRUE insert padding characters on the right side
instead of the left side.}
}
\value{
Returns a character vector.
}
\description{
Concatenate multiple variables inside a data frame into a new variable. An
automatic or individual padding can be applied. The padding character can be
chosen freely.

The function can also be used to give a single variable a padding.
}
\examples{
# Example data frame
my_data <- dummy_data(100)

# Concatenate variables as provided
my_data[["id1"]] <- my_data |> concat(household_id, state, age)

# Concatenate variables with leading zeros. Each variable will
# receive an individual padding length according to their
# longest value.
my_data[["id2"]] <- my_data |> concat(household_id, state, age,
                                      padding_char = "0")

# Concatenate variables with individual character and lengths.
my_data[["id2"]] <- my_data |> concat(household_id, state, age,
                                      padding_char   = "_",
                                      padding_length = c(5, 3, 4))

# Padding a single variable in place
my_data[["state"]] <- my_data |> concat(state, padding_char = "0")

}
\seealso{
Other character manipulating functions: \code{\link[=sub_string]{sub_string()}}, \code{\link[=remove_blanks]{remove_blanks()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keep_drop.R
\name{keep_dropp}
\alias{keep_dropp}
\alias{keep}
\alias{dropp}
\title{Keep and Drop Variables Inside a Data Frame}
\usage{
keep(data_frame, ..., order_vars = FALSE)

dropp(data_frame, ...)
}
\arguments{
\item{data_frame}{A data frame which should be reduced to (keep) or by (drop) the specified variables.}

\item{...}{The variable names to keep/drop.}

\item{order_vars}{\code{\link[=keep]{keep()}}: At the end variables are ordered as specified in the command.}
}
\value{
Returns a reduced data table.
}
\description{
\code{\link[=keep]{keep()}} enables you to put in a vector of variable names which then are kept
inside the given data frame. All other variables are dropped.

\code{\link[=dropp]{dropp()}} enables you to put in a vector of variable names which then are dropped
from the given data frame. All other variables are kept.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Call function
new_dt1 <- my_data |> keep(year, age, sex)
new_dt2 <- my_data |> keep(weight, income, education, sex, order_vars = TRUE)
new_dt3 <- my_data |> dropp(year, age, sex)

# Also works with characters
new_dt4 <- my_data |> keep("year", "age", "sex")
new_dt5 <- my_data |> dropp("year", "age", "sex")

# Or variable names stored as a character vector
var_names <- c("age", "sex", "income", "weight")

new_dt6 <- my_data |> keep(var_names)
new_dt7 <- my_data |> dropp(var_names)

# You can also keep or drop a range of variables
new_dt8 <- my_data |> keep(year, state:income)
new_dt9 <- my_data |> dropp(year, state:income)

# You can also use the colon as a placeholder for any text
start1 <- my_data |> keep("s:") # Variable names start with "s"
start2 <- my_data |> dropp("s:")

end1 <- my_data |> keep(":id") # Variable names end with "id"
end2 <- my_data |> dropp(":id")

contain1 <- my_data |> keep(":on:") # Variable names which contain "on"
contain2 <- my_data |> dropp(":on:")

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/if_else.R
\name{if_else}
\alias{if_else}
\alias{if.}
\alias{else_if.}
\alias{else.}
\title{If - Else if - Else Statements}
\usage{
if.(data_frame, condition, ...)

else_if.(data_frame, condition, ...)

else.(data_frame, ...)
}
\arguments{
\item{data_frame}{A data frame on which to apply an if statement.}

\item{condition}{The condition on which a value should be passed to a variable.}

\item{...}{The assignment of what should happen when condition becomes TRUE.}
}
\value{
Returns a data frame with conditionally computed variables.
}
\description{
These functions make if statements more readable. Especially if an if block becomes bigger
it can be hard to read with multiple nested if_else statements. With these new functions
if blocks can be written like in other languages with a clear and simpler structure. In
addition not only for one variable can a new value be assigned, but for multiple.

\code{\link[=if.]{if.()}} always creates a new variable if the given variable name is not part of the given
data frame. If there already is a variable with the given name, the existing values will
be overwritten if the condition is TRUE.

If no new variable is provided, \code{\link[=if.]{if.()}} will select observations by the given condition
instead.

\code{\link[=else_if.]{else_if.()}} only acts if there already is a variable with the given name. Only NA values
will get new values if condition is TRUE. The existing values will not be overwritten.

\code{\link[=else.]{else.()}} only acts if there already is a variable with the given name. Sets every
remaining NA in given variable to the given value.
}
\details{
The loop you can use within these functions is based on the 'SAS' do-over-loop. This
type of loop iterates over every vector that appears in the loop in parallel.
Means that in the first iteration all the first vector elements are used, in the
second iteration all second elements of every vector, and so on. With this loop
you don't have the need to construct an outer loop, but can directly pass in
different vectors and let the function handle the loop inside.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Call function
new_df <- my_data |>
         if.(age < 18,             age_group = "under 18") |>
    else_if.(age >= 18 & age < 65, age_group = "18 to under 65") |>
    else.   (                      age_group = "65 and older")

# Or with multiple variables
new_df <- my_data |>
         if.(age < 18,             age_group = "under 18"      , age_num = 1L) |>
    else_if.(age >= 18 & age < 65, age_group = "18 to under 65", age_num = 2L) |>
    else.   (                      age_group = "65 and older",   age_num = 3L)

# NOTE: As in other languages the following if blocks won't produce the same result.
#       if.() will overwrite existing values, while else_if.() will not.
state_df <- my_data |>
         if.(state == 1, state_a = "State 1") |>
    else_if.(state < 11, state_a = "West") |>
    else.   (            state_a = "East")

state_df <- state_df |>
      if.(state == 1, state_b = "State 1") |>
      if.(state < 11, state_b = "West") |>
    else.(            state_b = "East")

# Add multiple variables based on single conditions
multi_df <- my_data |>
         if.(age < 18 & education == "low",                  var1 = 1,
                                                             var2 = TRUE,
                                                             var3 = "Category 1") |>
    else_if.(age >= 18 & education \%in\% c("middle", "high"), var1 = 2,
                                                             var2 = FALSE,
                                                             var3 = "Category 2") |>
    else.   (                                                var1 = 3,
                                                             var2 = FALSE,
                                                             var3 = "Category 3")

# Use if.() as a do-over-loop. In this kind of loop all vectors will be
# advanced one iteration at a time in parallel.
money    <- c("income", "expenses", "balance", "probability")
new_vars <- c("var1", "var2", "var3", "var4")
result   <- c(1, 2, 3, 4)

do_over_df <- my_data |>
      if.(money > 0, new_vars = result) |>
    else.(           new_vars = 0)

# It is also possible to select character expressions based on whether they
# start/end with or contain a certain text.
text_select_df <- my_data |>
    if.(income_class == "01.:",  start    = 1) |>
    if.(income_class == ":more", end      = 1) |>
    if.(education    == ":i:",   contains = 1)

# Select observations by condition instead of generating new variable
subset_df <- my_data |> if.(sex == 1)

# Explicitly delete observations
subset_df <- my_data |> if.(sex == 1, delete)

# Select all non NA observations by variable
subset_df <- my_data |> if.(sex)

# All these functions can be used in a do_if() situation and are aware of
# overarching conditions.
do_if_df <- my_data |>
    do_if(state < 11) |>
          if.(age < 18, new_var = 1) |>
        else.(          new_var = 2) |>
    else_do() |>
          if.(age < 18, new_var = 3) |>
        else.(          new_var = 4) |>
    end_do()

}
\seealso{
\code{\link[=ifelse]{ifelse()}} with multiple conditions and 'SAS' writing style: \code{\link[=ifelse_multi]{ifelse_multi()}}

The following functions can make use of the \code{\link[=do_if]{do_if()}} filter variables:

Conditions: \code{\link[=if.]{if.()}}, \code{\link[=else_if.]{else_if.()}}, \code{\link[=else.]{else.()}}

Filter Data Frame: \code{\link[=where.]{where.()}}

Create new Variables: \code{\link[=compute.]{compute.()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_master.R
\name{build_master}
\alias{build_master}
\title{Build a Master Script From Folder}
\usage{
build_master(
  dir,
  master_name = "Master",
  author = "",
  with_structure = TRUE,
  with_run_all = TRUE,
  with_run_folder = TRUE,
  with_monitor = FALSE
)
}
\arguments{
\item{dir}{The folder structure which contains the scripts to build upon.}

\item{master_name}{The file name which should be written.}

\item{author}{Authors name to be put in the header.}

\item{with_structure}{Whether the folder structure as tree should be written
to the master script.}

\item{with_run_all}{Whether a section, which let's the user run all scripts,
should be written to the master script.}

\item{with_run_folder}{Whether a section, which let's the user run all scripts from a
specific folder, should be written to the master script.}

\item{with_monitor}{FALSE by default. If TRUE, outputs two charts to visualize the
time consumption of the individual scripts.}
}
\value{
Returns the script as character vector and saves it as markdown file.
}
\description{
\code{\link[=build_master]{build_master()}} reads a given folder structure, which contains scripts, and builds
a master script as a markdown file.
}
\details{
The function works with folder structures that look like this:

root/

\if{html}{\out{<div class="sourceCode">}}\preformatted{ subfolder1/

     script1.R

     script2.R

     ....R

 subfolder2/

     script3.R

     script4.R

     ....R

 .../

     ....R
}\if{html}{\out{</div>}}
}
\examples{
# Example export file paths
# NOTE: These tempfiles are only for the examples. In reality you just call the
# main function and put in your desired path and name directly.
temp_file <- tempfile(fileext = ".rstheme")
file_name <- basename(tools::file_path_sans_ext(temp_file))

# Example master
build_master(dir         = dirname(temp_file),
             master_name = file_name)

# Manual cleanup for example
unlink(temp_file)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/if_else.R
\name{do_if}
\alias{do_if}
\alias{else_do}
\alias{end_do}
\alias{end_all_do}
\title{Lock In A Condition}
\usage{
do_if(data_frame, condition)

else_do(data_frame)

end_do(data_frame)

end_all_do(data_frame)
}
\arguments{
\item{data_frame}{A data frame on which to apply filters.}

\item{condition}{The condition lock in.}
}
\value{
Returns a data frame with a new filter variable.
}
\description{
Creates a filter variable based on the given condition. This variable can be
accessed by \code{\link[=if.]{if.()}}, \code{\link[=else_if.]{else_if.()}}, \code{\link[=else.]{else.()}}, \code{\link[=where.]{where.()}} and \code{\link[=compute.]{compute.()}}, enabling these
functions to work with an overarching condition. This function can also be
used to nest multiple overarching conditions.

\code{\link[=else_do]{else_do()}}: Checks for existing filter variables and reverses the condition
of the last filter variable.

\code{\link[=end_do]{end_do()}}: Drops the last filter variable.

\code{\link[=end_all_do]{end_all_do()}}: Drops all filter variables.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Create a simple do-if-block
do_if_df <- my_data |>
    do_if(state < 11) |>
          if.(age < 18, new_var = 1) |>
        else.(          new_var = 2) |>
    else_do() |>
          if.(age < 18, new_var = 3) |>
        else.(          new_var = 4) |>
    end_do()

# do_if() can also be nested
do_if_df <- my_data |>
    do_if(state < 11) |>
        do_if(sex == 1) |>
              if.(age < 18, new_var = 1) |>
            else.(          new_var = 2) |>
        else_do() |>
              if.(age < 18, new_var = 3) |>
            else.(          new_var = 4) |>
        end_do() |>
    else_do() |>
        do_if(sex == 1) |>
              if.(age < 18, new_var = 5) |>
            else.(          new_var = 6) |>
        else_do() |>
              if.(age < 18, new_var = 7) |>
            else.(          new_var = 8) |>
        end_do() |>
    end_do()

# NOTE: Close the do-if-blocks with end_do() to remove the temporary logical
#       filter variables.

# Probably a logical filter variable is exactly what you want. In this case
# just run do_if() without closing the block.
logic_filter_df <- my_data |> do_if(state < 11)

}
\seealso{
The following functions can make use of the \code{\link[=do_if]{do_if()}} filter variables:

Conditions: \code{\link[=if.]{if.()}}, \code{\link[=else_if.]{else_if.()}}, \code{\link[=else.]{else.()}}

Filter Data Frame: \code{\link[=where.]{where.()}}

Create new Variables: \code{\link[=compute.]{compute.()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content_report.R
\name{content_report}
\alias{content_report}
\title{Get Detailed Summary About A Data Frame}
\usage{
content_report(data_frame, output = "console", monitor = FALSE)
}
\arguments{
\item{data_frame}{The data frame to get the content information from.}

\item{output}{The following output formats are available: console (default) or text.}

\item{monitor}{FALSE by default. If TRUE, outputs two charts to visualize the
functions time consumption.}
}
\value{
Returns a list containing the global information as well as a data table containing
the per variable information.
}
\description{
Prints a summary of a data frames contents, including details such as variable names,
types, unique values, missings and min/max values. It also tells you the number of
observations and variables present in the data frame, memory usage and the number
of duplicate observations.
}
\details{
\code{\link[=content_report]{content_report()}} is based on the 'SAS' procedure Proc Contents, which provides
a summary of global information one one hand like number of observations and variables among
many others and on the other hand shows per variable information like type and length.

'R' doesn't store the same information in a data frame like 'SAS', but there are many
useful information to get a quick overview of a data frame. With this function you don't
need to look at each variable individually. You can simply run it over a data frame and
get values for: number of unique values, missing values (absolute and relative), min and
max value as well as the top value.
}
\examples{
# Example data frame
my_data <- dummy_data(100)

content_report(my_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinytest_reporter.R
\name{reporter}
\alias{reporter}
\alias{report_test_results}
\alias{test_package}
\alias{test_single_file}
\title{Print Styled Tinytest Results}
\usage{
report_test_results(tiny_results, utf8 = .qol_messages[["format"]][["utf8"]])

test_package(package_name, multithread = FALSE)

test_single_file(file_name)
}
\arguments{
\item{tiny_results}{The results produced by tinytest.}

\item{utf8}{Whether to display complex characters or just plain text.}

\item{package_name}{Name of the package to test.}

\item{multithread}{FALSE by default. Whether to run tests multithreaded.
NOTE: To make this work you have to manually run \code{\link[devtools:install]{devtools::install()}} first.}

\item{file_name}{Name of the file to test.}
}
\value{
Returns the results.
}
\description{
Styles the results of tinytest to get a better visual overview.

\code{\link[=test_package]{test_package()}}: Runs all set up unit tests with tinytest and outputs the results with the
custom reporter.

\code{\link[=test_single_file]{test_single_file()}}: Runs a single unit test file with tinytest.
}
\examples{
# Example results
result_file <- system.file("extdata", "qol_tinytest_results.fst", package = "qol")
results     <- load_file(dirname(result_file), basename(result_file))

# Display results
results |> report_test_results()

# Normally you would do this:
# tinytest::test_all("PATH TO PACKAGE") |> report_test_results()
# or
# tinytest::test_package("PACKAGE NAME", testdir = "inst/tinytest") |> report_test_results()

# To test the whole package with the custom reporter use:
# test_package("qol")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{style_options}
\alias{style_options}
\alias{set_style_options}
\alias{get_style_options}
\alias{reset_style_options}
\alias{close_file}
\alias{set_labels}
\alias{get_labels}
\alias{reset_qol_options}
\title{Set Global Styling Options For Excel Workbooks}
\usage{
set_style_options(..., save_file = NULL)

get_style_options(from_file = NULL)

reset_style_options()

close_file()

set_labels(...)

get_labels()

reset_qol_options()
}
\arguments{
\item{...}{\code{\link[=set_labels]{set_labels()}}: Put in the variable names and their respective labels.}

\item{save_file}{A full file path to an RDS file in which global style options
should be stored.}

\item{from_file}{A full file path to an RDS file in which global style options
are stored.}
}
\value{
\code{\link[=set_style_options]{set_style_options()}}: Returns modified global styling options.

\code{\link[=get_style_options]{get_style_options()}}: List of global styling options.

\code{\link[=reset_style_options]{reset_style_options()}}: Returns default global styling options.

\code{\link[=close_file]{close_file()}}: List of global styling options with file = NULL.

\code{\link[=set_labels]{set_labels()}}: List of variable and statistic labels.

\code{\link[=get_labels]{get_labels()}}: List of variable and statistic labels.

\code{\link[=reset_qol_options]{reset_qol_options()}}: Returns default global options.
}
\description{
Modify Styling options for Excel workbooks. Available parameters can be seen in
\code{\link[=excel_output_style]{excel_output_style()}} or \code{\link[=number_format_style]{number_format_style()}}.

\code{\link[=set_style_options]{set_style_options()}} sets the styling options for Excel workbooks globally.These
options are used by all tabulation and output functions, which are capable of
exporting styled outputs.

\code{\link[=get_style_options]{get_style_options()}} Prints out the currently set global styling options.

\code{\link[=reset_style_options]{reset_style_options()}} Resets global style options to the default parameters.
This includes all options set with \code{\link[=set_style_options]{set_style_options()}}, \code{\link[=set_labels]{set_labels()}},
\code{\link[=set_titles]{set_titles()}} and \code{\link[=set_footnotes]{set_footnotes()}}.

\code{\link[=close_file]{close_file()}} is a simple, more readable wrapper for setting file parameter to NULL.

\code{\link[=set_labels]{set_labels()}}: Can set variable and statistic labels globally so that they don't
have to be provided in every output function separately.

\code{\link[=get_labels]{get_labels()}}: Get the globally stored variable and statistic labels as a list
and print the contents to the console.

\code{\link[=reset_qol_options]{reset_qol_options()}} Resets global options to the default parameters.
}
\examples{
# This function can process any parameter from excel_output_style() or
# number_format_style() and sets the option globally.
set_style_options(save_path    = "C:/My Projects/",
                  sum_decimals = 8)

get_style_options()

# Reset all the style options including variable and statistic labels as well
# as titles and footnotes.
reset_style_options()

# Reset the file parameter
close_file()

# Set variable and statistic labels globally, retrieve and reset them
set_labels(age_gr = "Group of ages",
           status = "Current status",
           pct    = "\%")

# To reset labels do this
set_labels(NULL)

get_labels()

# Reset all globally set options
reset_qol_options()

}
\seealso{
Functions that use global styling options: \code{\link[=any_table]{any_table()}}, \code{\link[=frequencies]{frequencies()}},
\code{\link[=crosstabs]{crosstabs()}}, \code{\link[=export_with_style]{export_with_style()}}.

Functions that use global variable and statistic labels: \code{\link[=any_table]{any_table()}}, \code{\link[=frequencies]{frequencies()}},
\code{\link[=crosstabs]{crosstabs()}}.

Functions that also use global variable labels: \code{\link[=export_with_style]{export_with_style()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/excel_helpers.R
\name{modify_output_style}
\alias{modify_output_style}
\title{Modify Style for 'Excel' Table Outputs}
\usage{
modify_output_style(style_to_modify, ...)
}
\arguments{
\item{style_to_modify}{A pre created style where only certain elements should be
modified while the rest is kept as is.}

\item{...}{Pass in names and corresponding new values for existing style elements.}
}
\value{
Returns a modified list of named style options.
}
\description{
Modify a previously created style with \code{\link[=excel_output_style]{excel_output_style()}}.
}
\details{
\code{\link[=modify_output_style]{modify_output_style()}} is based on the Output Delivery System (ODS) in 'SAS',
which provides efficient and readable ways to set up different table styles.

With the output style you have full control over the table design. There is no
need to think about calculating the right place to input a background color or a
border of a certain type and how to do this in a loop for multiple cells. Just
input colors, borders, font styles, etc. for the different table parts and everything
else is handled by the functions capable of using styles.

The concept basically is: design over complex calculations.
}
\examples{
# For default values
excel_style <- excel_output_style()

# Set specific options, the rest will be kept as is
excel_style <- excel_style |> modify_output_style(sheet_name      = "Sheet",
                                                  title_font_bold = FALSE)

# For cells with no background color pass an empty string
excel_style <- excel_style |> modify_output_style(table_back_color = "")

}
\seealso{
Creating a custom table style: \code{\link[=excel_output_style]{excel_output_style()}},
\code{\link[=number_format_style]{number_format_style()}}, \code{\link[=modify_number_formats]{modify_number_formats()}}.

Global style options: \code{\link[=set_style_options]{set_style_options()}}, \code{\link[=set_labels]{set_labels()}}.

Functions that can handle styles: \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}}, \code{\link[=any_table]{any_table()}},
\code{\link[=export_with_style]{export_with_style()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renaming.R
\name{add_extension}
\alias{add_extension}
\title{Add Extensions to Variable Names}
\usage{
add_extension(data_frame, from, extensions, reuse = "none")
}
\arguments{
\item{data_frame}{The data frame in which variables should gain extensions to their name.}

\item{from}{The position of the variable inside the data frame at which to start the renaming.}

\item{extensions}{The extensions to add.}

\item{reuse}{"none" by default, meaning only the provided extensions will be set. E.g. if
there are two extensions provided, two variables will be renamed. If "last", the last provided
extension will be used for every following variable until the end of the data frame. If "repeat",
the provided extensions will be repeated from the first one for every following variable until
the end of the data frame.}
}
\value{
Returns a data frame with extended variable names.
}
\description{
Renames variables in a data frame by adding the desired extensions to the original names.
This can be useful if you want to use pre summarised data with \code{\link[=any_table]{any_table()}}, which needs
the value variables to have the statistic extensions.
}
\examples{
# Example data frame
my_data <- dummy_data(10)

# Add extensions to variable names
new_names1 <- my_data |> add_extension(5, c("sum", "pct"))
new_names2 <- my_data |> add_extension(5, c("sum", "pct"), reuse = "last")
new_names3 <- my_data |> add_extension(5, c("sum", "pct"), reuse = "alternate")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variable_selection.R
\name{inverse}
\alias{inverse}
\title{Get Variable Names Which Are Not Part Of The Given Vector}
\usage{
inverse(data_frame, var_names)
}
\arguments{
\item{data_frame}{The data frame from which to take the variable names.}

\item{var_names}{A character vector of variable names.}
}
\value{
Returns the inverse vector of variable names compared to the given vector.
}
\description{
If you have stored variable names inside a character vector, this function gives you
the inverse variable name vector.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Get variable names
var_names <- c("year", "age", "sex")
other_names <- my_data |> inverse(var_names)

# Can also be used to just get all variable names
all_names <- my_data |> inverse()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conversion.R
\name{convert_arguments}
\alias{convert_arguments}
\alias{args_to_char}
\alias{dots_to_char}
\alias{get_origin_as_char}
\title{Convert Function Arguments to Character Vector}
\usage{
args_to_char(argument)

dots_to_char(...)

get_origin_as_char(original, substituted)
}
\arguments{
\item{argument}{Function argument to convert.}

\item{...}{Used for variable names listed in ... without the need to put them in c() or list().}

\item{original}{The data frame which contains the columns to be checked.}

\item{substituted}{The grouping variables which potentially form unique combinations.}
}
\value{
Returns a character vector.
}
\description{
\code{\link[=args_to_char]{args_to_char()}}: Converts any argument passed as a single character or symbol as well
as character vectors or vector of symbols back as character vector.

\code{\link[=dots_to_char]{dots_to_char()}}: When you define a function and want the user to be able to pass variable
names without the need to have them stored in a vector c() or list() beforehand and without
putting the names into quotation marks, you can convert this variable list passed as ...
into a character vector.

Note: If the user passes a list of characters it is returned as given.

\code{\link[=get_origin_as_char]{get_origin_as_char()}} is a wrapper that allows to retrieve the original contents
of the provided variable, whether called directly or nested in multiple function calls,
as a character vector.
}
\examples{
# Example function with function parameter
print_vnames <- function(parameter){
    var_names <- args_to_char(substitute(parameter))
    print(var_names)
}

print_vnames(age)
print_vnames("age")
print_vnames(c(age, sex, income, weight))
print_vnames(c("age", "sex", "income", "weight"))

# You can also pass in a character vector, if you have stored variable names elsewhere
var_names <- c("age", "sex", "income", "weight")
print_vnames(var_names)

# If you plan to use the function within other functions, better use get_origin_as_char()
print_vnames <- function(parameter){
    var_names <- get_origin_as_char(parameter, substitute(parameter))
    print(var_names)
}

another_function <- function(parameter){
    print_vnames(parameter)
}

another_function("age")
another_function(c("age", "sex", "income", "weight"))

# Example function with ellipsis
print_vnames <- function(...){
    var_names <- dots_to_char(...)
    print(var_names)
}

print_vnames(age)
print_vnames("age")
print_vnames(age, sex, income, weight)
print_vnames("age", "sex", "income", "weight")

# You can also pass in a character vector, if you have stored variable names elsewhere
var_names <- c("age", "sex", "income", "weight")
print_vnames(var_names)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/error_handling.R
\name{error_handling}
\alias{error_handling}
\alias{resolve_intersection}
\alias{part_of_df}
\alias{remove_doubled_values}
\alias{check_weight}
\title{Error Handling}
\usage{
resolve_intersection(base, vector_to_check, check_only = FALSE)

part_of_df(data_frame, var_names, check_only = FALSE)

remove_doubled_values(var_names)

check_weight(data_frame, var_names)
}
\arguments{
\item{base}{The base vector from which to remove any intersecting values.}

\item{vector_to_check}{The vector for which intersections should be checked.}

\item{check_only}{Returns a list of invalid entries instead of a vector. Additionally
it doesn't throw a warning.}

\item{data_frame}{A data frame in which to look up variable names.}

\item{var_names}{A character vector of variable names.}
}
\value{
Returns a vector or list.
}
\description{
\code{\link[=resolve_intersection]{resolve_intersection()}}: Compares if two vectors have intersecting values. If TRUE,
removes the intersection values from the base vector

\code{\link[=part_of_df]{part_of_df()}}: Check if variable names are part of a data frame. If not, remove them
from the given vector.

\code{\link[=remove_doubled_values]{remove_doubled_values()}}: Remove values from a vector that appear more than once.

\code{\link[=check_weight]{check_weight()}}: Check if a weight variable was provided. If TRUE, check whether it
can be used else add a temporary weight variable.
}
\examples{
# Resolve intersection between two vectors
vec1 <- c("a", "b", "c", "d")
vec2 <- c("e", "f", "a", "g")

vec1 <- resolve_intersection(vec1, vec2)

# Check if variables are part of a data frame
my_data   <- dummy_data(100)
var_names <- c("year", "state", "age", "test")

var_names <- my_data |> part_of_df(var_names)

# Remove doubled values
var_names <- c("year", "state", "state", "age")

var_names <- remove_doubled_values(var_names)

# Check the provided weight variable
var_names <- my_data |> check_weight("weight")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loading.R
\name{stack_data}
\alias{stack_data}
\title{Stack Multiple Data Frames}
\usage{
stack_data(..., id = FALSE, compress = NULL, guessing_rows = 100)
}
\arguments{
\item{...}{Put in multiple data frames to stack them in the provided order.}

\item{id}{Adds an ID column to indicate the different data frames.}

\item{compress}{No compression by default. If compression receives any value, \code{\link[=stack_data]{stack_data()}}
will convert character variables to numeric or integer where possible. If set to "factor"
all non numeric character variables will be converted to factors.}

\item{guessing_rows}{100 by default. \code{\link[=stack_data]{stack_data()}} takes a sample of rows to determine of
which type variables are.}
}
\value{
Returns a stacked data frame.
}
\description{
Stacks multiple data frames and matches column names.
}
\examples{
# Example data frames
my_data1 <- dummy_data(100)
my_data2 <- dummy_data(100)
my_data3 <- dummy_data(100)
my_data4 <- dummy_data(100)
my_data5 <- dummy_data(100)

# Stack data frames
stacked_df <- stack_data(my_data1,
                         my_data2,
                         my_data3,
                         my_data4,
                         my_data5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renaming.R
\name{rename_multi}
\alias{rename_multi}
\title{Rename One Or More Variables}
\usage{
rename_multi(data_frame, ...)
}
\arguments{
\item{data_frame}{The data frame which contains the variable names to be renamed.}

\item{...}{Pass in variables to be renamed in the form: "old_var" = "new_var".}
}
\value{
Returns a data_frame with renamed variables.
}
\description{
Can rename one or more existing variable names into the corresponding new variable
names in one go.
}
\examples{
# Example data frame
my_data <- dummy_data(10)

# Rename multiple variables at once
new_names_df <- my_data |> rename_multi(sex   = var1,
                                        age   = var2,
                                        state = var3)

# Also works with variable names in quotation marks
new_names_df <- my_data |> rename_multi("sex"   = "var1",
                                        "age"   = "var2",
                                        "state" = "var3")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recode.R
\name{recode}
\alias{recode}
\alias{recode.}
\alias{recode_multi}
\title{Recode New Variables With Formats}
\usage{
recode.(data_frame, ...)

recode_multi(data_frame, ..., convert = FALSE)
}
\arguments{
\item{data_frame}{A data frame which contains the the original variables to recode.}

\item{...}{\code{\link[=recode.]{recode.()}} Pass in the original variable name that should be recoded
along with the corresponding format container in the form: variable = format.

In \code{\link[=recode_multi]{recode_multi()}} multiple variables can be recoded in one go and multilabels
can be applied. This overwrites the original variables and duplicates rows if
multilabels are applied. In occasions were you want to use format containers to
afterwards perform operations with other packages, you can make use of this
principle with this function.}

\item{convert}{FALSE by default. If TRUE converts recoded variables to numeric
or character depending on the input format instead of leaving them as factors.}
}
\value{
\code{\link[=recode.]{recode.()}}: Returns a vector with recoded values.

\code{\link[=recode_multi]{recode_multi()}}: Returns a data frame with the newly recoded variable.
}
\description{
Instead of writing multiple if-clauses to recode values into a new variable,
you can use formats to recode a variable into a new one.
}
\details{
\code{\link[=recode.]{recode.()}} is based on the 'SAS' function put(), which provides an efficient
and readable way, to generate new variables with the help of formats.

When creating a format you can basically write code like you think: This new
category consists of these original values. And after that you just apply these
new categories to the original values to create a new variable. No need for multiple
if_else statements.
}
\examples{
# Example formats
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

# Example data frame
my_data <- dummy_data(1000)

# Call function
my_data[["age_group1"]] <- my_data |> recode.(age = age.)

# Formats can also be passed as characters
my_data[["age_group2"]] <- my_data |> recode.(age = "age.")

# Multilabel recode
sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

income. <- interval_format(
    "Total"              =    0:100000,
    "below 500"          =    0:500,
    "500 to under 1000"  =  500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

# recode_multi() can not only apply multiple recodings, but it can also
# apply multilabels.
# NOTE: Recoding will always be in place. When applying multilabels the
#       result data frame will have more observations than before.
multi_data <- my_data |> recode_multi(sex = sex., income = income.)

}
\seealso{
Creating formats: \code{\link[=discrete_format]{discrete_format()}} and \code{\link[=interval_format]{interval_format()}}.

Functions that also make use of formats: \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}},
\code{\link[=any_table]{any_table()}}, \code{\link[=summarise_plus]{summarise_plus()}}, \code{\link[=transpose_plus]{transpose_plus()}}, \code{\link[=sort_plus]{sort_plus()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarise_plus.R
\name{summarise_plus}
\alias{summarise_plus}
\title{Fast And Powerful Yet Simple To Use Summarise}
\usage{
summarise_plus(
  data_frame,
  class = NULL,
  values,
  statistics = c("sum", "freq"),
  formats = list(),
  types = "",
  weight = NULL,
  nesting = "deepest",
  merge_back = FALSE,
  na.rm = .qol_options[["na.rm"]],
  print_miss = .qol_options[["print_miss"]],
  monitor = .qol_options[["monitor"]],
  notes = TRUE
)
}
\arguments{
\item{data_frame}{A data frame to summarise.}

\item{class}{A vector containing all grouping variables.}

\item{values}{A vector containing all variables that should be summarised.}

\item{statistics}{Available functions:
\itemize{
\item "sum"       -> Weighted and unweighted sum
\item "sum_wgt"   -> Sum of all weights
\item "freq"      -> Unweighted frequency
\item "freq_g0"   -> Unweighted frequency of all values greater than zero
\item "pct_group" -> Weighted and unweighted percentages within the respective group
\item "pct_total" -> Weighted and unweighted percentages compared to the grand total
\item "mean"      -> Weighted and unweighted mean
\item "median"    -> Weighted and unweighted median
\item "mode"      -> Weighted and unweighted mode
\item "min"       -> Minimum
\item "max"       -> Maximum
\item "sd"        -> Weighted and unweighted standard deviation
\item "variance"  -> Weighted and unweighted standard variance
\item "first"     -> First value
\item "last"      -> Last value
\item "pn"        -> Weighted and unweighted percentiles (any p1, p2, p3, ... possible)
Uses the SAS default way of calculating percentiles, which is Type 2 described in \code{\link[=quantile]{quantile()}}.
\item "missing"   -> Missings generated by the value variables
}}

\item{formats}{A list in which is specified which formats should be applied to which
class variables.}

\item{types}{A character vector specifying the different combinations of group
variables which should be computed when using nesting = "all". If left empty all
possible combinations will be computed.}

\item{weight}{Put in a weight variable to compute weighted results.}

\item{nesting}{The predefined value is "deepest" meaning that only the fully
nested version of all class variables will be computed. If set to "all", all
possible combinations will be computed in one data table. The option "single"
only outputs the ungrouped summary of all class variables in one data table.}

\item{merge_back}{Newly summarised variables can be merged back to the original
data frame if TRUE. Only works if nested = "deepest" and no formats are defined.}

\item{na.rm}{FALSE by default. If TRUE removes all NA values from the class variables.}

\item{print_miss}{FALSE by default. If TRUE outputs all possible categories of the
grouping variables based on the provided formats, even if there are no observations
for a combination.}

\item{monitor}{FALSE by default. If TRUE, outputs two charts to visualize the
functions time consumption.}

\item{notes}{TRUE by default. Prints notifications about NA values produced by
class variables during summarise.}
}
\value{
Returns a summarised \code{data.table}.

If \code{nesting} is set to \code{"all"} or \code{"single"}, the returned table automatically includes three hierarchical metadata variables:
\itemize{
\item \strong{\code{TYPE}} (character): The active grouping combination for the row (e.g., \code{"year+sex"}), or \code{"total"} for the grand total.
\item \strong{\code{TYPE_NR}} (integer): A unique sequential identifier assigned to each distinct grouping combination (\code{TYPE}).
\item \strong{\code{DEPTH}} (integer): The number of active grouping variables in the row's combination (e.g., \code{0} for \code{"total"}, \code{1} for single variables, etc.).
}
}
\description{
\code{\link[=summarise_plus]{summarise_plus()}} creates a new aggregated data table with the desired grouping.
It can output only the deepest nested combination of the grouping variables (default)
or you can also output every possible combination of the grouping variables at once,
with just one small change. Besides the normal summary functions like sum, mean
or median, you can also calculate their respective weighted version by just
setting a weight variable.
}
\details{
\code{\link[=summarise_plus]{summarise_plus()}} is based on the 'SAS' procedure Proc Summary, which provides
efficient and readable ways to perform complex aggregations.

Normally you would compute new categorical variables beforehand - probably even in
different forms, if you wanted to have different categorizations - and bloat up
the data set. After all this recoding footwork you could finally use multiple
summaries to compute all the stats you need to then put them back together. With this
function this is no more necessary.

In \code{\link[=summarise_plus]{summarise_plus()}} you put in the original data frame and let the recoding happen
via format containers. This is very efficient, since new variables and categories
are only created just before the summarise happens.

Additionally you can specify whether you only want to produce the all nested version
of all group variables or whether you want to produce every possible combination in
one go. All with a single option.

The function is optimized to always take the fastest route, depending on the options
specified.
}
\examples{
# NOTE: The used threads are set here to prevent a NOTE from the CRAN checks,
#       otherwise this is not necessary.
old_threads <- data.table::getDTthreads()
data.table::setDTthreads(1)

# Example formats
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

income. <- interval_format(
    "Total"              =    0:100000,
    "below 500"          =    0:500,
    "500 to under 1000"  =  500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

# Example data frame
my_data <- dummy_data(1000)

# Summarise the data  with all class variables nested
all_nested <- my_data |>
    summarise_plus(class      = c(year, sex, age),
                   values     = income,
                   statistics = c("sum", "pct_group", "pct_total", "sum_wgt", "freq"),
                   formats    = list(sex = sex., age = age.),
                   weight     = weight,
                   nesting    = "deepest",
                   na.rm      = TRUE)

# Output all possible combinations of the provided class variables. Meaning:
# Total, combinations of 2, combinations of 3, ...
all_possible <- my_data |>
    summarise_plus(class      = c(year, sex, age, income),
                   values     = c(probability),
                   statistics = c("sum", "p1", "p99", "min", "max", "freq", "freq_g0"),
                   formats    = list(sex    = sex.,
                                     age    = age.,
                                     income = income.),
                   weight     = weight,
                   nesting    = "all",
                   na.rm      = TRUE)

# Formats can also be passed as characters
single <- my_data |>
    summarise_plus(class      = c(year, age, sex),
                   values     = weight,
                   statistics = c("sum", "mean"),
                   formats    = list(sex = "sex.", age = "age."),
                   nesting    = "single")

# Summarise and merge results back to the original data frame
merge_back <- my_data |>
    summarise_plus(class      = c(year, age, sex),
                   values     = weight,
                   statistics = c("sum", "mean"),
                   nesting    = "deepest",
                   merge_back = TRUE)

# Select only certain variable combinations of all the possible combinations
certain_types <- my_data |>
    summarise_plus(class      = c(year, sex, age),
                   values     = c(probability),
                   statistics = c("sum", "mean", "freq"),
                   formats    = list(sex = sex.,
                                     age = age.),
                   types      = c("year", "year + age", "age + sex"),
                   weight     = weight,
                   nesting    = "all",
                   na.rm      = TRUE)

# You can also select repeating combinations faster like this
combi_types <- my_data |>
    summarise_plus(class      = c(year, sex, age, education),
                   values     = weight,
                   statistics = "sum",
                   formats    = list(sex = sex.,
                                     age = age.),
                   types      = "year + sex + (age, age + education, education)",
                   nesting    = "all",
                   na.rm      = TRUE)

# Select specific statistics for specific variables
specific_stats <- my_data |>
    summarise_plus(class      = c(year, sex),
                   statistics = list("sum"       = c(weight, income),
                                     "mean"      = expenses,
                                     "pct_group" = balance),
                   formats    = list(sex = sex.),
                   na.rm      = TRUE)

# Works without grouping
no_group <- my_data |> summarise_plus(values = income)

# And also without values
no_values  <- my_data |> summarise_plus(class = c(year, sex, age))
no_nothing <- my_data |> summarise_plus()

# NOTE: The used threads are set here to prevent a NOTE from the CRAN checks,
#       otherwise this is not necessary.
data.table::setDTthreads(old_threads)

}
\seealso{
Creating formats: \code{\link[=discrete_format]{discrete_format()}} and \code{\link[=interval_format]{interval_format()}}.

Functions that also make use of formats: \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}},
\code{\link[=any_table]{any_table()}}, \code{\link[=recode.]{recode.()}}, \code{\link[=recode_multi]{recode_multi()}}, \code{\link[=transpose_plus]{transpose_plus()}}, \code{\link[=sort_plus]{sort_plus()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variable_selection.R
\name{vars_between}
\alias{vars_between}
\title{Get All Variable Names Between Two Variables}
\usage{
vars_between(data_frame, from, to)
}
\arguments{
\item{data_frame}{The data frame from which to take the variable names.}

\item{from}{Starting variable of variable range.}

\item{to}{Ending variable of variable range.}
}
\value{
Returns a character vector of variable names.
}
\description{
Get all the variable names inside a data frame between two variables (including the provided ones)
as a character vector
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Get variable names
var_names <- my_data |> vars_between(state, income)

# Get variable names in reverse order
vars_reverse <- my_data |> vars_between(income, state)

# If you only provide "from" or "to" you get all variable names from a point to
# the end or from the beginning to a given point.
vars_from <- my_data |> vars_between(state)
vars_to   <- my_data |> vars_between(to = state)

# Or just get all variable names
vars_all <- my_data |> vars_between()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multi_join.R
\name{multi_join}
\alias{multi_join}
\title{Join Multiple Data Frames In One Go}
\usage{
multi_join(
  data_frames,
  on,
  how = "left",
  keep_indicators = FALSE,
  monitor = .qol_options[["monitor"]]
)
}
\arguments{
\item{data_frames}{A list of data frames to join together. The second and all
following data frames will be joined on the first one.}

\item{on}{The key variables on which the data frames should be joined. If a
character vector is provided, the function assumes all the variables are in every
data frame. To join on different  variable names a list of character vectors has
to be provided.}

\item{how}{A character vector containing the join method names. Available methods are:
left, right, inner, full, outer, left_inner and right_inner.}

\item{keep_indicators}{FALSE by default. If TRUE, a variable for each data frame
is created, which indicates whether a data frame provides values.}

\item{monitor}{FALSE by default. If TRUE, outputs two charts to visualize the
functions time consumption.}
}
\value{
Returns a single data frame with joined variables from all given data frames.
}
\description{
Join two or more data frames together in one operation. \code{\link[=multi_join]{multi_join()}} can handle
multiple different join methods and can join on differently named variables.
}
\details{
\code{\link[=multi_join]{multi_join()}} is based on the 'SAS' Data-Step function Merge. Merge is capable of
joining multiple data sets together at once, with a very basic syntax.

Provide the dataset names, the variables, on which they should be joined and after
a full join is complete, the user can decide which parts of the joins should remain
in the final dataset.

\code{\link[=multi_join]{multi_join()}} tries to keep the simplicity, while giving the user the power, to
do more joins at the same time. Additionally to what Merge can do, this function
also makes use of the Proc SQL possibility to join datasets on different variable
names.
}
\examples{
# Example data frames
df1 <- data.frame(key = c(1, 1, 1, 2, 2, 2),
                  a   = c("a", "a", "a", "a", "a", "a"))

df2 <- data.frame(key = c(2, 3),
                  b   = c("b", "b"))

# See all different joins in action
join_methods <- c("left", "right", "inner", "full", "outer", "left_inner", "right_inner")
joined_data  <- list()

for (method in seq_along(join_methods)){
    joined_data[[method]] <- multi_join(list(df1, df2),
                                        on  = "key",
                                        how = join_methods[[method]])
}

# Left join on more than one key
df1b <- data.frame(key1 = c(1, 1, 1, 2, 2, 2),
                   key2 = c("a", "a", "a", "a", "a", "a"),
                   a    = c("a", "a", "a", "a", "a", "a"))

df2b <- data.frame(key1 = c(2, 3),
                   key2 = c("a", "a"),
                   b    = c("b", "b"))

left_joined <- multi_join(list(df1b, df2b), on = c("key1", "key2"))

# Join more than two data frames
df3 <- data.frame(key = c(1, 2),
                  c   = c("c", "c"))

multiple_joined <- multi_join(list(df1, df2, df3), on = "key")

# You can also use different methods for each join
multiple_joined2 <- multi_join(list(df1, df3, df2),
                               on  = "key",
                               how = c("left", "right"))

# Joining on different variable names
df1c <- data.frame(key1 = c(1, 1, 1, 2, 2, 2),
                   key2 = c("a", "a", "a", "a", "a", "a"),
                   a    = c("a", "a", "a", "a", "a", "a"))

df2c <- data.frame(var1 = c(2, 3),
                   var2 = c("a", "a"),
                   b    = c("b", "b"))

df3c <- data.frame(any  = c(1, 2),
                   name = c("a", "a"),
                   c    = c("c", "c"))

multiple_joined3 <- multi_join(list(df1c, df2c, df3c),
                               on = list(df1c = c("key1", "key2"),
                                         df2c = c("var1", "var2"),
                                         df3c = c("any", "name")))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{qol_news}
\alias{qol_news}
\title{Go To GitHub NEWS Page}
\usage{
qol_news()
}
\value{
URL.
}
\description{
Opens browser and goes to the Github NEWS page
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/small_helpers.R
\name{add_variable_range}
\alias{add_variable_range}
\title{Add Empty Variables In A Given Range}
\usage{
add_variable_range(data_frame, var_range)
}
\arguments{
\item{data_frame}{A data frame to add variables to.}

\item{var_range}{A range of variables to add, provided in the form: var_name1:var_name10.}
}
\value{
Returns a data frame with added variables.
}
\description{
Add empty variables to a data frame in the provided range. Basically does in a
data frame, what paste0("age", 1:10) does for a vector.
}
\examples{
# Example data frames
my_data <- dummy_data(100)

# Add variable range
my_data <- my_data |> add_variable_range(status1:status12)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/strings.R
\name{sub_string}
\alias{sub_string}
\title{Retrieve A Substring From A Character}
\usage{
sub_string(data_frame, variable, from = NULL, to = NULL, case_sensitive = TRUE)
}
\arguments{
\item{data_frame}{A data frame which contains the variables to concatenate.}

\item{variable}{A character variable to extract parts from.}

\item{from}{The names of the variables to concatenate.}

\item{to}{A single character which will be used to fill up the empty places.}

\item{case_sensitive}{TRUE by default. When a character expression is passed as
from or to it makes a difference whether a letter is written in upper or lower case.
Pass FALSE to handle upper and lower case equally.}
}
\value{
Returns parts of a character vector.
}
\description{
\code{\link[=sub_string]{sub_string()}} can extract parts of a character from the left side, right side
or from the middle. It is also able to start or end at specific letter sequences
instead of positions.
}
\examples{
# Example data frame
my_data <- dummy_data(100)

# Extract text from the left
my_data[["left"]] <- my_data |> sub_string(education, to = 2)

# Extract text from the right
my_data[["right"]] <- my_data |> sub_string(education, from = 2)

# Extract text from the middle
my_data[["middle"]] <- my_data |> sub_string(education, from = 2, to = 3)

# Find text and extract from the left
my_data[["left2"]] <- my_data |> sub_string(education, to = "l")

# Find text and extract from the right
my_data[["right2"]] <- my_data |> sub_string(education, from = "l")

# Find text and extract from the middle
my_data[["middle2"]] <- my_data |> sub_string(education, from = "i", to = "l")

}
\seealso{
Other character manipulating functions: \code{\link[=concat]{concat()}}, \code{\link[=remove_blanks]{remove_blanks()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/strings.R
\name{remove_blanks}
\alias{remove_blanks}
\title{Remove Blanks}
\usage{
remove_blanks(data_frame, variable, which = "all")
}
\arguments{
\item{data_frame}{A data frame which contains the character variables from which blanks
should be removed.}

\item{variable}{Variable name of the one from which to remove blanks.}

\item{which}{"all" by default. Can be "leading", "trailing", "trim", "normalize" or "all".
Determines which blanks should be removed}
}
\value{
Returns a character vector with removed blanks.
}
\description{
Removes leading and trailing blanks or both. Can also remove all blanks from a
character or normalize multiple blanks to single ones.
}
\examples{
# Example data frame
my_data <- dummy_data(100)
my_data[["blanks"]] <- " This  is  a  test "

# Remove blanks
my_data[["leading"]]   <- my_data |> remove_blanks(blanks, which = "leading")
my_data[["trailing"]]  <- my_data |> remove_blanks(blanks, which = "trailing")
my_data[["trim"]]      <- my_data |> remove_blanks(blanks, which = "trim")
my_data[["all"]]       <- my_data |> remove_blanks(blanks, which = "all")
my_data[["normalize"]] <- my_data |> remove_blanks(blanks, which = "normalize")

}
\seealso{
Other character manipulating functions: \code{\link[=concat]{concat()}}, \code{\link[=sub_string]{sub_string()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loading.R
\name{save_load}
\alias{save_load}
\alias{save_file}
\alias{save_file_multi}
\alias{load_file}
\alias{load_file_multi}
\title{Save And Load FST And RDS Files}
\usage{
save_file(
  data_frame,
  path,
  file,
  keep = NULL,
  where = NULL,
  compress = 100,
  protect = TRUE,
  ...
)

save_file_multi(
  data_frame_list,
  file_list,
  keep_list = NULL,
  compress = 100,
  protect = TRUE
)

load_file(path, file, keep = NULL, where = NULL, ...)

load_file_multi(file_list, keep_list = NULL, stack_files = TRUE)
}
\arguments{
\item{data_frame}{The data fame to be saved.}

\item{path}{The file path to save to/load from.}

\item{file}{The file name including the file extension.}

\item{keep}{The variables to keep in the file.}

\item{where}{A condition on which to subset the observations.}

\item{compress}{The amount of compression to use from 0 to 100. The lower the value,
the larger the file size and faster the saving speed. Uses maximum compression by
default.}

\item{protect}{TRUE by default. Throws an error if a file already exists. If FALSE,
overwrites existing files.}

\item{...}{Used internally to suppress messages, when using the multi save/load version.}

\item{data_frame_list}{A list of data frames.}

\item{file_list}{A character vector containing full file paths.}

\item{keep_list}{Can be a single variable name, a vector or a list of vectors containing
the variables to keep per file. If there are fewer list entries than files to load, the
last list element will be used repeatedly.}

\item{stack_files}{TRUE by default. Stacks data frames after loading them. If FALSE,
returns all data frames in a list.}
}
\value{
Returns a data frame. \code{\link[=load_file_multi]{load_file_multi()}} can also return a list of data frame.
}
\description{
\code{\link[=save_file]{save_file()}}: Saves fst and rds files. Offers variable selection and observation
subsetting. By default the function has a write protection, which has to be
explicitly turned off to be able to overwrite files.

\code{\link[=save_file_multi]{save_file_multi()}}: Saves multiple files.

\code{\link[=load_file]{load_file()}}: Loads fst and rds files. Provided variables to keep are read in case
insensitive and are returned in provided order. Additionally a subset can be defined
directly.

\code{\link[=load_file_multi]{load_file_multi()}}: Loads multiple files and stores them into a list or directly
stacks the files into one data frame.
}
\examples{
# Example data frame
my_data <- dummy_data(100)

# Save files
# NOTE: Normally you would pass in the path and file as character. For the
#       examples this is handled differently to provide runnable examples.
my_data |> save_file(path = tempdir(),
                     file = "testfile.fst")
my_data |> save_file(path = tempdir(),
                     file = "testfile.rds")

# Save file and only keep specific variables
# NOTE: Since the temporary file already exists now if you run the above code,
#       all the following save operations would throw errors because by default
#       the function write protects existing files. So for the following examples
#       the write protection is turned off.
my_data |> save_file(path    = tempdir(),
                     file    = "testfile.fst",
                     keep    = c(sex, age, state),
                     protect = FALSE)

# Save file and subset observations
my_data |> save_file(path    = tempdir(),
                     file    = "testfile.fst",
                     where   = sex == 1 & age > 65,
                     protect = FALSE)

# Example lists
my_df_list <- list(dummy_data(10),
                   dummy_data(10))

file1 <- file.path(tempdir(), "first.fst")
file2 <- file.path(tempdir(), "second.rds")
my_file_list <- list(file1, file2)

# Save multiple files at once
save_file_multi(data_frame_list = my_df_list,
                file_list       = my_file_list,
                protect         = FALSE)

# Save multiple files and only keep specific variables
save_file_multi(data_frame_list = my_df_list,
                file_list       = my_file_list,
                keep_list       = c(sex, age, state),
                protect         = FALSE)

# Save multiple files and keep different variables per data frame
save_file_multi(data_frame_list = my_df_list,
                file_list       = my_file_list,
                keep_list       = list(c(person_id, first_person),
                                       c(NUTS3, income, weight)),
                protect         = FALSE)

unlink(c(file1, file2,
         file.path(tempdir(), "testfile.fst"),
         file.path(tempdir(), "testfile.rds")))

# Example files
fst_file <- system.file("extdata", "qol_example_data_fst.fst", package = "qol")
rds_file <- system.file("extdata", "qol_example_data_rds.rds", package = "qol")

# Load file
my_fst <- load_file(path = dirname(fst_file),
                    file = basename(rds_file))
my_rds <- load_file(path = dirname(rds_file),
                    file = basename(rds_file))

# Load file and only keep specific variables
# NOTE: Variable names can be written case insensitive. Meaning if a variable
#       is stored as "age" and you write "AGE" in keep, the function will find
#       the variable and rename it to "AGE".
my_fst_keep<- load_file(path = dirname(fst_file),
                        file = basename(rds_file),
                        keep = c(AGE, INCOME_class, State, weight))

# Load file and subset observations
my_fst_where <- load_file(path  = dirname(fst_file),
                          file  = basename(rds_file),
                          where = sex == 1 & age > 65)

# Load multiple files and stack them
stack_files <- load_file_multi(c(fst_file, rds_file))

# Load multiple files and output them in a list
list_files <- load_file_multi(file_list   = c(fst_file, rds_file),
                              stack_files = FALSE)

# Load multiple files and only keep specific variables
all_files_keep <- load_file_multi(file_list = c(fst_file, rds_file),
                                  keep_list = c(Sex, AGE, stAte))

# Load multiple files and keep different variables per data frame
all_files_diff <- load_file_multi(file_list = c(fst_file, rds_file),
                                  keep_list = list(c(Person_ID, First_Person),
                                                   c(nuts3, Income, WEIGHT)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loading.R
\name{libname}
\alias{libname}
\title{Check If Path Exists And Retrieve Files}
\usage{
libname(path, get_files = FALSE, extensions = NULL, recursive = FALSE)
}
\arguments{
\item{path}{A folder path.}

\item{get_files}{FALSE by default. If TRUE returns a named character vector containing
file paths.}

\item{extensions}{Specify file extensions to be kept in the list when retrieving files.}

\item{recursive}{FALSE by default. If TRUE scans the provided folder for additional files
inside sub folders. Only work in combination with get_files = TRUE.}
}
\value{
Returns the given file path or a named character vector containing file paths.
}
\description{
\code{\link[=libname]{libname()}} checks if a given path exists and writes a message in the console accordingly.
Optional all files from the given path can be retrieved as a named character vector.
}
\examples{
my_path   <- libname("C:/My_Path/")
file_list <- libname("C:/My_Path/", get_files = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/small_helpers.R
\name{get_integer_length}
\alias{get_integer_length}
\title{Get Integer Length}
\usage{
get_integer_length(variable)
}
\arguments{
\item{variable}{The integer variable from which to get the length.}
}
\value{
Returns a vector with the number of digits places.
}
\description{
Get the number of digits of an integer variable.
}
\examples{
# Example data frame
my_data <- dummy_data(100)

my_data[["age_length"]] <- get_integer_length(my_data[["age"]])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/export_with_style.R
\name{export_with_style}
\alias{export_with_style}
\title{Export Data Frame With Style}
\usage{
export_with_style(
  data_frame,
  titles = .qol_options[["titles"]],
  footnotes = .qol_options[["footnotes"]],
  var_labels = .qol_options[["var_labels"]],
  workbook = NULL,
  style = .qol_options[["excel_style"]],
  column_align = NULL,
  output = .qol_options[["output"]],
  print = .qol_options[["print"]],
  monitor = .qol_options[["monitor"]]
)
}
\arguments{
\item{data_frame}{A data frame to print.}

\item{titles}{Specify one or more table titles.}

\item{footnotes}{Specify one or more table footnotes.}

\item{var_labels}{A list in which is specified which label should be printed for
which variable instead of the variable name.}

\item{workbook}{Insert a previously created workbook to expand the sheets instead of
creating a new file.}

\item{style}{A list of options can be passed to control the appearance of 'Excel' outputs.
Styles can be created with \code{\link[=excel_output_style]{excel_output_style()}}.}

\item{column_align}{Individually align the data frame columns in the final 'Excel' output.}

\item{output}{The following output formats are available: excel and excel_nostyle.}

\item{print}{TRUE by default. If TRUE prints the output, if FALSE doesn't print anything. Can be used
if one only wants to catch the output workbook.}

\item{monitor}{FALSE by default. If TRUE outputs two charts to visualize the functions time consumption.}
}
\value{
Returns a list with the data table containing the results for the table, the formatted
'Excel' workbook and the meta information needed for styling the final table.
}
\description{
\code{\link[=export_with_style]{export_with_style()}} prints a data frame as an individually styled 'Excel' table. Titles,
footnotes and labels for variable names can optionally be added.
}
\details{
\code{\link[=export_with_style]{export_with_style()}} is based on the 'SAS' procedure Proc Print, which outputs the data frame
as is into a styled table.
}
\examples{
# Example data frame
my_data <- dummy_data(100)
set_style_options(background_color = "")

# Define style
set_style_options(column_widths = c(2, 15, 15, 15, 9))

# Define titles and footnotes. If you want to add hyperlinks you can do so by
# adding "link:" followed by the hyperlink to the main text. Linking to another
# cell works with "cell:". To link to a file use "file:" an pass the full file
# path afterwards.
set_titles("This is title number 1",
           "This is title number 2 link: https://cran.r-project.org/",
           "This is title number 3 cell: W22",
           "This is title number 4 file: C:/MyFolder/MyFile.txt")

set_footnotes("This is footnote number 1",
              "This is footnote number 2 file: C:/MyFolder/MyFile.txt",
              "This is footnote number 3 cell: W22",
              "This is footnote number 4 link: https://cran.r-project.org/")

# Print styled data frame
my_data |> export_with_style()

# Retrieve formatted workbook, original table and meta information for further usage
export_list <- my_data |> export_with_style()

# To save a table as xlsx file you have to set the path and filename in the
# style element
# Example files paths
table_file <- tempfile(fileext = ".xlsx")

# Note: Normally you would directly input the path ("C:/MyPath/") and
#       name ("MyFile.xlsx").
#       With the set_style_options you can also set a table style globally.
set_style_options(save_path  = dirname(table_file),
                  file       = basename(table_file),
                  sheet_name = "MyTable")

my_data |> export_with_style()

# In case you have a good amount of tables, you want to combine in a single
# workbook, you can also catch the outputs and combine them afterwards in one go.
set_style_options(sheet_name = "sheet1")
tab1 <- my_data |> export_with_style(print = FALSE)

set_titles(NULL)
set_style_options(sheet_name = "sheet2")
tab2 <- my_data |> export_with_style(print = FALSE)

set_footnotes(NULL)
set_style_options(sheet_name = "sheet3")
tab3 <- my_data |> export_with_style(print = FALSE, output = "excel_nostyle")

# Every of the above tabs is a list, which contains the data table, an unstyled
# workbook and the meta information needed for the individual styling. These tabs
# can be input into the following function, which reads the meta information,
# styles each table individually and combines them as separate sheets into a
# single workbook.
combine_into_workbook(tab1, tab2, tab3)
combine_into_workbook(tab1, tab2, tab3, file = table_file)

# Manual cleanup for example
unlink(table_file)

# Global options are permanently active until the current R session is closed.
# There are also functions to reset the values manually.
reset_style_options()
reset_qol_options()
close_file()

set_print(TRUE)
set_style_options(background_color = "FFFFFF")

}
\seealso{
Creating a custom table style: \code{\link[=excel_output_style]{excel_output_style()}}, \code{\link[=modify_output_style]{modify_output_style()}},
\code{\link[=number_format_style]{number_format_style()}}, \code{\link[=modify_number_formats]{modify_number_formats()}}.

Global style options: \code{\link[=set_style_options]{set_style_options()}}, \code{\link[=set_labels]{set_labels()}}.

Other global options: \code{\link[=set_titles]{set_titles()}}, \code{\link[=set_footnotes]{set_footnotes()}}, \code{\link[=set_print]{set_print()}}, \code{\link[=set_monitor]{set_monitor()}},
\code{\link[=set_na.rm]{set_na.rm()}}, \code{\link[=set_print]{set_print()}}, \code{\link[=set_print_miss]{set_print_miss()}}, \code{\link[=set_output]{set_output()}}.

Combine Excel workbooks: \code{\link[=combine_into_workbook]{combine_into_workbook()}}.

Functions that can handle styles: \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}}, \code{\link[=any_table]{any_table()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/excel_helpers.R
\name{excel_output_style}
\alias{excel_output_style}
\title{Style for 'Excel' Table Outputs}
\usage{
excel_output_style(
  save_path = NULL,
  file = NULL,
  sheet_name = "Table",
  font = "Arial",
  column_widths = "auto",
  row_heights = "auto",
  title_heights = NULL,
  header_heights = NULL,
  subheader_heights = NULL,
  table_heights = NULL,
  footnote_heights = NULL,
  start_row = 2,
  start_column = 2,
  freeze_col_header = FALSE,
  freeze_row_header = FALSE,
  filters = TRUE,
  grid_lines = TRUE,
  by_as_subheaders = FALSE,
  background_color = "FFFFFF",
  header_back_color = "FFFFFF",
  header_font_color = "000000",
  header_font_size = 10,
  header_font_bold = TRUE,
  header_alignment = "center",
  header_stat_merging = "block",
  header_wrap = "1",
  header_indent = 0,
  header_borders = TRUE,
  header_border_color = "000000",
  subheader_back_color = "FFFFFF",
  subheader_font_color = "000000",
  subheader_font_size = 10,
  subheader_font_bold = TRUE,
  subheader_alignment = "center",
  subheader_wrap = "1",
  subheader_indent = 0,
  subheader_borders = TRUE,
  subheader_border_color = "000000",
  cat_col_back_color = "FFFFFF",
  cat_col_font_color = "000000",
  cat_col_font_size = 10,
  cat_col_font_bold = FALSE,
  cat_col_alignment = "left",
  cat_col_wrap = "1",
  cat_col_indent = 1,
  cat_col_borders = TRUE,
  cat_col_border_color = "000000",
  table_back_color = "FFFFFF",
  table_font_color = "000000",
  table_font_size = 10,
  table_font_bold = FALSE,
  table_alignment = "right",
  table_indent = 1,
  table_borders = FALSE,
  table_border_color = "000000",
  as_heatmap = FALSE,
  heatmap_low_color = "F8696B",
  heatmap_middle_color = "FFFFFF",
  heatmap_high_color = "63BE7B",
  box_back_color = "FFFFFF",
  box_font_color = "000000",
  box_font_size = 10,
  box_font_bold = TRUE,
  box_alignment = "center",
  box_wrap = "1",
  box_indent = 0,
  box_borders = TRUE,
  box_border_color = "000000",
  number_formats = number_format_style(),
  title_font_color = "000000",
  title_font_size = 10,
  title_font_bold = TRUE,
  title_alignment = "left",
  footnote_font_color = "000000",
  footnote_font_size = 8,
  footnote_font_bold = FALSE,
  footnote_alignment = "left",
  na_symbol = "."
)
}
\arguments{
\item{save_path}{If NULL, opens the output as temporary file. Otherwise specify an output path.}

\item{file}{If NULL, opens the output as temporary file. Otherwise specify a filename with extension.}

\item{sheet_name}{Name of the sheet inside the workbook to which the output shall be written.
If multiple outputs are produced in one go, the sheet name additionally receives a running number.}

\item{font}{Set the font to be used for the entire output.}

\item{column_widths}{Specify whether column widths should be set automatically and individually or
if a numeric vector is passed each column width can be specified manually. If a table has more
columns than column widths are provided, the last given column width will be repeated until the
end of the table.}

\item{row_heights}{Specify whether row heights should be set automatically and individually or
if a numeric vector is passed each row height can be specified manually. If a table has more
rows than row heights are provided, the last given row height will be repeated until the
end of the table.}

\item{title_heights}{Set individual row heights for the titles only.}

\item{header_heights}{Set individual row heights for the table header only.}

\item{subheader_heights}{Set individual row heights for the table subheader only.}

\item{table_heights}{Set individual row heights for the table body only.}

\item{footnote_heights}{Set individual row heights for the footnotes only.}

\item{start_row}{The row in which the table starts.}

\item{start_column}{The column in which the table starts.}

\item{freeze_col_header}{Whether to freeze the column header so that it is always visible
while scrolling down the document.}

\item{freeze_row_header}{Whether to freeze the row header so that it is always visible
while scrolling sideways in the document.}

\item{filters}{Whether to set filters in the column header, when exporting a data frame.}

\item{grid_lines}{Whether to show grid lines or not.}

\item{by_as_subheaders}{Whether to format by variables as subheaders in one table instead
of single tables on multiple sheets.}

\item{background_color}{Background cell color for any cell which isn't covered by the other
background color options.}

\item{header_back_color}{Background cell color of the table header.}

\item{header_font_color}{Font color of the table header.}

\item{header_font_size}{Font size of the table header.}

\item{header_font_bold}{Whether to print the table header in bold letters.}

\item{header_alignment}{Set the text alignment of the table header.}

\item{header_stat_merging}{Set how far the statistics row is merged. Allowed are "all",
"block" (default), "none".}

\item{header_wrap}{Whether to wrap the texts in the table header.}

\item{header_indent}{Indentation level of the table header.}

\item{header_borders}{Whether to draw borders around the table header cells.}

\item{header_border_color}{Borders colors of the table header cells.}

\item{subheader_back_color}{Background cell color of the table subheader.}

\item{subheader_font_color}{Font color of the table subheader.}

\item{subheader_font_size}{Font size of the table subheader.}

\item{subheader_font_bold}{Whether to print the table subheader in bold letters.}

\item{subheader_alignment}{Set the text alignment of the table subheader.}

\item{subheader_wrap}{Whether to wrap the texts in the table subheader.}

\item{subheader_indent}{Indentation level of the table subheader.}

\item{subheader_borders}{Whether to draw borders around the table subheader cells.}

\item{subheader_border_color}{Borders colors of the table subheader cells.}

\item{cat_col_back_color}{Background cell color of the category columns inside the table.}

\item{cat_col_font_color}{Font color of the category columns inside the table.}

\item{cat_col_font_size}{Font size of the category columns inside the table.}

\item{cat_col_font_bold}{Whether to print the category columns inside the table in bold letters.}

\item{cat_col_alignment}{Set the text alignment of the category columns inside the table.}

\item{cat_col_wrap}{Whether to wrap the texts in the category columns inside the table.}

\item{cat_col_indent}{Indentation level of the category columns inside the table.}

\item{cat_col_borders}{Whether to draw borders around the category columns inside the table.}

\item{cat_col_border_color}{Borders colors of the category columns inside the table.}

\item{table_back_color}{Background color of the inner table cells.}

\item{table_font_color}{Font color of the inner table cells.}

\item{table_font_size}{Font size of the inner table cells.}

\item{table_font_bold}{Whether to print the inner table cells in bold numbers}

\item{table_alignment}{Set the text alignment of the inner table cells.}

\item{table_indent}{Indentation level of the inner table cells.}

\item{table_borders}{Whether to draw borders around the inner table cells.}

\item{table_border_color}{Borders colors of the inner table cells.}

\item{as_heatmap}{Whether to lay a conditional formatting over the values.}

\item{heatmap_low_color}{The color for lower values in the conditional formatting.}

\item{heatmap_middle_color}{The color for middle values in the conditional formatting.}

\item{heatmap_high_color}{The color for high values in the conditional formatting.}

\item{box_back_color}{Background color of the left box in table header.}

\item{box_font_color}{Font color of the left box in table header.}

\item{box_font_size}{Font size of the left box in table header.}

\item{box_font_bold}{Whether to print the left box in table header in bold letters.}

\item{box_alignment}{Set the text alignment of the left box in table header.}

\item{box_wrap}{Whether to wrap the texts in the left box in table header.}

\item{box_indent}{Indentation level of the left box in table header.}

\item{box_borders}{Whether to draw borders around the left box in table header.}

\item{box_border_color}{Borders colors of the left box in table header.}

\item{number_formats}{Put in a list of number formats which should be assigned to
the different stats. Number formats can be created with \code{\link[=number_format_style]{number_format_style()}}.}

\item{title_font_color}{Font color of the titles.}

\item{title_font_size}{Font size of the tables titles.}

\item{title_font_bold}{Whether to print the tables titles in bold letters.}

\item{title_alignment}{Set the text alignment of the titles.}

\item{footnote_font_color}{Font color of the footnotes}

\item{footnote_font_size}{Font size of the tables footnotes}

\item{footnote_font_bold}{Whether to print the tables footnotes in bold letters.}

\item{footnote_alignment}{Set the text alignment of the footnotes.}

\item{na_symbol}{Define the symbol that should be used for NA values.}
}
\value{
Returns a list of named style options.
}
\description{
Set different options which define the visual output of 'Excel' tables produced
by \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}} and \code{\link[=any_table]{any_table()}}.
}
\details{
\code{\link[=excel_output_style]{excel_output_style()}} is based on the Output Delivery System (ODS) in 'SAS',
which provides efficient and readable ways to set up different table styles.

With the output style you have full control over the table design. There is no
need to think about calculating the right place to input a background color or a
border of a certain type and how to do this in a loop for multiple cells. Just
input colors, borders, font styles, etc. for the different table parts and everything
else is handled by the functions capable of using styles.

The concept basically is: design over complex calculations.
}
\examples{
# For default values
excel_style <- excel_output_style()

# Set specific options, the rest will be set to default values
excel_style <- excel_output_style(font       = "Calibri",
                                  sheet_name = "My_Output")

# For cells with no background color pass an empty string
excel_style <- excel_output_style(table_back_color = "")

}
\seealso{
Creating a custom table style: \code{\link[=modify_output_style]{modify_output_style()}},
\code{\link[=number_format_style]{number_format_style()}}, \code{\link[=modify_number_formats]{modify_number_formats()}}.

Global style options: \code{\link[=set_style_options]{set_style_options()}}, \code{\link[=set_labels]{set_labels()}}.

Functions that can handle styles: \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}}, \code{\link[=any_table]{any_table()}},
\code{\link[=export_with_style]{export_with_style()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/excel_helpers.R
\name{modify_number_formats}
\alias{modify_number_formats}
\title{Modify Number Formats Used by \code{\link[=any_table]{any_table()}}}
\usage{
modify_number_formats(formats_to_modify, ...)
}
\arguments{
\item{formats_to_modify}{Pre created number formats where only certain elements
should be modified while the rest is kept as is.}

\item{...}{Pass in names and corresponding new values for existing number formats.}
}
\value{
Returns a modified list of number format options.
}
\description{
Modify previously created number formats with \code{\link[=number_format_style]{number_format_style()}}.
}
\details{
\code{\link[=modify_number_formats]{modify_number_formats()}} is based on 'SAS' number formats and the Output Delivery
System (ODS), which provides efficient and readable ways to set up different table
styles.

With the number format style you have full control over formatting numbers according
to the different statistics. There is no need to think about calculating the right
place to input the number formats and how to do this in a loop for multiple cells.
Just input the different number formats and decimals for the different statistics
and everything else is handled by the functions capable of using number styles.

The concept basically is: design over complex calculations.
}
\examples{
# For default values
format_list <- number_format_style(pct_excel    = "0.00000000",
                                   pct_decimals = 8)

# Set specific options, the rest will be kept as is
format_list <- format_list |> modify_number_formats(sum_excel = "#,###,##0.000")

# IMPORTANT: Don't forget to add individual formats to an excel style, otherwise
# they won't come into affect.
excel_style <- excel_output_style(number_formats = format_list)

}
\seealso{
Creating a custom table style: \code{\link[=excel_output_style]{excel_output_style()}}, \code{\link[=modify_output_style]{modify_output_style()}},
\code{\link[=number_format_style]{number_format_style()}}.

Global style options: \code{\link[=set_style_options]{set_style_options()}}, \code{\link[=set_labels]{set_labels()}}.

Functions that can handle styles: \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}}, \code{\link[=any_table]{any_table()}},
\code{\link[=export_with_style]{export_with_style()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/macro.R
\name{macro}
\alias{macro}
\alias{apply_macro}
\title{Resolve Macro Variables In A Text}
\usage{
macro(text)

apply_macro(character_vector)
}
\arguments{
\item{text}{A text containing macro variables to resolve.}

\item{character_vector}{A vector containing multiple character expressions in which
to resolve macro variables.}
}
\value{
\code{\link[=macro]{macro()}}: Returns a character.

\code{\link[=apply_macro]{apply_macro()}}: Returns a character vector.
}
\description{
Variables written with a leading "&" inside a text will be resolved into the
character or numeric expression the corresponding object is holding.
}
\details{
Macro variables in 'SAS' can be set up with \%Let and can act as global accessible
variables. This in itself is nothing special to 'R' because basically every object
created outside a function is a global variable.

To use these global objects within a text one has to use e.g. paste0("The current year is: ", year).
With the macro function one can write it like this: macro("The current year is &year").

So where is the benefit? If implemented within a function, a parameter like "title" or
"footnote" in the tabulation functions, can resolve these variables without the need
of another function. You can just pass the character expression "The current year is &year"
and the implementation inside the function resolves the macro variable directly in place.
}
\examples{
# Resolving macro variable in single character
year <- 2026

text <- macro("The current year is &year")

# You can also combine multiple macro variables
some_variable <- "current"
current2026   <- "The current year is"

text_combi <- macro("&&some_variable&year &year")

# Resolving macro variable in character vector
char_vector <- c("The current year is &year",
                 "The &some_variable year is &year",
                 "&&some_variable&year &year")

text_vector <- apply_macro(char_vector)

}
\seealso{
Functions that can make use of macro variables: \code{\link[=any_table]{any_table()}}, \code{\link[=frequencies]{frequencies()}},
\code{\link[=crosstabs]{crosstabs()}}, \code{\link[=export_with_style]{export_with_style()}}, \code{\link[=summarise_plus]{summarise_plus()}}

Within the tabulation functions titles, footnotes, var_labels and stat_labels
can resolve macros. Additionally they can be used in the \code{\link[=any_table]{any_table()}} rows and
columns parameter and in the \code{\link[=summarise_plus]{summarise_plus()}} types parameter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/retain.R
\name{retain}
\alias{retain}
\alias{running_number}
\alias{mark_case}
\alias{retain_value}
\alias{retain_stat}
\alias{retain_variables}
\title{Different Facets of Retain}
\usage{
running_number(data_frame, by = NULL, sort = FALSE, group_nr = FALSE)

mark_case(data_frame, by = NULL, first = TRUE)

retain_value(data_frame, values, by = NULL)

retain_stat(data_frame, values, by = NULL, statistics = "sum")

retain_variables(data_frame, ..., order_last = FALSE)
}
\arguments{
\item{data_frame}{The data frame in which to compute retained variables.}

\item{by}{By group in which to compute the retained variable.}

\item{first}{\code{\link[=mark_case]{mark_case()}}: If TRUE marks the first case within a group, otherwise
the last case.}

\item{values}{One or multiple variables of which a value should be retained.}

\item{statistics}{Available functions:
\itemize{
\item "sum"       -> Weighted and unweighted sum
\item "sum_wgt"   -> Sum of all weights
\item "freq"      -> Unweighted frequency
\item "freq_g0"   -> Unweighted frequency of all values greater than zero
\item "pct_group" -> Weighted and unweighted percentages within the respective group
\item "pct_total" -> Weighted and unweighted percentages compared to the grand total
\item "mean"      -> Weighted and unweighted mean
\item "median"    -> Weighted and unweighted median
\item "mode"      -> Weighted and unweighted mode
\item "min"       -> Minimum
\item "max"       -> Maximum
\item "sd"        -> Weighted and unweighted standard deviation
\item "variance"  -> Weighted and unweighted standard variance
\item "first"     -> First value
\item "last"      -> Last value
\item "pn"        -> Weighted and unweighted percentiles (any p1, p2, p3, ... possible)
Uses the SAS default way of calculating percentiles, which is Type 2 described in \code{\link[=quantile]{quantile()}}.
\item "missing"   -> Missings generated by the value variables
}}

\item{...}{\code{\link[=retain_variables]{retain_variables()}}: Put in single variable names or variable ranges
(var_name1:var_name10) which should be ordered to the front or back of the data frame.
It is also possible to provide none existent variable names which will then be added
to the data frame.}

\item{retain_variables}{\code{\link[=retain_variables]{retain_variables()}}: FALSE by default. If TRUE puts the
variables at the end of the data frame instead of the beginning.}
}
\value{
\code{\link[=running_number]{running_number()}}: Returns a vector containing a running number.

\code{\link[=mark_case]{mark_case()}}: Returns a vector containing a the marking for first or last cases.

\code{\link[=retain_value]{retain_value()}}: Returns a vector containing a retained value.

\code{\link[=retain_stat]{retain_stat()}}: Returns a vector containing a retained sum.

\code{\link[=retain_variables]{retain_variables()}}: Return the data frame with reordered variables.
}
\description{
These retain functions all have one thing in common: transferring a value from one case
to the next. What they make out of this functionality can be quiet different. Therefor
there is a function for each different use case.

\code{\link[=running_number]{running_number()}} computes running numbers in a data frame. Without specifying
a by variable results in the row number. With by variable computes the running
number within each group of expressions.

\code{\link[=mark_case]{mark_case()}} sets a flag for the first or last case within the provided by group.

\code{\link[=retain_value]{retain_value()}} retains the first value for all cases of the same group and saves
it into a new variable.

\code{\link[=retain_stat]{retain_stat()}} retains the chosen statistic for all cases of the same group and saves
it into a new variable.

\code{\link[=retain_variables]{retain_variables()}} orders the provided variables to the front or back of the
data frame. If a variable is not part of the data frame it will be added with
all NA values at the desired position.
}
\details{
The functions listed here are based on the 'SAS' function retain. On a very basic level retain
can do two things, depending on the position in the 'SAS' code: It can either sort variables
column wise or it can - since it works row wise - remember a value from one row to the next.
The functions here concentrate on the second part.

Remembering a value from a previous observation offers multiple use cases. E.g. always adding
+1 to the previous case creates a running number. Or if an observation knows the value of the
previous one, it can check whether it is of the same value or another, e.g. to mark first or last
cases within a group.

In it's simplest form it can remember a value from the first observation and transfer it to all
other observations.

All of these functions work on the whole data frame as well as on groups, e.g. to transfer
a value from the first person in a household to all other persons of the same household.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Get row numbers
my_data[["run_nr"]] <- my_data |> running_number()

# Running number per variable expression
my_data[["run_nr_by"]] <- my_data |> running_number(by = year)

# Running number within a group of variables which need to be sorted beforehand
my_data[["run_nr_by_sort"]] <- my_data |>
    running_number(by   = c(sex, education, income_class),
                   sort = TRUE)

# Running number within a group of variables which need to be sorted beforehand
my_data[["run_group"]] <- my_data |>
    running_number(by       = c(sex, education, income_class),
                   sort     = TRUE,
                   group_nr = TRUE)

# Mark first and last cases
my_data[["first"]] <- my_data |> mark_case(by = household_id)
my_data[["last"]]  <- my_data |> mark_case(by    = household_id,
                                           first = FALSE)

# Retain first value without grouping
my_data[["first_weight"]] <- my_data |> retain_value(value = weight)

# Retain first value inside a group for multiple variables
my_data[, c("household_weight", "household_icome")] <- my_data |>
    retain_value(values = c(weight, income),
                 by     = c(state, household_id))

# Retain sum without grouping
my_data[["total_sum"]] <- my_data |> retain_stat(values = weight)

# Retain sum inside a group for multiple variables
my_data[, c("weight_sum", "income_sum")] <- my_data |>
    retain_stat(values = c(weight, income),
                by     = c(year, state, household_id))

# Retain different statistics
my_data[, c("income_freq", "income_max")] <- my_data |>
    retain_stat(values     = income,
                by         = c(year, state, household_id),
                statistics = c("freq", "max"))

# Retain columns inside data frame, which orders them to the front
my_data <- my_data |> retain_variables(age, sex, income)

# Retain columns inside data frame, but order them to the end.
# Variable ranges can also be used.
my_data <- my_data |> retain_variables(age:income, order_last = TRUE)

# Retain columns inside data frame and add new variables with all NA values
my_data <- my_data |> retain_variables(age, sex, income, status1:status5)

# You can also use the colon as a placeholder for any text
start1   <- my_data |> retain_variables("s:")   # Variable names start with "s"
end1     <- my_data |> retain_variables(":id")  # Variable names end with "id"
contain1 <- my_data |> retain_variables(":on:") # Variable names which contain "on"

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/if_else.R
\name{where.}
\alias{where.}
\title{Filter Data Frame With Direct View}
\usage{
where.(data_frame, condition = NULL, keep = NULL)
}
\arguments{
\item{data_frame}{A data frame on which to apply filters.}

\item{condition}{The condition on which to filter observations.}

\item{keep}{The Variables to keep in the result data frame.}
}
\value{
Returns a filtered data frame.
}
\description{
Filter observations and variables and directly view the result on screen.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Get a quick filtered view
my_data |> where.(sex == 1 & age < 25,
                  c(sex, age, household_id, education))

}
\seealso{
\code{\link[=ifelse]{ifelse()}} with multiple conditions and 'SAS' writing style: \code{\link[=ifelse_multi]{ifelse_multi()}}

The following functions can make use of the \code{\link[=do_if]{do_if()}} filter variables:

Conditions: \code{\link[=if.]{if.()}}, \code{\link[=else_if.]{else_if.()}}, \code{\link[=else.]{else.()}}

Filter Data Frame: \code{\link[=where.]{where.()}}

Create new Variables: \code{\link[=compute.]{compute.()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/any_table.R
\name{combine_into_workbook}
\alias{combine_into_workbook}
\title{Combine Multiple Tables Into One Workbook}
\usage{
combine_into_workbook(
  ...,
  file = NULL,
  output = "excel",
  print = TRUE,
  monitor = FALSE
)
}
\arguments{
\item{...}{Provide any number of result lists output by \code{\link[=any_table]{any_table()}}.}

\item{file}{If NULL, opens the output as temporary file. If a filename with path
is specified, saves the output to the specified path.}

\item{output}{The following output formats are available: excel and excel_nostyle.}

\item{print}{TRUE by default. If TRUE prints the output, if FALSE doesn't print anything. Can be used
if one only wants to catch the combined workbook.}

\item{monitor}{FALSE by default. If TRUE outputs two charts to visualize the functions time consumption.}
}
\value{
A fully styled workbook containing the provided tables.
}
\description{
Combines any number of tables created with \code{\link[=any_table]{any_table()}} into one workbook
and styles them according to their meta information.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)
my_data[["person"]] <- 1

# Formats
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

education. <- discrete_format(
    "Total"            = c("low", "middle", "high"),
    "low education"    = "low",
    "middle education" = "middle",
    "high education"   = "high")

# Define style
set_style_options(column_widths = c(2, 15, 15, 15, 9))

# Define titles and footnotes. If you want to add hyperlinks you can do so by
# adding "link:" followed by the hyperlink to the main text.
set_titles("This is title number 1 link: https://cran.r-project.org/",
           "This is title number 2 cell: W22",
           "This is title number 3 file: C:/MyFolder/MyFile.docx",
           "This is title number 4")
set_footnotes("This is footnote number 1 cell: W22",
              "This is footnote number 2 file: C:/MyFolder/MyFile.docx",
              "This is footnote number 3 link: https://cran.r-project.org/",
              "This is footnote number 4")

# Catch the output and additionally use the options:
# print = FALSE and output = "excel_nostyle".
# This skips the styling and output part, so that the function runs faster.
# The styling is done later on.
set_print(FALSE)
set_output("excel_nostyle")
set_style_options(sheet_name = "big table")

tab1 <- my_data |> any_table(rows       = c("sex + age", "sex", "age"),
                             columns    = c("year", "education + year"),
                             values     = weight,
                             statistics = c("sum", "pct_group"),
                             pct_group  = c("sex", "age", "education", "year"),
                             formats    = list(sex = sex., age = age.,
                                               education = education.),
                             na.rm      = TRUE)

set_style_options(sheet_name = "age_sex")

tab2 <- my_data |> any_table(rows       = "age",
                             columns    = "sex",
                             values     = weight,
                             statistics = "sum",
                             formats    = list(sex = sex., age = age.),
                             na.rm      = TRUE)

set_style_options(sheet_name = "edu_year")

tab3 <- my_data |> any_table(rows       = "education",
                             columns    = "year",
                             values     = weight,
                             statistics = "pct_group",
                             formats    = list(education = education.),
                             na.rm      = TRUE)

# Every of the above tabs is a list, which contains the data table, an unstyled
# workbook and the meta information needed for the individual styling. These
# tabs can be input into the following function, which reads the meta information,
# styles each table individually and combines them as separate sheets into a single workbook.
combine_into_workbook(tab1, tab2, tab3)

# Reset the global options afterwards
set_print(TRUE)
set_output("excel")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/small_helpers.R
\name{setcolorder_by_pattern}
\alias{setcolorder_by_pattern}
\title{Order Columns by Variable Name Patterns}
\usage{
setcolorder_by_pattern(data_frame, pattern)
}
\arguments{
\item{data_frame}{The data frame to be ordered.}

\item{pattern}{The pattern which is used for ordering the data frame columns.}
}
\value{
Returns a reordered data frame with the ordered variables at the end.
}
\description{
Order variables in a data frame based on a pattern rather than whole variable
names. E.g. grab every variable that contains "sum" in it's name and order
them together so that they appear next to each other.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Summarise data
all_nested <- my_data |>
    summarise_plus(class      = c(year, sex),
                   values     = c(weight, income),
                   statistics = c("sum", "pct_group", "pct_total", "sum_wgt", "freq"),
                   weight     = weight,
                   nesting    = "deepest",
                   na.rm      = TRUE)

# Set a different column order
new_order <- all_nested |> setcolorder_by_pattern(c("pct", "freq", "sum"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/if_else.R
\name{ifelse_multi}
\alias{ifelse_multi}
\title{Do Multiple ifelse At Once}
\usage{
ifelse_multi(data_frame, ..., else. = NA, do_if = NULL, na.rm = TRUE)
}
\arguments{
\item{data_frame}{A data frame on which to apply multiple ifelse statements.}

\item{...}{The conditions as character and the assignments in the form:
"condition" = value, "condition" = value, ...}

\item{else.}{The default value which is applied when no condition results in
an assignment.}

\item{do_if}{Define an overarching condition that will be used on all other conditions.}

\item{na.rm}{TRUE by default. Sets the default value even though a condition results
in NA. If FALSE leaves NA as missing value.}
}
\value{
Returns a vector with a conditionally computed variable.
}
\description{
\code{\link[=ifelse_multi]{ifelse_multi()}} handles multiple conditions and value assignments at once by
nesting multiple ifelse statements. It is special in the case that it takes the
conditions as unevaluated characters to be able to parse them before evaluation.
This enables a SAS like writing style of the conditions.
}
\details{
The function takes in conditions as characters to be able to parse the conditions.
The parsing allows to write conditions in a SAS like way. For example the condition

"age >= 15 & age < 65" can be written as "15 <= age < 65"

Additionally and/or (case insensitive) are recognized keywords which will be translated into &/|.

"age >= 15 and age < 65 or sex == 1" becomes "age >= 15 & age < 65 | sex == 1"

And \code{\link[=macro]{macro()}} variables are allowed in any spot, e.g.:

YEAR <- 2025\
"year == &YEAR" translates to "year == 2025".

Single = get translated to ==, the SAS not ^= is translated to !=.

In SAS you can write "age in (1 4 7 15 21)" or "age not in (2, 4, 37, 82)" which
gets translated into "age \%in\% (1, 4, 7, 15, 21)" and "!(age \%in\% (2, 4, 37, 82))".

When using == or != operators you can check if character expressions start with
(text:), end with (:text) or contain a certain text (:text:). These also get
translated.
}
\examples{
# Example data frame
my_data <- dummy_data(100)

# Simple ifelse statement
my_data[["under18"]]    <- my_data |> ifelse_multi(" age < 18 " = 1,       else. = 0)
my_data[["middle_age"]] <- my_data |> ifelse_multi(" 15 <= age < 65 " = 1, else. = 0)

my_data[["age_gr"]] <- my_data |>
    ifelse_multi(" age < 18 "       = "under 18",
                 " 18 <= age < 25 " = "18 to under 25",
                 " 25 <= age < 50 " = "25 to under 50",
                 " 50 <= age < 65 " = "50 to under 65",
                 " 65 <= age      " = "65 and more")

# With overarching do_if condition
my_data[["age_gr_edu"]] <- my_data |>
    ifelse_multi(do_if = " education in ('middle' 'high') ",
                     " age < 18 "       = "under 18",
                     " 18 <= age < 25 " = "18 to under 25",
                     " 25 <= age < 50 " = "25 to under 50",
                     " 50 <= age < 65 " = "50 to under 65",
                     " 65 <= age      " = "65 and more")

# And/or translation
my_data[["and"]] <- my_data |> ifelse_multi(" age > 65 and sex = 1 " = 1,
                                            " age > 65 and sex = 2 " = 2,
                                            else. = 0)

my_data[["or"]] <- my_data |> ifelse_multi(" age > 65 or sex = 1 " = 1,
                                           " age > 65 or sex = 2 " = 2,
                                           else. = 0)

# "in" translation
my_data[["in"]] <- my_data |> ifelse_multi(" age in (1 10 25 65 90) " = 1, else. = 0)

# Colon translation: start/ends with and contains
my_data[["start"]]    <- my_data |> ifelse_multi(" education == 'lo:' "  = 1, else. = 0)
my_data[["end"]]      <- my_data |> ifelse_multi(" education == ':le' "  = 1, else. = 0)
my_data[["contains"]] <- my_data |> ifelse_multi(" education == ':ig:' " = 1, else. = 0)

# Macro variables can be integrated in any place
variable     <- "age"
age_to_check <- 18
value_to_set <- "under 18"

my_data[["macro"]] <- my_data |>
    ifelse_multi(" &variable < &age_to_check " = "&value_to_set",
                 else.                         = "other")

# NA translation
my_data[["NA"]]    <- my_data |> ifelse_multi(" age == .       " = 1, else. = 0)
my_data[["notNA"]] <- my_data |> ifelse_multi(" education != . " = 1, else. = 0)

# Pass in existing variable values
my_data[["income_mix"]] <- my_data |>
    ifelse_multi(" age < 50 " = income, else. = expenses)

}
\seealso{
Conditions: \code{\link[=if.]{if.()}}, \code{\link[=else_if.]{else_if.()}}, \code{\link[=else.]{else.()}}

Filter Data Frame: \code{\link[=where.]{where.()}}

Create new Variables: \code{\link[=compute.]{compute.()}}

Resolve macro variables: \code{\link[=macro]{macro()}}, \code{\link[=apply_macro]{apply_macro()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/small_helpers.R
\name{round_values}
\alias{round_values}
\alias{round_multi}
\title{Round Values With Half Rounded Up And Multiples Of X}
\usage{
round_values(values, digits = 0, multiple = NULL)

round_multi(
  data_frame,
  variables,
  new_names = NULL,
  digits = 0,
  multiple = NULL
)
}
\arguments{
\item{values}{Numeric values to round.}

\item{digits}{The number of decimal places the values should be rounded to.}

\item{multiple}{The multiple to round the values to.}

\item{data_frame}{The data frame in which the variables to be rounded are found.}

\item{variables}{The variable names of which to round the values.}

\item{new_names}{The new names of the rounded variables. If provided adds variables,
if not overwrites the existing variables with rounded values.}
}
\value{
\code{\link[=round_values]{round_values()}}: Returns rounded values.

\code{\link[=round_multi]{round_multi()}}: Returns a data frame with rounded values.
}
\description{
This function rounds values according to DIN 1333 (round half up) or as an
alternative in multiples of a given value.

\code{\link[=round_multi]{round_multi()}}: Rounds multiple variables at once inside a data frame.
}
\examples{
# With vectors
round_numbers1 <- round_values(c(-0.5, -0.4, 0.1, 0.49, 0.5, 1.5, 2.5, 3.2))
round_numbers2 <- round_values(c(-0.5, -0.49, 0.17, 0.499, 0.51, 1.549, 2.51, 3.25),
                               digits = 1)
round_numbers3 <- round_values(c(-0.3, -0.24, 1.17, 2.749, 0.25, 1.549, 2.75, 3.25),
                               multiple = 0.5)

# With a data frame
my_data <- dummy_data(100)

my_data[["income_round1"]] <- my_data[["income"]] |> round_values()
my_data[["income_round2"]] <- my_data[["income"]] |> round_values(multiple = 100)


# Round multiple variables in a data frame
my_data <- my_data |> round_multi(variables = c(income,  expenses,  balance),
                                  new_names = c(incomeR, expensesR, balanceR),
                                  digits = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/monitor.R
\name{monitor}
\alias{monitor}
\alias{monitor_start}
\alias{monitor_end}
\alias{monitor_next}
\alias{monitor_plot}
\title{Monitor Time Consumption}
\usage{
monitor_start(monitor_df, section, group = "Total")

monitor_end(monitor_df)

monitor_next(monitor_df, section, group = "Total")

monitor_plot(monitor_df, by = "section", draw_plot = TRUE)
}
\arguments{
\item{monitor_df}{A data table in which the delta times with their respective section names are stored.}

\item{section}{A named section for which to store delta times.}

\item{group}{Optionally pass a broader group name to be able to plot summarised delta times in addition to the detailed ones.
"Total" as default value.}

\item{by}{Use "section" for a detailed plot and "group" for summarised categories.}

\item{draw_plot}{Conditionally draw plots. TRUE by default.}
}
\value{
Returns a small data table with section-, group-names and corresponding delta times.
}
\description{
The monitor functions offer a simple way to keep track of timings and visualize them in charts.
If used throughout a longer syntax it is useful to identify bottlenecks or just get a better feeling
which passages take more time than others.

\code{\link[=monitor_start]{monitor_start()}} starts a new timing and adds this as an observation to the monitoring
data table. Pass NULL as monitor_df if you call the function for the first time to create a new
monitoring data table.

\code{\link[=monitor_end]{monitor_end()}} ends the current timing and calculates corresponding delta.

\code{\link[=monitor_next]{monitor_next()}} ends the current timing and calculates corresponding delta. In addition directly starts
a new timing for a new section.

\code{\link[=monitor_plot]{monitor_plot()}} outputs two charts to visualize the saved delta times.
}
\examples{
# Example data frame
monitor_df <- NULL |> monitor_start("Generate data frame", "Preparation")

my_data <- dummy_data(1000)

# Formats
monitor_df <- monitor_df |> monitor_next("Create formats", "Preparation")

age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:65,
    "65 and older"   = 65:100)

sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

# Evaluations
monitor_df <- monitor_df |> monitor_next("Nested summarise", "Summarise")

all_nested <- my_data |>
    summarise_plus(class      = c(year, sex, age),
                   values     = income,
                   statistics = c("sum", "pct_group", "pct_total", "sum_wgt", "freq"),
                   formats    = list(sex = "sex.", age = "age."),
                   weight     = weight,
                   nesting    = "deepest",
                   na.rm      = TRUE)

monitor_df <- monitor_df |> monitor_next("All summarise", "Summarise")

all_possible <- my_data |>
    summarise_plus(class      = c(year, sex, age),
                   values     = c(probability),
                   statistics = c("sum", "p1", "p99", "min", "max", "freq", "freq_g0"),
                   formats    = list(sex    = "sex.",
                                     age    = "age."),
                   weight     = weight,
                   nesting    = "all",
                   na.rm      = TRUE)

monitor_df <- monitor_df |> monitor_end()

# For detailed plot
monitor_df |> monitor_plot()

# For summarised plot
monitor_df |> monitor_plot(by = "group")

# NOTE: The more complex functions in this package have a detailed monitoring
#       integrated which can be viewed by setting the argument 'monitor' to TRUE.

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_rstheme.R
\name{build_rstheme}
\alias{build_rstheme}
\title{Build a Theme From Scratch}
\usage{
build_rstheme(
  file_path,
  theme_name = "qol_green",
  dark_theme = TRUE,
  editor_background = "#062625",
  editor_headline = "#3B3B3B",
  editor_font = "#C3B79D",
  toolbar = "#2E2E2E",
  tab = "#3B3B3B",
  selected_tab = "#062625",
  line_number = "#C3B79D",
  print_margin = "#3B3B3B",
  cursor = "#CCCCCC",
  selection = "#1B436E",
  smart_highlight = "#3686dc",
  bracket_highlight = "#595959",
  active_line = "#202324",
  whitespace = "#CCCCCC",
  debug_line = "#F18889",
  scrollbar = "#3B3B3B",
  scrollbar_hover = "#595959",
  scrollbar_active = "#BFBFBF",
  class_name = "#BEDD1A",
  keyword = "#FFC90E",
  language_constant = "#FFC90E",
  function_name = "#C3B79D",
  numeric = "#C93F3F",
  string = "#63C2C9",
  regex = "#E8E6E3",
  variable = "#E8E6E3",
  comment = "#32CD32",
  symbol = "#C3B79D",
  console_code = "#C3B79D",
  markdown_code = "#083332"
)
}
\arguments{
\item{file_path}{The path to which the theme file should be saved.}

\item{theme_name}{The themes name.}

\item{dark_theme}{Handles some elements not covered with the other parameters.}

\item{editor_background}{Base background color in the editor.}

\item{editor_headline}{Mostly used for the headlines of the environment panel.}

\item{editor_font}{Base font color of the editor.}

\item{toolbar}{Base toolbar and frame color.}

\item{tab}{Color of inactive tabs.}

\item{selected_tab}{Color of active tabs.}

\item{line_number}{The color of the line numbers on the left.}

\item{print_margin}{Color of the vertical line showing the print margin.}

\item{cursor}{Cursor color.}

\item{selection}{The background color of the current selection.}

\item{smart_highlight}{Background color of smart highlighted words.}

\item{bracket_highlight}{Background color of highlighted bracket pairs.}

\item{active_line}{Color for the active line the cursor is in.}

\item{whitespace}{Color for whitespace characters.}

\item{debug_line}{Color of the current debug line.}

\item{scrollbar}{Color of the scrollbars.}

\item{scrollbar_hover}{Highlight color when hovering over a scrollbar.}

\item{scrollbar_active}{Highlight color when clicking on a scrollbar.}

\item{class_name}{Code color for class names (like package names).}

\item{keyword}{Code color for fixed keywords (like function, if, else).}

\item{language_constant}{Code color for language constants (like the @ keywords).}

\item{function_name}{Code color for base and package functions.}

\item{numeric}{Code color for numeric values.}

\item{string}{Code color for string values.}

\item{regex}{Code color for regex expressions.}

\item{variable}{Code color for variables, parameters and arguments.}

\item{comment}{Code color for comments.}

\item{symbol}{Code Color of symbols (like <-, brackets).}

\item{console_code}{Color of executed Code in the Console.}

\item{markdown_code}{Background color of code passages in a markdown file.}
}
\value{
Saves a complete theme file.
}
\description{
Build your own theme by just setting up the colors for the different parts of
RStudio. A theme file will be exported which can be added by going to:

Tools -> Global Options -> Appearance -> Add
}
\details{
In the 'SAS Enterprise Guide' the user is able to not only choose a given theme, but
to also pick the colors for the different parts of the editor by themselves. Everyone
has a different taste of what colors look pleasing to the eyes, so you should be able
to choose them by yourself.
}
\examples{
# Example export file paths
# NOTE: These tempfiles are only for the examples. In reality you just call the
# main function and put in your desired path and name directly.
temp_file <- tempfile(fileext = ".rstheme")
file_name <- basename(tools::file_path_sans_ext(temp_file))

# Example theme
build_rstheme(file_path         = dirname(temp_file),
              theme_name        = file_name,
              editor_background = "#417291",
              editor_headline   = "#602BCA",
              editor_font       = "#C75C48")

# Manual cleanup for example
unlink(temp_file)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_format.R
\name{formats}
\alias{formats}
\alias{discrete_format}
\alias{interval_format}
\title{Create Format Container}
\usage{
discrete_format(...)

interval_format(..., include_lower = TRUE, include_upper = FALSE)
}
\arguments{
\item{...}{List all the desired recodings/recoding ranges. Every element containsa text for
the new category name and the values/value ranges which should be recoded into this new category.}

\item{include_lower}{TRUE by default. Whether to include lower bound value in
interval formats.}

\item{include_upper}{FALSE by default. Whether to include upper bound value in
interval formats.}
}
\value{
Returns a data table which contains the values/value ranges with the corresponding labels
}
\description{
Create a format container which stores discrete or interval values
with corresponding labels that can be applied by using functions capable of
using formats.

Create a format container independent from any data frame. Define which values
should be recoded into which new categories, if the format is applied to
a variable in a data frame.
It is possible to assign a single value to multiple new categories to create
a multilabel.
It is recommended to let format names end with a dot to make them stand out.
}
\details{
The concept of having formats as molds or stencils to put the data through, is inspired by
'SAS' formats. In 'SAS' formats are defined with the procedure Proc Formats, which is adapted
with \code{\link[=discrete_format]{discrete_format()}} and \code{\link[=interval_format]{interval_format()}}. Here you can define, which values
should be transferred into which result categories. This is completely detached from the data
your working with.

The great thing about this is, that one can not only label and recode values, but one can also
define so called multilabels. Meaning, one original value can be transferred into multiple
result categories.

A cell in a data frame can only hold one distinct value, which is normally a good thing.
But let's say you want to convert single ages into age categories. The age "3" for example
could go into the category "under 6", but also in "under 12", "under 18" and "total". Normally
you would compute additional variables, which hold the different categorizations, or you could also
double up the observations for each category. Both ways would just bloat up the data frame and
cost additional memory, particularly if you work with big data sets.

With these format containers, you just keep a small reference of original values and result
categories. Formats and data find their way together only just before computing the results,
meaning the original data frame can be passed into a function capable of handling formats (see below),
without any data transformation beforehand. You just tell the function which format should
be applied to which variable. That's it. The function handles the rest and outputs all the
desired categories.

This method is very memory efficient, readable and user friendly for creating larger and more
complex outputs at the same time.
}
\examples{
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

education. <- discrete_format(
    "Total"            = c("low", "middle", "high"),
    "low education"    = "low",
    "middle education" = "middle",
    "high education"   = "high")

income. <- interval_format(
    "Total"              =    0:100000,
    "below 500"          =    0:500,
    "500 to under 1000"  =  500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

state. <- discrete_format(
    "Germany"                       = 1:16,
    "Schleswig-Holstein"            = 1,
    "Hamburg"                       = 2,
    "Lower Saxony"                  = 3,
    "Bremen"                        = 4,
    "North Rhine-Westphalia"        = 5,
    "Hesse"                         = 6,
    "Rhineland-Palatinate"          = 7,
    "Baden-Württemberg"             = 8,
    "Bavaria"                       = 9,
    "Saarland"                      = 10,
    "West"                          = 1:10,
    "Berlin"                        = 11,
    "Brandenburg"                   = 12,
    "Mecklenburg-Western Pomerania" = 13,
    "Saxony"                        = 14,
    "Saxony-Anhalt"                 = 15,
    "Thuringia"                     = 16,
    "East"                          = 11:16)

# With discrete formats you can specify the keyword "other" to
# catch any other value not covered by the explicitly specified values.
age. <- discrete_format(
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = "other")

# With interval formats you can also use the keywords "low" and "high" to
# catch everything from the lowest to the highest values, in case one doesn't
# know exactly what the lowest and highest values are.
income. <- interval_format(
    "Total"              = c("low", "high"),
    "below 500"          = c("low", 500),
    "500 to under 1000"  = 500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = c(2000, "high"))

}
\seealso{
Functions that can handle formats: \code{\link[=summarise_plus]{summarise_plus()}}, \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}},
\code{\link[=any_table]{any_table()}}, \code{\link[=recode.]{recode.()}}, \code{\link[=recode_multi]{recode_multi()}}, \code{\link[=transpose_plus]{transpose_plus()}}, \code{\link[=sort_plus]{sort_plus()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code_statistics.R
\name{code_statistics}
\alias{code_statistics}
\title{Analyze R Scripts And Print Out Statistics}
\usage{
code_statistics(paths, recursive = FALSE, output_per = "overall")
}
\arguments{
\item{paths}{A full folder path in which there are 'R' scripts to be analyzed. Can
be a single path or a vector of paths.}

\item{recursive}{FALSE by default. If TRUE scans the provided paths recursive
for sub folders and the 'R' scripts within them.}

\item{output_per}{"overall" by default. Determines how detailed the results are
printed. Valid options are "overall", "folder" or "file".}
}
\value{
Returns the results as a data frame.
}
\description{
\code{\link[=code_statistics]{code_statistics()}} reads in a folder or entire folder structure, grabs all
'R' script files and scans the contents for different patterns. It then outputs
a small report which shows of which parts the code consists.
}
\examples{
# Run this function on a directory, actually containing R script files
# to see some results.
code_statistics(tempdir())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/frequencies.R
\name{frequencies}
\alias{frequencies}
\title{Display Frequency Tables of Single Variables}
\usage{
frequencies(
  data_frame,
  variables,
  formats = c(),
  by = c(),
  weight = NULL,
  means = FALSE,
  titles = .qol_options[["titles"]],
  footnotes = .qol_options[["footnotes"]],
  style = .qol_options[["excel_style"]],
  output = .qol_options[["output"]],
  na.rm = .qol_options[["na.rm"]],
  print_miss = .qol_options[["print_miss"]],
  print = .qol_options[["print"]],
  monitor = .qol_options[["monitor"]]
)
}
\arguments{
\item{data_frame}{A data frame in which are the variables to tabulate.}

\item{variables}{A vector of single variables to create frequency tables for.}

\item{formats}{A list in which is specified which formats should be applied to which variables.}

\item{by}{Compute tables stratified by the expressions of the provided variables.}

\item{weight}{Put in a weight variable to compute weighted results.}

\item{means}{FALSE by default. If TRUE prints a small summarising table which contains mean, sd,
min, max, total freq and missing values.}

\item{titles}{Specify one or more table titles.}

\item{footnotes}{Specify one or more table footnotes.}

\item{style}{A list of options can be passed to control the appearance of 'Excel' outputs.
Styles can be created with \code{\link[=excel_output_style]{excel_output_style()}}.}

\item{output}{The following output formats are available: console (default), text,
excel and excel_nostyle.}

\item{na.rm}{FALSE by default. If TRUE removes all NA values from the variables.}

\item{print_miss}{FALSE by default. If TRUE outputs all possible categories of the
grouping variables based on the provided formats, even if there are no observations
for a combination.}

\item{print}{TRUE by default. If TRUE prints the output, if FALSE doesn't print anything. Can be used
if one only wants to catch the output data frame.}

\item{monitor}{FALSE by default. If TRUE, outputs two charts to visualize the functions time consumption.}
}
\value{
Returns a list of two data tables containing the results for the frequency tables.
}
\description{
\code{\link[=frequencies]{frequencies()}} produces two kinds of tables for a quick overview of single variables.
The first table is for a broader overview and contains mean, sd, min, max, freq and missings.
The second table is the actual frequency table which shows the weighted sums, percentages
and unweighted frequencies per expression.
}
\details{
\code{\link[=frequencies]{frequencies()}} is based on the 'SAS' procedure Proc Freq, which provides
efficient and readable ways to output frequency tables.

To create a frequency table you only need to provide a single variable.
Nothing special about this. The real power comes into play, when you output your
tables as a fully styled 'Excel' workbook. Setting up a custom, reusable style is
as easy as setting up options like: provide a color for the table header, set the
font size for the row header, should borders be drawn for the table cells yes/no,
and so on.

You also can provide multiple single variables to generate multiple, evenly designed
tables, all at once. For just a quick overview, rather than fully designed
tables, you can also just output the tables in ASCII style format.
}
\examples{
# Example data frame
my_data <- dummy_data(1000)

# Define titles and footnotes.
set_titles("This is title number 1",
           "This is title number 2",
           "This is title number 3")

set_footnotes("This is footnote number 1",
              "This is footnote number 2",
              "This is footnote number 3")

# Output frequencies tables
my_data |> frequencies(sex)
my_data |> frequencies(c(age, education),
                       weight = weight)

# Also works with characters
my_data |> frequencies("sex")
my_data |> frequencies(c("age", "education"),
                       weight = "weight")

# Applying
sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

my_data |> frequencies(sex, formats = list(sex = sex.))

# Split frequencies by expressions of another variable
my_data |> frequencies(sex, by = education)

# Get a list with two data tables for further usage
result_list <- my_data |> frequencies(sex, formats = list(sex = sex.))

# Output in text file
my_data |> frequencies(sex, output = "text")

# Output to Excel
my_data |> frequencies(sex, output = "excel")

# If you want to add hyperlinksto titles and footnotes you can do so by
# adding "link:" followed by the hyperlink to the main text. Linking to another
# cell works with "cell:". To link to a file use "file:" an pass the full file
# path afterwards.
set_titles("This is title number 1",
           "This is title number 2 link: https://cran.r-project.org/",
           "This is title number 3 cell: W22",
           "This is title number 4 file: C:/MyFolder/MyFile.txt")

set_footnotes("This is footnote number 1",
              "This is footnote number 2 file: C:/MyFolder/MyFile.txt",
              "This is footnote number 3 cell: W22",
              "This is footnote number 4 link: https://cran.r-project.org/")

# Individual styling can also be passed directly
my_style <- excel_output_style(header_back_color = "0077B6",
                               font              = "Times New Roman")

my_data |> frequencies(sex, output = "excel", style = my_style)

# To save a table as xlsx file you have to set the path and filename in the
# style element
# Example files paths
table_file <- tempfile(fileext = ".xlsx")

# Note: Normally you would directly input the path ("C:/MyPath/") and name ("MyFile.xlsx").
#       With the set_style_options you can also set a table style globally.
set_style_options(save_path  = dirname(table_file),
                  file       = basename(table_file),
                  sheet_name = "MyTable")

my_data |> frequencies(sex, output = "excel")

# Manual cleanup for example
unlink(table_file)

# Global options are permanently active until the current R session is closed.
# There are also functions to reset the values manually.
reset_style_options()
reset_qol_options()
close_file()

}
\seealso{
Creating a custom table style: \code{\link[=excel_output_style]{excel_output_style()}}, \code{\link[=modify_output_style]{modify_output_style()}},
\code{\link[=number_format_style]{number_format_style()}}, \code{\link[=modify_number_formats]{modify_number_formats()}}.

Global style options: \code{\link[=set_style_options]{set_style_options()}}, \code{\link[=set_labels]{set_labels()}}.

Other global options: \code{\link[=set_titles]{set_titles()}}, \code{\link[=set_footnotes]{set_footnotes()}}, \code{\link[=set_print]{set_print()}}, \code{\link[=set_monitor]{set_monitor()}},
\code{\link[=set_na.rm]{set_na.rm()}}, \code{\link[=set_print]{set_print()}}, \code{\link[=set_print_miss]{set_print_miss()}}, \code{\link[=set_output]{set_output()}}.

Creating formats: \code{\link[=discrete_format]{discrete_format()}} and \code{\link[=interval_format]{interval_format()}}.

Functions that can handle formats and styles: \code{\link[=crosstabs]{crosstabs()}}, \code{\link[=any_table]{any_table()}}.

Additional functions that can handle styles: \code{\link[=export_with_style]{export_with_style()}}

Additional functions that can handle formats: \code{\link[=summarise_plus]{summarise_plus()}}, \code{\link[=recode.]{recode.()}},
\code{\link[=recode_multi]{recode_multi()}}, \code{\link[=transpose_plus]{transpose_plus()}}, \code{\link[=sort_plus]{sort_plus()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renaming.R
\name{replace_except}
\alias{replace_except}
\title{Replace Patterns While Protecting Exceptions}
\usage{
replace_except(vector, pattern, replacement, exceptions = NULL)
}
\arguments{
\item{vector}{A vector containing the texts, where a pattern should be replaced.}

\item{pattern}{The pattern that should be replaced.}

\item{replacement}{The new pattern, which replaces the old one.}

\item{exceptions}{A character vector containing exceptions, which should not be altered.}
}
\value{
Returns a vector with replaced pattern.
}
\description{
Replaces a provided pattern with another, while protecting exceptions. Exceptions can
contain the given pattern, but won't be changed during replacement.
}
\examples{
# Vector, where underscores should be replaced
underscores <- c("my_variable", "var_with_underscores", "var_sum", "var_pct_total")

# Extensions, where underscores shouldn't be replaced
extensions <- c("_sum", "_pct_group", "_pct_total", "_pct_value", "_pct", "_freq_g0",
                "_freq", "_mean", "_median", "_mode", "_min", "_max", "_first",
                "_last", "_p1", "_p2", "_p3", "_p4", "_p5", "_p6", "_p7", "_p8", "_p9",
                "sum_wgt", "_sd", "_variance", "_missing")

# Replace
new_vector <- underscores |> replace_except("_", ".", extensions)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transpose_plus.R
\name{transpose_plus}
\alias{transpose_plus}
\title{Fast And Powerful Yet Simple To Use Transpose}
\usage{
transpose_plus(
  data_frame,
  preserve = NULL,
  pivot = NULL,
  values = NULL,
  formats = c(),
  weight = NULL,
  na.rm = .qol_options[["na.rm"]],
  monitor = .qol_options[["monitor"]]
)
}
\arguments{
\item{data_frame}{A data frame to transpose}

\item{preserve}{Variables to keep and preserve in their current form.}

\item{pivot}{A vector that provides the expressions of single variables or
variable combinations that should be transposed. To nest variables use the form:
"var1 + var2 + var3 + ...".}

\item{values}{A vector containing all value variables that should be transposed.}

\item{formats}{A list in which is specified which formats should be applied to which
variables.}

\item{weight}{Put in a weight variable to compute weighted results.}

\item{na.rm}{FALSE by default. If TRUE removes all NA values from the preserve and
pivot variables.}

\item{monitor}{FALSE by default. If TRUE, outputs two charts to visualize the
functions time consumption.}
}
\value{
Returns a transposed data table.
}
\description{
\code{\link[=transpose_plus]{transpose_plus()}} is able to reshape a data frame from long to wide and from
wide to long. In the long to wide transposition variables can be nested or
placed side by side. With the wide to long transposition it is also possible
to transpose multiple variables at once.

Additionally \code{\link[=transpose_plus]{transpose_plus()}} is able to weight results before transposing
them from long to wide.

The function also makes use of formats, which means you don't need to create
variables storing the new variable names before transposition. You can just
use formats to name the new variables and with multilabels you can even generate
new variable expressions at the same time.
}
\details{
\code{\link[=transpose_plus]{transpose_plus()}} is just very loosely based on the 'SAS' procedure Proc Transpose,
and the possibilities of a Data-Step transposition using loops.

The transposition methods 'SAS' has to offer are actually fairly weak. Which is weird
because all tools are there to have another powerful function. So \code{\link[=transpose_plus]{transpose_plus()}}
tries to create the function 'SAS' should have.

The function is able to interpret which transposition direction the user wants by just
looking at what the user provided with the function parameters. For a long to wide
transposition it is natural to just provide variables to transpose. While it is also
just natural to provide new variable names when transposing from wide to long. That alone
reduces the number of parameters the user has to enter to perform a simple transposition.

The real magic happens when formats come into play. With their help you can not only name
new variables or their expressions, but you can also generate completely new expressions
with no effort, just with the help of multilabels.
}
\examples{
# Example formats
age. <- discrete_format(
    "Total"          = 0:100,
    "under 18"       = 0:17,
    "18 to under 25" = 18:24,
    "25 to under 55" = 25:54,
    "55 to under 65" = 55:64,
    "65 and older"   = 65:100)

sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

sex2. <- discrete_format(
    "Total"  = c("Male", "Female"),
    "Male"   = "Male",
    "Female" = "Female")

income. <- interval_format(
    "Total"              =    0:100000,
    "below 500"          =    0:500,
    "500 to under 1000"  =  500:1000,
    "1000 to under 2000" = 1000:2000,
    "2000 and more"      = 2000:100000)

# Example data frame
my_data <- dummy_data(1000)

# Transpose from long to wide and use a multilabel to generate additional categories
long_to_wide <- my_data |>
    transpose_plus(preserve = c(year, age),
                   pivot    = c("sex", "education"),
                   values   = income,
                   formats  = list(sex = sex., age = age.),
                   weight   = weight,
                   na.rm    = TRUE)

# Transpose back from wide to long and put results below each other. To trigger
# this behavior every list entry in pivot has to have a different name.
wide_to_long <- long_to_wide |>
    transpose_plus(preserve = c(year, age),
                   pivot    = list(sex       = c("Total", "Male", "Female"),
                                   education = c("low", "middle", "high")))

# Transpose from long to wide and use a multilabel to generate additional categories
long_to_wide <- my_data |>
    transpose_plus(preserve = c(year, age),
                   pivot    = "sex",
                   values   = c(income, weight),
                   formats  = list(sex = sex., age = age.),
                   weight   = weight,
                   na.rm    = TRUE) |>
    rename_multi("income_Total"  = "Total",
                 "income_Male"   = "Male",
                 "income_Female" = "Female")

# Transpose back from wide to long but this time put results side by side.
# To do that every list entry has to have the same name. The values parameter
# is then used to give the new value variables a name. For the expressions of
# the new categorical variable the variable names from the first pivot list
# entry are used.
wide_to_long <- long_to_wide |>
    transpose_plus(preserve = c(year, age),
                   values   = c(income, weight),
                   pivot    = list(sex = c("Total", "Male", "Female"),
                                   sex = c("weight_Total", "weight_Male", "weight_Female")))

# Nesting variables in long to wide transposition
nested <- my_data |>
    transpose_plus(preserve = c(year, age),
                   pivot    = "sex + education",
                   values   = income,
                   formats  = list(sex = sex., age = age.),
                   weight   = weight,
                   na.rm    = TRUE)

# Or both, nested and un-nested, at the same time
both <- my_data |>
    transpose_plus(preserve = c(year, age),
                   pivot    = c("sex + education", "sex", "education"),
                   values   = income,
                   formats  = list(sex = sex., age = age.),
                   weight   = weight,
                   na.rm    = TRUE)

}
\seealso{
Creating formats: \code{\link[=discrete_format]{discrete_format()}} and \code{\link[=interval_format]{interval_format()}}.

Functions that also make use of formats: \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}},
\code{\link[=any_table]{any_table()}}, \code{\link[=recode.]{recode.()}}, \code{\link[=recode_multi]{recode_multi()}}, \code{\link[=sort_plus]{sort_plus()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/small_helpers.R
\name{fuse_variables}
\alias{fuse_variables}
\title{Fuse Multiple Variables}
\usage{
fuse_variables(
  data_frame,
  new_variable_name,
  variables_to_fuse,
  drop_original_vars = TRUE
)
}
\arguments{
\item{data_frame}{A data frame with variables to fuse.}

\item{new_variable_name}{The name of the new fused variable.}

\item{variables_to_fuse}{A vector with the variables that should be fused together.}

\item{drop_original_vars}{Whether to drop or keep the original values. TRUE by default.}
}
\value{
Returns a data frame without the variables TYPE, TYPE_NR and DEPTH.
}
\description{
When you have a situation where you have multiple variables with different NA
values that happen to be in different places (where one variable has a value the
other is NA and vice versa) you can fuse these together to a single variable.
}
\examples{
# Example format
sex. <- discrete_format(
    "Total"  = 1:2,
    "Male"   = 1,
    "Female" = 2)

# Example data frame
my_data <- dummy_data(1000)

# Call function
all_possible <- my_data |>
    summarise_plus(class      = c(year, sex),
                   values     = c(income, probability),
                   statistics = c("sum", "mean", "freq"),
                   formats    = list(sex = "sex."),
                   weight     = weight,
                   nesting    = "all",
                   na.rm      = TRUE)

all_possible <- all_possible[DEPTH <= 1] |>
    fuse_variables("fusion", c("year", "sex"))

# NOTE: You can generally use this function to fuse variables. What is done in
#       multiple steps above can be achieved by just using nested = "single" in
#       summarise_plus.
single <- my_data |>
    summarise_plus(class      = c(year, sex),
                   values     = c(income, probability),
                   statistics = c("sum", "mean", "freq"),
                   formats    = list(sex = "sex."),
                   weight     = weight,
                   nesting    = "single",
                   na.rm      = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qol.R
\docType{package}
\name{qol-package}
\alias{qol}
\alias{qol-package}
\title{qol - Quality of Life}
\description{
This package brings some quality of life concepts and functions inspired
by 'SAS' to 'R'. The main goal is to make descriptive evaluations easier, so
one can create bigger and more complex outputs in less time with less code.
Introducing format containers with multilabels, a more powerful summarise, which
is capable to output every possible combination of the provided grouping variables
in one go, tabulation functions which can create any table in different styles
and other more readable functions.

In addition it offers an error handling which often catches errors and just
let's your code flow, even if there are small errors. You always get an
understandable message which helps you to get rid of the problem.

The package builds on the incredibly fast data.table and collapse packages for
maximum speed and on the wonderful openxlsx2 package for maximum style.
}
\section{Imports}{

data.table, collapse, openxlsx2, fst
}

\section{Minimal R Version}{

4.1.0 or higher
}

\section{Functions}{

Creating formats: \code{\link[=discrete_format]{discrete_format()}}, \code{\link[=interval_format]{interval_format()}}, \code{\link[=expand_formats]{expand_formats()}}

Summarisation and tabulation: \code{\link[=summarise_plus]{summarise_plus()}}, \code{\link[=frequencies]{frequencies()}}, \code{\link[=crosstabs]{crosstabs()}}, \code{\link[=any_table]{any_table()}}, \code{\link[=export_with_style]{export_with_style()}}, \code{\link[=combine_into_workbook]{combine_into_workbook()}}

Creating a custom table style: \code{\link[=excel_output_style]{excel_output_style()}}, \code{\link[=modify_output_style]{modify_output_style()}},
\code{\link[=number_format_style]{number_format_style()}}, \code{\link[=modify_number_formats]{modify_number_formats()}}

Recoding: \code{\link[=recode.]{recode.()}}, \code{\link[=recode_multi]{recode_multi()}}

Selecting: \code{\link[=keep]{keep()}}, \code{\link[=dropp]{dropp()}}, \code{\link[=inverse]{inverse()}}, \code{\link[=vars_between]{vars_between()}}

Joining: \code{\link[=multi_join]{multi_join()}}

Transposing: \code{\link[=transpose_plus]{transpose_plus()}}

Sorting: \code{\link[=sort_plus]{sort_plus()}}

If-statement: \code{\link[=if.]{if.()}}, \code{\link[=else_if.]{else_if.()}}, \code{\link[=else.]{else.()}}, \code{\link[=ifelse_multi]{ifelse_multi()}}, \code{\link[=where.]{where.()}}, \code{\link[=do_if]{do_if()}}, \code{\link[=else_do]{else_do()}}, \code{\link[=end_do]{end_do()}}, \code{\link[=end_all_do]{end_all_do()}}

Monitoring: \code{\link[=monitor_start]{monitor_start()}}, \code{\link[=monitor_end]{monitor_end()}}, \code{\link[=monitor_next]{monitor_next()}}, \code{\link[=monitor_plot]{monitor_plot()}}

Renaming: \code{\link[=rename_pattern]{rename_pattern()}}, \code{\link[=add_extension]{add_extension()}}, \code{\link[=remove_stat_extension]{remove_stat_extension()}}, \code{\link[=replace_except]{replace_except()}}

Retaining: \code{\link[=running_number]{running_number()}}, \code{\link[=mark_case]{mark_case()}}, \code{\link[=retain_value]{retain_value()}}, \code{\link[=retain_stat]{retain_stat()}}, \code{\link[=retain_variables]{retain_variables()}}

Character Manipulation: \code{\link[=concat]{concat()}}, \code{\link[=sub_string]{sub_string()}}, \code{\link[=remove_blanks]{remove_blanks()}}

Generate dummy data: \code{\link[=dummy_data]{dummy_data()}}

Conversion: \code{\link[=args_to_char]{args_to_char()}}, \code{\link[=dots_to_char]{dots_to_char()}}, \code{\link[=get_origin_as_char]{get_origin_as_char()}} \code{\link[=convert_numeric]{convert_numeric()}}, \code{\link[=convert_factor]{convert_factor()}},
\code{\link[=hex_to_256]{hex_to_256()}}, \code{\link[=hex_to_ansi]{hex_to_ansi()}}

Macros: \code{\link[=macro]{macro()}}, \code{\link[=apply_macro]{apply_macro()}}

Saving And Loading: \code{\link[=libname]{libname()}}, \code{\link[=stack_data]{stack_data()}}, \code{\link[=save_file]{save_file()}}, \code{\link[=save_file_multi]{save_file_multi()}}, \code{\link[=load_file]{load_file()}}, \code{\link[=load_file_multi]{load_file_multi()}}

Reporting: \code{\link[=content_report]{content_report()}}, \code{\link[=code_statistics]{code_statistics()}}, \code{\link[=report_test_results]{report_test_results()}}, \code{\link[=test_package]{test_package()}}, \code{\link[=test_single_file]{test_single_file()}}

Messages: \code{\link[=print_message]{print_message()}}, \code{\link[=print_headline]{print_headline()}}, \code{\link[=print_start_message]{print_start_message()}}, \code{\link[=print_closing]{print_closing()}}, \code{\link[=print_step]{print_step()}},
\code{\link[=set_up_custom_message]{set_up_custom_message()}}, \code{\link[=get_message_stack]{get_message_stack()}}, \code{\link[=set_no_print]{set_no_print()}}, \code{\link[=print_stack_as_messages]{print_stack_as_messages()}}, \code{\link[=convert_square_brackets]{convert_square_brackets()}}

Calculation: \code{\link[=compute.]{compute.()}}, \code{\link[=row_calculation]{row_calculation()}}

Small helpers: \code{\link[=setcolorder_by_pattern]{setcolorder_by_pattern()}}, \code{\link[=drop_type_vars]{drop_type_vars()}}, \code{\link[=fuse_variables]{fuse_variables()}},
\code{\link[=get_excel_range]{get_excel_range()}}, \code{\link[=get_integer_length]{get_integer_length()}}, \code{\link[=round_values]{round_values()}}

Split data frame: \link{split_by}

Error handling: \code{\link[=resolve_intersection]{resolve_intersection()}}, \code{\link[=part_of_df]{part_of_df()}}, \code{\link[=remove_doubled_values]{remove_doubled_values()}}, \code{\link[=check_weight]{check_weight()}}

Global options: \code{\link[=set_style_options]{set_style_options()}}, \code{\link[=get_style_options]{get_style_options()}}, \code{\link[=reset_style_options]{reset_style_options()}}, \code{\link[=close_file]{close_file()}},
\code{\link[=set_labels]{set_labels()}}, \code{\link[=get_labels]{get_labels()}}, \code{\link[=set_print]{set_print()}}, \code{\link[=get_print]{get_print()}}, \code{\link[=set_monitor]{set_monitor()}},
\code{\link[=get_monitor]{get_monitor()}}, \code{\link[=set_na.rm]{set_na.rm()}}, \code{\link[=get_na.rm]{get_na.rm()}}, \code{\link[=set_output]{set_output()}}, \code{\link[=get_output]{get_output()}}, \code{\link[=set_titles]{set_titles()}},
\code{\link[=get_titles]{get_titles()}}, \code{\link[=set_footnotes]{set_footnotes()}}, \code{\link[=get_footnotes]{get_footnotes()}}, \code{\link[=set_print_miss]{set_print_miss()}}, \code{\link[=get_print_miss]{get_print_miss()}},
\code{\link[=set_threads]{set_threads()}}, \code{\link[=get_threads]{get_threads()}}

Other: \code{\link[=build_master]{build_master()}}, \code{\link[=build_rstheme]{build_rstheme()}}, \code{\link[=free_memory]{free_memory()}}, \code{\link[=qol_news]{qol_news()}}, \code{\link[=qol_chat]{qol_chat()}}
}

\section{Snippets}{

\preformatted{
snippet splus
    summarise_plus(class      = c(var1, var2, ...),
                   values     = c(var1, var2, ...),
                   statistics = c("pct_group", "sum", "sum_wgt", "freq"),
                   weight     = weight_var,
                   formats    = list(var = format., ...),
                   nesting    = "deepest")

snippet if.
         if.(condition, var = value) |>
    else_if.(condition, var = value) |>
    else.   (           var = value)

snippet freq
    frequencies(variables = c(var1, var2, ...),
                weight    = weight_var,
                formats   = list(var = "format.", ...),
                titles    = c(),
                footnotes = c())

snippet cross
    crosstabs(rows       = row_var,
              columns    = col_var,
              statistics = c("sum", "pct_row", "pct_column", "pct_total", "freq"),
              weight    = weight_var,
              formats   = list(var = format., ...),
              titles    = c(),
              footnotes = c())

snippet any
    any_table(rows        = c("var1 + var2 + ...", "var1"),
              columns     = c("var3", "var3 + var4 + ..."),
              values      = c("value_var1", "value_var2"),
              statistics  = c("sum", "pct_group", "pct_value", "freq"),
              pct_group   = c("var1", "var2"),
              pct_value   = list(new_var = "numerator / denominator"),
              weight    = weight_var,
              formats     = list(var = format., ...),
              titles      = c(),
              footnotes   = c(),
              var_labels  = list("var1" = "My label", ...),
              stat_labels = list("pct" = "%"),
              box         = "")
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/s3rdia/qol}
  \item \url{https://s3rdia.github.io/qol/}
  \item \url{https://deepwiki.com/s3rdia/qol}
  \item \url{https://s3rdia.github.io/qol_blog/}
}

}
\author{
Tim Siebenmorgen
}
\keyword{internal}
